package applications.trakla2.server;

import java.io.*;
import java.rmi.*;
import java.rmi.registry.*;
import java.rmi.server.*;
import java.util.*;

import applications.trakla2.datalogging.ExerciseData;



public class ExerciseServerImpl extends UnicastRemoteObject implements ExerciseServer {

    private static final long serialVersionUID = -4652300639427762388L;

    /**
     * The server prints an message "Still running" into the debug-log
     * every DEBUG_MSG_MINUTES minutes.
     */
    public static final int DEBUG_MSG_MINUTES = 5;

    static String dataDir = "";
    static PrintStream log = null;
    static PrintStream errorLog = null;
    static PrintStream debugLog = null;

    SubmissionHandler submit;
    ObjectSerializer objectSerializer;


    public ExerciseServerImpl(String givenDataDir) throws RemoteException, IOException {
        //Create and export a new UnicastRemoteObject object
        //using the particular supplied port.
        super(1099);
        dataDir = givenDataDir;
        submit = createSubmissionHandler();
        objectSerializer = new ObjectSerializer(this);
    }


    public ExerciseSession newSession() throws RemoteException {
        return new ExerciseSessionImpl(this);
    }


    protected String save(ExerciseData data) {
        return objectSerializer.save(data);
    }




    public static void main(String[] args) throws Exception {

        if (args.length < 1) {
            System.err.println("Usage: java trakla2.server.ExerciseServerImpl <properties file>");
            System.exit(1);
        }

        // load properties
        BufferedInputStream in = new BufferedInputStream(new FileInputStream(args[0]));
        System.getProperties().load(in);
        in.close();

        // set log file
        String dataDir = System.getProperty("trakla2.server.DATADIR");
	// check that dataDir exists and create it if it doesn't
	File f = new File(dataDir + "/logs/");
	if (!f.exists())
	    f.mkdirs();
        log = openPrintStream(dataDir + "/logs/trakla2.log");
        errorLog = openPrintStream(dataDir + "/logs/trakla2-error.log");
        debugLog = openPrintStream(dataDir + "/logs/trakla2-debug.log");


        // create server object

        ExerciseServerImpl server = null;
        if (dataDir == null) {
            server = new ExerciseServerImpl(".");
        } else {
            server = new ExerciseServerImpl(dataDir);
        }


        // Create fixed port socket factory to pass firewall.

        try {
            RMISocketFactory.setSocketFactory(new FixedPortRMISocketFactory());
        } catch (IOException e) {
            e.printStackTrace(errorLog);
            return;
	}


        int port = getPort();

        // Create registry that runs in an application and bind the
        // specified name to the server object.
        Registry r = null;
        try {
            r = LocateRegistry.createRegistry(port);
        } catch (RemoteException re) {
            error("Registry could not be exported in port "
                  + port + " !");
            return;
        }
        try {
            r.bind(ExerciseServer.SERVICE_NAME, server);
            debug("Service '" + ExerciseServer.SERVICE_NAME +
                  "' succesfully registered in registry (port " +
                  port + ").");
        } catch (AlreadyBoundException abe) {
            error("Service '" + ExerciseServer.SERVICE_NAME +
                  "' is already bound (port " + port + ").");
            return;
        } catch (Exception e) {
            error("Failed to bind service object");
            error("Exception: " + e);
            return;
        }


        // create thread to print log message every hour
        Thread t = new Thread() {
                public void run() {
                    while (true) {
                        try {
                            sleep(DEBUG_MSG_MINUTES*60*1000);
                        } catch (InterruptedException e) {}
                        debug("Still running...");
                    }
                }
            };
        t.start();
        debug("Server running.");
    }




    static synchronized void log(String msg)  {
        log.println(msg);
        log.flush();
    }


    static synchronized void error(String msg)  {
        errorLog.println(new Date().toString() + ": " + msg);
        errorLog.flush();
    }


    static synchronized void debug(String msg)  {
        debugLog.println(new Date().toString() + ": " + msg);
        debugLog.flush();
    }


    private static PrintStream openPrintStream(String fileName)  throws IOException {
        File f = new File(fileName);
        if (!f.exists())
            f.createNewFile();

        // append file
        return new PrintStream(new FileOutputStream(f.getPath(),true));
    }



    private static int getPort() throws NumberFormatException {
        String port = System.getProperty("trakla2.server.PORT");
        if (port != null) {
            return Integer.valueOf(port).intValue();
        } else {
            return -1;
        }
    }


    protected long nextSeed() {
        return new Date().getTime();
    }

    protected static SubmissionHandler createSubmissionHandler() {
	String h = System.getProperty("trakla2.submissionhandler");
	if (h == null)
	    return null;
	SubmissionHandler handler;
	try {
	    Class submitClass = Class.forName(h);
	    handler = (SubmissionHandler)submitClass.newInstance();
	} catch (Exception e) {
	    error("Failed to create SubmissionHandler: " + e.getMessage());
	    return null;
	}
	return handler;
    }
}
