package applications.trakla2.util;

import java.awt.Component;
import java.awt.FileDialog;
import java.awt.GridBagConstraints;
import java.awt.MenuItem;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.*;

import java.io.File;
import java.io.FileInputStream;
import java.io.ObjectInputStream;


import content.interfaces.AWTComponentUtilizer;
import content.interfaces.ConfigureVisualType;
import content.interfaces.HasRotatedContainers;
import content.interfaces.LayoutExercise;
import content.interfaces.SimulationExercise;
import content.interfaces.StyledExercise;

import prototype.ui.StartFrame;

import matrix.animation.Animator;
import matrix.simulation.VisualTypeConf;
import matrix.structures.FDT.FDT;
import matrix.uitools.StructurePanel;
import matrix.util.Note;
import matrix.visual.VisualContainer;
import matrix.visual.VisualType;
import applications.trakla2.datalogging.ExerciseData;
import applications.trakla2.datalogging.GradeData;
import applications.trakla2.datalogging.GradeToSubmissionData;
import applications.trakla2.datalogging.SubmissionData;

/**
 * The SubmissionViewer class encloses all the required methods to view submission data
 * collected by the datalogging package. These methods were originally in the StartFrame class.
 *
 * @author Otto Seppl
 */
public class SubmissionViewer {

    /**
     * This method can be used to add a menu for viewing TRAKLA2 submissions.
     * The mthod is typically called by the main method in the StartFrame class.
     *
     * @param parent Parent frame for the FileDialog
     * @return the new menuitem to be added to the menu
     */

	public static MenuItem addTrakla2FileMenu(final Frame parent) {
		MenuItem mi = new MenuItem("Open Trakla2 answer...");
		mi.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				openTrakla2Answer(parent);
			}});
		return mi;
	}

    private static Frame getNewFrame(StructurePanel p) {
        return new StartFrame(p);
    }

    /**
     * This method can be used to view TRAKLA2 submissions.
     * it opens up a new file dialog for choosing the desired submission and
     * then opens a window for viewing the submission itself.
     *
     * @param parent Parent frame for new windows
     */
	public static void openTrakla2Answer(Frame parent) {
		FileDialog fd = new FileDialog(parent,
				"Select file",
				FileDialog.LOAD);
		fd.setVisible(true);
		String fileName = fd.getDirectory() + fd.getFile();
		if (fileName == null)
			return;

		try {
			FileInputStream istream = new FileInputStream(new File(fileName));
			ObjectInputStream p = new ObjectInputStream(istream);
			ExerciseData data = (ExerciseData)p.readObject();
			System.out.println("Loaded: " + data);
			istream.close();

            openTrakla2Answer(parent, data);
		}
		catch (Exception e) {
			e.printStackTrace();
			Note.err(parent, e.getMessage());
			Note.show(parent, e.getMessage());

		}
	}

    /**
     * This method can be used to view TRAKLA2 submissions.
     * It opens a window for viewing the submission data given as a parameter.
     *
     * If there was an exercise description it will also be shown.
     *
     * @param parent Parent frame for new windows
     * @param data The exercise data for the submission to be viewed
     */

    public static void openTrakla2Answer(Frame parent, ExerciseData data){
    	Frame newFrame = null;

        if (data instanceof SubmissionData || data instanceof GradeData){
            newFrame = getNewFrame(reinitializeStructurePanel(data));        
        }else{
            Note.show(data, "The object does not contain any kind of submission data.");
            return;
        }

        //This code shows the exercise description in a Dialog window
        //so that the answer can viewed in proper context
        try{
            content.interfaces.SimulationExerciseModel scm = (content.interfaces.SimulationExerciseModel)(Class.forName(data.getExerciseClassName()).newInstance());
            scm.setSeed(data.getSeed());
            scm.init();
            String description = scm.getDescription();
            if (description.trim().length() != 0){

                int maxWidth   = 0;
                int height     = 1;
                int cursor     = 0;
                int nextCursor = description.indexOf("\n", cursor);

                while(nextCursor != -1){
                	height++;
                    maxWidth = Math.max(maxWidth, nextCursor-cursor);
                    cursor = nextCursor + 1;
                    nextCursor = description.indexOf("\n", cursor);
                }

                maxWidth = Math.max(maxWidth, description.length()-cursor);


                TextArea ta = new java.awt.TextArea(description,height+1,maxWidth+1);

                Dialog d = new Dialog(newFrame, "Exercise description", false);
                d.add(ta);
                d.pack();
                d.setVisible(true);
            }
          }catch(Exception e){}

    }

    public static StructurePanel reinitializeStructurePanel(ExerciseData data){        
        StructurePanel sp;
        SubmissionData sdata;
        
        if (data instanceof SubmissionData) {
            sdata = (SubmissionData)data;
        }else if (data instanceof GradeData){
            sdata = new GradeToSubmissionData((GradeData)data);
        }else throw new RuntimeException("The data is not valid submission data.");
            
        
        FDT[] ss   = sdata.getStudentSolution().getStructures();
        Animator a = sdata.getStudentSolution().getAnimator();
        
        sp = new StructurePanel(a);
        
        Object exercise = null;
        try {
            exercise = Class.forName(sdata.getExerciseClassName()).newInstance();
        } catch (InstantiationException e) {
            e.printStackTrace();
        } catch (IllegalAccessException e) {
            e.printStackTrace();
        } catch (ClassNotFoundException e) {
            e.printStackTrace();
        }
        
        VisualType[] vt;
        
        if (exercise instanceof LayoutExercise && exercise instanceof StyledExercise){
            vt = sp.addStructures(
                    ss, 
                    ((StyledExercise)exercise).getStructureVisualisations(),
                    ((LayoutExercise)exercise).getConstraints());
        }else if (exercise instanceof LayoutExercise){
            vt = sp.addStructures(
                    ss, 
                    ((LayoutExercise)exercise).getConstraints());
        }else if (exercise instanceof StyledExercise){
            vt = sp.addStructures(
                    ss, 
                    ((StyledExercise)exercise).getStructureVisualisations());
        }else{
            vt = sp.addStructures(ss);
        }

        if (exercise instanceof SimulationExercise){
            String[] names = ((SimulationExercise)exercise).getStructureNames();
            for (int i = 0; i < names.length; i++) {
                vt[i].setName(names[i]);
            }
        }
        
        // rotates containers when needed
        if(exercise instanceof HasRotatedContainers){
            boolean[] rotateInfo = ((HasRotatedContainers)exercise).getContainerRotations();
            for (int i = 0; i < vt.length; i++) {
                if (vt[i] instanceof VisualContainer)
                    ((VisualContainer)vt[i]).setRotated(rotateInfo[i]);
            }     
        }

        if (exercise instanceof ConfigureVisualType) {
            VisualTypeConf[] conf = ((ConfigureVisualType)exercise).conf();
            for (int i = 0; i < vt.length; i++) {
                vt[i].configure(conf[i]);
            }
        }
        
        
        // Some work required here (or copy from exercisePanel)
        /*
        if (exercise instanceof ButtonExercise) {
            JButton[] buttons = ExercisePanel.makeButtons((ButtonExercise)exercise);
            ExercisePanel.layoutButtons(exercise, buttons);
        }
        */
        if (exercise instanceof AWTComponentUtilizer) {
            // The exercise contains additional components -> add them.
            Component[] components = ((AWTComponentUtilizer)exercise).getAWTComponents();
            GridBagConstraints[] gbc = ((AWTComponentUtilizer)exercise).getAWTComponentConstraints();
            if (gbc == null) {
                for (int i = 0; i < components.length; i++) {
                    sp.add(components[i]);
                }
            } else {
                for (int i = 0; i < components.length; i++) {
                    sp.add(components[i], gbc[i]);
                }
            }
            // Give the application reference to the exercise object.
            // Might, e.g., be needed for validating the application. 
            ((AWTComponentUtilizer)exercise).setApplication(sp);
        }

        return sp;
    }
    
    
}
