/*
 * Created on May 10, 2005
 *
 */
package content.assessment;

import matrix.animation.Animator;
import matrix.structures.FDT.FDT;
import content.interfaces.JudgeBlocks;
import content.interfaces.MultipleQuestions;
import content.interfaces.SimulationExerciseModel;
import content.interfaces.SimulationExerciseSelfAssessment;


public abstract class AssessmentStrategy {

    /** 
     * For each FDT in the user solution of a <code>SimulationExercise</code> the number of 
     * states corresponding to correct states in the model answer is counted.
     * The number of correct states for each FDT is returned in the instance of <code>Judgemet</code>
     * returned by this method. The current state of the user animator a is the last
     * state judge will evaluate.
     *
     * @param user an array of the user's FDTs
     * @param mod the corresponding model answer for this exercise
     * @param a the animator of the user exercise
     */

    
    public Judgement judge(FDT[] user, Animator a, SimulationExerciseModel exercise){
        
        /* There are two types of exercises available :
         *  - Exercises assessed by the exercise class itself      : (SimulationExerciseSelfAssessment)
         *  - Exercises assessed by comparing simulation sequences : the rest
         * 
         * One exercise can also consist of several exercises.
         * In this case it is a MultipleQuestions exercise.
         * 
         * If the exercise is neither of these special cases it is assessed using
         * the judgeNormal method.
         *           
         * This method dispatches forward the judging procedure to the correct method.
         */
                
        if (exercise instanceof MultipleQuestions)
            return judge(a, (MultipleQuestions) exercise);        
        else if (exercise instanceof SimulationExerciseSelfAssessment)
            return judge(user, a, (SimulationExerciseSelfAssessment) exercise);
        else
            return judgeNormal(user, a, exercise);
    }

    /**
     * In some questions the FDT:s are graded together. The change in structure A
     * must happen together with structure B etc. Such functionality is requested
     * with the JudgeBlocks interface
     * 
     * If however it is not of any importance how the changes in different structures
     * happen related to other structures we judge them separately.
     *
     * @param user an array of the user's FDTs
     * @param mod the corresponding model answer for this exercise
     * @param a the animator of the user exercise
     */

    private Judgement judgeNormal(FDT[] user, Animator a, SimulationExerciseModel exercise){            

        /*
         JUAN - UNCOMMENT THIS PART!!! 
         
        if (exercise instanceof AlgorithmVariants){
            
            AnimatedStructures[] variants = ((AlgorithmVariants)exercise).createVariantSequences();
            String[] names = ((AlgorithmVariants)exercise).getVariantNames();
                       
            for (int i=0; i<variants.length; i++){
                Judgement jResult = 
                    Jury.getInstance().getJudgementSeparateFDTs(
                        user,
                        a,                        
                        variants[i].getStructures(), 
                        variants[i].getAnimator());
                    
                
                
                Note.out(null, names[i]+" :\t"+jResult.getPoints()[0]+"\n");
            }
            
        }
        */
        
        if (exercise instanceof JudgeBlocks)
            return getJudgementJoinedFDTs(user, a, exercise);
        else
            return getJudgementSeparateFDTs(user, a, exercise);        
    }    
    
    /**
     * This method creates the Judgement for exercises that grade themselves
     * 
     * @param user      The structures used by the user
     * @param a         The animator used by the user
     * @param exercise  The exercise being solved
     * @return          The result of the judging
     */
    
    private Judgement judge(FDT[] user, Animator a, SimulationExerciseSelfAssessment exercise) {
        //Otto removed the synchronized keyword 15.3.2005
        
        // Init
        String[] names = exercise.getStructureNames();
        int[] points   = exercise.judge();

        // Create judgement with length of points-array
        if (names == null || points == null)
            return new Judgement(0);
        
        int maxPoints = exercise.getMaxPoints();        
        Judgement J = new Judgement(names, points, maxPoints, exercise.getFeedbackString());

        // Return the created judgement
        if (J.getPointSum() == maxPoints)
            J.setComplete();
        return J;
    }

    /**
     * This method creates the Judgement for exercises that consist of subexercises
     * 
     * @param user      The structures used by the user
     * @param a         The animator used by the user
     * @param exercise  The exercise being solved
     * @return          The result of the judging
     */

    private Judgement judge(Animator a, MultipleQuestions exercise) {
    	FDT[] user = exercise.getAnswer();
        
        Judgement j= judgeNormal(user, a, (SimulationExerciseModel)exercise);
        for (int i=1; i<exercise.numberOfQuestions(); i++) {            
            exercise.changeAnimator(i);
            user = exercise.getAnswer();
            j.join(judgeNormal(user, a, (SimulationExerciseModel)exercise));
        }
        exercise.changeAnimator(0);
        return j;
    }

    /**
     * This method creates the judgement for a simulation exercise when we want
     * to grade the different user structures as a group
     * 
     * @param user          The structures used by the user
     * @param userAnimator  The Animator used by the user
     * @param exercise      The exercise being solved
     * @return              The result of the judging
     */
    
    protected abstract Judgement getJudgementJoinedFDTs(FDT[] user, Animator userAnimator, SimulationExerciseModel exercise);
    
    /**
     * This method creates the judgement for a simulation exercise when we want
     * to grade the different user structures separately of each other
     * 
     * @param user          The structures used by the user
     * @param userAnimator  The Animator used by the user
     * @param exercise      The exercise being solved
     * @return              The result of the judging
     */
    
    protected abstract Judgement getJudgementSeparateFDTs(FDT[] user, Animator a, SimulationExerciseModel exercise);

    /**
     * This method winds the animators to the end and compares the structures to find
     * whether the final states match
     *  
     * @param userAnimator      The Animator used by the user
     * @param solverAnimator    The Animator used by the solving method
     * @param userSolution      User's structures
     * @param modelSolution     Solving methods structures
     * @return                  true if the final states match, false otherwise
     */
    
    public static boolean isFinished(Animator userAnimator, Animator solverAnimator, FDT[] userSolution, FDT[] modelSolution) {
        userAnimator.end();
        solverAnimator.end();
        return allEqual(userSolution, modelSolution);
    }

    /**
     * This method compares two arrays of structures to find
     * whether the they match
     *  
     * @param a     first array of structures
     * @param b     second array of structures
     * @return      true if the states match, false otherwise
     */    
    
    public static boolean allEqual(FDT[] a, FDT[] b){
        for (int i=0; i<a.length; i++)
            if (a[i].equals(b[i]) == false)
                return false;
        return true;
    }

}
