/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package content.assessment;

import java.util.Vector;
import content.interfaces.Exercise;

/**
 * This class contains a criticism for an exercise, and is created
 * and returned by the judge() -method in Exercise interface.
 * @see Exercise
 */
public class DeprecatedJudgement{

    /** Points got from the exercise */
    private int[] points;
    private int maxPoints;
    private String[] names;
    private Vector notes = new Vector();
    private boolean complete = false;
    private int errorStateIndex = -1;
    public DeprecatedJudgement next = null;

    public boolean equals(Judgement j){
        int[] otherPoints = j.getPoints();
        int   otherMaxpoints = j.getMax();
        
        if (otherMaxpoints != maxPoints)
            return false;
        
    	for (int i=0; i<points.length; i++){
    		if (points[i] != otherPoints[i])
                return false;
        }
    	
        //Not comparing notes or names
        
        if (j.isComplete() != complete)
        	return false;
    
        if (j.getErrorStateIndex() != errorStateIndex)
            return false;
        
        if (next != null){
        	if (j.next == null)
                return false;
            
            return next.equals(j.next);
        }
        
        return true;           
    }
    
    /**
     * Constructs a new Judgement
     */
    public DeprecatedJudgement(int i) {
	points = new int[i];
	names = new String[i];
    }
  
    /**
     * Returns the points got from the exercise judged.
     * @return the points
     */
    public int[] getPoints() {
	return points;
    }
  
    /**
     * Sets the points got from the exercise judged.
     * @param i the points
     */
    public void setPoints(int i, int pos) {
	if(pos > -1 && pos < points.length)
	    points[pos] = i;
    }
  
    public void addPoints(int[] newPoints) {
	int[] oldPoints = points;
	points = new int[oldPoints.length + newPoints.length];
	int i;
	for (i=0; i<oldPoints.length; i++)
	    points[i] = oldPoints[i];
	for (int j=0; j<newPoints.length; j++)
	    points[i+j] = newPoints[j];
    }

    public void setMax(int max) {
	maxPoints = max;
    }

    public int getMax() { return maxPoints; }

    public void setNames(String[] names) { this.names = names; }

    public String[] getNames() { return this.names; }
  
    public void setComplete() {
	complete = !complete;
    }
  
    public boolean isComplete() {
	return complete;
    }


    public void setErrorStateIndex(int value) {
	errorStateIndex = value;
    }


    public int getErrorStateIndex() {
	return errorStateIndex;
    }

    public boolean errorStateIndexSet() {
	return errorStateIndex >= 0;
    }

  
    public void addNote(Object note) {
	notes.addElement(note);
    } 

    /** Append new judgement at the end of the list in case of Multiple Questions. */
    public void join(DeprecatedJudgement j2) {
	DeprecatedJudgement last = this;
	while (last.next != null)
	    last = last.next;
	last.next = j2;
    }
}
