/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: Judgement.java,v 1.4 2007/09/05 16:55:38 oseppala Exp $ */

package content.assessment;

import java.util.Vector;

import content.interfaces.Exercise;

/**
 * This class contains a criticism for an exercise, and is created
 * and returned by the judge() -method in Exercise interface.
 * @see Exercise
 */
public class Judgement {
	
	/** Points got from the exercise */
	private int[] points;
	private int maxPoints;
    private String feedbackString;
	private String[] names;
	private Vector notes        = new Vector();
	private boolean complete    = false;

	// Marks first MODEL state not reached
	private int errorStateIndex = -1;
    private int pointSum        = 0;
    
    // Marks last USER state that matches a MODEL state
    private int finalUserState = 0;
    
    // Marks that there were model states not reached or extra
    // user states after the final state of the model algorithm
    private boolean errorStates = false; 

	// If the user sequence contains only the beginning state
	// of the model sequence and no has no user states after
	// that the answer is considered EMPTY  
    private boolean empty = false;
    
    public void setFinalUserState(int value){
        finalUserState = value;
    }
    
    public int getFinalUserState(){
        return finalUserState;
    }
    
    public void setErrorStates(){
    	errorStates = true;
    }
    
    public boolean hasErrorStates(){
    	return errorStates;
    }
    
    public void setEmpty(){
    	empty = true;
    }
    
    public boolean wasEmpty(){
        return empty;
    }
    
    public String getFeedbackString() {
        return feedbackString;
    }
    
	public Judgement next = null;
	
	/**
	 * Constructs a new Judgement
	 */
	public Judgement(int i) {
		points = new int[i];
		names = new String[i];
	}
	
    public Judgement(String[] names, int[] points, int maxPoints, String feedbackString){
        this(names, points, maxPoints);
        this.feedbackString = feedbackString;
    }

    public Judgement(String[] names, int[] points, int maxPoints){
		this.names  = (String[])names.clone();
        this.points = (int[])points.clone();
        updatePointSum();
        
        this.maxPoints = maxPoints;
	}

    public Judgement(String[] names, int pointSize, int maxPoints){
        this(names, new int[pointSize], maxPoints);
    }

    
    private int updatePointSum(){
        pointSum = 0;
        for (int i=0; i < points.length;i++) {
            pointSum+=points[i];
        }
        return pointSum;
    }
    
    public int getPointSum(){
        return pointSum;
    }
    
    
	/**
	 * Returns the points got from the exercise judged.
	 * @return the points
	 */
	public int[] getPoints() {
		return points;
	}
	
	/**
	 * Sets the points got from the exercise judged.
	 * @param i the points
	 */
	public void setPoints(int i, int pos) {
		if(pos > -1 && pos < points.length)
			points[pos] = i;
        
        updatePointSum();
	}
	
	public void addPoints(int[] newPoints) {
		int[] oldPoints = points;
		points = new int[oldPoints.length + newPoints.length];
		int i;
		for (i=0; i<oldPoints.length; i++)    
			points[i] = oldPoints[i];
		for (int j=0; j<newPoints.length; j++)
			points[i+j] = newPoints[j];
        
        updatePointSum();
	}
	
	public void setMax(int max) {
		maxPoints = max;
	}
	
	public int getMax() { return maxPoints; }
	
    // TODO: This information is never used anywhere - It could be factored out, but would change the Judgement
    // class public interface. Wait for the right moment
	public String[] getNames() { return this.names; }
	
	public void setComplete() {
		complete = !complete;
	}
	
	public boolean isComplete() {
		return complete;
	}
	
	
	public void setErrorStateIndex(int value) {
		errorStateIndex = value;
	}
	
	
	public int getErrorStateIndex() {
		return errorStateIndex;
	}
	
	public boolean errorStateIndexSet() {
		return errorStateIndex >= 0;
	}
	
	
	public void addNote(Object note) {
		notes.addElement(note);
	} 
	
	/** Append new judgement at the end of the list in case of Multiple Questions. */
	public void join(Judgement j2) {
		Judgement last = this;
		while (last.next != null)
			last = last.next;
		last.next = j2;
	}
    
    public String toString(){
        if (next==null)
            return "[Judgement : "+pointSum+"]";
    	return "[Judgement : "+pointSum+"]->" + next.toString();
    }
}
