/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package content.exercises;

import content.interfaces.*;
import content.exercises.structures.*;
import matrix.animation.Animator;
import matrix.simulation.VisualTypeConf;
import matrix.structures.FDT.FDT;
import matrix.structures.FDT.probe.*;
import matrix.util.RandomKey;
import java.util.Random;

/**
 * CommonTrie exercise where specified strings in the input array should
 * be inserted to common trie. Insertion is done by expanding the correct subtrees
 * in the tree and changing the false bit to true when appropriate. Expanding
 * of the subtrees and changing of the false/true bit is done by clicking the keys
 * of the corresponding nodes.
 * 
 * Trie structure in the exercise is configured to be rotated through HasRotatedContainers
 * interface and configured to be flipped in y direction through VisualTypeConf.
 */
public class CommonTrie_Insert implements SimulationExerciseModel,
StyledExercise, ModelAnswerNames, ConfigureVisualType, JudgeBlocks, HasRotatedContainers {
    
    // ATTRIBUTES

    private Table inputTable;
    private ClickableCommonTrie commonTrie;
    
    private long seed = 1;
    
    String[] keys;
    
    // METHODS
    
    public FDT[] solve() {
        Table table = (Table)getInitialStructures()[0];
        ClickableCommonTrie trie = (ClickableCommonTrie)getInitialStructures()[1];
        
        String tmpString = "";
        ClickableCommonTrie tempTrie;
        char character;
        
        for (int i = 0; i < table.size(); i++) {
            tmpString = ((Key)table.getObject(i)).getComparisonString().substring(1);//removing first character (")
            tempTrie = trie;
            
            while (!tmpString.equals("\"")) {
                character = tmpString.charAt(0);
                
                Animator.getActiveAnimator().startOperation();
                tempTrie.setSubTreeChar(character);
                Animator.getActiveAnimator().endOperation();
                
                tempTrie = (ClickableCommonTrie)tempTrie.getSubTree(character);
                
                tmpString = tmpString.substring(1);
            }
            
            Animator.getActiveAnimator().startOperation();
            if (((Key)tempTrie.getElement()).getComparisonString().equals("false")) {
                tempTrie.toggleBoolean();
            }
            Animator.getActiveAnimator().endOperation();
            
        }
        
        return new FDT[] { table, trie };
    }
    
    
    public FDT[] makeModelAnswer() {

        return solve();
        
    }
    
    public String[] getStructureVisualisations() {
        
        return new String[] { "array", "layered tree" };
        
    }
    
    public String[] getModelAnswerVisualisations() {
        
        return new String[] { "array", "layered tree" };
        
    }
    
    public String[] getModelAnswerNames() {
        
        return new String[] { "Input strings", "Common Trie" };
        
    }
    
    public VisualTypeConf[] conf() {
        
        VisualTypeConf table = new VisualTypeConf();
        
        VisualTypeConf trie = new VisualTypeConf();
        trie.enable("matrix.visual.VisualKey", VisualTypeConf.HIGHLIGHT_OPERATION);
        trie.setVisualConfValue("matrix.visual.VisualLayeredTree", VisualTypeConf.FLIPPEDY, "true");
        
        return new VisualTypeConf[] { table, trie };
        
    }
    
    public long getSeed() {
        
        return seed;
    }
    
    public String[] getStructureNames() {
        
        return new String[] { "Input strings", "Common Trie" };
        
    }
    
    public FDT[] getAnswer() {
        
        return new FDT[] { inputTable, commonTrie };
    }
    
    public FDT[] getInitialStructures() {
                
        return new FDT[] { new Table(keys), new ClickableCommonTrie() };
        
    }
    
    public FDT[] init() {
        Random random = new Random(seed);
        //how many keys to create
        int numOfKeys = RandomKey.createIntKey(random, 8, 13);
        keys = new String[numOfKeys];
        String tempKey;
        
        int lengthOfKey = 0;
        for (int i = 0; i < keys.length; i++) {
            lengthOfKey = RandomKey.createIntKey(random, 1, 6);
            tempKey = "\"";
            for (int j = 0; j < lengthOfKey; j++) {
                tempKey = tempKey + RandomKey.keyIt(RandomKey.createIntKey(random, 3, 6));
            }
            keys[i] = tempKey + "\"";
        }
        //randomly select position for empty key
        int emptyStringPos = RandomKey.createIntKey(random, 0, numOfKeys);
        keys[emptyStringPos] = "\"\"";
        //randomly select the duplicate and a position for it 
        int duplicateFromPos = RandomKey.createIntKey(random, 0, numOfKeys);
        while (duplicateFromPos == emptyStringPos) {
            duplicateFromPos = RandomKey.createIntKey(random, 0, numOfKeys);
        }
        int duplicateToPos = RandomKey.createIntKey(random, 0, numOfKeys);
        while (Math.abs(duplicateToPos - duplicateFromPos) < 2 || duplicateToPos == emptyStringPos) {
            duplicateToPos = RandomKey.createIntKey(random, 0, numOfKeys);
        }
        keys[duplicateToPos] = keys[duplicateFromPos];

        inputTable = new Table(keys);
        commonTrie = new ClickableCommonTrie();
        
        return new FDT[] { inputTable, commonTrie };
        
    }
    
    public void setSeed(long v) {
        
        this.seed = v;
    }
    
    public String getDescription() {
        
        return "";
    }


    public boolean[] getContainerRotations() {
        return new boolean[]{false,true};
    }

    public boolean[] getModelAnswerRotations() {
        return getContainerRotations();
    }

    static final long serialVersionUID = -4820871225917227169L;
}
