/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2006 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// Change log
// ==========
// 24.05.06 oseppala : changed the class to use Tables instead of VirtualArrays
//                     made some variable names better, changed one field into
//                     a local variable, fixes to english variable names,
//                     made the hash parameters constant variables

package content.exercises;

import java.util.Random;

import content.interfaces.ConfigureVisualType;
import content.interfaces.ModelAnswerNames;
import content.interfaces.ComparableExercise;
import content.interfaces.SwapBehaviour;

import matrix.animation.Animator;
import matrix.simulation.VisualTypeConf;
import matrix.structures.FDT.FDT;
import matrix.structures.FDT.probe.Key;
import matrix.structures.FDT.probe.Table;

public class Hash_DoubleHash implements ComparableExercise, ModelAnswerNames, ConfigureVisualType, SwapBehaviour {

    // Size > 7 !!!!
    private Table inputDataTable;
    private Table userSolutionTable;

    static final int size = 19;
    static final int keySize = 10;
    static final int min = 100;
    static final int max = 999;
    static final int duplicates = 5;
    static final int moreDuplicates = 1;

    static final int firstHashParameter = 3;
    static final int secondHashParameter = 7;
    
    // This variable is for convenience reasons (the inputDataTable does not need to be parsed)
    private int inputData[] = new int[keySize];
    public static final boolean DEBUG = false;

    private long seed = 1;

    
    public Hash_DoubleHash() {
    }

    /**
     * In this exercise a drag operation performs a swap
     * @return
     */
    public boolean getSwapBehaviour() {
        return true;
    }
    

    /**
     * Get the value of seed.
     * @return value of seed.
     */
    public long getSeed() {
        return seed;
    }

    /**
     * Set the value of seed.
     * @param v  Value to assign to seed.
     */
    public void setSeed(long  v) {
        this.seed = v;
    }

    public FDT[] init() {
	// new random
        Random R = new Random(seed);
	// create random data
        for(int i = 0; i < keySize-duplicates; i++) {
            inputData[i] = Math.abs(R.nextInt()%(max-min))+min;
        }

        // Create duplicates
        for(int i = keySize-duplicates; i < keySize; i++) {
            
            do {
                inputData[i] = inputData[(0+i+duplicates-keySize)%(duplicates-moreDuplicates)];
                inputData[i] -= (R.nextInt(keySize))*size;
            } while(inputData[i]<min || inputData[i] > max);
        }

        inputDataTable = new Table(inputData);
        userSolutionTable = new Table();
        userSolutionTable.setObject(null, 0);
        userSolutionTable.setObject(null, size-1);
        return new FDT[] {inputDataTable, userSolutionTable};
    }


    public FDT[] init(Object o) {
        return init();
    }

    public String[] getStructureNames() {
        return new String[] { "Stream of Keys", "Hash table" };
    }

    public String getDescription() {
        String s = "The hash function is h(k) = (k + "+firstHashParameter+") mod "+size+"\n"+
	           "The second hash function is h(k) = "+secondHashParameter+" - (k mod "+secondHashParameter+")\n"; 
        return s;
    }

    public FDT[] solve() {
        FDT[] initialStructures = getInitialStructures();
        Table modelSolution = (Table)initialStructures[1];
        Table tbl = (Table)initialStructures[0];        
        Animator a = Animator.getActiveAnimator();
        for(int i = 0; i < tbl.size(); i++) {
            a.startOperation();
            probe(inputData[i], modelSolution);
            a.endOperation();
        }
        return new FDT[] { modelSolution };
    }



    public FDT[] makeModelAnswer() {
        return solve();
    }


    public FDT[] getInitialStructures() {
        Table resultTable = new Table();
        resultTable.setObject(null, 0);
        resultTable.setObject(null, size-1);
        return new FDT[] { new Table(inputData), resultTable };
    }

    public FDT[] getSimulatedStructures() {
        return new FDT[] { inputDataTable, userSolutionTable };
    }

    private void probe(int key, Table probedTable) {
        int n = 0;
        // hash key
        int hashValue = (key + firstHashParameter) % size;
        int doubleHashValue = secondHashParameter - (key % secondHashParameter);
        if (DEBUG) {
            System.out.println(this+ "i: " + hashValue);
            System.out.println(this+ "i2: " + doubleHashValue);
        }
        Object o;
        while(n < size) {
            o = probedTable.getObject((hashValue+n*doubleHashValue)%size);
            if(o == null) {
                probedTable.setObject(new Key(key), (hashValue+n*doubleHashValue)%size);
                return;
            }
            n++;
        }
    }


    public String[] getModelAnswerNames() {
        String[] s = { "Hash Table" };
        return s;
    }


    public FDT[] getAnswer() {
        return new FDT[] { userSolutionTable };
    }

    public ComparableExercise getModelAnswer() {
        return this;
    }

    public VisualTypeConf[] conf() {
        VisualTypeConf tabl = new VisualTypeConf();
        tabl.enable("matrix.visual.VisualKey", VisualTypeConf.DRAG_OPERATION);
        tabl.enable("matrix.visual.VisualKey", VisualTypeConf.HIGHLIGHT_OPERATION);
        VisualTypeConf hashtabl = new VisualTypeConf();
        hashtabl.enable("matrix.visual.VisualKey", VisualTypeConf.HIGHLIGHT_OPERATION);
        hashtabl.enable("matrix.visual.VisualKey", VisualTypeConf.DROP_OPERATION);
        hashtabl.enable("matrix.visual.VisualArrayComponent", VisualTypeConf.HIGHLIGHT_OPERATION);
        hashtabl.enable("matrix.visual.VisualArrayComponent", VisualTypeConf.DROP_OPERATION);
        return new VisualTypeConf[] { tabl, hashtabl };
    }

    public boolean canRecover() {
        return true;
    }

    public int[] getCompareIndices() {
        return new int[] {0};
    }

    static final long serialVersionUID = -37655555550211L;


	public int[] getGradeIndices() {
        return new int[] {0};
	}
}
