package content.exercises.sda;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Rectangle;
import java.util.Iterator;
import java.util.Random;

import content.exercises.sda.structures.GeometryFunctions;
import content.exercises.sda.structures.PointGenerator;
import content.exercises.sda.structures.SimpleStack;
import content.exercises.sda.structures.Triangulation;
import content.exercises.sda.structures.PointGenerator.IllegalConstraintsException;
import matrix.animation.Animator;
import matrix.structures.FDT.FDT;
import matrix.structures.spatial.PaintingStyleDecorator;
import matrix.structures.spatial.FDT.probe.LineSegment;
import matrix.structures.spatial.FDT.probe.Point;
import matrix.util.Note;

/**
 * <p>
 * An exercise for inserting additional points to a complete
 * Delaunay triangulation. The answer is checked against 
 * a correct triangulation and points are awarded for correct
 * Delaunay edges and deducted for incorrect ones.
 * </p> 
 * <p>
 * The algorithm for inserting the points: <br />
 * <ol>
 * 	<li>Find the triangle that contains the point being inserted.</li>
 * 	<li>Split the triangle into three new triangles by "connecting"
 * 		the point to the three vertices of the triangle.</li>
 * 	<li>Recursively test each edge/triangle pair for the delaunay
 * 		conditions and flip the diagonal if necessary, that is, 
 * 		in the quadrilateral formed by the two triangles switch
 * 		the diagonal. (Lawson's local optimization)</li>
 * </ol>
 * The Delaunay test can be performed by checking for points that are
 * inside the circumcircle of a triangle or by checking that 
 * the configuration maximizes the minimum angle in the triangles.
 * </p>
 * 
 * @author Juha Helminen
 *
 */
public class DelaunayPointInsert extends VoronoiConstruction {

    // For debugging
    private static final boolean STATIC_SEED = false;
    private static final long STATIC_SEED_VALUE = 1181113351228L;

    // The number of points in the given Delaunay triangulation.
    private static final int INITIAL_INPUT_SIZE = 10;
	// The number of points to be inserted in the triangulation.
    private static final int USER_INPUT_SIZE = 3; 
    // The minimum allowed distance between two points.
    private static final int MIN_DISTANCE = 100;
    // The minimum allowed distance from the points to the border.
    private static final int BORDER_X = 100;
    private static final int BORDER_Y = 100;
    // The size of the exercise Area.
    private static final int WIDTH = 450;
    private static final int HEIGHT = 350;

    private long seed = System.currentTimeMillis();
    
    // The style for the points that are being inserted in the triangulation.
    protected static final Color INSERT_POINT_COLOR = Color.BLUE;
    protected static final PaintingStyleDecorator insertPointDecorator;
    static {
    	insertPointDecorator = new PaintingStyleDecorator();
        insertPointDecorator.setLineColor(INSERT_POINT_COLOR);
        insertPointDecorator.setStroke(2, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL);
    }
    
    
    /* (non-Javadoc)
     * @see content.exercises.sda.VoronoiConstruction#getInitialStructures()
     */
    public FDT[] getInitialStructures() {
        SimpleStack problemVis = new SimpleStack(new Rectangle(WIDTH, HEIGHT, 0, 0));
        Triangulation triangulation = new Triangulation(new StudentVisualElements(), WIDTH, HEIGHT);
        for (int i = 0; i < INITIAL_INPUT_SIZE; i++) {
            triangulation.addVertex(this.input[i]);
        }
        for (int i = INITIAL_INPUT_SIZE; i < INITIAL_INPUT_SIZE + USER_INPUT_SIZE ; i++) {
            triangulation.addVertex(this.input[i], insertPointDecorator);
        }
        problemVis.addSubArea(triangulation);

        return new FDT[]{ problemVis, triangulation };
    }
    
    
	/* (non-Javadoc)
	 * @see content.exercises.sda.VoronoiConstruction#init()
	 */
	public FDT[] init() {
        final int BORDER_POINT_COUNT = 8;
        // For debugging
        Note.out(this, "The random number generator seed for this exercise is: " + this.getSeed());
        Random r = new Random(this.getSeed());

        // Generate the exercise input.
        final int MAX_TRIES = 200;
        int tryCounter = 0;
        java.awt.Point[] randomInput;
        java.awt.Point[][] initialTriangles;
        while (true) {
        	// Generate the points for the initial Delaunay triangulation.
	        try {
	            randomInput = PointGenerator.generatePoints(BORDER_X, BORDER_Y,
	                                                        WIDTH - BORDER_X, HEIGHT - BORDER_Y,
	                                                        INITIAL_INPUT_SIZE - BORDER_POINT_COUNT, MIN_DISTANCE, r);
	        } catch (IllegalConstraintsException e) {
	            Note.err(this, e.getMessage());
	            Note.show(this, "Could not generate input for the exercise. Please restart the exercise.");
	            return null;
	        }
	
	        // The border points.
	        this.input = new java.awt.Point[randomInput.length + BORDER_POINT_COUNT + USER_INPUT_SIZE];
	        for (int i = 0; i < randomInput.length; i++) {
	            this.input[i] = randomInput[i];
	        }
	        this.input[randomInput.length] = new java.awt.Point(BORDER_X / 4, BORDER_Y / 4);
	        this.input[randomInput.length + 1] = new java.awt.Point(BORDER_X / 4, HEIGHT - BORDER_Y / 4);
	        this.input[randomInput.length + 2] = new java.awt.Point(WIDTH - BORDER_X / 4, HEIGHT - BORDER_Y / 4);
	        this.input[randomInput.length + 3] = new java.awt.Point(WIDTH - BORDER_X / 4, BORDER_Y / 4);
	        this.input[randomInput.length + 4] = new java.awt.Point(WIDTH / 2, BORDER_Y / 6);
	        this.input[randomInput.length + 5] = new java.awt.Point(BORDER_X / 6, HEIGHT / 2);
	        this.input[randomInput.length + 6] = new java.awt.Point(WIDTH - BORDER_X / 6, HEIGHT / 2);
	        this.input[randomInput.length + 7] = new java.awt.Point(WIDTH / 2, HEIGHT - BORDER_Y / 6);

	        
	        // Generate the points to be inserted inside randomly selected triangles in
	        // the initial Delaunay triangulation.
            java.awt.Point[] initialInput = new java.awt.Point[this.input.length - USER_INPUT_SIZE];
            for (int i = 0; i < initialInput.length; i++) {
                initialInput[i] = this.input[i];
            }
            initialTriangles = GeometryFunctions.delaunay(initialInput);

	        if (initialTriangles == null) {
                if (tryCounter++ >= MAX_TRIES) {
                    Note.err(this, "Could not generate input for the exercise.");
                    Note.show(this, "Could not generate input for the exercise. Please restart the exercise.");
                    return null;
                } else {
                	continue;
                }
	        }
            
            boolean[] insertPointTriangles = new boolean[initialTriangles.length];
            int done = 0;
            int next;
            // USER_INPUT_SIZE - 1, because the last point is added separately
            // to satisfy  specific requirements (see below)
            while (done < USER_INPUT_SIZE - 1) {
            	next = r.nextInt(insertPointTriangles.length);
            	if (!insertPointTriangles[next]) {
            		done++;
            		insertPointTriangles[next] = true;
            	}
            }
            done = 0;
            for (int i = 0; i < initialTriangles.length; i++) {
            	if (insertPointTriangles[i] && (done < USER_INPUT_SIZE - 1)) {
            		// create the point
            		int x = (initialTriangles[i][0].x + initialTriangles[i][1].x + initialTriangles[i][2].x) / 3;
            		int y = (initialTriangles[i][0].y + initialTriangles[i][1].y + initialTriangles[i][2].y) / 3;
            		this.input[INITIAL_INPUT_SIZE + done++] = new java.awt.Point(x, y);
            	}
            }
            
            // Add the last point inside a triangle that shares an edge with
            // a triangle that already contains a point.
            loop:
	            for (int i = 0; i < initialTriangles.length; i++) {
	            	if (insertPointTriangles[i]) {
	            		// This triangle has a point inside it, 
	            		// find a neighbouring triangle
	                    for (int j = 0; j < initialTriangles.length; j++) {
	                    	if (!insertPointTriangles[j]) {
	                    		for (int k = 0; k < initialTriangles[i].length; k++) {
		                    		for (int l = 0; l < initialTriangles[j].length; l++) {
			                            if ((initialTriangles[i][k].equals(initialTriangles[j][l]) && 
			                            	 initialTriangles[i][(k + 1) % 3].equals(initialTriangles[j][(l + 1) % 3])) ||
			                            	(initialTriangles[i][k].equals(initialTriangles[j][(l + 1) % 3]) && 
				                             initialTriangles[i][(k + 1) % 3].equals(initialTriangles[j][l]))) {
				                        		int x = (initialTriangles[j][0].x + initialTriangles[j][1].x + initialTriangles[j][2].x) / 3;
				                        		int y = (initialTriangles[j][0].y + initialTriangles[j][1].y + initialTriangles[j][2].y) / 3;
				                           		this.input[INITIAL_INPUT_SIZE + USER_INPUT_SIZE - 1] = new java.awt.Point(x, y);
				                           		break loop;
			                            }
		                    		}
	                    		}
	                    	}
	                    }
	            	}
	            }
            
            // Test the complete exercise input for the illegal case of 4 (or more) 
            // cocircular/concyclic points.
	        if (GeometryFunctions.delaunay(this.input) == null) {
                if (tryCounter++ >= MAX_TRIES) {
                    Note.err(this, "Could not generate input for the exercise.");
                    Note.show(this, "Could not generate input for the exercise. Please restart the exercise.");
                    return null;
                }
	        } else {
	        	break;
	        }
        }

        // Calculate the number of edges in the correct triangulation.
        // Precondition: the border points define the convex hull of the point set
        this.edges = BORDER_POINT_COUNT + BORDER_POINT_COUNT + (INITIAL_INPUT_SIZE + USER_INPUT_SIZE - 1 - BORDER_POINT_COUNT) * 3;

        FDT[] structures = this.getInitialStructures();
        SimpleStack problemVis = (SimpleStack)structures[0];
        this.circumcircleVis = new SimpleStack();
        problemVis.addSubArea(this.circumcircleVis);
        this.triangulation = (Triangulation)structures[1];
        this.triangulation.setDrawVoronoi(Triangulation.VORONOI_NONE);
        this.triangulation.setDrawVoronoiVertices(false);

        // Add the initial triangulation.
        for (int i = 0; i < initialTriangles.length; i++) {
            this.triangulation.addEdge(initialTriangles[i][0], initialTriangles[i][1]);
            this.triangulation.addEdge(initialTriangles[i][1], initialTriangles[i][2]);
            this.triangulation.addEdge(initialTriangles[i][2], initialTriangles[i][0]);
        }
        
        return new FDT[]{problemVis};
	}
	
	
    public FDT[] solve() {
        this.correctEdges = 0;
        this.incorrectEdges = 0;
        this.missingEdges = 0;
        FDT[] modelStructures = this.getInitialStructures();
        FDT[] comparisonStructures = this.getInitialStructures();
        Triangulation modelTriangulation = (Triangulation)modelStructures[1];
        modelTriangulation.setDrawVoronoi(Triangulation.VORONOI_REGIONS);
        modelTriangulation.setDrawVoronoiVertices(false);
        modelTriangulation.setDrawDelaunay(false);
        Triangulation comparisonTriangulation = (Triangulation)comparisonStructures[1];
        comparisonTriangulation.setDrawVoronoiVertices(false);
        
        // Generate the points to be inserted inside randomly selected triangles in
        // the initial Delaunay triangulation.
        java.awt.Point[] initialInput = new java.awt.Point[this.input.length - USER_INPUT_SIZE];
        for (int i = 0; i < initialInput.length; i++) {
            initialInput[i] = this.input[i];
        }
        java.awt.Point[][] initialTriangles = GeometryFunctions.delaunay(initialInput);

        // Add the initial triangulation.
        for (int i = 0; i < initialTriangles.length; i++) {
            modelTriangulation.addEdge(initialTriangles[i][0], initialTriangles[i][1]);
            modelTriangulation.addEdge(initialTriangles[i][1], initialTriangles[i][2]);
            modelTriangulation.addEdge(initialTriangles[i][2], initialTriangles[i][0]);
            comparisonTriangulation.addEdge(initialTriangles[i][0], initialTriangles[i][1]);
            comparisonTriangulation.addEdge(initialTriangles[i][1], initialTriangles[i][2]);
            comparisonTriangulation.addEdge(initialTriangles[i][2], initialTriangles[i][0]);
        }

        
        Animator a = Animator.getActiveAnimator();
        a.startOperation();

        modelTriangulation.clearEdges();
        comparisonTriangulation.clearEdges();
        
        java.awt.Point[][] triangles = GeometryFunctions.delaunay(this.input);
        for (int i = 0; i < triangles.length; i++) {
            modelTriangulation.addEdge(triangles[i][0], triangles[i][1]);
            modelTriangulation.addEdge(triangles[i][1], triangles[i][2]);
            modelTriangulation.addEdge(triangles[i][2], triangles[i][0]);
        }

        Iterator modelEdges = modelTriangulation.edges();
        LineSegment edge;
        while (modelEdges.hasNext()) {
            edge = (LineSegment)modelEdges.next();
            if (this.triangulation.hasEdge(edge.getP1(), edge.getP2())) {
                comparisonTriangulation.addEdge(edge.getP1(), edge.getP2());
                this.correctEdges++;
            } else {
                comparisonTriangulation.addEdge(edge.getP1(), edge.getP2(), missingDelaunayDecorator);
                this.missingEdges++;
            }
        }

        Iterator studentEdges = this.triangulation.edges();
        while (studentEdges.hasNext()) {
            edge = (LineSegment)studentEdges.next();
            if (!modelTriangulation.hasEdge(edge.getP1(), edge.getP2())) {
                comparisonTriangulation.addEdge(edge.getP1(), edge.getP2(), wrongDelaunayDecorator);
                this.incorrectEdges++;
            }
        }
        
        a.endOperation();
        
        return new FDT[]{ modelStructures[0], comparisonStructures[0] };
    }
	
	
    /* (non-Javadoc)
     * @see content.interfaces.SimulationExercise#getSeed()
     */
    public long getSeed() {
        // The seed might be set to a constant value for debugging purposes.
        if (STATIC_SEED) {
            return DelaunayPointInsert.STATIC_SEED_VALUE;
        } else {
            return this.seed;
        }
    }

    /* (non-Javadoc)
     * @see content.interfaces.SimulationExercise#setSeed(long)
     */
    public void setSeed(long seed) {
        this.seed = seed;
    }
    
    protected void clearPointColors() {
        // Clear point coloring
        Iterator vertices = DelaunayPointInsert.this.triangulation.vertices();
        Point point;
        loop:
            while (vertices.hasNext()) {
                point = (Point)vertices.next();
                for (int i = INITIAL_INPUT_SIZE; i < INITIAL_INPUT_SIZE + USER_INPUT_SIZE ; i++) {
                    if (point.getPoint().equals(this.input[i])) {
                        point.setPaintingStyleDecorator(insertPointDecorator);
                        continue loop;
                    }
                }
                point.setPaintingStyleDecorator(vertexDecorator);
            }
    }
    
    private static final long serialVersionUID = -7127809006061268311L;
}
