package content.exercises.sda;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.LayoutManager;
import java.awt.Rectangle;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.util.Random;

import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;

import matrix.animation.Animator;
import matrix.simulation.VisualTypeConf;
import matrix.structures.CDT.CDT;
import matrix.structures.CDT.probe.QueueImpl;
import matrix.structures.FDT.FDT;
import matrix.structures.FDT.probe.Key;
import matrix.structures.memory.VirtualObject;
import matrix.structures.simulationextensions.Selectable;
import matrix.structures.spatial.PaintingStyleDecorator;
import matrix.structures.spatial.FDT.probe.Point;
import matrix.structures.spatial.FDT.probe.PolyLine;
import matrix.structures.spatial.FDT.probe.Polygon;
import matrix.util.Application;
import matrix.util.Note;
import content.exercises.sda.structures.PointGenerator;
import content.exercises.sda.structures.PolygonGenerator;
import content.exercises.sda.structures.SimpleStack;
import content.interfaces.AWTComponentUtilizer;
import content.interfaces.ConfigureVisualType;
import content.interfaces.Exercise;
import content.interfaces.JudgeBlocks;
import content.interfaces.LayoutExercise;
import content.interfaces.ModelAnswerNames;
import content.interfaces.SimulationExercise;
import content.interfaces.SimulationExerciseModel;
import content.interfaces.StyledExercise;

/**
 * An exercise for the Douglas-Peucker line simplification algorithm.
 * 
 * @author Juha Helminen
 */
public class DouglasPeucker implements Exercise, SimulationExercise, 
									   SimulationExerciseModel, ConfigureVisualType,
									   AWTComponentUtilizer, ModelAnswerNames,
                                       JudgeBlocks, LayoutExercise, StyledExercise {

    // For debugging
	private static final boolean STATIC_SEED = false;
	private static final long STATIC_SEED_VALUE = 1176811897861L;

	// ** Exercise configuration **
	private static final int POLYLINE_VERTEX_COUNT = 15;
	private static final int X_BORDER = 20;
	private static final int Y_BORDER = 25;
	private static final int WIDTH = 450;
	private static final int HEIGHT = 280;
	private static final int TOLERANCE = 35;
    // The minimum allowed distance between two vertices.
    private static final int MIN_VERTEX_VERTEX_DISTANCE = 37;
    // The minimum allowed distance between a polygon vertex and edge.
    private static final int MIN_VERTEX_EDGE_DISTANCE = 37;
    
    // ** Color and style scheme **
    private static final float LEGEND_FONT_SIZE = 14f; 
    private static final Color TOLERANCE_COLOR = Color.GRAY;
    private static final Color TOLERANCE_FILL_COLOR = Color.GREEN;
    private static final Color POLYLINE_COLOR = Color.WHITE;
    private static final Color VERTEX_COLOR = Color.BLUE;
    private static final Color TESTLINE_COLOR = Color.RED;
    private static final Color SIMPLIFIED_POLYLINE_COLOR = Color.RED;
    private static final Color DISTANCE_OUT_COLOR = Color.YELLOW;
    private static final Color DISTANCE_IN_COLOR = Color.ORANGE;
    private static final PaintingStyleDecorator toleranceDecorator;
    private static final PaintingStyleDecorator polylineDecorator;
    private static final PaintingStyleDecorator testLineDecorator;
    private static final PaintingStyleDecorator distanceOutDecorator;
    private static final PaintingStyleDecorator distanceInDecorator;
    private static final PaintingStyleDecorator simplifiedPolylineDecorator;
    private static final PaintingStyleDecorator vertexDecorator;
    static {
        float[] toleranceDashPattern = { 10, 10 };
        toleranceDecorator = new PaintingStyleDecorator();
        toleranceDecorator.setLineColor(TOLERANCE_COLOR);
        toleranceDecorator.setFillColor(TOLERANCE_FILL_COLOR);
        toleranceDecorator.setTransparency(0.10);
        toleranceDecorator.setStroke(2, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL, 10, toleranceDashPattern, 0);

        polylineDecorator = new PaintingStyleDecorator();
        polylineDecorator.setLineColor(POLYLINE_COLOR);
        polylineDecorator.setStroke(3, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL);

        float[] testLineDashPattern = { 5 , 5, 10, 5 };
        testLineDecorator = new PaintingStyleDecorator();
        testLineDecorator.setLineColor(TESTLINE_COLOR);
        testLineDecorator.setStroke(3, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL, 5, testLineDashPattern, 0);

        float[] distanceDashPattern = { 5 , 3 };

        distanceOutDecorator = new PaintingStyleDecorator();
        distanceOutDecorator.setLineColor(DISTANCE_OUT_COLOR);
        distanceOutDecorator.setStroke(3, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL, 3, distanceDashPattern, 0);

        distanceInDecorator = new PaintingStyleDecorator();
        distanceInDecorator.setLineColor(DISTANCE_IN_COLOR);
        distanceInDecorator.setStroke(3, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL, 3, distanceDashPattern, 0);

        simplifiedPolylineDecorator = new PaintingStyleDecorator();
        simplifiedPolylineDecorator.setLineColor(SIMPLIFIED_POLYLINE_COLOR);
        simplifiedPolylineDecorator.setStroke(3, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL);

        vertexDecorator = new PaintingStyleDecorator();
        vertexDecorator.setLineColor(VERTEX_COLOR);
    }
    

	// ** Structures for the exercise. **
    private java.awt.Point[] polylineVertices;

    // Student structures.
    private StudentStack studentStack;
    private StudentList studentList;
    private VirtualObject anchor;
    private VirtualObject floater;
    
    // Student problem visualisations.
    private SimpleStack toleranceVis;
    private SimpleStack simplifiedLineVis;
    private SimpleStack distanceVis;

	transient private VirtualDoubleField distance;
	
    private long seed = System.currentTimeMillis();


    // ** Implementation of interface Exercise **

	public String getDescription() {
		return "Simplify the polyline following the Douglas-Peucker line " + 
			   "simplification algorithm.";
	}

	// ** Implementation of interface SimulationExercise **
	
	public FDT[] getAnswer() {
		return new FDT[]{ this.studentStack, this.studentList };
	}

	public FDT[] getInitialStructures() {
        SimpleStack polylineVis = new SimpleStack();

        SelectablePoint[] points = new SelectablePoint[this.polylineVertices.length];
        for (int i = 0; i <this. polylineVertices.length; i++) {
            points[i] = new SelectablePoint(this.polylineVertices[i]);
            points[i].setIndex(i);
            points[i].setPaintingStyleDecorator(vertexDecorator);
            polylineVis.ADTPush(points[i]);
        }
        PolyLine polyline = new PolyLine(points, "");
        polyline.setPaintingStyleDecorator(polylineDecorator);
        polylineVis.ADTPush(polyline);
        
		return new FDT[]{ polylineVis };
	}

	public String[] getStructureNames() {
		return new String[]{ "Polyline", "Stack", "Stack top", "The simplified polyline" };
	}

	public FDT[] init() {
        // For debugging
        Note.out(this, "The random number generator seed for this exercise is: " + this.getSeed());
        Random r = new Random(this.getSeed());

        this.anchor = new VirtualObject(null, "anchor");
        this.floater = new VirtualObject(null, "floater");

        final int MAX_TRIES = 80;
        int tryCounter = 0;
        while (true) {
            // Generate the input.
            try {
                this.polylineVertices = PointGenerator.generatePoints(X_BORDER,
                                                           Y_BORDER,
                                                           WIDTH - X_BORDER,
                                                           HEIGHT - Y_BORDER,
                                                           POLYLINE_VERTEX_COUNT,
                                                           MIN_VERTEX_VERTEX_DISTANCE,
                                                           r);
                // Order the points in such a manner that they form the contour of a polygon.
                PolygonGenerator.generatePolygon2Peasants(this.polylineVertices);
                Line2D[] edges = createPolygonEdges(this.polylineVertices);
                for (int j = 0; j < this.polylineVertices.length; j++) {
                    if (!checkDistance2(this.polylineVertices[j], edges, MIN_VERTEX_EDGE_DISTANCE)) {
                        throw new Exception("Failed to generate the polygon.");
                    }
                }
                // success
                break;
            } catch (Exception e) {
                if (tryCounter++ >= MAX_TRIES) {
                    Note.err(this, e.getMessage());
                    Note.show(this, "Could not generate input for the exercise. Please restart the exercise.");
                    return null;
                }
            }
        }
        
        SimpleStack polylineVis = (SimpleStack)this.getInitialStructures()[0];
        
        SimpleStack problemVis = new SimpleStack(new Rectangle(WIDTH, HEIGHT, 0, 0));
        this.toleranceVis = new SimpleStack();
        this.distanceVis = new SimpleStack();
        this.simplifiedLineVis = new SimpleStack();
        problemVis.addSubArea(polylineVis);
        problemVis.addSubArea(simplifiedLineVis);
        problemVis.addSubArea(toleranceVis);
        problemVis.addSubArea(distanceVis);
        this.studentStack = new StudentStack();
        this.studentList = new StudentList();
		
        return new FDT[]{problemVis,
                         this.studentStack.getArrayVisualization(), 
                         this.studentStack, 
                         this.studentList};
	}

    private static Line2D[] createPolygonEdges(java.awt.Point[] vertices) {
        Line2D[] edges = null;
        if(vertices.length > 1) {
            edges = new Line2D[vertices.length];
            for (int i = 0; i < vertices.length - 1; i++) {
                edges[i] = new Line2D.Double(vertices[i], vertices[i + 1]);
            }
            // Close the polygon.
            edges[vertices.length - 1] = new Line2D.Double(vertices[vertices.length - 1], 
                                                           vertices[0]);
        }
        return edges;
    }

    private static boolean checkDistance2(Point2D point, 
                                          Line2D[] lines, 
                                          int minDistance) {
        for (int i = 0; i < lines.length; i++) {
            if (lines[i].getP1().equals(point) ||
                lines[i].getP2().equals(point)) {
                continue;
            } else if (lines[i].ptSegDist(point) < minDistance) {
                return false;
            }
        }
        return true;
    }
    
    public long getSeed() {
		// Seed might be set to a constant value for debugging purposes.
		if (STATIC_SEED) {
			return DouglasPeucker.STATIC_SEED_VALUE;
		} else {
			return this.seed;
		}
	}

	public void setSeed(long seed) {
		this.seed = seed;
	}
	

	// ** Implementation of interface SimulationExercisemodel **

	public FDT[] makeModelAnswer() {
		return solveHelp();
	}

	public FDT[] solve() {
		FDT[] f = solveHelp();
		return new FDT[]{ f[2], f[3] };
	}

	private FDT[] solveHelp() {
		SimpleStack solutionVis = new SimpleStack(new Rectangle(WIDTH, HEIGHT, 0, 0));
        SimpleStack toleranceVis = new SimpleStack();
        SimpleStack polylineVis = new SimpleStack();
        SimpleStack distanceVis = new SimpleStack();
        SimpleStack simplifiedLineVis = new SimpleStack();
        solutionVis.addSubArea(polylineVis);
        solutionVis.addSubArea(simplifiedLineVis);
        solutionVis.addSubArea(toleranceVis);
        solutionVis.addSubArea(distanceVis);
        SimpleStack s = new SimpleStack();
        QueueImpl q = new QueueImpl();
		
        SelectablePoint[] points = new SelectablePoint[polylineVertices.length];
        for (int i = 0; i < polylineVertices.length; i++) {
            points[i] = new SelectablePoint(polylineVertices[i]);
            points[i].setIndex(i);
            points[i].setPaintingStyleDecorator(vertexDecorator);
            polylineVis.ADTPush(points[i]);
        }
        PolyLine polyline = new PolyLine(points, "");
        polyline.setPaintingStyleDecorator(polylineDecorator);
        polylineVis.ADTPush(polyline);

        Animator a = Animator.getActiveAnimator();
		// Initialize structures.
		SelectablePoint anchor = points[0];
		SelectablePoint floater = points[points.length - 1];
		a.startOperation();
		q.insert(anchor);
		a.endOperation();
		a.startOperation();
		s.ADTPush(floater);
		this.drawTolerance(anchor, floater, toleranceVis);
		while (!s.ADTIsEmpty()) {
			// Select the vertex which is farthest from
			// the line from the anchor to the floater.
			// Find the vertex with  maximum distance to 
			// the line segment anchor->floater.
			SelectablePoint farthest = floater;
			double maxDistance = -1;
			Line2D line = new Line2D.Double(anchor.getPoint2D(), floater.getPoint2D());
			for (int i = anchor.getIndex() + 1; i < floater.getIndex(); i++) {
				if (line.ptSegDist(points[i].getPoint2D()) > maxDistance) {
					maxDistance = line.ptSegDist(points[i].getPoint2D());
					farthest = points[i];
				}
			}
			// Draw a line to the vertex who is farthest from 
			// the line from the anchor to the floater.
			
			if (maxDistance <= TOLERANCE) {
				// All the vertices are inside the tolerance area
				// -> eliminate them and move the anchor.
				a.endOperation();
				a.startOperation();
				q.insert(s.ADTPop());
				toleranceVis.clear();
				this.drawSimplifiedLine(anchor, floater, simplifiedLineVis);
				anchor = floater;
				floater = (SelectablePoint)s.ADTPeek();
				if (floater != null) {
					this.drawTolerance(anchor, floater, toleranceVis);
				}
			} else {
				// The point is outside the tolerance area and
				// becomes the new floater.
				drawDistance(anchor, floater, farthest, distanceVis, distanceOutDecorator);
				a.endOperation();
				a.startOperation();
				distanceVis.clear();
				floater = farthest;
				s.insert(floater);
                toleranceVis.clear();
				this.drawTolerance(anchor, floater, toleranceVis);
			}
		}		
		a.endOperation();
		return new FDT[]{ solutionVis, s.getArrayVisualization(), s, q };
	}


	// ** Implementation of interface ModelAnswerNames **

	public String[] getModelAnswerNames() {
		return new String[]{ "Polyline", "Stack", "Stack top", "The simplified polyline" };
	}

	
	// ** Implementation of interface ConfigureVisualType **
	
	public VisualTypeConf[] conf() {
		VisualTypeConf polyline = new VisualTypeConf();
		polyline.enable("matrix.visual.VisualArea$AreaVisualization", VisualTypeConf.HIGHLIGHT_OPERATION);
		polyline.enable("matrix.visual.VisualArea$AreaVisualization", VisualTypeConf.DRAG_OPERATION);
		VisualTypeConf stackTop = new VisualTypeConf();
		stackTop.enable("matrix.visual.VisualArray", VisualTypeConf.DROP_OPERATION);
		stackTop.enable("matrix.visual.VisualArray", VisualTypeConf.HIGHLIGHT_OPERATION);
		stackTop.setVisualConfValue("matrix.visual.VisualArray", VisualTypeConf.INDEXED, "false");
		stackTop.enable("matrix.visual.VisualArrayComponent", VisualTypeConf.HIGHLIGHT_OPERATION);
		stackTop.enable("matrix.visual.VisualArrayComponent", VisualTypeConf.DRAG_OPERATION);
        stackTop.enable("matrix.visual.VisualStruct-matrix.structures.spatial.FDT.probe.Point", VisualTypeConf.MINIMIZED);
        VisualTypeConf stack = new VisualTypeConf();
        stack.setVisualConfValue("matrix.visual.VisualArray", VisualTypeConf.INDEXED, "false");
        stack.enable("matrix.visual.VisualStruct-matrix.structures.spatial.FDT.probe.Point", VisualTypeConf.MINIMIZED);
		VisualTypeConf queue = new VisualTypeConf();
		queue.enable("matrix.visual.VisualList", VisualTypeConf.DROP_OPERATION);
		queue.enable("matrix.visual.VisualList", VisualTypeConf.HIGHLIGHT_OPERATION);
        queue.enable("matrix.visual.VisualStruct-matrix.structures.spatial.FDT.probe.Point", VisualTypeConf.MINIMIZED);
		return new VisualTypeConf[]{ polyline, stack, stackTop, queue };
	}
    
    // ** Implementation of interface StyledExercise **
    
    public String[] getStructureVisualisations() {
        return new String[]{ "2D", "array", "array", "list"};
    }

    public String[] getModelAnswerVisualisations() {
        return this.getStructureVisualisations();
    }
    
    // ** Implementation of interface LayoutExercise **

    public GridBagConstraints[] getConstraints() {
        GridBagConstraints[] gbc = new GridBagConstraints[4];
        for (int i = 0; i < 4; i++) {
            gbc[i] = new GridBagConstraints();
        }
        Insets insets = new Insets(3, 3, 0, 0);
        gbc[0].insets = insets;
        gbc[1].insets = insets;
        gbc[2].insets = new Insets(3, 0, 0, 0);
        gbc[3].insets = insets;
        gbc[0].gridy = 0;
        gbc[0].gridwidth = 3;
        gbc[1].gridy = 1;
        gbc[2].gridy = 1;
        gbc[3].gridy = 2;
        gbc[3].gridwidth = 6;
        gbc[3].fill = GridBagConstraints.HORIZONTAL;
        return new GridBagConstraints[]{ gbc[0], gbc[1], gbc[2], gbc[3] };
    }

    public GridBagConstraints[] getModelAnswerConstraints() {
        return this.getConstraints();
    }

    
    // ** Implementation of interface AWTComponentUtilizer **

    class VirtualDoubleField extends JTextField {
        VirtualObject value;
        double textValue;
        
        public VirtualDoubleField() {
            super("NA", 8);
            this.value = new VirtualObject(null, "value");
            this.value.setObject(new Double(Double.NaN));
            this.textValue = Double.NaN;
        }

        public synchronized void setValue(double value) {
            this.value.setObject(new Double(value));
            this.validateText();
        }
        
        private synchronized double getValue() {
            return ((Double)this.value.getObject()).doubleValue();
        }
        
        private synchronized void validateText() {
            if ((this.textValue != this.getValue()) &&
                (!(Double.isNaN(this.textValue) && Double.isNaN(this.getValue())))) {
                this.textValue = this.getValue();
                if (Double.isNaN(this.getValue())) {
                    super.setText("NA");
                } else {
                    //super.setText(String.format("%.2f", new Object[]{ new Double(this.getValue()) })); 1.5-specific code removed -jtn
                    String str = String.valueOf(this.getValue());
                    if (str.length() > 9) {
                        super.setText(str.substring(0, 9));
                    } else {
                        super.setText(str);
                    }
                }
            }
        }

        public void validate() {
            this.validateText();
            super.validate();
        }
        
        private static final long serialVersionUID = -3666908688650017002L;
    }
    
    class MyPanel extends JPanel {

        public MyPanel(LayoutManager layout) {
            super(layout);
        }

        public void validate() {
            // Force validating of subcomponents
            int componentCount = this.getComponentCount();
            for (int i = 0; i < componentCount; i++) {
                this.getComponent(i).validate();
            }
            super.validate();
        }

        private static final long serialVersionUID = 3699821165607871718L;
    }
    
	public void setApplication(Application a) {
        // Reference to the application is not needed.
	}

	public Component[] getAWTComponents() {
		GridBagLayout gbLayout = new GridBagLayout();
		MyPanel gridPanel = new MyPanel(gbLayout);
        JLabel initialLabel = new JLabel("Initial polyline");
        Font font = initialLabel.getFont().deriveFont(LEGEND_FONT_SIZE); 
        initialLabel.setForeground(POLYLINE_COLOR);
        initialLabel.setFont(font);
        JLabel pointsLabel = new JLabel("Polyline vertices");
        pointsLabel.setForeground(VERTEX_COLOR);
        pointsLabel.setFont(font);
        JLabel simplifiedLabel = new JLabel("Simplified polyline");
        simplifiedLabel.setForeground(SIMPLIFIED_POLYLINE_COLOR);
        simplifiedLabel.setFont(font);
		JLabel toleranceLabel = new JLabel("Tolerance");
        toleranceLabel.setForeground(TOLERANCE_COLOR);
        toleranceLabel.setFont(font);
		JTextField tolerance = new JTextField(String.valueOf(TOLERANCE), 4);
		tolerance.setEditable(false);
        tolerance.setHorizontalAlignment(JTextField.CENTER);
        tolerance.setFont(font);
		JLabel distanceLabel = new JLabel("Distance");
        distanceLabel.setForeground(DISTANCE_OUT_COLOR);
        distanceLabel.setFont(font);
		this.distance = new VirtualDoubleField();
		this.distance.setEditable(false);
        this.distance.setHorizontalAlignment(JTextField.CENTER);
		this.distance.setFont(font);
        
		GridBagConstraints gbc = new GridBagConstraints();
        gridPanel.setBackground(new Color(192, 192, 192));
        gbc.gridx = GridBagConstraints.REMAINDER;
        gbc.insets = new Insets(10, 5, 0, 0);
        gridPanel.add(initialLabel, gbc);
        gridPanel.add(pointsLabel, gbc);
        gridPanel.add(simplifiedLabel, gbc);
        gridPanel.add(toleranceLabel, gbc);
        gbc.insets = new Insets(2, 5, 0, 0);
        gridPanel.add(tolerance, gbc);
        gbc.insets = new Insets(10, 5, 0, 0);
        gridPanel.add(distanceLabel, gbc);
        gbc.insets = new Insets(2, 5, 0, 0);
        gridPanel.add(distance, gbc);
        return new Component[]{ gridPanel };
	}

    public GridBagConstraints[] getAWTComponentConstraints() {
        return null;
    }
    
	
	// ** Other methods **
 
    private void switchFloater(Point point) {
        this.toleranceVis.clear();
		this.setFloater(point);
		if (this.getAnchor() != null) {
            drawTolerance(this.getAnchor(), this.getFloater(), this.toleranceVis);
		}
	}
	
	private void switchAnchor(Point point) {
    	this.setFloater((Point)this.studentStack.ADTPeek());            	
		if (this.getAnchor() == null) {
			// No anchor has been set, i.e. the simplified polyline list is empty
		} else {
			// Connect the previous anchor, that is, the vertex at the end of the current 
			// simplification to the new anchor.
            drawSimplifiedLine(this.getAnchor(), point, this.simplifiedLineVis);
		}
		this.setAnchor(point);
        this.toleranceVis.clear();
		if (this.getFloater() != null) {
			// If the floater is set, that is, there is at least one vertex in the stack,
			// draw the candidate line.
            drawTolerance(this.getAnchor(), this.getFloater(), this.toleranceVis);
		}
	}

	private void drawTolerance(Point anchor, Point floater, SimpleStack stack) {
		double[] unitNormal = new double[2];
		double length;
		double dy = anchor.getY() - floater.getY();
		if (dy == 0) {
			// The line segment is horizontal -> normal is vertical
			unitNormal[0] = 0;
			unitNormal[1] = 1;
		} else {
			unitNormal[0] = 1;
			unitNormal[1] = -(anchor.getX() - floater.getX()) / dy;
			length = Math.sqrt((unitNormal[0] * unitNormal[0]) + (unitNormal[1] * unitNormal[1]));
			unitNormal[0] /= length;
			unitNormal[1] /= length;
		}
		drawTestLine(anchor, floater, stack);
        int divisions = 10;
        double precision = Math.PI / divisions;
        double a = Math.atan(unitNormal[1] / unitNormal[0]);
        Point[] points = new Point[divisions * 2 + 1];
        int sign = 1;
        int j;
        if (anchor.getX() >= floater.getX()) {
            if (anchor.getY() > floater.getY()) {
                sign = -1;
            } else {
                sign = 1;
            }
        } else {
            if (anchor.getY() >= floater.getY()) {
                sign = -1;
            } else {
                sign = 1;
            }
        }
        for (int i = 0; i < points.length / 2; i++) {
            j = sign * i;
            points[i] = new Point(anchor.getX() + TOLERANCE * Math.cos(a - j * precision), anchor.getY() + TOLERANCE * Math.sin(a - j * precision), "");
        }
        for (int i = points.length / 2; i < points.length; i++) {
            j = sign * i;
            points[i] = new Point(floater.getX() + TOLERANCE * Math.cos(a - j * precision), floater.getY() + TOLERANCE * Math.sin(a - j * precision), "");
        }
        Polygon polygon = new Polygon(points, "");
        polygon.setPaintingStyleDecorator(toleranceDecorator);
        stack.ADTPush(polygon);
	}
    
	private void drawTestLine(Point anchor, Point floater, SimpleStack stack) {
        PolyLine polyLine = new PolyLine(new Point[]{ anchor, floater }, "");
        polyLine.setPaintingStyleDecorator(testLineDecorator);
        stack.ADTPush(polyLine);
	}

    private void drawSimplifiedLine(Point anchor, Point floater, SimpleStack stack) {
        PolyLine polyLine = new PolyLine(new Point[]{ anchor, floater }, "");
        polyLine.setPaintingStyleDecorator(simplifiedPolylineDecorator);
        stack.ADTPush(polyLine);
    }

    private void drawDistance() {
		if ((this.getAnchor() != null) && (this.getFloater() != null)) {
			Line2D l1 = new Line2D.Double(this.getAnchor().getPoint2D(), this.getFloater().getPoint2D());
            double distance = l1.ptSegDist(((Point)this.selectedPoint.getObject()).getPoint2D());
			this.setDistance(distance);
            if (distance <= TOLERANCE) {
                this.drawDistance(this.getAnchor(), 
                                  this.getFloater(), 
                                  (Point)this.selectedPoint.getObject(), 
                                  this.distanceVis,
                                  distanceInDecorator);
            } else {
                this.drawDistance(this.getAnchor(), 
                                  this.getFloater(), 
                                  (Point)this.selectedPoint.getObject(), 
                                  this.distanceVis,
                                  distanceOutDecorator);
            }
		}
	}

	private void drawDistance(Point anchor, 
                              Point floater, 
                              Point p, 
                              SimpleStack stack,
                              PaintingStyleDecorator psd) {
		double[] va = new double[2];
		double[] vb = new double[2];
		double[] vc = new double[2];
		// Vector anchor->floater
		va[0] = floater.getX() - anchor.getX(); 
		va[1] = floater.getY() - anchor.getY();
		// Vector anchor->p
		vb[0] = p.getX() - anchor.getX(); 
		vb[1] = p.getY() - anchor.getY();
		// Vector floater->p
		vc[0] = p.getX() - floater.getX(); 
		vc[1] = p.getY() - floater.getY();
		// Use dot product to determine how the point is located relative to the line segment.
        PolyLine polyLine;
        if ((va[0] * vb[0] + va[1] * vb[1]) < 0) {
			if ((-va[0] * vc[0] + -va[1] * vc[1]) < 0) {
				double[] normal = new double[2];
				double dy = anchor.getY() - floater.getY();
				if (dy == 0) {
					normal[0] = 0;
					normal[1] = 1;
				} else {
					normal[0] = 1;
					normal[1] = -(anchor.getX() - floater.getX()) / dy;
				}
				Line2D l1 = new Line2D.Double(anchor.getPoint2D(), floater.getPoint2D());
				Line2D l2 = new Line2D.Double(p.getPoint2D(), new Point2D.Double(p.getX() + normal[0], p.getY() + normal[1]));
                polyLine = new PolyLine(new Point[]{    getLineLineIntersection(l1, l2), p }, "");
                polyLine.setPaintingStyleDecorator(psd);
                stack.ADTPush(polyLine);
			} else {
                polyLine = new PolyLine(new Point[]{ anchor,    p }, "");
                polyLine.setPaintingStyleDecorator(psd);
                stack.ADTPush(polyLine);
			}
		} else {
			if ((-va[0] * vc[0] + -va[1] * vc[1]) < 0) {
                polyLine = new PolyLine(new Point[]{ floater, p }, "");
                polyLine.setPaintingStyleDecorator(psd);
                stack.ADTPush(polyLine);
			} else {
				double[] normal = new double[2];
				double dy = anchor.getY() - floater.getY();
				if (dy == 0) {
					normal[0] = 0;
					normal[1] = 1;
				} else {
					normal[0] = 1;
					normal[1] = -(anchor.getX() - floater.getX()) / dy;
				}
				Line2D l1 = new Line2D.Double(anchor.getPoint2D(), floater.getPoint2D());
				Line2D l2 = new Line2D.Double(p.getPoint2D(), new Point2D.Double(p.getX() + normal[0], p.getY() + normal[1]));
                polyLine = new PolyLine(new Point[]{    getLineLineIntersection(l1, l2), p }, "");
                polyLine.setPaintingStyleDecorator(psd);
                stack.ADTPush(polyLine);
			}
		}
	}
	
	/**
	 * Calculate and return the intersection of two lines.
	 */
	private static Point getLineLineIntersection(Line2D l1, Line2D l2) {
		double ax = l1.getX1(), ay = l1.getY1(),
			   bx = l2.getX1(), by = l2.getY1(),
			   cx = l1.getX2(), cy = l1.getY2(),
			   dx = l2.getX2(), dy = l2.getY2();
		Point p = new Point(0, 0, "");
		p.setX(det(det(ax, ay, cx, cy), 
				   ax - cx,
				   det(bx, by, dx, dy), 
				   bx - dx) /
			   det(ax - cx, 
				   ay - cy, 
				   bx - dx, 
				   by - dy));
		p.setY(det(det(ax, ay, cx, cy), 
				   ay - cy,
				   det(bx, by, dx, dy), 
				   by - dy) /
			   det(ax - cx, 
				   ay - cy, 
				   bx - dx, 
				   by - dy));
		return p;
	}
	
	private static double det(double a, double b, double c, double d){
		return (a * d) - (b * c);
	}

	private void setFloater(Point p) {
		this.floater.setObject(p);
	}
	
	private void setAnchor(Point p) {
		this.anchor.setObject(p);
	}
	
	private Point getAnchor() {
		return (Point)this.anchor.getObject();
	}

	private Point getFloater() {
		return (Point)this.floater.getObject();
	}
	
	private void setDistance(double value) {
		this.distance.setValue(value);
	}


    class StudentStack extends SimpleStack {

		public StudentStack() {
			super();
		}

		public void ADTPush(Object o) {
            clearSelection();
			super.ADTPush(o);
			switchFloater((Point)o);
		}
		
		private static final long serialVersionUID = -2881294854736131824L;
	}
	
	class StudentList extends QueueImpl {

		public StudentList() {
            super();
		}

		public CDT insert(Object o) {
		    if (o instanceof Key) {
		        o = ((Key)o).getObject();
		    }
		    if (o != null) {
                clearSelection();
    			if (o.equals(DouglasPeucker.this.studentStack.ADTPeek())) {
    			    DouglasPeucker.this.studentStack.ADTPop();
    			}
    			switchAnchor((Point)o);
    			return super.insert(o);
		    } else {
		        return this;
		    }
		}
		
		private static final long serialVersionUID = -7376696877829029332L;
	}

    
    // ** The class and methods allowing the selection of points 
    // ** and execution of commands based on this.
    
    private VirtualObject selectedPoint = new VirtualObject(null, "selectedPoint");

    private synchronized void clearSelection() {
        this.distanceVis.clear();
        this.distance.setValue(Double.NaN);
        this.select(null);
    }

    private synchronized void select(SelectablePoint p) {
        this.selectedPoint.setObject(p);
        if (p != null) {
            drawDistance();
        }
    }
    
    private synchronized boolean hasSelection() {
        return (this.getSelection() != null);
    }
    
    private synchronized SelectablePoint getSelection() {
        return (SelectablePoint)this.selectedPoint.getObject();
    }
    
    class SelectablePoint extends Point implements Selectable {
        private int index;
		
		public SelectablePoint(Point2D p) {
			super(p);
			setDraggable(true);
			this.index = 0;
		}
		
		public long getSelectionTime() {
            if (isSelected()) {
                return Selectable.clock.getNextTimeStamp();
            } else {
                return Selectable.NOT_SELECTED;
            }
		}

		public void setSelect(long globalClockTimestamp) {
            if ((getAnchor() != null) && (getFloater() != null)) {
                Animator a = Animator.getActiveAnimator();
                a.startOperation();
                if (isSelected()) {
                    clearSelection();
                } else if (globalClockTimestamp != Selectable.NOT_SELECTED) {
                    if (hasSelection()) {
                        clearSelection();
                    }
                    select(this);
                }
                a.endOperation();
            }
		}

		private boolean isSelected() {
            return (getSelection() == this);
		}
		
		public void setIndex(int index) {
            // Start indexing from 1.
			this.setVisualizationString(String.valueOf(index + 1));
			this.index = index;
		}
		
		public int getIndex() {
			return this.index;
		}

        private static final long serialVersionUID = 4536065312896843619L;
	}

    private static final long serialVersionUID = -1730502700302671164L;
}
