package content.exercises.sda;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.GridBagConstraints;
import java.awt.Insets;
import java.awt.Rectangle;
import java.awt.geom.Point2D;
import java.util.HashMap;
import java.util.Random;
import java.util.Vector;

import matrix.animation.Animator;
import matrix.simulation.VisualTypeConf;
import matrix.structures.CDT.CDT;
import matrix.structures.CDT.interfaces.PrioritySupplied;
import matrix.structures.CDT.probe.BinHeap;
import matrix.structures.CDT.probe.QueueImpl;
import matrix.structures.FDT.FDT;
import matrix.structures.FDT.LinkedList;
import matrix.structures.FDT.probe.Key;
import matrix.structures.FDT.probe.LinkedListImpl;
import matrix.structures.FDT.probe.Table;
import matrix.structures.memory.VirtualObject;
import matrix.structures.simulationextensions.DeleteEnabled;
import matrix.structures.simulationextensions.Selectable;
import matrix.structures.spatial.PaintingStyleDecorator;
import matrix.structures.spatial.FDT.probe.Point;
import matrix.structures.spatial.FDT.probe.PolyLine;
import matrix.structures.util.MatrixComparable;
import matrix.util.Note;
import content.exercises.sda.structures.GeometryFunctions;
import content.exercises.sda.structures.PointGenerator;
import content.exercises.sda.structures.SimpleStack;
import content.interfaces.ConfigureVisualType;
import content.interfaces.JudgeBlocks;
import content.interfaces.LayoutExercise;
import content.interfaces.ModelAnswerNames;
import content.interfaces.SimulationExerciseModel;
import content.interfaces.StyledExercise;
/**
 * Exercise testing the line sweep algorithm for finding line segment intersections. The implementation is
 * from Computational Geometry by de Berg, van Kreveld et al. As per book, the sweep line goes from the top
 * of the set to the bottom (in the direction of the y axis, in other words).
 * @author jtn
 *
 */
public class IntersectionsLineSweep implements SimulationExerciseModel, StyledExercise, ConfigureVisualType, ModelAnswerNames, JudgeBlocks, LayoutExercise{

	/**
	 * 
	 */
	private static final long serialVersionUID = -8563709064280996856L;
	// The structure used for creating the area visualization for the lines, points, etc.
	private SimpleStack problem, model;
	// the heap where to store the points handled by the algorithm
	private BinHeap heap, modelHeap;
	// the neighbour structure used: a table in this case.
	private Table neighbours, modelNeighbours;
	// lineMap contains a mapping from points to lines. Modified only during
	// initialization; therefore a HashMap can be used instead of
	// structures included in the animator.
	// Similarly pairMap maps the endpoints of a line for the solver, and found stores 
	// intersections found by the solver
	private HashMap lineMap, pairMap, found; 
	// the output structure.
	private LinkedList output, modelOutput;
	// input, stored for the model answer generator.
	private Point[] input;
	// for storing intersections used during the execution of the algorithm
	// OwnTable quick hack, since matrix structures tend not to have search 
	// implemented! 
	private OwnTable foundIntersections;
	// number of edges, maximum x and y coordinates, maximum number of iterations in
	// the initialization.
	private int nroEdges, maxX, maxY, maxIter;
	// random seed
	private long seed;
	// Random
	private Random rnd;
	// the sweep line is stored here during the execution of the algorithm.
	private VirtualObject sweepLine;
	
	// Minimum distance between points in the input
	public static final int MIN_POINT_DIST = 40;
	// Minimum distance from a point to a line (that its not part of)
	public static final int MIN_LINE_DIST = 10;
	// minimum distance between intersections
	public static final int MIN_INTER_DIST = 10;
	public static final int MIN_INTERSECTIONS = 2;
	// buffer around the area
	public static final int BUFFER = 5;
	
	// decorator for the sweep line.
	public static final PaintingStyleDecorator lineDecorator;
	public static final Color SWEEP_COLOR = Color.ORANGE;
	
	// used in debugging. If USE_STATIC is true the static seed value is used.
	public static final boolean DEBUG=true;
	public static final long STATIC_SEED = 1186490374807L;
	public static final boolean USE_STATIC=false;

	static {
	    float[] dash = { 5 , 5 };
		lineDecorator = new PaintingStyleDecorator();
        lineDecorator.setLineColor(SWEEP_COLOR);
        lineDecorator.setStroke(2, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL,5,dash,0);

	}
	
	public IntersectionsLineSweep() {
		nroEdges = 6;
		maxX = 300;
		maxY = 300;
		maxIter = 100;
		seed = 0;
	}
	
	public FDT[] solve() {
		FDT[] ret = makeModelAnswer();
		if(ret == null|| ret.length < 4) {
			return null;
		}
		// returns the heap, the neighbour structure and output structure.
		// the array is used for visualization and student interaction and
		// is not needed for grading
		return new FDT[] {ret[0], ret[2], ret[3]};
	}

	public FDT[] makeModelAnswer() {
		// create and initialize data structures.
		modelHeap = new BinHeap();
		modelNeighbours = new Table();
		modelNeighbours.setObject(null,0);
		modelNeighbours.setObject(null,nroEdges-1);
		modelOutput = new QueueImpl();
		model = new SimpleStack(new Rectangle(maxX+15, maxY+5,0,0));
		pairMap = new HashMap();
		found = new HashMap();
		// problem initialization
		for(int i = 0; i < nroEdges; i++) {
			model.ADTPush(input[i*2]);
			model.ADTPush(input[i*2+1]);
			modelHeap.insert(input[2*i]);
			modelHeap.insert(input[i*2+1]);
			pairMap.put(input[i*2], input[i*2+1]);
			pairMap.put(input[i*2+1], input[i*2]);
			PolyLine p = new PolyLine(new Point[] {input[i*2], input[i*2+1]}, "");
			model.ADTPush(p);
		}
		// create sweep line, required only for visualization.
		double[] sweepx = new double[2];
		double[] sweepy = new double[2];
		sweepx[0] = 0;
		sweepx[1] = maxX;
		sweepy[0] = sweepy[1] = maxY;
		PolyLine sweep = new PolyLine(sweepx, sweepy,"");
		sweep.setPaintingStyleDecorator(lineDecorator);
		model.ADTPush(sweep);
		Animator a = Animator.getActiveAnimator();
		// main loop for solving the problem
		while(!modelHeap.ADTIsEmpty()) {
			a.startOperation();
			PriorityPoint p = (PriorityPoint)modelHeap.ADTdeleteMin();
			// remove old sweep line from the area and prepare the new
			model.ADTPop();
			sweepy[0] = sweepy[1] = p.getY();
			sweep = new PolyLine(sweepx, sweepy,"");
			sweep.setPaintingStyleDecorator(lineDecorator);
			if(p instanceof Intersection) {
				handleIntersection((Intersection)p);
			} else {
				Point p1 = getPair(p);
				if(p1.getY() < p.getY() || (p1.getY() == p.getY() && p1.getX() > p.getX())) {
					handleFirstPoint(p);
				} else {
					handleSecondPoint(p);
				}
			}
			// put new sweepline to the area
			model.ADTPush(sweep);
			a.endOperation();
			
		}
		return new FDT[] {modelHeap, model, modelNeighbours, modelOutput};
	}

	/**
	 * Handles the second (lower) endpoint of a given line in the problem set.
	 * The corresponding line is removed from the neighbour structure, and the
	 * new neighbours thus created are tested for new intersections.
	 * @param p
	 */
	private void handleSecondPoint(PriorityPoint p) {
		PolyLine l = (PolyLine)lineMap.get(p);
		Point[] ps = l.getPoints();
		Point pp;
		if(ps[0] == p) {
			pp = ps[1];
		} else if(ps[1] == p) {
			pp = ps[0];
		} else {
			Note.err(this, "Line corresponding to second end point not found");
			return;
		}
		int i = removeFromTable(pp);
		if(i < 0) {
			Note.err(this, "Line corresponding to second endpoint not neighbour structure");
			return;
		}
		Point p1 = (Point)modelNeighbours.getObject(i);
		Point p2 = (Point)modelNeighbours.getObject(i-1);
		if(p1 != null && p2 != null) {
			checkNewIntersection(p1,p2,(int)p.getY(),(int)p.getX());
		}
	}

	
	/**
	 * Removes the parameter point from the modelNeighbours table, shifting all
	 * remaining objects in the table appropriately in order to not leave a hole in it.
	 * @param p
	 * @return index from which the point was removed, or -1 if the point wasn't in the table
	 */
	private int removeFromTable(Point p) {
		int count = modelNeighbours.getFirst();
		int last = modelNeighbours.getLast();
		while(count <= last) {
			Point c = (Point)modelNeighbours.getObject(count);
			if(c == p) {
				shiftLeft(count);
				return count;
			}
			count++;
		}
		return -1;
	}

	/**
	 * Shift the items from the indices larger than the parameter int to the 
	 * left in the table. Effectively deletes the parameter index from the table
	 * without leaving any empty indices.
	 * @param count
	 */
	private void shiftLeft(int count) {
		int last = modelNeighbours.getLast();
		while(count < last) {
			modelNeighbours.setObject(modelNeighbours.getObject(count+1), count);
			count++;
		}
		modelNeighbours.setObject(null,count);
	}
	
	/**
	 * Handles the first (higher) endpoint of a line. Adds the point to the
	 * appropriate position in the neighbour structure and checks if it 
	 * intersects either of its neighbouring lines.
	 * @param p
	 */
	private void handleFirstPoint(PriorityPoint p) {
		int x = (int)p.getX();
		int y = (int)p.getY();
		int i = addToTable(p);
		Point p1 = (Point)modelNeighbours.getObject(i-1);
		Point p2 = (Point)modelNeighbours.getObject(i+1);
		if(p1 != null) {
			checkNewIntersection(p,p1,y,x);
		}
		if(p2 != null) {
			checkNewIntersection(p,p2,y,x);
		}		
	}

	/**
	 * Adds the parameter point to the modelNeighbour table. Takes into
	 * account the current y coordinate of the sweep line and the 
	 * x coordinates (in the current y coordinate) of the lines already 
	 * in the table. 
	 * @param p
	 * @return index where the point was added
	 */
	private int addToTable(Point p) {
		int count = modelNeighbours.getFirst();
		int last = modelNeighbours.getLast()+1;
		while(count < last) {
			Point current = (Point)modelNeighbours.getObject(count);
			if(current == null) {
				modelNeighbours.setObject(p,count);
				return count;
			}
			if(currentXCoordinate(current,p.getY()) > p.getX()) {
				shiftRight(count);
				modelNeighbours.setObject(p,count);
				return count;
			}
			count++;
		}
		return -1;
	}
	
	/**
	 * Shifts the objects in the modelNeighbours to the right
	 * starting with the parameter index. In effect creates an
	 * "empty" index for a new object.
	 * @param count
	 */
	private void shiftRight(int count) {
		int last = modelNeighbours.getLast();
		Object current = modelNeighbours.getObject(count);
		while(count <= last) {
			Object tmp = modelNeighbours.getObject(count);
			modelNeighbours.setObject(current,count);
			current = tmp;
			count++;
		}
	}

	/**
	 * Calculates the x coordinate of the line defined
	 * by the parameter start in the current y coordinate, defined
	 * by the parameter  
	 * @param start
	 * @param y
	 * @return the value of the current x coordinate
	 */
	private double currentXCoordinate(Point start, double y) {
		PolyLine line = (PolyLine)lineMap.get(start);
		Point[] ps = line.getPoints();
		Point end;
		if(ps[0] == start) {
			end = ps[1];
		} else if(ps[1] == start){
			end = ps[0];
		} else {
			Note.err(this, "Error: point mismatch");
			return -1;
		}
		if(end.getX() == start.getX()) {
			return end.getX();
		} else if(end.getX() < start.getX()) {
			return getCoord(end, start, y);
		} else {
			return getCoord(start, end, y);
		}
	}

	/**
	 * Helper function that calculates the actual coordinate, using
	 * simple geometry (the y=ax+b line equation)
	 * @param start
	 * @param end
	 * @param y
	 * @return value of the x coordinate
	 */
	private double getCoord(Point start, Point end, double y) {
		double a = (end.getY() - start.getY()) / (end.getX() - start.getX());
		double b = end.getY() - a*end.getX(); 
		return (y-b)/a;
	}
	
	/**
	 * Handles an intersection. Swaps the two lines in the neighbour structure
	 * and checks if they make intersection with their new neighbours. 
	 * @param p
	 */
	private void handleIntersection(Intersection p) {
		((QueueImpl)modelOutput).ADTEnqueue(p);
		PolyLine[] l = p.getLines();
		Point p1, p2, p3, p4;
		p1 = getStartPoint(l[0]);
		p2 = getStartPoint(l[1]);
		// swap the points in the array
		// i and j are the indices of the points to be swapped
		int i, j, count, up;
		up = modelNeighbours.getLast();
		i = j = -1;		
		for(count = modelNeighbours.getFirst(); count <= up;count++) {
			if(modelNeighbours.getObject(count) == p1) {
				i = count;
			}
			if(modelNeighbours.getObject(count) == p2) {
				j = count;
			}
		}
		// Rest of the code assumes i < j 
		if(i > j) {
			int tmp1 = i;
			i = j;
			j = tmp1;
			Point tmp2 = p1;
			p1 = p2;
			p2 = tmp2;
		}
		// sanity checks
		if(i < 0 || j < 0) {
			Note.err(this, "Cannot swap points, one of them is missing from neighbour structure");
			return;
		}
		if(j-i != 1) {
			Note.err(this, "Intersecting lines not next to each other in the neighbour structure");
			return;
		}
		modelNeighbours.swap(i,j);
		// check for new intersections
		p3 = (Point)modelNeighbours.getObject(j+1);
		p4 = (Point)modelNeighbours.getObject(i-1);
		// p1 (in index i) and its new neighbour in index j+1
		if(p3 != null) {
			checkNewIntersection(p1,p3,(int)p.getY(), (int)p.getX());
		}
		if(p4 != null) {
			checkNewIntersection(p2,p4,(int)p.getY(), (int)p.getX());
		}
	}

	// checks if the lines defined by the two parameter points intersect
	// in a position that is below the current sweepline position
	private void checkNewIntersection(Point p1, Point p2, int y, int x) {
		PolyLine l1, l2;
		l1 = (PolyLine)lineMap.get(p1);
		l2 = (PolyLine)lineMap.get(p2);
		Point[] l1s, l2s;
		l1s = l1.getPoints();
		l2s = l2.getPoints();
		java.awt.Point inter = GeometryFunctions.segmentIntersection(l1s[0],l1s[1],l2s[0],l2s[1]);
		if(inter != null && !found.containsKey(inter) && (inter.getY() < y || inter.getY() == y && inter.getX() > x)) {
			found.put(inter,inter);
			Intersection i = new Intersection(inter, "("+p1.getVisualizationString()+","+p2.getVisualizationString()+")", l1,l2);
			model.ADTPush(i);
			modelHeap.ADTinsert(i);
		}		
	}
	
	// fetches the starting point for the given polyline.
	private Point getStartPoint(PolyLine l) {
		Point[] ps = l.getPoints();
		if(ps[0].getY() == ps[1].getY()) {
			if(ps[0].getX() < ps[1].getX()) {
				return ps[0];
			} else {
				return ps[1];
			}
		} else if(ps[0].getY() < ps[1].getY()) {
			return ps[1];
		}
		return ps[0];
	}
	
	private Point getPair(PriorityPoint p) {
		return (Point)pairMap.get(p);
	}

	public FDT[] init() {
		try {
			rnd = new Random(getSeed());
			if(DEBUG) {
				Note.out(this, "Seed is: " + getSeed());
			}
			boolean usable = false;
			int iters = 0;
			// initialize the instance variables used by this exercise.
			heap = new ExBinHeap();
			neighbours = new ExTable();
			neighbours.setObject(null,0);
			neighbours.setObject(null,nroEdges-1);
			output = new QueueImpl();
			sweepLine = new VirtualObject(null, null, "(Visualization) sweepline");
			foundIntersections = new OwnTable();
			// loop for creating the input.
			initloop: 
				while(!usable && iters < maxIter) { 
					iters++;
					usable = true;
					// generate the point set.
					java.awt.Point[] points = generatePoints();
					if(points == null) {
						if(DEBUG) {
							// Points are created using valid constraints, therefore
							// if this happens something's very wrong with the point generator.
							Note.out(this, "No point set generated");
						}
						usable = false;
						continue;
					}
					// the rectangle is bit larger than the actual area in order to create
					// a small buffer zone in order to, for example, have room for point labels etc.
					problem = new SimpleStack(new Rectangle(maxX+15, maxY+5,0,0));
					Vector intersections = new Vector();
					PolyLine[] lines = new PolyLine[nroEdges];
					input = new Point[points.length];
					lineMap = new HashMap();
					// generate input
					for(int i = 0; i < nroEdges;i++) {
						Point p1 = new PriorityPoint(points[i*2], ""+i*2);
						Point p2 = new PriorityPoint(points[i*2+1], ""+(i*2+1));
						PolyLine p = new PolyLine(new Point[] {p1,p2}, ""+i); 
						lines[i] = p;
						// add points and lines to the area, and map points to lines
						problem.ADTPush(p);
						problem.ADTPush(p1);
						problem.ADTPush(p2);
						input[i*2] = p1;
						input[i*2+1] = p2;
						lineMap.put(p1,p);
						lineMap.put(p2,p);
					}
					// check input, first count the number of intersections
					int intersectionNro = 0;
					for(int i = 0; i < lines.length; i++) {
						PolyLine p = lines[i];
						Point p1, p2;
						Point[] ps = p.getPoints();
						p1 = ps[0];
						p2 = ps[1];
						for(int j = i+1; j < lines.length; j++) {
							ps = lines[j].getPoints();
							Point p3, p4;
							p3 = ps[0];
							p4 = ps[1];
							java.awt.Point sect = GeometryFunctions.segmentIntersection(p1,p2,p3,p4);
							if(sect != null) {
								intersectionNro++;
								intersections.add(sect);
							}
						}
					}
					if(intersectionNro < MIN_INTERSECTIONS) {
						if(DEBUG) {
							Note.out(this, "Too few intersections");
						}
						usable = false;
						continue initloop;
					}
					// check that no point in the input is too close to
					// a line segment created by the other points.
					for(int i = 0; i < points.length; i++) {
						java.awt.Point curr = points[i];
						for(int j = 0; j < points.length; j = j+2) {
							if(i == j || i == j+1) {
								continue;
							}
							if(pointLineDistance(points[j], points[j+1], curr) < MIN_LINE_DIST) {
								if(DEBUG){
									Note.out(this, "Point too close to line, cannot work");
								}
								usable = false;
								continue initloop;								
							}
						}
						// Check that points are not too close to intersections
						for(int j = 0; j < intersections.size(); j++) {
							if(pointPointDistance(curr,(java.awt.Point)intersections.get(j)) < MIN_INTER_DIST) {
								if(DEBUG) {
									Note.out(this, "Points too close to intersection, cannot work");
								}
								usable = false;
								continue initloop;								
							}
						}
					}
					// check that intersections are not too close to each other.
					for(int i = 0; i < intersections.size(); i++) {
						java.awt.Point curr = (java.awt.Point)intersections.get(i);
						for(int j = 0; j < i; j++) {
							java.awt.Point p = (java.awt.Point)intersections.get(j);
							if(pointPointDistance(curr,p) < MIN_INTER_DIST) {
								if(DEBUG) {
									Note.out(this, "Intersections too close, cannot work");
								}
								usable = false;
								continue initloop;
							}
						}
					}
				}
			if(!usable) {
				Note.out(this, "No usable input gained");
				//XXX Here we should include some sort of static input creation!
				problem = new SimpleStack(new Rectangle(maxX+10, maxY+10,0,0));
				return new FDT[] {heap, problem, neighbours, output};				
			}
			// input checked, initialize
			for(int i = 0; i < input.length; i++) {
				heap.insert(input[i]);
			}
			// create initial position of the sweep line.
			double[] x = new double[2];
			double[] y = new double[2];
			x[0] = 0;
			x[1] = maxX;
			y[0] = y[1] = maxY;
			PolyLine line = new PolyLine(x,y,"");
			line.setPaintingStyleDecorator(lineDecorator);
			sweepLine.setObject(line);
			problem.ADTPush(line);
			return new FDT[] {heap, problem, neighbours, output};
		} catch(Exception e) {
			Note.err(this, "error: " + e.getMessage());
			e.printStackTrace();
		}
		return null;
	}

	// generates the input points. The area where they are generated is first divided into four sections
	// in order to get a bit more even distribution of points. Once all the points are generated, 
	// they are stored in an array that is then permuted. The order of the points in the array will
	// mark the point pairs that are used to generate the lines in the exercise.
	private java.awt.Point[] generatePoints() {
		java.awt.Point[] points;
		int edges = nroEdges;
		try {
			int x = maxX/2;
			int y = maxY/2;
			if(edges%2 == 1) {
				edges++;
			}
			// points are generated in four quarters of the area in order to get a bit more 
			// evenly distributed set of points.
			points = new java.awt.Point[2*edges];
			java.awt.Point[] p1,p2,p3,p4;
			p1 = PointGenerator.generatePoints(BUFFER,BUFFER, x-BUFFER,y-BUFFER,edges/2,MIN_POINT_DIST,rnd);
			p2 = PointGenerator.generatePoints(x+BUFFER,BUFFER, maxX,y-BUFFER,edges/2,MIN_POINT_DIST,rnd);
			p3 = PointGenerator.generatePoints(BUFFER,y+BUFFER, x-BUFFER,maxY,edges/2,MIN_POINT_DIST,rnd);
			p4 = PointGenerator.generatePoints(x+BUFFER,y+BUFFER, maxX,maxY,edges/2,MIN_POINT_DIST,rnd);
			int i = 0;
			for(int j = 0; j < p1.length;j++) {
				points[i] = p1[j];
				i++;
			}
			for(int j = 0; j < p2.length;j++) {
				points[i] = p2[j];
				i++;
			}
			for(int j = 0; j < p3.length;j++) {
				points[i] = p3[j];
				i++;
			}
			for(int j = 0; j < p4.length;j++) {
				points[i] = p4[j];
				i++;
			}
			//points = PointGenerator.generatePoints(BUFFER,BUFFER,maxX,maxY,2*edges,MIN_POINT_DIST,new Random());
		}catch(PointGenerator.IllegalConstraintsException e) {
			Note.out(this, "Illegal constraints, could not generate points");
			return null;
		}
		points = createPermutation(points, rnd);
		return points;
	}

	// creates a permutation for a given point array, using Knuth Shuffle.
	private java.awt.Point[] createPermutation(java.awt.Point[] points, Random rnd) {
		for(int i = 0; i < points.length; i++) {
			int r = rnd.nextInt(i+1);
			java.awt.Point swap = points[r];
			points[r] = points[i];
			points[i] = swap;
		}
		return points;
	}


	// Distance between points. Not really needed, since Point contains the same functionality.
	private double pointPointDistance(java.awt.Point p1, java.awt.Point p2) {
		return Math.sqrt((p1.getX()-p2.getX())*(p1.getX()-p2.getX()) + (p1.getY()-p2.getY())*(p1.getY()-p2.getY()));
	}
	
	// The distance of point p from the line determined by points line1 and line2
	// defined at http://mathworld.wolfram.com/Point-LineDistance2-Dimensional.html
	private double pointLineDistance(java.awt.Point line1, java.awt.Point line2, java.awt.Point p) {
		double t1 = line2.getX()-line1.getX();
		double t2 = line2.getY()-line1.getY();
		return Math.abs((t1)*(line1.getY()-p.getY()) - (line1.getX()-p.getX())*t2) / Math.sqrt((t1*t1 + t2*t2));
	}

	// Adds an intersection to the area visualization, if the two lines intersect, and the intersection has not been found before.
	// the two parameter points will determine the lines for which the intersection is tested.
	private void addIntersection(Point p1, Point p2) {
		if(p1 == null || p2 == null) {
			return;
		}
		PolyLine line1 = (PolyLine)lineMap.get(p1);
		PolyLine line2 = (PolyLine)lineMap.get(p2);
		if(line1 == null || line2 == null || line1 == line2) {
			return;
		}
		Point[] lp1 = line1.getPoints();
		Point[] lp2 = line2.getPoints();
		java.awt.Point inter = GeometryFunctions.segmentIntersection(lp1[0],lp1[1],lp2[0],lp2[1]);
		if(inter == null) {
			return;
		}
		
		Intersection i = new Intersection(inter,"("+p1.getVisualizationString()+","+p2.getVisualizationString()+")",line1,line2);
		// the intersection has already been found previously
		if(foundIntersections.contains(i)) {
			return;
		}
		foundIntersections.add(i);
		if(problem.ADTPeek() == sweepLine.getObject()) {
			Object o = problem.ADTPop();
			problem.ADTPush(i);
			problem.ADTPush(o);
		} else {
			problem.ADTPush(i);
		}
		
	}
	
	public void setSeed(long seed) {
		this.seed = seed;
	}

	public long getSeed() {
		if(USE_STATIC) {
			return STATIC_SEED;
		}
		return seed;
	}

	public String[] getStructureNames() {
		String[] s = { "Priority queue",  "Line segments", "adjacency structure", "output" };
		return s;
	}

	public FDT[] getAnswer() {
		return new FDT[] {heap, neighbours, output};
	}

	public FDT[] getInitialStructures() {
		return new FDT[] { new ExBinHeap(), new SimpleStack(), new ExTable(), new LinkedListImpl() };
	}

	public String getDescription() {
		return "Find line intersections using line sweep.";
	}

	public String[] getStructureVisualisations() {
		return new String[] {"layered tree", "2D", "array", "list"};
	}

	public String[] getModelAnswerVisualisations() {
		return getStructureVisualisations();
	}

	public VisualTypeConf[] conf() {
		VisualTypeConf heap = new VisualTypeConf();
		heap.enable("matrix.visual.VisualLayeredTree", VisualTypeConf.DROP_OPERATION);
		heap.enable("matrix.visual.VisualLayeredTree", VisualTypeConf.HIGHLIGHT_OPERATION);
		heap.enable("matrix.visual.VisualLayeredTreeComponent", VisualTypeConf.HIGHLIGHT_OPERATION);
		heap.enable("matrix.visual.VisualStruct", VisualTypeConf.HIGHLIGHT_OPERATION);
		heap.enable("matrix.visual.VisualStruct", VisualTypeConf.DRAG_OPERATION);
		heap.enable("matrix.visual.VisualStruct", VisualTypeConf.DROP_OPERATION);

		VisualTypeConf area = new VisualTypeConf();
		area.enable("matrix.visual.VisualArea$AreaVisualization", VisualTypeConf.HIGHLIGHT_OPERATION);
		area.enable("matrix.visual.VisualArea$AreaVisualization", VisualTypeConf.DRAG_OPERATION);
		VisualTypeConf neighbour = new VisualTypeConf();
		neighbour.enable("matrix.visual.VisualStruct", VisualTypeConf.HIGHLIGHT_OPERATION);
		neighbour.enable("matrix.visual.VisualStruct", VisualTypeConf.DRAG_OPERATION);
		//neighbour.enable("matrix.visual.VisualKey", VisualTypeConf.DROP_OPERATION);
		neighbour.enable("matrix.visual.VisualArrayComponent", VisualTypeConf.HIGHLIGHT_OPERATION);
		//neighbour.enable("matrix.visual.VisualListComponent", VisualTypeConf.DRAG_OPERATION);
		neighbour.enable("matrix.visual.VisualArrayComponent", VisualTypeConf.DROP_OPERATION);
		//neighbour.enable("matrix.visual.VisualArray", VisualTypeConf.DROP_OPERATION);
		//neighbour.enable("matrix.visual.VisualArray", VisualTypeConf.HIGHLIGHT_OPERATION);
		//neighbour.setVisualConfValue("matrix.visual.VisualList", VisualTypeConf.ROTATED, "true");
		VisualTypeConf cuts = new VisualTypeConf();
		cuts.enable("matrix.visual.VisualList", VisualTypeConf.DROP_OPERATION);
		cuts.enable("matrix.visual.VisualList", VisualTypeConf.HIGHLIGHT_OPERATION);
		cuts.enable("matrix.visual.VisualKey", VisualTypeConf.DROP_OPERATION);
		return new VisualTypeConf[] {heap,area,neighbour,cuts};
	}

	public String[] getModelAnswerNames() {
		return getStructureNames();
	}

	public GridBagConstraints[] getConstraints() {
		Insets i = new Insets(10, 10, 10, 10);
		GridBagConstraints gbc[] = new GridBagConstraints[4];
		GridBagConstraints c = new GridBagConstraints();
		c.insets = i;
		c.weightx = 1.0;
		gbc[0] = c;
		gbc[2] = c;
		c = new GridBagConstraints();
		c.insets = i;
		c.weightx = 1.0;
		c.gridwidth = GridBagConstraints.REMAINDER;
		gbc[1] = c;
		gbc[3] = c;
 		return gbc;
	}

	public GridBagConstraints[] getModelAnswerConstraints() {
		return getConstraints();
	}
	
	// The subclass of point used in this exercise. Can be dragged from the area, but not selected, has a priority, and
	// creates the priorities in a manner that will give those keys with high y coordinate
	// a high priority. If two points have equal y coordinates, the one with smaller x 
	// coordinate has higher priority.
	public class PriorityPoint extends Point implements PrioritySupplied, Selectable {

		/**
		 * 
		 */
		private static final long serialVersionUID = 7279204502134059511L;
		private Key priority;
		private long stamp;
		
		public PriorityPoint(Point2D point, String s) {
			super(point,s);
			setDraggable(true);
			int tmp = maxY - (int)point.getY();
			String k = createString(tmp);
			tmp = (int)point.getX();
			k += createString(tmp);
			priority = new Key(k);
			stamp = Selectable.NOT_SELECTED;
		}

		private String createString(int i) {
			if(i < 10) {
				return "00"+i;
			} else if(i < 100) {
				return "0"+i;
			}
			return ""+i;
		}
		
		public void setPriorityElement(MatrixComparable newPriority) {
		}

		public MatrixComparable getPriorityElement() {
			return priority;
		}

		public void setSelect(long globalClockTimestamp) {
			//stamp = globalClockTimestamp;			
		}

		public long getSelectionTime() {
			return stamp;
		}
		
	}
	
	// Encapsulates an intersection. Has a different label and contains information
	// about the lines that intersect. Equals-method works only with other
	// instances of Intersection.
	public class Intersection extends PriorityPoint {
		/**
		 * 
		 */
		private static final long serialVersionUID = -3698909099692893256L;
		private PolyLine[] l;
		
		public Intersection(Point2D point, String s, PolyLine p1, PolyLine p2) {
			super(point,s);
			l = new PolyLine[2];
			l[0] = p1;
			l[1] = p2;
		}

		public boolean equals(Object o) {
			if(o instanceof Intersection) {
				Intersection i = (Intersection)o;
				PolyLine[] lines = i.getLines();
				return ((l[0] == lines[0] && l[1] == lines[1])||(l[0]==lines[1] && l[1]==lines[0]));
			}
			return false;
		}
		
		public PolyLine[] getLines() {
			return l;
		}
		
	}
    
    public class ExBinHeap extends BinHeap implements DeleteEnabled {

        /**
		 * 
		 */
		private static final long serialVersionUID = 2736173988925124065L;

		public boolean shouldDelete() {
            return true;
        }
        
        // delete tweaked to move the sweep line to the y coordinate
        // of the point that is deleted.
        public CDT delete(Object o) {
        	Point p = (Point)getObject(0);
        	double[] x = new double[2];
        	double[] y = new double[2];
        	x[0] = 0;
        	x[1] = maxX;
        	y[0] = y[1] = p.getY();
        	if(problem.ADTPeek() == sweepLine.getObject()) {
        		PolyLine line = new PolyLine(x,y,"");
        		line.setPaintingStyleDecorator(lineDecorator);
        		Animator a = Animator.getActiveAnimator();
        		a.startOperation();
        		problem.ADTPop();
        		sweepLine.setObject(line);
        		problem.ADTPush(line);
        		a.endOperation();
        	}
        	return super.delete(o);
        }
        
    }

    public class ExTable extends Table implements DeleteEnabled, CDT {

        /**
		 * 
		 */
		private static final long serialVersionUID = -5029619654125194032L;

		public boolean shouldDelete() {
            return true;
        }

        // insert-method adds the parameter object to the position 0
        // in the array. It will also test whether the line 
        // associated with the inserted point will intersect
        // with the line associated with the point previously in 
        // index 0.
        public CDT insert(Object o) {
            int currIndex = getFirst();
            int lastIndex = getLast();
            Object curr = o;
            Object comp = getObject(currIndex);
            Object tmp;
            while(currIndex <= lastIndex) {
                tmp = getObject(currIndex);
                setObject2(curr, currIndex);
                curr = tmp;
                currIndex++;
            }
            addIntersection((Point)o, (Point)comp);
            return this;
        }

        public void setObject2(Object o, int i) {
        	super.setObject(o,i);
        }
        
        // Setobject will now swap contents of two indices, if
        // the method is called for an object that is already 
        // contained in the array. For new objects, it will
        // move contents of the array to the right in order to
        // make room. Also, intersections are tested for.
        public void setObject(Object o, int i) {
        	if(o == null) {
        		setObject2(o,i);
        	}
        	int curr = getFirst();
        	int last = getLast();
        	// check if the current object is already in the
        	// table, if it is, swap
        	while(curr <= last) {
        		if(getObject(curr) == o) {
        			swap(i, curr);
        			addIntersections(i);
        			addIntersections(curr);
        			return;
        		}
        		curr++;
        	}
        	// if the execution reaches here, the object
        	// is new. Move the objects after it to the right
        	// in the array to make room.
        	Object next = o;
        	curr = i;
        	while(next != null) {
        		Object tmp = getObject(curr);            	
            	setObject2(next,curr);
            	next = tmp;
            	curr++;
        	}
        	addIntersections(i);
        }

        private void addIntersections(int i) {
			Point p = (Point)getObject(i);
			Point p1 = (Point)getObject(i-1);
			Point p2 = (Point)getObject(i+1);
			addIntersection(p,p1);
			addIntersection(p,p2);        	
        }
        
        // deletes the parameter object from the array. Other objects in the
        // array are shifted to the left, so that no empty indices are left
        // between objects. Also, intersections between the neighbours of the
        // deleted object are tested.
        public CDT delete(Object o) {
        	int curr = getFirst();
        	int last = getLast();
        	Object obj = getObject(curr);
        	while(obj != o && curr < last) {
        		curr++;
        		obj = getObject(curr);
        	}
        	if(curr == last) {
        		setObject2(null,curr);
        		return this;
        	}
        	int index = curr;
        	while(curr < last) {
        		setObject2(getObject(curr+1),curr);
        		curr++;
        	}
        	setObject2(null, curr);
        	Point p1 = (Point)getObject(index);
        	Point p2 = (Point)getObject(index-1);
        	addIntersection(p1,p2);
        	return this;
        }

        public Object search(Object o) {
            // not needed
            return null;
        }

        public CDT getNewInstance() {
            // not needed
            return null;
        }
        
    }
    
    // helper class for storing the intersections found in the algorithm.
    public class OwnTable extends Table {
    	
    	private int size;
    	
    	public OwnTable() {
    		super();
    		size = 0;
    	}

    	public void add(Object o) {
    		setObject(o,size);
    		size++;
    	}
    	
    	public boolean contains(Object o) {
    		if(o == null) {
    			return false;
    		}
    		for(int i=0; i < size; i++) {
    			if(o.equals(getObject(i))) {
    				return true;
    			}
    		}
    		return false;
    	}
    }
    
}
