package content.exercises.sda;

import matrix.simulation.VisualTypeConf;
import matrix.structures.FDT.FDT;
import matrix.structures.FDT.LinkedList;
import content.exercises.sda.structures.ExerciseGeometricGraph;
import content.exercises.sda.structures.ExerciseGeometricGraphVertex;
import content.interfaces.*;

import matrix.structures.FDT.probe.Key;
import matrix.structures.CDT.CDT;
import matrix.structures.CDT.probe.BinHeap;
import matrix.structures.CDT.probe.QueueImpl;
import matrix.structures.FDT.probe.LinkedListImpl;
import matrix.structures.FDT.substructures.Vertex;
import matrix.structures.simulationextensions.DeleteEnabled;
import matrix.decoration.StyleSheet;
import matrix.decoration.StyleSheetAdapter;
import matrix.decoration.Styled;

import matrix.animation.Animator;

import matrix.util.Note;

import java.awt.GridBagConstraints;
import java.util.Random;
import java.util.Vector;
import java.util.HashMap;
import java.awt.Insets;
import java.io.Serializable;


/**
 * Line sweep for finding line segment intersections. This exercise is based on a simplified version of the algorithm described
 * in Berg et. al, Computational Geometry: Algorithms and Applications: 20-27. The implemented algorithm cannot
 * handle cases where more than two line segments intersect in a given point. 
 * @author jtn
 *
 */
public class LineSweep implements SimulationExerciseModel, StyledExercise, ConfigureVisualType, ModelAnswerNames, JudgeBlocks, LayoutExercise {

	/**
	 * 
	 */
	private static final long serialVersionUID = -4378498028473158549L;
	private MyGraph graph, modelGraph;
	//private GraphAdapter wrapper;
	private BinHeap heap, modelHeap;
	private LinkedList neighbours, neighbourModel;
	private LinkedList cuts, modelCuts;
	private Point[] input;
	private int nroEdges, minIntersections, minDist, maxIter;
	private float maxX, maxY;
	private HashMap found;
	
	//private VirtualInteger counter;
	
	
	public LineSweep() {
		nroEdges = 6;
		minIntersections = 2;
		maxX = 300;
		maxY = 200;
		minDist = 15; // minimum distance between line end points
		maxIter = 10000;
	}
	
	public FDT[] solve() {
		//VirtualInteger tmpCounter = counter;
		//counter = new VirtualInteger(nroEdges*2-1);
		modelHeap = new BinHeap();
		modelGraph = new MyGraph();
		neighbourModel = new QueueImpl();
		modelCuts = new QueueImpl();
		found = new HashMap();
		boolean add = true;
		for(int i = 0; i < input.length;i++) {
			Point p = input[i];
			Key k = makeKey(p);
			modelHeap.insert(k);
		}
		Animator a = Animator.getActiveAnimator();
		while(!modelHeap.ADTIsEmpty()) {
			a.startOperation();
			Key k = (Key)modelHeap.deleteMin();
			Point p = (Point)k.getObject();
			Point pair = p.getPair();
			// Intersections always have id -1
			//if(p.getId() == -1) {
			if(p instanceof Intersection) {
				intersectionPoint(k);
			} else if(p.getX() <= pair.getX()) { // sweep in the direction of positive X-axis
				firstPoint(k);
			} else {
				secondPoint(k);
			}
			a.endOperation();
		}
		//counter = tmpCounter;
		FDT[] f = { modelHeap, neighbourModel, modelCuts };		
		return f;
	}

	/**
	 * Handles a given Point value in the solve method if this point is an
	 * intersection point The point is wrapped inside the parameter Key value.
	 * 
	 * At an intersection point the two line segments that intersect are swapped in the
	 * neighbour structure. After the swap new intersections are checked between the 
	 * line segments and their new neighbours.
	 * 
	 * @param k the Key containing a Point
	 */
	private void intersectionPoint(Key k) {
		Intersection i = (Intersection)k.getObject();
		((QueueImpl)modelCuts).insert(k);
		// Check if this indeed is an intersection point!
		Point first = i.getFirst();
		Point second = i.getSecond();
		LinkedList current = neighbourModel;
		LinkedList prev = null;
		while(current != null) {
			Key currentKey = (Key)current.getElement();
			Point currentPoint = (Point)currentKey.getObject();
			if(currentPoint == first || currentPoint == second) {
				// Correct position found, swap current and next list element
				LinkedList next = current.getNext();
				Key nextKey = (Key)next.getElement();
				// There should always be a next element, since the condition is true when first of the two lines is found 
				//therefore no null-checking is required. 
				current.setElement(nextKey);
				next.setElement(currentKey);
				//if(current == neighbourModel.getNext()) {
				//	neighbourModel.setNext(next);
				//}
				if(prev != null) {
					//prev.setNext(next);
					//now compare the points in prev and next to see if they intersect				
					Key tmp = (Key)prev.getElement();
					Point prevPoint = (Point)tmp.getObject();
					tmp = (Key)current.getElement();
					Point currPoint = (Point)tmp.getObject();
					Point inter = intersection(prevPoint,currPoint);
					String vs = getVisualizationString(inter);
					if(inter != null && inter.getX() > i.getX()&& !found.containsKey(vs)) { // if new intersection is found, add it to heap
						found.put(vs,vs);
						addPoint(inter,modelGraph);
						modelHeap.insert(makeKey(inter));
					}
				}
				// compare current to its successor
				LinkedList nextnext = next.getNext();
				if(nextnext != null) {
					Key tmp = (Key)nextnext.getElement();
					Point nextnextPoint = (Point)tmp.getObject();
					Point inter = intersection(currentPoint, nextnextPoint);
					String vs = getVisualizationString(inter);
					if(inter != null && inter.getX() > i.getX()&& !found.containsKey(vs)) { // if new intersection is found, add it to heap
						found.put(vs,vs);
						addPoint(inter, modelGraph);
						modelHeap.insert(makeKey(inter));
					}
				}
				return;
			}
			prev = current;
			current = current.getNext();
		}
	}
	
	/**
	 *  Handles a given Point value in the solve method if this point's X coordinate is
	 *  smaller than its pair's. The point is wrapped inside the parameter Key value.
	 *  
	 *  When a new line segment is encountered, it is added to the correct position in the 
	 *  neighbour structure. After this new intersections are checked between the new line
	 *  segment and its neighbours.
	 *  
	 * @param k a Key containing a Point
	 */
	private void firstPoint(Key k) {
		Point p = (Point)k.getObject();
		Point pair = p.getPair();
		addPoint(p, modelGraph);
		LinkedList current = neighbourModel;
		LinkedList prev = null;
		LinkedList next = null;
		// special case, if the neighbourModel is empty
		if(current.getElement() == null) {
			current.setElement(k);
			return;
		}
		// The elements in the neighbourModel are arranged in 
		// increasing order of Y-coordinate. Therefore, we 
		// iterate through the elements until we find an element which
		// has larger Y-coordinate than the current element.
		while(current != null) {
			Key currentKey = (Key)current.getElement();
			Point currentPoint = (Point)currentKey.getObject();
			float y = currentYCoordinate(currentPoint, p.getX());
			if(y > p.getY()) {
				next = new LinkedListImpl(currentKey);
				current.setElement(k);
				next.setNext(current.getNext());
				current.setNext(next);
				break;
			}
			prev = current;
			current = (LinkedListImpl)current.getNext();
		}
		Key nextKey, previousKey;
		Point previousPoint, nextPoint;
		// if current is null, we have arrived at the end of the list
		if(current == null) {
			current = new LinkedListImpl();
			current.setElement(k);
			prev.setNext(current);
			previousKey = (Key)prev.getElement();
			previousPoint = (Point)previousKey.getObject();
			Point inter = intersection(p, previousPoint);
			if(inter != null) {
				String vs = getVisualizationString(inter);
				found.put(vs,vs);
				addPoint(inter, modelGraph);
				modelHeap.insert(makeKey(inter));
			}
		} else {
			// now current holds the new Point, prev its smaller neighbour and next its larger neighbour
			// Now we check for intersections and add those points to the heap
			nextKey = (Key)next.getElement();
			nextPoint = (Point)nextKey.getObject();
			Point inter = intersection(p, nextPoint);
			if(inter != null) {
				String vs = getVisualizationString(inter);
				found.put(vs,vs);
				addPoint(inter, modelGraph);
				modelHeap.insert(makeKey(inter));
			}
			inter = null;
			if(prev != null) {
				previousKey = (Key)prev.getElement();
				previousPoint = (Point)previousKey.getObject();
				inter = intersection(p, previousPoint);
				if(inter != null) {
					String vs = getVisualizationString(inter);
					found.put(vs,vs);
					addPoint(inter, modelGraph);
					modelHeap.insert(makeKey(inter));
				}
			}
		}
	}

	/** 
	 * Handles a given Point value in the solve method if this point's X coordinate is
	 * larger than its pair's. The point is wrapped inside the paremeter Key value.
	 * 
	 * When the second endpoint of a line segment is encountered, the line segment is removed from
	 * the neighbour structure. After deletion, it is checked whether the deleted item's neighbours
	 * intersect, and if a new intersection is found, it is added to the stack.
	 * 
	 * @param k a Key containing a Point
	 */
	private void secondPoint(Key k) {
		Point p = (Point)k.getObject();
		LinkedList current = neighbourModel;
		LinkedList prev = null;
		while(current != null) {
			Key currentKey = (Key)current.getElement();
			Point currentPoint = (Point)currentKey.getObject();
			currentPoint = currentPoint.getPair();
			if(currentPoint == p) {
				LinkedList next = current.getNext();
				if(prev != null) {
					prev.setNext(next);
					// if next is not null, see if prev and next intersect
					if(next != null) {
						Key prevKey = (Key)prev.getElement();
						Key nextKey = (Key)next.getElement();
						Point prevPoint = (Point)prevKey.getObject();
						Point nextPoint = (Point)nextKey.getObject();
						Point inter = intersection(prevPoint, nextPoint);
						String vs = getVisualizationString(inter);
						if(inter != null && inter.getX() > p.getX() && !found.containsKey(vs)) {
							found.put(vs,vs);
							addPoint(inter, modelGraph);
							modelHeap.insert(makeKey(inter));
						}
					}
					return;
				} else {
					if(next != null) {
						neighbourModel.setElement(next.getElement());
						neighbourModel.setNext(next.getNext());
						//neighbourModel = next;
					} else {
						neighbourModel.setElement(null);
					}
					return;
				}
			}
			prev = current;
			current = current.getNext();
		}
	}
	
	/**
	 *  Adds the parameter point and its possible pair to the graph. An intersection point has no pair and
	 *  therefore no new line segment is added in that case.
	 *  
	 * @param p the Point to be added
	 * @return the int id of the point added. If two points are added, the id of the first is returned
	 */
	private void addPoint(Point p, MyGraph gg) {
		if(p instanceof Intersection) {
		//if(p.getId() == -1) {
			//int vid = gg.getNextId();
			gg.addVertex((int)p.getX(), (int)p.getY(), p);
			return;
		} else {
			Point pair = p.getPair();
			//int id = gg.getNextId();
			//int pid = gg.getNextId();
			MyVertex v1 = (MyVertex)gg.addVertex((int)p.getX(), (int)p.getY(), p);
			MyVertex v2 = (MyVertex)gg.addVertex((int)pair.getX(), (int)pair.getY(), pair);
			v1.addSuccessor(v2);
			v2.addSuccessor(v1);
			//v1.setReferenceInformation(0, "ID "+p.getId()/2);
			//v2.setReferenceInformation(0, "ID "+p.getId()/2);
			return;
		}
	}
		
	public FDT[] makeModelAnswer() {
		solve();
		FDT[] f = {modelHeap, neighbourModel, modelGraph, modelCuts };
		return f;
	}

	public String[] getModelAnswerNames() {
		String[] s = { "Priority queue",  "adjacency structure", "Line segments", "output" };
		return s;
	}

	public FDT[] init() {
		//Random r = new Random(getSeed());
		Random r = new Random();
		heap = new OwnBinHeap();
		graph = new MyGraph();
		neighbours = new OwnLinkedList();
		cuts = new QueueImpl();
		input = new Point[nroEdges*2];
		//counter = new VirtualInteger(-1);
		Vector intersections = new Vector();
		int nroIntersections = 0;
		int iters = 0;
		// input generation
		initloop:
		/* This loop wil create the given input. Loop will exit when a good input
		 * (one which has enough intersections) is found or if more than maxIter
		 * iterations have been completed. 
		*/
		while(nroIntersections <= minIntersections && iters < maxIter) {
			//counter.assign(-1);
			iters++;
			//input.clear(); // drop all previous elements
			//Note.out(this, "initlooping");
			nroIntersections = 0;
			for(int i = 0; i < nroEdges; i++) { // generate new edges
				Point a, b;
				int ida, idb, x1, x2, y1, y2;
				ida = 2*i;
				idb = 2*i+1;
				//ida = counter.inc();
				//idb = counter.inc();
				x1 = r.nextInt((int)maxX);
				do{
					x2 = r.nextInt((int)maxX);
				} while(Math.abs(x1-x2) < minDist);
				y1 = r.nextInt((int)maxY);
				do{
					y2 = r.nextInt((int)maxX);
				} while(Math.abs(y1-y2) < minDist);
				a = new Point(ida, x1, y1);
				b = new Point(idb, x2,y2);
				a.setPair(b);
				b.setPair(a);
				input[2*i] = a;
				input[2*i+1] = b;
				//addPoint(a, graph);
			}
			intersections.clear();

			// Check input if it can be accepted. This is a stupid O(n^2) check where each edge in turn is
			// compared to each other edge.
			for(int i = 0; i < nroEdges; i++) { 
				Point a = input[2*i];
				Point b = input[2*i+1];
				for(int j = 0; j < nroEdges; j++) {
					Point comp = input[2*j];
					if(a == comp) { // the point we're currently comparing to
						continue;
					}
					Point cpair = comp.getPair();
					if(a.pointEqual(comp) || a.pointEqual(cpair) || b.pointEqual(comp) || b.pointEqual(cpair)) {
						// two points in same position, cannot work. Discard input
						nroIntersections = 0;
						continue initloop;
					}
					if(j < i) {
						Point intersection = intersection(a, comp);
						if(intersection == null) { // check for parallel lines
							if(pointDistance(a, b, comp) == 0  
									&& determinant(a.getX()-b.getX(), a.getY()-b.getY(), comp.getX()- cpair.getX(), 
											comp.getY() - cpair.getY()) == 0) {
								// Parallel lines that cross each other, cannot work. Discard input
								nroIntersections = 0;
								continue initloop;
							}
						} else {
							nroIntersections++;
							intersections.add(intersection);
						}
					}

					if(pointDistance(comp,cpair, a) < minDist || pointDistance(comp, cpair, b) < minDist) {
						// too close to line created by other lines, discard
						nroIntersections = 0;
						continue initloop;
					}
					
				}
			}
			if(nroIntersections < minIntersections) {
				nroIntersections = 0;
				continue;
			}
			// Check that intersections are not too close to each other

			for(int i = 0; i < intersections.size(); i++) {
				Point a = (Point)intersections.get(i);
				for(int j = 0; j < intersections.size(); j++) {
					Point b = (Point)intersections.get(j);
					if(a == b) {
						continue;
					}
					if(!a.equals(b) && (Math.abs(a.getX()-b.getX()) < minDist/2 || Math.abs(a.getY()-b.getY()) < minDist/2)) {
						nroIntersections = 0;
						continue initloop;
					}
				}
			}
			
			
		}
		// If maximum number of iterations has been reached, return dummy exercise.
		// XXX This must be replaced with pre-generated exercise. 
		if(iters >= maxIter) {
			return new FDT[] {heap, neighbours, graph, cuts};
		}
		// when we've reached here, valid input has been gained, and we can initialize the data structures
		for(int i = 0; i < input.length; i++) {
			Point p = input[i];
			Point pair = p.getPair();
			Key k = makeKey(p);
			heap.insert(k);
			if(p.getId() < pair.getId()) {
				addPoint(p, graph);
			}
		}
		Key k = (Key)heap.getObject(0); // get topmost object from heap
		Point top = (Point)k.getObject();
		FDT[] s = {heap, neighbours, graph, cuts };
		//FDT[] s = {heap, neighbours, cuts };
		return s;
	}

	/**
	 *  Creates a Key object which has a correct comparison string from the given Point parameter. 
	 *  The comparison string is the X coordinate of the Point with enough zeroes before the real value
	 *  to make the string comparison of the coordinate to give same value than int comparison. 
	 *  This is done in order to make BinHeap implementation work correctly.
	 * @param p the Point for which the Key value is to be constructed
	 * @return a Key that contains the parameter Point
	 */
	private Key makeKey(Point p) {
		String cmp = getComparisonString(p);
		Key k = new Key(p, cmp);
		cmp = getVisualizationString(p);
		k.setVisualizationString(cmp);
		return k;
	}

	// creates comparison string for a key in heap
	private String getComparisonString(Point p) {
		String max = "" + maxX;
		String cmp = "" + p.getX();
		cmp = cmp.substring(0, cmp.indexOf('.')+2);
		while(cmp.length() < max.length()) {
			cmp = "0" + cmp;
		}
		return cmp;
	}

	// creates visualization string for a key in the heap.
	private String getVisualizationString(Point p) {
		if(p == null) {
			return "";
		}
		String vis = "";
		if(p instanceof Intersection) {
		//if(p.getId() == -1) { // intersection
			Intersection i = (Intersection)p;
			if(i.getFirst().getX() < i.getSecond().getX()) {
				vis = "" + (i.getFirst().getId()) + "," + (i.getSecond().getId());
			} else {
				vis = "" + (i.getSecond().getId()) + "," + (i.getFirst().getId());
			}
		} else { // line section endpoint
			vis = ""+p.getId();
			//Point first = p;
			//Point second = p.getPair();
			//vis = "ID " + (first.getId() / 2);
		}
		return vis;
	}
	
	public void setSeed(long seed) {
		
	}

	public long getSeed() {
		return 0;
	}

	public String[] getStructureNames() {
		return getModelAnswerNames();
	}

	public FDT[] getAnswer() {
		FDT[] f = {heap, neighbours, cuts };
		return f;
	}

	public FDT[] getInitialStructures() {
		FDT[] s = { new BinHeap(), new QueueImpl(), new MyGraph(), new QueueImpl() }; 
		//FDT[] s = { new BinHeap(), new QueueImpl(), new QueueImpl() };
		return s;
	}

	public String getDescription() {
		return "this is a test for finding line intersections using line sweep.";
	}

	public VisualTypeConf[] conf() {
		VisualTypeConf heap = new VisualTypeConf();
		heap.enable("matrix.visual.VisualLayeredTree", VisualTypeConf.DROP_OPERATION);
		heap.enable("matrix.visual.VisualLayeredTree", VisualTypeConf.HIGHLIGHT_OPERATION);
		heap.enable("matrix.visual.VisualKey", VisualTypeConf.HIGHLIGHT_OPERATION);
		heap.enable("matrix.visual.VisualKey", VisualTypeConf.DRAG_OPERATION);
		heap.enable("matrix.visual.VisualKey", VisualTypeConf.DROP_OPERATION);
		VisualTypeConf graph = new VisualTypeConf();
		graph.setVisualConfValue("matrix.visual.VisualGeometricGraph", matrix.visual.VisualGraph.DIRECTED, "false");
		graph.setVisualConfValue("matrix.visual.VisualGeometricGraph", VisualTypeConf.FLIPPEDY, "true");
		//graph.enable("matrix.visual.VisualReference", VisualTypeConf.HIGHLIGHT_OPERATION);
		graph.enable("matrix.visual.VisualKey", VisualTypeConf.HIGHLIGHT_OPERATION);
		graph.enable("matrix.visual.VisualKey", VisualTypeConf.DRAG_OPERATION);
		VisualTypeConf neighbour = new VisualTypeConf();
		neighbour.enable("matrix.visual.VisualKey", VisualTypeConf.HIGHLIGHT_OPERATION);
		neighbour.enable("matrix.visual.VisualKey", VisualTypeConf.DRAG_OPERATION);
		//neighbour.enable("matrix.visual.VisualKey", VisualTypeConf.DROP_OPERATION);
		neighbour.enable("matrix.visual.VisualListComponent", VisualTypeConf.HIGHLIGHT_OPERATION);
		//neighbour.enable("matrix.visual.VisualListComponent", VisualTypeConf.DRAG_OPERATION);
		neighbour.enable("matrix.visual.VisualListComponent", VisualTypeConf.DROP_OPERATION);
		neighbour.enable("matrix.visual.VisualList", VisualTypeConf.DROP_OPERATION);
		neighbour.enable("matrix.visual.VisualList", VisualTypeConf.HIGHLIGHT_OPERATION);
		neighbour.setVisualConfValue("matrix.visual.VisualList", VisualTypeConf.ROTATED, "true");
		VisualTypeConf cuts = new VisualTypeConf();
		cuts.enable("matrix.visual.VisualList", VisualTypeConf.DROP_OPERATION);
		cuts.enable("matrix.visual.VisualList", VisualTypeConf.HIGHLIGHT_OPERATION);
		cuts.enable("matrix.visual.VisualKey", VisualTypeConf.DROP_OPERATION);
		VisualTypeConf[] conf = {heap, neighbour, graph, cuts };
		return conf;
	}

	
	/**
	 *  Calculates the y coordinate of the parameter line segment in the parameter x coordinate
	 * @param p a Point representing a line segment
	 * @param x the current X coordinate
	 * @return the Y coordinate of the line segment in the current X coordinate
	 */
	private float currentYCoordinate(Point p, float x) {
		// first calculate the parameters of the line
		Point pair = p.getPair();
		float a = (p.getY() - pair.getY()) / (p.getX() - pair.getX());
		float b = pair.getY() - a*pair.getX();
		return a*x+b;
	}
	
	/**
	 * Calculates the intersection point of the line segments represented by the parameter points.
	 * The calculation is done as per instructions at http://mathworld.wolfram.com/Line-LineIntersection.html
	 * It is assumed that the parameters represent separate line segments that intersect in one single point.
	 * If the intersection point does not exist or is outside the segment areas, null is returned
	 * 
	 * @param p a Point representing one line segment
	 * @param q a Point representing one line segment
	 * @return Point containing the intersection or null if none found
	 */
	private Point intersection(Point p, Point q) {
		if(p instanceof Intersection || q instanceof Intersection) {
			return null;
		}
		float x1, x2, x3, x4, x;
		float y1, y2, y3, y4, y;
		float tmp;
		Point ppair = p.getPair();
		Point qpair = q.getPair();
		x1 = p.getX();
		y1 = p.getY();
		x2 = ppair.getX();
		y2 = ppair.getY();
		x3 = q.getX();
		y3 = q.getY();
		x4 = qpair.getX();
		y4 = qpair.getY();
		x = determinant(determinant(x1,y1,x2,y2), x1-x2, determinant(x3,y3,x4,y4), x3-x4);
		tmp = determinant(x1-x2, y1-y2, x3-x4, y3-y4);
		// case of parallel lines
		if(tmp == 0) {
			return null;
		}
		x = x / tmp;
		y = determinant(determinant(x1,y1,x2,y2), y1-y2, determinant(x3,y3,x4,y4), y3-y4);
		tmp = determinant(x1-x2, y1-y2, x3-x4, y3-y4);
		if(tmp == 0) {
			return null;
		}
		y = y / tmp;
		Intersection ret = new Intersection(-1,p,q, x ,y);
		// check if intersection inside line segments
		if(((x > x1 && x < x2) || (x < x1 && x > x2)) 
				&& ((y > y1 && y < y2) || (y < y1 && y > y2))
				&& ((x > x3 && x < x4) || (x < x3 && x > x4)) 
				&& ((y > y3 && y < y4) || (y < y3 && y > y4))) {
			return ret;			
		}
		return null;
	}

	private float determinant(float a, float b, float c, float d) {
		return a*d - b*c;
	}
	
	// The distance of point p from the line determined by points line1 and line2
	// defined at http://mathworld.wolfram.com/Point-LineDistance2-Dimensional.html
	private float pointDistance(Point line1, Point line2, Point p) {
		float t1 = line2.getX()-line1.getX();
		float t2 = line2.getY()-line1.getY();
		return Math.abs((t1)*(line1.getY()-p.getY()) - (line1.getX()-p.getX())*t2) / (float)Math.sqrt((t1*t1 + t2*t2));
	}
	
	// Point used to represent line segment end in this exercise.
	class Point implements Serializable {
		float x, y;
		int id;
		Point pair;
		
		Point(int id, float x, float y) {
			this.id = id;
			this.x = x;
			this.y = y;
		}
		
	
		float getX() { return x; }
		void setX(float i) { x = i; }
		
		float getY() { return y; };
		void setY(float i) { y = i; }

		int getId() { return id; }
		void setId(int i) { id = i; }
		
		Point getPair() { return pair; }
		void setPair(Point i) { pair = i; }
		
		public String toString() { return ""; }

		public boolean equals(Object o) {
			if(o == null) {
				return false;
			}
			if(o instanceof Point) {
				Point p = (Point)o;
				if(p.getX() == getX() && p.getY() == getY() && p.getId() == getId()) {
					Point op = getPair();
					Point cp = p.getPair();
					if (op.getX() == cp.getX() && op.getY() == cp.getY() && op.getId() == cp.getId()) {
						return true;
					}
				}
			}
			return false;
		}
		
		public boolean pointEqual(Point p) {
			return (p.getX() == getX() && p.getY() == getY());
		}
	
	}

	public String[] getStructureVisualisations() {
		return getModelAnswerVisualisations();
	}

	public String[] getModelAnswerVisualisations() {
		String[] s = { "layered tree", "list", "geometric graph", "list" };
		//String[] s = { "layered tree", "list", "list" };
		return s;
	}
	
	class Intersection extends Point {
		Point first, second;
		
		Intersection(int id, Point p, Point q, float x, float y) {
			super(id, x, y);
			first = p;
			second = q;
		}

		Point getFirst() { return first; }
		Point getSecond() { return second; }

		public boolean equals(Object o) {
			if(o == null) {
				return false;
			}
			if(o instanceof Intersection) {
				Intersection i = (Intersection)o;
				if(i.getX() == getX() && i.getY() == getY()) {
					if(first != null && second != null) {
						return (first.equals(i.getFirst()) && second.equals(i.getSecond())) ||
							(first.equals(i.getSecond()) && second.equals(i.getFirst()));
					} else if(first == null && second != null) {
						return (second.equals(i.getFirst()) && i.getSecond() == null) ||
							(second.equals(i.getSecond()) && i.getFirst() == null);
					} else if(first != null && second == null) {
						return (first.equals(i.getFirst()) && i.getSecond() == null) ||
							(first.equals(i.getSecond()) && i.getFirst() == null);
					} else {
						return i.getFirst() == null && i.getSecond() == null;
					}
				}
			}
			return false;
		}
		
	}

	// Binheap class used for the heaps in this exercise.
	class OwnBinHeap extends BinHeap implements DeleteEnabled {
		
		public boolean shouldDelete() {
			return true;
		}

		public CDT insert(Object o) {
			if(o == null) {
				return this;
			}
			if(o instanceof Point) {
				return super.insert(makeKey((Point)o));
			} else if (o instanceof Key) {
				Key k = (Key)o;
				String s = k.getVisualizationString();
				if(s == null || s.equals("")) {
					Point p = (Point)k.getObject();
					k = makeKey(p);
				}
				return super.insert(k);
			}
			return this;
		}
/*
		public CDT  delete(Object u) {
			CDT ret = super.delete(u);
			Object o = getObject(0); // get the root and add it to the graph if it starts a new line segment
			if(o instanceof Key) {
				Key k = (Key)o;
				Point p = (Point)k.getObject();
				if(p.getId() % 2 == 0){
					addPoint(p, graph);
					wrapper.updateNotify();
				}
			}
			return ret;
		}
*/		
	}

	// Linkedlist class used for linked lists in this exercise.
	class OwnLinkedList extends LinkedListImpl implements CDT, DeleteEnabled {
		
		public OwnLinkedList() {
			super();
		}
		
		public OwnLinkedList(Object key) {
			super(key);
		}
		
		public boolean shouldDelete() {
			return true;
		}
		
		public LinkedList getNewNode(Object key) {
			return new OwnLinkedList(key);
		}

		private OwnLinkedList findDuplicate(Object o) {
			LinkedList current = neighbours;
			while(current != null) {
				Object key = current.getElement();
				if(key == o) {
					return (OwnLinkedList)current;
				}
				current = current.getNext();
			}
			return null;
		}
		
		public CDT insert(Object o) {
			Note.out(this, "insert");
			if(neighbours.getElement() == null) { // empty list
				neighbours.setElement(o);
			} else {
				// check if node with the same key already exists
				OwnLinkedList oldNode = findDuplicate(o);
				if(oldNode == null) { // New insert
					OwnLinkedList node = (OwnLinkedList)getNewNode(neighbours.getElement());
					node.setNext2(neighbours.getNext());
					neighbours.setElement(o);
					((OwnLinkedList)neighbours).setNext2(node);					
				} else {
					oldNode.setElement(neighbours.getElement());
					neighbours.setElement(o);
				}
				checkList(neighbours.getElement());
/*				
				OwnLinkedList node = (OwnLinkedList)getNewNode(neighbours.getElement());
				node.setNext2(neighbours.getNext());
				neighbours.setElement(o);
				((OwnLinkedList)neighbours).setNext2(node);
				checkList((OwnLinkedList)neighbours);
*/
			}
			return (CDT)neighbours;
		}

		public CDT delete(Object o) {
			Note.out(this, "delete");
			if(o == null) {
				return this;
			}
			OwnLinkedList current = (OwnLinkedList)neighbours; 
			OwnLinkedList prev = null;
			while(current != null) {
				Object key = current.getElement();
				if(o.equals(key)) {
					if(current == neighbours) {
						OwnLinkedList next = (OwnLinkedList)current.getNext();
						if(next == null) {
							current.setElement(null);
						} else {
							current.setElement(next.getElement());
							current.setNext2(next.getNext());
							if(current.getNext() != null) {
								next = (OwnLinkedList)current.getNext();
								Key tmp = (Key)current.getElement();
								Point p1, p2;
								p1 = (Point)tmp.getObject();
								tmp = (Key)next.getElement();
								p2 = (Point)tmp.getObject();
								Point inter = intersection(p1,p2); {
									if(inter != null && !alreadyFound(inter)) {
										addPoint(inter, graph);
										//graph.setKey(id, inter);
										//wrapper.updateNotify();
									}
								}
							}
							return (CDT)neighbours;
						}
					} else {
						prev.setNext2(current.getNext());
						if(prev.getNext() != null) {
							Key tmp = (Key)prev.getElement();
							Point p1, p2;
							p1 = (Point)tmp.getObject();
							tmp = (Key)prev.getNext().getElement();
							p2 = (Point)tmp.getObject();
							Point inter = intersection(p1, p2);
							if(inter != null && !alreadyFound(inter)) {
								addPoint(inter, graph);
								//graph.setKey(id, inter);
								//wrapper.updateNotify();
							}
						}
						return (CDT)neighbours;
					}
				}
				prev = current;
				current = (OwnLinkedList)current.getNext();
			}
			return (CDT)neighbours;
		}

		public Object search(Object o) {
			return null;
		}

		public CDT getNewInstance() {
			return new OwnLinkedList();
		}

		public boolean hasNext() {
			return getNext() != null;
		}
		
		public void setNext2(LinkedList l) {
			super.setNext(l);
		}
		
		public void setNext(LinkedList l) {
			Note.out(this, "setnext");
			if(l == null) {
				return;
			}
			LinkedList next = getNext();
			if(next != null) {
				Object o = l.getElement();
				Object c = next.getElement();
				if(o == c) {
					next.setElement(getElement());
					super.setElement(o);
					checkList(next.getElement());
					checkList(o);
					return;
				}
			}
			super.setNext(l);
			Object o = l.getElement();
			OwnLinkedList current = (OwnLinkedList)neighbours;
			if(l == current) {
				Note.out(this, "neighbours found");
			}
			OwnLinkedList prev = null;
			while(current != null) {
				Object c = current.getElement();
				if(o == c && current != l) { // duplicate list node
					if(prev != null) {
						prev.setNext(current.getNext()); 
					} else {
						next = current.getNext();
						if(next != null) {
							current.setElement(next.getElement());
							current.setNext(next.getNext());							
						} else {
							Note.err(this, "Duplicate found in list, but list contains only one element. Contradiction.");
						}
					}
					break;
				}				
				prev = current;
				current = (OwnLinkedList)current.getNext();
			}
			
			checkList(l.getElement());
		}

		protected void checkList(Object o) {
			if(o == null) {
				return;
			}
			OwnLinkedList current = (OwnLinkedList)neighbours;
			OwnLinkedList prev = null;
			while(current != null) {
				Object cobj = current.getElement();
				if(cobj == o) {
					Point prevP, newP, nextP, inter;
					Key tmp = (Key)cobj;
					newP = (Point)tmp.getObject();
					if(prev != null) {
						tmp = (Key)prev.getElement();
						prevP = (Point)tmp.getObject();
						inter = intersection(newP, prevP);
						if(inter != null && !alreadyFound(inter)) {
							addPoint(inter, graph);
							//graph.setKey(id, inter);
						}
					}
					if(current.getNext() != null) {
						tmp = (Key)current.getNext().getElement();
						nextP = (Point)tmp.getObject();
						inter = intersection(newP, nextP);
						if(inter != null && !alreadyFound(inter)) {
							addPoint(inter, graph);
							//graph.setKey(id, inter);
						}
					}
					break;
				}
				prev = current;
				current = (OwnLinkedList)current.getNext();
			}
			//wrapper.updateNotify();
			
		}
	
		private boolean alreadyFound(Point p) {
			LinkedList check = cuts;
			while(check != null) {
				if(p.equals(check.getElement())) {
					return true;
				}
				check = check.getNext();
			}
			return false;
		}

	}
		
	// class for creating non-random test input.
	class NonRandom implements Serializable {
		
		private int[] nros = {10, 10, 40, 60, 12, 55, 80, 20, 50, 8, 70, 70, 95, 20, 110, 40 };
		private int i;
		
		NonRandom() {
			i = 0;
		}
		
		public int nextInt(int foo) {
			int ret = nros[i];
			i = (i+1)%16;
			return ret;
		}
	}

	class MyGraph extends ExerciseGeometricGraph /*implements Styled */{

		//private VirtualInteger counter;
		private StyleSheet ss;
		
		public MyGraph() {
			super();
			//counter = new VirtualInteger(-1);
			ss = new StyleSheetAdapter();
		}
/*		
		public int getNextId() {
			return counter.inc();
		}
*/		
		public StyleSheet getStyleSheet() {
			return ss;
		}
	
		public Vertex getNewVertex(int x, int y, Object key) {
			String l = "";
			if(key instanceof Point) {				
				l = getVisualizationString((Point)key);
			} 
			MyVertex v = new MyVertex(x,y,key);
			v.setLabel(l);
			return v;
		}
		
	}
	
	class MyVertex extends ExerciseGeometricGraphVertex implements /*ReferenceInformationDecorator,*/ Styled {

		private String refInfo;
		private String label;
		//private int id;
		private StyleSheet ss;
		
		public MyVertex(int x, int y, Object s) {
			super(x,y,s);
			/*
			if(s == null) {
				label = "";
			} else {
				label = s.toString();
			}
*/
			refInfo = "";
			//id = i;
			label = "";
			ss = new StyleSheetAdapter() {
				public boolean isLabelEnabled() { return true; }
				public String getLabel() { return label; }
			};
		}

		public void setLabel(String s) {
			label = s;
		}
/*		
		public boolean isReferenceInfoBoxEnabled() {
			return false;
		}

		public boolean isReferenceToolTipInfoEnabled() {
			return true;
		}

		// hacked, since each node will have only one reference from it
		public String getReferenceInformation(int i) {
			return refInfo;
		}

		public void setReferenceInformation(int i, String s) {
			refInfo = s;
		}

		public boolean isInfoBoxEnabled() {
			return false;
		}

		public boolean isToolTipInfoEnabled() {
			return false;
		}

		public void setInfo(String s) {
			
		}

		public String getInfo() {
			return null;
		}

		public int getWaitTime() {
			return 2;
		}

		public void setWaitTime(int t) {
		}
*/
		public StyleSheet getStyleSheet() {
			return ss;
		}
		
	}
	
	public GridBagConstraints[] getConstraints() {
		Insets i = new Insets(10, 10, 10, 10);
		GridBagConstraints gbc[] = new GridBagConstraints[4];
		GridBagConstraints c = new GridBagConstraints();
		c.insets = i;
		c.weightx = 1.0;
		gbc[0] = c;
		gbc[1] = c;
		c = new GridBagConstraints();
		c.insets = i;
		c.weightx = 1.0;
		c.gridwidth = GridBagConstraints.REMAINDER;
		gbc[2] = c;
		gbc[3] = c;
 		return gbc;
	}

	public GridBagConstraints[] getModelAnswerConstraints() {
		return getConstraints();
	}

}
