package content.exercises.sda;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Component;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.Rectangle;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.geom.GeneralPath;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.util.Comparator;
import java.util.Iterator;
import java.util.Random;
import java.util.TreeSet;

import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSpinner;
import javax.swing.JTextField;
import javax.swing.SpinnerNumberModel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import matrix.simulation.VisualTypeConf;
import matrix.structures.FDT.FDT;
import matrix.structures.code.probe.PseudoCode;
import matrix.structures.simulationextensions.Selectable;
import matrix.structures.spatial.PaintingStyleDecorator;
import matrix.structures.spatial.FDT.probe.Point;
import matrix.structures.spatial.FDT.probe.PolyLine;
import matrix.structures.spatial.FDT.probe.Polygon;
import matrix.util.Application;
import matrix.util.Note;
import content.exercises.sda.structures.PointGenerator;
import content.exercises.sda.structures.PolygonGenerator;
import content.exercises.sda.structures.SimpleStack;
import content.interfaces.AWTComponentUtilizer;
import content.interfaces.ConfigureVisualType;
import content.interfaces.LayoutExercise;
import content.interfaces.ModelAnswerNames;
import content.interfaces.SimulationExercise;
import content.interfaces.SimulationExerciseSelfAssessment;
import content.interfaces.StyledExercise;

/**
 * An exercise for the ray intersection method in solving the
 * point in polygon problem. 
 * 
 * A point is outside the polygon if a ray drawn from the point 
 * to infinity crosses an even number of polygon edges and 
 * inside if the number of crossings is odd (the special case of
 * rays crossing vertices is not handled in this exercise). 
 * 
 * @author Juha Helminen
 */
public class PointInPolygon implements SimulationExercise, SimulationExerciseSelfAssessment, 
									   AWTComponentUtilizer, LayoutExercise, StyledExercise,
									   ConfigureVisualType, ModelAnswerNames {

    // For debugging
	private static final boolean STATIC_SEED = false;
	private static final long STATIC_SEED_VALUE = 1167829464618L;

	
    // ** Exercise configuration **
    
	// The number of vertices in the polygon.
	private static final int POLYGON_VERTEX_COUNT = 30;
    // The minimum allowed distance between two vertices.
    private static final int MIN_VERTEX_VERTEX_DISTANCE = 10;
    // The minimum allowed number of ray intersections.
    private static final int MIN_INTERSECTION_COUNT = 3;
	// The minimum distance along the x-axis from a vertex to the visualization border.
	private static final int X_BORDER = 20;
	// The minimum distance along the x-axis from a vertex to the visualization border.
	private static final int Y_BORDER = 3;
	// Minimum distance of the input point and the intersection points along the x-axis.
	private static final int MIN_IS_IS_DISTANCE = 8;
    // The minimum allowed distance between the test point and the polygon edges.
    private static final int MIN_POINT_EDGE_DISTANCE = 8;
    // The minimum allowed distance between a polygon vertex and edge.
    private static final int MIN_VERTEX_EDGE_DISTANCE = 8;
    // The minimum allowed distance between the ray and the vertices.
    private static final int MIN_RAY_VERTEX_DISTANCE = 4;
	// The dimensions of the visualization area.
	private static final int WIDTH = 500;
	private static final int HEIGHT = 320;
	// Limits for the number of intersections for the made up tests.
	private static final int MIN1 = 1;
	private static final int MAX1 = 9;
	private static final int MIN2 = 10;
	private static final int MAX2 = 19;
	
    
    // ** Exercise color and style scheme **

    private static final Color RAY_COLOR = Color.YELLOW;
    private static final Color POLYGON_COLOR = Color.WHITE;
    private static final Color TEST_POINT_COLOR = Color.BLUE;
    private static final Color INTERSECTION_COLOR = Color.RED;
    private static final PaintingStyleDecorator intersectionDecorator;
    private static final PaintingStyleDecorator testPointDecorator;
    private static final PaintingStyleDecorator rayDecorator;
    private static final PaintingStyleDecorator polygonDecorator;
    static {
        float[] rayDashPattern = { 1, 5 };
        rayDecorator = new PaintingStyleDecorator();
        rayDecorator.setLineColor(RAY_COLOR);
        rayDecorator.setStroke(2, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL, 5, rayDashPattern, 0);
        
        intersectionDecorator = new PaintingStyleDecorator();
        intersectionDecorator.setLineColor(INTERSECTION_COLOR);
        intersectionDecorator.setStroke(1, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL);

        testPointDecorator = new PaintingStyleDecorator();
        testPointDecorator.setLineColor(TEST_POINT_COLOR);
        testPointDecorator.setStroke(1, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL);
        

        polygonDecorator = new PaintingStyleDecorator();
        polygonDecorator.setLineColor(POLYGON_COLOR);
        polygonDecorator.setStroke(2, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL);
    }
    
    // Used to define the maximum legal value for the JSpinner component.
	private static final int MAX_NUMBER_OF_INTERSECTIONS = POLYGON_VERTEX_COUNT;
	// The background color of the answer panel.
	private static final Color ANSWER_PANEL_COLOR = new Color(192, 192, 192);

	private long seed = System.currentTimeMillis();

	
    // ** Structures for the exercise. **
    
    // The vertices of the polygon.
	private java.awt.Point[] polygonVertices;
    // The points to be tested.
    private java.awt.Point[] testPoints;
    
    // Student
    private Boolean[] studentAnswers;
    private int[] studentIntersections;
    private boolean[] raysDrawn;

    // Model
	private boolean[] modelAnswers;
    private int[] modelIntersections;
	
    // The visualizations of the polygon, points and rays.
	private SimpleStack problem;
    // Model answer visual elements.
    private SimpleStack solution;
    private PseudoCode solutionPanel;


    // ** Implementation of interface SimulationExercise **

	public FDT[] getAnswer() {
		return new FDT[0];
	}

	public FDT[] getInitialStructures() {
        SimpleStack problem = new SimpleStack(new Rectangle(WIDTH, HEIGHT, 0, 0));
        Polygon polygon = createPolygon(this.polygonVertices);
        problem.ADTPush(polygon);
        ClickablePoint[] points = createPoints(this.testPoints);
        for (int i = 0; i < points.length; i++) {
            problem.ADTPush(points[i]);
        }
        return new FDT[]{ problem };
	}
	
	public String[] getStructureNames() {
		return new String[]{ "Point in polygon problem" };
	}

	public FDT[] init() {
        // For debugging
        long seed = this.getSeed();
        Note.out(this, "The random number generator seed for this exercise is: " + seed);
        Random r = new Random(seed);

        // Initialize fields.
        this.studentAnswers = new Boolean[2 + 2];
		this.studentIntersections = new int[2];
		for (int i = 0; i < this.studentIntersections.length; i++) {
			this.studentIntersections[i] = -1;
		}
		this.raysDrawn = new boolean[2];
		this.modelIntersections = new int[2 + 2];
		this.modelAnswers = new boolean[2 + 2];

        final int MAX_TRIES = 40;
        int tryCounter = 0;
        while (true) {
            // Generate the exercise input.
            try {
                // Generate a set of points for the polygon.
                this.polygonVertices = PointGenerator.generatePoints(X_BORDER,
                                                                     Y_BORDER,
                                                                     WIDTH - X_BORDER,
                                                                     HEIGHT - Y_BORDER,
                                                                     POLYGON_VERTEX_COUNT,
                                                                     MIN_VERTEX_VERTEX_DISTANCE,
                                                                     r);
                // Order the points in such a manner that they form the contour of a polygon.
                PolygonGenerator.generatePolygon2Peasants(this.polygonVertices);
               
                Line2D[] edges = createPolygonEdges(this.polygonVertices);
                for (int j = 0; j < this.polygonVertices.length; j++) {
                    if (!checkDistance2(this.polygonVertices[j], edges, MIN_VERTEX_EDGE_DISTANCE)) {
                        throw new Exception("Failed to generate the polygon.");
                    }
                }
                
                // Generate the test points.
                this.testPoints = generateTestPoints(polygonVertices, 
                                                     new Rectangle(X_BORDER, 
                                                                   Y_BORDER, 
                                                                   WIDTH / 4, 
                                                                   HEIGHT - 2 * Y_BORDER), 
                                                     r);
                // success
                break;
            } catch (Exception e) {
                if (tryCounter++ >= MAX_TRIES) {
                    Note.err(this, e.getMessage());
                    Note.show(this, "Could not generate input for the exercise. Please restart the exercise.");
                    return null;
                }
            }
        }
        
        // Generate the intersection counts for the made up -rays.
        this.modelIntersections[2] = r.nextInt(MAX1 - MIN1 + 1) + MIN1;
        this.modelIntersections[3] = r.nextInt(MAX2 - MIN2 + 1) + MIN2;

        this.problem = (SimpleStack)this.getInitialStructures()[0];
        // Generate the solution.
        this.solution = this.createSolution();
        this.solutionPanel = this.createSolutionPanel();
        
        return new FDT[]{ this.problem };
	}
    
    private static Polygon createPolygon(java.awt.Point[] vertices) {
        Point[] points = new Point[vertices.length];
        for (int i = 0; i < vertices.length; i++) {
            points[i] = new Point(vertices[i]);
        }
        Polygon polygon = new Polygon(points, ""); 
        polygon.setPaintingStyleDecorator(polygonDecorator);
        return polygon;
    }
    
    private ClickablePoint[] createPoints(java.awt.Point[] points) {
        ClickablePoint[] newPoints = new ClickablePoint[points.length];
        for (int i = 0; i < points.length; i++) {
            // i + 1 - start point indexing from 1
            newPoints[i] = new ClickablePoint(points[i], i + 1);
            newPoints[i].setPaintingStyleDecorator(testPointDecorator);
        }
        return newPoints;
    }
    
    private static java.awt.Point[] generateTestPoints(java.awt.Point[] vertices, 
                                                       Rectangle limits, 
                                                       Random r) throws Exception {
        java.awt.Point[] points = new java.awt.Point[2];
        boolean randomBool = r.nextBoolean();
        points[0] = generateTestPoint(vertices, 
                                      new Rectangle(limits.x,
                                                    limits.y + limits.height / 2 + limits.height % 2 + 5,
                                                    limits.width,
                                                    limits.height / 2 - 5), 
                                      randomBool, 
                                      r);
        points[1] = generateTestPoint(vertices, 
                                      new Rectangle(limits.x,
                                                    limits.y,
                                                    limits.width,
                                                    limits.height / 2 + limits.height % 2 - 5), 
                                      !randomBool, 
                                      r);
        return points;
    }
    
    
    private static java.awt.Point generateTestPoint(java.awt.Point[] vertices, 
                                                    Rectangle limits,
                                                    boolean inside,
                                                    Random r) throws Exception {
        final int MAX_TRIES = 60;
        int tryCounter = 0;
        
        java.awt.Point point;
        int x = 0, y = 0;

        GeneralPath polygon = createPolygonPath(vertices);
        Line2D[] edges = createPolygonEdges(vertices);
        Line2D ray;

        do {
            if (tryCounter++ >= MAX_TRIES) {
                throw new Exception("Failed to choose a test point.");
            }
            x = r.nextInt((int)limits.getMaxX() - (int)limits.getMinX() + 1)
                + (int)limits.getMinX();
            y = r.nextInt((int)limits.getMaxY() - (int)limits.getMinY() + 1)
                + (int)limits.getMinY();
            point = new java.awt.Point(x, y);
            ray = new Line2D.Double(point, 
                                    new Point2D.Double(WIDTH, point.getY()));
        } while ((polygon.contains(point) != inside) || 
                 (!checkIntersections(ray, edges)) ||
                 (!checkDistance(point, edges, MIN_POINT_EDGE_DISTANCE) ||
                 (!checkDistance(ray, vertices, MIN_RAY_VERTEX_DISTANCE))));
        return point;
    }
    

    private static GeneralPath createPolygonPath(java.awt.Point[] vertices) {
        GeneralPath path = new GeneralPath();
        if(vertices.length > 0) {
            path.moveTo((float)vertices[0].getX(), (float)vertices[0].getY());
            for (int i = 1; i < vertices.length; i++) {
                path.lineTo((float)vertices[i].getX(), (float)vertices[i].getY());
            }
            // Close the path.
            path.closePath();
        }
        return path;
    }

    private static Line2D[] createPolygonEdges(java.awt.Point[] vertices) {
        Line2D[] edges = null;
        if(vertices.length > 1) {
            edges = new Line2D[vertices.length];
            for (int i = 0; i < vertices.length - 1; i++) {
                edges[i] = new Line2D.Double(vertices[i], vertices[i + 1]);
            }
            // Close the polygon.
            edges[vertices.length - 1] = new Line2D.Double(vertices[vertices.length - 1], 
                                                           vertices[0]);
        }
        return edges;
    }
    
    private static boolean checkDistance(Point2D point, 
                                         Line2D[] lines, 
                                         int minDistance) {
        for (int i = 0; i < lines.length; i++) {
            if (lines[i].ptSegDist(point) < minDistance) {
                return false;
            }
        }
        return true;
    }

    private static boolean checkDistance2(Point2D point, 
                                          Line2D[] lines, 
                                          int minDistance) {
        for (int i = 0; i < lines.length; i++) {
            if (lines[i].getP1().equals(point) ||
                lines[i].getP2().equals(point)) {
                continue;
            } else if (lines[i].ptSegDist(point) < minDistance) {
                return false;
            }
        }
        return true;
    }

    private static boolean checkDistance(Line2D line, 
                                         Point2D[] points, 
                                         int minDistance) {
        for (int i = 0; i < points.length; i++) {
            if (line.ptSegDist(points[i]) < minDistance) {
                return false;
            }
        }
        return true;
    }
    
    
    private static boolean checkIntersections(Line2D line, Line2D[] lines) {
        TreeSet intersections = new TreeSet(new Comparator() {
            public int compare(Object arg0, Object arg1) {
                Point2D p0 = (Point2D)arg0;
                Point2D p1 = (Point2D)arg1;
                if (p0.getX() < p1.getX()) {
                    return -1;
                }
                // Let the TreeSet object contain duplicates.
                /* else if (p0.getX() == p1.getX()) {
                    return 0;
                }*/ else {
                    return 1;
                }
            }
        });
        intersections.add(line.getP1());
        Point2D isp;
        int numberOfIntersections = 0;
        for (int i = 0; i < lines.length; i++) {
            if (line.intersectsLine(lines[i])) {
                isp = getIntersection(line, lines[i]);
                intersections.add(isp);
                numberOfIntersections++;
            }
        }
        
        // Check the intersection distances.
        Iterator ispIterator = intersections.iterator();
        Point2D previous = (Point2D)ispIterator.next();
        Point2D current;
        while (ispIterator.hasNext()) {
            current = (Point2D)ispIterator.next();
            if (current.distance(previous) < MIN_IS_IS_DISTANCE) {
                return false;
            }
            previous = current;
        }
        
        // Check the number of intersections.
        if (numberOfIntersections < MIN_INTERSECTION_COUNT) {
            return false;
        } else {
            return true;
        }
    }
    
    
    private static Point2D getIntersection(Line2D l1, Line2D l2) {
        double ax = l1.getX1(), ay = l1.getY1(),
               bx = l2.getX1(), by = l2.getY1(),
               cx = l1.getX2(), cy = l1.getY2(),
               dx = l2.getX2(), dy = l2.getY2();
        Point2D p = new Point2D.Double(0, 0);
        p.setLocation(det(det(ax, ay, cx, cy), 
                          ax - cx,
                          det(bx, by, dx, dy), 
                          bx - dx) /
                      det(ax - cx, 
                          ay - cy, 
                          bx - dx, 
                          by - dy), 
                      0);
        p.setLocation(p.getX(),
                      det(det(ax, ay, cx, cy), 
                              ay - cy,
                              det(bx, by, dx, dy), 
                              by - dy) /
                          det(ax - cx, 
                              ay - cy, 
                              bx - dx, 
                              by - dy));
        return p;
    }
    
    private static double det(double a, double b, double c, double d){
        return (a * d) - (b * c);
    }

    private SimpleStack createSolution() {
        SimpleStack solution = (SimpleStack)this.getInitialStructures()[0];
        java.awt.Point testPoint;
        Point intersection;
        Line2D.Double l1, l2;
        // Draw the rays and intersections & calculate them.
        for (int i = 0; i < this.testPoints.length; i++) {
            // Set the number of intersections to zero and answers to false.
            this.modelIntersections[i] = 0;
            this.modelAnswers[i] = false;
            testPoint = this.testPoints[i];
            drawRay(testPoint, solution);
            l1 = new Line2D.Double(testPoint, new Point2D.Double(WIDTH, testPoint.getY()));
            int j;
            // Test for intersection with each edge of the polygon.
            for (j = 0; j < this.polygonVertices.length - 1; j++) {
                l2 = new Line2D.Double(polygonVertices[j], polygonVertices[j + 1]);
                if (l1.intersectsLine(l2)) {
                    this.modelIntersections[i]++;
                    this.modelAnswers[i] = !this.modelAnswers[i];
                    intersection = new Point(getIntersection(l1, l2), "");
                    intersection.setPaintingStyleDecorator(intersectionDecorator);
                    solution.ADTPush(intersection);
                }
            }
            l2 = new Line2D.Double(polygonVertices[j], polygonVertices[0]);
            if (l1.intersectsLine(l2)) {
                this.modelIntersections[i]++;
                this.modelAnswers[i] = !this.modelAnswers[i];
                intersection = new Point(getIntersection(l1, l2), "");
                intersection.setPaintingStyleDecorator(intersectionDecorator);
                solution.ADTPush(intersection);
            }
        }
        return solution;
    }
    
    private PseudoCode createSolutionPanel() {
        // Generate answers for the made up -rays.
        for (int i = 0; i < 2; i++) {
            this.modelAnswers[2 + i] = ((this.modelIntersections[2 + i] % 2) == 1); 
        }
        
        // Create the text panel showing the correct answers.
        // Ray and point indexing starts from 1
        String[] text = new String[2 + 2];
        for (int i = 0; i < 2; i++) {
            text[i] = new String();
            text[i] += "Ray " + (i + 1) + " intersections: " + this.modelIntersections[i] + 
                       "  =>  " + 
                       "Point " + (i + 1) + (this.modelAnswers[i] ? " IS " : " IS NOT ") +
                       "inside the polygon.";
        }
        for (int i = 2; i < (2 + 2); i++) {
            text[i] = new String();
            text[i] += "Ray " + (i + 1) + " intersections: " + this.modelIntersections[i] + 
                       "  =>  " + 
                       "Point " + (i + 1) + (this.modelAnswers[i] ? " IS " : " IS NOT ") +
                       "inside the polygon.";
        }
        text[0] = "_1 " + text[0];
        text[text.length - 1] += "_/1";
        PseudoCode solutionPanel = new PseudoCode();
        solutionPanel.setCodeLines(text);
        solutionPanel.setLocation(1);
        return solutionPanel;
    }

    
	public long getSeed() {
		// Seed might be set to a constant value for debugging purposes.
		if (STATIC_SEED) {
			return PointInPolygon.STATIC_SEED_VALUE;
		} else {
			return this.seed;
		}
	}

	public void setSeed(long seed) {
		this.seed = seed;
	}


	// ** Implementation of interface Exercise **

	public String getDescription() {
		return "Solve the point in polygon problem using the ray intersection method.";
	}


	// ** Implementation of interface SimulationExerciseSelfAssesment **
	
	public int getMaxPoints() {
		// 1 point for drawing a ray.
		// 1 point for entering a correct number of intersections.
		// 1 point for answering correctly to a question.
		return (2 * 3 + 2);
	}

	public int[] judge() {
        int grade = 0;
        // 1 point for drawing a ray.
        for (int i = 0; i < raysDrawn.length; i++) {
            if (raysDrawn[i]) {
                grade++;
            }
        }
        // 1 point for entering a correct number of intersections.
        for (int i = 0; i < 2; i++) {
            if (studentIntersections[i] == modelIntersections[i]) {
                    grade++;
            }
        }
        // 1 point for answering correctly to a question.
		for (int i = 0; i < modelAnswers.length; i++) {
			if ((studentAnswers[i] != null) && 
				(studentAnswers[i].booleanValue() == modelAnswers[i])) {
				grade++;
			}
		}
		return new int[]{ grade };
	}

    public String getFeedbackString() {
        // TODO Auto-generated method stub
        return null;
    }

	// ** Implementation of interface SimulationExerciseModel **
	
	public FDT[] makeModelAnswer() {
		return new FDT[]{ this.solution, this.solutionPanel };
	}
	
    public FDT[] solve() {
        return new FDT[0];
    }

	
	// ** Implementation of interface AWTComponentUtilizer **
	
	public void setApplication(Application a) {
        // Storing the application is unnecessary in this exercise.
	}

	
	public Component[] getAWTComponents() {
		GridBagLayout gbLayout = new GridBagLayout();
		JPanel gridPanel = new JPanel(gbLayout);
		JLabel labels[] = new JLabel[4];
		JSpinner spinners[] = new JSpinner[2];
		JTextField values[] = new JTextField[2];
        JLabel questions[] = new JLabel[4];
        JComboBox answers[] = new JComboBox[4];
        for (int i = 0; i < 4; i++) {
            // i + 1 - Start ray numbering from 1
        	labels[i] = new JLabel("Ray " + (i + 1) + " intersections");
        	if (i >= 2) {
        		values[i - 2] = 
        			new JTextField(new Integer(this.modelIntersections[i]).toString());
        		values[i - 2].setEditable(false);
        	} else {
	        	spinners[i] = new JSpinner(new SpinnerNumberModel(this.studentIntersections[i], -1, MAX_NUMBER_OF_INTERSECTIONS, 1));
	        	spinners[i].addChangeListener(new spinnerChangeListener(spinners[i], i));
        	}
        	questions[i] = new JLabel("Is point " + (i + 1) + " inside the polygon?");
        	answers[i] = new JComboBox(new Object[]{ new String("ANSWER"), new String("NO"), new String("YES") });
            answers[i].addItemListener(new answerBoxItemListener(answers[i], i));
        }
        
        GridBagConstraints gbc = new GridBagConstraints();
        gridPanel.setBackground(ANSWER_PANEL_COLOR);
        gbc.gridwidth = 1;
        for (int i = 0; i < 4; i++) {
        	gbc.gridy = i + 1;
            gbc.insets = new Insets(0, 0, 0, 5);
        	gridPanel.add(labels[i], gbc);
            gbc.insets = new Insets(0, 0, 0, 30);
        	if (i >= 2) {
        		gridPanel.add(values[i - 2], gbc);
        	} else {
        		gridPanel.add(spinners[i], gbc);
        	}
            gbc.insets = new Insets(0, 0, 0, 5);
        	gridPanel.add(questions[i], gbc);
            gbc.insets = new Insets(0, 0, 0, 0);
        	gridPanel.add(answers[i], gbc);
        }
        return new Component[]{ gridPanel };
	}

    public GridBagConstraints[] getAWTComponentConstraints() {
        return null;
    }
    
	class answerBoxItemListener implements ItemListener {
		private int rayNumber;
		private JComboBox comboBox;
		private boolean answered;
		
		public answerBoxItemListener(JComboBox comboBox, int rayNumber) {
			this.rayNumber = rayNumber;
			this.comboBox = comboBox;
			this.answered = false;
		}
		
		public void itemStateChanged(ItemEvent e) {
			if (e.getStateChange() == ItemEvent.SELECTED) {
				setAnswer((String)e.getItem(), this.rayNumber);
			} else if ((!answered) &&
					   (e.getStateChange() == ItemEvent.DESELECTED)) {
					this.comboBox.removeItem("ANSWER");
					this.answered = true;
			}
		}

		private void setAnswer(String answer, int rayNumber) {
			if (answer.equalsIgnoreCase("yes")) {
				studentAnswers[rayNumber] = new Boolean(true);
			} else if (answer.equalsIgnoreCase("no")) {
				studentAnswers[rayNumber] = new Boolean(false);
			} else {
				studentAnswers[rayNumber] = null;
			}
		}
	}
	
	class spinnerChangeListener implements ChangeListener {
		private int rayNumber;
		private JSpinner spinner;
		
		public spinnerChangeListener(JSpinner spinner, int rayNumber) {
			this.rayNumber = rayNumber;
			this.spinner = spinner; 
		}

		public void stateChanged(ChangeEvent e) {
			studentIntersections[this.rayNumber] = ((Integer)this.spinner.getValue()).intValue();
			if (studentIntersections[this.rayNumber] > -1) {
				((SpinnerNumberModel)this.spinner.getModel()).setMinimum(new Integer(0));
			}
		}
		
	}
	
	
	// ** Implementation of interface LayoutExercise **
	
	public GridBagConstraints[] getConstraints() {
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		gbc.insets = new Insets(2, 0, 6, 0);
		return new GridBagConstraints[]{ gbc };
	}

	public GridBagConstraints[] getModelAnswerConstraints() {
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		gbc.insets = new Insets(2, 0, 0, 0);
		return new GridBagConstraints[]{ gbc, gbc };
	}


	// ** Implementation of interface StyledExercise **

	public String[] getModelAnswerVisualisations() {
		return new String[]{ "2D", "pseudocode" };
	}

	public String[] getStructureVisualisations() {
		return new String[]{ "2D" };
	}


	// ** Implementation of interface StyledExercise **

	public VisualTypeConf[] conf() {
		VisualTypeConf polygon = new VisualTypeConf();
		VisualTypeConf answerPanel = new VisualTypeConf();
		polygon.enable("matrix.visual.VisualArea$AreaVisualization", 
					   VisualTypeConf.HIGHLIGHT_OPERATION);
		return new VisualTypeConf[]{ polygon, answerPanel };
	}
	
	
	// ** Implementation of interface ModelAnswerNames **

	public String[] getModelAnswerNames() {
		return new String[]{ "Point in polygon problem solution using the ray intersection method.", 
							 "Correct answers:"};
	}
	
	
    // ** Other **
    
    private static void drawRay(Point2D p, SimpleStack t) {
        PolyLine ray = new PolyLine(new Point[]{new Point(p), new Point(WIDTH, p.getY(), "")}, "Ray " + p);
        ray.setPaintingStyleDecorator(rayDecorator);
		t.ADTPush(ray);
	}

	
    private class ClickablePoint extends Point implements Selectable {

        // ** Constructors **
        
        public ClickablePoint(Point2D point, int id) {
            super(point.getX(), point.getY(), String.valueOf(id));
        }
        
        
        // ** Implementation of interface Selectable **

        public long getSelectionTime() {
            return Selectable.NOT_SELECTED;
        }

        public void setSelect(long globalClockTimestamp) {
            // - 1 because the rays are numbered starting from 1
            if (!raysDrawn[Integer.parseInt(super.toString()) - 1]) {
                drawRay(this.getPoint2D(), problem);
                // Mark the ray as having been drawn.
                raysDrawn[Integer.parseInt(super.toString()) - 1] = true;
            }
        }
        
        private static final long serialVersionUID = -165816825139970075L;
    }

    private static final long serialVersionUID = 1836335111352230389L;


}
