package content.exercises.sda;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Component;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.LayoutManager;
import java.awt.Rectangle;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.geom.GeneralPath;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.util.Comparator;
import java.util.Iterator;
import java.util.Random;
import java.util.Stack;
import java.util.TreeSet;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSpinner;
import javax.swing.SpinnerModel;
import javax.swing.SpinnerNumberModel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import matrix.animation.Animator;
import matrix.simulation.VisualTypeConf;
import matrix.structures.FDT.Array;
import matrix.structures.FDT.FDT;
import matrix.structures.FDT.Struct;
import matrix.structures.memory.VirtualObject;
import matrix.structures.simulationextensions.Selectable;
import matrix.structures.spatial.Area;
import matrix.structures.spatial.PaintingStyleDecorator;
import matrix.structures.spatial.FDT.probe.Point;
import matrix.structures.spatial.FDT.probe.PolyLine;
import matrix.structures.spatial.FDT.probe.SpatialElement;
import matrix.util.Application;
import matrix.util.Note;
import content.exercises.sda.structures.PointGenerator;
import content.exercises.sda.structures.PolygonGenerator;
import content.exercises.sda.structures.PolygonRTree;
import content.exercises.sda.structures.SimpleStack;
import content.exercises.sda.structures.PolygonRTree.LeafFunction;
import content.exercises.sda.structures.PolygonRTree.PolygonRTreeNode;
import content.exercises.sda.structures.PolygonRTree.SelectFunction;
import content.interfaces.AWTComponentUtilizer;
import content.interfaces.ConfigureVisualType;
import content.interfaces.JudgeBlocks;
import content.interfaces.LayoutExercise;
import content.interfaces.ModelAnswerNames;
import content.interfaces.SimulationExercise;
import content.interfaces.SimulationExerciseModel;
import content.interfaces.StyledExercise;

/**
 * An exercise for the ray intersection method in solving the
 * point in polygon problem where the polygon (its edges) is 
 * stored in an r-tree. 
 * 
 * A point is outside the polygon if a ray drawn from the point 
 * to infinity crosses an even number of polygon edges and 
 * inside if the number of crossings is odd (the special case of
 * rays crossing vertices is not handled in this exercise). 
 * 
 * @author Juha Helminen
 */
public class PointInPolygonRTree implements SimulationExercise, StyledExercise, 
                                            ConfigureVisualType, AWTComponentUtilizer,
                                            LayoutExercise, SimulationExerciseModel, 
                                            ModelAnswerNames, JudgeBlocks {

    // For debugging
    private static final boolean STATIC_SEED = false;
    private static final long STATIC_SEED_VALUE = 1165453059890L;

    private long seed = System.currentTimeMillis();

    // ** Exercise configuration **
    // The number of vertices in the polygon.
    private static final int POLYGON_VERTEX_COUNT = 18;
    // The minimum allowed distance between two vertices.
    private static final int MIN_VERTEX_VERTEX_DISTANCE = 10;
    // The minimum allowed number of ray intersections.
    private static final int MIN_INTERSECTION_COUNT = 3;
    // The minimum distance along the x-axis from a vertex to the visualization border.
    private static final int X_BORDER = 8;
    // The minimum distance along the x-axis from a vertex to the visualization border.
    private static final int Y_BORDER = 3;
    // Minimum distance of the input point and the intersection points along the x-axis.
    private static final int MIN_IS_IS_DISTANCE = 8;
    // The minimum allowed distance between the test point and the polygon edges.
    private static final int MIN_POINT_EDGE_DISTANCE = 8;
    // The minimum allowed distance between a polygon vertex and edge.
    private static final int MIN_VERTEX_EDGE_DISTANCE = 8;
    // The minimum allowed distance between the ray and the vertices.
    private static final int MIN_RAY_VERTEX_DISTANCE = 6;
    // The dimensions of the visualization area.
    private static final int WIDTH = 600;
    private static final int HEIGHT = 200;
    // 2 is for now the only allowed value.
    private static final int NUMBER_OF_TESTS = 2;
    
    // ** Exercise color and style scheme **
    // The background color of the answer panel.
    private static final Color ANSWER_PANEL_COLOR = new Color(192, 192, 192);
    private static final Color TEST_POINT_COLOR = Color.BLUE;
    private static final Color RAY_COLOR = Color.YELLOW;
    private static final Color EDGE_COLOR = Color.WHITE;
    private static final Color HIDDEN_EDGE_COLOR = Color.BLACK;
    private static final Color INTERSECTED_EDGE_COLOR = Color.WHITE;
    private static final Color INTERSECTION_COLOR = Color.RED;
    private static final Color AREA_COLOR = Color.GRAY;
    private static final Color AREA_FILL_COLOR = Color.GRAY;
    private static final Color SELECTED_AREA_COLOR = Color.GRAY;
    private static final Color SELECTED_AREA_FILL_COLOR = Color.DARK_GRAY;
    private static final Color INTERSECTED_AREA_COLOR = Color.GRAY;
    private static final Color INTERSECTED_AREA_FILL_COLOR = Color.DARK_GRAY;
    private static final PaintingStyleDecorator areaDecorator;
    private static final PaintingStyleDecorator selectedAreaDecorator;
    private static final PaintingStyleDecorator intersectedAreaDecorator;
    private static final PaintingStyleDecorator edgeDecorator;
    private static final PaintingStyleDecorator intersectedEdgeDecorator;
    private static final PaintingStyleDecorator intersectionDecorator;
    private static final PaintingStyleDecorator testPointDecorator;
    private static final PaintingStyleDecorator hiddenEdgeDecorator;    
    private static final PaintingStyleDecorator rayDecorator;
    static {
        float[] areaDashPattern = { 4, 2 };
        areaDecorator = new PaintingStyleDecorator();
        areaDecorator.setLineColor(AREA_COLOR);
        areaDecorator.setFillColor(AREA_FILL_COLOR);
        areaDecorator.setTransparency(0.10);
        areaDecorator.setStroke(1, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL, 5, areaDashPattern, 0);
        
        intersectedAreaDecorator = new PaintingStyleDecorator();
        intersectedAreaDecorator.setLineColor(INTERSECTED_AREA_COLOR);
        intersectedAreaDecorator.setFillColor(INTERSECTED_AREA_FILL_COLOR);
        intersectedAreaDecorator.setTransparency(0.10);
        intersectedAreaDecorator.setStroke(1, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL);

        selectedAreaDecorator = new PaintingStyleDecorator();
        selectedAreaDecorator.setLineColor(SELECTED_AREA_COLOR);
        selectedAreaDecorator.setFillColor(SELECTED_AREA_FILL_COLOR);
        selectedAreaDecorator.setTransparency(0.10);
        selectedAreaDecorator.setStroke(1, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL);

        float[] rayDashPattern = { 1, 5 };
        rayDecorator = new PaintingStyleDecorator();
        rayDecorator.setLineColor(RAY_COLOR);
        rayDecorator.setStroke(2, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL, 5, rayDashPattern, 0);

        edgeDecorator = new PaintingStyleDecorator();
        edgeDecorator.setLineColor(EDGE_COLOR);
        edgeDecorator.setStroke(2, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL);
        
        intersectedEdgeDecorator = new PaintingStyleDecorator();
        intersectedEdgeDecorator.setLineColor(INTERSECTED_EDGE_COLOR);
        intersectedEdgeDecorator.setStroke(2, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL);

        float[] modelEdgeDashPattern = { 4, 4 };
        hiddenEdgeDecorator = new PaintingStyleDecorator();
        hiddenEdgeDecorator.setLineColor(HIDDEN_EDGE_COLOR);
        hiddenEdgeDecorator.setStroke(1, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL, 4, modelEdgeDashPattern, 0);

        intersectionDecorator = new PaintingStyleDecorator();
        intersectionDecorator.setLineColor(INTERSECTION_COLOR);
        intersectionDecorator.setStroke(1, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL);

        testPointDecorator = new PaintingStyleDecorator();
        testPointDecorator.setLineColor(TEST_POINT_COLOR);
        testPointDecorator.setStroke(1, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL);
    }
    
    
    // ** Structures for the exercise. **
    private java.awt.Point[] polygonVertices;
    private java.awt.Point[] testPoints;
    // Student
    /**
     * This stack stores the problem representation: the polygon 
     * stored in the tree, the tested points, the ray(s) and 
     * the intersections.
     */
    private SimpleStack problem;
    private PolygonRTree tree;
    private Slot ray;
    
    private AnswerPanel answerPanel;
    private Slot[] answers;
    private Slot[] intersections;

    
    // ** Implementation of interface SimulationExercise **
    
    public String getDescription() {
        return "Solve the point-in-polygon problem using the ray intersection method.";
    }

    public FDT[] getInitialStructures() {
        SimpleStack problem = new SimpleStack(new Rectangle(WIDTH, HEIGHT, 0, 0));
        Slot ray = new Slot();

        PolygonRTree.PolygonRTreeNode.setSplitAlgorithm(PolygonRTree.PolygonRTreeNode.SPLIT_QUADRATIC);
        PolygonRTree.resetIDCounter();
        PolygonRTree tree = new PolygonRTree();
        tree.setPaintingStyleDecorator(areaDecorator);

        // Create the polygon edges and store them in the R-tree.
        this.insertPolygon(this.polygonVertices, tree, edgeDecorator);

        // Store the tree in the stack.
        problem.ADTPush(tree); 

        // Create the test points.
        ClickablePoint p;
        for (int i = 0; i < this.testPoints.length; i++) {
            // Start indexing from 1.
            p = new ClickablePoint(testPoints[i].x, 
                                   testPoints[i].y,
                                   String.valueOf(i + 1));
            p.setPaintingStyleDecorator(testPointDecorator);
            problem.ADTPush(p);
        }

        return new FDT[]{ problem, tree, ray };
    }
    
    
    public String[] getStructureNames() {
        return new String[]{ "Point-in-polygon problem.", 
                             "2-3-R-tree containing polygon edges." };
    }

    public FDT[] init() {
        // For debugging
        long seed = this.getSeed();
        Note.out(this, "The random number generator seed for this exercise is: " + seed);
        Random r = new Random(seed);

        PolygonRTree.PolygonRTreeNode.setSplitAlgorithm(PolygonRTree.PolygonRTreeNode.SPLIT_QUADRATIC);
        PolygonRTree tree;

        final int MAX_TRIES = 40;
        int tryCounter = 0;
        while (true) {
            // Generate the exercise input.
            try {
                tree = new PolygonRTree();

                // Generate a set of points for the polygon.
                this.polygonVertices = PointGenerator.generatePoints(X_BORDER,
                                                                     Y_BORDER,
                                                                     WIDTH - X_BORDER,
                                                                     HEIGHT - Y_BORDER,
                                                                     POLYGON_VERTEX_COUNT,
                                                                     MIN_VERTEX_VERTEX_DISTANCE,
                                                                     r);
                // Order the points in such a manner that they form the contour of a polygon.
                PolygonGenerator.generatePolygon2Peasants(this.polygonVertices);
               
                Line2D[] edges = createPolygonEdges(this.polygonVertices);
                for (int j = 0; j < this.polygonVertices.length; j++) {
                    if (!checkDistance2(this.polygonVertices[j], edges, MIN_VERTEX_EDGE_DISTANCE)) {
                        throw new Exception("Failed to generate the polygon.");
                    }
                }
                
                // Create the polygon edges and store them in the R-tree.
                this.insertPolygon(this.polygonVertices, tree, edgeDecorator);

                // Generate the test points.
                this.testPoints = generateTestPoints(polygonVertices, 
                                                     tree.getBoundingPolygon(),
                                                     r);
                // success
                break;
            } catch (Exception e) {
                if (tryCounter++ >= MAX_TRIES) {
                    Note.err(this, e.getMessage());
                    Note.show(this, "Could not generate input for the exercise. Please restart the exercise.");
                    return null;
                }
            }
        }

        FDT[] structures = this.getInitialStructures();
        this.problem = (SimpleStack)structures[0];
        this.tree = (PolygonRTree)structures[1];
        this.ray = (Slot)structures[2];
        FindIntersections findIntersections = 
            new FindIntersections(this.ray, this.problem, edgeDecorator, 
                                  edgeDecorator, intersectedEdgeDecorator);
        NodeSelector nodeSelector = new NodeSelector(ray);
        this.tree.setLeafFunction(findIntersections);
        this.tree.setSelectFunction(nodeSelector);
        this.tree.setVisualizeMinimized(false);
        
        
        this.intersections = new Slot[NUMBER_OF_TESTS];
        for (int i = 0; i < this.intersections.length; i++) {
            this.intersections[i] = new Slot();
            this.intersections[i].setElement(new Integer(0));
        }

        this.answers = new Slot[NUMBER_OF_TESTS];
        for (int i = 0; i < this.answers.length; i++) {
            this.answers[i] = new Slot();
        }
        
        this.answerPanel = new AnswerPanel(new Slot[]{ this.intersections[0], 
                this.answers[0],
                this.intersections[1],
                this.answers[1] });

        return new FDT[]{ this.problem, this.tree };
    }

    /**
     * Creates the edges (<code>Polyline</code>) corresponding to 
     * the given polygon, sets the their style and inserts them 
     * into the given tree.
     * 
     * @param polygonVertices   A set of points which are ordered to form 
     *                          the contour of a simple poloygon.
     * @param tree              The tree in which to insert the polygon.
     * @param psd               The <code>PaintingStyleDecorator</code> to be used for
     *                          the polygon edges.
     */
    private void insertPolygon(Point2D[] polygonVertices, PolygonRTree tree, PaintingStyleDecorator psd) {
        PolyLine polyline;
        final Point first = new Point((int)polygonVertices[0].getX(), 
                                      (int)polygonVertices[0].getY(),
                                      "");
        Point previous = first, current;
        for (int i = 0; i < polygonVertices.length - 1; i++) {
            current = new Point((int)polygonVertices[i + 1].getX(), 
                                (int)polygonVertices[i + 1].getY(),
                                "");
            polyline = new PolyLine(new Point[]{ previous, current }, "E" + i);
            polyline.setPaintingStyleDecorator(psd);
            tree.insert(polyline);
            previous = current;
        }
        // Connect the endpoint and starting point.
        polyline = new PolyLine(new Point[]{ previous, first }, "E" + (polygonVertices.length - 1));
        polyline.setPaintingStyleDecorator(psd);
        tree.insert(polyline);
    }

    private static java.awt.Point[] generateTestPoints(java.awt.Point[] vertices, 
                                                       Rectangle limits, 
                                                       Random r) throws Exception {
        java.awt.Point[] points = new java.awt.Point[2];
        boolean randomBool = r.nextBoolean();
        points[0] = generateTestPoint(vertices, 
                                      new Rectangle(limits.x,
                                                    limits.y + limits.height / 2 + limits.height % 2 + 5,
                                                    limits.width / 3,
                                                    limits.height / 2 - 5), 
                                      randomBool, 
                                      r);
        points[1] = generateTestPoint(vertices, 
                                      new Rectangle(limits.x,
                                                    limits.y,
                                                    limits.width / 3,
                                                    limits.height / 2 + limits.height % 2 - 5), 
                                      !randomBool, 
                                      r);
        return points;
    }
    
    private static java.awt.Point generateTestPoint(java.awt.Point[] vertices, 
                                                    Rectangle limits,
                                                    boolean inside,
                                                    Random r) throws Exception {
        final int MAX_TRIES = 60;
        int tryCounter = 0;
        
        java.awt.Point point;
        int x = 0, y = 0;

        GeneralPath polygon = createPolygonPath(vertices);
        Line2D[] edges = createPolygonEdges(vertices);
        Line2D ray;
        Line2D extendedRay;

        do {
            if (tryCounter++ >= MAX_TRIES) {
                throw new Exception("Failed to choose a test point.");
            }
            x = r.nextInt((int)limits.getMaxX() - (int)limits.getMinX() + 1)
                + (int)limits.getMinX();
            y = r.nextInt((int)limits.getMaxY() - (int)limits.getMinY() + 1)
                + (int)limits.getMinY();
            point = new java.awt.Point(x, y);
            ray = new Line2D.Double(point, 
                                    new Point2D.Double(WIDTH, point.getY()));
            extendedRay = new Line2D.Double(new Point2D.Double(0, point.getY()),
                                            new Point2D.Double(WIDTH, point.getY()));
        } while ((polygon.contains(point) != inside) || 
                 (!checkIntersections(ray, edges)) ||
                 (!checkDistance(point, edges, MIN_POINT_EDGE_DISTANCE) ||
                 (!checkDistance(extendedRay, vertices, MIN_RAY_VERTEX_DISTANCE))));
        return point;
    }

    private static GeneralPath createPolygonPath(java.awt.Point[] vertices) {
        GeneralPath path = new GeneralPath();
        if(vertices.length > 0) {
            path.moveTo((float)vertices[0].getX(), (float)vertices[0].getY());
            for (int i = 1; i < vertices.length; i++) {
                path.lineTo((float)vertices[i].getX(), (float)vertices[i].getY());
            }
            // Close the path.
            path.closePath();
        }
        return path;
    }

    private static Line2D[] createPolygonEdges(java.awt.Point[] vertices) {
        Line2D[] edges = null;
        if(vertices.length > 1) {
            edges = new Line2D[vertices.length];
            for (int i = 0; i < vertices.length - 1; i++) {
                edges[i] = new Line2D.Double(vertices[i], vertices[i + 1]);
            }
            // Close the polygon.
            edges[vertices.length - 1] = new Line2D.Double(vertices[vertices.length - 1], 
                                                           vertices[0]);
        }
        return edges;
    }
    
    private static boolean checkDistance(Point2D point, 
                                         Line2D[] lines, 
                                         int minDistance) {
        for (int i = 0; i < lines.length; i++) {
            if (lines[i].ptSegDist(point) < minDistance) {
                return false;
            }
        }
        return true;
    }

    private static boolean checkDistance2(Point2D point, 
                                          Line2D[] lines, 
                                          int minDistance) {
        for (int i = 0; i < lines.length; i++) {
            if (lines[i].getP1().equals(point) ||
                lines[i].getP2().equals(point)) {
                continue;
            } else if (lines[i].ptSegDist(point) < minDistance) {
                return false;
            }
        }
        return true;
    }
    
    private static boolean checkDistance(Line2D line, 
                                         Point2D[] points, 
                                         int minDistance) {
        for (int i = 0; i < points.length; i++) {
            if (line.ptSegDist(points[i]) < minDistance) {
                return false;
            }
        }
        return true;
    }
    
    private static boolean checkIntersections(Line2D line, Line2D[] lines) {
        TreeSet intersections = new TreeSet(new Comparator() {
            public int compare(Object arg0, Object arg1) {
                Point2D p0 = (Point2D)arg0;
                Point2D p1 = (Point2D)arg1;
                if (p0.getX() < p1.getX()) {
                    return -1;
                }
                // Let the TreeSet object contain duplicates.
                /* else if (p0.getX() == p1.getX()) {
                    return 0;
                }*/ else {
                    return 1;
                }
            }
        });
        intersections.add(line.getP1());
        Point2D isp;
        int numberOfIntersections = 0;
        for (int i = 0; i < lines.length; i++) {
            if (line.intersectsLine(lines[i])) {
                isp = getIntersection(line, lines[i]);
                intersections.add(isp);
                numberOfIntersections++;
            }
        }
        
        // Check the intersection distances.
        Iterator ispIterator = intersections.iterator();
        Point2D previous = (Point2D)ispIterator.next();
        Point2D current;
        while (ispIterator.hasNext()) {
            current = (Point2D)ispIterator.next();
            if (current.distance(previous) < MIN_IS_IS_DISTANCE) {
                return false;
            }
            previous = current;
        }
        
        // Check the number of intersections.
        if (numberOfIntersections < MIN_INTERSECTION_COUNT) {
            return false;
        } else {
            return true;
        }
    }

    public long getSeed() {
        // Seed might be set to a constant value for debugging purposes.
        if (STATIC_SEED) {
            return PointInPolygonRTree.STATIC_SEED_VALUE;
        } else {
            return this.seed;
        }
    }

    public void setSeed(long seed) {
        this.seed = seed;
    }

    public FDT[] getAnswer() {
        return new FDT[]{ this.tree, this.ray, this.answerPanel };
    }
    
    
    // ** Implementation of interface StyledExercise **

    public String[] getModelAnswerVisualisations() {
        return new String[]{ "2D", "layered tree", "struct"};
    }

    public String[] getStructureVisualisations() {
        return new String[]{ "2D", "layered tree"};
    }

    
    // ** Implementation of interface ModelAnswerNames **

    public String[] getModelAnswerNames() {
        return new String[]{ "Point-in-polygon problem.", 
                             "2-3-R-tree containing polygon edges.",
                             "Answers"};
    }
    
    
    // ** Implementation of interface SimulationExerciseModel **
    
    public FDT[] makeModelAnswer() {
        FDT[] f = solveHelp();
        return new FDT[]{ f[0], f[1], f[3] };
    }
    
    public FDT[] solve() {
        FDT[] f = solveHelp();
        return new FDT[]{ f[1], f[2], f[3] };
    }

    public FDT[] solveHelp() {
        Slot[] answers = new Slot[NUMBER_OF_TESTS];
        for (int i = 0; i < answers.length; i++) {
            answers[i] = new Slot();
        }
        Slot[] intersections = new Slot[NUMBER_OF_TESTS];
        for (int i = 0; i < intersections.length; i++) {
            intersections[i] = new Slot();
            intersections[i].setElement(new Integer(0));
        }
        
        AnswerPanel answerPanel = new AnswerPanel(new Slot[]{ intersections[0], 
                                                              answers[0],
                                                              intersections[1],
                                                              answers[1] });

        FDT[] structures = this.getInitialStructures();
        SimpleStack problem = (SimpleStack)structures[0];
        PolygonRTree tree = (PolygonRTree)structures[1];
        Slot ray = (Slot)structures[2];
        FindIntersections findIntersections = 
            new FindIntersections(ray, problem, hiddenEdgeDecorator, 
                                  edgeDecorator, intersectedEdgeDecorator,
                                  intersections);
        NodeSelector nodeSelector = new NodeSelector(ray);
        tree.setLeafFunction(findIntersections);
        tree.setSelectFunction(nodeSelector);
        tree.setVisualizeMinimized(true);
        tree.minimize();
        problem.clear();
        problem.ADTPush(tree);
        for (int j = 0; j < testPoints.length; j++) {
            problem.ADTPush(new ClickablePoint(testPoints[j].x, 
                                               testPoints[j].y,
                                               String.valueOf(j + 1)));
        }
        
        PolygonRTree.PolygonRTreeNode node = null;
        Stack stack = new Stack();
        Animator a = Animator.getActiveAnimator();
        for (int i = 0; i < this.testPoints.length; i++) {
            a.startOperation();
            if (ray.getElement() != null) {
                tree.minimize();
                problem.clear();
                problem.ADTPush(tree);
                for (int j = 0; j < testPoints.length; j++) {
                    problem.ADTPush(new ClickablePoint(testPoints[j].x, 
                                                       testPoints[j].y,
                                                       String.valueOf(j + 1)));
                }
            }
            this.drawRay(testPoints[i], problem, ray, i);
            a.endOperation();
            
            node = (PolygonRTreeNode)tree.getElement();
            stack.push(node);
            while (!stack.empty()) {
                node = (PolygonRTreeNode)stack.pop();
                a.startOperation();
                node.setSelect(Selectable.clock.getNextTimeStamp());
                a.endOperation();
                if (node.getBoundingPolygon().
                        intersectsLine(new Line2D.Double(((PolyLine)ray.getElement()).getPoints()[0].getPoint2D(),
                                                         ((PolyLine)ray.getElement()).getPoints()[1].getPoint2D()))) {
                    a.startOperation();
                    node.setSelect(Selectable.clock.getNextTimeStamp());
                    a.endOperation();
                    if (!node.isLeaf()) {
                        Area[] subAreas = node.getSubAreas();
                        for (int j = subAreas.length - 1; j >= 0; j--) {
                            stack.push(subAreas[j]);
                        }
                    }
                }
            }
            a.startOperation();
            answers[i].setElement(new Boolean(((((Integer)intersections[i].getElement()).intValue() % 2) == 1)));
            a.endOperation();
        }
        return new FDT[]{ problem, tree, ray, answerPanel };
    }


    // ** Implementation of interface ConfigureVisualType **
    
    public VisualTypeConf[] conf() {
        // Area visualization
        VisualTypeConf area = new VisualTypeConf();
        area.enable("matrix.visual.VisualArea$AreaVisualization", VisualTypeConf.HIGHLIGHT_OPERATION);
        // R-tree tree representation
        VisualTypeConf tree = new VisualTypeConf();
        tree.enable("matrix.visual.VisualLayeredTreeComponent", VisualTypeConf.HIGHLIGHT_OPERATION);
        tree.enable("matrix.visual.VisualStruct-matrix.structures.spatial.FDT.probe.PolyLine", VisualTypeConf.MINIMIZED);
        return new VisualTypeConf[] { area, tree, new VisualTypeConf() };
    }

    
    // ** Implementation of interface AWTComponentUtilizer **
    
    private transient Application application;
    
    public void setApplication(Application a) {
        this.application = a;
    }

    public Application getApplication() {
        return this.application;
    }

    public Component[] getAWTComponents() {
        GridBagLayout gbLayout = new GridBagLayout();
        MyPanel gridPanel = new MyPanel(gbLayout);
        JLabel labels[] = new JLabel[NUMBER_OF_TESTS];
        MySpinner spinners[] = new MySpinner[NUMBER_OF_TESTS];
        JLabel questions[] = new JLabel[NUMBER_OF_TESTS];
        MyComboBox answers[] = new MyComboBox[NUMBER_OF_TESTS];

        SpinnerChangeListener spinnerListener;
        AnswerBoxItemListener answerListener;
        for (int i = 0; i < NUMBER_OF_TESTS; i++) {
            labels[i] = new JLabel("Ray " + (i + 1) + " intersections");
            spinners[i] = new MySpinner(new SpinnerNumberModel(((Integer)this.intersections[i].getElement()).intValue(), 
                    0, 
                    POLYGON_VERTEX_COUNT, 1), i);
            spinnerListener = new SpinnerChangeListener(spinners[i], i);
            spinners[i].setSpinnerListener(spinnerListener);
            spinners[i].addChangeListener(spinnerListener);
            questions[i] = new JLabel("Is point " + (i + 1) + " inside the polygon?");
            answers[i] = new MyComboBox(new Object[]{ new String("ANSWER"), new String("NO"), new String("YES") }, i);
            answerListener = new AnswerBoxItemListener(i);
            answers[i].setItemListener(answerListener);
            answers[i].addItemListener(answerListener);
        }
        
        GridBagConstraints gbc = new GridBagConstraints();
        gridPanel.setBackground(ANSWER_PANEL_COLOR);
        gbc.gridwidth = 1;
        for (int i = 0; i < NUMBER_OF_TESTS; i++) {
            gbc.gridy = i + 1;
            gbc.insets = new Insets(0, 0, 0, 5);
            gridPanel.add(labels[i], gbc);
            gbc.insets = new Insets(0, 0, 0, 30);
            gridPanel.add(spinners[i], gbc);
            gbc.insets = new Insets(0, 0, 0, 5);
            gridPanel.add(questions[i], gbc);
            gbc.insets = new Insets(0, 0, 0, 0);
            gridPanel.add(answers[i], gbc);
        }
        return new Component[]{ gridPanel };
    }

    public GridBagConstraints[] getAWTComponentConstraints() {
        GridBagConstraints gbc = new GridBagConstraints();
        gbc.gridwidth = GridBagConstraints.REMAINDER;
        gbc.insets = new Insets(0, 2, 2, 2);
        return new GridBagConstraints[]{ gbc};
    }

    class AnswerBoxItemListener implements ItemListener {
        private int rayNumber;
        private boolean ignoreStateChange;
        
        public AnswerBoxItemListener(int rayNumber) {
            this.rayNumber = rayNumber;
            this.ignoreStateChange = false;
        }
        
        public void itemStateChanged(ItemEvent e) {
            if (!this.ignoreStateChange) {
                if (e.getStateChange() == ItemEvent.SELECTED) {
                    setAnswer((String)e.getItem(), this.rayNumber);
                    this.ignoreStateChange = false;
                }
            }
        }

        private void setAnswer(String str, int rayNumber) {
            Animator a = Animator.getActiveAnimator();
            a.startOperation();
            if (str.equalsIgnoreCase("yes")) {
                answers[rayNumber].setElement(new Boolean(true));
            } else if (str.equalsIgnoreCase("no")) {
                answers[rayNumber].setElement(new Boolean(false));
            } else {
                answers[rayNumber].setElement(null);
            }
            a.endOperation();
            getApplication().validate();
            getApplication().validateAnimator();
        }

        public void setIgnoreStateChange(boolean ignoreStateChange) {
            this.ignoreStateChange = ignoreStateChange;
        }
    }
    
    class SpinnerChangeListener implements ChangeListener {
        private int rayNumber;
        private MySpinner spinner;
        private boolean ignoreStateChange;
        
        public SpinnerChangeListener(MySpinner spinner, int rayNumber) {
            this.rayNumber = rayNumber;
            this.spinner = spinner;
            this.ignoreStateChange = false;
        }
        
        public void setIgnoreStateChange(boolean ignoreStateChange) {
            this.ignoreStateChange = ignoreStateChange;
        }

        public void stateChanged(ChangeEvent e) {
            if (!this.ignoreStateChange) {
                Animator a = Animator.getActiveAnimator();
                a.startOperation();
                intersections[this.rayNumber].setElement((Integer)this.spinner.getValue());
                a.endOperation();
                getApplication().validate();
                getApplication().validateAnimator();
            }
            this.ignoreStateChange = false;
        }
        
    }
    
    class MyPanel extends JPanel {

        public MyPanel(LayoutManager layout) {
            super(layout);
        }

        public void validate() {
            // Force validating of subcomponents
            int componentCount = this.getComponentCount();
            for (int i = 0; i < componentCount; i++) {
                this.getComponent(i).validate();
            }
            super.validate();
        }

        private static final long serialVersionUID = 3699821165607871718L;
    }


    /**
     * This is called when the student clicks a test point. 
     * Draws a ray from the point to the rightmost edge of the 
     * viewing area.
     * 
     * @param p The test point clicked.
     */
    private void drawStudentRay(Point p) {
        if ((this.ray.getElement() == null) || 
            ((((PolyLine)this.ray.getElement()).getPoints()[0].getX() != p.getX()) &&
             (((PolyLine)this.ray.getElement()).getPoints()[0].getY() != p.getX()))) {
            Animator a = Animator.getActiveAnimator();
            a.startOperation();
            if (this.ray.getElement() != null) {
                this.tree.minimize();
                this.problem.clear();
                this.problem.ADTPush(this.tree);
                for (int i = 0; i < testPoints.length; i++) {
                    this.problem.ADTPush(new ClickablePoint(testPoints[i].x, 
                                                            testPoints[i].y,
                                                            String.valueOf(i + 1)));
                }
            }
            this.drawRay(p.getPoint2D(), this.problem, this.ray, Integer.parseInt(p.toString()) - 1);
            a.endOperation();
            this.getApplication().validate();
        }
    }

    /**
     * Creates the ray for the given points, inserts it into the 
     * given stack and sets it as the current ray.
     * 
     * @param p     The test point.
     * @param s     The structure containing the problem representation.
     * @param ray   The holder for the current ray.
     */
    private void drawRay(Point2D p, SimpleStack s, Slot ray, int index) {
        PolyLine polyline = new PolyLine(new Point[]{new Point(p), 
                                                     new Point(WIDTH, p.getY(), "")},
                                         String.valueOf(index));
        polyline.setPaintingStyleDecorator(rayDecorator);
        s.ADTPush(polyline);
        ray.setElement(polyline);
    }
    
    
    // ** Implementation of interface LayoutExercise **
    
    public GridBagConstraints[] getConstraints() {
        GridBagConstraints gbc = new GridBagConstraints();
        gbc.gridwidth = GridBagConstraints.REMAINDER;
        gbc.insets = new Insets(2, 2, 2, 2);
        return new GridBagConstraints[]{ gbc, gbc };
    }

    public GridBagConstraints[] getModelAnswerConstraints() {
        GridBagConstraints gbc = new GridBagConstraints();
        gbc.gridwidth = GridBagConstraints.REMAINDER;
        gbc.insets = new Insets(2, 2, 2, 2);
        return new GridBagConstraints[]{ gbc, gbc, gbc };
    }
    
    private static Point2D getIntersection(Line2D l1, Line2D l2) {
        double ax = l1.getX1(), ay = l1.getY1(),
               bx = l2.getX1(), by = l2.getY1(),
               cx = l1.getX2(), cy = l1.getY2(),
               dx = l2.getX2(), dy = l2.getY2();
        Point2D p = new Point2D.Double(0, 0);
        p.setLocation(det(det(ax, ay, cx, cy), 
                          ax - cx,
                          det(bx, by, dx, dy), 
                          bx - dx) /
                      det(ax - cx, 
                          ay - cy, 
                          bx - dx, 
                          by - dy), 
                      0);
        p.setLocation(p.getX(),
                      det(det(ax, ay, cx, cy), 
                              ay - cy,
                              det(bx, by, dx, dy), 
                              by - dy) /
                          det(ax - cx, 
                              ay - cy, 
                              bx - dx, 
                              by - dy));
        return p;
    }
    
    private static double det(double a, double b, double c, double d){
        return (a * d) - (b * c);
    }
    
    
    /**
     * The execute()-method finds and marks the intersections when 
     * a leaf node is maximized.
     */
    private class FindIntersections implements LeafFunction {
        private final Slot ray;
        private final SimpleStack stack;
        private final PaintingStyleDecorator hiddenEdgeDecorator;
        private final PaintingStyleDecorator edgeDecorator;
        private final PaintingStyleDecorator intersectedEdgeDecorator;
        private Slot[] intersections; 
        
        public FindIntersections(Slot ray, SimpleStack stack, 
                                 PaintingStyleDecorator hiddenEdgeDecorator, 
                                 PaintingStyleDecorator edgeDecorator, 
                                 PaintingStyleDecorator intersectedEdgeDecorator) {
            this.ray = ray;
            this.stack = stack;
            this.hiddenEdgeDecorator = hiddenEdgeDecorator; 
            this.edgeDecorator = edgeDecorator;
            this.intersectedEdgeDecorator = intersectedEdgeDecorator;
            this.intersections = null;
        }

        public FindIntersections(Slot ray, SimpleStack stack, 
                PaintingStyleDecorator hiddenEdgeDecorator, 
                PaintingStyleDecorator edgeDecorator, 
                PaintingStyleDecorator intersectedEdgeDecorator,
                Slot[] intersections) {
            this(ray, stack, hiddenEdgeDecorator, edgeDecorator, intersectedEdgeDecorator);
            this.intersections = intersections;
        }
        
        public void execute(PolygonRTreeNode node, boolean minimized) {
            if (!minimized) {
                // Visualize possible intersections with the ray.
                if (ray.getElement() != null) {
                    Line2D rayLine = new Line2D.Double(((PolyLine)ray.getElement()).getPoints()[0].getPoint2D(),
                                                       ((PolyLine)ray.getElement()).getPoints()[1].getPoint2D());
                    SpatialElement[] lines = node.getSpatialElements();
                    Point2D isp;
                    Line2D line;
                    Point[] points;
                    Point intersection;
                    for (int i = 0; i < lines.length; i++) {
                        points = ((PolyLine)lines[i]).getPoints();
                        line = new Line2D.Double(points[0].getPoint2D(), 
                                                 points[1].getPoint2D());
                        if (rayLine.intersectsLine(line)) {
                            if (this.intersections != null) {
                                int isCount = 
                                    ((Integer)this.intersections[Integer.parseInt(((PolyLine)this.ray.getElement()).getVisualizationString())].
                                            getElement()).intValue();
                                this.intersections[Integer.parseInt(((PolyLine)this.ray.getElement()).toString())].
                                    setElement(new Integer(++isCount));
                            }
                            lines[i].setPaintingStyleDecorator(this.intersectedEdgeDecorator);
                            isp = getIntersection(rayLine, line);
                            intersection = new Point(isp.getX(), isp.getY(), "");
                            intersection.setPaintingStyleDecorator(intersectionDecorator);
                            stack.ADTPush(intersection);
                        } else {
                            lines[i].setPaintingStyleDecorator(this.edgeDecorator);
                        }
                    }
                }
            } else {
                SpatialElement[] lines = node.getSpatialElements();
                for (int i = 0; i < lines.length; i++) {
                    lines[i].setPaintingStyleDecorator(this.hiddenEdgeDecorator);
                }
            }
        }

        private static final long serialVersionUID = -7001410834193210519L;
    }
    
    /**
     * Switches the painting style decorators according to the selection 
     * state changes.
     */
    private class NodeSelector implements SelectFunction {
        private final Slot ray;
        
        public NodeSelector(Slot ray) {
            this.ray = ray;
        }
        
        public void execute(PolygonRTreeNode node, boolean select) {
            if (select) {
                // Visualize possible intersections with the ray.
                if ((this.ray.getElement() != null) &&
                    node.getBoundingPolygon().
                        intersectsLine(
                                new Line2D.Double(((PolyLine)this.ray.getElement()).getPoints()[0].getPoint2D(),
                                                  ((PolyLine)this.ray.getElement()).getPoints()[1].getPoint2D()))) {
                        node.setPaintingStyleDecorator(intersectedAreaDecorator);
                } else {
                    node.setPaintingStyleDecorator(selectedAreaDecorator);
                }
            } else {
                node.setPaintingStyleDecorator(areaDecorator);
            }
        }

        private static final long serialVersionUID = 6807519423094364799L;
    }
        
    private class ClickablePoint extends Point implements Selectable {

        // ** Constructors **
        
        public ClickablePoint(int x, int y, String visualizationString) {
            super(x, y, visualizationString);
        }
        
        
        // ** Implementation of interface Selectable **

        public long getSelectionTime() {
            return Selectable.NOT_SELECTED;
        }

        public void setSelect(long globalClockTimestamp) {
            drawStudentRay(this);
        }
        
        private static final long serialVersionUID = 1L;
    }
    
    /**
     * An FDT wrapper for one data element visualized as
     * a single-item array.
     */
    private class Slot implements FDT, Array {
        private VirtualObject contents;
        
        public Slot() {
            this.contents = new VirtualObject(this, "contents");
        }
        
        public Slot(Object o) {
            this();
            this.setElement(o);
        }

        // ** Implementation of interface FDT **

        public Object getElement() {
            return this.contents.getObject();
        }

        public void setElement(Object o) {
            this.contents.setObject(o);
        }

        public boolean equals(Object obj) {
            if (obj instanceof Slot) {
                if (this.getElement() == null) {
                    if (((Slot)obj).getElement() == null) {
                        return true;
                    } else {
                        return false;
                    }
                } else {
                    return this.getElement().equals(((Slot)obj).getElement());
                }
            } else {
                return false;
            }
        }

        
        // ** Implementation of interface Array **

        public int getFirst() {
            return 0;
        }

        public int getLast() {
            return 0;
        }

        public Object getObject(int address) {
            if (address == 0) {
                return this.getElement();
            } else {
                return null;
            }
        }

        public void setFirst(int i) {
            // disabled
        }

        public void setLast(int i) {
            // disabled
        }

        public void setObject(Object o, int address) {
            // disabled
        }

        private static final long serialVersionUID = -3420257164917054541L;
    }

    
    private class AnswerPanel implements FDT, Struct {
        private Slot[] answers;
        
        public AnswerPanel(Slot[] answers) {
            this.answers = answers;
        }

        public Object getField(int location) {
            switch (location) {
                case 0 : return "";
                case 1 : return this.answers[0].getElement();
                case 2 : return (this.answers[1].getElement() == null) ? "ANSWER" : ((((Boolean)this.answers[1].getElement()).booleanValue() == true) ? "YES" : "NO");
                case 3 : return this.answers[2].getElement();
                case 4 : return (this.answers[3].getElement() == null) ? "ANSWER" : ((((Boolean)this.answers[3].getElement()).booleanValue() == true) ? "YES" : "NO");
                default : return null;
            }
        }

        public String getFieldName(int location) {
            switch (location) {
                case 0 : return "";
                case 1 : return "Ray 1 intersections:";
                case 2 : return "Is point 1 inside?";
                case 3 : return "Ray 2 intersections:";
                case 4 : return "Is point 2 inside?";
                default : return null;
            }
        }

        public String[] getFieldNames() {
            return new String[]{ "Answers", 
                                 "Ray 1 intersections:",
                                 "Is point 1 inside?",
                                 "Ray 2 intersections:",
                                 "Is point 2 inside?"};
        }

        public Object[] getFields() {
            return new Object[]{ "Answers", 
                                 this.answers[0], 
                                 (this.answers[1].getElement() == null) ? "ANSWER" : ((((Boolean)this.answers[1].getElement()).booleanValue() == true) ? "YES" : "NO"),
                                 this.answers[2],
                                 (this.answers[3].getElement() == null) ? "ANSWER" : ((((Boolean)this.answers[3].getElement()).booleanValue() == true) ? "YES" : "NO")};
        }

        public void setField(Object value, int location) {
        }

        public Object getElement() {
            return null;
        }

        public void setElement(Object o) {
        }

        public boolean equals(Object obj) {
            if (obj instanceof AnswerPanel) {
                AnswerPanel answerPanel = (AnswerPanel)obj;
                for (int i = 0; i < this.answers.length; i++) {
                    if (!this.answers[i].equals(answerPanel.answers[i])) {
                        return false;
                    }
                }
                return true;
            } else {
                return false;
            }
        }
        
        private static final long serialVersionUID = 1009938486841895362L;
    }
    
    private class MySpinner extends JSpinner {
        private Integer rayNumber;
        private SpinnerChangeListener spinnerListener; 
        
        public MySpinner(SpinnerModel spinnerModel, int rayNumber) {
            super(spinnerModel);
            this.rayNumber = new Integer(rayNumber);
            this.spinnerListener = null;
        }
        
        public void setSpinnerListener(SpinnerChangeListener spinnerListener) {
            this.spinnerListener = spinnerListener;
        }
        
        public void validate() {
            if (((Integer)intersections[this.rayNumber.intValue()].getElement()).intValue() != ((Integer)this.getValue()).intValue()) {
                if (this.spinnerListener != null) {
                    this.spinnerListener.setIgnoreStateChange(true);
                }
                this.setValue(intersections[this.rayNumber.intValue()].getElement());
            }
            super.validate();
        }
        
        private static final long serialVersionUID = 1274090539886238015L;
    }
    
    private class MyComboBox extends JComboBox {
        private Integer rayNumber;
        private AnswerBoxItemListener itemListener; 
        
        public MyComboBox(Object[] items, int rayNumber) {
            super(items);
            this.rayNumber = new Integer(rayNumber);
            this.itemListener = null;
        }
        
        public void setItemListener(AnswerBoxItemListener itemListener) {
            this.itemListener = itemListener;
        }

        public void validate() {
            if (!(((answers[this.rayNumber.intValue()].getElement() == null) &&  this.getModel().getSelectedItem().equals("ANSWER")) ||
                  ((answers[this.rayNumber.intValue()].getElement() != null) && (((Boolean)answers[this.rayNumber.intValue()].getElement()).booleanValue()) &&  this.getModel().getSelectedItem().equals("YES")) ||
                  ((answers[this.rayNumber.intValue()].getElement() != null) && (!((Boolean)answers[this.rayNumber.intValue()].getElement()).booleanValue()) &&  this.getModel().getSelectedItem().equals("NO")))) {
                if (this.itemListener != null) {
                    this.itemListener.setIgnoreStateChange(true);
                }
                if (answers[this.rayNumber.intValue()].getElement() == null) {
                    this.getModel().setSelectedItem(new String("ANSWER"));
                } else if (((Boolean)answers[this.rayNumber.intValue()].getElement()).booleanValue()) {
                    this.getModel().setSelectedItem(new String("YES"));
                } else {
                    this.getModel().setSelectedItem(new String("NO"));
                }
                if (this.itemListener != null) {
                    this.itemListener.setIgnoreStateChange(false);
                }
            }
            super.validate();
        }

        private static final long serialVersionUID = 3972698053568077578L;
    }
    
    
    private static final long serialVersionUID = -5627366663682954388L;
}
