package content.exercises.sda;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Component;
import java.awt.GridBagConstraints;
import java.awt.Rectangle;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Point2D;
import java.util.Iterator;
import java.util.Random;

import matrix.animation.Animator;
import matrix.structures.FDT.FDT;
import matrix.structures.simulationextensions.Selectable;
import matrix.structures.spatial.PaintingStyleDecorator;
import matrix.structures.spatial.FDT.probe.Circle;
import matrix.structures.spatial.FDT.probe.CubicBezier;
import matrix.structures.spatial.FDT.probe.LineSegment;
import matrix.structures.spatial.FDT.probe.Point;
import matrix.structures.spatial.FDT.probe.Polygon;
import matrix.structures.spatial.FDT.probe.Triangle;
import matrix.util.Note;
import content.exercises.sda.structures.GeometryFunctions;
import content.exercises.sda.structures.HashTable;
import content.exercises.sda.structures.SimpleStack;
import content.exercises.sda.structures.SinglyLinkedList;
import content.exercises.sda.structures.Triangulation;
import content.exercises.sda.structures.Triangulation.VisualElements;

/**
 * <p>An exercise for constructing the crust and anti-crust/skeleton 
 * of a point set as described by C. Gold and J. Snoeyink
 * (A One-Step Crust and Skeleton Extraction Algorithm (2003)).</p>
 * 
 * <p>The crust and skeleton are extracted by first constructing 
 * the Delaunay triangulation and Voronoi diagram of the point set, 
 * after which each of the Delaunay-Voronoi-edge pairs is either 
 * assigned to the crust or anti-crust according to a local test. 
 * The delaunay edges of the dual pairs assigned to the crust form 
 * the actual crust and the voronoi edges of the other pairs form 
 * the anti-crust, the endo- and exo-skeleton.</p>
 * 
 * <p>According to the local test a Delaunay-Voronoi-pair belongs 
 * to the crust if the Delaunay edge has a circle empty of the 
 * corresponding Voronoi vertices and the Voronoi edge intersects 
 * the Delaunnay edge, otherwise it's part of the skeleton/anti-crust. 
 * These conditions can be checked with the inCircle-test by examining 
 * a circle drawn through the two Delaunay vertices and one of the two 
 * Voronoi vertices. If the other Voronoi vertex is inside the circle 
 * or the Voronoi edge does not intersect the Delaunay edge, 
 * the Delaunay-Voronoi-pair belongs to the skeleton, otherwise it's 
 * part of the crust.</p>
 * 
 * <p>To perform the test, the student can select any three points and 
 * the points located inside the circumcircle of these three points 
 * are highlighted.</p>
 * 
 * @author Juha Helminen
 * @see Triangulation VoronoiConstruction
 */
public class PolygonSkeleton extends VoronoiConstruction {

    // For debugging
    private static final boolean STATIC_SEED = false;
    private static final long STATIC_SEED_VALUE = 1188206555187L;

    // The number of points.
    private static final int INPUT_SIZE = 12;
    // The minimum allowed distance between voronoi vertices.
    private static final int VORONOI_VERTEX_MIN_DISTANCE = 16;

    // The size of the exercise visualization.
    private static final int WIDTH = 550;
    private static final int HEIGHT = 420;

    private long seed = System.currentTimeMillis();

    // The style for the points that are being inserted in the triangulation.
    protected static final Color VERTEX_SELECTED_COLOR = Color.RED;
    protected static final Color VERTEX_FOCUS_COLOR = Color.RED;
    protected static final Color VERTEX_SELECTED_FOCUS_COLOR = Color.MAGENTA;
    protected static final Color SHAPE_COLOR = new Color(110, 110, 110);
    protected static final Color DELAUNAY_COLOR = Color.WHITE;
    protected static final Color DELAUNAY_FOCUS_COLOR = Color.RED;
    protected static final Color VORONOI_COLOR = new Color(120, 120, 130);
    protected static final Color VORONOI_SELECTED_COLOR = new Color(70, 0, 0);
    protected static final Color VORONOI_FOCUS_COLOR = Color.RED;
    protected static final Color VORONOI_SELECTED_FOCUS_COLOR = Color.RED; 
    protected static final Color VORONOI_VERTEX_COLOR = new Color(100, 100, 150);
    protected static final Color VORONOI_VERTEX_SELECTED_COLOR = Color.RED;
    protected static final Color VORONOI_VERTEX_FOCUS_COLOR = Color.RED;
    protected static final Color VORONOI_VERTEX_SELECTED_FOCUS_COLOR = Color.MAGENTA;
    protected static final Color WRONG_SKELETON_COLOR = Color.RED;
    protected static final PaintingStyleDecorator vertexDecorator;
    protected static final PaintingStyleDecorator voronoiVertexDecorator;
    protected static final PaintingStyleDecorator skeletonDecorator;
    protected static final PaintingStyleDecorator voronoiDecorator;
    protected static final PaintingStyleDecorator crustDecorator;
    protected static final PaintingStyleDecorator delaunayDecorator;
    protected static final PaintingStyleDecorator missingSkeletonDecorator;
    protected static final PaintingStyleDecorator wrongSkeletonDecorator;
    protected static final PaintingStyleDecorator shapeDecorator;
    static {
        vertexDecorator = new PaintingStyleDecorator();
        vertexDecorator.setLineColor(POINT_COLOR);
        vertexDecorator.setSelectionLineColor(VERTEX_SELECTED_COLOR);
        vertexDecorator.setFocusLineColor(VERTEX_FOCUS_COLOR);
        vertexDecorator.setSelectionFocusLineColor(VERTEX_SELECTED_FOCUS_COLOR);
        vertexDecorator.setStroke(2, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL);

        voronoiVertexDecorator = new PaintingStyleDecorator();
        voronoiVertexDecorator.setLineColor(VORONOI_VERTEX_COLOR);
        voronoiVertexDecorator.setSelectionLineColor(VORONOI_VERTEX_SELECTED_COLOR);
        voronoiVertexDecorator.setFocusLineColor(VORONOI_VERTEX_FOCUS_COLOR);
        voronoiVertexDecorator.setSelectionFocusLineColor(VORONOI_VERTEX_SELECTED_FOCUS_COLOR);
        voronoiVertexDecorator.setStroke(2, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL);

        skeletonDecorator = new PaintingStyleDecorator();
        skeletonDecorator.setLineColor(VORONOI_SELECTED_COLOR);
        skeletonDecorator.setSelectionLineColor(VORONOI_SELECTED_COLOR);
        skeletonDecorator.setFocusLineColor(VORONOI_FOCUS_COLOR);
        skeletonDecorator.setSelectionFocusLineColor(VORONOI_SELECTED_FOCUS_COLOR);
        skeletonDecorator.setStroke(2, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL);

        float[] delaunayDashPattern = { 2, 7 };
        delaunayDecorator = new PaintingStyleDecorator();
        delaunayDecorator.setLineColor(DELAUNAY_COLOR);
        delaunayDecorator.setSelectionLineColor(DELAUNAY_COLOR);
        delaunayDecorator.setFocusLineColor(DELAUNAY_FOCUS_COLOR);
        delaunayDecorator.setSelectionFocusLineColor(DELAUNAY_FOCUS_COLOR);
        delaunayDecorator.setStroke(1, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL, 0, delaunayDashPattern, 0);

        float[] voronoiDashPattern = { 5, 4 };
        voronoiDecorator = new PaintingStyleDecorator();
        voronoiDecorator.setLineColor(VORONOI_COLOR);
        voronoiDecorator.setFocusLineColor(VORONOI_FOCUS_COLOR);
        voronoiDecorator.setStroke(1, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL, 5, voronoiDashPattern, 0);

        crustDecorator = new PaintingStyleDecorator();
        crustDecorator.setLineColor(DELAUNAY_COLOR);
        crustDecorator.setFocusLineColor(DELAUNAY_FOCUS_COLOR);
        crustDecorator.setStroke(2, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL);

        float[] missingCrustDashPattern = { 2, 7 };
        missingSkeletonDecorator = new PaintingStyleDecorator();
        missingSkeletonDecorator.setLineColor(VORONOI_SELECTED_COLOR);
        missingSkeletonDecorator.setStroke(2, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL, 0, missingCrustDashPattern, 0);

        wrongSkeletonDecorator = new PaintingStyleDecorator();
        wrongSkeletonDecorator.setLineColor(WRONG_SKELETON_COLOR);
        wrongSkeletonDecorator.setStroke(2, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL);
        
        shapeDecorator = new PaintingStyleDecorator();
        shapeDecorator.setLineColor(SHAPE_COLOR);
        shapeDecorator.setFill(false);
        shapeDecorator.setStroke(3, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL);
    }

    
    protected class StudentVisualElements extends VisualElements {

        protected Point newVertex(java.awt.Point vertex) {
            return new SelectablePoint(vertex, vertexDecorator);
        }

        protected LineSegment newDelaunayEdge(java.awt.Point orig, java.awt.Point dest) {
            return new SelectableDelaunayLineSegment(orig, dest, crustDecorator);
        }

        protected Triangle newTriangle(java.awt.Point[] vertices) {
            return new Triangle(vertices);
        }

        protected Point newVoronoiVertex(java.awt.Point vertex) {
            return new SelectablePoint(vertex, voronoiVertexDecorator);
        }

        protected LineSegment newVoronoiEdge(java.awt.Point orig, java.awt.Point dest) {
            return new SelectableVoronoiLineSegment(orig, dest, voronoiDecorator);
        }

        protected Polygon newVoronoiRegion(java.awt.Point[] vertices) {
            return new Polygon(vertices, skeletonDecorator);
        }

        protected void hideDelaunay() {}

        protected void showDelaunay() {}

        protected void hideVoronoi() {}

        private static final long serialVersionUID = -428811880509503010L;
    }

    
    protected class ModelVisualElements extends StudentVisualElements {
        protected LineSegment newDelaunayEdge(java.awt.Point orig, java.awt.Point dest) {
            return new LineSegment(orig, dest, crustDecorator);
        }

        protected LineSegment newVoronoiEdge(java.awt.Point orig, java.awt.Point dest) {
            return new LineSegment(orig, dest, skeletonDecorator);
        }

        private static final long serialVersionUID = 6915993847297893L;
    }

    
    /*
     * (non-Javadoc)
     * 
     * @see content.exercises.sda.VoronoiConstruction#getInitialStructures()
     */
    public FDT[] getInitialStructures() {
        SimpleStack problemVis = new SimpleStack(new Rectangle(WIDTH, HEIGHT, 0, 0));
        return new FDT[] { problemVis };
    }

    private FDT[] getInitialStructures(VisualElements visualElements) {
        SimpleStack problemVis = (SimpleStack) this.getInitialStructures()[0];
        Triangulation triangulation = new Triangulation(visualElements, WIDTH, HEIGHT);
        triangulation.setDrawVoronoi(Triangulation.VORONOI_EDGES);
        triangulation.setDrawTriangles(false);
        for (int i = 0; i < this.input.length; i++) {
            triangulation.addVertex(this.input[i]);
        }

        problemVis.addSubArea(triangulation);
        
        SimpleStack shape = new SimpleStack();
        problemVis.addSubArea(shape);

        java.awt.Point[][] initialTriangles = GeometryFunctions.delaunay(this.input);
        for (int i = 0; i < initialTriangles.length; i++) {
            triangulation.addEdge(initialTriangles[i][0], initialTriangles[i][1]);
            triangulation.addEdge(initialTriangles[i][1], initialTriangles[i][2]);
            triangulation.addEdge(initialTriangles[i][2], initialTriangles[i][0]);
        }

        return new FDT[] { problemVis, triangulation, shape };
    }

    
    /*
     * (non-Javadoc)
     * 
     * @see content.exercises.sda.VoronoiConstruction#init()
     */
    public FDT[] init() {
        // For debugging
        Note.out(this, "The random number generator seed for this exercise is: " + this.getSeed());
        Random r = new Random(this.getSeed());

        // Generate the exercise input.
        final int MAX_TRIES = 1000;
        final int MAX_TRIES2 = 50;
        int tryCounter = 0;
        int tryCounter2 = 0;
        final double CENTER_RADIUS = 100;
        final int MIN_DEVIATION = 40;
        final int MAX_DEVIATION = 70;
        final int DEVIATION_RADIUS = 110;
        final int ANGLE_DEVIATION_SUM_MAX = 100;
        double angle;
        boolean[] deviation;
        int x, y;
        int trans, lastDev, lastTrans;
        LineSegment[] crust;
        FDT[] structures;
        java.awt.Point[][] initialTriangles;
        java.awt.Point[] centers;
        while (true) {
            try {
                // The set of points is generated by slightly translating in varying amounts 
                // points chosen around an ellipse.
                this.input = new java.awt.Point[INPUT_SIZE];
                deviation = new boolean[DEVIATION_RADIUS];
                lastDev = DEVIATION_RADIUS / 2;
                lastTrans = MIN_DEVIATION + (MAX_DEVIATION - MIN_DEVIATION) / 2;
                for (int i = 0; i < this.input.length; i++) {
                    angle = ((2 * Math.PI) / INPUT_SIZE) * i;
                    trans = (r.nextBoolean() ? 1 : -1) * (r.nextInt(MAX_DEVIATION - MIN_DEVIATION + 1) + MIN_DEVIATION);
                    tryCounter2 = 0;
                    while ((Math.abs(lastTrans - trans) > ANGLE_DEVIATION_SUM_MAX) || 
                           (lastDev + trans < 0) || 
                           (lastDev + trans >= DEVIATION_RADIUS) || 
                           (deviation[lastDev + trans])) {
                        trans = (r.nextBoolean() ? 1 : -1) * (r.nextInt(MAX_DEVIATION - MIN_DEVIATION + 1) + MIN_DEVIATION);
                        if (tryCounter2++ > MAX_TRIES2) {
                            throw new Exception("Point generation failed.");
                        }
                    }
                    lastTrans = trans;
                    lastDev = lastDev + trans;
                    deviation[lastDev] = true;
                    x = WIDTH / 2 + (int)Math.floor(Math.cos(angle) * (CENTER_RADIUS * 1.2 + lastDev));
                    y = HEIGHT / 2 + (int)Math.floor(Math.sin(angle) * (CENTER_RADIUS + lastDev));
                    this.input[i] = new java.awt.Point(x, y);
                }

                // Test the exercise input for the illegal case of 4
                // (or more) cocircular/concyclic points.
                initialTriangles = GeometryFunctions.delaunay(this.input);
                if (initialTriangles == null) {
                    throw new Exception("4 cocircular points.");
                }

                // Check that the voronoi vertices aren't too close to each other.
                centers = new java.awt.Point[initialTriangles.length];
                Ellipse2D circumcircle;
                for (int i = 0; i < centers.length; i++) {
                    circumcircle = GeometryFunctions.circumcircle(initialTriangles[i]);
                    centers[i] = new java.awt.Point((int)circumcircle.getCenterX(), 
                                                    (int)circumcircle.getCenterY());
                }
                for (int i = 0; i < centers.length; i++) {
                    for (int j = i + 1; j < centers.length; j++) {
                        if (centers[i].distance(centers[j]) < VORONOI_VERTEX_MIN_DISTANCE) {
                            throw new Exception("Voronoi vertices are too close to each other.");
                        }
                    }
                }

                // Check that the crust forms a simple polygon.
                structures = this.getInitialStructures(new StudentVisualElements());
                this.triangulation = (Triangulation) structures[1];
                crust = this.triangulation.crust();
                boolean[] visited = new boolean[crust.length];
                visited[0] = true;
                int visitedCount = 1;
                java.awt.Point start, next;
                start = crust[0].getP1();
                next = crust[0].getP2();
                boolean nextFound;
                
                while (visitedCount < crust.length) {
                    nextFound = false;
                    for (int i = 1; i < crust.length; i++) {
                        if (!visited[i]) {
                            if (crust[i].getP1().equals(next)) {
                                next = crust[i].getP2();
                                visited[i] = true;
                                visitedCount++;
                                nextFound = true;
                                break;
                            } else if (crust[i].getP2().equals(next)) {
                                next = crust[i].getP1();
                                visited[i] = true;
                                visitedCount++;
                                nextFound = true;
                                break;
                            }
                        }
                    }
                    if (!nextFound) {
                        throw new Exception("The crust is not a simple polygon.");
                    }
                }
                if (!start.equals(next)) {
                    throw new Exception("The crust is not a simple polygon.");
                }
                
                // success
                break;
            } catch (Exception e) {
                if (tryCounter++ >= MAX_TRIES) {
                    Note.err(this, e.getMessage());
                    Note.show(this,
                              "Could not generate input for the exercise. Please restart the exercise.");
                    return null;
                }
            }
        }

        SimpleStack problemVis = (SimpleStack) structures[0];
        this.circumcircleVis = new SimpleStack();
        problemVis.addSubArea(this.circumcircleVis);

        SimpleStack spline;
        spline = (SimpleStack)structures[2];
        Point2D[][] bezier = GeometryFunctions.splineInterpolation(this.input);
        for (int i = 0; i < bezier.length; i++) {
            spline.ADTPush(new CubicBezier(bezier[i][0].getX(), bezier[i][0].getY(), 
                                           bezier[i][1].getX(), bezier[i][1].getY(), 
                                           bezier[i][2].getX(), bezier[i][2].getY(), 
                                           bezier[i][3].getX(), bezier[i][3].getY(), 
                                           shapeDecorator));
        }
        
        return new FDT[] { problemVis };
    }

    
    public FDT[] solve() {
        this.correctEdges = 0;
        this.incorrectEdges = 0;
        this.missingEdges = 0;
        FDT[] modelStructures = this.getInitialStructures(new ModelVisualElements());
        FDT[] comparisonStructures = this.getInitialStructures(new StudentVisualElements());

        Triangulation modelTriangulation = (Triangulation) modelStructures[1];
        modelTriangulation.setDrawVoronoi(Triangulation.VORONOI_EDGES);
        modelTriangulation.setDrawVoronoiVertices(true);
        modelTriangulation.setDrawOnlyCrust(true);
        Triangulation comparisonTriangulation = (Triangulation) comparisonStructures[1];
        comparisonTriangulation.setDrawVoronoi(Triangulation.VORONOI_EDGES);
        comparisonTriangulation.setDrawVoronoiVertices(true);

        SimpleStack spline;
        spline = (SimpleStack)modelStructures[2];
        Point2D[][] bezier = GeometryFunctions.splineInterpolation(this.input);
        for (int i = 0; i < bezier.length; i++) {
            spline.ADTPush(new CubicBezier(bezier[i][0].getX(), bezier[i][0].getY(), 
                                           bezier[i][1].getX(), bezier[i][1].getY(), 
                                           bezier[i][2].getX(), bezier[i][2].getY(), 
                                           bezier[i][3].getX(), bezier[i][3].getY(), 
                                           shapeDecorator));
        }
        
        LineSegment[] modelSkeleton = modelTriangulation.skeleton();
        
        // The number of skeleton edges in the correctly constructed skeleton. 
        this.edges = modelSkeleton.length;
        Iterator studentSkeleton;
        LineSegment skeletonEdge, delaunayEdge;
        boolean foundEdge;
        for (int i = 0; i < modelSkeleton.length; i++) {
            studentSkeleton = this.selectedLines.iteratorKeys();
            foundEdge = false;
            while (studentSkeleton.hasNext()) {
                skeletonEdge = (LineSegment)studentSkeleton.next();
                if ((modelSkeleton[i].getP1().equals(skeletonEdge.getP1()) && 
                     modelSkeleton[i].getP2().equals(skeletonEdge.getP2())) || 
                    (modelSkeleton[i].getP1().equals(skeletonEdge.getP2()) && 
                     modelSkeleton[i].getP2().equals(skeletonEdge.getP1()))) {
                    foundEdge = true;
                    break;
                }
            }

            comparisonTriangulation.getDelaunayEdge(modelSkeleton[i].getP1(), 
                                                    modelSkeleton[i].getP2()).setPaintingStyleDecorator(delaunayDecorator);
            if (foundEdge) {
                // The student has chosen a correct edge.
                delaunayEdge = modelTriangulation.getDelaunayEdge(modelSkeleton[i].getP1(), 
                                                                  modelSkeleton[i].getP2());
                comparisonTriangulation.getVoronoiEdge(delaunayEdge.getP1(), 
                                                       delaunayEdge.getP2()).setPaintingStyleDecorator(skeletonDecorator);
                this.correctEdges++;
            } else {
                // The student has left out a skeleton edge.
                delaunayEdge = modelTriangulation.getDelaunayEdge(modelSkeleton[i].getP1(), 
                                                                  modelSkeleton[i].getP2());
                comparisonTriangulation.getVoronoiEdge(delaunayEdge.getP1(), 
                                                       delaunayEdge.getP2()).setPaintingStyleDecorator(missingSkeletonDecorator);
                this.missingEdges++;
            }
        }

        studentSkeleton = this.selectedLines.iteratorKeys();
        while (studentSkeleton.hasNext()) {
            skeletonEdge = (LineSegment)studentSkeleton.next();
            foundEdge = false;
            for (int i = 0; i < modelSkeleton.length; i++) {
                if ((modelSkeleton[i].getP1().equals(skeletonEdge.getP1()) && 
                     modelSkeleton[i].getP2().equals(skeletonEdge.getP2())) || 
                    (modelSkeleton[i].getP1().equals(skeletonEdge.getP2()) && 
                     modelSkeleton[i].getP2().equals(skeletonEdge.getP1()))) {
                    foundEdge = true;
                    break;
                }
            }
            if (!foundEdge) {
                // The student has selected an edge that does not belong to the skeleton.
                delaunayEdge = modelTriangulation.getDelaunayEdge(skeletonEdge.getP1(), 
                                                                  skeletonEdge.getP2());
                comparisonTriangulation.getVoronoiEdge(delaunayEdge.getP1(), 
                                                       delaunayEdge.getP2()).setPaintingStyleDecorator(wrongSkeletonDecorator);
                this.incorrectEdges++;
            }
        }

        return new FDT[] { modelStructures[0], comparisonStructures[0] };
    }

    public String getDescription() {
        return "Construct the skeleton and the crust of the given point set.";
    }

    public String[] getStructureNames() {
        return new String[] { "Delaunay Triangulation / Voronoi Diagram" };
    }

    public String[] getModelAnswerNames() {
        return new String[] {
                "The original shape and the crust and skeleton of the sampled points",
                "Correct (solid brown), incorrect (solid red) and missing (dashed brown) skeleton edges." };
    }

    /*
     * (non-Javadoc)
     * 
     * @see content.interfaces.SimulationExercise#getSeed()
     */
    public long getSeed() {
        // The seed might be set to a constant value for debugging purposes.
        if (STATIC_SEED) {
            return PolygonSkeleton.STATIC_SEED_VALUE;
        } else {
            return this.seed;
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see content.interfaces.SimulationExercise#setSeed(long)
     */
    public void setSeed(long seed) {
        this.seed = seed;
    }

    public Component[] getAWTComponents() {
        return new Component[0];
    }

    public GridBagConstraints[] getAWTComponentConstraints() {
        return new GridBagConstraints[0];
    }

    
    // ** Methods for the circumcircle test **

    private void circumcircle(java.awt.Point p1, java.awt.Point p2, java.awt.Point p3) {
        Animator a = Animator.getActiveAnimator();
        a.startOperation();

        // Draw circle
        Ellipse2D circumcircle = GeometryFunctions.circumcircle(new java.awt.Point[] { p1, p2, p3 });
        Circle circle = new Circle(circumcircle.getCenterX(), 
                                   circumcircle.getCenterY(), 
                                   circumcircle.getWidth() / 2);
        circle.setPaintingStyleDecorator(circumcircleDecorator);
        this.circumcircleVis.ADTPush(circle);

        // Color vertices if they are inside the circle.
        colorPoints(p1, p2, p3);

        a.endOperation();
    }

    private void clearCircle() {
        if (!this.circumcircleVis.ADTIsEmpty()) {
            Animator a = Animator.getActiveAnimator();
            a.startOperation();

            // Clear the circumcircle.
            this.circumcircleVis.clear();

            // Clear point colors.
            clearPointColors();

            a.endOperation();
        }
    }

    protected void clearPointColors() {
        // Clear point coloring.
        Iterator vertices = PolygonSkeleton.this.triangulation.vertices();
        while (vertices.hasNext()) {
            ((Point)vertices.next()).setPaintingStyleDecorator(vertexDecorator);
        }
        // Clear voronoi vertex coloring.
        vertices = PolygonSkeleton.this.triangulation.voronoiVertices();
        while (vertices.hasNext()) {
            ((Point)vertices.next()).setPaintingStyleDecorator(voronoiVertexDecorator);
        }
    }

    // Color the points that are located inside the given triangle's circumcircle.
    private void colorPoints(java.awt.Point p1, java.awt.Point p2, java.awt.Point p3) {
        java.awt.Point[] points = new java.awt.Point[] { p1, p2, p3 };
        GeometryFunctions.ccwOrder(points);
        Iterator vertices = PolygonSkeleton.this.triangulation.vertices();
        Point point;
        whileloop: while (vertices.hasNext()) {
            point = (Point) vertices.next();
            if (p1.equals(point.getPoint())) {
                point.setPaintingStyleDecorator(circlepointDecorator);
                continue whileloop;
            }
            if (p2.equals(point.getPoint())) {
                point.setPaintingStyleDecorator(circlepointDecorator);
                continue whileloop;
            }
            if (p3.equals(point.getPoint())) {
                point.setPaintingStyleDecorator(circlepointDecorator);
                continue whileloop;
            }

            if (GeometryFunctions.inCircleTest(point.getPoint(), points[0], points[1], points[2]) > 0) {
                point.setPaintingStyleDecorator(inpointDecorator);
            }
        }
        vertices = PolygonSkeleton.this.triangulation.voronoiVertices();
        whileloop: while (vertices.hasNext()) {
            point = (Point) vertices.next();
            if (p1.equals(point.getPoint())) {
                point.setPaintingStyleDecorator(circlepointDecorator);
                continue whileloop;
            }
            if (p2.equals(point.getPoint())) {
                point.setPaintingStyleDecorator(circlepointDecorator);
                continue whileloop;
            }
            if (p3.equals(point.getPoint())) {
                point.setPaintingStyleDecorator(circlepointDecorator);
                continue whileloop;
            }

            if (GeometryFunctions.inCircleTest(point.getPoint(), points[0], points[1], points[2]) > 0) {
                point.setPaintingStyleDecorator(inpointDecorator);
            }
        }
    }

    
    // ** The class and methods for selectable points. **

    /**
     * The currently selected points.
     */
    protected SinglyLinkedList selectedPoints = new SinglyLinkedList();
    protected static long SELECTED = Selectable.clock.getNextTimeStamp();

    protected synchronized void selectPoint(Point p) {
        if (!this.selectedPoints.contains(p)) {
            this.selectedPoints.insert(p);
        }
    }

    protected synchronized void clearPointSelection(Point p) {
        this.selectedPoints.delete(p);
    }

    protected synchronized void clearPointSelections() {
        this.selectedPoints.clear();
    }

    protected synchronized boolean isPointSelected(Point p) {
        return this.selectedPoints.contains(p);
    }

    protected synchronized SinglyLinkedList getSelectedPoints() {
        return this.selectedPoints;
    }

    /**
     * A point that implements the selection functionality through the
     * Selectable interface.
     * 
     * @see Selectable
     */
    class SelectablePoint extends Point implements Selectable {

        public SelectablePoint(java.awt.Point point) {
            super(point);
            // Set focusable to make selecting possible.
            this.setDraggable(true);
        }

        public SelectablePoint(java.awt.Point point, PaintingStyleDecorator psd) {
            this(point);
            this.setPaintingStyleDecorator(psd);
        }

        /*
         * (non-Javadoc)
         * 
         * @see matrix.structures.simulationextensions.Selectable#getSelectionTime()
         */
        public long getSelectionTime() {
            if (this.isSelected()) {
                return PolygonSkeleton.SELECTED;
            } else {
                return Selectable.NOT_SELECTED;
            }
        }

        /*
         * (non-Javadoc)
         * 
         * @see matrix.structures.simulationextensions.Selectable#setSelect(long)
         */
        public void setSelect(long globalClockTimestamp) {
            Animator a = Animator.getActiveAnimator();
            a.startOperation();
            if (isSelected()) {
                // The point is already selected -> clear the selection.
                PolygonSkeleton.this.clearCircle();
                this.clearSelection();
            } else if (globalClockTimestamp != Selectable.NOT_SELECTED) {
                PolygonSkeleton.this.clearCircle();
                PolygonSkeleton.this.selectPoint(this);
                SinglyLinkedList selectedPoints = PolygonSkeleton.this.getSelectedPoints();
                if (selectedPoints.size() == 3) {
                    // Thre points have been selected -> draw a circle through them.
                    Iterator points = selectedPoints.iterator();
                    SelectablePoint point1 = (SelectablePoint) points.next();
                    SelectablePoint point2 = (SelectablePoint) points.next();
                    SelectablePoint point3 = (SelectablePoint) points.next();
                    point1.clearSelection();
                    point2.clearSelection();
                    point3.clearSelection();
                    PolygonSkeleton.this.circumcircle(point1.getPoint(), point2.getPoint(), point3.getPoint());
                }
            }
            a.endOperation();
        }

        private void clearSelection() {
            // Remove the stored selection.
            PolygonSkeleton.this.clearPointSelection(this);
        }

        private boolean isSelected() {
            return PolygonSkeleton.this.isPointSelected(this);
        }

        public boolean equals(Object k) {
            return (k == this);
        }

        public int hashCode() {
            return this.getIdentityString().hashCode();
        }

        private static final long serialVersionUID = -94066704145446270L;
    }

    // ** The class and methods for a selectable line segment. **

    /**
     * The currently selected points.
     */
    protected HashTable selectedLines = new HashTable();

    protected synchronized void selectLine(LineSegment line) {
        if (!this.selectedLines.containsKey(line)) {
            this.selectedLines.put(line, null);
        }
    }

    protected synchronized void clearLineSelection(LineSegment line) {
        this.selectedLines.remove(line);
    }

    protected synchronized boolean isLineSelected(LineSegment line) {
        return this.selectedLines.containsKey(line);
    }

    /**
     * A line segment that implements the selection functionality through 
     * the Selectable interface.
     * 
     * @see Selectable
     */
    class SelectableVoronoiLineSegment extends LineSegment implements Selectable {

        public SelectableVoronoiLineSegment(java.awt.Point orig, java.awt.Point dest, PaintingStyleDecorator psd) {
            super(orig, dest, psd);
            // Set focusable to make selecting possible.
            this.setDraggable(true);
        }

        /*
         * (non-Javadoc)
         * 
         * @see matrix.structures.simulationextensions.Selectable#getSelectionTime()
         */
        public long getSelectionTime() {
            if (this.isSelected()) {
                return PolygonSkeleton.SELECTED;
            } else {
                return Selectable.NOT_SELECTED;
            }
        }

        /*
         * (non-Javadoc)
         * 
         * @see matrix.structures.simulationextensions.Selectable#setSelect(long)
         */
        public void setSelect(long globalClockTimestamp) {
            Animator a = Animator.getActiveAnimator();
            a.startOperation();
            if (isSelected()) {
                PolygonSkeleton.this.clearPointSelections();
                PolygonSkeleton.this.clearCircle();
                // The line segment is already selected -> clear the selection.
                this.clearSelection();
                LineSegment delaunayEdge = 
                    PolygonSkeleton.this.triangulation.getDelaunayEdge(this.getP1(), this.getP2());
                this.setPaintingStyleDecorator(voronoiDecorator);
                delaunayEdge.setPaintingStyleDecorator(crustDecorator);
            } else if (globalClockTimestamp != Selectable.NOT_SELECTED) {
                PolygonSkeleton.this.clearPointSelections();
                PolygonSkeleton.this.clearCircle();
                PolygonSkeleton.this.selectLine(this);
                LineSegment delaunayEdge = 
                    PolygonSkeleton.this.triangulation.getDelaunayEdge(this.getP1(), this.getP2());
                this.setPaintingStyleDecorator(skeletonDecorator);
                delaunayEdge.setPaintingStyleDecorator(delaunayDecorator);
            }
            a.endOperation();
        }

        private void clearSelection() {
            // Remove the stored selection.
            PolygonSkeleton.this.clearLineSelection(this);
        }

        private boolean isSelected() {
            return PolygonSkeleton.this.isLineSelected(this);
        }

        public boolean equals(Object k) {
            return (k == this);
        }

        public int hashCode() {
            return this.getIdentityString().hashCode();
        }

        private static final long serialVersionUID = 3166816181043756932L;
    }

    /**
     * A line segment that implements the selection functionality through 
     * the Selectable interface.
     * 
     * @see Selectable
     */
    class SelectableDelaunayLineSegment extends LineSegment implements Selectable {

        public SelectableDelaunayLineSegment(java.awt.Point orig, java.awt.Point dest, PaintingStyleDecorator psd) {
            super(orig, dest, psd);
            // Set focusable to make selecting possible.
            this.setDraggable(true);
        }

        /*
         * (non-Javadoc)
         * 
         * @see matrix.structures.simulationextensions.Selectable#getSelectionTime()
         */
        public long getSelectionTime() {
            return Selectable.NOT_SELECTED;
        }

        /*
         * (non-Javadoc)
         * 
         * @see matrix.structures.simulationextensions.Selectable#setSelect(long)
         */
        public void setSelect(long globalClockTimestamp) {
            Animator a = Animator.getActiveAnimator();
            a.startOperation();
            LineSegment voronoiEdge = 
                PolygonSkeleton.this.triangulation.getVoronoiEdge(this.getP1(), this.getP2());
            if (isSelected(voronoiEdge)) {
                PolygonSkeleton.this.clearPointSelections();
                PolygonSkeleton.this.clearCircle();
                // The line segment is already selected -> clear the selection.
                this.clearSelection(voronoiEdge);
                this.setPaintingStyleDecorator(crustDecorator);
                voronoiEdge.setPaintingStyleDecorator(voronoiDecorator);
            } else if (globalClockTimestamp != Selectable.NOT_SELECTED) {
                PolygonSkeleton.this.clearPointSelections();
                PolygonSkeleton.this.clearCircle();
                PolygonSkeleton.this.selectLine(voronoiEdge);
                this.setPaintingStyleDecorator(delaunayDecorator);
                voronoiEdge.setPaintingStyleDecorator(skeletonDecorator);
            }
            a.endOperation();
        }

        private void clearSelection(LineSegment voronoiEdge) {
            // Remove the stored selection.
            PolygonSkeleton.this.clearLineSelection(voronoiEdge);
        }

        private boolean isSelected(LineSegment voronoiEdge) {
            return PolygonSkeleton.this.isLineSelected(voronoiEdge);
        }

        public boolean equals(Object k) {
            return (k == this);
        }

        public int hashCode() {
            return this.getIdentityString().hashCode();
        }

        private static final long serialVersionUID = -4942773732690065490L;
    }

    private static final long serialVersionUID = -918806677240285998L;
}
