package content.exercises.sda;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Component;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.LayoutManager;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Random;
import java.util.Vector;

import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;

import matrix.animation.Animator;
import matrix.simulation.VisualTypeConf;
import matrix.structures.FDT.FDT;
import matrix.structures.memory.VirtualObject;
import matrix.structures.spatial.PaintingStyleDecorator;
import matrix.structures.spatial.SpatialComparable;
import matrix.structures.spatial.CDT.probe.RTree;
import matrix.structures.spatial.CDT.probe.RTreeNode;
import matrix.structures.spatial.FDT.probe.Polygon;
import matrix.structures.spatial.FDT.probe.SpatialElement;
import matrix.structures.util.SelectionVisitor;
import matrix.util.Application;
import matrix.util.Note;
import content.exercises.sda.structures.ExerciseRTree;
import content.exercises.sda.structures.PointGenerator;
import content.exercises.sda.structures.PolygonGenerator;
import content.exercises.sda.structures.SimpleStack;
import content.exercises.sda.structures.ExerciseRTree.NodeSelector;
import content.interfaces.AWTComponentUtilizer;
import content.interfaces.ConfigureVisualType;
import content.interfaces.LayoutExercise;
import content.interfaces.ModelAnswerNames;
import content.interfaces.SimulationExerciseModel;
import content.interfaces.StyledExercise;
import content.interfaces.SwapBehaviour;

/**
 * Exercise for the insert operation of an R-Tree. 
 * 
 * @see ExerciseRTree, RTree, RTreeNode
 * @author Juha Helminen
 */
public class RTreeInsert implements SimulationExerciseModel, ConfigureVisualType,  
									ModelAnswerNames, LayoutExercise, StyledExercise, 
									SwapBehaviour, AWTComponentUtilizer {

    // For debugging
	private static final boolean STATIC_SEED = false;
	private static final long STATIC_SEED_VALUE = 1170669610657L;

    // The possible input types and input type slection.
    private static final int POLYGON_MAP = 0;
    private static final int POLYGONS = 1;
    private static final int INPUT = POLYGON_MAP;
    
    // The number of polygons in the input stream when POLYGONS is chosen as the input type.
    private static final int INPUT_SIZE = 13;
    
	// These control characteristics of the generated data in the input stream.
	private static final int WIDTH = 270;
	private static final int HEIGHT = 180;
	// R-tree node capacity; must be min <= Sup(MAX/2), e.g. (1, 2) and (2, 3).
	private static final int RTREE_MIN_ENTRIES = 2;
	private static final int RTREE_MAX_ENTRIES = 3;
	// SPLIT_QUADRATIC or SPLIT_LINEAR
	private static final int SPLIT_ALGORITHM = RTreeNode.SPLIT_QUADRATIC;
	// If true, internal nodes are split automatically when needed (maximum number of entries is exceeded).
	// If false, the user has to split the nodes themselves.
	private static final boolean AUTOMATIC_SPLIT_PROPAGATION = false;
	// If true, the user can select and split internal nodes.
	private static final boolean ALLOW_INTERNAL_NODE_SPLITS = true;

	// ** Color and style scheme **
    private static final Color SPATIAL_ELEMENT_COLOR = Color.BLACK;
    private static final Color SPATIAL_ELEMENT_FILL_COLOR = Color.DARK_GRAY;
    private static final Color INDEXED_SPATIAL_ELEMENT_COLOR = Color.WHITE;
    private static final Color INDEXED_SPATIAL_ELEMENT_FILL_COLOR = Color.GREEN;
    private static final Color NEXT_SPATIAL_ELEMENT_COLOR = Color.WHITE;
    private static final Color NEXT_SPATIAL_ELEMENT_FILL_COLOR = Color.GREEN;
    private static final Color SELECTED_SPATIAL_ELEMENT_COLOR = Color.GREEN;
    private static final Color SELECTED_SPATIAL_ELEMENT_FILL_COLOR = Color.GREEN;
    private static final Color AREA_COLOR = Color.GRAY;
    private static final Color AREA_FILL_COLOR = Color.GRAY;
    private static final Color SELECTED_AREA_COLOR = Color.GRAY;
    private static final Color SELECTED_AREA_FILL_COLOR = Color.DARK_GRAY;
    private static final PaintingStyleDecorator areaDecorator;
    private static final PaintingStyleDecorator selectedAreaDecorator;
    private static final PaintingStyleDecorator spatialElementDecorator;
    private static final PaintingStyleDecorator indexedSpatialElementDecorator;
    private static final PaintingStyleDecorator nextSpatialElementDecorator;
    private static final PaintingStyleDecorator selectedSpatialElementDecorator;
    static {
        float[] areaDashPattern = { 4, 2 };
        areaDecorator = new PaintingStyleDecorator();
        areaDecorator.setLineColor(AREA_COLOR);
        areaDecorator.setFillColor(AREA_FILL_COLOR);
        areaDecorator.setTransparency(0.10);
        areaDecorator.setStroke(1, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL, 5, areaDashPattern, 0);

        selectedAreaDecorator = new PaintingStyleDecorator();
        selectedAreaDecorator.setLineColor(SELECTED_AREA_COLOR);
        selectedAreaDecorator.setFillColor(SELECTED_AREA_FILL_COLOR);
        selectedAreaDecorator.setTransparency(0.10);
        selectedAreaDecorator.setStroke(1, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL);

        float[] dashPattern = { 2 , 5};
        spatialElementDecorator = new PaintingStyleDecorator();
        spatialElementDecorator.setLineColor(SPATIAL_ELEMENT_COLOR);
        spatialElementDecorator.setTransparency(0.07);
        spatialElementDecorator.setFillColor(SPATIAL_ELEMENT_FILL_COLOR);
        spatialElementDecorator.setStroke(1, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL, 5, dashPattern, 0);

        indexedSpatialElementDecorator = new PaintingStyleDecorator();
        indexedSpatialElementDecorator.setLineColor(INDEXED_SPATIAL_ELEMENT_COLOR);
        indexedSpatialElementDecorator.setTransparency(0.07);
        indexedSpatialElementDecorator.setFillColor(INDEXED_SPATIAL_ELEMENT_FILL_COLOR);
        indexedSpatialElementDecorator.setStroke(1, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL);
        
        nextSpatialElementDecorator = new PaintingStyleDecorator();
        nextSpatialElementDecorator.setLineColor(NEXT_SPATIAL_ELEMENT_COLOR);
        nextSpatialElementDecorator.setFillColor(NEXT_SPATIAL_ELEMENT_FILL_COLOR);
        nextSpatialElementDecorator.setTransparency(0.07);
        nextSpatialElementDecorator.setStroke(2, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL);

        selectedSpatialElementDecorator = new PaintingStyleDecorator();
        selectedSpatialElementDecorator.setLineColor(SELECTED_SPATIAL_ELEMENT_COLOR);
        selectedSpatialElementDecorator.setFillColor(SELECTED_SPATIAL_ELEMENT_FILL_COLOR);
        selectedSpatialElementDecorator.setTransparency(0.2);
        selectedSpatialElementDecorator.setStroke(2, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL);
    }

    private long seed = System.currentTimeMillis();
    
    
    // ** Exercise Structures **
    // exercise input
    private java.awt.Point[][] polygons;
	// Structures for the student.
    private InputStack inputStream;
	private ExerciseRTree tree;
    
    private transient VirtualDoubleField enlargement;
    private transient VirtualDoubleField area;
	

	// ** Implementation of interface SimulationExerciseModel **
	
	public FDT[] solve() {
		FDT[] f = solveHelp();
		return new FDT[] { f[1] };
	}
	
	public FDT[] makeModelAnswer() {
		FDT[] f = solveHelp();
        SimpleStack problem = new SimpleStack(new Rectangle(WIDTH, HEIGHT, 0, 0));
        problem.insert(f[1]);
		return new FDT[] { f[0], problem, f[0], f[1] };
	}

	private FDT[] solveHelp() {
        // Structures for the model answer.
        RTreeNode.resetIDCounter();
        RTree modelTree = new RTree(RTREE_MIN_ENTRIES, RTREE_MAX_ENTRIES);
        InputStack modelInputStream = (InputStack)getInitialStructures()[0]; 
        modelTree.setPaintingStyleDecorator(areaDecorator);
		Animator a = Animator.getActiveAnimator();
		while (!modelInputStream.ADTIsEmpty()) {
			a.startOperation();
			modelTree.insert(modelInputStream.ADTPop());
			a.endOperation();
		}		
		return new FDT[]{ modelInputStream, modelTree };
	}
	

	// ** Implementation of interface SimulationExercise **
	
	public FDT[] init() {
        // For debugging
        Note.out(this, "The random number generator seed for this exercise is: " + this.getSeed());
        Random r = new Random(this.getSeed());
        
        // ** Generate the input for the exercise. **
        if (INPUT == POLYGON_MAP) {
            this.polygons = 
                PolygonGenerator.findPolygons(
                        PolygonGenerator.generatePolygonMapAsGraph(WIDTH, HEIGHT, 90, 90, 30));
            // Randomize the order of the polygons.
            Vector vector = new Vector(this.polygons.length);
            for (int i = 0; i < this.polygons.length; i++) {
                vector.add(this.polygons[i]);
            }
            for (int i = 0; i < this.polygons.length; i++) {
                this.polygons[i] = (java.awt.Point[])vector.remove(r.nextInt(vector.size()));
            }
        } else /* if (INPUT == POLYGONS)*/{
            final int MAX_TRIES = 200;
            final int BORDER = 2;
            final int MAX_POLYGON_VERTEX_COUNT = 5;
            final int POLYGON_VERTEX_DISTANCE = 30;
            final int POLYGON_SIZE = 80;
            final int MIN_VERTEX_EDGE_DISTANCE = 13;
            int x, y;
            this.polygons = new java.awt.Point[INPUT_SIZE][];
            for (int i = 0; i < this.polygons.length; i++) {
                int tryCounter = 0;
                while (true) {
                    try {
                        x = r.nextInt(WIDTH - BORDER - POLYGON_SIZE + 1) + BORDER;
                        y = r.nextInt(HEIGHT - BORDER - POLYGON_SIZE + 1) + BORDER;
                        this.polygons[i] = PointGenerator.generatePoints(new Rectangle(x,
                                                                                       y,
                                                                                       POLYGON_SIZE,
                                                                                       POLYGON_SIZE),
                                                                         r.nextInt(MAX_POLYGON_VERTEX_COUNT - 3 + 1) + 3, 
                                                                         POLYGON_VERTEX_DISTANCE, 
                                                                         r);
                        // Arrange the points into a polygon.
                        PolygonGenerator.generatePolygon2Peasants(this.polygons[i]);
                        
                        Line2D[] edges;
                        edges = createPolygonEdges(this.polygons[i]);
                        for (int j = 0; j < this.polygons[i].length; j++) {
                            if (!checkDistance(this.polygons[i][j], edges, MIN_VERTEX_EDGE_DISTANCE)) {
                                throw new Exception("Failed to generate polygons.");
                            }
                        }
                        
                        break;
                    } catch (Exception e) {
                        if (tryCounter++ >= MAX_TRIES) {
                            Note.err(this, e.getMessage());
                            Note.show(this, "Could not generate input for the exercise. Please restart the exercise.");
                            return null;
                        }
                    }
                }
            }
        }
        
        ExerciseRTree.resetIDCounter();
        this.tree = 
            new ExerciseRTree(RTREE_MIN_ENTRIES, RTREE_MAX_ENTRIES, new Selector(), areaDecorator);
        this.inputStream = (InputStack)getInitialStructures()[0]; 
        RTreeNode.setSplitAlgorithm(RTreeInsert.SPLIT_ALGORITHM);
        ExerciseRTree.setAutomaticSplitPropagation(RTreeInsert.AUTOMATIC_SPLIT_PROPAGATION);
        ExerciseRTree.setInternalNodeSplits(RTreeInsert.ALLOW_INTERNAL_NODE_SPLITS);
        
        SimpleStack problem = new SimpleStack(new Rectangle(WIDTH, HEIGHT, 0, 0));
        problem.insert(this.tree);

		return new FDT[]{this.inputStream, problem, this.inputStream, this.tree};
	}

    private static Line2D[] createPolygonEdges(java.awt.Point[] vertices) {
        Line2D[] edges = null;
        if(vertices.length > 1) {
            edges = new Line2D[vertices.length];
            for (int i = 0; i < vertices.length - 1; i++) {
                edges[i] = new Line2D.Double(vertices[i], vertices[i + 1]);
            }
            // Close the polygon.
            edges[vertices.length - 1] = new Line2D.Double(vertices[vertices.length - 1], 
                                                           vertices[0]);
        }
        return edges;
    }

    private static boolean checkDistance(Point2D point, 
                                         Line2D[] lines, 
                                         int minDistance) {
        for (int i = 0; i < lines.length; i++) {
            if (lines[i].getP1().equals(point) ||
                lines[i].getP2().equals(point)) {
                continue;
            } else if (lines[i].ptSegDist(point) < minDistance) {
                return false;
            }
        }
        return true;
    }
    
	public void setSeed(long seed) {
		this.seed = seed;
	}

	public long getSeed() {
		// Seed might be set to a constant value for debugging purposes.
		if (STATIC_SEED) {
			return RTreeInsert.STATIC_SEED_VALUE;
		} else {
			return this.seed;
		}
	}
	
	public String[] getStructureNames() {
		return new String[] { "Input", "R-Tree", "Input stream", "R-tree" };
	}

	public FDT[] getAnswer() {
		return new FDT[] { tree };
	}

	public FDT[] getInitialStructures() {
        InputStack inputStack = new InputStack(new Rectangle(WIDTH, HEIGHT, 0, 0), nextSpatialElementDecorator); 
        // Create the polygons.
        Polygon polygon;
        for (int i = 0; i < this.polygons.length; i++) {
            // Start indexing from 1.
            polygon = new Polygon(this.polygons[i], String.valueOf(polygons.length - i));
            polygon.setPaintingStyleDecorator(spatialElementDecorator);
            inputStack.ADTPush(polygon);
        }
        return new FDT[]{inputStack};
	}

	
	// ** Implementation of interface Exercise **
	
	public String getDescription() {
		return "Insert the spatial data from the input stream into " +
			   "the " + RTREE_MIN_ENTRIES + "-" + RTREE_MAX_ENTRIES + "-" + 
			   "R-Tree executing splits when necessary."; 
	}
	
	
	// ** Implementation of interface ConfigureVisualType **
	
	public VisualTypeConf[] conf() {
		// Input stream array representation
		VisualTypeConf stream = new VisualTypeConf();
		stream.setVisualConfValue("matrix.visual.VisualArray", VisualTypeConf.INDEXED, "false");
		stream.enable("matrix.visual.VisualArrayComponent", VisualTypeConf.DRAG_OPERATION);
		stream.enable("matrix.visual.VisualArrayComponent", VisualTypeConf.HIGHLIGHT_OPERATION);
        stream.enable("matrix.visual.VisualStruct-matrix.structures.spatial.FDT.probe.Polygon", VisualTypeConf.MINIMIZED);

		// R-tree tree representation
		VisualTypeConf tree = new VisualTypeConf();
		tree.enable("matrix.visual.VisualLayeredTreeComponent", VisualTypeConf.HIGHLIGHT_OPERATION);
		tree.enable("matrix.visual.VisualKey", VisualTypeConf.HIGHLIGHT_OPERATION);
        tree.enable("matrix.visual.VisualStruct", VisualTypeConf.DRAG_OPERATION);
        tree.enable("matrix.visual.VisualStruct", VisualTypeConf.HIGHLIGHT_OPERATION);
		tree.enable("matrix.visual.VisualStruct", VisualTypeConf.DROP_OPERATION);
        tree.enable("matrix.visual.VisualStruct-matrix.structures.spatial.FDT.probe.Polygon", VisualTypeConf.MINIMIZED);
        tree.enable("matrix.visual.VisualArrayComponent", VisualTypeConf.HIGHLIGHT_OPERATION);
		tree.enable("matrix.visual.VisualArrayComponent", VisualTypeConf.DROP_OPERATION);

		return new VisualTypeConf[] { new VisualTypeConf(), new VisualTypeConf(), stream, tree };
	}

	
	// ** Implementation of interface GetModelAnswerNames **
	
	public String[] getModelAnswerNames() {
		return getStructureNames();
	}

	
	// ** Implementation of interface LayoutExercise **

	public GridBagConstraints[] getConstraints() {
        GridBagConstraints inputVis = new GridBagConstraints();
//        inputVis.gridwidth = 6;
        inputVis.insets = new Insets(2, 2, 2, 2);
        inputVis.gridx = 0;
        inputVis.gridy = 0;
        inputVis.gridwidth = 2;
//        inputVis.fill = GridBagConstraints.BOTH;
        GridBagConstraints treeVis = new GridBagConstraints();
//        treeVis.gridwidth = GridBagConstraints.REMAINDER;
        treeVis.gridx = 2;
        treeVis.gridy = 0;
        treeVis.gridwidth = 2;
        treeVis.insets = new Insets(2, 0, 2, 2);
//        treeVis.fill = GridBagConstraints.BOTH;
        GridBagConstraints input = new GridBagConstraints();
//        input.gridwidth = 1;
        input.insets = new Insets(0, 2, 2, 2);
        input.gridy = 1;
        input.gridx = 0;
        input.gridwidth = 1;
//        input.fill = GridBagConstraints.BOTH;
        GridBagConstraints tree = new GridBagConstraints();
//        tree.gridwidth = GridBagConstraints.REMAINDER;
        tree.insets = new Insets(0, 0, 2, 2);
        tree.gridx = 1;
        tree.gridy = 1;
        tree.gridwidth = 3;
//        tree.fill = GridBagConstraints.BOTH;
        tree.weightx = 1;
        return new GridBagConstraints[]{ inputVis, treeVis, input, tree};
	}

	public GridBagConstraints[] getModelAnswerConstraints() {
		return getConstraints();
	}

	
	// ** Implementation of interface StyledExercise **
	
	public String[] getStructureVisualisations() {
		return new String[] {"2D", "2D", "array", "layered tree" };
	}

	public String[] getModelAnswerVisualisations() {
		return getStructureVisualisations();
	}

	
	// ** Implementation of interface SwapBehaviour **
	
	public boolean getSwapBehaviour() {
		return true;
	}

    
    // ** Implementation of interface AWTComponentUtilizer ** 

	// This need not be serialized.
    private transient Application application;
    
    class MyPanel extends JPanel {

        public MyPanel(LayoutManager layout) {
            super(layout);
        }

        public void validate() {
            // Force validating of subcomponents
            int componentCount = this.getComponentCount();
            for (int i = 0; i < componentCount; i++) {
                this.getComponent(i).validate();
            }
            super.validate();
        }

        private static final long serialVersionUID = 3699821165607871718L;
    }

    class VirtualDoubleField extends JTextField {
        VirtualObject value;
        double textValue;
        
        public VirtualDoubleField(int columns) {
            super("NA", columns);
            this.value = new VirtualObject(null, "value");
            this.value.setObject(new Double(Double.NaN));
            this.textValue = Double.NaN;
        }

        public void setValue(double value) {
            this.value.setObject(new Double(value));
            this.validateText();
        }
        
        private double getValue() {
            return ((Double)this.value.getObject()).doubleValue();
        }
        
        private void validateText() {
            if ((this.textValue != this.getValue()) &&
                (!(Double.isNaN(this.textValue) && Double.isNaN(this.getValue())))) {
                this.textValue = this.getValue();
                if (Double.isNaN(this.getValue())) {
                    super.setText("NA");
                } else {
                    super.setText(Integer.toString((int)Math.round(this.getValue())));
                }
            }
        }

        public void validate() {
            this.validateText();
            super.validate();
        }
        
        private static final long serialVersionUID = 131324268706044705L;
    }

    public Component[] getAWTComponents() {
    	// The button for the split operation.
        JButton splitButton = new JButton("Split / Split & Insert");
        splitButton.addActionListener(new ActionListener() {
        	public void actionPerformed(ActionEvent evt) {
                split();
            }
        });

        GridBagLayout gbLayout = new GridBagLayout();
        MyPanel gridPanel = new MyPanel(gbLayout);
        GridBagConstraints gbc = new GridBagConstraints();
        gridPanel.setBackground(new Color(192, 192, 192));
        
        JLabel enlargementLabel = new JLabel("Enlargement:");
        enlargementLabel.setForeground(Color.BLACK);
        this.enlargement = new VirtualDoubleField(Integer.toString(WIDTH * HEIGHT).length() + 2);
        this.enlargement.setEditable(false);
        this.enlargement.setHorizontalAlignment(JTextField.CENTER);
        JLabel areaLabel = new JLabel("Area:");
        areaLabel.setForeground(Color.BLACK);
        this.area = new VirtualDoubleField(Integer.toString(WIDTH * HEIGHT).length() + 2);
        this.area.setEditable(false);
        this.area.setHorizontalAlignment(JTextField.CENTER);

        gbc.gridwidth = GridBagConstraints.REMAINDER;
        gbc.insets = new Insets(2, 2, 2, 2);
        gridPanel.add(splitButton, gbc);
        gbc.gridwidth = GridBagConstraints.RELATIVE;
        gridPanel.add(enlargementLabel);
        gbc.gridwidth = GridBagConstraints.REMAINDER;
        gridPanel.add(this.enlargement, gbc);
        gbc.gridwidth = GridBagConstraints.RELATIVE;
        gridPanel.add(areaLabel);
        gbc.gridwidth = GridBagConstraints.REMAINDER;
        gridPanel.add(this.area, gbc);
        
        return new Component[]{ gridPanel };
    }
    
    public GridBagConstraints[] getAWTComponentConstraints() {
        GridBagConstraints gbc = new GridBagConstraints();
        gbc.gridx = 0;
        gbc.gridy = 2;
        gbc.gridwidth = 2;
        gbc.insets = new Insets(2, 2, 2, 2);
        return new GridBagConstraints[]{ gbc };
    }

    public void setApplication(Application a) {
        this.application = a;
    }
    
    
    /**
     * Splits the node selected by the user when the user presses
     * the split button. If there are multiple selections, the 
     * operation is done to the one last selected. 
     */
    private void split() {
    	// Get all the selections currently active in the tree.
    	ArrayList selections = SelectionVisitor.getSelectedInOrder(this.tree, false);

        if (!selections.isEmpty()) {
            // Execute the split.
            Animator.getActiveAnimator().startOperation();

            if (((ExerciseRTree)selections.get(selections.size() - 1)).isLeaf()) {
                // button function = split & insert
                SpatialComparable nextInput = (SpatialElement)inputStream.ADTPop();
                // Split the selected node
                ((ExerciseRTree)selections.get(selections.size() - 1)).splitThis(nextInput);
                Animator.getActiveAnimator().endOperation();
                this.application.validate();
                this.application.validateAnimator();
            } else {
                // button function = split
                // Split the selected node
                ((ExerciseRTree)selections.get(selections.size() - 1)).splitThis(null);
                Animator.getActiveAnimator().endOperation();
                this.application.validate();
                this.application.validateAnimator();
            }
        }
    }
    
    class Selector implements NodeSelector, Serializable {
        private VirtualObject selection;
        
        public Selector() {
            this.selection = new VirtualObject(null, "selection");
        }
        
        public synchronized void select(ExerciseRTree node, boolean selected) {
            if (selected) {
                if (this.selection.getObject() != null) {
                    select((ExerciseRTree)this.selection.getObject(), false);
                }
                this.selection.setObject(node);
                node.setPaintingStyleDecorator(selectedAreaDecorator);
                SpatialElement[] spatialElements = node.getSpatialElements();
                for (int i = 0; i < spatialElements.length; i++) {
                    spatialElements[i].setPaintingStyleDecorator(selectedSpatialElementDecorator);
                }
                // enlargement area
                if (inputStream.ADTPeek() != null) {
                    enlargement.setValue(ExerciseRTree.enlargementArea(inputStream.ADTPeek(), node));
                    area.setValue(node.getBoundingPolygon().getWidth() * node.getBoundingPolygon().getHeight());
                } else {
                    enlargement.setValue(Double.NaN);
                    area.setValue(Double.NaN);
                }
            } else {
                if (isSelected(node)) {
                    this.selection.setObject(null);
                }
                node.setPaintingStyleDecorator(areaDecorator);
                SpatialElement[] spatialElements = node.getSpatialElements();
                for (int i = 0; i < spatialElements.length; i++) {
                    
                    spatialElements[i].setPaintingStyleDecorator(indexedSpatialElementDecorator);
                }
                enlargement.setValue(Double.NaN);
                area.setValue(Double.NaN);      
            }
        }

        public synchronized boolean isSelected(ExerciseRTree node) {
            if (selection.getObject() == node) {
                return true;
            } else {
                return false;
            }
        }
        
        public synchronized void clearSelection() {
            if (this.selection.getObject() != null) {
                select((ExerciseRTree)this.selection.getObject(), false);
            }
        }

        private static final long serialVersionUID = 27188362702827449L;
    }
    
    
    class InputStack extends SimpleStack {
        
        public InputStack(Rectangle boundingBox, PaintingStyleDecorator topDecorator) {
            super(boundingBox, topDecorator);
        }
        
        public Object ADTPop() {
            SpatialElement spatialElement = (SpatialElement)super.ADTPop();
            if (spatialElement != null) {
                spatialElement.setPaintingStyleDecorator(indexedSpatialElementDecorator);
            }
            return spatialElement;
        }
        private static final long serialVersionUID = 8237666678622411219L;
    }
    
    private static final long serialVersionUID = 7687384595704045070L;
}
