package content.exercises.sda;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Component;
import java.awt.GridBagConstraints;
import java.awt.Insets;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Random;
import java.util.Vector;

import javax.swing.JButton;

import matrix.animation.Animator;
import matrix.simulation.VisualTypeConf;
import matrix.structures.CDT.CDT;
import matrix.structures.CDT.probe.QueueImpl;
import matrix.structures.CDT.probe.RBTree;
import matrix.structures.CDT.probe.RedBlackNode;
import matrix.structures.FDT.Array;
import matrix.structures.FDT.FDT;
import matrix.structures.FDT.Struct;
import matrix.structures.FDT.Tree;
import matrix.structures.memory.VirtualBoolean;
import matrix.structures.memory.VirtualObject;
import matrix.structures.simulationextensions.DeleteEnabled;
import matrix.structures.simulationextensions.Selectable;
import matrix.structures.spatial.PaintingStyleDecorator;
import matrix.structures.spatial.FDT.probe.Point;
import matrix.structures.spatial.FDT.probe.PolyLine;
import matrix.structures.spatial.FDT.probe.Polygon;
import matrix.structures.util.MatrixComparable;
import matrix.util.Application;
import matrix.util.Note;
import matrix.util.Traverse;
import content.exercises.sda.structures.PointGenerator;
import content.exercises.sda.structures.PolygonGenerator;
import content.exercises.sda.structures.SimpleStack;
import content.interfaces.AWTComponentUtilizer;
import content.interfaces.ConfigureVisualType;
import content.interfaces.Exercise;
import content.interfaces.JudgeBlocks;
import content.interfaces.LayoutExercise;
import content.interfaces.ModelAnswerNames;
import content.interfaces.SimulationExercise;
import content.interfaces.SimulationExerciseModel;
import content.interfaces.StyledExercise;

/**
 * Exercise for finding the set of vertices in a polygon set that are visible 
 * from a given reference point using a rotational sweep algorithm. 
 * 
 * @author Juha Helminen
 */
public class VisibleVertices implements Exercise, SimulationExercise, ConfigureVisualType, 
                                        SimulationExerciseModel, LayoutExercise, 
                                        AWTComponentUtilizer, ModelAnswerNames, JudgeBlocks,
                                        StyledExercise {

    // For debugging
    private static final boolean STATIC_SEED = false;
    private static final long STATIC_SEED_VALUE = 1172505534424L;

    // ** Exercise configuration **
    private static final int WIDTH = 250;
    private static final int HEIGHT = 400;
    // The polygon area border.
    private static final int BORDER = 12;
    // The minimum allowed distance between a polygon vertex and edge.
    private static final int MIN_VERTEX_EDGE_DISTANCE = 13;
    // The minimum allowed distance between a ray and a vertex.
    private static final int MIN_RAY_VERTEX_DISTANCE = 4;
    private static final int POLYGON0_VERTEX_COUNT = 3;
    private static final int POLYGON1_VERTEX_COUNT = 4;
    private static final int POLYGON2_VERTEX_COUNT = 5;
    // The minimum allowed distance between polygon's vertices.
    private static final int POLYGON0_VERTEX_DISTANCE = 25;
    private static final int POLYGON1_VERTEX_DISTANCE = 43;
    private static final int POLYGON2_VERTEX_DISTANCE = 46;
    
    
    // ** Color and style scheme **
    private static final Color VERTEX_COLOR = Color.BLUE;
    private static final Color REF_POINT_COLOR = Color.ORANGE;
    private static final Color EVENT_POINT_COLOR = Color.MAGENTA;
    private static final Color VISIBLE_VERTEX_COLOR = Color.RED;
    private static final Color SELECTED_EDGE_COLOR = Color.GREEN;
    private static final Color EDGE_COLOR = Color.WHITE;
    private static final Color POLYGON_FILL_COLOR = Color.GREEN;
    private static final Color INTERSECTED_EDGE_COLOR = Color.WHITE;
    private static final Color RAY_COLOR = Color.YELLOW;
    private static final PaintingStyleDecorator refPointDecorator;
    private static final PaintingStyleDecorator eventPointDecorator;
    private static final PaintingStyleDecorator visibleVertexDecorator;
    private static final PaintingStyleDecorator vertexDecorator;
    private static final PaintingStyleDecorator edgeDecorator;
    private static final PaintingStyleDecorator selectedEdgeDecorator;
    private static final PaintingStyleDecorator polygonFillDecorator;
    private static final PaintingStyleDecorator intersectedEdgeDecorator;
    private static final PaintingStyleDecorator rayDecorator;
    static {
        vertexDecorator = new PaintingStyleDecorator();
        vertexDecorator.setLineColor(VERTEX_COLOR);

        refPointDecorator = new PaintingStyleDecorator();
        refPointDecorator.setLineColor(REF_POINT_COLOR);

        eventPointDecorator = new PaintingStyleDecorator();
        eventPointDecorator.setLineColor(EVENT_POINT_COLOR);

        visibleVertexDecorator = new PaintingStyleDecorator();
        visibleVertexDecorator.setLineColor(VISIBLE_VERTEX_COLOR);

        selectedEdgeDecorator = new PaintingStyleDecorator();
        selectedEdgeDecorator.setLineColor(SELECTED_EDGE_COLOR);
        selectedEdgeDecorator.setStroke(3, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL);
        
        float[] polygonDashPattern = { 5 , 5 };
        edgeDecorator = new PaintingStyleDecorator();
        edgeDecorator.setLineColor(EDGE_COLOR);
        edgeDecorator.setStroke(2, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL, 5, polygonDashPattern, 0);

        polygonFillDecorator = new PaintingStyleDecorator();
        polygonFillDecorator.setTransparency(0.07);
        polygonFillDecorator.setLineColor(EDGE_COLOR);
        polygonFillDecorator.setFillColor(POLYGON_FILL_COLOR);
        polygonFillDecorator.setOutline(false);

        intersectedEdgeDecorator = new PaintingStyleDecorator();
        intersectedEdgeDecorator.setLineColor(INTERSECTED_EDGE_COLOR);
        intersectedEdgeDecorator.setStroke(2, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL);

        float[] rayDashPattern = { 2, 5, 2, 5 };
        rayDecorator = new PaintingStyleDecorator();
        rayDecorator.setLineColor(RAY_COLOR);
        rayDecorator.setStroke(2, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL, 2, rayDashPattern, 0);
    }

    
    // ** Structures for the exercise. **
    
    // The set of polygons.
    private java.awt.Point[][] polygons;
    // The viewing point for which the visible vertices are computed.
    private java.awt.Point viewpoint;

    // * student *
    // The visual representation of the problem.
    private SimpleStack problem;
    // The vertices of the polygons in cw order starting from the positive x-axis.
    private SimpleStack vertexStack;
    // The vertex currently being examined.
    private Slot currentVertex;
    // The status structure containing the edges intersected by the current ray in the
    // order in which they are intersected starting from the viewing point.
    private EdgeRBTree statusTree;
    // The list of vertices visible from the viewing point.
    private VisibleVertexList visibleVertices;
    // The ray.
    private SimpleStack rayVis;
    
    private long seed = System.currentTimeMillis();
    

    // ** Implementation of interface Exercise **
    
    public String getDescription() {
        return "Find the vertices visible from the viewing point by following a rotational sweep algorithm.";
    }

    
    // ** Implementation of interface SimulationExercise **
    
    public FDT[] getAnswer() {
        return new FDT[]{ this.statusTree, this.visibleVertices, this.vertexStack};
    }

    public FDT[] getInitialStructures() {
        SimpleStack problem = new SimpleStack(new Rectangle(WIDTH, HEIGHT, 0, 0));
        Slot currentVertex = new Slot();
        SimpleStack vertexStack = new SimpleStack();
        EdgeRBTree statusTree = new EdgeRBTree();
        VisibleVertexList visibleVertices = new VisibleVertexList();
        
        // Create the reference point.
        Point viewpoint = new Point(this.viewpoint);
        viewpoint.setPaintingStyleDecorator(refPointDecorator);
        problem.ADTPush(viewpoint);

        // Create the vertices.
        int vertexCount = 0;
        for (int i = 0; i < this.polygons.length; i++) {
            vertexCount += this.polygons[i].length;
        }
        Vertex[] vertices = new Vertex[vertexCount];
        int k = 0;
        for (int i = 0; i < this.polygons.length; i++) {
            for (int j = 0; j < this.polygons[i].length; j++) {
                vertices[k++] = new Vertex(this.polygons[i][j].x, 
                                           this.polygons[i][j].y);
            }
        }
        
        // Create the polygons.
        k = 0;
        Polygon polygon;
        Point[] polygonVertices;
        for (int i = 0; i < this.polygons.length; i++) {
            polygonVertices = new Point[this.polygons[i].length];
            for (int j = 0; j < this.polygons[i].length - 1; j++) {
                this.newLine(k + j, k + j + 1, vertices, problem);
                polygonVertices[j] = vertices[k + j];
            }
            this.newLine(k + this.polygons[i].length - 1, k, vertices, problem);
            polygonVertices[this.polygons[i].length - 1] = vertices[k + this.polygons[i].length - 1];
            polygon = new Polygon(polygonVertices, "");
            polygon.setPaintingStyleDecorator(polygonFillDecorator);
            problem.ADTPush(polygon);
            k += this.polygons[i].length;
        }
        
        // Sort the vertices in clockwise order relative to the reference point 
        // starting from the positive x-axis.
        this.sortCW(vertices);

        // Index the vertices according to the clockwise order.
        for (int i = 0; i < vertices.length; i++) {
            ((Vertex)vertices[i]).setIndex(i);
            vertices[i].setPaintingStyleDecorator(vertexDecorator);
            problem.ADTPush(vertices[i]);
        }

        // Fill the input stack.
        for (int i = vertices.length - 1; i >= 0; i--) {
            vertexStack.ADTPush(vertices[i]);
        }
        
        // Visualize the algorithm by drawing an imaginary ray sweeping the area around the viewing point.
        SimpleStack rayVis = new SimpleStack();
        problem.addSubArea(rayVis);
        this.drawRay(new java.awt.Point(WIDTH, this.viewpoint.y), rayVis);

        return new FDT[]{ problem, 
                          currentVertex, 
                          vertexStack, 
                          statusTree, 
                          visibleVertices,
                          rayVis};
    }

    
    public FDT[] init() {
        Note.out(this, "The random number generator seed for this exercise is: " + this.getSeed());
        Random r = new Random(this.getSeed());
        
        // ** Generate the input for the exercise. **
        final int MAX_TRIES = 120;
        int tryCounter = 0;
        this.viewpoint = new java.awt.Point(WIDTH / 2, HEIGHT - HEIGHT / 5);
        this.polygons = new java.awt.Point[3][];
        while (true) {
            try {
                this.polygons[0] = PointGenerator.generatePoints(new Rectangle(BORDER, 
                                                                               HEIGHT / 3 + BORDER / 2,
                                                                               WIDTH / 3 - BORDER - BORDER / 2,
                                                                               HEIGHT / 3 - BORDER / 2),
                                                                 POLYGON0_VERTEX_COUNT, 
                                                                 POLYGON0_VERTEX_DISTANCE, 
                                                                 r);
                this.polygons[1] = PointGenerator.generatePoints(new Rectangle(WIDTH / 3 + BORDER / 2, 
                                                                               HEIGHT / 3 + BORDER / 2,
                                                                               2 * WIDTH / 3 - BORDER - BORDER / 2,
                                                                               HEIGHT / 3 - BORDER / 2),
                                                                 POLYGON1_VERTEX_COUNT, 
                                                                 POLYGON1_VERTEX_DISTANCE, 
                                                                 r);
                this.polygons[2] = PointGenerator.generatePoints(new Rectangle(BORDER, 
                                                                               BORDER,
                                                                               WIDTH - 2 * BORDER,
                                                                               HEIGHT / 3 - BORDER - BORDER / 2),
                                                                 POLYGON2_VERTEX_COUNT, 
                                                                 POLYGON2_VERTEX_DISTANCE, 
                                                                 r);
                // Arrange the points into polygons.
                for (int i = 0; i < this.polygons.length; i++) {
                    PolygonGenerator.generatePolygon2Peasants(this.polygons[i]);
                }
                
                Line2D[] edges;
                for (int i = 0; i < this.polygons.length; i++) {
                    edges = createPolygonEdges(this.polygons[i]);
                    for (int j = 0; j < this.polygons[i].length; j++) {
                        if (!checkDistance(this.polygons[i][j], edges, MIN_VERTEX_EDGE_DISTANCE)) {
                            throw new Exception("Failed to generate polygons.");
                        }
                    }
                }
                
                for (int i = 0; i < this.polygons.length; i++) {
                    for (int j = 0; j < this.polygons[i].length; j++) {
                        for (int k = 0; k < this.polygons.length; k++) {
                            for (int l = 0; l < this.polygons[k].length; l++) {
                                if (!((i == k) && (j == l))) {
                                    if (new Line2D.Double(this.viewpoint, castRay(this.polygons[i][j])).
                                            ptLineDist(this.polygons[k][l]) < MIN_RAY_VERTEX_DISTANCE) {
                                        throw new Exception("Could not generate polygons.");
                                    }
                                }
                            }
                        }
                    }
                }
                break;
            } catch (Exception e) {
                if (tryCounter++ >= MAX_TRIES) {
                    Note.err(this, e.getMessage());
                    Note.show(this, "Could not generate input for the exercise. Please restart the exercise.");
                    return null;
                }
            }
        }

        FDT[] structures = this.getInitialStructures();
        this.problem = (SimpleStack)structures[0];
        this.currentVertex = (Slot)structures[1];
        this.vertexStack = (SimpleStack)structures[2];
        this.statusTree = (EdgeRBTree)structures[3];
        this.visibleVertices = (VisibleVertexList)structures[4];
        this.rayVis = (SimpleStack)structures[5];
        
        return new FDT[]{this.problem, 
                         this.currentVertex,
                         this.vertexStack,
                         this.statusTree,
                         this.visibleVertices};
    }
    
    private static Line2D[] createPolygonEdges(java.awt.Point[] vertices) {
        Line2D[] edges = null;
        if(vertices.length > 1) {
            edges = new Line2D[vertices.length];
            for (int i = 0; i < vertices.length - 1; i++) {
                edges[i] = new Line2D.Double(vertices[i], vertices[i + 1]);
            }
            // Close the polygon.
            edges[vertices.length - 1] = new Line2D.Double(vertices[vertices.length - 1], 
                                                           vertices[0]);
        }
        return edges;
    }

    private static boolean checkDistance(Point2D point, 
                                         Line2D[] lines, 
                                         int minDistance) {
        for (int i = 0; i < lines.length; i++) {
            if (lines[i].getP1().equals(point) ||
                lines[i].getP2().equals(point)) {
                continue;
            } else if (lines[i].ptSegDist(point) < minDistance) {
                return false;
            }
        }
        return true;
    }

    public String[] getStructureNames() {
        return new String[]{ "", "Current vertex", "Next vertex", "Status structure", "The visible vertices" };
    }

    public long getSeed() {
        // Seed might be set to a constant value for debugging purposes.
        if (STATIC_SEED) {
            return VisibleVertices.STATIC_SEED_VALUE;
        } else {
            return this.seed;
        }
    }

    public void setSeed(long seed) {
        this.seed = seed;
    }
    
    
    // ** Implementation of interface SimulationExerciseModel **
    
    public FDT[] makeModelAnswer() {
        return solveHelp();
    }


    public FDT[] solve() {
        FDT[] f = solveHelp();
        return new FDT[]{ f[3], f[4], f[2] };
    }
    
    
    private FDT[] solveHelp() {
        // Initialize solution structures.
        FDT[] structures = this.getInitialStructures();
        SimpleStack solution = (SimpleStack)structures[0];
        Slot currentVertex = (Slot)structures[1];
        SimpleStack vertexStack = (SimpleStack)structures[2];
        EdgeRBTree modelTree = (EdgeRBTree)structures[3];
        VisibleVertexList visibleVertices = (VisibleVertexList)structures[4];
        SimpleStack rayVis = (SimpleStack)structures[5];
        PolyLine ray;

        Animator a = Animator.getActiveAnimator();
        Point[] rayPoints;
        Line2D rayLine;
        Point[] edgePoints;
        Vector edges;
        
        while (!vertexStack.ADTIsEmpty()) {
            // Rotate the ray, i.e. move on to the next vertex in cw order.
            a.startOperation();
            rotateRay(modelTree, currentVertex, vertexStack, rayVis);
            a.endOperation();

            ray = (PolyLine)rayVis.ADTPeek();
            rayPoints = ray.getPoints();
            rayLine = new Line2D.Double(rayPoints[0].getPoint2D(), rayPoints[1].getPoint2D());
            edges = ((Vertex)currentVertex.getElement()).getEdges();

            // Delete edges from tree.
            // Test for edges CCW to the ray and remove them from the tree.
            a.startOperation();
            for (int k = 0; k < edges.size(); k++) {
                edgePoints = ((PolyLine)edges.get(k)).getPoints();
                if (((Vertex)currentVertex.getElement()) == edgePoints[0]) {
                    if (!isClockWise(edgePoints[1].getPoint2D(), rayLine)) {
                        modelTree.delete(modelTree.createEdge((PolyLine)edges.get(k)));
                    }
                } else {
                    if (!isClockWise(edgePoints[0].getPoint2D(), rayLine)) {
                        modelTree.delete(modelTree.createEdge((PolyLine)edges.get(k)));
                    }
                }
            }
            a.endOperation();

            // test visibility - select leftmost edge
            EdgeRBTree.Edge edge = modelTree.selectMin();
            Line2D lineOfSight = 
                new Line2D.Double(rayLine.getP1(), 
                                  ((Vertex)currentVertex.getElement()).getPoint2D());
            if ((edge == null) || 
                (!lineOfSight.intersectsLine(
                        new Line2D.Double(edge.getLine().getPoints()[0].getPoint2D(),
                                          edge.getLine().getPoints()[1].getPoint2D())))) {
                a.startOperation();
                visibleVertices.insert(currentVertex.getElement());
                a.endOperation();
            }

            if (!vertexStack.ADTIsEmpty()) {
                // Insert edges into the tree.
                // Test for edges CW to the ray and insert them into the tree.
                a.startOperation();
                for (int k = 0; k < edges.size(); k++) {
                    edgePoints = ((PolyLine)edges.get(k)).getPoints();
                    if (((Vertex)currentVertex.getElement()) == edgePoints[0]) {
                        if (isClockWise(edgePoints[1].getPoint2D(), rayLine)) {
                            modelTree.insert(edges.get(k));
                        }
                    } else {
                        if (isClockWise(edgePoints[0].getPoint2D(), rayLine)) {
                            modelTree.insert(edges.get(k));
                        }
                    }
                }
                a.endOperation();
            }
        }
            
        return new FDT[]{ solution, currentVertex, vertexStack, modelTree, visibleVertices };
    }
    
    
    private boolean isClockWise(Point2D p, Line2D line) {
        if (line.relativeCCW(p) > 0) {
            return true;
        } else {
            return false;
        }
    }
    
    
    // ** Implementation of interface ModelAnswerNames **
    
    public String[] getModelAnswerNames() {
        return new String[]{ "", "Current vertex", "Next vertex", "Status structure", "The visible vertices"};
    }


    // ** Implementation of interface ConfigureVisualType **
    
    public VisualTypeConf[] conf() {
        VisualTypeConf polygons = new VisualTypeConf();
        polygons.enable("matrix.visual.VisualArea$AreaVisualization", VisualTypeConf.HIGHLIGHT_OPERATION);
        polygons.enable("matrix.visual.VisualArea$AreaVisualization", VisualTypeConf.DRAG_OPERATION);
        VisualTypeConf statusStructure = new VisualTypeConf();
        statusStructure.enable("matrix.visual.VisualLayeredTree", VisualTypeConf.DROP_OPERATION);
        statusStructure.enable("matrix.visual.VisualLayeredTree", VisualTypeConf.HIGHLIGHT_OPERATION);
        statusStructure.enable("matrix.visual.VisualStruct", VisualTypeConf.HIGHLIGHT_OPERATION);
        statusStructure.enable("matrix.visual.VisualStruct", VisualTypeConf.DRAG_OPERATION);
        statusStructure.enable("matrix.visual.VisualStruct-content.exercises.sda.VisibleVertices$EdgeRBTree$Edge", VisualTypeConf.MINIMIZED);
        VisualTypeConf queue = new VisualTypeConf();
        VisualTypeConf slot = new VisualTypeConf();
        slot.setVisualConfValue("matrix.visual.VisualArray", VisualTypeConf.INDEXED, "false");
        slot.enable("matrix.visual.VisualStruct-matrix.structures.spatial.FDT.probe.Point", VisualTypeConf.MINIMIZED);
        slot.enable("matrix.visual.VisualStruct-matrix.structures.spatial.FDT.probe.Point", VisualTypeConf.HIGHLIGHT_OPERATION);
        slot.enable("matrix.visual.VisualStruct-matrix.structures.spatial.FDT.probe.Point", VisualTypeConf.DRAG_OPERATION);
        VisualTypeConf stack = new VisualTypeConf();
        stack.setVisualConfValue("matrix.visual.VisualArray", VisualTypeConf.INDEXED, "false");
        stack.enable("matrix.visual.VisualStruct-matrix.structures.spatial.FDT.probe.Point", VisualTypeConf.MINIMIZED);
        queue.enable("matrix.visual.VisualList", VisualTypeConf.DROP_OPERATION);
        queue.enable("matrix.visual.VisualList", VisualTypeConf.HIGHLIGHT_OPERATION);
        queue.enable("matrix.visual.VisualStruct-matrix.structures.spatial.FDT.probe.Point", VisualTypeConf.MINIMIZED);
        VisualTypeConf trash = new VisualTypeConf();
        trash.enable("matrix.visual.VisualTrash", VisualTypeConf.DROP_OPERATION);
        trash.enable("matrix.visual.VisualTrash", VisualTypeConf.HIGHLIGHT_OPERATION);
        return new VisualTypeConf[]{ polygons, slot, stack, statusStructure, queue };
    }


    // ** Implementation of interface StyledExercise **
    
    public String[] getModelAnswerVisualisations() {
        return getStructureVisualisations();
    }

    public String[] getStructureVisualisations() {
        return new String[]{ "2D", "array", "array", "layered tree", "list"};
    }
    
    
    // ** Implementation of interface LayoutExercise **

    public GridBagConstraints[] getConstraints() {
        GridBagConstraints area = new GridBagConstraints();
        area.insets = new Insets(2, 2, 2, 4);;
        area.gridheight = 3;
        area.gridx = 0;
        area.gridy = 0;
        GridBagConstraints currentStack = new GridBagConstraints();
        currentStack.insets = new Insets(2, 0, 2, 1);;
        currentStack.gridy = 0;
        currentStack.fill = GridBagConstraints.BOTH;
        GridBagConstraints nextStack = new GridBagConstraints();
        nextStack.insets = new Insets(2, 0, 2, 2);;
        nextStack.gridy = 0;
        nextStack.fill = GridBagConstraints.BOTH;
        GridBagConstraints tree = new GridBagConstraints();
        tree.insets = new Insets(2, 0, 2, 2);;
        tree.gridy = 2;
        tree.gridx = 1;
        tree.gridwidth = 4;
        GridBagConstraints list = new GridBagConstraints();
        list.insets = new Insets(0, 0, 2, 2);;
        list.gridy = 1;
        list.gridx = 1;
        list.gridwidth = 4;
        list.fill = GridBagConstraints.HORIZONTAL;
        list.anchor = GridBagConstraints.NORTH;
        return new GridBagConstraints[]{ area, currentStack, nextStack, tree, list };
    }


    public GridBagConstraints[] getModelAnswerConstraints() {
        GridBagConstraints[] constraints = getConstraints();
        constraints[2].insets = new Insets(2, 0, 2, 160);
        return constraints;
    }

    
    // ** Implementation of interface AWTComponentUtilizer **

    private transient Application application;
    
    public void setApplication(Application a) {
        this.application = a; 
    }

    public Application getApplication() {
        return this.application; 
    }

    public Component[] getAWTComponents() {
        JButton rotateRay = new JButton("Rotate");
        rotateRay.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent evt) {
                rotateRay(statusTree, currentVertex, vertexStack, rayVis);
            }
        });
        return new Component[]{ rotateRay };
    }
    
    public GridBagConstraints[] getAWTComponentConstraints() {
        GridBagConstraints rotateButton = new GridBagConstraints();
        rotateButton.insets = new Insets(2, 0, 2, 100);;
        rotateButton.gridx = 4;
        rotateButton.gridy = 0;
        return new GridBagConstraints[]{ rotateButton };
    }

    private void rotateRay(EdgeRBTree tree, 
                           Slot currentVertex, 
                           SimpleStack vertexStack, 
                           SimpleStack rayVis) {
        if (!vertexStack.ADTIsEmpty()) {

            Animator a = Animator.getActiveAnimator();
            a.startOperation();

            if (tree.selection.getObject() != null) {
                ((EdgeRBTree.Edge)tree.selection.getObject()).clearSelection();
            }

            if ((currentVertex.getElement() != null) && (!((Vertex)currentVertex.getElement()).isVisible())) {
                ((Vertex)currentVertex.getElement()).setPaintingStyleDecorator(vertexDecorator);
            }
            currentVertex.setElement(vertexStack.ADTPop());
            ((Vertex)currentVertex.getElement()).setPaintingStyleDecorator(eventPointDecorator);
            this.drawRay(((Vertex)currentVertex.getElement()).getPoint2D(), rayVis);
            a.endOperation();
            
            this.getApplication().validateAnimator();
            this.getApplication().validate();
        }
    }
    
    private Point2D castRay(Point2D direction) {
        double[] v = new double[2];
        v[0] = direction.getX() - this.viewpoint.x;
        v[1] = direction.getY() - this.viewpoint.y;
        double length = Math.sqrt((v[0] * v[0]) + (v[1] * v[1]));
        v[0] /= length; v[1] /= length;
        v[0] *= Math.max(WIDTH, HEIGHT);
        v[1] *= Math.max(WIDTH, HEIGHT);
        return new Point2D.Double(this.viewpoint.x + v[0], this.viewpoint.y + v[1]);
    }

    private PolyLine drawRay(Point2D direction, SimpleStack vis) {
        Point2D destination = castRay(direction);
        PolyLine polyline = new PolyLine(new Point2D[]{ this.viewpoint, destination}, "");
        polyline.setPaintingStyleDecorator(rayDecorator);
        vis.clear();
        vis.ADTPush(polyline);
        return polyline;
    }

    
    private class EdgeRBTree extends RBTree implements DeleteEnabled {
        
        public EdgeRBTree() {
            super();
        }

        public CDT insert(Object o) {
            if (o instanceof PolyLine) {
                Animator a = Animator.getActiveAnimator();
                a.startOperation();
                if (selection.getObject() != null) {
                    ((Edge)selection.getObject()).clearSelection();
                }
                ((PolyLine)o).setPaintingStyleDecorator(intersectedEdgeDecorator);
                CDT cdt =  super.insert(new Edge((PolyLine)o));
                a.endOperation();
                return cdt;
            } else {
                return this;
            }
        }

        public CDT delete(Object o) {
            if (o instanceof Edge) {
                Animator a = Animator.getActiveAnimator();
                a.startOperation();
                if (selection.getObject() != null) {
                    ((Edge)selection.getObject()).clearSelection();
                }
                ((Edge)o).getLine().setPaintingStyleDecorator(edgeDecorator);
                CDT cdt =  super.delete(o);
                a.endOperation();
                return cdt;
            }
            return this;
        }
        
        // ** Implementation of interface DeleteEnabled **

        public boolean shouldDelete() {
            return true;
        }

        public CDT getNewInstance() {
            return new EdgeRBTree();
        }

        public Tree getNewNode(Object key) {
            return new RBNode((MatrixComparable)key);
        }

        public FDT getNewStructure() {
            return new RBNode() ;
        }
        
        public Edge selectMin() {
            return ((RBNode)this.getElement()).selectMin();
        }
        
        public boolean equals(Object o) {
            if (o instanceof EdgeRBTree) {
                if (this.selection.getObject() == null) {
                    if (((EdgeRBTree)o).selection.getObject() == null) {
                        return getElement().equals(((RBTree)o).getElement()); 
                    }
                } else {
                    if (this.selection.getObject().equals(((EdgeRBTree)o).selection.getObject())) {
                        return getElement().equals(((RBTree)o).getElement()); 
                    }
                }
                return false;
            } else {
                return false;
            }
        }

        public Edge createEdge(PolyLine polyline) {
            return new Edge(polyline);
        }
        
        private VirtualObject selection = new VirtualObject(null, "selection");
                
        private class RBNode extends RedBlackNode implements Selectable {
            
            public RBNode() {
                super();
            }
            
            public RBNode(MatrixComparable key) {
                super(key);
            }

            public Tree getNewNode(Object o) {
                return new RBNode((MatrixComparable)o);
            }

            public int getSubTreeCount() {
                if (this.isLeaf()) {
                    return 0;
                } else {
                    return super.getSubTreeCount();
                }
            }

            // ** Implementation of interface Selectable **
            
            public long getSelectionTime() {
                if (isSelected()) {
                    return Selectable.clock.getNextTimeStamp();
                } else {
                    return Selectable.NOT_SELECTED;
                }
            }
            
            public void clearSelection() {
                if (!this.empty()) {
                    ((Edge)this.getElement()).clearSelection();
                }
            }
            
            public void setSelect(long globalClockTimestamp) {
                if (!this.empty()) {
                    ((Edge)this.getElement()).setSelect(globalClockTimestamp);
                }
            }

            private boolean isSelected() {
                if (this.empty()) {
                    return false;
                } else {
                    return ((Edge)this.getElement()).isSelected();
                }
            }

            public Edge selectMin() {
                if ((this.getLeft() == null)) {
                    // The minimum node was found.
                    if (!this.empty()) {
                        ((Selectable)this.getElement()).
                            setSelect(Selectable.clock.getNextTimeStamp());
                    }
                    return (Edge)this.getElement();
                } else {
                    return ((RBNode)this.getLeft()).selectMin();
                }
            }

            public String toString() {
                return (this.empty() ? "empty" : getElement().toString()); 
            }

            public boolean equals(Object o) {
                if (o == null) {
                    return false;
                } else if (o instanceof RedBlackNode) {
                    if (this.empty()) {
                        return ((RedBlackNode)o).empty();
                    } else if (((RedBlackNode)o).empty()) {
                        return false;
                    } else {
                        return Traverse.traverseInOrder(this).equals(Traverse.traverseInOrder((RedBlackNode)o));
                    }
                } else {
                    return false;
                }
            }

            private static final long serialVersionUID = -5870513895516853861L;
        }


        
        public class Edge implements FDT, Struct, Comparable, MatrixComparable, Selectable {

            private PolyLine line;
            private final String visualization;
            
            // ** Constructors **
            
            public Edge(PolyLine polyline) {
                this.line = polyline;
                if (this.getOrigin().getIndex() < this.getDestination().getIndex()) {
                    this.visualization = this.getOrigin() + " -> " + this.getDestination();
                } else {
                    this.visualization = this.getDestination() + " -> " + this.getOrigin();
                }
            }
            
            
            // ** Getters **
            
            public Vertex getOrigin() {
                return (Vertex)this.getLine().getPoints()[0];
            }
            
            public Vertex getDestination() {
                return (Vertex)this.getLine().getPoints()[1];
            }
            
            public PolyLine getLine() {
                return this.line;
            }

            
            // ** Implementation of interface FDT **
            
            public Object getElement() {
                return null;
            }

            public void setElement(Object o) {
                // disabled
            }


            // ** Implementation of interface Struct **
            
            public Object getField(int location) {
                return this.getFields()[location];
            }

            public String getFieldName(int location) {
                return this.getFieldNames()[location];
            }

            public String[] getFieldNames() {
                return new String[]{ "Label", "Origin", "Destination" };
            }

            public Object[] getFields() {
                return new Object[]{ this.toString(),
                                     this.getOrigin(),
                                     this.getDestination() };
            }

            public void setField(Object value, int location) {
                // disabled
            }

            
            // ** Implementation of interface Selectable **
            
            public long getSelectionTime() {
                if (isSelected()) {
                    return Selectable.clock.getNextTimeStamp();
                } else {
                    return Selectable.NOT_SELECTED;
                }
            }
            
            public void clearSelection() {
                if (selection.getObject() != null) {
                    selection.setObject(null);
                    this.getLine().setPaintingStyleDecorator(intersectedEdgeDecorator);
                }
            }
            
            public void setSelect(long globalClockTimestamp) {
                if (!this.isSelected()) {
                    Animator a = Animator.getActiveAnimator();
                    a.startOperation();
                    if (selection.getObject() != null) {
                        ((Edge)selection.getObject()).clearSelection();
                    }
                    selection.setObject(this);
                    this.getLine().setPaintingStyleDecorator(selectedEdgeDecorator);
                    a.endOperation();
                }
            }

            public boolean isSelected() {
                if (selection.getObject() == this) {
                    return true;
                } else {
                    return false;
                }
            }
            

            // ** Implementation of interface Comparable **
            
            
            private int compare(Edge e, Point reference) {
                Line2D l = new Line2D.Double(e.getOrigin().getPoint2D(), e.getDestination().getPoint2D());
                if (l.relativeCCW(viewpoint) == l.relativeCCW(reference.getPoint2D())) {
                    return 1;
                } else {
                    return -1;
                }
            }

            public int compareTo(Object o) {
                if (o instanceof Edge) {
                    Edge e = (Edge)o;
                    Vertex reference; 
                    if ((this.getOrigin() == e.getOrigin()) &&
                        (this.getDestination() == e.getDestination())) {
                        return 0;
                    } else if (this.getOrigin() == e.getOrigin()) {
                        reference = e.getDestination();
                    } else if (this.getDestination() == e.getDestination()) {
                        reference = e.getOrigin();
                    } else if (this.getOrigin() == e.getDestination()){
                        reference = e.getOrigin();
                    } else if (this.getDestination() == e.getOrigin()) {
                        reference = e.getDestination();
                    } else {
                        // No common vertex.
                        int comparison1 = compare(this, e.getOrigin());
                        int comparison2 = compare(this, e.getDestination());
                        if (comparison1 == comparison2) {
                            return comparison1;
                        } else {
                            return -compare(e, this.getOrigin());
                        }
                    }
                    return compare(this, reference);
                } else {
                    return -1;
                }
            }

            
            // ** Implementation of interface MatrixComparable **
            
            public boolean eq(MatrixComparable k) {
                return (this.compareTo(k) == 0);
            }


            public boolean geq(MatrixComparable k) {
                return (this.compareTo(k) >= 0);
            }


            public String getComparisonString() {
                return null;
            }


            public boolean gt(MatrixComparable k) {
                return (this.compareTo(k) > 0);
            }


            public boolean leq(MatrixComparable k) {
                return (this.compareTo(k) <= 0);
            }


            public boolean lt(MatrixComparable k) {
                return (this.compareTo(k) < 0);
            }


            public boolean equals(Object obj) {
                if (obj instanceof Edge) {
                   if (this == obj)  {
                       return true;
                   } else {
                       Edge e = (Edge)obj;
                       return (((this.getOrigin().equals(e.getOrigin())) &&
                                (this.getDestination().equals(e.getDestination()))) ||
                               ((this.getDestination().equals(e.getOrigin())) &&
                                (this.getOrigin().equals(e.getDestination()))));
                   }
                } else {
                    return false;
                }
            }

            public String toString() {
                return this.visualization; 
            }
            
            private static final long serialVersionUID = -6073875663605256061L;
        }
        
        private static final long serialVersionUID = 6726236697302260437L;
    }

    private void sortCW(Vertex[] points) {
        // Sort the vertices in CW-order relative to the reference point,
        // starting from the positive x-axis.
        Arrays.sort(points, new Comparator() {
            public int compare(Object arg0, Object arg1) {
                double[] veca = {((Point)arg0).getX() - viewpoint.x, ((Point)arg0).getY() - viewpoint.y};
                double[] vecb = {((Point)arg1).getX() - viewpoint.x, ((Point)arg1).getY() - viewpoint.y};
                if ((veca[1] > 0) && (vecb[1] <= 0)) {
                    // a up, b down
                    return 1;
                } else if ((vecb[1] > 0) && (veca[1] <= 0)) {
                    // b up, a down
                    return -1;
                } else {
                    // both down or up
                    if (veca[1] <= 0) {
                        // a down -> b down
                        if (veca[0] > 0) {
                            // a right
                            if (vecb[0] <= 0) {
                                // b left
                                return -1;
                            } else {
                                // a, b down right
                                double lengtha = Math.sqrt((veca[0] * veca[0]) + (veca[1] * veca[1])); 
                                double lengthb = Math.sqrt((vecb[0] * vecb[0]) + (vecb[1] * vecb[1]));
                                veca[1] /= lengtha;
                                vecb[1] /= lengthb;
                                if (veca[1] < vecb[1]) {
                                    return 1;
                                } else if (veca[1] > vecb[1]) {
                                    return -1;
                                } else {
                                    return ((lengtha < lengthb) ? -1 : 1);
                                }
                            }
                        } else {
                            // a left
                            if (vecb[0] > 0) {
                                // b right
                                return 1;
                            } else {
                                // a, b down left
                                double lengtha = Math.sqrt((veca[0] * veca[0]) + (veca[1] * veca[1])); 
                                double lengthb = Math.sqrt((vecb[0] * vecb[0]) + (vecb[1] * vecb[1]));
                                veca[1] /= lengtha;
                                vecb[1] /= lengthb;
                                if (veca[1] < vecb[1]) {
                                    return -1;
                                } else if (veca[1] > vecb[1]) {
                                    return 1;
                                } else {
                                    return ((lengtha < lengthb) ? -1 : 1);
                                }
                            }
                        }
                    } else {
                        // a up -> b up
                        if (veca[0] > 0) {
                            // a right
                            if (vecb[0] <= 0) {
                                // b left
                                return 1;
                            } else {
                                // a, b up right
                                double lengtha = Math.sqrt((veca[0] * veca[0]) + (veca[1] * veca[1])); 
                                double lengthb = Math.sqrt((vecb[0] * vecb[0]) + (vecb[1] * vecb[1]));
                                veca[1] /= lengtha;
                                vecb[1] /= lengthb;
                                if (veca[1] < vecb[1]) {
                                    return 1;
                                } else if (veca[1] > vecb[1]) {
                                    return -1;
                                } else {
                                    return ((lengtha < lengthb) ? -1 : 1);
                                }
                            }
                        } else {
                            // a left
                            if (vecb[0] > 0) {
                                // b right
                                return -1;
                            } else {
                                // a, b up left
                                double lengtha = Math.sqrt((veca[0] * veca[0]) + (veca[1] * veca[1])); 
                                double lengthb = Math.sqrt((vecb[0] * vecb[0]) + (vecb[1] * vecb[1]));
                                veca[1] /= lengtha;
                                vecb[1] /= lengthb;
                                if (veca[1] < vecb[1]) {
                                    return -1;
                                } else if (veca[1] > vecb[1]) {
                                    return 1;
                                } else {
                                    return ((lengtha < lengthb) ? -1 : 1);
                                }
                            }
                        }
                    }
                }
            }
        });
    }
    
    private PolyLine newLine(int origin, int destination, Vertex[] vertices, SimpleStack vis) {
        PolyLine polyline = new PolyLine(new Point[]{vertices[origin], 
                                                     vertices[destination]}, 
                                         "");
        polyline.setPaintingStyleDecorator(edgeDecorator);
        polyline.setDraggable(true);
        vis.ADTPush(polyline);
        // Add to the adjacency list.
        vertices[origin].connect(polyline);
        vertices[destination].connect(polyline);
        return polyline;
    }


    /**
     * Class representing an indexed vertex in a geometric graph.
     */
    class Vertex extends Point {

        private int index;
        VirtualBoolean visible;
        Vector edgeList;
        
        private Vertex(double x, double y, int index) {
            super(x, y, "");
            this.setDraggable(true);
            this.setIndex(index);
            this.visible = new VirtualBoolean(this, "visible");
            this.visible.assign(false);
            this.edgeList = new Vector(2);
        }

        public Vertex(int x, int y, int index) {
            this((double)x, (double)y, index);
        }

        public Vertex(int x, int y) {
            this(x, y, 0);
        }
        
        public Vertex(double x, double y) {
            this((int)x, (int)y, 0);
        }

        public void connect(PolyLine edge) {
            this.edgeList.add(edge);
        }
        
        public Vector getEdges() {
            return this.edgeList;
        }

        public void setIndex(int index) {
            // Start indexing from 1.
            this.setVisualizationString(String.valueOf(index + 1));
            this.index = index;
        }
        
        public int getIndex() {
            return this.index;
        }
        
        public void setVisible(boolean isVisible) {
            this.visible.assign(isVisible);
        }
        
        public boolean isVisible() {
            return this.visible.eval();
        }
        
        public boolean equals(Object obj) {
            if (obj instanceof Vertex) {
                Vertex v = (Vertex)obj;
                if (this.getIndex() == v.getIndex()) {
                    return true;
                } else {
                    return false;
                }
            } else {
                return false;
            }
        }
        
        private static final long serialVersionUID = -3812663500905338607L;
    }

    
    /**
     * List for the visible vertices.
     * The style of inserted elements is changed.
     */
    class VisibleVertexList extends QueueImpl {

        public VisibleVertexList() {
            super();
        }

        public CDT insert(Object o) {
            if (o instanceof Point) {
                ((Point)o).setPaintingStyleDecorator(visibleVertexDecorator);
                ((Vertex)o).setVisible(true);
                return super.insert(o);
            } else {
                return this;
            }
        }
        
        private static final long serialVersionUID = -905179150187239828L;
    }


    /**
     * An FDT wrapper for one data element visualized as
     * a single-item array.
     */
    private class Slot implements FDT, Array {
        private VirtualObject contents;
        
        public Slot() {
            this.contents = new VirtualObject(this, "contents");
        }
        
        // ** Implementation of interface FDT **

        public Object getElement() {
            return this.contents.getObject();
        }

        public void setElement(Object o) {
            this.contents.setObject(o);
        }

        public boolean equals(Object obj) {
            if (obj instanceof Slot) {
                if (this.getElement() == null) {
                    if (((Slot)obj).getElement() == null) {
                        return true;
                    } else {
                        return false;
                    }
                } else {
                    return this.getElement().equals(((Slot)obj).getElement());
                }
            } else {
                return false;
            }
        }

        
        // ** Implementation of interface Array **

        public int getFirst() {
            return 0;
        }

        public int getLast() {
            return 0;
        }

        public Object getObject(int address) {
            if (address == 0) {
                return this.getElement();
            } else {
                return null;
            }
        }

        public void setFirst(int i) {
            // disabled
        }

        public void setLast(int i) {
            // disabled
        }

        public void setObject(Object o, int address) {
            // disabled
        }

        private static final long serialVersionUID = -3420257164917054541L;
    }

    private static final long serialVersionUID = 5790922905277382211L;
}



