package content.exercises.sda;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Component;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.LayoutManager;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.geom.Ellipse2D;
import java.util.Iterator;
import java.util.Random;

import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;

import matrix.animation.Animator;
import matrix.simulation.VisualTypeConf;
import matrix.structures.FDT.FDT;
import matrix.structures.memory.VirtualObject;
import matrix.structures.simulationextensions.Selectable;
import matrix.structures.spatial.PaintingStyleDecorator;
import matrix.structures.spatial.FDT.probe.Circle;
import matrix.structures.spatial.FDT.probe.LineSegment;
import matrix.structures.spatial.FDT.probe.Point;
import matrix.structures.spatial.FDT.probe.Polygon;
import matrix.structures.spatial.FDT.probe.Triangle;
import matrix.util.Application;
import matrix.util.Note;
import content.exercises.sda.structures.GeometryFunctions;
import content.exercises.sda.structures.PointGenerator;
import content.exercises.sda.structures.SimpleStack;
import content.exercises.sda.structures.Triangulation;
import content.exercises.sda.structures.PointGenerator.IllegalConstraintsException;
import content.exercises.sda.structures.Triangulation.VisualElements;
import content.interfaces.AWTComponentUtilizer;
import content.interfaces.ConfigureVisualType;
import content.interfaces.Exercise;
import content.interfaces.LayoutExercise;
import content.interfaces.ModelAnswerNames;
import content.interfaces.SimulationExercise;
import content.interfaces.SimulationExerciseModel;
import content.interfaces.SimulationExerciseSelfAssessment;
import content.interfaces.StyledExercise;

/**
 * <p>
 * An exercise for constructing the Voronoi diagram of a
 * given set of points.
 * </p><p>
 * The diagram is constructed incrementally
 * by selecting the edges of the equivalent Delaunay TIN-model
 * (Triangular Irregular Network). Edges are added one by one by
 * selecting pairs of points.
 * </p><p>
 * The Delaunay test can be performed by checking for points that are
 * inside the circumcircle of a triangle or by seeing that the configuration
 * maximizes the minimum angle in the triangles.
 * </p>
 *
 * @author Juha Helminen
 *
 */
public class VoronoiConstruction implements Exercise, SimulationExercise,
                                            SimulationExerciseModel, StyledExercise,
                                            SimulationExerciseSelfAssessment, ModelAnswerNames,
                                            AWTComponentUtilizer, ConfigureVisualType,
                                            LayoutExercise {

    // For debugging
    private static final boolean STATIC_SEED = false;
    private static final long STATIC_SEED_VALUE = 1182154339937L;

    // The number of points.
    private static final int INPUT_SIZE = 14;
    // The minimum allowed distance between two points.
    private static final int MIN_DISTANCE = 70;
    // The minimum allowed distance from the points to the border.
    private static final int BORDER_X = 60;
    private static final int BORDER_Y = 60;
    // The size of the exercise Area.
    private static final int WIDTH = 450;
    private static final int HEIGHT = 350;

    private long seed = System.currentTimeMillis();


    // ** Color and style scheme **

    protected static final Color POINT_COLOR = Color.BLACK;
    protected static final Color INPOINT_COLOR = Color.MAGENTA;
    protected static final Color CIRCLEPOINT_COLOR = Color.RED;
    protected static final Color VORONOI_VERTEX_COLOR = Color.GRAY;
    protected static final Color DELAUNAY_POINT_COLOR = Color.WHITE;
    protected static final Color VORONOI_COLOR = Color.GRAY;
    protected static final Color VORONOI_FILL_COLOR = Color.BLACK;
    protected static final double VORONOI_TRANSPARENCY = 0;
    protected static final Color DELAUNAY_COLOR = Color.WHITE;
    protected static final Color DELAUNAY_FILL_COLOR = Color.BLACK;
    protected static final double DELAUNAY_TRANSPARENCY = 0;
    protected static final Color MISSING_DELAUNAY_COLOR = Color.WHITE;
    protected static final Color WRONG_DELAUNAY_COLOR = Color.RED;
    protected static final Color BISECTOR_COLOR = Color.RED;
    protected static final Color CIRCUMCIRCLE_COLOR = Color.RED;
    protected static final Color CIRCUMCIRCLE_CENTER_COLOR = Color.DARK_GRAY;
    protected static final Color CIRCUMCIRCLE_FILL_COLOR = Color.RED;
    protected static final double CIRCUMCIRCLE_TRANSPARENCY = 0.035;
    protected static final Color TRIANGLE_FILL_COLOR = Color.WHITE;
    protected static final double TRIANGLE_TRANSPARENCY = 0.12;
    protected static final PaintingStyleDecorator vertexDecorator;
    protected static final PaintingStyleDecorator inpointDecorator;
    protected static final PaintingStyleDecorator circlepointDecorator;
    protected static final PaintingStyleDecorator voronoiVertexDecorator;
    protected static final PaintingStyleDecorator delaunayPointDecorator;
    protected static final PaintingStyleDecorator voronoiDecorator;
    protected static final PaintingStyleDecorator delaunayDecorator;
    protected static final PaintingStyleDecorator missingDelaunayDecorator;
    protected static final PaintingStyleDecorator wrongDelaunayDecorator;
    protected static final PaintingStyleDecorator circumcircleDecorator;
    protected static final PaintingStyleDecorator circumcircleCenterDecorator;
    protected static final PaintingStyleDecorator triangleDecorator;
    static {
        vertexDecorator = new PaintingStyleDecorator();
        vertexDecorator.setLineColor(POINT_COLOR);
        vertexDecorator.setStroke(2, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL);

        inpointDecorator = new PaintingStyleDecorator();
        inpointDecorator.setLineColor(INPOINT_COLOR);
        inpointDecorator.setStroke(2, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL);
        
        circlepointDecorator = new PaintingStyleDecorator();
        circlepointDecorator.setLineColor(CIRCLEPOINT_COLOR);
        circlepointDecorator.setStroke(2, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL);

        voronoiVertexDecorator = new PaintingStyleDecorator();
        voronoiVertexDecorator.setLineColor(VORONOI_VERTEX_COLOR);
        voronoiVertexDecorator.setStroke(2, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL);

        delaunayPointDecorator = new PaintingStyleDecorator();
        delaunayPointDecorator.setLineColor(DELAUNAY_POINT_COLOR);
        delaunayPointDecorator.setStroke(2, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL);

        voronoiDecorator = new PaintingStyleDecorator();
        voronoiDecorator.setLineColor(VORONOI_COLOR);
        voronoiDecorator.setFillColor(VORONOI_FILL_COLOR);
        voronoiDecorator.setTransparency(VORONOI_TRANSPARENCY);
        voronoiDecorator.setStroke(2, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL);

        delaunayDecorator = new PaintingStyleDecorator();
        delaunayDecorator.setLineColor(DELAUNAY_COLOR);
        delaunayDecorator.setFillColor(DELAUNAY_FILL_COLOR);
        delaunayDecorator.setTransparency(DELAUNAY_TRANSPARENCY);
        delaunayDecorator.setStroke(2, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL);

        float[] missingDelaunayDashPattern = { 10, 10 };
        missingDelaunayDecorator = new PaintingStyleDecorator();
        missingDelaunayDecorator.setLineColor(MISSING_DELAUNAY_COLOR);
        missingDelaunayDecorator.setStroke(2, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL, 10, missingDelaunayDashPattern, 0);

        wrongDelaunayDecorator = new PaintingStyleDecorator();
        wrongDelaunayDecorator.setLineColor(WRONG_DELAUNAY_COLOR);
        wrongDelaunayDecorator.setStroke(2, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL);

        float[] circumcircleDashPattern = { 10, 10 };
        circumcircleDecorator = new PaintingStyleDecorator();
        circumcircleDecorator.setLineColor(CIRCUMCIRCLE_COLOR);
        circumcircleDecorator.setFillColor(CIRCUMCIRCLE_FILL_COLOR);
        circumcircleDecorator.setTransparency(CIRCUMCIRCLE_TRANSPARENCY);
        circumcircleDecorator.setStroke(1, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL, 10, circumcircleDashPattern, 0);

        circumcircleCenterDecorator = new PaintingStyleDecorator();
        circumcircleCenterDecorator.setLineColor(CIRCUMCIRCLE_CENTER_COLOR);
        circumcircleCenterDecorator.setStroke(2, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL);

        triangleDecorator = new PaintingStyleDecorator();
        triangleDecorator.setFillColor(TRIANGLE_FILL_COLOR);
        triangleDecorator.setTransparency(TRIANGLE_TRANSPARENCY);
        triangleDecorator.setOutline(false);
    }


    // ** Structures for the exercise. **
    protected Triangulation triangulation;
    // Triangle circumcircle visualization   
    protected SimpleStack circumcircleVis;

    // Information for grading.
    protected int edges;
    protected int correctEdges;
    protected int incorrectEdges;
    protected int missingEdges;

    private VirtualDoubleField minAngle = new VirtualDoubleField();

    // The exercise input.
    protected java.awt.Point[] input;



    //  ** Implementation of interface SimulationExercise **

    protected class StudentVisualElements extends VisualElements {

        protected Point newVertex(java.awt.Point vertex) {
            return new SelectablePoint(vertex, vertexDecorator);
        }

        protected LineSegment newDelaunayEdge(java.awt.Point orig, java.awt.Point dest) {
            return new ClickableLineSegment(orig, dest, delaunayDecorator);        
        }
        
        protected Triangle newTriangle(java.awt.Point[] vertices) {
            return new SelectableTriangle(vertices, triangleDecorator);
        }
        
        protected Point newVoronoiVertex(java.awt.Point vertex) {
            return new Point(vertex, voronoiVertexDecorator);
        }
        
        protected LineSegment newVoronoiEdge(java.awt.Point orig, java.awt.Point dest) {
            return new LineSegment(orig, dest, voronoiDecorator);        
        }
        
        protected Polygon newVoronoiRegion(java.awt.Point[] vertices) {
            return new Polygon(vertices, voronoiDecorator);        
        }

        protected void hideDelaunay() {
            if (VoronoiConstruction.this.hasTriangleSelection()) {
                VoronoiConstruction.this.clearTriangleSelection();
            }
            Iterator vertices = VoronoiConstruction.this.triangulation.vertices();
            Point vertex;
            while (vertices.hasNext()) {
                vertex = (Point)vertices.next();
                if (VoronoiConstruction.this.triangulation.vertexEdges(vertex.getPoint()).hasNext()) {
                    vertex.setPaintingStyleDecorator(delaunayPointDecorator);
                }
            }
        }
        
        protected void showDelaunay() {
            VoronoiConstruction.this.clearPointColors();
        }
        
        protected void hideVoronoi() {}

        private static final long serialVersionUID = 4482392082455412991L;
    }
    
    /* (non-Javadoc)
     * @see content.interfaces.SimulationExercise#getInitialStructures()
     */
    public FDT[] getInitialStructures() {
        SimpleStack problemVis = new SimpleStack(new Rectangle(WIDTH, HEIGHT, 0, 0));

        return new FDT[]{ problemVis };
    }

    /* (non-Javadoc)
     * @see content.interfaces.SimulationExercise#init()
     */
    public FDT[] init() {
        // For debugging
        Note.out(this, "The random number generator seed for this exercise is: " + this.getSeed());
        Random r = new Random(this.getSeed());

        // Generate the exercise input.
        final int BORDER_POINT_COUNT = 8;
        final int MAX_TRIES = 2000;
        int tryCounter = 0;
        java.awt.Point[] randomInput;
        while (true) {
	        try {
	            randomInput = PointGenerator.generatePoints(BORDER_X, BORDER_Y,
	                                                        WIDTH - BORDER_X, HEIGHT - BORDER_Y,
	                                                        INPUT_SIZE - BORDER_POINT_COUNT, MIN_DISTANCE, r);
	        } catch (IllegalConstraintsException e) {
	            Note.err(this, e.getMessage());
	            Note.show(this, "Could not generate input for the exercise. Please restart the exercise.");
	            return null;
	        }
	
	        this.input = new java.awt.Point[randomInput.length + BORDER_POINT_COUNT];
	        for (int i = 0; i < randomInput.length; i++) {
	            this.input[i] = randomInput[i];
	        }
	        this.input[randomInput.length] = new java.awt.Point(BORDER_X / 2, BORDER_Y / 2);
	        this.input[randomInput.length + 1] = new java.awt.Point(BORDER_X / 2, HEIGHT - BORDER_Y / 2);
	        this.input[randomInput.length + 2] = new java.awt.Point(WIDTH - BORDER_X / 2, HEIGHT - BORDER_Y / 2);
	        this.input[randomInput.length + 3] = new java.awt.Point(WIDTH - BORDER_X / 2, BORDER_Y / 2);
	        this.input[randomInput.length + 4] = new java.awt.Point(WIDTH / 2, BORDER_Y / 3);
	        this.input[randomInput.length + 5] = new java.awt.Point(BORDER_X / 3, HEIGHT / 2);
	        this.input[randomInput.length + 6] = new java.awt.Point(WIDTH - BORDER_X / 3, HEIGHT / 2);
	        this.input[randomInput.length + 7] = new java.awt.Point(WIDTH / 2, HEIGHT - BORDER_Y / 3);

	        if (GeometryFunctions.delaunay(this.input) == null) {
                if (tryCounter++ >= MAX_TRIES) {
                    Note.err(this, "Could not generate input for the exercise.");
                    Note.show(this, "Could not generate input for the exercise. Please restart the exercise.");
                    return null;
                }
	        } else {
	        	break;
	        }
        }

        // Calculate the number of edges in the correct triangulation.
        // Precondition: the border points define the convex hull of the point set
        this.edges = BORDER_POINT_COUNT + BORDER_POINT_COUNT + (INPUT_SIZE - 1 - BORDER_POINT_COUNT) * 3;

        FDT[] structures = this.getInitialStructures();
        SimpleStack problemVis = (SimpleStack)structures[0];
        this.circumcircleVis = new SimpleStack();
        problemVis.addSubArea(this.circumcircleVis);
        
        this.triangulation = new Triangulation(new StudentVisualElements(), WIDTH, HEIGHT);
        for (int i = 0; i < this.input.length; i++) {
            this.triangulation.addVertex(this.input[i]);
        }
        problemVis.addSubArea(this.triangulation);

        this.triangulation.setDrawVoronoi(Triangulation.VORONOI_REGIONS);
        this.triangulation.setDrawTrianglesVoronoi(false);
        this.triangulation.setDrawVoronoiVertices(false);

        return new FDT[]{problemVis};
    }

    /* (non-Javadoc)
     * @see content.interfaces.SimulationExercise#getStructureNames()
     */
    public String[] getStructureNames() {
        return new String[]{ "Delaunay Triangulation / Voronoi Diagram" };
    }

    /* (non-Javadoc)
     * @see content.interfaces.SimulationExercise#getAnswer()
     */
    public FDT[] getAnswer() {
        return new FDT[]{ this.triangulation };
    }

    /* (non-Javadoc)
     * @see content.interfaces.SimulationExercise#getSeed()
     */
    public long getSeed() {
        // The seed might be set to a constant value for debugging purposes.
        if (STATIC_SEED) {
            return VoronoiConstruction.STATIC_SEED_VALUE;
        } else {
            return this.seed;
        }
    }

    /* (non-Javadoc)
     * @see content.interfaces.SimulationExercise#setSeed(long)
     */
    public void setSeed(long seed) {
        this.seed = seed;
    }



    //  ** Implementation of interface SimulationExerciseModel **


    public FDT[] makeModelAnswer() {
        return this.solve();
    }

    public FDT[] solve() {
        this.correctEdges = 0;
        this.incorrectEdges = 0;
        this.missingEdges = 0;
        FDT[] modelStructures = this.getInitialStructures();
        FDT[] comparisonStructures = this.getInitialStructures();

        Triangulation modelTriangulation = new Triangulation(new StudentVisualElements(), WIDTH, HEIGHT);
        modelTriangulation.setDrawVoronoi(Triangulation.VORONOI_REGIONS);
        modelTriangulation.setDrawVoronoiVertices(false);
        modelTriangulation.setDrawDelaunay(false);
        Triangulation comparisonTriangulation = new Triangulation(new StudentVisualElements(), WIDTH, HEIGHT);
        comparisonTriangulation.setDrawVoronoi(Triangulation.VORONOI_NONE);
        comparisonTriangulation.setDrawVoronoiVertices(false);

        for (int i = 0; i < this.input.length; i++) {
            modelTriangulation.addVertex(this.input[i]);
            comparisonTriangulation.addVertex(this.input[i]);
        }
        ((SimpleStack)modelStructures[0]).addSubArea(modelTriangulation);
        ((SimpleStack)comparisonStructures[0]).addSubArea(comparisonTriangulation);

        java.awt.Point[][] triangles = GeometryFunctions.delaunay(this.input);
        for (int i = 0; i < triangles.length; i++) {
            modelTriangulation.addEdge(triangles[i][0], triangles[i][1]);
            modelTriangulation.addEdge(triangles[i][1], triangles[i][2]);
            modelTriangulation.addEdge(triangles[i][2], triangles[i][0]);
        }

        Iterator modelEdges = modelTriangulation.edges();
        LineSegment edge;
        while (modelEdges.hasNext()) {
            edge = (LineSegment)modelEdges.next();
            if (this.triangulation.hasEdge(edge.getP1(), edge.getP2())) {
                comparisonTriangulation.addEdge(edge.getP1(), edge.getP2());
                this.correctEdges++;
            } else {
                comparisonTriangulation.addEdge(edge.getP1(), edge.getP2(), missingDelaunayDecorator);
                this.missingEdges++;
            }
        }

        Iterator studentEdges = this.triangulation.edges();
        while (studentEdges.hasNext()) {
            edge = (LineSegment)studentEdges.next();
            if (!modelTriangulation.hasEdge(edge.getP1(), edge.getP2())) {
                comparisonTriangulation.addEdge(edge.getP1(), edge.getP2(), wrongDelaunayDecorator);
                this.incorrectEdges++;
            }
        }

        return new FDT[]{ modelStructures[0], comparisonStructures[0] };
    }



    //  ** Implementation of interface ModelAnswerNames **

    public String[] getModelAnswerNames() {
        return new String[]{"Voronoi Diagram", "Correct (solid white), incorrect (solid red) and missing (dashed white) delaunay edges"};
    }




    //  ** Implementation of interface Exercise **

    /* (non-Javadoc)
     * @see content.interfaces.Exercise#getDescription()
     */
    public String getDescription() {
        return "Construct the Delaunay triangulation and Voronoi diagram of the given point set.";
    }



    //  ** Implementation of interface StyledExercise **

    public String[] getModelAnswerVisualisations() {
        return new String[]{"2D", "2D"};
    }

    public String[] getStructureVisualisations() {
        return new String[]{"2D"};
    }



    //  ** Implementation of interface StyledExercise **

    public int getMaxPoints() {
    	// Give 1 point for every correct edge in the triangulation.
        return this.edges;
    }

    public int[] judge() {
        this.solve();
        // Give 1 point for every correct edge in the triangulation and
        // deduct 1 point for every incorrect one.
        return new int[]{Math.max(0, this.correctEdges - this.incorrectEdges)};
    }

    public String getFeedbackString() {
        return this.correctEdges +  " correct edge(s),\n" +
               this.incorrectEdges + " incorrect edge(s) and\n" +
               this.missingEdges + " edge(s) missing.";
    }



    //  ** Implementation of interface AWTComponentUtilizer **

    transient protected Application application;

    public void setApplication(Application application) {
        this.application = application;
    }

    public Component[] getAWTComponents() {
    	// The show/hide voronoi button
        final boolean currentVoronoiState = VoronoiConstruction.this.triangulation.isVoronoiDrawn();
        final String showVoronoi = "Show Voronoi";
        final String hideVoronoi = "Hide Voronoi";
        final JButton drawVoronoiButton = new JButton(currentVoronoiState ? hideVoronoi : showVoronoi);
        drawVoronoiButton.addActionListener(new ActionListener() {
            private boolean state = currentVoronoiState;
            public void actionPerformed(ActionEvent evt) {
                this.state = !this.state;
                drawVoronoiButton.setText((this.state ? hideVoronoi : showVoronoi));
                VoronoiConstruction.this.triangulation.setDrawVoronoi(this.state ? Triangulation.VORONOI_REGIONS : Triangulation.VORONOI_NONE);
                VoronoiConstruction.this.application.validate();
            }
        });

    	// The show/hide delaunay button
        final boolean currentDelaunayState = VoronoiConstruction.this.triangulation.isDelaunayDrawn();
        final String showDelaunay = "Show Delaunay";
        final String hideDelaunay = "Hide Delaunay";
        final JButton drawDelaunayButton = new JButton(currentDelaunayState ? hideDelaunay : showDelaunay);
        drawDelaunayButton.addActionListener(new ActionListener() {
            private boolean state = currentDelaunayState;
            public void actionPerformed(ActionEvent evt) {
                this.state = !this.state;
                drawDelaunayButton.setText((this.state ? hideDelaunay : showDelaunay));
                VoronoiConstruction.this.triangulation.setDrawDelaunay(this.state);
                VoronoiConstruction.this.application.validate();
            }
        });

        JLabel angleLabel = new JLabel("The minimum angle in the selected triangle:");
        angleLabel.setForeground(Color.BLACK);

        GridBagLayout gbLayout = new GridBagLayout();
        MyPanel gridPanel = new MyPanel(gbLayout);
        gridPanel.setBackground(new Color(192, 192, 192));
        GridBagConstraints gbc = new GridBagConstraints();

        gbc.insets = new Insets(4, 4, 4, 4);
        gridPanel.add(angleLabel);
        gbc.gridwidth = GridBagConstraints.REMAINDER;
        gridPanel.add(this.minAngle, gbc);
        gbc.gridwidth = GridBagConstraints.RELATIVE;
        gridPanel.add(drawVoronoiButton, gbc);
        gbc.gridwidth = GridBagConstraints.REMAINDER;
        gridPanel.add(drawDelaunayButton, gbc);

        return new Component[]{ gridPanel };
    }

    public GridBagConstraints[] getAWTComponentConstraints() {
        GridBagConstraints gbc = new GridBagConstraints();
        gbc.gridx = 0;
        gbc.gridy = 1;
        gbc.insets = new Insets(2, 2, 2, 2);
        return new GridBagConstraints[]{ gbc };
    }



    //  ** Implementation of interface LayoutExercise **

    public GridBagConstraints[] getConstraints() {
        GridBagConstraints gbc = new GridBagConstraints();
        Insets insets = new Insets(3, 3, 0, 0);
        gbc.insets = insets;
        return new GridBagConstraints[]{ gbc };
    }

    public GridBagConstraints[] getModelAnswerConstraints() {
        GridBagConstraints gbc = new GridBagConstraints();
        Insets insets = new Insets(3, 3, 0, 0);
        gbc.insets = insets;
        return new GridBagConstraints[]{ gbc, gbc };
    }



    //  ** Implementation of interface ConfigureVisualType **

    public VisualTypeConf[] conf() {
        VisualTypeConf area = new VisualTypeConf();
        area.enable("matrix.visual.VisualArea$AreaVisualization", VisualTypeConf.HIGHLIGHT_OPERATION);
        return new VisualTypeConf[]{ area, new VisualTypeConf() };
    }

    
    
    
    
    
    // ** Circumcircle visualization **
    
    private void circumcircle(Triangle triangle) {
    	Animator a = Animator.getActiveAnimator();
    	a.startOperation();
        
        VoronoiConstruction.this.minAngle.setValue(Math.toDegrees(GeometryFunctions.minAngle(triangle)));
        
        // Draw circle
        Ellipse2D circumcircle = GeometryFunctions.circumcircle(triangle.getPointsAwt());
        Circle circle = new Circle(circumcircle.getCenterX(), 
        						   circumcircle.getCenterY(), 
        						   circumcircle.getWidth() / 2);
        circle.setPaintingStyleDecorator(circumcircleDecorator);
        this.circumcircleVis.ADTPush(circle);
        
        // Draw circle center
        if (this.triangulation.isVoronoiDrawn()) {
            Point center = circle.getCenter();
            center.setPaintingStyleDecorator(circumcircleCenterDecorator);
            this.circumcircleVis.ADTPush(center);
        }
        
        // Color vertices if they are inside the circle
        colorPoints(triangle);

        a.endOperation();
    }
    
    private void clearCircumcircle() {
    	Animator a = Animator.getActiveAnimator();
    	a.startOperation();
    	
    	// Clear circumcircle
    	this.circumcircleVis.clear();
    	
    	// Clear angle
        VoronoiConstruction.this.minAngle.setValue(Double.NaN);

        // Clear point colors
        clearPointColors();
        
        a.endOperation();
    }

    protected void clearPointColors() {
        // Clear point coloring
        Iterator vertices = VoronoiConstruction.this.triangulation.vertices();
        Point point;
        while (vertices.hasNext()) {
            point = (Point)vertices.next();
            point.setPaintingStyleDecorator(vertexDecorator);
        }
    }
    
    private void colorPoints(Triangle triangle) {
    	java.awt.Point[] trianglePoints = triangle.getPointsAwt();
		GeometryFunctions.ccwOrder(trianglePoints);
		Iterator vertices = VoronoiConstruction.this.triangulation.vertices();
    	Point point;
    	whileloop:
	    	while (vertices.hasNext()) {
	    		point = (Point)vertices.next();
	    		for (int i = 0; i < trianglePoints.length; i++) {
	    			if (trianglePoints[i].equals(point.getPoint())) {
	    				point.setPaintingStyleDecorator(circlepointDecorator);
	    				continue whileloop;
	    			}
	    		}
	    		
	    		if (GeometryFunctions.inCircleTest(point.getPoint(), 
  												   trianglePoints[0], 
												   trianglePoints[1], 
												   trianglePoints[2]) > 0) {
	        		point.setPaintingStyleDecorator(inpointDecorator);
	    		}
	    	}
    }
    
    
    
    
    class VirtualDoubleField extends JTextField {
        VirtualObject value;
        double textValue;
        
        public VirtualDoubleField() {
            super("", 8);
            this.setHorizontalAlignment(JTextField.CENTER);
            this.setEditable(false);
            this.value = new VirtualObject(null, "value");
            this.value.setObject(new Double(Double.NaN));
            this.textValue = Double.NaN;
        }

        public synchronized void setValue(double value) {
            this.value.setObject(new Double(value));
            this.validateText();
        }
        
        private synchronized double getValue() {
            return ((Double)this.value.getObject()).doubleValue();
        }
        
        private synchronized void validateText() {
            if ((this.textValue != this.getValue()) &&
                (!(Double.isNaN(this.textValue) && Double.isNaN(this.getValue())))) {
                this.textValue = this.getValue();
                if (Double.isNaN(this.getValue())) {
                    super.setText("");
                } else {
                    String str = String.valueOf(this.getValue());
                    if (str.length() > 9) {
                        super.setText(str.substring(0, 9));
                    } else {
                        super.setText(str);
                    }
                }
            }
        }

        public void validate() {
            this.validateText();
            super.validate();
        }
        
        private static final long serialVersionUID = -3666908688650017002L;
    }

    
    class MyPanel extends JPanel {

        public MyPanel(LayoutManager layout) {
            super(layout);
        }

        public void validate() {
            // Force validating of subcomponents
            int componentCount = this.getComponentCount();
            for (int i = 0; i < componentCount; i++) {
                this.getComponent(i).validate();
            }
            super.validate();
        }

        private static final long serialVersionUID = 3699821165607871718L;
    }

    
    
    
    // ** The class and methods for selectable points. **

    /**
     * The currently selected point or null.
     */
    protected VirtualObject selectedPoint = new VirtualObject(null, "selectedPoint");
    protected static long SELECTED = Selectable.clock.getNextTimeStamp();

    protected synchronized void selectPoint(SelectablePoint p) {
        this.selectedPoint.setObject(p);
    }

    protected synchronized void clearPointSelection() {
        this.selectedPoint.setObject(null);
    }

    protected synchronized Point getPointSelection() {
        return (SelectablePoint)this.selectedPoint.getObject();
    }

    protected synchronized boolean hasPointSelection() {
        return (this.getPointSelection() != null);
    }

    /**
     * A point that implements the selection functionality through
     * the Selectable interface.
     *
     * @see Selectable
     */
    class SelectablePoint extends Point implements Selectable {

		public SelectablePoint(java.awt.Point point) {
            super(point);
            // Set focusable to make selecting possible.
            this.setDraggable(true);
        }

        public SelectablePoint(java.awt.Point point, PaintingStyleDecorator psd) {
            this(point);
            this.setPaintingStyleDecorator(psd);
        }

        /* (non-Javadoc)
         * @see matrix.structures.simulationextensions.Selectable#getSelectionTime()
         */
        public long getSelectionTime() {
            if (this.isSelected()) {
                return VoronoiConstruction.SELECTED;
            } else {
                return Selectable.NOT_SELECTED;
            }
        }

        /* (non-Javadoc)
         * @see matrix.structures.simulationextensions.Selectable#setSelect(long)
         */
        public void setSelect(long globalClockTimestamp) {
            
            Animator a = Animator.getActiveAnimator();
            a.startOperation();
            if (isSelected()) {
                // The point is already selected -> clear the selection.
                this.clearSelection();
            } else if (globalClockTimestamp != Selectable.NOT_SELECTED) {
                VoronoiConstruction.this.clearTriangleSelection();
                if (VoronoiConstruction.this.hasPointSelection()) {
                    // Another point is already selected.
                    Point p1 = VoronoiConstruction.this.getPointSelection();
                    VoronoiConstruction.this.clearPointSelection();
                    SelectablePoint p2 = this;
                    if (!VoronoiConstruction.this.triangulation.addEdge(p1.getPoint(), p2.getPoint())) {
                        VoronoiConstruction.this.triangulation.removeEdge(p1.getPoint(), p2.getPoint());
                    }
                } else {
                    // Store the point as the selection.
                    VoronoiConstruction.this.selectPoint(this);
                }
            }
            a.endOperation();
        }

        private void clearSelection() {
            // Remove the stored selection.
            VoronoiConstruction.this.clearPointSelection();
        }

        private boolean isSelected() {
            return (VoronoiConstruction.this.getPointSelection() == this);
        }

		private static final long serialVersionUID = -94066704145446270L;
    }



    // ** The class and methods for a clickable polyline for implementing the flip functionality. **

    /**
     * A point that implements the selection functionality through
     * the Selectable interface.
     *
     * @see Selectable
     */
    class ClickableLineSegment extends LineSegment implements Selectable {

		public ClickableLineSegment(java.awt.Point orig, java.awt.Point dest, PaintingStyleDecorator psd) {
            super(orig, dest, psd);
            // Set focusable to make selecting possible.
            this.setDraggable(true);
        }

        /* (non-Javadoc)
         * @see matrix.structures.simulationextensions.Selectable#getSelectionTime()
         */
        public long getSelectionTime() {
            return Selectable.NOT_SELECTED;
        }

        /* (non-Javadoc)
         * @see matrix.structures.simulationextensions.Selectable#setSelect(long)
         */
        public void setSelect(long globalClockTimestamp) {
            if (globalClockTimestamp != Selectable.NOT_SELECTED) {
                Animator a = Animator.getActiveAnimator();
                a.startOperation();

                VoronoiConstruction.this.clearTriangleSelection();
                VoronoiConstruction.this.clearPointSelection();
                
                Point[] points = this.getPoints();
                VoronoiConstruction.this.triangulation.flipEdge(points[0].getPoint(), points[1].getPoint());
                
                a.endOperation();
            }
        }

        private static final long serialVersionUID = 385931983355206745L;
    }




    
    // ** The class and methods for a selectable triangle. **
    
    /**
     * The currently selected triangle or null.
     */
    protected VirtualObject selectedTriangle = new VirtualObject(null, "selectedTriangle");

    protected synchronized void selectTriangle(SelectableTriangle t) {
        this.selectedTriangle.setObject(t);
        this.circumcircle(t);
    }

    protected synchronized void clearTriangleSelection() {
        this.selectedTriangle.setObject(null);
        this.clearCircumcircle();
    }

    protected synchronized SelectableTriangle getTriangleSelection() {
        return (SelectableTriangle)this.selectedTriangle.getObject();
    }

    protected synchronized boolean hasTriangleSelection() {
        return (this.getTriangleSelection() != null);
    }

    
    /**
     * A triangle that implements the selection functionality through
     * the Selectable interface.
     *
     * @see Selectable
     */
    class SelectableTriangle extends Triangle implements Selectable {

        public SelectableTriangle(java.awt.Point[] points, PaintingStyleDecorator psd) {
            super(points, psd);
            // Set focusable to make selecting possible.
            this.setDraggable(true);
        }

        /* (non-Javadoc)
         * @see matrix.structures.simulationextensions.Selectable#getSelectionTime()
         */
        public long getSelectionTime() {
            return Selectable.NOT_SELECTED;
        }

        /* (non-Javadoc)
         * @see matrix.structures.simulationextensions.Selectable#setSelect(long)
         */
        public void setSelect(long globalClockTimestamp) {
            Animator a = Animator.getActiveAnimator();
            a.startOperation();
            if (this.isSelected()) {
                // The triangle is already selected -> clear the selection.
                this.clearSelection();
            } else if (globalClockTimestamp != Selectable.NOT_SELECTED) {
                this.clearSelection();
                // Store the triangle as the selection.
                VoronoiConstruction.this.selectTriangle(this);
            }
            a.endOperation();
        }

        private void clearSelection() {
            // Remove the stored selection.
            VoronoiConstruction.this.clearTriangleSelection();
        }

        private boolean isSelected() {
            return (VoronoiConstruction.this.getTriangleSelection() == this);
        }

        private static final long serialVersionUID = 5706197086702426910L;
    }
    
    private static final long serialVersionUID = 7723309499475075939L;
}
