package content.exercises.sda.structures;

import matrix.structures.FDT.probe.Key;
import matrix.structures.FDT.probe.Table;
import matrix.structures.memory.VirtualInteger;
import matrix.structures.memory.VirtualObject;
import matrix.structures.simulationextensions.Selectable;
import matrix.structures.spatial.PaintingStyleDecorator;
import matrix.structures.spatial.SpatialComparable;
import matrix.structures.spatial.CDT.probe.RTreeNode;
import matrix.structures.spatial.FDT.probe.SpatialElement;

/**
 * R-Tree implementation with extra support for simulation.
 * 
 * @see RTreeNode
 * @author Juha Helminen
 */
public class ExerciseRTree extends RTreeNode implements Selectable {
    private VirtualObject paintingStyleDecorator = 
        new VirtualObject(null, "PaintingStyleDecorator");
    private static boolean AUTOMATIC_SPLIT_PROPAGATION = true;
	private static boolean ALLOW_INTERNAL_NODE_SPLITS = false;
	private static VirtualInteger idCounter = new VirtualInteger(-1, null, "ExerciseRTree.idCounter");

	// Used to implement the needed simulation functionality for the exercise.
	private EntryTable entries;

	// Used to implement the Selectable interface.
	private NodeSelector nodeSelector;

	
	// ** Constructors **

	public ExerciseRTree(int minEntries, int maxEntries, NodeSelector nodeSelector, PaintingStyleDecorator psd) {
		this(minEntries, maxEntries, true, nodeSelector, psd);
	}

	public ExerciseRTree(int minEntries, int maxEntries, boolean isLeaf, NodeSelector nodeSelector, PaintingStyleDecorator psd) {
		this(minEntries, maxEntries, isLeaf, ExerciseRTree.getNewId(), nodeSelector, psd);
	}

	public ExerciseRTree(int minEntries, int maxEntries, boolean isLeaf, int id, NodeSelector nodeSelector, PaintingStyleDecorator psd) {
		super(null, minEntries, maxEntries, isLeaf, id);
		this.entries = new EntryTable();
		// Set the bounds of the array, so that the empty slots are also visualized.
		this.entries.setFirst(0);
		this.entries.setLast(maxEntries - 1);
        this.nodeSelector = nodeSelector;
        this.paintingStyleDecorator.setObject(psd);
	}


	// ** Overridden methods **
    
    public boolean isLeaf() {
        return super.isLeaf();
    }

	public void setElement(Object o) {}

	public synchronized static void resetIDCounter() {
		ExerciseRTree.idCounter.assign(-1);
	}

	protected synchronized static int getNewId() {
		return ExerciseRTree.idCounter.inc();
	}

	protected Table getEntries() {
		return this.entries;
	}

	protected void insertEntry(Object entry) {
		if (entry == null) {
			return;
		} else {
			if (entry instanceof ExerciseRTree) {
				((ExerciseRTree)entry).setParent(this);
			}
			// Insert at the first empty slot.
			int i = 0;
			while (this.getEntry(i++) != null) {}
			this.setEntry(entry, i - 1);
		}
	}
	
	protected void deleteEntry(int index) {
		this.setEntry(null, index);
		// This is done in the setObject()-method of the entry table.
		//this.numberOfEntries.dec();
		for (int i = index; i <= this.getEntries().getLast(); i++) {
			this.setEntry(this.getEntry(i + 1), i);
		}
	}

    protected RTreeNode getNewNode(int minEntries, int maxEntries, boolean isLeaf, int id) {
        return new ExerciseRTree(minEntries, maxEntries, isLeaf, id, this.nodeSelector, (PaintingStyleDecorator)this.paintingStyleDecorator.getObject());
    }
    
    
    public boolean hasPaintingStyleDecorator() {
        return (this.paintingStyleDecorator.getObject() != null);
    }

    public PaintingStyleDecorator getPaintingStyleDecorator() {
        return (PaintingStyleDecorator)this.paintingStyleDecorator.getObject();
    }

    public void setPaintingStyleDecorator(PaintingStyleDecorator psd) {
        this.paintingStyleDecorator.setObject(psd);
    }
	
	// ** Methods for executing the split operation. **

    /**
	 * Splits this node as if the insertion of the key given as parameter
	 * in this node caused it.
	 */
	public void splitThis(SpatialComparable key) {
        clearSelection();
		if (this.isLeaf()) {
            this.insertEntry(key);
			// Create two new nodes and assign the keys of this node into them.
			ExerciseRTree[] newNodes = (ExerciseRTree[])this.split();
			if (this.getParent() != null) {
				// Change the old node into the new.
				((ExerciseRTree)this.getParent()).replace(this, newNodes[0]);
				// Propagate the changes up the tree and also
				// add the other node created in the split.
				if (ExerciseRTree.AUTOMATIC_SPLIT_PROPAGATION) {
					newNodes[0].adjustTree(newNodes[1]);
				} else {
					((ExerciseRTree)this.getParent()).insertNextTo(newNodes[1], newNodes[0]);
				}
			} else {
				// Splitting root
				this.getEntries().setLast(this.getMaxEntries() - 1);
				this.setLeaf(false);
				this.setId(ExerciseRTree.getNewId());
				this.insertEntry(newNodes[0]);
				this.insertEntry(newNodes[1]);
			}

            
//            // Delete the key that should be inserted next,
//			// it's in one of the new nodes
//			newNodes[0].deleteEntry(key);
//			newNodes[1].deleteEntry(key);

        
        } else {
			if (ExerciseRTree.ALLOW_INTERNAL_NODE_SPLITS) {
				ExerciseRTree[] newNodes = (ExerciseRTree[])this.split();
				if (this.getParent() != null) {
					// Change the old node into the new.
					((ExerciseRTree)this.getParent()).replace(this, newNodes[0]);
					// Propagate the changes up the tree and also
					// add the other node created in the split.
					if (ExerciseRTree.AUTOMATIC_SPLIT_PROPAGATION) {
						newNodes[0].adjustTree(newNodes[1]);
					} else {
						((ExerciseRTree)this.getParent()).insertNextTo(newNodes[1], newNodes[0]);
					}
				} else {
					// Splitting root
					this.setId(ExerciseRTree.getNewId());
					this.insertEntry(newNodes[0]);
					this.insertEntry(newNodes[1]);
				}
			}
		}
	}

	private static void insertionSort(ExerciseRTree node) {
		SpatialElement value;
		int j;
		for (int i = 1; i <= node.getEntries().getLast(); i++) {
			value = (SpatialElement)node.getEntry(i); 
			for (j = i - 1; j >= 0; j--) {
				if ((node.getEntry(j) != null) &&
					((value == null) ||
				     (((SpatialElement)node.getEntry(j)).getComparisonString().compareTo(value.getComparisonString()) < 0))) {
					break;
				} else {
					node.setEntry(node.getEntry(j), j + 1);
				}
			}
			node.setEntry(value, j + 1);
		}
	}

	protected RTreeNode[] split() {
		ExerciseRTree[] nodes = new ExerciseRTree[2];
		// nodes[0] will replace the split node -> copy its ID
		nodes[0] = new ExerciseRTree(this.getMinEntries(), this.getMaxEntries(), this.isLeaf(), this.getId(), this.nodeSelector, (PaintingStyleDecorator)this.paintingStyleDecorator.getObject());
		nodes[1] = new ExerciseRTree(this.getMinEntries(), this.getMaxEntries(), this.isLeaf(), this.nodeSelector, (PaintingStyleDecorator)this.paintingStyleDecorator.getObject());
		this.splitDivideEntries(nodes);
		return nodes;
	}

	/**
	 * @param newSibling The extra node created when split occured or 
	 *                   null if there was no split.
	 */
	protected void adjustTree(ExerciseRTree newSibling) {
		RTreeNode parent = this.getParent();
		// Add the newSibling.
		if (newSibling == null) {
		} else if (!((ExerciseRTree)parent).isFull()) {
			// Insert the new node next to the one split (= this).
			((ExerciseRTree)parent).insertNextTo(newSibling, this);
		} else {
			// Insert the new node.
			((ExerciseRTree)parent).insertEntry(newSibling);
			ExerciseRTree[] newNodes = (ExerciseRTree[])((ExerciseRTree)parent).split();
			if (((ExerciseRTree)parent).getParent() != null) {
				((ExerciseRTree)((ExerciseRTree)parent).getParent()).replace(parent, newNodes[0]);
				newNodes[0].adjustTree(newNodes[1]);
			} else {
				((ExerciseRTree)parent).getEntries().setLast(this.getMaxEntries() - 1);
				// Parent is the root node ->
				// The creation of a new root is emulated by giving it a new id,
				// while newNodes[0] takes its old id.
				((ExerciseRTree)parent).setId(ExerciseRTree.getNewId());
				((ExerciseRTree)parent).insertEntry(newNodes[0]);
				((ExerciseRTree)parent).insertEntry(newNodes[1]);
			}
		}
	}

	public static void setAutomaticSplitPropagation(boolean bool) {
		if (!bool) {
			// Must allow splitting of internal nodes, if split propagation is disabled.
			ExerciseRTree.ALLOW_INTERNAL_NODE_SPLITS = true;
		}
		ExerciseRTree.AUTOMATIC_SPLIT_PROPAGATION = bool;
	}

	public static void setInternalNodeSplits(boolean bool) {
		if (!ExerciseRTree.AUTOMATIC_SPLIT_PROPAGATION) {
			// Must allow splitting of internal nodes, if split propagation is disabled.
			ExerciseRTree.ALLOW_INTERNAL_NODE_SPLITS = true;
		} else {
			ExerciseRTree.ALLOW_INTERNAL_NODE_SPLITS = bool;
		}
	}
	
	
	// ** Implementation of interface Selectable **

	public long getSelectionTime() {
        if (this.isSelected()) {
            return Selectable.clock.getNextTimeStamp();
        } else {
            return Selectable.NOT_SELECTED;
        }
	}

	public void setSelect(long globalClockTimestamp) {
        if (this.isSelected() || 
            (globalClockTimestamp == Selectable.NOT_SELECTED)) {
            this.nodeSelector.select(this, false);
        } else {
            this.nodeSelector.select(this, true);
        }
        
	}

    public void reSelect() {
        this.nodeSelector.select(this, true);
    }

    public void deSelect() {
        this.setSelect(Selectable.NOT_SELECTED);
    }

	private boolean isSelected() {
		return this.nodeSelector.isSelected(this);
	}

    private void clearSelection() {
        this.nodeSelector.clearSelection();
    }
	
	// ** Table for storing the keys in the nodes **
	
	/**
	 * Table fo storing the keys in a node
	 * Needed to handle the user's swap, drag and drop operations.
	 */
	private class EntryTable extends Table {

		public EntryTable() {
			super();
		}

		public void setObject(Object o, int address) {
			if (o == null) {
				if (getObject(address) != null) {
					// Removing an entry
					super.setObject(o, address);
					numberOfEntries.dec();
					if (isEmpty()) {
						setBoundingBox(null);
					} else {
						setBoundingBox(calculateBoundingBox());
					}
					fixAncestors();
				} else {
					// Replacing null with null
				}
			} else if (o instanceof Key) {
				this.setObject(((Key)o).getObject(), address);
			} else if (o instanceof SpatialComparable) {
				if (getObject(address) == null) {
					// Adding an entry
					if (getBoundingBox() == null) {
						setBoundingBox(((SpatialComparable)o).getBoundingBox());
					} else {
						setBoundingBox(add(getBoundingBox(), ((SpatialComparable)o).getBoundingBox()));
					}
					super.setObject(o, address);
					numberOfEntries.inc();
				} else {
					// Replacing an entry
					super.setObject(o, address);
					setBoundingBox(calculateBoundingBox());
				}
				fixAncestors();
			} else if (o instanceof ExerciseRTree) {
				if (getObject(address) == null) {
					// Adding an entry
					if (getBoundingBox() == null) {
						setBoundingBox(((ExerciseRTree)o).getBoundingBox());
					} else {
						setBoundingBox(add(getBoundingBox(), ((ExerciseRTree)o).getBoundingBox()));
					}
					super.setObject(o, address);
					numberOfEntries.inc();
				} else {
					// Replacing an entry
					super.setObject(o, address);
					setBoundingBox(calculateBoundingBox());
				}
				fixAncestors();
			}
            // If a node is selected, clear the selection.
            clearSelection();
		}

		private static final long serialVersionUID = 5184081427693179452L;
	}

	/**
	 * Update the obunding rectangles of the ancestors of this node.
     * Called after the contents of this node have been changed.
	 */
	private void fixAncestors() {
		ExerciseRTree parent = (ExerciseRTree)this.getParent();
		if (parent == null) {
			return;
		} else {
			parent.setBoundingBox(parent.calculateBoundingBox());
			parent.fixAncestors();
		}
	}
    
    public interface NodeSelector {
        public void select(ExerciseRTree node, boolean selected);
        public boolean isSelected(ExerciseRTree node);
        public void clearSelection();
    }
    
	private static final long serialVersionUID = -6511359780308366089L;
}
