package content.exercises.sda.structures;

import java.awt.geom.Ellipse2D;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;

import matrix.structures.spatial.FDT.probe.Point;
import matrix.structures.spatial.FDT.probe.Polygon;

/**
 * This class contains some commonly used functions required for the 
 * calculations of basic geometry in the SDA exercises.
 * 
 * All methods should be static and as general as possible
 * 
 * @author Jussi Nikander
 *
 */

public class GeometryFunctions {


	/**
	 * Calculates the angle between vectors b-a and b-c. The angle returned is between
	 * 0 <= angle <= Pi. 
	 * @param a
	 * @param b
	 * @param c
	 * @return
	 */
	public static double angle(Point a, Point b, Point c) {
		double i1, i2, j1, j2;
		i1 = a.getX() - b.getX();
		i2 = c.getX() - b.getX();
		j1 = a.getY() - b.getY();
		j2 = c.getY() - b.getY();
		return angle(i1,i2,j1,j2);
	}

	/**
	 * Calculates the angle between the vectors (i1,j1) and (i2,j2). The angle returned
	 * is between 0 <= angle <= Pi
	 * @param i1
	 * @param i2
	 * @param j1
	 * @param j2
	 * @return
	 */
	public static double angle(double i1, double i2, double j1, double j2) {
		return Math.acos((i1*i2 + j1*j2)/(Math.sqrt(i1*i1+j1*j1)*Math.sqrt(i2*i2+j2*j2)));		
	}
	
	/**
	 * Calculates whether point c is to the clockwise direction from the vector b-a.
	 * @param a
	 * @param b
	 * @param c
	 * @return
	 */
	public static boolean isClockwise(Point a, Point b, Point c) {
		double i1,i2,j1,j2;
		i1 = a.getX() - b.getX();
		i2 = c.getX() - b.getX();
		j1 = a.getY() - b.getY();
		j2 = c.getY() - b.getY();
		return isClockwise(i1,i2,j1,j2);
	}

	/**
	 * Tells whether the angle between vectors (i1,j1) and (i2,j2) is smaller in the clocwise
	 * direction or not. Will return false if the vectors paraller in either direction.
	 * @param i1
	 * @param i2
	 * @param j1
	 * @param j2
	 * @return
	 */
	public static boolean isClockwise(double i1, double i2, double j1, double j2) {
		double angle1, angle2;
		angle1 = angleToPositiveX(i1,j1);
		angle2 = angleToPositiveX(i2,j2);
		return Math.sin(angle1-angle2) > 0;
	}

	/**
	 * Tells whether the angle between vectors (i1,j1) and (i2,j2) is smaller in the counterclocwise
	 * direction or not. Will return false if the vectors paraller in either direction.
	 * @param i1
	 * @param i2
	 * @param j1
	 * @param j2
	 * @return
	 */
	public static boolean isCounterClockwise(double i1, double i2, double j1, double j2) {
		double angle1, angle2;
		angle1 = angleToPositiveX(i1,j1);
		angle2 = angleToPositiveX(i2,j2);
		return Math.sin(angle2-angle1) > 0;
	}

	/**
	 * Calculates the counterclockwise angle between the vector (i,j) and the positive
	 * x axis. The size of the angle is 0 <= angle < 2*Pi.
	 * @param i
	 * @param j
	 * @return
	 */
	private static double angleToPositiveX(double i, double j) {
		double angle = angle(1,i,0,j);
		if(j < 0) {
			angle = 2*Math.PI - angle;
		}
		return angle;
	}
	
	/**
	 * Calculates the clockwise angle from line segment b-a to line segment b-c
	 * 
	 * Therefore, if vector b-a is in the direction of the positive y-axis and vector
	 * b-c in the direction of the positive x-axis, the clockwise angle between the 
	 * two vectors is Pi/2. If, on the other hand, b-a is in the direction of the negative
	 * y-axis, the angle between the two vectors is 3*Pi/2.
	 * 
	 * Angle is returned in radians.
	 * 
	 * @param a
	 * @param b
	 * @param c
	 * @return
	 */
//	public static double clockWiseAngle(Point a, Point b, Point c) {
//		double i1,i2,j1,j2;
//		i1 = a.getX() - b.getX();
//		i2 = c.getX() - b.getX();
//		j1 = a.getY() - b.getY();
//		j2 = c.getY() - b.getY();
//		return clockWiseAngle(i1,i2,j1,j2);
//	}

	public static double clockWiseAngle(double i1, double i2, double j1, double j2) {
		double angle1, angle2;
		angle1 = angleToPositiveX(i1,j1);
		angle2 = angleToPositiveX(i2,j2);
		if(angle1 >= angle2) {
			return angle1-angle2;
		} else {
			return Math.PI*2 - angle1 + angle2;
		}		
	}
	
    /**
     * Calculates the clockwise angle from line segment b-a to line segment b-c
     * 
     * Therefore, if vector b-a is in the direction of the positive y-axis and vector
     * b-c in the direction of the positive x-axis, the clockwise angle between the 
     * two vectors is Pi/2. If, on the other hand, b-a is in the direction of the negative
     * y-axis, the angle between the two vectors is 3*Pi/2.
     * 
     * Angle is returned in radians.
     * 
     * @param a
     * @param b
     * @param c
     * @return
     */
    public static double clockWiseAngle(Point2D a, Point2D b, Point2D c) {
    	double[] ba = new double[]{ a.getX() - b.getX(), a.getY() - b.getY()};
    	double[] bc = new double[]{ c.getX() - b.getX(), c.getY() - b.getY()};
    	if (collinear(a, b, c)) {
    		if (ba[0] > 0) {
    			if (bc[0] > 0) {
    				return 0;
    			}
    		} else if (ba[0] < 0) {
    			if (bc[0] < 0) {
    				return 0;
    			}
    		} else if (ba[1] > 0) {
    			if (bc[1] > 0) {
    				return 0;
    			}
    		} else if (ba[1] < 0) {
    			if (bc[1] < 0) {
    				return 0;
    			}
    		}
    	}
    	double baLength = Math.sqrt(ba[0] * ba[0] + ba[1] * ba[1]);
    	double bcLength = Math.sqrt(bc[0] * bc[0] + bc[1] * bc[1]);
    	double dotProduct = ba[0] * bc[0] + ba[1] * bc[1];
    	double angle = Math.acos(dotProduct / (baLength * bcLength));
    	double crossProduct = ba[0] * bc[1] - ba[1] * bc[0];
    	if (crossProduct <= 0) {
    		return angle;
    	} else {
    		return (2 * Math.PI - angle);
    	}
    }

    /**
	 * Calculates the counterclockwise angle from linesegment b-a to line segment b-c.
	 * 
	 * @param a
	 * @param b
	 * @param c
	 * @return
	 */
//	public static double counterClockWiseAngle(Point a, Point b, Point c) {
//		double i1,i2,j1,j2;
//		i1 = a.getX() - b.getX();
//		i2 = c.getX() - b.getX();
//		j1 = a.getY() - b.getY();
//		j2 = c.getY() - b.getY();
//		return counterClockWiseAngle(i1,i2,j1,j2);
//	}

    /**
     * Calculates the counterclockwise angle from linesegment b-a to line segment b-c.
     * 
     * @param a
     * @param b
     * @param c
     * @return
     */
    public static double counterClockWiseAngle(Point2D a, Point2D b, Point2D c) {
        return clockWiseAngle(c, b, a);
    }
    
	/**
	 * Counterclockwise angle from vector (i1,j1) to vector (i2,j2).
	 * @param i1
	 * @param i2
	 * @param j1
	 * @param j2
	 * @return
	 */
	public static double counterClockWiseAngle(double i1, double i2, double j1, double j2) {
		double angle1, angle2;
		angle1 = angleToPositiveX(i1,j1);
		angle2 = angleToPositiveX(i2,j2);
		if(angle2 >= angle1) {
			return angle2-angle1;
		} else {
			return Math.PI*2 - angle2 + angle1;
		}
	}
	
	/**
	 * Calculates the determinant of the matrix ((a,b),(c,d)). Used to calculate,
	 * for example, line-line intersection.
	 * 
	 * @param a
	 * @param b
	 * @param c
	 * @param d
	 * @return
	 */
	public static double determinant(double a, double b, double c, double d) {
		return a*d - b*c;
	}

	/**
 	 * Calculates whether the lines represented by the parameter points intersect.
 	 * First line is defined by (x1,y1),(x2,y2) the second line by (x3,y3),(x4,y4).
 	 * 
	 * The calculation is done as per instructions at http://mathworld.wolfram.com/Line-LineIntersection.html
	 * It is assumed that the parameters represent separate lines that may intersect in one single point.
	 *
     * @param x1
     * @param x2
     * @param y1
     * @param y2
     * @param x3
     * @param x4
     * @param y3
     * @param y4
     * @return True if line segments intersect, false otherwise
     */
    public static boolean linesIntersect(int x1, int x2, int y1, int y2, int x3, int x4, int y3, int y4) {
    	return determinant(x1-x2, y1-y2, x3-x4, y3-y4) != 0;
    }

    public static boolean linesIntersect(Point a1, Point a2, Point b1, Point b2) {
    	return linesIntersect((int)a1.getX(), (int)a2.getX(),(int)a1.getY(),(int)a2.getY(),(int)b1.getX(),(int)b2.getX(),(int)b1.getY(),(int)b2.getY());
    }
    
	/**
 	 * Calculates whether the line segments represented by the parameter points intersect.
 	 * First line segment is defined by (x1,y1),(x2,y2) the second line segment by (x3,y3),(x4,y4).
 	 * 
	 * The calculation is done as per instructions at http://mathworld.wolfram.com/Line-LineIntersection.html
	 * It is assumed that the parameters represent separate line segments that intersect in one single point.
	 * If the lines intersect in an area between the parameter points, true is returned.
	 *
     * @param x1
     * @param x2
     * @param y1
     * @param y2
     * @param x3
     * @param x4
     * @param y3
     * @param y4
     * @return True if line segments intersect, false otherwise
     */
    public static boolean segmentsIntersect(int x1, int x2, int y1, int y2, int x3, int x4, int y3, int y4) {
		double x = determinant(determinant(x1,y1,x2,y2), x1-x2, determinant(x3,y3,x4,y4), x3-x4);
		double tmp = determinant(x1-x2, y1-y2, x3-x4, y3-y4);
		// case of parallel lines
		if(tmp == 0) {
			return false;
		}
		x = x / tmp;
		double y = determinant(determinant(x1,y1,x2,y2), y1-y2, determinant(x3,y3,x4,y4), y3-y4);
		tmp = determinant(x1-x2, y1-y2, x3-x4, y3-y4);
		/* Unneeded, the already done with x coordinate
		if(tmp == 0) {
			return false;
		}
		*/
		y = y / tmp;
		// check if intersection inside line segments
		if(((x > x1 && x < x2) || (x < x1 && x > x2)) 
				&& ((y > y1 && y < y2) || (y < y1 && y > y2))
				&& ((x > x3 && x < x4) || (x < x3 && x > x4)) 
				&& ((y > y3 && y < y4) || (y < y3 && y > y4))) {
			return true;			
		}
		return false;
    }

    public static boolean segmentsIntersect(Point a1, Point a2, Point b1, Point b2) {
    	return segmentsIntersect((int)a1.getX(), (int)a2.getX(),(int)a1.getY(),(int)a2.getY(),(int)b1.getX(),(int)b2.getX(),(int)b1.getY(),(int)b2.getY());
    }
    
    /**
 	 * Calculates the intersection point of the lines represented by the parameter points.
 	 * First line is defined by (x1,y1),(x2,y2) the second line by (x3,y3),(x4,y4).
 	 * The intersection is returned as java.awt.Point (since creating new Matrix Points is a laborous
 	 * process that involves the Animator) or null if the lines do not intersect. 
 	 * 
	 * The calculation is done as per instructions at http://mathworld.wolfram.com/Line-LineIntersection.html
	 * It is assumed that the parameters represent separate line segments that intersect in one single point.
	 *
     * @param x1
     * @param x2
     * @param y1
     * @param y2
     * @param x3
     * @param x4
     * @param y3
     * @param y4
     * @return a new Point representing the point of intersection or null if the line segments do not intersect
     */
    public static java.awt.Point lineIntersection(int x1, int x2, int y1, int y2, int x3, int x4, int y3, int y4) {
		double x = determinant(determinant(x1,y1,x2,y2), x1-x2, determinant(x3,y3,x4,y4), x3-x4);
		double tmp = determinant(x1-x2, y1-y2, x3-x4, y3-y4);
		// case of parallel lines
		if(tmp == 0) {
			return null;
		}
		x = x / tmp;
		double y = determinant(determinant(x1,y1,x2,y2), y1-y2, determinant(x3,y3,x4,y4), y3-y4);
		tmp = determinant(x1-x2, y1-y2, x3-x4, y3-y4);
		/* Unneeded, the already done with x coordinate
		if(tmp == 0) {
			return false;
		}
		*/
		y = y / tmp;
		// check if intersection inside line segments
		return new java.awt.Point((int)x,(int)y);			
    }

    public static java.awt.Point lineIntersection(Point a1, Point a2, Point b1, Point b2) {
    	return lineIntersection((int)a1.getX(), (int)a2.getX(),(int)a1.getY(),(int)a2.getY(),(int)b1.getX(),(int)b2.getX(),(int)b1.getY(),(int)b2.getY());
    }

    public static java.awt.Point segmentIntersection(Point a1, Point a2, Point b1, Point b2) {
    	return segmentIntersection((int)a1.getX(), (int)a2.getX(),(int)a1.getY(),(int)a2.getY(),(int)b1.getX(),(int)b2.getX(),(int)b1.getY(),(int)b2.getY());
    }

    public static java.awt.Point segmentIntersection(Point2D a1, Point2D a2, Point2D b1, Point2D b2) {
        return segmentIntersection((int)a1.getX(), (int)a2.getX(),(int)a1.getY(),(int)a2.getY(),(int)b1.getX(),(int)b2.getX(),(int)b1.getY(),(int)b2.getY());
    }

    public static java.awt.Point lineIntersection(Line2D l1, Line2D l2) {
    	return lineIntersection((int)l1.getX1(), (int)l1.getX2(), (int)l1.getY1(), (int)l1.getY2(),
    							(int)l2.getX1(), (int)l2.getX2(), (int)l2.getY1(),(int)l2.getY2());
    }

    /**
 	 * Calculates the intersection point of the line segments represented by the parameter points.
 	 * First line segment is defined by (x1,y1),(x2,y2) the second line segment by (x3,y3),(x4,y4).
 	 * The intersection is returned as java.awt.Point (since creating new Matrix Points is a laborous
 	 * process that involves the Animator) or null if the line segments do not intersect. 
 	 * 
	 * The calculation is done as per instructions at http://mathworld.wolfram.com/Line-LineIntersection.html
	 * It is assumed that the parameters represent separate line segments that intersect in one single point.
	 * If the lines intersect outside an area between the parameter points, null is returned.
	 *
     * @param x1
     * @param x2
     * @param y1
     * @param y2
     * @param x3
     * @param x4
     * @param y3
     * @param y4
     * @return a new Point representing the point of intersection or null if the line segments do not intersect
     */
    public static java.awt.Point segmentIntersection(int x1, int x2, int y1, int y2, int x3, int x4, int y3, int y4) {
		double x = determinant(determinant(x1,y1,x2,y2), x1-x2, determinant(x3,y3,x4,y4), x3-x4);
		double tmp = determinant(x1-x2, y1-y2, x3-x4, y3-y4);
		// case of parallel lines
		if(tmp == 0) {
			return null;
		}
		x = x / tmp;
		double y = determinant(determinant(x1,y1,x2,y2), y1-y2, determinant(x3,y3,x4,y4), y3-y4);
		tmp = determinant(x1-x2, y1-y2, x3-x4, y3-y4);
		/* Unneeded, the already done with x coordinate
		if(tmp == 0) {
			return false;
		}
		*/
		y = y / tmp;
		// check if intersection inside line segments
		if(((x > x1 && x < x2) || (x < x1 && x > x2)) 
				&& ((y > y1 && y < y2) || (y < y1 && y > y2))
				&& ((x > x3 && x < x4) || (x < x3 && x > x4)) 
				&& ((y > y3 && y < y4) || (y < y3 && y > y4))) {
			return new java.awt.Point((int)x,(int)y);			
		}
		return null;    	
    }
    
    
    /**
     * Constructs a delaunay triangulation of the given set of points. 
     * This is done with a very inefficient brute-force approach where 
     * delaunay triangles are found by iterating over all the possible 
     * triangles and testing every other point in the set with the 
     * circumcircle test against the triangle, that is, the triangle 
     * is a delaunay if any other point in the set lies outside 
     * the circumcircle defined by its three vertices.
     * 
     * @param points The set of points for which to construct the triangulation.
     * @return The delaunay triangles in no specific order. Each triangle is
     *         a single item in the array, which in turn is an array of three
     *         points, the vertices of the triangle in counterclockwise order.
     *         Null is returned if the point set contains 4 or more cocircular/
     *         concyclic points, in this case the triangulation isn't unambiguous.
     */
    public static java.awt.Point[][] delaunay(java.awt.Point[] points) {
        ArrayList triangles = new ArrayList();
        double testValue = 0;
        java.awt.Point[] ccw = new java.awt.Point[3];
        // For each triple of points test that any other point lies outside the circumcircle
        // defined by the three points.
        for (int i = 0; i < points.length - 2; i++) {
            for (int j = i + 1; j < points.length - 1; j++) {
                for (int k = j + 1; k < points.length; k++) {
                    // Determine the counterclockwise order of the points to construct 
                	// an oriented circle for the circumcircle-test.
                	ccw[0] = points[i]; ccw[1] = points[j]; ccw[2] = points[k];
                	ccwOrder(ccw);
                    // Test that any other points lie outside the circumcircle.
                    for (int l = 0; l < points.length; l++) {
                        if ((l == i) || (l == j) || (l == k)) {
                            continue;
                        }
                        // Test if points[l] is inside the circumcircle defined by the triangle 
                        // formed with points[i], points[j], points[k] 
                        testValue = inCircleTest(points[l], ccw[0], ccw[1], ccw[2]);
                        if (testValue > 0) {
                        	break;
                        } else if (testValue == 0) {
                        	// 4 cocircular points
                        	return null;
                        }
                    }
                    if (testValue <= 0) {
                        // There are no points that lie inside the circumcircle of 
                        // points[i], points[j], points[k] -> we've found a delaunay triangle.
                        triangles.add(new java.awt.Point[]{ccw[0], ccw[1], ccw[2]});
                    }
                }
            }
        }
        java.awt.Point[][] triangulation = new java.awt.Point[triangles.size()][3];
        triangles.toArray(triangulation);
        return triangulation;
    }
    
    
    public static Ellipse2D circumcircle(Point2D[] trianglePoints) {
        // Compute the circumcircle as described in
        // http://mathworld.wolfram.com/Circumcircle.html
        int i = 0, j = 1, k = 2;
        double a =
            GeometryFunctions.det(new double[][]{
                    {trianglePoints[i].getX(), trianglePoints[i].getY(), 1},
                    {trianglePoints[j].getX(), trianglePoints[j].getY(), 1},
                    {trianglePoints[k].getX(), trianglePoints[k].getY(), 1}});
        double bx =
            -GeometryFunctions.det(new double[][]{
                     {trianglePoints[i].getX() * trianglePoints[i].getX() + trianglePoints[i].getY() * trianglePoints[i].getY(), trianglePoints[i].getY(), 1},
                     {trianglePoints[j].getX() * trianglePoints[j].getX() + trianglePoints[j].getY() * trianglePoints[j].getY(), trianglePoints[j].getY(), 1},
                     {trianglePoints[k].getX() * trianglePoints[k].getX() + trianglePoints[k].getY() * trianglePoints[k].getY(), trianglePoints[k].getY(), 1}});
        double by =
            GeometryFunctions.det(new double[][]{
                    {trianglePoints[i].getX() * trianglePoints[i].getX() + trianglePoints[i].getY() * trianglePoints[i].getY(), trianglePoints[i].getX(), 1},
                    {trianglePoints[j].getX() * trianglePoints[j].getX() + trianglePoints[j].getY() * trianglePoints[j].getY(), trianglePoints[j].getX(), 1},
                    {trianglePoints[k].getX() * trianglePoints[k].getX() + trianglePoints[k].getY() * trianglePoints[k].getY(), trianglePoints[k].getX(), 1}});
        double x = -bx / (2 * a);
        double y = -by / (2 * a);
        double c =
            -GeometryFunctions.det(new double[][]{
                    {trianglePoints[i].getX() * trianglePoints[i].getX() + trianglePoints[i].getY() * trianglePoints[i].getY(), trianglePoints[i].getX(), trianglePoints[i].getY()},
                    {trianglePoints[j].getX() * trianglePoints[j].getX() + trianglePoints[j].getY() * trianglePoints[j].getY(), trianglePoints[j].getX(), trianglePoints[j].getY()},
                    {trianglePoints[k].getX() * trianglePoints[k].getX() + trianglePoints[k].getY() * trianglePoints[k].getY(), trianglePoints[k].getX(), trianglePoints[k].getY()}});
        double r = Math.sqrt(bx * bx + by * by - 4 * a * c) / (2 * Math.abs(a));
        return new Ellipse2D.Double(x - r, y - r, 2 * r, 2 * r);
    }

    
    
    /**
     * @param convexPolygon Must be convex
     * @return
     */
    public static double minAngle(Polygon convexPolygon) {
        Point[] points = convexPolygon.getPoints();
        if (points.length > 2) {
            double minAngle =
                Math.min(GeometryFunctions.angle(points[points.length - 2], points[points.length - 1], points[0]),
                		 GeometryFunctions.angle(points[points.length - 1], points[0], points[1]));
            for (int i = 0; i < points.length - 2; i++) {
                minAngle =
                    Math.min(minAngle,
                             GeometryFunctions.angle(points[i], points[i + 1], points[i + 2]));
            }
            return minAngle;
        } else {
            return 0;
        }
    }
    

    private static boolean collinear(Point2D p0, Point2D p1, Point2D p2) {
    	// The area of a triangle  determined by three points will be zero iff they are collinear.
    	return ((p0.getX() * (p1.getY() - p2.getY()) + 
    			p1.getX() * (p2.getY() - p0.getY()) + 
    			p2.getX() * (p0.getY() - p1.getY())) == 0);
    }

    
    private static class CcwComparator implements Comparator {
    	private java.awt.Point center;
    	private java.awt.Point reference;
    	
    	public CcwComparator(java.awt.Point center, java.awt.Point reference) {
    		this.center = center;
    		this.reference = reference;
		}
    	
    	public int compare(Object arg0, Object arg1) {
    		java.awt.Point p0 = (java.awt.Point)arg0;
    		java.awt.Point p1 = (java.awt.Point)arg1;
    		double angle0, angle1;
    		if (p0.equals(this.center)) {
    			return -1;
    		} else if (p1.equals(this.center)) {
    			return 1;
    		} else {
				angle0 = GeometryFunctions.counterClockWiseAngle(this.reference, this.center, p0);
				angle1 = GeometryFunctions.counterClockWiseAngle(this.reference, this.center, p1);
				double difference = angle0 - angle1; 
				if (difference < 0 ) {
					return -1;
				} else if (difference > 0 ) {
					return 1;
				} else {
					double distanceDiff = p0.distance(this.center) - p1.distance(this.center);
					if (distanceDiff < 0) {
						return -1;
					} else if (distanceDiff > 0) {
						return 1;
					} else {
						return 0;
					}
				}
    		}
    	}
    	
    }
    
    public static void ccwSort(java.awt.Point points[], java.awt.Point center, java.awt.Point reference) {
    	Comparator comparator = new CcwComparator(center, reference);
    	Arrays.sort(points, comparator); 
    }
    
    // must be only three points
    public static void ccwOrder(Point2D trianglePoints[]) {
    	if (GeometryFunctions.counterClockWiseAngle(trianglePoints[0], trianglePoints[1], trianglePoints[2]) < Math.PI) {
    	    Point2D temp = trianglePoints[0];
    		trianglePoints[0] = trianglePoints[1];
    		trianglePoints[1] = temp;
    	} else {
    		return;
    	}
    }

//    // Incircle will return 0 if all four points are cocircular, a positive value if d is inside the circle, and negative if d is outside.
//    // Points c1, c2, c3 must be ordered counterclockwise (oriented circle)
//    public static double inCircleTest(java.awt.Point p, 
//    										 java.awt.Point c1, java.awt.Point c2, java.awt.Point c3) {
//        // Test if p is inside the circumcircle defined by the triangle 
//        // formed with c1, c2, c3 
//        // as described in Guibas, L. and Stolfi, J. 1985 - Primitives for the manipulation 
//        // of general subdivisions and the computation of Voronoi. 
//    	return
//    		det(new double[][]{
//                    {c1.x, c1.y, c1.x * c1.x + c1.y * c1.y, 1},
//                    {c2.x, c2.y, c2.x * c2.x + c2.y * c2.y, 1},
//                    {c3.x, c3.y, c3.x * c3.x + c3.y * c3.y, 1},
//                    {p.x, p.y, p.x * p.x + p.y * p.y, 1}});
//    }
    
    public static double inCircleTest(Point2D p, Point2D c1, Point2D c2, Point2D c3) {
        // Test if p is inside the circumcircle defined by the triangle 
        // formed with c1, c2, c3 
        // as described in Guibas, L. and Stolfi, J. 1985 - Primitives for the manipulation 
        // of general subdivisions and the computation of Voronoi. 
        return
            det(new double[][]{
                    {c1.getX(), c1.getY(), c1.getX() * c1.getX() + c1.getY() * c1.getY(), 1},
                    {c2.getX(), c2.getY(), c2.getX() * c2.getX() + c2.getY() * c2.getY(), 1},
                    {c3.getX(), c3.getY(), c3.getX() * c3.getX() + c3.getY() * c3.getY(), 1},
                    {p.getX(), p.getY(), p.getX() * p.getX() + p.getY() * p.getY(), 1}});
    }

    /**
     * Computes the determinant of a square matrix using the technique 
     * known as Laplacian determinant expansion by minors as decribed 
     * in http://mathworld.wolfram.com/DeterminantExpansionbyMinors.html.
     * 
     * @param matrix The matrix as an array of rows, that is, the first
     *               index is the row and the second index the column.
     * @return The determinant or Double.NaN if the given array is null, 
     *         zero-length or the number of rows does not equal the number 
     *         of colums.
     */
    public static double det(double[][] matrix) {
      if ((matrix == null) || 
          (matrix.length < 1) || 
          (matrix[0] == null) || 
          (matrix.length != matrix[0].length)) {
          return Double.NaN;
      } else {
          if (matrix.length == 1) {
              return matrix[0][0];
          } else {
              double det = 0;
              for (int minor = 0; minor < matrix[0].length; minor++) {
                  double[][] minorMatrix = new double[matrix.length - 1][matrix[0].length - 1];  
                  for (int row = 1, minorRow = 0; row < matrix.length; row++, minorRow++) {
                      for (int column = 0, minorColumn = 0; column < matrix[0].length; column++) {
                          if (column == minor) {
                              continue;
                          } else {
                              minorMatrix[minorRow][minorColumn++] = matrix[row][column];
                          }
                      }
                  }
                  det = det + Math.pow(-1, minor) * matrix[0][minor] * det(minorMatrix);
              }
              return det;
          }
      }
        
    }
    

    /**
     * <p>
     * An implementation of the gaussian elimination algorithm, a method 
     * for finding the solutions of a system of linear equations, that is,
     * solving for equations of the form Ax = y, where A and y are known.
     * </p>
     * 
     * @param a A square matrix with as many rows/columns as y has rows.
     *          The matrix is stored in the array in row,column-order.
     * @param y A matrix/vector with as many rows as matrix a has rows/columns.
     *          The matrix is stored in the array in row,column-order.
     * @return  The solution matrix that equals the size the matrix/vector y. 
     *          The matrix is stored in the array in row,column-order.
     */
    public static double[][] gaussianElimination(double[][] a, double[][] y) {
        // Construct the augmented matrix [a|y]
        double[][] matrix = new double[a.length][a[0].length + y[0].length];
        int rows = matrix.length;
        int columns = matrix[0].length;
        for (int i = 0; i < rows; i++) {
            for (int j = 0; j < columns; j++) {
                if (j < a[0].length) {
                    matrix[i][j] = a[i][j];
                } else {
                    matrix[i][j] = y[i][j - a[0].length];
                }
            }
        }
        
        // Convert the augmented matrix into upper triangular form with
        // elementary matrix operations.
        int i = 0;
        int j = 0;
        int maxi;
        double tempValue;
        while ((i < rows) && (j < columns)) {
            // Find the pivot value in column j starting from row i:
            maxi = i;
            for (int k = i + 1; k < rows; k++) {
                if (Math.abs(matrix[k][j]) > Math.abs(matrix[maxi][j])) {
                    maxi = k;
                }
            }
            if (matrix[maxi][j] != 0) {
                // Swap rows i and maxi
                for (int w = 0; w < columns; w++) {
                    tempValue = matrix[i][w];
                    matrix[i][w] = matrix[maxi][w];
                    matrix[maxi][w] = tempValue;
                }
                // a[i][j] contains the old value of a[maxi][j] ->
                // divide each entry in row i by A[i][j]
                tempValue = matrix[i][j];
                for (int w = 0; w < columns; w++) {
                    matrix[i][w] = matrix[i][w] / tempValue;
                }
                for (int u = i + 1; u < rows; u++) {
                    // subtract A[u][j] * row i from row u
                    tempValue = matrix[u][j];
                    for (int w = 0; w < columns; w++) {
                        matrix[u][w] = matrix[u][w] - (tempValue * matrix[i][w]);
                    }
                }
                i++;
            }
            j++;
        }
        
        // The solution matrix/vector
        double x[][] = new double[y.length][y[0].length];
        // Back substitution
        for (int u = 0; u < y[0].length; u++) {
            if (matrix[a.length - 1][a[0].length - 1] != 0) {
                x[x.length - 1][u] =  
                    matrix[matrix.length - 1][a.length + u] / matrix[a.length - 1][a[0].length - 1];
            }
    
            for (int k = rows - 2; k >= 0; k--) {
                x[k][u] = matrix[k][a.length + u];
                for (int l = k + 1; l < rows; l++) {
                    x[k][u] = x[k][u] - matrix[k][l] * x[l][u];
                }
                x[k][u] = x[k][u] / matrix[k][k];
            }
        }
        return x;
    }
    
    
    /**
     * Constructs a cubic spline interpolation for the given data points that
     * define a closed path (the last data point connects to the first one).
     * The interpolation curve is constructed using b-splines and then 
     * described as a series of cubic bezier curves.
     * 
     * @param dataPoints Data points that define a closed path which is to be 
     *                   interpolated using splines.
     * @return           An array of the control points for bezier curves of the 
     *                   interpolation curve, that is, an array of 4-item arrays
     *                   which are the four control points used to define a cubic
     *                   bezier curve. 
     */
    public static Point2D[][] splineInterpolation(Point2D[] dataPoints) {
        // [row][column]
        // Formulate the equation used to solve the interpolation curve.
        // It is a matrix equation of the form Ax=y, where A is a tridiagonal
        // square matrix and y is derived from the data points.
        double[][] y = new double[dataPoints.length][2];
        // the tridiagonal matrix A
        double[][] matrix141 = new double[dataPoints.length][dataPoints.length];
        for (int i = 0; i < matrix141.length; i++) {
            for (int j = 0; j < matrix141.length; j++) {
                if (i == j) {
                    matrix141[i][j] = 4;
                } else if ((j == i - 1) || (j == i + 1)) {
                    matrix141[i][j] = 1;
                } else {
                    matrix141[i][j] = 0;
                }
            }
        }
        matrix141[0][matrix141.length - 1] = 1;
        matrix141[matrix141.length - 1][0] = 1;

        for (int i = 0; i < y.length; i++) {
            y[i][0] = 6 * dataPoints[i].getX();
            y[i][1] = 6 * dataPoints[i].getY();
        }
        
        // Solve the matrix equation to obtain the b-spline control points.
        double[][] bspline = GeometryFunctions.gaussianElimination(matrix141, y);

        // Construct the cubic bezier control points from the b-spline control points.
        Point2D.Double[][] bezier = new Point2D.Double[dataPoints.length][4];
        double vector[] = new double[2];
        for (int i = 0; i < dataPoints.length - 1; i++) {
            vector[0] = bspline[i + 1][0] - bspline[i][0];
            vector[1] = bspline[i + 1][1] - bspline[i][1];
            vector[0] /= 3; vector[1] /= 3;
            bezier[i][0] = new Point2D.Double(dataPoints[i].getX(), dataPoints[i].getY());
            bezier[i][1] = new Point2D.Double(bspline[i][0] + vector[0], bspline[i][1] + vector[1]);
            bezier[i][2] = new Point2D.Double(bspline[i + 1][0] - vector[0], bspline[i + 1][1] - vector[1]);
            bezier[i][3] = new Point2D.Double(dataPoints[i + 1].getX(), dataPoints[i + 1].getY());
        }
        vector[0] = bspline[0][0] - bspline[bspline.length - 1][0];
        vector[1] = bspline[0][1] - bspline[bspline.length - 1][1];
        vector[0] /= 3; vector[1] /= 3;
        bezier[bezier.length - 1][0] = new Point2D.Double(dataPoints[dataPoints.length - 1].getX(), 
                                                          dataPoints[dataPoints.length - 1].getY());
        bezier[bezier.length - 1][1] = new Point2D.Double(bspline[bspline.length - 1][0] + vector[0], 
                                                          bspline[bspline.length - 1][1] + vector[1]);
        bezier[bezier.length - 1][2] = new Point2D.Double(bspline[0][0] - vector[0], 
                                                          bspline[0][1] - vector[1]);
        bezier[bezier.length - 1][3] = new Point2D.Double(dataPoints[0].getX(), 
                                                          dataPoints[0].getY());
        
        return bezier;
    }
    
}
