package content.exercises.sda.structures;

import java.io.Serializable;

import matrix.structures.FDT.Tree;
import matrix.structures.memory.VirtualBoolean;
import matrix.structures.memory.VirtualObject;
import matrix.structures.simulationextensions.Minimized;
import matrix.structures.simulationextensions.Selectable;
import matrix.structures.spatial.Area;
import matrix.structures.spatial.PaintingStyleDecorator;
import matrix.structures.spatial.SpatialComparable;
import matrix.structures.spatial.CDT.probe.RTree;
import matrix.structures.spatial.CDT.probe.RTreeNode;
import matrix.structures.spatial.FDT.probe.SpatialElement;
import matrix.util.Note;

public class PolygonRTree extends RTree {
    
    private VirtualObject selection;
    private LeafFunction leafFunction;
    private SelectFunction selectFunction;
    private boolean visualizeMinimized;

    public PolygonRTree() {
        this.root = new VirtualObject(this, "root");
        this.root.setObject(new PolygonRTreeNode());
        this.selection = new VirtualObject(this, "root");
        this.leafFunction = null;
        this.selectFunction = null;
        this.visualizeMinimized = false;
    }

    public static void resetIDCounter() {
        PolygonRTreeNode.resetIDCounter();
    }
    
    public void minimize() {
        ((PolygonRTreeNode)this.getElement()).minimizeSubTree();
    }
    
    public interface LeafFunction extends Serializable {
        public void execute(PolygonRTreeNode node, boolean minimized);
    }
    
    public interface SelectFunction extends Serializable {
        public void execute(PolygonRTreeNode node, boolean select);
    }

    public void setLeafFunction(LeafFunction leafFunction) {
        this.leafFunction = leafFunction;
    }
    
    public void setSelectFunction(SelectFunction selectFunction) {
        this.selectFunction = selectFunction;
    }
    
    public void setVisualizeMinimized(boolean visualizeMinimized) {
        this.visualizeMinimized = visualizeMinimized;
    }

    public class PolygonRTreeNode extends RTreeNode implements Minimized, Selectable {

        private VirtualBoolean isMinimized;
        
        // ** Constructors **
        
        public PolygonRTreeNode() {
            super();
            this.isMinimized = new VirtualBoolean(true, this, "isMinimized");
        }

        public PolygonRTreeNode(SpatialComparable key) {
            super(key);
            this.isMinimized = new VirtualBoolean(true, this, "isMinimized");
        }

        public PolygonRTreeNode(int minEntries, int maxEntries, boolean isLeaf, int id,
                                PaintingStyleDecorator psd) {
            super(minEntries, maxEntries, isLeaf, id);
            this.isMinimized = new VirtualBoolean(true, this, "isMinimized");
            // The split children inherit the painting style decorator.
            this.setPaintingStyleDecorator(psd);
        }

        
        // ** Implementation of interface Minimized **

        public boolean isMinimized() {
            return this.isMinimized.eval();
        }

        public void setMinimized(boolean value) {
            this.isMinimized.assign(value);
            if (this.isLeaf()) {
                leafFunction.execute(this, value);
            }
        }


        // ** Implementation of interface Selectable **
        
        public long getSelectionTime() {
            if (isSelected()) {
                return Selectable.clock.getNextTimeStamp();
            } else {
                return Selectable.NOT_SELECTED;
            }
        }
        
        private void selectFunctionExecute(PolygonRTreeNode node, boolean select)  {
            if (selectFunction != null) {
                selectFunction.execute(node, select);
            }
        }

        public void clearSelection() {
            if (selection.getObject() != null) {
                selectFunctionExecute((PolygonRTreeNode)selection.getObject(), false);
                selection.setObject(null);
            }
        }
        
        public void setSelect(long globalClockTimestamp) {
            if (this.isSelected()) {
                if (this.isMinimized()) {
                    this.setMinimized(!isMinimized());
                }
            } else {
                if (selection.getObject() != null) {
                    selectFunctionExecute((PolygonRTreeNode)selection.getObject(), false);
                }
                selection.setObject(this);
                selectFunctionExecute(this, true);
            }
        }

        private boolean isSelected() {
            if (selection.getObject() == this) {
                return true;
            } else {
                return false;
            }
        }
        

        // ** Overridden methods in interface Area **

        public SpatialElement[] getSpatialElements() {
            if (this.isMinimized() && !visualizeMinimized) {
                return new SpatialElement[0];
            } else {
                return super.getSpatialElements();
            }
        }

        public Area[] getSubAreas() {
            if (this.isMinimized() && !visualizeMinimized) {
                return new Area[0];
            } else {
                return super.getSubAreas();
            }
        }

        // ** Overridden methods in interface Tree **

        public int getSubTreeCount() {
            if (this.isMinimized()) {
                return 0;
            } else {
                return super.getSubTreeCount();
            }
        }
        
        public Tree[] getSubTrees() {
            if (this.isMinimized()) {
                return new Tree[0];
            } else {
                return super.getSubTrees();
            }
        }

        public Tree getNewNode(Object key) {
            if (key instanceof SpatialComparable) {
                return new PolygonRTreeNode((SpatialComparable)key);
            } else {
                Note.warning(this, "Cannot create a new node with key of type " + key.getClass());
                return null;
            }
        }

        protected RTreeNode getNewNode(int minEntries, int maxEntries, boolean isLeaf, int id) {
            return new PolygonRTreeNode(minEntries, maxEntries, isLeaf, id, this.getPaintingStyleDecorator());
        }

        
        

        // ** Overridden methods in interface LabelDecorator **

        public boolean isLabelEnabled() {
            if (visualizeMinimized) {
                return super.isLabelEnabled();
            } else {
                if (this.isMinimized() || this.isSelected() || this.isLeaf()) {
                    return super.isLabelEnabled();
                } else {
                    return false;
                }
            }
        }


        // ** Overridden methods in interface FDT **
        
        public Object getElement() {
            if (this.isMinimized()) {
                return null;
            } else {
                return super.getElement();
            }
        }
        
        
        // ** Other overridden methods **
        
        public boolean equals(Object obj) {
            if (obj instanceof PolygonRTree) {
                return this.equals(((PolygonRTree)obj).getElement());
            } else if (obj instanceof PolygonRTreeNode) {
                // Comparison is solely based on the state of the minimizations.
                PolygonRTreeNode node = (PolygonRTreeNode)obj;
                if ((this.isMinimized() == node.isMinimized()) &&
                    (this.isSelected() == node.isSelected())) {
                    if (this.isMinimized() || this.isLeaf()) {
                        return true;
                    } else {
                        // Compare children.
                        for (int i = 0; i < this.numberOfEntries(); i++) {
                            if (!this.getEntry(i).equals(node.getEntry(i))) {
                                return false;
                            }
                        }
                        // Children were equal.
                        return true;
                    }
                } else {
                    return false;
                }
            } else {
                return false;
            }
        }
        
        public String toString() {
            return "";
        }

        public boolean isLeaf() {
            return super.isLeaf();
        }
        
        
        // ** Other methods **

        public void minimizeSubTree() {
            this.clearSelection();
            this.minimize();
        }
        
        private void minimize() {
            this.setMinimized(true);
            if (!this.isLeaf()) {
                for (int i = 0; i < this.numberOfEntries(); i++) {
                    ((PolygonRTreeNode)this.getEntry(i)).minimize();
                }
            }
        }
        
        private static final long serialVersionUID = 6954905693255977590L;
    }
    
    private static final long serialVersionUID = 8901660830219891489L;
}
