package content.exercises.sda.structures;

import java.awt.Rectangle;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.Vector;

import matrix.structures.ADT.Stack;
import matrix.structures.CDT.CDT;
import matrix.structures.FDT.Array;
import matrix.structures.FDT.FDT;
import matrix.structures.FDT.probe.Key;
import matrix.structures.memory.VirtualArray;
import matrix.structures.memory.VirtualInteger;
import matrix.structures.memory.VirtualObject;
import matrix.structures.spatial.Area;
import matrix.structures.spatial.PaintingStyleDecorator;
import matrix.structures.spatial.SpatialComparable;
import matrix.structures.spatial.FDT.probe.SpatialElement;
import matrix.util.Note;

/**
 * A stack implementation which can be visualized in several ways.
 * 
 * The basic visualization is through the interface <code>Array</code>
 * as a single-item array, which represents the top of the stack.
 * 
 * If the stack contains elements of class <code>SpatialComparable</code> 
 * or <code>Area</code> these can be visualized via the interface 
 * <code>Area</code>.
 * 
 * Finally, the contents of the whole stack can be visualized 
 * as an array, which is provided by the method 
 * <code>getArrayVisualization()</code>. <br><br>
 * 
 * Additionally, if one intends to store elements of class
 * <code>SpatialElement</code> in the stack, the top of the 
 * stack can be assigned to always have a certain 
 * <code>PaintingStyleDecorator</code>.
 * 
 * @author Juha Helminen
 */
public class SimpleStack implements FDT, CDT, Array, Stack, Area {
    
    private VirtualObject paintingStyleDecorator;
    private VirtualInteger stackPointer;
    private PaintingStyleDecorator savedDecorator;
    private final SimpleArray contents;
    private final Rectangle boundingBox;
    private final PaintingStyleDecorator topDecorator;
    private final LinkedHashSet subAreas;

    // ** Constructors **

    public SimpleStack() {
        this(new Rectangle(0, 0, 0, 0));
    }

    public SimpleStack(Rectangle boundingBox) {
        this(boundingBox, null);
    }

    /**
     * @param boundingBox   The bounding box for the <code>Area</code>-visualization.
     * @param topDecorator  Style decorator to be used for the top element of the stack,
     *                      if it's of type <code>SpatialElement</code>.
     */
    public SimpleStack(Rectangle boundingBox, PaintingStyleDecorator topDecorator) {
        this.boundingBox = boundingBox;
        this.topDecorator = topDecorator;
        this.contents = new SimpleArray();
        this.paintingStyleDecorator = new VirtualObject(null, "PaintingStyleDecorator");
        this.stackPointer = new VirtualInteger(0, this, "stackPointer");
        this.savedDecorator = null;
        this.subAreas = new LinkedHashSet();
    }

    // ** Implementation of interface FDT **
    
    public Object getElement() {
        // XXX
        Note.warning(this, "Operation not implemented.");
        return null;
    }

    public void setElement(Object o) {
        // XXX
        Note.warning(this, "Operation not implemented.");
    }

    // ** Implementation of interface CDT **
    
    public CDT insert(Object o) {
        this.ADTPush(o);
        return this;
    }

    public CDT delete(Object o) {
        // XXX
        Note.warning(this, "Operation not implemented.");
        return this;
    }
    
    public CDT getNewInstance() {
        // XXX
        Note.warning(this, "Operation not implemented.");
        return null;
    }
    
    public Object search(Object o) {
        Note.warning(this, "Operation not implemented.");
        return null;
    }
    
    // ** Implementation of interface Array **

    public int getFirst() {
        return 0;
    }

    public int getLast() {
        return 0;
    }

    public Object getObject(int address) {
        if (address == 0) {
            return this.ADTPeek();
        } else {
            return null;
        }
    }

    public void setFirst(int i) {
        Note.warning(this, "Operation not implemented.");
    }

    public void setLast(int i) {
        Note.warning(this, "Operation not implemented.");
    }

    public void setObject(Object o, int address) {
        if (address == 0) {
            if ((o == null) ||
                ((o instanceof Key) && (((Key)o).getObject() == null))) {
                this.ADTPop();
            } else {
                this.ADTPop();
                this.ADTPush(o);
            }
        }
    }
    
    
    // ** Implementation of interface Stack **
    
    public void ADTPush(Object o) {
        if (this.topDecorator != null) {
            if (this.ADTPeek() instanceof SpatialElement) {
                ((SpatialElement)this.ADTPeek()).setPaintingStyleDecorator((PaintingStyleDecorator)this.savedDecorator);
            }
            if (o instanceof SpatialElement) {
                this.savedDecorator = ((SpatialElement)o).getPaintingStyleDecorator();
                ((SpatialElement)o).setPaintingStyleDecorator(this.topDecorator);
            }
        }
        this.contents.setObject(o, this.stackPointer());
        this.stackPointer.inc();
    }
    
    public Object ADTPeek() {
        return this.contents.getObject(this.stackPointer() - 1);
    }

    public Object ADTPop() {
        if (!this.ADTIsEmpty()) {
            Object top = this.ADTPeek();
            if (this.topDecorator != null) {
                if (top instanceof SpatialElement) {
                    ((SpatialElement)top).setPaintingStyleDecorator((PaintingStyleDecorator)this.savedDecorator);
                }
            }
            this.stackPointer.dec();
            this.contents.setObject(null, this.stackPointer());
            this.contents.setLast(this.stackPointer() - 1);
            if (this.topDecorator != null) {
                if ((!this.ADTIsEmpty()) && 
                    (this.ADTPeek() instanceof SpatialElement)) {
                    this.savedDecorator = ((SpatialElement)this.ADTPeek()).getPaintingStyleDecorator();
                    ((SpatialElement)this.ADTPeek()).setPaintingStyleDecorator(this.topDecorator);
                }
            }
            return top;
        } else {
            return null;
        }
    }
    
    public boolean ADTIsEmpty() {
        return (this.stackPointer() == 0);
    }
    
    // ** Implementation of interface Area **

    public void addSubArea(Area a) {
        if (a != null) {
            this.subAreas.add(a);
        }
    }

    public void removeSubArea(Area a) {
        // TODO replace subarea structure with one whose states are stored in the animator
//        if (a != null) {
//            this.subAreas.remove(a);
//        }
    }

    public Area[] getSubAreas() {
        Vector vec = new Vector();
        for (int i = this.contents.getFirst(); i <= this.contents.getLast(); i++) {
            if ((this.contents.getObject(i) != null) && 
                ((this.contents.getObject(i) instanceof Area) && 
                 (!(this.contents.getObject(i) instanceof SpatialComparable)))) {
                vec.add(this.contents.getObject(i));
            }
        }
        Iterator iterator = this.subAreas.iterator();
        while (iterator.hasNext()) {
            vec.add(iterator.next());
        }
        Area[] areas = new Area[vec.size()];
        vec.toArray(areas);
        return areas;
    }

    public void delete(SpatialComparable e) {
        // XXX
        Note.warning(this, "Operation not implemented.");
    }

    public void insert(SpatialComparable e) {
        this.ADTPush(e);
    }

    public Rectangle getBoundingPolygon() {
        return this.boundingBox;
    }

    public SpatialElement[] getSpatialElements() {
        Vector vec = new Vector();
        for (int i = this.contents.getFirst(); i <= this.contents.getLast(); i++) {
            if ((this.contents.getObject(i) != null) && 
                (this.contents.getObject(i) instanceof SpatialComparable)) {
                vec.add(this.contents.getObject(i));
            }
        }
        SpatialElement[] spatialElements = new SpatialElement[vec.size()];
        vec.toArray(spatialElements);
        return spatialElements;
    }
    
    public boolean hasPaintingStyleDecorator(){
        return (paintingStyleDecorator.getObject() != null);
    }

    public void setPaintingStyleDecorator(PaintingStyleDecorator psd){
        paintingStyleDecorator.setObject(psd);
    }

    public PaintingStyleDecorator getPaintingStyleDecorator(){
        return (PaintingStyleDecorator)paintingStyleDecorator.getObject();
    }

    
    // ** Other methods **
    
    public void clear() {
        while (!this.ADTIsEmpty()) {
            this.ADTPop();
        }
    }
    
    private int stackPointer() {
        return this.stackPointer.eval();
    }
    
    /**
     * @return An array visualizing the contents of this stack.
     */
    public Array getArrayVisualization() {
        return new ImmutableArray(this.contents);
    }
    
    public boolean equals(Object obj) {
        if (obj == this) {
            return true;
        } else {
            if (obj instanceof SimpleStack) {
                SimpleStack stack = (SimpleStack)obj;
                if ((this.stackPointer() == stack.stackPointer()) &&
                    (this.contents.equals(stack.contents))) {
                    return true;
                } else {
                    return false;
                }
            } else {
                return false;
            }
        }
    }
    
    public String toString() {
        if (this.ADTIsEmpty()) {
            return "[ EMPTY ]";
        } else {
            String str = "[ ";
            for (int i = 0; i < this.stackPointer(); i++) {
                str += " " + this.contents.getObject(i);
            }
            str += " TOP ]";
            return str;
        }
    }
    
    /**
     * FDT-wrapper for a VirtualArray.
     */
    private static class SimpleArray extends VirtualArray implements FDT, Array {
        
        public SimpleArray() {
            super(null, "contents");
        }

        public Object getElement() {
            return null;
        }

        public void setElement(Object o) {
        }

        private static final long serialVersionUID = -2292778504768648701L;
    }
    
    /**
     * A wrapper class for transforming a mutable array into 
     * an immutable one.
     */
    private static final class ImmutableArray implements Array {

        private final Array array;
        
        public ImmutableArray(Array array) {
            this.array = array;
        }
        
        public Object getElement() {
            return this.array.getElement();
        }

        public void setElement(Object o) {
            // disabled
        }

        public int getFirst() {
            return this.array.getFirst();
        }

        public int getLast() {
            return this.array.getLast();
        }

        public Object getObject(int address) {
            return this.array.getObject(address);
        }

        public void setFirst(int i) {
            // disabled
        }

        public void setLast(int i) {
            // disabled
        }

        public void setObject(Object o, int address) {
            // disabled
        }

        private static final long serialVersionUID = -6709777536081678052L;
    }
    
    private static final long serialVersionUID = -5043763957501793067L;
}

 
