/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: ClickableCommonTrie.java,v 1.11 2006/12/12 14:21:26 oseppala Exp $ */

package content.exercises.structures;

import java.awt.Color;

import matrix.structures.FDT.*;
import matrix.structures.FDT.probe.Key;
import matrix.structures.memory.*;
import matrix.util.RandomKey;
import matrix.util.Note;
import matrix.animation.Animator;
import matrix.decoration.LabelDecorator;
import matrix.decoration.StyleSheet;
import matrix.decoration.StyleSheetAdapter;
import matrix.decoration.Styled;

/** 
 * A common trie which is used in CommonTrie_Insert exercise.
 * Keys in the nodes of this ClickableCommonTrie can be clicked. The structure
 * should be used only by clicking keys in its nodes (and not dragging or dropping).
 * Subtrees of ClickableCommonTrie can be other ClickableCommonTries or EmptyTrieNode.
 * This depends on whether the corresponding subtree node has already been expanded or not 
 * (by clicking the character in the subtree node).
 */
public class ClickableCommonTrie implements FDT, Tree, LabelDecorator {


    /**
     * Contains the content of this trie node.
     * Possible values are keys "true" and "false"
     */
    private VirtualObject content;
    
    /** 
     * Children contains the subtrees. The subtrees corresponds to characters
     * a,b,c,d... etc.
     * */
    private VirtualArray children;

    /**
     * Number of characters (that is, subtrees in one node)
     */
    private int numberOfChars = 0;
    
    /** 
     * The constructor for an empty trie.
     * This constructs a trie whose every node has three children (a,b,c).
     */
    public ClickableCommonTrie() {
        numberOfChars = 3;
        content = new VirtualObject(this, "Content of the trie node");

        //actual value of the node is false
        content.setObject(new ClickableBooleanKey("false"));

        children = new VirtualArray(this, "Children of the trie node");
        children.setLast(numberOfChars - 1); //by default three children

        //characters are a,b,c
        for(int i = 0; i < children.size(); i++) {
            children.setObject(new EmptyTrieNode(RandomKey.keyIt(i + 3).charAt(0)), i);
        }
    
    }

    /** 
     * The constructor for a new root node.
     * 
     * @param numOfChars Number of characters (and number of subtrees).
     */
    public ClickableCommonTrie(int numOfChars) {
        
        if (numOfChars > 26) {
            Note.out(this, "Maximum number of different characters in ClickableCommonTrie is 26. Using 26.");
            numberOfChars = 26;
        } else if (numOfChars < 2) {
            Note.out(this, "Minimum number of different characters in ClickableCommonTrie is 2. Using 2.");
            numberOfChars = 2;
        } else {
            numberOfChars = numOfChars;
        }

        content = new VirtualObject(this, "Content of the trie node");

        //actual value of the node is false
        content.setObject(new ClickableBooleanKey("false"));

        children = new VirtualArray(this, "Children of the trie node");
        children.setLast(numberOfChars - 1);

        //characters are a,b,c,d,...
        for(int i = 0; i < children.size(); i++) {
            children.setObject(new EmptyTrieNode(RandomKey.keyIt(i + 3).charAt(0)), i);
        }

    }

    // Methods for interface FDT

    /** Return the element value. */
    public Object getElement() {
        return content.getObject();
    }

    /** Set the element value. */
    public void setElement(Object v) {
        content.setObject(v);
    }

    // Methods for interface Tree

    /**
     * Returns the subtree count 
     */
    public int getSubTreeCount() {
        return children.size();
    }

    /**
     * Returns the subtrees.
     */
    public Tree[] getSubTrees() {
        Tree[] t = new Tree[getSubTreeCount()];
        for (int i=0; i<t.length; i++)
            t[i] = (Tree)children.getObject(i);
        return t;
    }

    // Methods for interface LabelDecorator 
    
    //label is not needed at all
    
    public boolean isLabelEnabled() {
        return false;
    }

    public void setLabelEnabled(boolean b) {}

    public void setLabel(String s) {}

    public String getLabel() { return ""; }

    //reference label is needed
    
    private boolean isReferenceLabelEnabled = true;
    
    public boolean isReferenceLabelEnabled() {
        return isReferenceLabelEnabled;
    }

    public void setReferenceLabelEnabled(boolean b) {
        isReferenceLabelEnabled = b;
    }

    public void setReferenceLabel(int i, String s) {
        //do nothing, because reference labels should not be changed
    }

    //i = 0 returns a, i=1 returns b, and so on
    public String getReferenceLabel(int i) {
        return RandomKey.keyIt(i + 3);
    }
    
    // end of interfaces

    /**
     * Toggles the boolean value in trie node.
     */
    public void toggleBoolean() {
        Key tempKey = (Key)content.getObject();
        String keyValue = tempKey.getVisualizationString();
        if ("true".equals(keyValue)) {
            content.setObject(new ClickableBooleanKey("false"));
        } else if ("false".equals(keyValue)) {
            content.setObject(new ClickableBooleanKey("true"));
        } else {
            Note.out(this, "Value of the key is not false or true.");
        }
    }
    

    /**
     * Sets the specified subtree from EmptyTrieNode to ClickableCommonTrie.
     * Parameter tells which subtree should be changed (a = 0, b = 1, c = 2,..).
     * @param characterAsInt tells which subtree should be changed (a = 0, b = 1,...)
     */
    public void setSubTreeChar(int characterAsInt) {

        int position = characterAsInt - 'a';
        
        if (children.getObject(position) instanceof ClickableCommonTrie) {
            //subtree is already created, do nothing
            return;
        } else {
            /* Subtree was EmptyTrieNode so change it to ClickableCommonTrie 
             * (thus expand this node).
            */
            children.setObject(new ClickableCommonTrie(numberOfChars), position);
        }
    }
    
    /**
     * Returns the specified subtree.
     * Parameter is a character a,b,c..
     * Characters maps to subtrees as follows:
     * a = 0, b = 1, c = 2,...
     * @param characterAsInt tells which subtree to return (a = 0, b = 1,...)
     * @return the sub tree that is specified as a character in a parameter
     */
    public Tree getSubTree(int characterAsInt) {
        int position = characterAsInt - 'a';
        return (Tree)children.getObject(position);
    }
    

    public String toString() {
        String tempString = getClass().getName() + "["; 
        tempString += content.getObject(); 
        tempString += "], Children[";

        for(int i = 0; i < children.size(); i++) {
            tempString += children.getObject(i);
            if (i < children.size() - 1) {
                tempString += ",";
            }
        }
        tempString += "]";
        
        return tempString;
    }

    /* StyleSheetAdapters for ClickableBooleanKey */    
    private static final StyleSheetAdapter trueStyle = new StyleSheetAdapter();
    private static final StyleSheetAdapter falseStyle = new StyleSheetAdapter();

    static{
        trueStyle.setDefaultBackgroundColor(Color.cyan);
        trueStyle.setVisitedBackgroundColor(Color.cyan);        
    }
    /**
     * This class is used as key in ClickableCommonTrie nodes.
     * It can be clicked which cause toggleBoolean method of 
     * ClickableCommonTrie to be called.
     */    
    private class ClickableBooleanKey extends Key implements Styled {
                


        /**
         * Constructor
         * @param s key as a string
         */
        public ClickableBooleanKey(String s) {
            super(s);
        }
        
        /**
         * This method is called through Matrix Framework when
         * user clicks this key. Clicking causes the toggleBoolean
         * method in ClickableCommonTrie to be called.
         */
        public void select() {
            Animator.getActiveAnimator().startOperation();
            toggleBoolean();
            Animator.getActiveAnimator().endOperation();
        }

      
        /**
         * Returns stylesheet which changes DefaultBackgroundColor and VisitedBackgroundColor
         * for keys that are "true" to Color.CYAN so that "true" nodes are more visible.
         */
        public StyleSheet getStyleSheet() {

            if (getVisualizationString().equals("true")) 
                return trueStyle;                    
            return falseStyle;
                
        }
        static final long serialVersionUID = 4337983200735279917L;
        
    }

    /**
     * Objects of this class are used as empty sub trees in
     * ClickableCommonTrie (thus before sub trees are clicked.)
     * It contains a ClickableKey which calls setSubTree method of
     * Empty Trie Node's parent (that is, ClickableCommonTrie) when clicked.  
     */
    private class EmptyTrieNode implements FDT, Tree {

        private Object content;
        private char name; //name of this empty trie node can be a,b,c,..
        
        /**
         * Constructor
         * @param name name of this empty trie node (a,b,c,d...)
         */
        public EmptyTrieNode(char name) {
            this.name = name;
            content =  new ClickableKey("\u2295");//plus operator inside a circle
        }
        
        /**
         * Returns the element in this node.
         */
        public Object getElement() {
            return content;
        }

        /**
         * Sets the element in this node.
         */
        public void setElement(Object o) {
            content = o;
        }

        /**
         * Returns the subtree count of this node.
         * It is always zero, because this represents leaf node 
         * of the trie.
         */
        public int getSubTreeCount() {
            return 0;
        }

        /**
         * Returns always null because there are no sub trees
         * in leaf node.
         */
        public Tree[] getSubTrees() {
            return null;
        }
     
        /**
         * Equals method of EmptyTrieNode.
         */
        public boolean equals(Object o) {
            if (!(o instanceof EmptyTrieNode)) {
                return false;
            } else {
                if (getElement() == null && ((EmptyTrieNode)o).getElement() != null) {
                    return false;
                } else if (getElement() == null && ((EmptyTrieNode)o).getElement() == null) {
                    return true;
                }
                return getElement().equals(((EmptyTrieNode)o).getElement());
            }
         
        }
        
        /**
         * This class is used as key in EmptyTrieNode nodes.
         * It can be clicked which cause setSubTree method of 
         * ClickableCommonTrie to be called.
         */
        private class ClickableKey extends Key {
            

            /**
             * Constructor
             * @param s key as a string
             */
            public ClickableKey(String s) {
                super(s);
             }
            
            /**
             * This method is called through Matrix Framework when
             * user clicks this key. Clicking causes the setSubTree
             * method in ClickableCommonTrie to be called.
             */
            public void select() {
                Animator.getActiveAnimator().startOperation();
                setSubTreeChar((int)name);
                Animator.getActiveAnimator().endOperation();
            }

            static final long serialVersionUID = -7564560916655787873L;
        }
        static final long serialVersionUID = 176929799319292907L;

        public void setSubTree(Tree t, int child) {
            // TODO Auto-generated method stub
            
        }

        public Tree getNewNode(Object key) {
            // TODO Auto-generated method stub
            return null;
        }
    }

    /**
     * Equals method of ClickableCommonTrie.
     */
    public boolean equals(Object o) {

        if (!(o instanceof ClickableCommonTrie)) {
            return false;
        } else {
            ClickableCommonTrie trie = (ClickableCommonTrie)o;

            if (getElement() == null && trie.getElement() != null) {
                return false;
            } else if (getElement() != null && !(getElement().equals(trie.getElement()))) {
                return false;
            }

            if (getSubTreeCount() != trie.getSubTreeCount()) {
                return false;
            }
            
            for (int i = 0; i < children.size(); i++) {
                if (children.getObject(i) == null && trie.getSubTrees()[i] != null) {
                    return false;
                } else if (children.getObject(i) != null && 
                           !(children.getObject(i).equals(trie.getSubTrees()[i]))) {
                    return false;
                }
            }
            
            return true;
        }
    }

    static final long serialVersionUID = 6137902580372630315L;

    public void setSubTree(Tree t, int child) {
        // TODO Auto-generated method stub
        
    }

    public Tree getNewNode(Object key) {
        // TODO Auto-generated method stub
        return null;
    }
}
