/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: ExerBTree.java,v 1.9 2006/12/12 14:20:52 oseppala Exp $ */

package content.exercises.structures;

import matrix.structures.CDT.CDT;
import matrix.structures.FDT.*;
import matrix.structures.FDT.probe.*;
import matrix.structures.memory.*;
import matrix.structures.other.Comparer;
import matrix.structures.util.MatrixComparable;
import matrix.util.Note;

public class ExerBTree implements CDT {
    
    private VirtualObject store;
    private BT root;
    
    public ExerBTree() {
        this(2);
    }
    
    public CDT getNewInstance(){
        return new ExerBTree();
    }
    
    public ExerBTree(int i) {
        if(i < 2)
            i = 2;
        store = new VirtualObject(this, "Root node of this BTree");
        root = new BT(i);
        store.setObject(root);
    }
    
    // Interface FDT
    
    public Object getElement() {
        return getRootObject();
    }
    
    public void setElement(Object o) {
        setRootObject(o);
    }
    
    //interface Tree
    
    public int getSubTreeCount() {
        root = getRootObject();
        if(root == null)
            return 0;
        return root.getSubTreeCount();
    }
    
    public Tree[] getSubTrees() {
        root = getRootObject();
        return root.getSubTrees();
    }
    
    //interface SimulationTree
    
    public void setSubTree(Tree t, int child) {
        root = getRootObject();
        root.setSubTree(t, child);
    }
    
    
    public Tree getNewNode(Object key) {
        root = getRootObject();
        return root.getNewNode(key);
    }
    
    //Interface CDT
    
    public CDT insert(Object o) {
        root = getRootObject();
        FDT tmp = root.insert(o);
        if(tmp != root)
            setRootObject(tmp);
        return this;
    }
    
    public CDT delete(Object o) {
        root = getRootObject();
        FDT tmp = root.delete(o);
        if(tmp != root) {
            setRootObject(tmp);
        }
        return this;
    }
    
    public Object search(Object o) {
        return this;
    }
    
    
    public boolean equals(Object o) {
        if(o instanceof ExerBTree) {
            return Comparer.compareIdentical(getRootObject(), ((ExerBTree)o).getRootObject());
        }
        return false;
    }
    
    private BT getRootObject() {
        return (BT) store.getObject();
    }
    
    private void setRootObject(Object o) {
        if(o instanceof BT) {
            store.setObject(o);
            root = getRootObject();
        }
    }
    
    protected class BT implements FDT, Tree, CommonTree {
        
        /**
         * The minimum degree for this B-Tree. If t is the minimum degree
         * of a B-Tree, each node, other than root, can have a minimum of
         * t-1 and maximum of 2t-1 keys. Root may have a minimum of
         * 1 and maximum of 2t-1 keys.
         */
        private int minimum;
        
        /** tells how many keys this node has */
        private VirtualInteger nroKeys;
        
        private VirtualInteger nroChildren;
        
        /** Stores the keys in this node */
        private Table contents;
        
        /** Stores the children of this node */
        private VirtualArray children;
        
        /** Constructs a new B-Tree with a minimum degree of 2 */
        public BT() {
            minimum = 2;
            nroKeys = new VirtualInteger(0, this, "Number of keys contained in the node");
            contents = new Table ("");
            contents.setObject(null, 0);
            contents.setObject(null, 2 * minimum - 2);
            children = new VirtualArray(this, "Children of this BTree node");
        }
        
        /**
         * Constructs a new B-Tree
         * @param i the minimum degree of this tree
         */
        public BT(int n) {
            this();
            minimum = n;
            for(int i = 0; i < 2 * minimum - 1; i++)
                contents.setObject(null, i);
        }
        
        public BT(Object o, int i) {
            this();
            minimum = i;
            for(int n = 0; i < 2 * minimum - 1; n++)
                contents.setObject(null, n);
            insertKey(o, 0);
        }
        
        public void split() {
            BT root = this;
            //Note.out(this, "SPLIT " + root);
            if (root == getRootObject()) { 
                root.getKeyCount(); //this call is needed because getKeyCount updates the value of nroKeys 
                //Note.out(this, "split root" + root.getKeyCount());
                BT newRoot = new BT(minimum);
                splitChild(newRoot, 0, root);
                if(newRoot != root)
                    setRootObject(newRoot);
            } else {
                root.getKeyCount(); //this call is needed because getKeyCount updates the value of nroKeys 
                //Note.out(this, "split else " + root.getKeyCount());
                splitChild(this);
            }
        }
        
        private void splitChild(BT child) {
            BT parent;
            int index;
            BT root = getRootObject();
            parent = getParent(root, child);
            for (index = 0; ;index++) {
                if (parent.getChild(index) == child)
                    break;
            }
            
            splitChild(parent, index, child);
        }
        
        private BT getParent(BT tree, BT child) {
            if (tree == null)
                return null;
            for (int i = 0; i < tree.getChildCount(); i++) {
                if (tree.getChild(i) == child)
                    return tree;
            }
            
            for (int i = 0; i < tree.getChildCount(); i++) {
                BT temp = getParent(tree.getChild(i), child);
                if (temp != null)
                    return temp;
            }
            return null;
        }
        
        /**
         * Retrieves the i'th key of this node
         * @param i the number i
         * @return the key
         */
        public Object getKey(int i) {
            return contents.getObject(i);
        }
        
        /**
         * Tells the number of keys this node has
         * @return the number of keys
         */
        public int getKeyCount() {
            int temp1 = nroKeys.eval();
            int temp2 = 0;
            
            for (int i = 0; i < contents.size(); i++) {
                if (contents.getObject(i) != null)
                    temp2 = i + 1;
            }
            if (temp2 > temp1)
                nroKeys.assign(temp2); //updates value of nroKeys
            return temp2;
        }
        
        /**
         * Tells if a node is leaf or not
         * @return true if node is leaf, false otherwise
         */
        public boolean isLeaf() {
            if(getChild(0) == null)
                return true;
            return false;
        }
        
        /**
         * Searches the node for a key matching the argument object. The
         * search tests for the equality of references.
         * @param o the object to be found
         * @return the index position of the object, if found, -1 otherwise
         */
        public int findKey(Object o) {
            for(int i = 0; i < nroKeys.eval(); i++) {
                if(!gt(o, getKey(i))) {
                    if(o == getKey(i))
                        return i;
                    else
                        return -1;
                }
            }
            return -1;
        }
        
        /**
         * Inserts the param object as the i'th key of this node.
         * The larger keys are pushed one step further
         * @param o the value of the new key
         * @param i the number i
         */
        public void insertKey(Object o, int i) {
            //	if(i > nroKeys.eval())
            //    Note.err(this, "insertKey: illegal value of i " + i);
            if(i == nroKeys.eval())
                setKey(o, i);
            else {
                Object tmp1 = null;
                Object tmp2 = getKey(i);
                for(int j = i; j < nroKeys.eval(); j++) {
                    tmp1 = tmp2;
                    tmp2 = getKey(j + 1);
                    setKey(tmp1, j + 1);
                }
                setKey(o, i);
            }
            nroKeys.inc();
        }
        
        // Sets the value of i'th key to o
        private void setKey(Object o, int i) {
            contents.setObject(o, i);
        }
        
        
        private void deleteKey(int i) {
            getKeyCount(); //this updates the value of nroKeys;
            Object tmp1 = null;
            Object tmp2 = null;
            for(int j = nroKeys.eval(); j > i; j--) {
                tmp1 = tmp2;
                tmp2 = getKey(j - 1);
                setKey(tmp1, j - 1);
            }
            nroKeys.dec();
        }
        
        /**
         * Retrives the i'th child of this node
         * @param i the number i
         * @return the child
         */
        public BT getChild(int i) {
            return (BT) children.getObject(i);
        }
        
        //Inserts the argument tree as the i'th child of this node
        
        public void insertChild(BT bt, int i) {
            if(i == nroKeys.eval())
                setChild(bt, i);
            else {
                BT tmp1 = null;
                BT tmp2 = getChild(i);
                for(int j = i; j < nroKeys.eval() + 1; j++) {
                    tmp1 = tmp2;
                    tmp2 = getChild(j + 1);
                    setChild(tmp1, j + 1);
                }
                setChild(bt, i);
            }
        }
        
        // Deletes the i'th child of this node and shifts other children appropriately
        private void deleteChild(int i) {
            BT tmp1 = null;
            BT tmp2 = null;
            for(int j = nroKeys.eval(); j > i - 1; j--) {
                tmp1 = tmp2;
                tmp2 = getChild(j);
                setChild(tmp1, j);
            }
            children.setLast(children.getLast() - 1);
        }
        
        
        public void setChild(BT bt, int i) {
            children.setObject(bt, i);
        }
        
        public int getChildCount() {
            return children.size();
        }
        
        // finds the position where the argument key would be in
        // node. This also finds the correct child where the key could be
        int findPos(Object o) {
            for(int i = 0; i < nroKeys.eval(); i++) {
                if(!gt(o, getKey(i)))
                    return i;
            }
            return nroKeys.eval();
        }
        
        
        // Tells if the (String) value of object a is greater than the value
        // of object b. All values are considered to be less than null
        
        private boolean gt(Object a, Object b) {
            if(b == null)
                return false;
            String as = a.toString();
            String bs = b.toString();
            int i = as.compareTo(bs);
            if(i <= 0)
                return false;
            return true;
        }
        
        // methods for interface FDT
        
        public Object getElement() {
            return contents;
        }
        
        public void setElement(Object o) {
            if(o instanceof Table && ((Table)o).size() == 2 * minimum - 1)
                contents = (Table)o;
            else
                Note.err(this, "Cannot insert " + o + " ,not a Table.");
        }
        
        //methods for interface Tree
        
        public int getSubTreeCount() {
            if(isLeaf())
                return 0;
            else
                return 2 * minimum - 1;
        }
        
        public Tree[] getSubTrees() {
            if (isLeaf())
                return new Tree[0];
            Tree[] t = new Tree[nroKeys.eval() + 1];
            for(int i = 0; i < nroKeys.eval() + 1; i++)
                t[i] = getChild(i);
            return t;
        }
        
        
        //methods for interface SimulationTree
        
        public void setSubTree(Tree t, int child) {
            if(t instanceof BT && child < nroKeys.eval() + 1)
                setChild((BT)t, child);
            else
                Note.err(this, "setSubTree: illegal arguments: " + t + " " + child);
        }
        
        public Tree getNewNode(Object key) {
            return new BT(key, minimum);
        }
        
        //methods for interface CDT
        
        public FDT insert(Object o) {
            if(o == null)
                return this;
            if(o instanceof Table) {
                Table t = (Table)o;
                BT root = this;
                for(int i = 0; i < t.size(); i++)
                    root = (BT) root.insert(t.getObject(i));
                return root;
            }
            MatrixComparable k;
            if(o instanceof MatrixComparable)
                k = (MatrixComparable)o;
            else
                k = new Key(o);
            if(getKeyCount() == 2 * minimum - 1) {
                BT newRoot = new BT(minimum);
                splitChild(newRoot, 0, this);
                insertHelp(newRoot, k);
                return newRoot;
            } else {
                insertHelp(this, k);
                return this;
            }
        }
        
        private void insertHelp(BT tree, Object o) {
            if(tree.isLeaf()) {
                Note.out(tree, "leaf found.");
                if(tree.getKeyCount() == minimum * 2 - 1)
                    Note.err(tree, "insertHelp encountered a full leaf node.");
                else {
                    int i = 0;
                    while(gt(o, tree.getKey(i)))
                        i++;
                    tree.insertKey(o, i);
                }
            } else {
                int i = 0;
                while(gt(o, tree.getKey(i)))
                    i++;
                if(tree.getChild(i).getKeyCount() == 2 * minimum - 1) {
                    splitChild(tree, i, tree.getChild(i));
                    if(gt(o, tree.getKey(i)))
                        insertHelp(tree.getChild(i + 1), o);
                    else
                        insertHelp(tree.getChild(i), o);
                }
                else
                    insertHelp(tree.getChild(i), o);
            }
        }
        
        
        // splits the full child of a parent tree. Parent is not full.
        // childNro is the index number of the child to be split.
        
        private void splitChild(BT parent, int childNro, BT child) {
            BT newChild = new BT(minimum);
            if(!child.isLeaf()) {
                for(int i = 0; i < minimum; i++) {
                    newChild.insertChild(child.getChild(minimum), i);
                    child.deleteChild(minimum);
                }
            }
            for(int i = 0; i < minimum - 1; i++) {
                newChild.insertKey(child.getKey(minimum), i);
                child.deleteKey(minimum);
            }
            parent.insertKey(child.getKey(minimum - 1), childNro);
            child.deleteKey(minimum - 1);
            parent.setChild(child, childNro);
            parent.insertChild(newChild, childNro + 1);
        }
        
        public FDT delete(Object o) {
            if(o instanceof MatrixComparable) {
                deleteHelp(this, o);
                if(getKeyCount() == 0) {
                    BT newRoot = getChild(0);
                    deleteChild(0);
                    return newRoot;
                }
            }
            else
                Note.out(this, "Cannot delete, not an Element");
            return this;
            
        }
        
        private void deleteHelp(BT tree, Object o) {
            int pos = tree.findKey(o);
            
            // if node is leaf, do simple delete
            
            if(tree.isLeaf()) {
                if(pos != -1)
                    tree.deleteKey(pos);
                else
                    Note.err(tree, "Delete: key not found.");
            } else {
                
                // if key is found in internal node, replace it with it
                // predecessor or successor, or if both immediate children have
                // minimum number of keys, merge them and delete the key recursively
                
                if(pos != -1) {
                    Object newKey;
                    if(tree.getChild(pos).getKeyCount() > minimum - 1) {
                        newKey = deleteLargest(tree.getChild(pos));
                        tree.setKey(newKey, pos);
                    } else if(tree.getChild(pos + 1).getKeyCount() > minimum - 1) {
                        newKey = deleteSmallest(tree.getChild(pos + 1));
                        tree.setKey(newKey, pos);
                    } else {
                        mergeChild(tree, pos);
                        deleteHelp(tree.getChild(pos), o);
                    }
                } else {
                    pos = tree.findPos(o);
                    deleteCheck(tree, pos);
                    if(pos <= tree.getKeyCount())
                        deleteHelp(tree.getChild(pos), o);
                    else
                        deleteHelp(tree.getChild(pos - 1), o);
                }
            }
        }
        
        
        
        
        // deletest the largest key in the argument tree and returns it
        Object deleteLargest(BT tree) {
            if(tree.isLeaf()) {
                Object tmp = tree.getKey(tree.getKeyCount() - 1);
                tree.deleteKey(tree.getKeyCount() - 1);
                return tmp;
            } else {
                deleteCheck(tree, tree.getKeyCount());
                return deleteLargest(tree.getChild(tree.getKeyCount()));
            }
        }
        
        // deletes the smallest key in the argument tree, and returns it
        Object deleteSmallest(BT tree) {
            if(tree.isLeaf()) {
                Object tmp = tree.getKey(0);
                tree.deleteKey(0);
                return tmp;
            } else {
                deleteCheck(tree, 0);
                return deleteSmallest(tree.getChild(0));
            }
        }
        
        public Object search() { return this; }
        
        
        
        // parent is the parent node, childNro is the number of the left child
        // of the two children to be merged.
        private void mergeChild(BT parent, int childNro) {
            BT left = parent.getChild(childNro);
            BT right = parent.getChild(childNro + 1);
            if(right == null || left == null)
                Note.err(this, "mergechild called for null child");
            else if(left.getKeyCount() != minimum - 1
                    || right.getKeyCount() != minimum - 1)
                Note.err(this, "mergechild called for child with illegal number of keys");
            else
                if(!left.isLeaf()) {
                    BT tmp1 = null;
                    BT tmp2 = right.getChild(0);
                    for(int i = 0; i < minimum; i++) {
                        tmp1 = tmp2;
                        tmp2 = right.getChild(i + 1);
                        left.insertChild(tmp1, minimum + i);
                    }
                }
            Object tmp1 = null;
            Object tmp2 = right.getKey(0);
            left.insertKey(parent.getKey(childNro), minimum - 1);
            for(int i = 0; i < minimum - 1; i++) {
                tmp1 = tmp2;
                tmp2 = right.getKey(i + 1);
                left.insertKey(tmp1, minimum + i);
            }
            parent.deleteChild(childNro + 1);
            parent.deleteKey(childNro);
        }
        
        
        
        // Checks the child where delete is going to be called recursively.
        // if the child has the minimum number of keys, it will either
        // move a key fro mthe right or left sibling to parent, and a key from parent
        // to the child (adjusting the parent's grandchildren appropriately)
        // or mergest the child with its right sibling.
        // the argument tree is the node to be checked, the int is the number of the
        // child recursion will descend to.
        
        void deleteCheck(BT tree, int pos) {
            
            // mc stands for middle child, it is the child node where recursion
            // will descend. lc is mc's left sibling. lc is mc's right sibling.
            
            BT lc, mc, rc;
            mc = tree.getChild(pos);
            if(pos != 0)
                lc = tree.getChild(pos - 1);
            else lc = null;
            if(pos != tree.getKeyCount())
                rc = tree.getChild(pos + 1);
            else rc = null;
            
            // if the child where recursion is to descend has minimum number of keys,
            // either move a key from the left or right sibling to parent, and a key
            // from parent to the child, or merge the child with its (right) sibling.
            
            if(mc.getKeyCount() == minimum - 1) {
                if(lc != null && lc.getKeyCount() > minimum - 1) {
                    if(!lc.isLeaf()) {
                        mc.insertChild(lc.getChild(lc.getKeyCount()), 0);
                        lc.deleteChild(lc.getKeyCount());
                    }
                    mc.insertKey(tree.getKey(pos - 1), 0);
                    tree.setKey(lc.getKey(lc.getKeyCount() - 1), pos - 1);
                    lc.deleteKey(lc.getKeyCount() - 1);
                } else if(rc != null && rc.getKeyCount() > minimum - 1) {
                    if(!rc.isLeaf()) {
                        mc.insertChild(rc.getChild(0), minimum);
                        rc.deleteChild(0);
                    }
                    mc.insertKey(tree.getKey(pos), mc.getKeyCount());
                    tree.setKey(rc.getKey(0), pos);
                    rc.deleteKey(0);
                } else {
                    if(pos < tree.getKeyCount())
                        mergeChild(tree, pos);
                    else
                        mergeChild(tree, pos - 1);
                }
            }
        }
        static final long serialVersionUID = 8167655877657039950L;
    }
    
    static final long serialVersionUID = -5436258313505218101L;
}
