package content.exercises.structures;

import java.awt.Color;

import matrix.structures.FDT.probe.Key;
import matrix.structures.FDT.probe.UndirectedGraphImpl;
import matrix.structures.FDT.substructures.Vertex;
import matrix.structures.memory.VirtualObject;

/**
 * This class is used in Dijsktra_Heap exercise. This graph is weighted and the actual distance of
 * the node from the start node is shown in the label of the corresponding node.
 *
 * @author kstalnac
 *
 */
public class ExerLabeledGraph_DH extends UndirectedGraphImpl {

    /** heap used in the exercise */
    private ExerLabeledHeap_DH heap;
    /** lastChosen tells which node was lastly selected to the spanning tree (and was colored to
     * black) */
    private VirtualObject lastChosen = new VirtualObject(null, this, "Last chosen node in the spanning tree");
    
    /**
     * Constructor
     */
    public ExerLabeledGraph_DH() {
        super();
    }

    /**
     * Sets the reference of the heap.
     * @param heap heap used in the exercise
     */
    public void setHeap(ExerLabeledHeap_DH heap) {
        this.heap = heap;
    }
    
    /**
     * Switch the labels visible for the graph nodes. (These could be weights etc.))
     */
    public void setReferenceLabelEnabled(boolean labeled) {
        super.setReferenceLabelEnabled(labeled);
        Vertex[] vertices = getVertices();
        for (int i = 0; i < vertices.length; i++) {
            ((PriorityExerVertex)vertices[i]).setReferenceLabelEnabled(isReferenceLabelEnabled());
        }
    }
    
    /**
     * Returns the distance of the node corresponding to the given key from the start node.
     * If lastChosen is null then distance is 0, otherwise the distance is lastChosen's 
     * distance + weight of the edge between lastChosen and the node corresponding to key.
     * If the node corresponding the key is not successor of lastChosen, then -1 is returned.
     * @param key key which distance is wanted
     * @return distance of the key (its corresponding node) from the start node
     */
    public int getDistance(Key key) {
        if (lastChosen.getObject() == null)
            return 0;
        else {
            PriorityExerVertex_DH last = (PriorityExerVertex_DH)lastChosen.getObject();
            PriorityExerVertex_DH succ = last.getSuccessor(key);
            if (succ == null)
                return Integer.MAX_VALUE;
            else {
                return last.getDistanceOfNode() + last.getWeight(succ);
            }
        }
    }
    
    /**
     * Sets the node corresponding to the key to the lastChosen variable.
     * Sets also the distance of the corresponding node to the distance parameter.
     * This is called, when a key is inserted to traverse order list (queue).
     */
    public void setLastChosen(Key key, int distance) {
        Vertex[] vertices = getVertices();
        for (int i = 0; i < vertices.length; i++) {
            PriorityExerVertex_DH pev = (PriorityExerVertex_DH)vertices[i];
            if (pev.getElement().equals(key)) {
                //set the distance and change the color of the node to black
                pev.setDistanceOfNode(distance);
                pev.setColor(Color.black);

                if (pev.getFrom() != null) {
                    //set the reference through which the algorithm came to new lastChosen node to black.
                    PriorityExerVertex[] adjacents = pev.getFrom().getAdjacentVertices();
                    for (int j = 0; j < adjacents.length; j++) {
                        if (adjacents[j] == pev) {
                            pev.getFrom().setReferenceColor(j, Color.black);
                            break;
                        }
                    }
                    //because there are always two references between nodes in undirectedGraph set the color of the other one also to black
                    adjacents = pev.getAdjacentVertices();
                    for (int j = 0; j < adjacents.length; j++) {
                        if (adjacents[j] == pev.getFrom()) {
                            pev.setReferenceColor(j, Color.black);
                            break;
                        }
                    }

                }
                //save the new lastChosen node
                lastChosen.setObject(pev);
            }
        }
    }
    
    /**
     * This method sets the from node (father) of the node corresponding to the given key 
     * to be the current node in lastChosen variable.   
     * This is called when node is inserted in the heap or its value is updated in the heap.
     */
    public void setFromNode(Key key) {
        Vertex[] vertices = getVertices();
        for (int i = 0; i < vertices.length; i++) {
            PriorityExerVertex_DH pev = (PriorityExerVertex_DH)vertices[i];
            if (pev.getElement().equals(key)) {
                
                PriorityExerVertex_DH from = pev.getFrom();
                if (from != null) { 
                    /* From node for the node corresponding to the given key has already been set.
                     * Thus, we have to change first the reference color of the old reference 
                     * back to normal.
                     * For undirectedGraph there are always two references between two nodes so 
                     * changing colors for both of them.
                     */
                    
                    PriorityExerVertex[] adjacents = from.getAdjacentVertices();
                    for (int j = 0; j < adjacents.length; j++) {
                        if (adjacents[j] == pev) {
                            from.setReferenceColor(j, Color.blue);
                            break;
                        }
                    }
                    
                    adjacents = pev.getAdjacentVertices();
                    for (int j = 0; j < adjacents.length; j++) {
                        if (adjacents[j] == from) {
                            pev.setReferenceColor(j, Color.blue);
                            break;
                        }
                    }
                    
                }
                
                //Set from node to be the current node in the lastChosen
                pev.setFrom((PriorityExerVertex_DH)lastChosen.getObject());
                //Set the color of the node to gray.
                pev.setColor(Color.gray);

                from = (PriorityExerVertex_DH)lastChosen.getObject();
                if (from != null) {
                    //color the referencies to gray
                    
                    PriorityExerVertex[] adjacents = from.getAdjacentVertices();
                    for (int j = 0; j < adjacents.length; j++) {
                        if (adjacents[j] == pev) {
                            from.setReferenceColor(j, Color.gray);
                            break;
                        }
                    }
                    
                    adjacents = pev.getAdjacentVertices();
                    for (int j = 0; j < adjacents.length; j++) {
                        if (adjacents[j] == from) {
                            pev.setReferenceColor(j, Color.gray);
                            break;
                        }
                    }
                }
            }
        }
    }

    private static final long serialVersionUID = -7428829300361011004L;
}
