package content.exercises.structures;

import matrix.decoration.LabelDecorator;
import matrix.structures.CDT.CDT;
import matrix.structures.CDT.probe.BinHeap;
import matrix.structures.FDT.BinaryTree;
import matrix.structures.FDT.probe.Key;
import matrix.structures.memory.VirtualArray;
import matrix.structures.util.MatrixComparable;

/**
 * This class is used in the Dijkstra_Heap exercise. This heap uses the priority values in the
 * nodes of the heap to maintain the heap property. This also informs the graph everytime a new
 * node is inserted in the heap.
 * 
 * @author kstalnac
 *
 */
public class ExerLabeledHeap_DH extends BinHeap implements LabelDecorator {

    /** reference to the graph */
    private ExerLabeledGraph_DH graph;
  
    /**
     * Constructor.
     */
    public ExerLabeledHeap_DH() {
        super();
    }

    /**
     * Constructor.
     * @param g graph
     */
    public ExerLabeledHeap_DH(ExerLabeledGraph_DH g) {
        graph = g;
    }
    
    /**
     * Constructor.
     * @param bt nodes of this heap in virtual array
     * @param curr this node's number in the virtual array
     */
    private ExerLabeledHeap_DH(VirtualArray bt, int curr) {
        this.bt = bt;
        this.curr = curr;
    }

    // methods for interface LabelDecorator
    
    //labels are always on
    public boolean isLabelEnabled() {
        return true;
    }
    
    public void setLabelEnabled(boolean b) {
    }

    public void setLabel(String s) {
    }

    /**
     * Label of this node is the corresponding priority value in the key of the node.
     */
    public String getLabel() {
        if ((ExerPriorityElement_DH)getPriorityElement(curr) != null) {
            int priority = ((ExerPriorityElement_DH)getPriorityElement(curr)).getPriority();
            if (priority == Integer.MAX_VALUE)
                return "\u221e";
            else
                return priority + "";
        }
        return "";
    }

    // reference labels are not needed
    public boolean isReferenceLabelEnabled() {
        return false;
    }
    public void setReferenceLabelEnabled(boolean b) {
    }
    public void setReferenceLabel(int i, String s) {
    }
    public String getReferenceLabel(int i) {
        return null;
    }
    
    //end of methods for interface LabelDecorator
    
    /**
     * Insert method that inserts the given key to the heap.
     * This also informs the graph after every insertion.
     */
    public CDT insert(Object o) {
        if (o instanceof ExerKeyWithPriority_DH) {
            return super.insert(o);
        } else if (o instanceof Key){
            Key tempKey = (Key)o;
            ExerKeyWithPriority_DH newKey = new ExerKeyWithPriority_DH(tempKey.getObject(), tempKey.getComparisonString());
            newKey.setOrig(tempKey.getOrig());
            //create a priority element for the new key, get its priority from the graph
            newKey.setPriorityElement(new ExerPriorityElement_DH(graph.getDistance(tempKey)));
            //set from node (father node) of the key
            graph.setFromNode(tempKey);
            
            return super.insert(newKey);
        } else {
            return super.insert(o);
        }
    }
    
    BinaryTree left;
    public BinaryTree getLeft() {
        if (leftChild(curr) < size()) {
            if (left == null)
                left = new ExerLabeledHeap_DH(bt, leftChild(curr));
        }
        if (bt.getObject(leftChild(curr)) == null) {
            left = null;
        }
        return left;
    }

    BinaryTree right;
    public BinaryTree getRight() {
        if (rightChild(curr) < size()) { 
            if (right == null) 
                right = new ExerLabeledHeap_DH(bt, rightChild(curr));
        }
        if (bt.getObject(rightChild(curr)) == null) {
            right = null;
        }
        return right;
    }

    /**
     * Returns the minimum key from the heap.
     * @return the minimum key of the heap
     */
    public Key getMinimumKey() { 
        return (Key)getObject(0);
    }
    
    /**
     * Updtes the priority of the given key.
     * @param key key which priority is to be updated
     * @param newPriority new priority for the key
     */
    public void updatePriorityOfKey(Key key, int newPriority) {
        ExerKeyWithPriority_DH tmpKey;
        for (int i = 0; i < bt.size(); i++) {
            tmpKey = (ExerKeyWithPriority_DH)bt.getObject(i);
            if (tmpKey != null && tmpKey.getComparisonString().equals(key.getComparisonString())) {
                //create new key with the new priority
                ExerKeyWithPriority_DH newKey = new ExerKeyWithPriority_DH(tmpKey.getObject(), tmpKey.getComparisonString(), newPriority); 
                //set the new key to right position in the virtual array
                bt.setObject(newKey, i);
                //bubble the updated key to correct position
                bubbleUpdatedKey(i);
                //update the from node (father) of the key in the graph 
                graph.setFromNode(key);
                return;
            }
        }
    }
    
    /**
     * Bubble updated key to the correct position.
     * @param i key which position should be updated
     */
    private void bubbleUpdatedKey(int i) {
        MatrixComparable p = getPriorityElement(parent(i));
        MatrixComparable t = getPriorityElement(i);
        if (p.gt(t)) {
            //bubble up in the heap
            bubble(i);
        } else {
            //bubble down in the heap
            bubbledown(i);
        }
    }

    /**
     * Returns the priority of the key in the heap.
     * @param key which priority should be returned
     * @return priority of the given key, 
     * returns Integer.MIN_VALUE if the parameter key is not found from the heap 
     */
    public int getPriorityOfKey(Key key) {
        ExerKeyWithPriority_DH tmpKey;
        for (int i = 0; i < bt.size(); i++) {
            tmpKey = (ExerKeyWithPriority_DH)bt.getObject(i);
            if (tmpKey != null && tmpKey.getComparisonString().equals(key.getComparisonString())) {
                return ((ExerPriorityElement_DH)tmpKey.getPriorityElement()).getPriority();
            }
            
        }
        return Integer.MIN_VALUE;
    }
    private static final long serialVersionUID = -5643990600911522538L;
}
