/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package content.exercises.structures;

import java.awt.Color;

import matrix.decoration.*;
import matrix.structures.FDT.substructures.Vertex;
import matrix.structures.memory.VirtualObject;
import matrix.structures.memory.VirtualArray;

/**
 * New Implementation of Vertex with coloring by user.
 *
 * @author Harri Salonen
 */

public class ExerVertex implements Vertex, Styled {
    
    public static final boolean DEBUG = false;
    
    private VirtualArray successors;
    private VirtualObject color = new VirtualObject(this, "Vertex color");
    private StyleSheet ss = null;
    
    
    /* successors of this vertex. successor[0] is the key of this vertex */
    
    public ExerVertex() {
        successors = new VirtualArray(this, "Successors of the node");
        successors.setObject("", 0);
        setColor(Color.white);
    }
    
    public ExerVertex(Object o) {
        this();
        setElement(o);
    }
    
    public boolean hasSuccessor(ExerVertex V) {
        Vertex[] successorList = getSuccessors();
        for(int i = 0; i < successorList.length; i++) 
            if (V.equals((ExerVertex)successorList[i]))
                return true;
        return false;
    }
    
    public void markVisited(boolean f) {
        if (f) 
            markVisited();
        else
            markUnVisited();
    }
    
    public void markUnVisited() {
        this.setColor(Color.blue);
    }
    
    // Public exercise methods allowed for this component
    
    public void markVisited() {
        this.setColor(Color.red);
    }
    
    public void markFinished() {
        this.setColor(Color.black);
        
    }
    
    public boolean visited() {
        return (this.getColor().equals(Color.red));
    }
    
    public boolean finished() {
        return (this.getColor().equals(Color.black));
    }
    
    public void printSuccessors() {
        Vertex[] successorList = getSuccessors();
        for(int i = 0; i < successorList.length; i++) {
            if (successorList[i] != null)
                if (DEBUG)
                    System.out.print(i+": "+successorList[i].getElement()+" ");
                else
                    if (DEBUG)
                        System.out.print(i+": null ");
        }
    }
    
    
    
    
    public ExerVertex getNewVertex(Object o) {
        return new ExerVertex(o);
    }
    
    
    /** Inserts the param vertex to the list of successor vertices */
    public void addSuccessor(Vertex v) {
        addSuccessor(v, successors.size());
    }
    
    /** Private method for actually adding successors or child trees to
     * the graph/tree.
     * Should not be called directly, use addSuccessor(Vertex) instead
     */
    private void addSuccessor(Vertex v, int pos) {
        int i = 1;
        for(; i <= pos && successors.getObject(i) != null; i++) {
            if(v == successors.getObject(i))
                return;
        }
        if(v instanceof ExerVertex) {
            successors.setObject(v, i);
        } else {
            ExerVertex tmp = getNewVertex(v.getElement());
            successors.setObject(tmp, i);
        }
        //((VertexStyleSheet)getStyleSheet()).addSuccessor(i);
    }
    
    /** Removes the param vertex from the list of successor vertices. */
    public void removeSuccessor(Vertex v) {
        //Note.out(this, "removing successor " + v);
        int i;
        ExerVertex tmp;
        for(i = 1; i <= successors.size(); i++) {
            tmp = (ExerVertex)successors.getObject(i);
            if(tmp != null && tmp.equals(v))
                break;
        }
        for(; i < successors.size(); i++)
            successors.setObject(successors.getObject(i + 1), i);
    }
    
    // interface Vertex
    
    public Vertex[] getSuccessors() {
        ExerVertex[] tmp = new ExerVertex[successors.size() - 1];
        for(int i = 1 /*successors.getFirst()*/ ; i < successors.size(); i++) {
            tmp[i - 1] = (ExerVertex) successors.getObject(i);
        }
        return tmp;
    }
    
    public void setSuccessors(Vertex[] v) {
        int i = 0;
        for(; i < v.length; i++)
            successors.setObject(v[i], i + 1);
        successors.setLast(i);
        //int last = successors.size();
        //for(; i < last; i++)
        //successors.setObject(null, i + 1);
    }
    
    public Object getElement() {
        return successors.getObject(0);
    }
    
    public void setElement(Object o) {
        successors.setObject(o, 0);
    }
    
    // interface Styled
    
    public StyleSheet getStyleSheet() {
        if(ss == null)
            ss = new StyleSheetAdapter() {
            public boolean isInfoEnabled() { return false; }
            public String getInfo() { return "Exercise vertex"; }
            public boolean isLabelEnabled() { return false; }
            public String getLabel() { return ""; }
            public Color getDefaultBackgoundColor() { return Color.white; }
            static final long serialVersionUID = -2360089479061355464L;  
        };
        
        ss.setDefaultBackgroundColor(getColor());
        return ss;
    }
    
    
    public void setColor(Color c) {
        color.setObject(c);
        getStyleSheet().setDefaultBackgroundColor(c);
    }
    
    public Color getColor() {
        return (Color)color.getObject();
    }
    
    public void setLabel(String s) {
        getStyleSheet().setLabel(s);
    }
    
    public String getLabel() {
        return getStyleSheet().getLabel();
    }
    
    public boolean equals(Object o) {
        if (o instanceof ExerVertex) 
            return ((this.getColor().equals(((ExerVertex)o).getColor()))
                    && (this.getElement().equals(((ExerVertex)o).getElement())));
        return false;
    }
    
    static final long serialVersionUID = -2360089479061355464L;
}
