/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2006 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package content.interfaces;

import matrix.simulation.VisualTypeConf;

/**
 * ConfigureVisualType interface is used to enable/disable dragging or dropping
 * from a structure and possible use of menus. It is also used to configure some
 * visualization related options - whether array indices are shown, if
 * structures should be flipped in horizontal or vertical direction and if
 * structures should highlight when they are operated on.
 * 
 * Only one method is needed to return an array of corresponding configure
 * objects (VisualTypeConf) for each data structure to be configured.
 * 
 * matrix.simulation.VisualTypeConf class contains several methods that can be
 * used to save and load properties (such as drag, drop, highlight) in order to
 * configure visualizations in exercises.
 * 
 * See, for example, the implementation of the content.exercises.InsertionSort
 * exercise. In conf method, several GUI operations are enabled by
 * 
 * <pre>
 * tabl1.enable(&quot;matrix.visual.VisualKey&quot;, VisualTypeConf.HIGHLIGHT_OPERATION);
 * tabl1.enable(&quot;matrix.visual.VisualKey&quot;, VisualTypeConf.DRAG_OPERATION);
 * tabl1.enable(&quot;matrix.visual.VisualKey&quot;, VisualTypeConf.DROP_OPERATION);
 * </pre>
 * 
 * In the above, all the keys in the array are highlighted when cursor is on the
 * key. One can also drag & drop keys from and to the array, correspondingly.
 * Note, however, that all the components need to be set separately. For
 * example, the highlight and drop for array components need to be set as well:
 * 
 * <pre>
 * tabl1.enable(&quot;matrix.visual.VisualArrayComponent&quot;,
 * 		VisualTypeConf.HIGHLIGHT_OPERATION);
 * tabl1.enable(&quot;matrix.visual.VisualArrayComponent&quot;,
 * 		VisualTypeConf.DROP_OPERATION);
 * </pre>
 * 
 * In addition to this, for example, "tmp" structure is set to be non-indexed by
 * 
 * <pre>
 * tabl2.setVisualConfValue(&quot;matrix.visual.VisualArray&quot;, VisualTypeConf.INDEXED,
 * 		&quot;false&quot;);
 * </pre>
 * 
 * How to implement new properties?
 * 
 * For example, there is a configure method for VisualArray. This method
 * overrides the configure method of VisualType. The value of
 * VisualTypeConf.INDEXED property is fetched from VisualTypeconf and the
 * indexes are set according to that value. At the end of the configure method,
 * the super class configure method is called for VisualType.
 * 
 * @see matrix.simulation.VisualTypeConf
 */
public interface ConfigureVisualType {

	/**
	 * Returns an array of corresponding configure objects (VisualTypeConf) for
	 * each data structure to be configured.
	 * 
	 * @return an array of VisualTypeConf objects
	 * @see matrix.simulation.VisualTypeConf
	 */
	public VisualTypeConf[] conf();

}
