/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: Animator.java,v 1.58 2007/09/03 13:14:53 oseppala Exp $ */

package matrix.animation;

import java.io.*;
import java.util.Iterator;
import java.util.Stack;
import java.util.StringTokenizer;

import matrix.structures.memory.MemoryStructure;
import matrix.util.*;


/**
 * <p>
 * Stores animated operations to memory and allows user to
 * browse operation history. For example when inserting Key to a binary tree
 * animator stores the states of the structure and the visualization can now
 * be  replayed step-by-step.
 * </p><p>
 * Normally only one animator is needed to store user's operation history.
 * Multiple animators are needed for example, when user is solving an
 * exercise, and when such exercise is judged. The animator -class keeps
 * track of the instance of Animator that is active at the moment. Usually
 * the active Animator is the Animator of the active window, but by using
 * the static <code>lock()</code> method the active Animator can be locked
 * so it cannot be changed (except by the holder of the lock) until released.
 * </p><p>
 * All animated operations are conserned as assignments. Every assignment
 * should be  bethween <code>startOperation()</code> and
 * <code>endOperation()</code> pair. These two methods constitute parenthesis
 * hierarcy. <code>undo()</code> and <code>redo()</code> do not go inside
 * this hierarcy - they just find the pair
 * to a next or a previous parenthesis.
 * </p><p>
 * ((A B)(C))(D)</p><p>
 * Capital letters mean animated assignment. From end animation can be
 * rewinded to the beginning with two <code>undo()</code> commands.
 * </p>
 *
 * @see AnimatedOperation
 * @author Jussi Nikander
 * @author Petri Tenhunen
 */
public class Animator implements Serializable {
    private Stack undo, redo;
    private String name;
    protected int parentheses;
    private static Animator activeAnimator = null;

    /** Unstarted end operations after current position. */
    private int endsAfter=0;

    /** Unended start operations before current position. */
    private int startsBefore=0;

    public final static String DEFAULT_NAME = "Matrix Animator";

    /**
     * Default constructor for animator named as defaultAnimator.
     * Currently animator's names are not used by matrix framework.
     */
    public Animator() {
        undo = new Stack();
        redo = new Stack();
        parentheses = 0;
        name = DEFAULT_NAME;
    }

    /**
     * Constructor for named animator
     *
     * @param name name of the new animator
     */
    private Animator(String name) {
        this();
        this.name = name;
    }



    /** debug method */
    public String print(Stack s) {
        return s.toString();
    }

    /** debug method */
    public static final String nestLines(String s, int depth){
        StringTokenizer st = new StringTokenizer(s,"\n");
        StringBuffer result = new StringBuffer();

        while (st.hasMoreTokens()){
            for (int i=0; i<depth; i++){
                result.append("|   ");
            }
            result.append(st.nextToken());
            result.append("\n");
        }
        return result.toString();
    }



    /** debug method */
    public void dump(PrintStream out) {
        out.println("Animator: " + this);

        out.println("Undo stack ====================");

        Iterator it = undo.iterator();

        int nestingDepth = 0;
        while (it.hasNext()){
            AnimatedOperation ao = (AnimatedOperation)it.next();

            if (ao instanceof endOperation)
                nestingDepth--;

            out.print(nestLines(ao.toString(), nestingDepth));

            if (ao instanceof startOperation)
                nestingDepth++;        
        }

        out.println("Redo stack ====================");

        it = redo.iterator();

        nestingDepth = 0;
        while (it.hasNext()){
            AnimatedOperation ao = (AnimatedOperation)it.next();

            if (ao instanceof startOperation)
                nestingDepth--;        

            out.print(nestLines(ao.toString(), nestingDepth));

            if (ao instanceof endOperation)
                nestingDepth++;

        }

    }

    /**
     * Set marker "start operation" for the Animator.
     * Undo and Redo always executes primitive operations until startOperation or
     * endOperation is encountered, respectively. For each startOperation there should
     * exists the corresponding endOperation and all the AnimatedOperations should
     * exist between startOperation and endOperation.
     * @see #endOperation()
     * @see startOperation
     */
    public startOperation startOperation() {
        setActiveAnimator(this);
        startOperation so = new startOperation();
        put(so);
        return so;
    }


    public startOperation startOperation(String annotation) {
        setActiveAnimator(this);
        startOperation so = new startOperation(annotation);
        put(so);
        return so;
    }

    /**
     * Set marker "end operation" for the Animator
     * Undo and Redo always executes primitive operations until startOperation or
     * endOperation is encountered, respectively. For each startOperation there should
     * exists the corresponding endOperation and all the AnimatedOperations should
     * exist between startOperation and endOperation.
     * @see #startOperation()
     * @see endOperation
     */
    public void endOperation() {
        put(new endOperation()); }

    /**
     * This method is used to execute an encapsulated Command pattern. given command
     * pattern is executed bethween starOperation and endOperation pair.
     *
     * @param c command pattern to encapsulate
     */
    public void operation(Command c) {
        startOperation();
        c.execute();
        endOperation();
    }

    /**
     * Set marker "start operation" for the selected Animator with prefix command ao.
     * Animator is originally designed for animating assignments (substitution of
     * evaluated right side into the left side of an expression). However, in order to
     * give a visual hint where the right side is evaluated, this method gives a way
     * to determine the evaluated operation as a prefix command attaches to the
     * assignment.
     * @see #startOperation()
     * @see startOperation
     */
    public startOperation startOperation(AnimatedOperation ao) {
        setActiveAnimator(this);
        startOperation so = new startOperation(ao);
        put(so);
        return so;
    }

    /**
     * Set a new marker operation to this animator. Markers
     */
    private void markerOperation(String marker) {
        put(new markerOperation(marker));
    }

    /**
     * Creates and puts a new animated assignment operation into the selected Animator.
     * @param ms the left side memory structure of the assigment
     * @param obj the right side of the assignment
     * @param index of the left side memory structure
     * @see Assignment
     */
    public void putAssignment(MemoryStructure ms,
            Object obj,
            int index) {
        //Note.out(this, "putAssignment, this=" + this);
        Assignment a = new Assignment(ms, obj, index);
        put(a);
    }

    /** Clear undo-stack. Because undo - stack is cleared <code>undo</undo> is disposed. */
    public void disposeUndo() {
        undo=new Stack();
        newEvent("dispose undo");

        /* Preserve nesting depth. */
        for(int i=0;i<startsBefore;i++)
            undo.push(new startOperation());
    }

    /**
     * Clear redo-stack. Because redo - stack is cleared <code>redo</undo> is disposed.
     * @return stack Containing removed elements from redo - stack.
     */
    public Stack disposeRedo() {
        Stack removed = new Stack();

        while(hasNextOperation())
            removed.push(redo.pop());

        /* Preserve nesting depth. */
        for(int i=0;i<endsAfter;i++)
            redo.push(new endOperation());

        return removed;
    }

    /**
     * This should not be used or at least should be used withs care. This pops all items
     * from argument stack and pushs Objects one by one into the redo-stack
     * Warning: this method ignores start-/endOperation balance completely; the
     * caller is reponsible for ensuring that it is maintained.
     *
     * @param s stack to add to the redo stack
     */
    public void setNewRedoStack(Stack s) {
        while (!s.empty())
            redo.push(s.pop());
    }

    /**
     * Move topmost animated operation from {@link #undo} to {@link #redo}.
     * @return The moved operation.
     */
    private AnimatedOperation moveFromUndoToRedo() {
        AnimatedOperation ao=(AnimatedOperation)undo.pop();
        if (ao instanceof startOperation)
            startsBefore--;
        if (ao instanceof endOperation)
            startsBefore++;
        redo.push(ao);
        if (ao instanceof endOperation)
            endsAfter++;
        if (ao instanceof startOperation)
            endsAfter--;

        return ao;
    }

    /**
     * Move topmost animated operation from {@link #redo} to {@link #undo}.
     * @return The moved operation.
     */
    private AnimatedOperation moveFromRedoToUndo() {
        AnimatedOperation ao=(AnimatedOperation)redo.pop();
        if (ao instanceof endOperation)
            endsAfter--;
        if (ao instanceof startOperation)
            endsAfter++;
        undo.push(ao);
        if (ao instanceof startOperation)
            startsBefore++;
        if (ao instanceof endOperation)
            startsBefore--;

        return ao;
    }

    /** undo last set of atomic operations */
    public void undo() {
        AnimatedOperation ao = null;
        int stack = 0;
        //Note.out(this, "<<<< UNDO " + name);
        do {
            /*
            Assert.notFalse(hasPreviousOperation(),
                            "No previous operation or missing startOperation");
             */
            ao = moveFromUndoToRedo();
            //Note.out(this, "Calling undo for ao " + ao + " of class " + (ao == null ? null: ao.getClass().getName()));
            ao.undo();
            //Note.out(this, "Returning from call from " + ao);
            if(ao instanceof endOperation)
                stack++;
            else if(ao instanceof startOperation)
                stack--;
            // marker as a very first operation
            if ((ao instanceof markerOperation) &&
                    (!hasPreviousOperation()))
                break;
            newEvent("step");
        } while (stack > 0 && hasPreviousOperation());
        if(stack > 0) {
            while(stack > 0) {
                redo.push(new startOperation());
                stack--;
            }
        }
        newEvent("undo");
        //Note.out(this, "<<<< UNDO DONE " + name);
    }

    /**
     * Is <code>undo()</code> possible or not.
     * @return true if and only if undo is possible.
     */
    public boolean hasPreviousOperation() { return !undo.empty(); }

    /** redo last (undone) set of operations */
    public void redo() {
        AnimatedOperation ao;
        int stack = 0;
        //Note.out(this, ">>>> REDO " + name);
        do {
            Assert.notFalse(hasNextOperation(),
            "No next operation or missing endOperation");
            ao = moveFromRedoToUndo();
            ao.redo();

            if(ao instanceof startOperation)
                stack++;
            else if(ao instanceof endOperation)
                stack--;
            newEvent("Step");
        } while (stack > 0);
        newEvent("redo");
    }

    /**
     * Is <code>redo()</code> possible or not.
     * @return true if and only if redo is possible.
     */
    public boolean hasNextOperation() { return !redo.empty(); }

    /**
     * Rewind to the last <code>markerOperation</code> with given name. If marker with
     * given text is not found assertion "Marker not found" is thrown.
     *
     * @param m Marker text to search when animating back.
     */
    public void rewind(String m) {
        AnimatedOperation ao;
        //Note.out(this, "<<<< REWIND " + name);
        do {
            Assert.notFalse(hasPreviousOperation(), "Marker not found");
            ao = moveFromUndoToRedo();
            ao.undo();
        } while(!(ao instanceof markerOperation));
        if (!((markerOperation)ao).isMarker(m))
            rewind(m);
        else
            newEvent("rewind");
    }

    /**
     * Rewind to the last <code>markerOperation</code> with given name, and
     * rewind past the beginning of the corresponding
     * <code>startOperation</code> to ensure that we stay at the same
     * hierarchy level (or step out). 
     *
     * If marker with given text is not found assertion "Marker not found"
     * fails.
     *
     * @param m Marker text to search when animating back.
     */
    public void rewindBlocks(String m) {
        AnimatedOperation ao;
        //Note.out(this, "<<<< REWIND " + name);
        int stack=0;

        do {
            do {
                Assert.notFalse(hasPreviousOperation(), "Marker not found");
                ao = moveFromUndoToRedo();
                ao.undo();
                if(ao instanceof endOperation)
                    stack++;
                else if(ao instanceof startOperation)
                    stack--;
            } while(!(ao instanceof markerOperation));
        } while(!((markerOperation)ao).isMarker(m));

        newEvent("rewind");

        do {
            ao = moveFromUndoToRedo();
            ao.undo();

            if(ao instanceof endOperation)
                stack++;
            else if(ao instanceof startOperation)
                stack--;

            if ((ao instanceof markerOperation) &&
                    (!hasPreviousOperation()))
                break;
            newEvent("step");
        } while (stack > 0 && hasPreviousOperation());
    }

    /**
     * Rewind the animator to the beginning.
     * 
     * @return number of steps taken to reach beginning
     */
    public int rewind() {
        int count = 0;
        while (hasPreviousOperation()){
            undo();
            count++;
        }
        newEvent("rewind");
        return count;
    }


    /**
     * Sets the animator to the end state.
     *  
     * @return number of steps taken to reach the end
     */
    public int end() {
        int count = 0;
        while (hasNextOperation()){
            redo();
            count++;
        }
        return count;
    }




    /**
     * Seek forward to the next <code>markerOperation</code> with given name. If marker with
     * given text is not found assertion "Marker not found" is thrown.
     *
     * @param m Marker text to search when animating forward.
     */
    protected void search(String m) {
        AnimatedOperation ao;
        //Note.out(this, "<<<< REWIND " + name);
        do {
            Assert.notFalse(hasNextOperation(), "Marker not found");
            ao = moveFromRedoToUndo();
            ao.redo();
        } while(!(ao instanceof markerOperation));
        if (!((markerOperation)ao).isMarker(m))
            search(m);
        else
            newEvent("search");
    }

    // Scrollbar
    /** Goes to the given state of the maximum number of max steps. */
    public void GoTo(int state, int max) {
        AnimatedOperation ao;
        if (getState() < state) {
            ao = moveFromRedoToUndo();
            ao.redo();
            GoTo(state, max);
        } if (getState() > state) {
            ao = moveFromUndoToRedo();
            ao.undo();
            GoTo(state, max);
        } else newEvent("goto");
    }

    /**
     * size of undo-stack.
     * @return size of the undo - stack.
     */
    public int getState() { return undo.size(); }

    /**
     * size of undo-stack + size of redo-stack
     * @return size of undo-stack + size of redo-stack
     */
    public int getMaximum() { return undo.size() + redo.size(); }

    /**
     * Put a new AnimatedOperation in Animator.
     *
     * @param ao Operation to put in Animator
     * @see AnimatedOperation
     */
    public void put(AnimatedOperation ao) {
        //Note.out(this, "put(" + ao + ") this=" + this );
        // remove adjacent start and end operations
        if(ao instanceof startOperation) {
            parentheses++;
        }
        if (ao instanceof endOperation) {
            if(parentheses > 0)
                parentheses--;
            AnimatedOperation peek = (AnimatedOperation)undo.pop();
            if (peek instanceof startOperation) {
                startsBefore--;
                return;
            }
            undo.push(peek);
        }
        if(parentheses <= 0 && !(ao instanceof endOperation))
            return;
        if (ao instanceof startOperation)
            startsBefore++;
        if (ao instanceof endOperation)
            startsBefore--;
        undo.push(ao);

        // cannot redo old undone operations anymore

        if ((redo != null) && (!redo.isEmpty()) && !(ao instanceof markerOperation)) {
            Note.out(this, "clear redo");
            redo = new Stack();

            /* Preserve nesting depth. */
            for(int i=0;i<endsAfter;i++)
                redo.push(new endOperation());
        }
        // update all observing windows
        if (ao instanceof endOperation) newEvent("put");
    }

    /**
     * Name of a selected Animator.
     * @return Name of a selected Animator
     */
    public String getName() { return name; }

    protected int getParentheses() { return parentheses; }


    protected void newEvent(String event) {
        //fireActionEvent(new ActionEvent(this, 0, event));
    }



    public static Animator getActiveAnimator() {
        if (activeAnimator == null)
            activeAnimator = new Animator();
        return activeAnimator;
    }

    public static void setActiveAnimator(Animator a) {
        activeAnimator = a;
    }

    /** 
     * Returns the current redo stack. You shoudn't manipulate the
     * Stack returned by this method unless you know what you are
     * doing. 
     */
    protected Stack getRedoStack() { return redo; }

    /** 
     * Returns the current undo stack. You shoudn't manipulate the
     * Stack returned by this method unless you know what you are
     * doing. 
     */
    protected Stack getUndoStack() { return undo; }

    static final long serialVersionUID = 6289787979389228045L;

    public int getStartsBefore() {
        return startsBefore;
    }

    public int getEndsAfter() {
        return endsAfter;
    }
}
