/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: VisualAnimator.java,v 1.38 2007/02/19 14:38:13 etanol Exp $ */

package matrix.animation;

import java.awt.*;
import javax.swing.*;
import java.awt.event.*;
import java.lang.reflect.InvocationTargetException;

/** Implementation for Animator Window. */
public class VisualAnimator extends JPanel {
    protected JButton undo, redo, begin, end, play, resetBeg;
    protected JScrollBar sb;
    protected Animator animator;
    private boolean buttonsLocked = false;
    private int waitTime = DEFAULT_WAIT_TIME;
    public static final int DEFAULT_WAIT_TIME = 500;

    public VisualAnimator(Animator a) {
        this.animator = a;
        initComponents();
        layoutComponents();
    }
    
    /**
     * Allows changing animator without creating a new VisualAnimator
     * @param a the new animator
     */
    public void changeAnimator(Animator a){
        this.animator = a;
        validate();
    }
    
    protected void initComponents() {
        undo = new JButton("Backward");
        redo = new JButton("Forward");
        begin = new JButton("Begin");
        end = new JButton("End");
        play = new JButton("Play");
        resetBeg = new JButton("Reset beginning here");
        undo.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                undo();
                fireActionEvent(e);
            }
        });

        redo.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                redo();
                fireActionEvent(e);
            }
        });

        begin.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                rewind();
                fireActionEvent(e);
            }
        });

        end.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                end();
                fireActionEvent(e);
            }
        });

        /*
         * After "Play" is pressed, repeat: - Wait waitTime ms. - Step forward in
         * animation. until the end of the animation is reached.
         * 
         * The step events are put on the AWT event queue to avoid
         * synchronisation problems between successive steps.
         */

        play.addActionListener(new ActionListener() {
            public void actionPerformed(final ActionEvent e) {
                new Thread() {
                    public void run() {
                        while (animator.hasNextOperation()) {
                            try {
                                Thread.sleep(waitTime);
                                try {
                                    EventQueue.invokeAndWait(new Runnable() {
                                        public void run() {
                                            if (animator.hasNextOperation()) {
                                                animator.redo();
                                                fireActionEvent((ActionEvent) e);
                                            }
                                        }
                                    });
                                } catch (InvocationTargetException ite) {
                                    ite.getTargetException().printStackTrace();
                                }
                            } catch (InterruptedException ie) {
                            }
                        }
                    }
                }.start();
            }
        });

        resetBeg.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                animator.disposeUndo();
                fireActionEvent(e);
            }
        });
        sb = new JScrollBar(JScrollBar.HORIZONTAL);
    }

    //This var is used to break the following cycle:
    //
    // AdjuctmentvalueChanged -> FireActionEvent -> Structurepanel$2.actionperformed
    // -> structurepanel.validateanimator -> visualanimator.validate
    // -> visualanimator.validatescrollbar -> system-calls -> AdjustmentvalueChanged
    private int stackDepth = 0;

    
    protected void layoutComponents() {
        setLayout(new BorderLayout(100, 10));
        JPanel p = new JPanel();

        p.add(undo);
        p.add(redo);
        p.add(begin);
        p.add(end);
        p.add(play);

        this.add(p, "North");
        this.add(sb, "Center");
        this.add(resetBeg, "South");

        sb.addAdjustmentListener(new AdjustmentListener() {

            
                public void adjustmentValueChanged(AdjustmentEvent e) {
                    if (stackDepth != 0)
                        return;                    
                    
                    GoTo(e.getValue(), sb.getMaximum());

                    stackDepth += 1;
                    fireActionEvent(new ActionEvent(this, 0, "adjust"));
                    stackDepth -=1;

                    //focus must be transfered to get the shortcut keys for
		    //animator to work
		    sb.transferFocus();
                }
            });

 		setResetButtonVisible(false);
        validate();
    }

    public void actionPerformed(ActionEvent event) {
        getParent().validate();
    }


    public void setResetButtonVisible(boolean value) {
        if (resetBeg != null)
            resetBeg.setVisible(value);
    }


    public void setPlayButtonVisible(boolean value) {
        if (play != null)
            play.setVisible(value);
    }


    protected void undo() {
        if (animator.hasPreviousOperation()) {
            animator.undo();
        }
    }


    protected void redo() {
        if (animator.hasNextOperation()) {
            animator.redo();
        }
    }


    protected void rewind() {
	while (animator.hasPreviousOperation()) {
	    animator.undo();

	    //by calling this in every step correct visualizations are created
	    fireActionEvent(new ActionEvent(this, 0, "undo")); 
	}
	//        animator.rewind();
    }


    protected void end() {
        while (animator.hasNextOperation()) {
            animator.redo();

	    //by calling this in every step correct visualizations are created
	    fireActionEvent(new ActionEvent(this, 0, "redo"));
	}
    }

    public void setButtonsLocked(boolean t) {
	buttonsLocked=t;
    }

    public void validate() {
        super.validate();
		if (buttonsLocked)
		    return;
        if (animator.hasNextOperation()) {
            if (play != null)
                play.setEnabled(true);
            if (redo != null)
                redo.setEnabled(true);
            if (end != null)
                end.setEnabled(true);
        } else {
            if (play != null)
                play.setEnabled(false);
            if (redo != null)
                redo.setEnabled(false);
            if (end != null)
                end.setEnabled(false);
        }

        if (animator.hasPreviousOperation()) {
            if (undo != null)
                undo.setEnabled(true);
            if (begin != null)
                begin.setEnabled(true);
        } else {
            if (undo != null)
                undo.setEnabled(false);
            if (begin != null)
                begin.setEnabled(false);
        }
        if (sb != null)
            validateScrollbar();
    }


    protected void validateScrollbar() {
        // XXX: Temporary fix
    	//stackDepth++;
        sb.setValues(animator.getState(), 1, 0, animator.getMaximum());
        //stackDepth--;
    }


    protected void GoTo(int index, int maximum) {
        animator.GoTo(index, maximum);
    }

    // AWT event handling
    protected ActionListener listener = null;

    /** Add new listener. */
    public void addActionListener(ActionListener a) {
        listener = AWTEventMulticaster.add(listener, a);
        //animator.addActionListener(a);
    }

    /** Remove listener. */
    public void removeActionListener(ActionListener a) {
        listener = AWTEventMulticaster.remove(listener, a);
        //animator.removeActionListener(a);
    }

    /** This method is invoked when user action takes place
     ** in this VisualAnimator.
     * All registered listeners are notified that they might be changed.
     */
    public void fireActionEvent(ActionEvent e) {
      if (listener != null) {
        //Note.out(this, "firing action to listener " + listener);
          listener.actionPerformed(e);
      }
    }
    // -------------------

    /** 
     * Sets the time to wait in one step when playing the animation. 
     */
    public void setWaitTime(int time) {
    waitTime = time;
    }


    static final long serialVersionUID = -11601511645348702L;

    public long getWaitTime() {
        return waitTime;
    }
}

