/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: BinHeap.java,v 1.13 2006/12/12 14:12:54 oseppala Exp $ */

package matrix.structures.CDT.probe;

import matrix.structures.CDT.CDT;
import matrix.structures.CDT.interfaces.PrioritySupplied;
import matrix.structures.FDT.probe.*;
import matrix.structures.memory.*;
import matrix.structures.other.IteratedOperations;
import matrix.structures.util.MatrixComparable;

import matrix.util.Note;

public class BinHeap extends BinTrei implements CDT, matrix.structures.ADT.PriorityQueue {
    private VirtualInteger size;

    public BinHeap() { size = new VirtualInteger(0, this, "size of the heap"); }

    public CDT getNewInstance(){
    	return new BinHeap();
    }
    
    //TODO: check if this is enough, could it be an element with null data?
    public boolean ADTIsEmpty(){
        return size.eval() == 0;        
    }
    
    public void ADTinsert(PrioritySupplied o){
    	insert(o);
    }
    
    public PrioritySupplied ADTdeleteMin(){
        return deleteMin();
    }
    
    public PrioritySupplied deleteMin(){
        PrioritySupplied root = (PrioritySupplied)this.getObject(0);
        delete(null);
        
        return root;
    }

    /**
     * Insertion to a heap is a normal heap insertion using the given priority
     */
    
    public CDT insert(Object o) {

        if (o instanceof PrioritySupplied) {
            setObject(o, size.eval()); //Insert the new key in the last position
            bubble(size.eval());       //PercolateUp
            size.inc();                //Increment element count 
            
            return this;
        } else if (o instanceof Table) {
            IteratedOperations.iteratedInsert(this, (Table)o);
        } else
            Note.show(this, "Cannot insert object of type " + o.getClass());

        return this;
    }

    /**
     * The delete implementation for heaps is deletemin 
     */
    public CDT delete(Object o) {
        //Get the last element
        MatrixComparable tmp = (MatrixComparable)getObject(size.eval() - 1);
        setObject(tmp, 0);                //Replace the root with the last element
        setObject(null, size.eval() - 1); //Set the last element to be null
        size.dec();                        //decrease the element count
        bubbledown(0);                     // Fix the heap property
        return this;
    }

    /**
     * The CDT search method is not implemented for heaps 
     * -
     * default is to return this
     */
    public Object search(Object o) { return null; }

    /**
     * Get the index of the left child of the given index
     */
    public int leftChild(int n) {
        if(n == 0) return 1;
        return 2 * n + 1;
    }

    /**
     * Get the index of the right child of the given index
     */
    public int rightChild(int n) {
        if(n == 0) return 2;
        return 2 * n + 2;
    }
    /**
     * Get the index of the parent child of the given index
     */

    public int parent(int n) {
        //Root has no parent, its root is itself
        if (n == 0)
            return 0;
        
        return (n-1)/2;
    }

    public MatrixComparable getPriorityElement(int i){
        PrioritySupplied ps = (PrioritySupplied)getObject(i);
        if (ps == null)
            return null;
                
        return ps.getPriorityElement(); 
    }
    
    /**
     * Normal percolate-down
     * @param i
     */
    
    protected void bubbledown(int i) {
        //Note.out(this, "Bubbledown: " + i);
        //Note.out(this, "lc: " + leftChild(i) + " rc: " + rightChild(i));
        if(i >= size.eval()) //Elements that have no children are not considered
            return;
        
        MatrixComparable p  = getPriorityElement(i);
        MatrixComparable lc = getPriorityElement(leftChild(i));
        MatrixComparable rc = getPriorityElement(rightChild(i));

        //If there is no left child, theres no right child either
        if(lc == null)
            return;
        
        //If the right child is null, consider only the left one
        if(rc == null) {
            if(p.gt(lc)) {
                swap(i, leftChild(i));
            }
            return;
        }
        
        //Otherwise find out the smaller child
        if(lc.gt(rc)) {
            //..and compare it to the parent
            if(p.gt(rc)) {
                //..and swap if necessary
                swap(i, rightChild(i));
                bubbledown(rightChild(i));
            }
        } else {
            if(p.gt(lc)) {
                swap(i, leftChild(i));
                bubbledown(leftChild(i));
            }
        }
    }

    protected void bubble(int n) {
        MatrixComparable p = getPriorityElement(parent(n));
        MatrixComparable t = getPriorityElement(n);
        
        if (p.gt(t)) {
            swap(n,parent(n));
            bubble(parent(n));
        }
    }

    static final long serialVersionUID = 8971892506567935031L;
}
