/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package matrix.structures.CDT.probe;

import matrix.structures.CDT.CDT;
import matrix.structures.FDT.*;
import matrix.structures.FDT.probe.*;
import matrix.structures.memory.*;
import matrix.structures.other.Comparer;
import matrix.structures.util.MatrixComparable;
import matrix.util.Note;
import matrix.util.Traverse;

/**
 * A Binary search tree. A binary search tree is a dictionary in the form
 * of a binary tree where the left subtree of each node contains smaller
 * keys than the node's key, and the right subtree contains larger keys.
 * Duplicate keys goes to the left branch of the node.
 *
 * @author Jussi Nikander
 */
public class BinSearchTree implements CDT,Tree {
    
    protected VirtualObject root;
    
    /** Constructs a new BinSearchTree */
    public BinSearchTree() {
        this(null);
    }
    
    public CDT getNewInstance(){
        return new BinSearchTree();
    }
    
    /**
     * Constructs a new BinSearchTree
     * @param o the key to the root node
     */
    public BinSearchTree(Object o) {
        root = new VirtualObject(this, "root node of the tree");
        root.setObject(getNewNode(o));
    }
    
    /**
     * Tests for the equality of this object and the parameter object.
     * @param o the Object tested against
     * @return true if the two objects are considered equal, false otherwise
     */
    public boolean equals(Object o) {
        if(o instanceof BinaryTree) {
            return Comparer.compareIdentical((BinaryTree)getElement(), (BinaryTree)o);
        }
        if(o instanceof BinSearchTree)
            return Comparer.compareIdentical((BinaryTree)getElement(), (BinaryTree)((BinSearchTree)o).getElement());
        return false;
    }
    
    public Object getElement() {
        return root.getObject();
    }
    
    public void setElement(Object o) {
        // To prevent setting this as the Element of this.
        if(o == this)
            return;
        if(o instanceof BinaryTree)
            root.setObject(o);
        else if (o == null)
            root.setObject(getNewStructure());
        else
            Note.err(this, "Cannot set non-BinaryTree as root");
    }
    
    /**
     * Tells if a tree is empty or not
     * @return true if the tree is empty, false otherwise
     */
    public boolean empty() {
        return getElement() == null;
    }
    
    /**
     * Creates a new empty Node.
     * @return a new node
     */
    public FDT getNewStructure() {
        return new BinTree() ;
    }
    
    /**
     * Creates a new node.
     * @param key the key of the new node
     * @return a new node
     */
    public Tree getNewNode(Object key) {
        return new BinTree(key);
    }
    
    // Methods for interface CDT
    
    public CDT insert(Object o) {
        if(o == null) 
            return this;
        if(o instanceof MatrixComparable) {
            setElement(insertHelp((MatrixComparable)o, (BinaryTree)getElement()));
            return this;
        }
        if(o instanceof Table) {
            Table t = (Table)o;
            for(int i = 0; i < t.size(); i++)
                insert(t.getObject(i));
            return this;
        }
        Key k = new Key(o);
        setElement(insertHelp(k, (BinaryTree)getElement()));
        return this;
    }
    
    /**
     * Helper function for insert.
     * @param k the new key to be inserted
     * @param tree the subtree currently being examined
     */
    protected BinaryTree insertHelp(MatrixComparable k, BinaryTree tree) {
        if(tree == null)
            return (BinaryTree) getNewNode(k);
        MatrixComparable key = (MatrixComparable)tree.getElement();
        if(key == null) {
            tree.setElement(k);
            return tree;
        }
        if(k.leq(key))
            tree.setLeft(insertHelp(k, tree.getLeft()));
        else
            tree.setRight(insertHelp(k, tree.getRight()));
        return tree;
    }
    
    public CDT delete(Object o) {
        if(o == null)
            return this;
        if(o instanceof Key) {
            setElement(deleteHelp((Key)o, (BinaryTree)getElement()));
            return this;
        }
        Note.show(this, "Cannot delete, not a Key");
        return this;
    }
    
    /**
     * Helper function for delete.
     * @param k the key to be deleted
     * @param tree the subtree currently being examined
     */
    protected BinaryTree deleteHelp(Key k, BinaryTree tree) {
        Key key = tree != null ? (Key)tree.getElement() : null;
        if (key == null) 
            return null;
        if(k.equals(key)) {  // found the node to be deleted
            BinaryTree left = tree.getLeft();
            BinaryTree right = tree.getRight();
            
            if(left == null)
                tree = tree.getRight();
            else if(right == null)
                tree = tree.getLeft();
            else {
                Key tmp = getLargest(left);
                tree.setElement(tmp);
                tree.setLeft(deleteLargest(left));
            }
        } 
        else if(k.lt(key))
            tree.setLeft(deleteHelp(k, tree.getLeft()));
        else
            tree.setRight(deleteHelp(k, tree.getRight()));
        return tree;
    }
    
    /**
     * Deletes the largest item in a subtree.
     * @param tree the subtree currently being examined
     */
    protected BinaryTree deleteLargest(BinaryTree tree) {
        if(tree.getRight() == null) {
            return tree.getLeft();
        } 
        else {
            tree.setRight(deleteLargest(tree.getRight()));
            return tree;
        }
    }
    
    
    /*Methods from Tree interface*/
    
    public Tree[] getSubTrees() {
        return ((BinTree)root.getObject()).getSubTrees();
    }
    
    public int getSubTreeCount() {
        if (((BinTree)root.getObject()).getLeft() != null || ((BinTree)root.getObject()).getRight() != null || ((BinTree)root.getObject()).getElement() != null)
            return 2;
        else
            return 0;
    }
    
    
    
    /**
     * Returns the height of a subtree
     * @param tree the subtree currently being examined
     * @return the largest height
     */
    protected int getLargestHeight(BinaryTree tree) {
        int right=0;
        int left=0;
        
        // Parse left subtree
        if(tree.getRight() != null)
            right = getLargestHeight(tree.getRight());
        
        // Parse right subtree
        if(tree.getLeft() != null)
            left = getLargestHeight(tree.getLeft());
        
        // Return largest
        return (right>left)?(right+1):(left+1);
    }
    
    /**
     * Returns the largest height of the tree
     * @return the largest height
     */
    public int getHeight() {
        return getLargestHeight((BinTree)getElement());
    }
    
    
    /**
     * Returns the largest key in a subtree
     * @param tree the subtree currently being examined
     * @return the largest key
     */
    protected Key getLargest(BinaryTree tree) {
        if(tree.getRight() == null)
            return (Key)tree.getElement();
        else
            return getLargest(tree.getRight());
    }
    
    public Object search(Object o) {
        return this;
    }
    
    public String inOrder() {
        return Traverse.traverseInOrder((BinaryTree)getElement());
    }
    
    public String levelOrder() {
        return Traverse.traverseLevelOrder((BinaryTree)getElement());
    }
    
    public String postOrder() {
        return Traverse.traversePostOrder((BinaryTree)getElement());
    }
    
    public String preOrder() {
        return Traverse.traversePreOrder((BinaryTree)getElement());
    }
    
    static final long serialVersionUID = 490933325492185329L;

    public void setSubTree(Tree t, int child) {
        // TODO Auto-generated method stub
        
    }
}
