/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: DigitalSearchTree.java,v 1.8 2006/05/26 11:05:13 vkaravir Exp $ */

package matrix.structures.CDT.probe;

import java.util.BitSet;

import matrix.structures.CDT.CDT;
import matrix.structures.FDT.*;
import matrix.structures.FDT.probe.*;
import matrix.structures.other.Comparer;
import matrix.util.KeyBuilder;

/**
 * A digital search tree. In a digital search tree keys are inserted into the
 * tree according to their bit representation. The class uses the services of
 * <code>KeyBuilder</code> -class to convert <code>String</code> keys
 * into a format it can do bit comparisons to. The keys are turned into
 * <code>BitSet</code>s that are compared bit-by-bit. Consult the documentation
 * of the <code>KeyBuilder</code> -class to find out what sort of keys should be used
 *
 * @author Jussi Nikander
 *
 */


public class DigitalSearchTree implements /*FDT,*/ BinaryTree, CDT, Tree {

  /** root of the digital search tree */
  protected LabeledBinTree root;

  /** a special head node that is root's parent */
  private LabeledBinTree head;

  /** Tells if the tree is empty or not */
  private boolean isEmpty;

  /**
   * Constructs a new <code>DigitalSearchTree</code>
   */
  public DigitalSearchTree() {
    root = new LabeledBinTree();
    head = new LabeledBinTree();
    isEmpty = true;
  }

  public CDT getNewInstance(){
    return new DigitalSearchTree();
  }

  
    /**
     * Constructs a new <code>DigitalSearchTree</code>
     * @param obj the contents of the root node
     * @param key the key to the root node
     */
    public DigitalSearchTree(Object obj, String key) {
        root = new LabeledBinTree(obj, key);
        head = new LabeledBinTree(null, "o");
        isEmpty = false;
    }

    public boolean equals(Object o) {
	if(o instanceof BinaryTree) {
	    if ((root==o)||(this==o))
		return true;
	    else
		return Comparer.compareIdentical((BinaryTree)root,
						 (BinaryTree)o);
	} else
	    return false;
    }

  /**
   * Creates a new <code>LabeledBinTree</code> and inserts it into the tree
   *
   * @param obj The <code>Object</code> that will be the contents of the node
   * @param key  <code>String</code> representation of the node's key
   */
    public void insert(Object obj, String key) {

        // if the tree was empty, new node is automatically made root

        if(isEmpty()) {
            root = new LabeledBinTree(obj, key);
            head.setRight(root);
            isEmpty = false;
            return;
        }

        LabeledBinTree parent = head;
        LabeledBinTree current = root;

        /* it is assumed that every different key gives an unique digital key, which
         * might not be true, depending on how KeyBuilder works.
         * Consult the documentation of matrix.util.KeyBuilder to find out the legal
         * values of keys.
         */

        BitSet digitalKey = KeyBuilder.getDigitalKey(key);
        int counter = 0;
        boolean right = true;

        while(counter < digitalKey.size()) {
            if(current == null) {
                current = new LabeledBinTree(obj, key);
                if(right) {
                    parent.setRight(current);
                } else {
                    parent.setLeft(current);
                }
                return;
            } else if(obj.equals(current.getContents())){
                return;
            } else if(digitalKey.get(counter++)) {
                parent = current;
                current = (LabeledBinTree) current.getRight();
                right = true;
            } else {
                parent = current;
                current = (LabeledBinTree) current.getLeft();
                right = false;
            }
        }
        return;
    }



    /**
     * Searches the digital search tree for a node with a certain key
     */
    public Object search(String key) {
        LabeledBinTree tmp = searchNode(key);
        return tmp.getContents();
    }

  /**
   * Searches the digital search tree for a node which has a certain key
   *
   * @param key the key to be searched
   * @return The <code>LabeledBinTree</code> containing the key
   */
  public LabeledBinTree searchNode(String key) {

    BitSet digitalKey = KeyBuilder.getDigitalKey(key);
    LabeledBinTree current = root;
    int counter = 0;

    while(counter < digitalKey.size()) {
        if(current == null) {
            return current;
        } else {
            if(key.equals(current.getKey())) {
                return current;
            } else {
                if(digitalKey.get(counter++)) {
                    current = (LabeledBinTree) current.getRight();
                } else {
                    current = (LabeledBinTree) current.getLeft();
                }
            }
        }
    }

    return current;
  }

  /**
   * deletes the node with a certain key from the tree
   * @param key the key of the node to be deleted
   */
    public void delete(String key) {

        if(isEmpty()) {
            return;
        }

        BitSet digitalKey = KeyBuilder.getDigitalKey(key);
        LabeledBinTree current = root;
        LabeledBinTree parent = head;
        int counter = 0;
        boolean right = true;

        while(counter < digitalKey.size()) {
            if(current == null) {
                return;
            } else if(key.equals(current.getKey())) {
                break;
            } else {
                parent = current;
                if(digitalKey.get(counter++)) {
                    current = (LabeledBinTree) current.getRight();
                    right = true;
                } else {
                    current = (LabeledBinTree) current.getLeft();
                    right = false;
                }
            }
        }
        if(current == null) {
            return;

            // at this point, the node which is to be deleted has been found
            // rest of this method does the actual delete -operation

        } else {
            if(current.getRight() == null && current.getLeft() == null) {
                if(right) {
                    parent.setRight(null);
                } else {
                    parent.setLeft(null);
                }
                if(parent == head) {
                    isEmpty = true;
                    root = new LabeledBinTree(null, null);
                }
                return;
            } else
                deletehelp(current);
        }
    }

    /**
     * Tells if the tree is empty or not.
     * @return <code>true</code> if empty <code>false</code> if not empty
     */
    public boolean isEmpty() {
        return root.getElement() == null;
    }


    /**
     * Returns the root node of the tree.
     * @return the root <code>LabeledBinTree</code>
     */
    public LabeledBinTree getRoot() {
        return root;
    }

    // Methods for interface FDT


    /**
     * Returns the contents of the root node.
     * @return contents of the root node
     */
    public Object getElement() {
        //return root.getElement();
	return root;

    }

    /**
     * Sets the contents of the root node.
     * @param o the new contents
     */
    public void setElement(Object o) {
	//root.setElement(o);
	root = (LabeledBinTree)o;
    }

    // Methods for interface Tree

    /**
     * Counts the number of children of the root node.
     * @return the number of children
     */
    public int getSubTreeCount() {
        return root.getSubTreeCount();
    }

    /**
     * Returns the children of the root node.
     * @return a <code>Tree</code> array containing the children of the root
     */
    public Tree[] getSubTrees() {
        return root.getSubTrees();
    }

    // Methods for interface BinaryTree

    /**
     * Returns the right child of root.
     * @return the right child
     */
    public BinaryTree getRight() {
        return root.getRight();
    }

    /**
     * Returns the left child of root.
     * @return the left child
     */
    public BinaryTree getLeft() {
        return root.getLeft();
    }

    /**
     * Sets the left child of the root.
     * @param p the new left child
     */
    public void setLeft(BinaryTree p) {
        root.setLeft(p);
    }

    /**
     * Sets the right child of the root.
     * @param p the new right child
     */
    public void setRight(BinaryTree p) {
        root.setRight(p);
    }

    /**
     * Tells if the root is a leaf node.
     * @return true if root is a leaf, false otherwise
     */
    public boolean isLeaf() {
        return root.isLeaf();
    }

    // Methods for interface CDT

    /**
     * Inserts an object into the tree.
     * @param o the object to be inserted
     */
    public CDT insert(Object o) {
        if(o instanceof Table) {
            Table t = (Table)o;
            for(int i = 0; i < t.size(); i++)
                insert(t.getObject(i));
            return this;
        }
        String key = o.toString();
        insert(o, key);
        return this;
    }

    /**
     * Deletes an object from the tree.
     * @param o the object to be deleted
     */
    public CDT delete(Object o) {
        String key = o.toString();
        delete(key);
        return this;
    }

    /**
     * Does nothing. This method is needed in the implementation
     * of the interface CDT.
     */
    public Object search(Object o) {
        return this;
    }

    // Methods for interface SimulationTree

    public void setSubTree(Tree t, int child) {
        root.setSubTree(t, child);
    }

    public Tree getNewNode(Object key) {
        return root.getNewNode(key);
    }

    private void deletehelp(LabeledBinTree tr) {

        Object contents;
        String key;

        if(tr.getLeft() != null) {
            contents = tr.getLeft().getElement();
            key = ((LabeledBinTree) tr.getLeft()).getKey();
            tr.setContents(contents, key);
            if(tr.getLeft().isLeaf())
                tr.setLeft(null);
            else
                deletehelp((LabeledBinTree)tr.getLeft());
        } else if(tr.getRight() != null) {
            contents = tr.getRight().getElement();
            key = ((LabeledBinTree) tr.getRight()).getKey();
            tr.setContents(contents, key);
            if(tr.getRight().isLeaf())
                tr.setRight(null);
            else
                deletehelp((LabeledBinTree)tr.getRight());
        } else {
            tr = null;
        }
        return;
    }



    static final long serialVersionUID = -5868442859609297671L;
}
