/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: LinkedListImplCDT.java,v 1.9 2007/09/03 13:40:30 oseppala Exp $ */

package matrix.structures.CDT.probe;

import matrix.structures.CDT.CDT;
import matrix.structures.FDT.*;
import matrix.structures.FDT.probe.*;
import matrix.structures.memory.*;
import matrix.structures.util.MatrixComparable;

/**
 * Implementation of a linked list. This linked list is implemented so that
 * a node of linked list can be deleted by calling delete for a key. That 
 * is a reason why this is a CDT structure.
 *  
 */
public class LinkedListImplCDT extends LinkedListImpl implements CDT, DoublyLinkedList {
    

    /** previous node in the linked list */
    private VirtualObject prev;


    /**
     * Constructs LinkedListImplCDT.
     *
     */
    public LinkedListImplCDT() {
		super();
		prev = new VirtualObject(this, "link to previous node");
    }

    public CDT getNewInstance(){
        return new LinkedListImplCDT();
    }

    
    /**
     * Constructs LinkedListImplCDT with given key.
     *
     * @param key the key of the new LinkedListImplCDT
     */
    public LinkedListImplCDT(Object key) {
		super();
		prev = new VirtualObject(this, "link to previous node");
		setElement(key);
    }

    /**
     * Returns new (LinkedList) node with a given key. 
     *
     * @param key key of new node
     * @return new (LinkedList) Node
     */    
    public LinkedList getNewNode(Object key) {
		return new LinkedListImplCDT(key);
    }


    /**
     * Inserts a new key to the beginning of the linked list.
     *
     * @param key new key
     * @return linked list after insertion
     */
    public CDT insert(Object key) {
		if (key == null)
			return getFirstOfList();
		if (key instanceof Key)
			if (((Key)key).toString() == "")
				return getFirstOfList();
		if (key instanceof MatrixComparable) {
			LinkedListImplCDT newNode = new LinkedListImplCDT();
			newNode.setElement(key);
			setFirstOfList(newNode);
			return getFirstOfList();
		}
		if (key instanceof Table) {
			Table t = (Table)key;
			for (int i = 0; i < t.size(); i++)
				insert(t.getObject(i));
			return getFirstOfList();
		}
		LinkedListImplCDT newNode = new LinkedListImplCDT();
		newNode.setElement(key);
		setFirstOfList(newNode);
		return getFirstOfList();
    }



    /**
     * Sets the given node to the beginning of the linked list.
     *
     * @param newNode new (LinkedList) node to be added to the
     * beginning of the linked list.
     */
    private void setFirstOfList(LinkedListImplCDT newNode) {
		if (getElement() == null && hasPrev() == false) {
	    	setElement(newNode.getElement());
		}
		else if (hasPrev() == false) {
	    	newNode.setNext(this);	    
	    	setPrev(newNode);
		}
		else {
	    	((LinkedListImplCDT)getPrev()).setFirstOfList(newNode);
		}
    }

    /**
     * Returns the first node of the linked list
     *
     * @return node from the beginning of the linked list
     */
    private LinkedListImplCDT getFirstOfList() {
		if (hasPrev() == false){
	    	return this;}
		else
	    	return ((LinkedListImplCDT)getPrev()).getFirstOfList();
    }


    /**
     * This method deletes the node containing the given key
     * from the linked list and returns the first node of 
     * linked list after deletion.
     *
     * @param o key to be deleted
     * @return returns always the first node of the linked list after deletion
     */

	public CDT delete(Object o) {
		if (hasPrev() == false && getNext() == null) { //only one node in the list
			if (o instanceof Key) {
				if(getElement() == ((Key)o).getOrig())
					setElement(null);
				else if (getElement() == o)
					setElement(null);
			}
			else {
				if (getElement() == o) {
				setElement(null);
				}
			}
			return this;
		}
		else if (hasPrev() == false) { //first node of the list
			LinkedListImplCDT temp = (LinkedListImplCDT)getNext();
			if (o instanceof Key) {
				if (getElement() == ((Key)o).getOrig()) {
					setNext(null);
					temp.setPrev(null);
					return temp;
				}
				else if (getElement() == o) {
					setNext(null);
					temp.setPrev(null);
					return temp;
				}
			}
			else {
				if (getElement() == o) {
					setNext(null);
					temp.setPrev(null);
					return temp;
				}
			}
			return temp.delete(o);
		}
		else if (getNext() == null) {  //last node of the list
			LinkedListImplCDT temp = (LinkedListImplCDT)getPrev();
			if (o instanceof Key) {
				if (getElement() == ((Key)o).getOrig()) {
					setPrev(null);
					temp.setNext(null);
					return temp.getFirstOfList();
				}
				else if (getElement() == o) {
					setPrev(null);
					temp.setNext(null);
					return temp.getFirstOfList();
				}
			}
			else {
				if (getElement() == o) {
					setPrev(null);
					temp.setNext(null);
					return temp.getFirstOfList();
				}
			}
			return getFirstOfList();
		}
		else { //node in the middle of a list
			LinkedListImplCDT next = (LinkedListImplCDT)getNext();
			LinkedListImplCDT prev = (LinkedListImplCDT)getPrev();
			if (o instanceof Key) {
				if (getElement() == ((Key)o).getOrig()) {
					getPrev().setNext(next);
					next.setPrev(prev);
					setNext(null);
					setPrev(null);
					return prev.getFirstOfList();
				}
				else if (getElement() == o) {
					getPrev().setNext(next);
					next.setPrev(prev);
					setNext(null);
					setPrev(null);
					return prev.getFirstOfList();
				}	
			}
			else {
				if (getElement() == o) {
					getPrev().setNext(next);
					next.setPrev(prev);
					setNext(null);
					setPrev(null);
					return prev.getFirstOfList();
				}
			}
		return next.delete(o);
		}
	}



    /**
     * Sets the previous node.
     *
     * @param o previous node
     */
    public void setPrev(DoublyLinkedList o) {
		prev.setObject(o);
    }


    /**
     * Returns the previous node.
     *
     * @return the previous node
     */
    public DoublyLinkedList getPrev() {
		return (DoublyLinkedList)prev.getObject();
    }


    /**
     * Tests whether there is a previous node.
     *
     * @return true, if there is a previous node
     */
    public boolean hasPrev() {
		return (getPrev() != null);
    } 


    public Object search(Object o) {
		return getFirstOfList();
    }



	static final long serialVersionUID = 7796907732463706271L;
}
