/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: QueueImpl.java,v 1.12 2007/09/05 10:34:06 oseppala Exp $ */

package matrix.structures.CDT.probe;

import matrix.structures.CDT.CDT;
import matrix.structures.FDT.*;
import matrix.structures.FDT.probe.*;
import matrix.structures.memory.*;
import matrix.structures.util.MatrixComparable;

/**
 * Implementation of a queue. This queue is implemented by using LinkedlistImpl.
 * This queue contains always a reference to the LinkedListImpl which is on the
 * top of the stack.
 *
 * @author kstalnac
 */
public class QueueImpl extends LinkedListImpl implements CDT, matrix.structures.ADT.Queue{
    

    /**First node of the queue. This has to be VirtualObject because otherwise 
     * animation doesn't work. */
    private VirtualObject firstNode;
    

    /**
     * Constructor. Constructs QueueImpl.
     *
     */
    public QueueImpl() {
        super();
        firstNode = new VirtualObject(new LinkedListImpl(), this, "first node in the queue");
    }

    
    
    
    /**
     * Constructs QueueImpl with given key.
     *
     * @param key the key of the new QueueImpl
     */
    public QueueImpl(Object key) {
        super();
        firstNode = new VirtualObject(new LinkedListImpl(), this, "first node in the queue");
        setElement(key);
    }

    public CDT getNewInstance(){
        return new QueueImpl();
    }
    
    
    /**
     * Sets the element of the firstLinkedListImpl of the queue
     * @see LinkedListImpl
     * @param o element to be set on the first of the queue
     */  
    public void setElement(Object o) {
        ((FDT) firstNode.getObject()).setElement(o);
    }
    
    /**
     * Returns the element from the first LinkedListImpl of the queue.
     * @see LinkedListImpl
     * @return element from the first LinkedListImpl of the queue
     */    
    public Object getElement() {
        return ((FDT) firstNode.getObject()).getElement();
    }
    
    /**
     * Returns the next LinkedListImpl from the first LinkedListImpl.
     * @see LinkedListImpl.
     * @return next LinkedListImpl
     */    
    public LinkedList getNext() {
        return ((LinkedList) firstNode.getObject()).getNext();
    }
    
    /**
     * Returns true if there is a next LinkedListImpl.
     * @see LinkedListImpl
     * @return true if next LinkedListImpl exists
     */   
    public boolean hasNext() {
        return ((LinkedList) firstNode.getObject()).hasNext();
    }

    /**
     * Sets the next LinkedListImpl of the first LinkedListImpl
     * @param o LinkedListImpl to be set to the next of the first LinkedListImpl
     */
    public void setNext(LinkedList o) {
        ((LinkedListImpl)firstNode.getObject()).setNext(o);
    }
    
    /**
     * Inserts a new key to the last position of the queue and returns the queue
     * after insertion. 
     *
     * @param key new key
     * @return queue after insertion
     */
    public CDT insert(Object key) {
        if (key == null)
            return this;
        if (key instanceof Key && ((Key)key).toString() == "")
            return this;
            
        if (((FDT) firstNode.getObject()).getElement() == null) { //Queue is empty
            ((FDT) firstNode.getObject()).setElement(key);
            return this;
        }
 
        if (key instanceof MatrixComparable) {
            LinkedListImpl newNode = new LinkedListImpl();
            newNode.setElement(key);
            setLastOfQueue(newNode);
            return this;
        }
        if (key instanceof Table) {
            Table t = (Table)key;

            for(int i = 0; i < t.size(); i++)
                insert(t.getObject(i));
            return this;
        }

        LinkedListImpl newNode = new LinkedListImpl();
        newNode.setElement(key);
        setLastOfQueue(newNode);
        return this;
    }


    /**
     * Sets the given node to the last position of the queue.
     *
     * @param newNode new (LinkedListImp) node to be added to the last position of 
     * the queue
     */
    protected void setLastOfQueue(LinkedListImpl newNode) {
        if (getElement() == null && hasNext() == false) { //queue is empty
            setElement(newNode.getElement());
        } else if (hasNext() == false) { 
            setNext(newNode);
        } else {
            LinkedList tmpNode = getNext();
            while (tmpNode.hasNext()) {
                tmpNode = tmpNode.getNext();
            }
            tmpNode.setNext(newNode);
        }
    }

    /**
     * Returns the first node of the queue
     *
     * @return node the first node of the queue
     */
    protected QueueImpl getFirstOfQueue() {
        return this;
    }
    
    /**
     * This method deletes the node containing the given key and all
     * its predecessors from the queue and returns 
     * the first node of queue after deletion.
     *
     * @param o key to be deleted
     * @return returns always the first node of the queue after 
     * deletion.
     */
    public CDT delete(Object o) {
        boolean isInStack = false;
        LinkedListImpl temp = (LinkedListImpl)this.firstNode.getObject();
        while(temp != null) {
            if(o instanceof Key) {
                if(temp.getElement() == ((Key)o).getOrig())
                    isInStack = true;
                else if(temp.getElement() == o)
                    isInStack = true;
            }
            else {
                if(temp.getElement() == o)
                    isInStack = true;
            }
            temp = (LinkedListImpl)temp.getNext();
        }
        if(isInStack)  //key to be removed is in the queue
            return dequeue(o);
        return getFirstOfQueue();
    }


    /**
     * This method dequeues the node containing the given key and all its
     * predecessors from the queue and returns the first
     * node of the queue after dequeue operation.
     *
     * @param o key to be dequeued from the queue
     * @return first node of the queue after dequeue operation
     */
    private CDT dequeue(Object o) {
        if (getNext() == null) { //only one node in the queue
            setElement(null);
            return this;
        } else { //this node is the first node
            LinkedListImpl temp = (LinkedListImpl) getNext();
            setNext(null);
            if (o instanceof Key) {
                if (getElement() == ((Key) o).getOrig()) {
                    firstNode.setObject(temp);
                    return this;
                } else if (getElement() == o) {
                    firstNode.setObject(temp);
                    return this;
                } else {
                    firstNode.setObject(temp);
                    return dequeue(o);
                }
            } else {
                if (getElement() == o) {
                    firstNode.setObject(temp);
                    return this;
                } else {
                    firstNode.setObject(temp);
                    return dequeue(o);
                }
            }
        }
    }        
    //----------------------------------------------------------//
    
    /**
     * This method puts the given element to the end of the queue.
     * @param element element to be put to the end of the queue
     */
    public void ADTEnqueue(Object element) {
        insert(element);
    }
    
    /**
     * This method gets (removes) the element from the beginning of the queue 
     * and returns it.
     * @return object removed from the beginning of the queue, or null if queue
     * is already empty
     */
    public Object ADTDequeue(){
        if (!isEmpty()) {
            Object o = getFirstOfQueue().getElement();
            delete(o);
            
            //Keys must be unwrapped - ask Ari about this
            if (o instanceof Key)
                return ((Key)o).getObject();
            return o;
        }
        return null;
    }
    
    /**
     * This method returns the first element of the queue (without removing it).
     * @return first element of the queue, or null if queue is empty
     */
    public boolean ADTIsEmpty() {
        return isEmpty();
    }
    
    public Object ADTPeek(){
        Object o = getFirstOfQueue().getElement();
        
        //Keys must be unwrapped - ask Ari about this
        if (o instanceof Key)
            return ((Key)o).getObject();
        return o;
    }    
    
    public Object search(Object o) {
        return getFirstOfQueue();
    }

    /**
     * Tests whether the queue is empty or not.
     * 
     * @return true, if the queue is empty
     */

    public boolean isEmpty() {
    	if(getNext() == null && getElement() == null) {
    		return true;
    	} else {
    		return false;
    	}
    }
    
    static final long serialVersionUID = -8291102470177338558L;
}
