/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


package matrix.structures.CDT.probe;
import java.awt.Color;

import matrix.decoration.*;
import matrix.structures.FDT.*;
import matrix.structures.FDT.probe.*;
import matrix.structures.memory.*;
import matrix.structures.other.Rotations;
import matrix.structures.util.MatrixComparable;
import matrix.util.Note;

/**
 * <p>Implementation of the RedBlackNode in RedBlackTree.
 *
 * @see RBTree
 * @author Petri Tenhunen
 */

public class RedBlackNode extends BinTree implements Styled {

    //TODO This should probably be VirtualBoolean
    private VirtualInteger red = new VirtualInteger(this, "color (red=1 black=0)");
    private VirtualObject parent;

    private StyleSheet myStyle = null;

    private boolean isColorChangeEnabled = false;

    /** default constructor for empty tree */
    public RedBlackNode() {
        super();
        parent = new VirtualObject(this, "parent of the node");
        red.assign(1);
    }


    /** Constructor for tree with one element */
    public RedBlackNode(MatrixComparable e) {
        super(e);
        parent = new VirtualObject(this, "parent of the node");
        red.assign(1);
    }

    /**
     * Tells if a tree is empty or not
     * @return true if the tree is empty, false otherwise
     */
    public boolean empty() {
        return getElement() == null;
    }


    public StyleSheet getStyleSheet() {
        if (myStyle != null)
            return myStyle;
        else {
            myStyle = new StyleSheetAdapter() {
                    public Color getDefaultBorderColor() {
                        return (empty() || !isRed()) ? Color.black: Color.red;
                    }
                    public Color getDefaultBackgroundColor() {
                        return (isRed()) ? Color.red : Color.black;
                    }
                    public Color getFocusedBackgroundColor() {
                        return (empty() || !isRed()) ? Color.black: Color.red;
                    }
                    public Color getFocusedBorderColor() {
                        return (empty() || !isRed()) ? Color.red : Color.black;
                    }

                    public Color getVisitedFocusedBackgroundColor() {
                        return (empty() || !isRed()) ? Color.black: Color.red;
                    }
                    public Color getVisitedBackgroundColor() {
                        return (empty() || !isRed()) ? Color.black: Color.red;
                    }

                    
                    static final long serialVersionUID = -1653377563227996332L;
                };
            return myStyle;
        }
    }

    public FDT insert(Object o) {
        if (o == null)
            return this;
        //if (o instanceof Table) {         //recursive insertion to Tables, but seems to be dead code since multiple insertions still work...
	//Table t = (Table)o;
        //    RedBlackNode r = this; //to store new roots
        //    for (int i=0; i<t.size(); i++)
        //        r = (RedBlackNode)r.insert(t.getObject(i));
        //    return r;
        //} else 
	if (o instanceof MatrixComparable) {
            if (empty()) {                       //inserted element is only element
                setElement((MatrixComparable)o);
                this.setRed(false);
                return this;
            }
            else {
                BinaryTree node=this, prev=this;
                do {
                    prev = node;
                    if (((MatrixComparable)node.getElement()).gt((MatrixComparable)o))
                        node = node.getLeft();
                    else
                        node = node.getRight();
                } while (node != null);  //new leaf will now be prev's child
                //node = new RedBlackNode((Element)o);
		node = (RedBlackNode)getNewNode((MatrixComparable)o);
                ((RedBlackNode)node).setParent((RedBlackNode)prev);
                if (((MatrixComparable)prev.getElement()).gt((MatrixComparable)o)) {
                    prev.setLeft(node);
                }else {
                    prev.setRight(node);
                }
                //((RedBlackNode)node).head = this.head;

                return rotateUp((RedBlackNode)node);
            }
        } else Note.show(this, "Cannot insert object of type " + o.getClass());
        return this;
    }


    /**
     * Searches for the node corresponding to the object given as a parameter
     * and deletes it in a way that preserves the red-black properties of 
     * the tree. 
     * 
     * @param o The object to be deleted.
     * @return The possibly new root node of the tree.
     */
    public RedBlackNode delete(Object o) {
	RedBlackNode nodex, nodey, nodez, root = this;

    	if (o == null)
	    return this; // Nothing to delete
	if (o instanceof MatrixComparable) {
    		// Find the node that is to be deleted
    		nodez = this.search((MatrixComparable)o);
    		if (nodez == null) {
    			// No node found corresponding to element o
    			return this;
    		} else {
    			// Note.out(this, "Deleting [<RedBlackNode> " + nodez + "] corresponding to [<Element> " + o + "]");
    			if ((nodez.getLeft() == null) || (nodez.getRight() == null)) {
    				nodey = nodez;
    			} else {
    				nodey = nodez.treeSuccessor();
    			}
   				if (nodey.getLeft() != null) {
   					nodex = (RedBlackNode)nodey.getLeft();
    			} else {
    				nodex = (RedBlackNode)nodey.getRight();
    			}
    			if (nodex != null) {
    				nodex.setParent(nodey.getParent());
    			}
    			if (nodey.getParent() == null) {
    				root = nodex;
    			} else {
    				if (nodey == nodey.getParent().getLeft()) {
    					nodey.getParent().setLeft(nodex);
    				} else {
    					nodey.getParent().setRight(nodex);
    				}
    			}
    			if (nodey != nodez) {
    				nodez.setElement(nodey.getElement());
    			}
    			// If node y is red, the red black properties still hold
    			if (nodey.isBlack()) {
    				// Restore the red-black tree properties
    				if (root != null) {
    					root = root.deleteFixup(nodex, nodey.getParent());
    				}
    			}
    		}
    	} else {
    		Note.show(this, "Cannot delete object of type " + o.getClass());
    	}
    	return root;
    }
    

    /**
     * Restores the red-black properties of the tree after deletion.
     * This method is called from 
     * <code>public RedBlackNode delete(Object o)</code>.
     * 1. Every node is either black or red.
     * 2. The root is black.
     * 3. Every leaf (null) is black.
     * 4. Red nodes do not have red children.
     * 5. For each node, all paths from the node to descendant leaves
     * contain the same number of black nodes.
     * 
     * @param nodex The node passed from the delete-algorithm.
     * @return The possibly new root node of the tree.
     */
    private RedBlackNode deleteFixup(RedBlackNode nodex, RedBlackNode parentx) {
    	RedBlackNode nodew, tmp, root = this;
		// Null nodes, i.e. leaf nodes are black
    	while ((nodex != root) && ((nodex == null) || (nodex.isBlack()))) {
    		if (nodex == parentx.getLeft()) {
    			nodew = (RedBlackNode)parentx.getRight();
				if (nodew.isRed()) {
					nodew.setRed(false);
					parentx.setRed(true);
					tmp = parentx.leftRotate();
					if (tmp != null) {
						root = tmp;
					}
					nodew = (RedBlackNode)parentx.getRight();
				}
				if (((nodew.getLeft() == null) || ((RedBlackNode)nodew.getLeft()).isBlack()) &&
					((nodew.getRight() == null) || (((RedBlackNode)nodew.getRight()).isBlack()))) {
					nodew.setRed(true);
					nodex = parentx;
					parentx = nodex.getParent();
				} else {
					if ((nodew.getRight() == null) || (((RedBlackNode)nodew.getRight()).isBlack())) {
						((RedBlackNode)nodew.getLeft()).setRed(false);
						nodew.setRed(true);
						tmp = nodew.rightRotate();
						if (tmp != null) {
							root = tmp;
						}
						nodew = (RedBlackNode)parentx.getRight();
					}
					nodew.setRed(parentx.isRed());
					parentx.setRed(false);
					((RedBlackNode)nodew.getRight()).setRed(false);
					tmp = parentx.leftRotate();
					if (tmp != null) {
						root = tmp;
					}
					nodex = root;
				}
    		} else {
    			nodew = (RedBlackNode)parentx.getLeft();
				if ((nodew != null) && nodew.isRed()) {
					nodew.setRed(false);
					parentx.setRed(true);
					tmp = parentx.rightRotate();
					if (tmp != null) {
						root = tmp;
					}
					nodew = (RedBlackNode)parentx.getLeft();
				}
				if (((nodew.getRight() == null) || ((RedBlackNode)nodew.getRight()).isBlack()) &&
					((nodew.getLeft() == null) || (((RedBlackNode)nodew.getLeft()).isBlack()))) {
					nodew.setRed(true);
					nodex = parentx;
					parentx = nodex.getParent();
				} else {
					if ((nodew.getLeft() == null) || (((RedBlackNode)nodew.getLeft()).isBlack())) {
						((RedBlackNode)nodew.getRight()).setRed(false);
						nodew.setRed(true);
						tmp = nodew.leftRotate();
						if (tmp != null) {
							root = tmp;
						}
						nodew = (RedBlackNode)parentx.getLeft();
					}
					nodew.setRed(parentx.isRed());
					parentx.setRed(false);
					((RedBlackNode)nodew.getLeft()).setRed(false);
					tmp = parentx.rightRotate();
					if (tmp != null) {
						root = tmp;
					}
					nodex = root;
        		}
    		}
    	}
    	nodex.setRed(false);
    	return root;
    }
    

    /**
     * Executes a left rotation on the node.
     * Preserves the binary-search-tree property.
     * This method is called from
     * <code>private RedBlackNode deleteFixup(RedBlackNode nodex, RedBlackNode parentx)</code>.
	 * <pre>
	 *       A              C
	 *      / \     =>     / \
	 *     B   C          A   E
	 *        / \        / \
	 *       D   E      B   D
	 * </pre></p>
     * 
     * @return Returns the new root node of the tree or null if unchanged.
     */
    private RedBlackNode leftRotate() {
    	RedBlackNode rChild, root = null;
    	rChild = (RedBlackNode)this.getRight();
    	// If there is no right child, the rotation cannot be done
    	if ((rChild != null) && (rChild.getElement() != null)) {
    		this.setRight(rChild.getLeft());
    		if (rChild.getLeft() != null) {
    			((RedBlackNode)rChild.getLeft()).setParent(this);
    		}
    		rChild.setParent(this.getParent());
    		rChild.setLeft(this);
        	if (this.getParent() == null) {
    			root = rChild;
    		} else {
    			if (this == this.getParent().getLeft()) {
    				this.getParent().setLeft(rChild);
    			} else {
    				this.getParent().setRight(rChild);
    			}
    		}
    		this.setParent(rChild);
    	}
    	return root;
    }
    

    /**
     * Executes a right rotation on the node.
     * Preserves the binary-search-tree property.
     * This method is called from
     * <code>private RedBlackNode deleteFixup(RedBlackNode nodex, RedBlackNode parentx)</code>.
	 * <pre>
	 *       A              B
	 *      / \     =>     / \
	 *     B   C          D   A
	 *    / \                / \
	 *   D   E              E   C
	 * </pre></p>
     * 
     * @return Returns the new root node of the tree or null if unchanged.
     */
    private RedBlackNode rightRotate() {
    	RedBlackNode lChild, root = null;
    	lChild = (RedBlackNode)this.getLeft();
    	// If there is no left child, the rotation cannot be done
    	if ((lChild != null) && (lChild.getElement() != null)) {
    		this.setLeft(lChild.getRight());
    		if (lChild.getRight() != null) {
    			((RedBlackNode)lChild.getRight()).setParent(this);
    		}
    		lChild.setParent(this.getParent());
    		lChild.setRight(this);
        	if (this.getParent() == null) {
        		root = lChild;
        	} else {
        		if (this == this.getParent().getRight()) {
        			this.getParent().setRight(lChild);
        		} else {
        			this.getParent().setLeft(lChild);
        		}
        	}
        	this.setParent(lChild);
    	}
    	return root;
    }

    
    /**
     * Finds the successor of the node.
     * 
     * @return The successor node or null if none exist.
     */
    private RedBlackNode treeSuccessor() {
    	if (this.getRight() != null) {
    		// The successor is the minimum of the right subtree
    		return ((RedBlackNode)this.getRight()).treeMinimum();
    	} else {
    		// The successor is the first ancestor, whose left subtree contains the node
    		RedBlackNode parent = this.getParent();
    		RedBlackNode child = this;
    		while ((parent != null) && 
    			   (child == parent.getRight())) {
    			child = parent;
    			parent = parent.getParent();
    		}
    		// Returns the successor or null if search failed
    		return parent;
    	}
    }
    

    /**
     * Finds the minimum node of the subtree, whose root the node is.
     * 
     * @return The minimum node.
     */
    private RedBlackNode treeMinimum() {
    	if ((this.getLeft() == null)) {
    		// The minimum node was found
    		return this;
    	} else {
    		return ((RedBlackNode)this.getLeft()).treeMinimum();
    	}
    }
    

    /**
     * Searches for a node containing the given element.
     * 
     * @param e The key to find.
     * @return The node corresponding to the key or null if not found or 
     * tree contains an erroneous node with no key value.
     */
    private RedBlackNode search(MatrixComparable e) {
    	if ((MatrixComparable)this.getElement() == null) {
    		// An incorrect node with no key was encountered
    		return null;
    	} else if (((MatrixComparable)this.getElement()).equals(e)) { 
    		// The node containing the element was found
    		return this;
    	} else  {
    		if (((MatrixComparable)this.getElement()).gt(e)) {
    			if (this.getLeft() == null) {
    				// Search failed
    				return null;
    			} else {
    				// Search the left subtree
    				return ((RedBlackNode)this.getLeft()).search(e);
    			}
    		} else {
    			if (this.getRight() == null) {
    				// Search failed
    				return null;
    			} else {
    				// Search the right subtree
    				return ((RedBlackNode)this.getRight()).search(e);
    			}
    		}
    	}
    }
    

    public Tree getNewNode(Object o) {
	return new RedBlackNode((MatrixComparable)o);
    }

    /**
     * Rotations for bottom up insertion. Normal binary search tree
     * insertion must applied first. This method is called from
     * <code>public FDT insert(Object o)</code>.
     *
     * @param x newly inserted leaf before balansing.
     * @return Root of the RedBlackNode after all rotations are performed.
     */
    public  RedBlackNode rotateUp(RedBlackNode x) {
        x.setRed(true);
        //Note.out(this,"Starting rotations");
        while ( x.getParent() != null && x.getParent().isRed() && x.isRed() ) {  //(not root) and red
            //Note.out(this,"in loop");
            //Note.out(this,"x: "+x);
            if (  x.getParent() == x.getParent().getParent().getLeft() ) { //K1
                //Note.out(this,"left branch");
                RedBlackNode y = (RedBlackNode)x.getParent().getParent().getRight();
                RedBlackNode pp = x.getParent().getParent();
                RedBlackNode ppp = pp.getParent(); //parent of rotated subtree
                RedBlackNode tmp;

                if ( y != null && y.isRed()) { //only swap colors
                    //Note.out(this,"color swap");
                    x.getParent().setRed(false);
                    y.setRed(false);
                    x.getParent().getParent().setRed(true);
                    x = x.getParent().getParent();


                } else if ( x == x.getParent().getRight()) {     //zig-zag
                    //Note.out(this,"zig-zag");
                    x.setRed(false);
                    x.getParent().getParent().setRed(true);
                    x.newParentsZigZagRight();
                    tmp = (RedBlackNode)Rotations.doubleWithLeftChild(pp);
                    if (ppp == null) {
                        return tmp;
                    } else if (ppp.getLeft() == pp) {
                        ppp.setLeft(tmp);
                    } else {
                        ppp.setRight(tmp);
                    }

                } else {
                    //Note.out(this,"zig");
                    x.getParent().getParent().setRed(true);
                    x.getParent().setRed(false);
                    x.newParentsZigRight();
                    tmp = (RedBlackNode) Rotations.withLeftChild(pp);

                    if (ppp == null) {
                        tmp.setRed(false);
                        tmp.setParent(null);
                        return tmp;
                    } else if (ppp.getRight() == pp) {
                        ppp.setRight(tmp);
                    } else {
                        ppp.setLeft(tmp);
                    }
                    tmp.setParent(ppp);
                }

            } else {
                //Note.out(this,"right branch");
                RedBlackNode y = (RedBlackNode)x.getParent().getParent().getLeft();
                RedBlackNode pp = x.getParent().getParent();
                RedBlackNode ppp = pp.getParent(); //parent of rotated subtree
                RedBlackNode tmp;

                if ( y != null && y.isRed()) { //only swap colors
                    //Note.out(this,"color swap");
                    x.getParent().setRed(false);
                    y.setRed(false);
                    x.getParent().getParent().setRed(true);
                    x = x.getParent().getParent();


                } else if ( x == x.getParent().getLeft()) {     //zig-zag
                    //Note.out(this,"zig-zag");
                    x.setRed(false);
                    x.getParent().getParent().setRed(true);
                    x.newParentsZigZagLeft();
                    tmp = (RedBlackNode)Rotations.doubleWithRightChild(pp);
                    if (ppp == null) {
                        tmp.setParent(null);
                        return tmp;
                    } else if (ppp.getLeft() == pp) {
                        ppp.setLeft(tmp);
                    } else {
                        ppp.setRight(tmp);
                    }

                } else {
                    //Note.out(this,"zig");
                    x.getParent().getParent().setRed(true);
                    x.getParent().setRed(false);
                    x.newParentsZigLeft();
                    tmp = (RedBlackNode) Rotations.withRightChild(pp);
                    if (ppp == null) {
                        tmp.setRed(false);
                        tmp.setParent(null);
                        return tmp;
                    } else if (ppp.getRight() == pp) {
                        ppp.setRight(tmp);
                    }else {
                        ppp.setLeft(tmp);
                    }
                    tmp.setParent(ppp);

                }
            }

        } // end while (x is not root and red)
        //Note.out(this,"new root is "+getRoot().getElement().toString());
        x.getRoot().setRed(false);    //set root to black
        return x.getRoot();
    }

    /**
     * This method is called before right-rotation is performed. If rotation is performed
     * to node <code>n</code> the method is called to node <code>n.getLeft()<code>
     */
    protected void newParentsZigRight() {
        this.getParent().getParent().setParent(this.getParent());
        if (this.getParent().getRight() != null)
            ((RedBlackNode)this.getParent().getRight()).setParent(this.getParent().getParent());
        this.getParent().setParent(this.getParent().getParent().getParent());
    }

    /**
     * This method is called before left-rotation is performed. If rotation is performed
     * to node <code>n</code> the method is called to node <code>n.getRight()<code>
     */
    protected void newParentsZigLeft() {
        this.getParent().getParent().setParent(this.getParent());
        if (this.getParent().getLeft() != null)
            ((RedBlackNode)this.getParent().getLeft()).setParent(this.getParent().getParent());
        this.getParent().setParent(this.getParent().getParent().getParent());
    }

    /**
     * This method is called before double-rotation to right is performed.
     * If rotation is performed to node <code>n</code> the method is called to node
     * <code>n.getLeftt().getRight()<code>
     */
    protected void newParentsZigZagRight() {
        RedBlackNode a = this.getParent();
        RedBlackNode b = a.getParent();
        if (this.getRight() != null)
            ((RedBlackNode)this.getRight()).setParent(b);
        if (this.getLeft() != null)
            ((RedBlackNode)this.getLeft()).setParent(a);
        this.setParent(b.getParent());
        a.setParent(this);
        b.setParent(this);
    }

    /**
     * This method is called before double-rotation to left is performed.
     * If rotation is performed to node <code>n</code> the method is called to node
     * <code>n.getRight().getLeft()<code>
     */
    protected void  newParentsZigZagLeft() {
        RedBlackNode a = this.getParent();
        RedBlackNode b = a.getParent();
        if (this.getRight() != null)
            ((RedBlackNode)this.getRight()).setParent(a);
        if (this.getLeft() != null)
            ((RedBlackNode)this.getLeft()).setParent(b);
        this.setParent(b.getParent());
        a.setParent(this);
        b.setParent(this);
    }

    /** debug */
    private String printAllParents() {
        String tmp = getElement() +" : "+ getParent() +"\n";
        if (getRight() != null)
            tmp += ((RedBlackNode)getRight()).printAllParents();
        if (getLeft() != null)
            tmp += ((RedBlackNode)getLeft()).printAllParents();
        return tmp;
    }




    /**
     * Returns root node of this redBlack-tree.
     *
     * @return root node from tree where current node is acting as subtree
     */
    public RedBlackNode getRoot() {
        RedBlackNode root = this;
        //return (RedBlackNode)head.getRight();
        while(root.getParent() != null)
            root = root.getParent();
        return root;
    }


    /**
     * Gets father of this subtree
     *
     * @return parent of this subtree or null, if this is the root.
     **/
    public RedBlackNode getParent() {
        return (RedBlackNode)this.parent.getObject();
    }


    /**
     * Set father information.
     *
     * @param p father for this subtree
     */
    public void setParent(RedBlackNode p) {
        this.parent.setObject(p);
    }


    /**
     * Set color information.
     *
     * @param red If true color is set to red. If false the color is
     *            set to black
     */
    public void setRed(boolean red) {
        if (red)
            this.red.assign(1);
        else
            this.red.assign(0);
    }


    /**
     * Color of the node.
     *
     * @return true if node is red
     */
    public boolean isRed() { return red.eval()==1; }

    public boolean isBlack() { 
    	return (red.eval() == 0); 
    }
    
    /** Color the whole tree with param c. 
     * Known usage: ExerRBTree
     */
    public void colorTree(Color c) {
	if (c != Color.black) 
	    setRed(true);
	else
	    setRed(false);
	if (getLeft() != null)
	    ((RedBlackNode)getLeft()).colorTree(c);
	if (getRight() != null)
	    ((RedBlackNode)getRight()).colorTree(c);
    }

    public void setSelect(long timeStamp) {
    	//Note.out(this, "select this node " + this);
    	if (isColorChangeEnabled())
    		setRed(!isRed());
    }

    /** This method sets the flag that indicates whether the select method actually
     * can change the color of this node. 
     */
    public void setColorChangeEnabled(boolean value) {
	isColorChangeEnabled = value;
    }

    public boolean isColorChangeEnabled() {
	return isColorChangeEnabled;
    }

    
    /**
     * validity[0] tells if root is valid
     * validity[1] tells if each path has equal number of black nodes
     * validity[2] tells if any red node has red children 
     */
    private boolean[] validity;
    
    /**
     * 
     * 
     * @return 
     */
    public boolean[] isValid() {
        validity = new boolean[3];
        validity[0] = validity[1] = validity[2] = true;
        // Test if the root node is black as expected.
    	if (isRed()) {
    	    validity[0] = false;
        }
    	//Note.out(this, "invoke left");
    	int left = isValidTree((RedBlackNode)getLeft());
    	/*
        if (left == -1) {
    	    validity[1] = false;
        }
        */
    	    
            //Note.out(this, "invoke right");
    	if (left != isValidTree((RedBlackNode)getRight())) {
         validity[1] = false;   
        }
        return validity;
    }
    
    /** return the number of black nodes. */
    private int isValidTree(RedBlackNode n) {
    	if (n == null) {
    	    return 1;
    	}
    	int left = isValidTree((RedBlackNode)n.getLeft());
    	int right = isValidTree((RedBlackNode)n.getRight());
        
    	if (/*(left == -1) ||*/ (left != right)) {
    	    //Note.out(this, "error (left != right) in " + n);
    	    //Note.out(this, "left = " + left);
    	    //Note.out(this, "right = " + right);
    	    validity[1] = false;
    	    //return -1;
        }
    	if (n.isRed()) {
    	    if (n.getLeft() != null) {
    	        if (((RedBlackNode)n.getLeft()).isRed()) {
    	            //Note.out(this, "error (left) in " + n);
    	            validity[2] = false;
                    //return -1;
    	        }
    	    }
    	    if (n.getRight() != null) {
    	        if (((RedBlackNode)n.getRight()).isRed()) {
    	            //Note.out(this, "error (right) in " + n);
    	            validity[2] = false;
                    //return -1;
    	        }
    	    }
    	    return left;
    	} else {
    	    return 1 + left;
    	}
    }

    public String toString() {
        return empty() ? "empty" : getElement().toString() + ", isRed: "+isRed(); }


    public boolean equals(Object o) {
        //System.out.println(""+this+" joo "+o);
      //Note.out(this,"equals");
        if (o instanceof RedBlackNode)
            return (compareIdentical(this, (RedBlackNode)o));
        return false;
    }

    protected static boolean compareIdentical(Tree a, Tree b) {

        if(a == b) {
          //Note.out(null,"case 1");
          return true;
        }
        if(a == null || b == null) return false;
        Object aObj = a.getElement();
        boolean aRed = ((RedBlackNode)a).isRed();
        Object bObj = b.getElement();
        boolean bRed = ((RedBlackNode)b).isRed();
        if(aObj == bObj && aObj == null) {
          //Note.out(null,"case 2");
          return true;
        }
        if(aObj == null || bObj == null) {
          //Note.out(null,"case 2");
            return false;
        }
        if(((!(aObj.equals(bObj))) && (!(bObj.equals(aObj)))) || (aRed != bRed) )
            return false;
        Tree[] aTree = a.getSubTrees();
        Tree[] bTree = b.getSubTrees();
        if(aTree.length != bTree.length)
            return false;
        for(int i = 0; i < aTree.length; i++)
            if(!compareIdentical(aTree[i], bTree[i]))
                return false;
        return true;
    }

    private boolean compareChilds(RedBlackNode other) {
        boolean right = false, left = false;

        if ( getRight() != null )
            right = getRight().equals(other.getRight());
        else if (other.getRight() == null)
            right = true;

        if ( getLeft() != null )
            left = getLeft().equals(other.getLeft());
        else if (other.getLeft() == null)
            left = true;

        return (left && right);
    }


    static final long serialVersionUID = -1653377563227996332L;
}
