package matrix.structures.FDT.probe;

import matrix.structures.memory.*;
import java.util.Vector;
import java.util.HashMap;
import matrix.structures.FDT.indexed.StaticGraph;
import matrix.structures.coordinates.*;
import matrix.decoration.StaticInformationDecorator;

/** 
 * Implementation of the <code>StaticGraph</code> interface where 
 * the vertices have coordinates. 
 * 
 * @author jtn
 *
 */
public class GeometricGraph implements StaticGraph, StaticXYCoordinates, StaticInformationDecorator {
	
	private VirtualArray keys;
	private VirtualArray xCoordinates;
	private VirtualArray yCoordinates;
	private VirtualArray edges;
	private VirtualInteger nroVertices;
	private VirtualInteger nroEdges;
	private Vector information;
	private HashMap referenceInformation;
	private boolean toolTipAvailable;
	
	public GeometricGraph(int vertices) {
		xCoordinates = new VirtualArray(this, "x-coordinates of the nodes");
		xCoordinates.setObject(null, vertices);
		yCoordinates = new VirtualArray(this, "y-coordinates of the nodes");
		yCoordinates.setObject(null, vertices);
		keys = new VirtualArray(this, "keys of the nodes");
		keys.setObject(null, vertices);
		this.edges = new VirtualArray(this, "edges in the graph");
		nroVertices = new VirtualInteger(vertices, this, "number of nodes");
		nroEdges = new VirtualInteger(0, this, "number of edges");
	}
	
	public GeometricGraph(int vertices, boolean tooltips) {
		this(vertices);
		if(tooltips) {
			toolTipAvailable = true;
			information = new Vector();
			referenceInformation = new HashMap();
		}
	}
	
	public GeometricGraph() {
	 this(0);
	}
	
	/* --- get methods --- */
	
	/** Returns the number of vertices in the graph. */
	public int getNrOfNodes() {
		return nroVertices.eval();
	}
	
	/** Gets the successors of the parameter vertex */
	public int[] getSuccessors(int i) {
		Vector v = new Vector(20, 10);
		int s = nroEdges.eval();
		for(int j = 0; j < s; j++) {
			GeometricEdge e = (GeometricEdge)edges.getObject(j);
			if(e != null && e.getSource() == i) {
				v.add(e);
			}
		}
		int[] successors = new int[v.size()];
		s = v.size();
		for(int j = 0; j < s; j++) {
			GeometricEdge e = (GeometricEdge)v.get(j);
			successors[j] = e.getDestination();
		}
		return successors;
	}
	
	/** Gets the X coordinate of the parameter vertex */
	
	public int getX(int i) {
 	    Integer ii = (Integer)xCoordinates.getObject(i);
	    if(ii == null) return 0;
	    return ii.intValue();
	}
	
	/** Gets the Y coordinate of the parameter vertex */
	
	public int getY(int i) {
 	    Integer ii = (Integer)yCoordinates.getObject(i);
	    if(ii == null) return 0;
	    return ii.intValue();

	}
	
	/* --- Set methods --- */
	
	/** Adds a new vertex to the graph. Returns the int id of the vertex. */
	
	public int addVertex() {
		int nro = nroVertices.eval();
		xCoordinates.setObject(new Integer(0), nro);
		yCoordinates.setObject(new Integer(0), nro);
		keys.setObject("", nro);
		nroVertices.inc();
		return nro;
	}
	
	/** Removes the parameter vertex from the graph. The id of all
	 * vertices with id larger than the parameter is decreased by one.
	 */
	
	public void removeVertex(int i) {
		int nro = nroVertices.eval();
		for(; i < nro; i++) {
			int x = getX(i+1);
			int y = getY(i+1);
			setCoordinates(i, x, y);
			Object key = keys.getObject(i+1);
			keys.setObject(key, i);
		}
		// XXX Now we have to go through the edge list and decrease the ids there!
		nroVertices.dec();		
	}
	
	/** Sets the coordinates of the parameter vertex. */
	
	public void setCoordinates(int vertex, int x, int y) {
		xCoordinates.setObject(new Integer(x), vertex);
		yCoordinates.setObject(new Integer(y), vertex);
	}
	
	public void setCoordinates(int id, float x, float y) {
		setCoordinates(id, Math.round(x), Math.round(y));
	}
	
	/** Adds a successor to the vertex. */
	
	public void addSuccessor(int vertex, int successor) {
		GeometricEdge e = new GeometricEdge(vertex, successor);
		edges.setObject(e, nroEdges.eval());
		nroEdges.inc();
	}
	
	/** Removes a successor from the parameter vertex. */
	
	public void removeSuccessor(int vertex, int successor) {
		boolean found = false;
		int size = edges.size();
		for(int i = 0; i < size; i++) {
			GeometricEdge e = (GeometricEdge)edges.getObject(i);
			if(!found && ((e.getSource()== vertex && e.getDestination() == successor) ||
					(e.getSource() == successor && e.getDestination() == vertex))) {
				found = true;
				continue;
			}
			if(found) {
				edges.setObject(edges.getObject(i),i-1);
			}
		}
		if(found) {
			nroEdges.dec();
		}
	}
    
	/* --- Methods inhereted from StaticGraph --- */
	
    public Object getElement() { return null; }

    public void setElement(Object o) {}

    public void setKey(int index, Object o) {
    	keys.setObject(o, index);
    }

    public Object getKey(int i) { 
    	Object ret = keys.getObject(i); 
    	if(ret == null) {
    		return "";
    	}
    	return ret;
    }

    static final long serialVersionUID = -6279427956655022649L;

	public boolean isInfoBoxEnabled() {
		return false;
	}

	public boolean isToolTipInfoEnabled() {
		return toolTipAvailable;
	}

	public void setInfo(int i, String s) {
		if(toolTipAvailable) {
			information.set(i, s);
		}
	}

	public String getInfo(int i) {
		if(toolTipAvailable && i < information.size() && i < nroVertices.eval()) {
			return (String)information.get(i);
		}
		return null;
	}

	public int getWaitTime() {
		return 0;
	}

	public void setWaitTime(int t) {
	}

	public boolean isReferenceInfoBoxEnabled() {
		return false;
	}

	public boolean isReferenceToolTipInfoEnabled() {
		return toolTipAvailable;
	}

	public String getReferenceInformation(int from, int refNro) {
		if(toolTipAvailable) {
			int[] succs = getSuccessors(from);
			if(refNro < succs.length) { 
				String key = "" + from + refNro;
				return (String)referenceInformation.get(key);
			}
		}
		return null;
	}

	public void setReferenceInformation(int from, int refNro, String s) {
		if(toolTipAvailable) {
			String key = ""+ from + refNro;
			referenceInformation.put(key, s);
		}		
	}
}

