/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: Key.java,v 1.2 2006/10/04 14:38:43 jtn Exp $ */

package matrix.structures.FDT.probe;

import java.io.IOException;
import java.io.ObjectInputStream;

import matrix.structures.CDT.interfaces.PrioritySupplied;
import matrix.structures.FDT.FDT;
import matrix.structures.util.MatrixComparable;
import matrix.util.Note;

public class Key implements PrioritySupplied, MatrixComparable, FDT
{
    
    /*
     * Identity, comparison and visualization strings for Key
     * ======================================================
     * 
     * It is not uncommon to have the same value exist multiple times
     * in the same structure. In some cases we need to identify these
     * keys as two different keys, in some cases they should be the same.
     * 
     * To solve this problem, three strings representing different
     * aspects of the key are given:
     * 
     *  + comparisonString
     *     This string should give the string in a form that cannot
     *     separate between identical keys. This can be used by
     *     different algorithms that need to do comparisons.
     *
     *          returned by : getComparisonString() and toString() [for compatibility with other structures]
     *
     *  + visualizationString
     *     This String is the one used to visualize the key. This String can be decorated
     *     with additional charaters.
     * 
     *          returned by : getVisualizationString()
     * 
     *  + identityString
     *     Two keys have the same identityString only if they "point" to the same
     *     data. Normally all identityStrings are different. The String is formed
     *     with a the cmpString and a numeric id.
     * 
     *     This string can for example be used to identify the exact same key that
     *     was used in an operation or to check whether elements were changed
     *     correctly in stable algorithms.
     * 
     *          returned by : getIdentityString()
    */
    
    private String comparisonString; //Formerly known as "key"
    private String identityString;
    private String visualizationString;
    
    private Object object;
    private Key orig = null;    
    
    // The following field should not be static. it is serialized with
    // the other data and can be used to recognize the version for deserialization
    // When the fields of this class change, change this value accordingly
    private final int CLASS_VERSION = 2;
    
    //Unique ids for keys
    private static int idCounter = 0;
    private int id;
    
    public static final String EMPTY = "";
    public static final Key Empty = new Key(EMPTY);
    
    private Key(){
    	Note.err(this, "This constructor should never be called");
    }
    
    public Key(Object o, String s){
        object = o;  
        id     = idCounter++;
        
    	comparisonString    = s;
        setOtherStrings();        
    }

    //this method is called also by the deserialization code
    private void setOtherStrings(){
        visualizationString = comparisonString;
        identityString      = comparisonString + "-" + id;        
    }
        
    public Key(Object o) {
        this((Object)o, ((String)((o != null)? o.toString(): EMPTY)));
    }

    public Key(String k) {
        this((Object)null, ((String)((k != null) ? k : EMPTY)));
    }

    public Key(int i) {
        this((Object)null, ((String)(String.valueOf(i))));
    }

    public Key getOrig() {
        return orig;
    }

    public Object getObject() { return object; }

    public void setOrig(Key orig) {
        this.orig = orig;
    }

    /* This method is invoked by VisualType each time this key is selected. */
    public void select() {
    }
    
    public boolean isSelected() {
        return false;
    }

    // operational interface (CDT) (not actually defined)
    public boolean equals(Key k) {
    	if (comparisonString == null)
            return false;
            
        return eq(k);
    }
    
    public boolean equals(Object k) {
        if (k instanceof Key)
            return equals((Key)k);
        else if (k instanceof String) {
            if ((comparisonString == null || comparisonString.equals(EMPTY)) &&
            	(k   == null              || k.equals(EMPTY)))
                return true;
            return comparisonString.equals((String)k);
        } else return false;
    }


    
    //Comparison String
    public String getComparisonString() {
        return comparisonString;
    }
    
    public void setComparisonString(String s) {
    	comparisonString = s;
    }

    //Identity String
    public String getIdentityString(){
    	return identityString;
    }
    
    //Visualization String
    public String getVisualizationString(){
    	return visualizationString;
    }

    public void setVisualizationString(String s) {
    	visualizationString = s;
    }
    
    public String toString() {
        return comparisonString; // Debugging code : "[" +id+" - "+key+ "]";
    }

  
    //Deserialization code
    private void readObject(ObjectInputStream s) throws IOException{
        java.io.ObjectInputStream.GetField fields = null;
        try{
            fields = s.readFields();
        }
        catch(Exception ClassNotFoundException){
        	Note.err(this, "Problems deserializing Keys");
            throw new IOException();
        }
        
        //These two fields are in all versions
        object = fields.get("object",null);
        orig   = (Key)fields.get("orig", "null");

        
        //if there is no field named "key", this is a new version 
        if (testForField(fields, "key") == false){
            id = fields.get("id",0);
            
            comparisonString    = (String)fields.get("comparisonString",    "unknown"); 
            identityString      = (String)fields.get("identityString",      "unknown");
            visualizationString = (String)fields.get("visualizationString", "unknown");
            
            if (fields.get("CLASS_VERSION",0) > CLASS_VERSION)
                Note.show(null, "Trying to de-serialize a newer serialization version in class Key");
        }
        else{
            id = fields.get("id",0);
            
            comparisonString    = (String)fields.get("key", "unknown"); 
            setOtherStrings();
        }
        
        if (id >= idCounter)
            idCounter = id+1;
        
    }
    
    private boolean testForField(java.io.ObjectInputStream.GetField fields, String fieldName) throws IOException{
    	try{
            fields.get(fieldName, "unknown");
            return true;
        }
        catch(IllegalArgumentException e){}
        return false;
    }
    
    // === DEPRECATED METHODS ====================================================
    private final String getStr() { return comparisonString; }          //DEPRECATED
    private final void setKey(Key k) { comparisonString = k.getStr(); } //DEPRECATED
    
    public void changeKeyValue(String newValue){
    	comparisonString    = newValue;
    	setOtherStrings();
    }
    
    // FDT interface
    
    public void setElement(Object o) {
        Note.err(this,
                 "Element: Cannot perform setElement("+o+") for primitive");
    }

    public Object getElement() {
        return this;
    }


    // PrioritySupplied interface
    //
    // for normal Keys the priority defaults to the comparison string
    //
    // Note : Priorities are compared as Strings, not numbers!!
    
    public void setPriorityElement(MatrixComparable e){
        changeKeyValue(e.getComparisonString());
    }

    public MatrixComparable getPriorityElement(){
        return this;
    }

    
    //Comparable interface
    /* (non-Javadoc)
     * @see matrix.structures.memory.Comparable#eq(matrix.structures.memory.Element)
     */
    public boolean eq(MatrixComparable k) {
        if (k == null)
            return (this.getComparisonString().equals(""));
        if(k instanceof LexInf)
            return false;
        return (this.getComparisonString().equals(k.getComparisonString()));
    }

    /* (non-Javadoc)
     * @see matrix.structures.memory.Comparable#lt(matrix.structures.memory.Element)
     */
    public boolean lt(MatrixComparable k) {
        if (k == null)
            return false;
        if(k instanceof LexInf)
            return true;
        return (this.getComparisonString().compareTo(k.getComparisonString()) < 0);
    }

    /* (non-Javadoc)
     * @see matrix.structures.memory.Comparable#gt(matrix.structures.memory.Element)
     */
    public boolean gt(MatrixComparable k) {
        if (k == null)
            return (this.getComparisonString().equals(""));
        if(k instanceof LexInf)
            return false;
        return (this.getComparisonString().compareTo(k.getComparisonString()) > 0);
    }

    /* (non-Javadoc)
     * @see matrix.structures.memory.Comparable#leq(matrix.structures.memory.Element)
     */
    public boolean leq(MatrixComparable k) {
        if(k instanceof LexInf)
            return true;
        return (this.eq(k) || this.lt(k));
    }

    /* (non-Javadoc)
     * @see matrix.structures.memory.Comparable#geq(matrix.structures.memory.Element)
     */
    public boolean geq(MatrixComparable k) {
        if(k instanceof LexInf)
            return false;
        return (this.eq(k) || this.gt(k));
    }
    
    static final long serialVersionUID = 2497751812742138576L;
}
