/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: VanillaTable.java,v 1.18 2006/12/12 14:11:09 oseppala Exp $ */

package matrix.structures.FDT.probe;

import matrix.structures.FDT.FDT;
import matrix.structures.memory.*;
import matrix.structures.other.Comparer;
import matrix.structures.util.MatrixComparable;
import matrix.util.RandomKey;
import matrix.structures.FDT.Array;

/**
 * A table of Objects that can be visualised as an Array in Matrix. It is possible
 * to create a "random table", which will randomize its keys each time it is
 * refreshed from the GUI. The random table uses two-character keys.
 */
public class VanillaTable implements FDT, Array {

    //    public static final int random = 0;
    protected boolean rnd = false;
    private VirtualArray contents;
    
    public void setElement(Object o){
        contents.setObject(o,contents.getFirst());
        contents.setLast(contents.getFirst());
    }
    
    public Object getElement(){
        return this;
    }
    
    /**
     * Constructs a new Table from the given <code>String</code>. Each char in
     * the string is stored in separate index of the <code>Table</code>.
     * @param s the contens of the Table.
     */
    public VanillaTable(String s, int zero) {
        contents = new VirtualArray(zero, this, "[contents of the table]");
        
        for (int i=0; i<s.length(); i++)
            contents.setObject(new Key(s.substring(i,i+1)), i+zero);
    }

    public VanillaTable(String s) { 
        this(s, 0); 
    }

    /**
     * Constructs a new <code>Table</code> from a given <code>String</code> array.
     * Each string in the array is stored in a separate index of the <code>Table</code>.
     * @param s the contents of the Table.
     */
    public VanillaTable(String[] s, int zero) {
        contents = new VirtualArray(zero, this, "[contents of the table]");
        
        for(int i=0; i < s.length; i++)
            contents.setObject(new Key(s[i]), i+zero);
    }

    public VanillaTable(String[] s) {
        this(s, 0);
    }

    public VanillaTable(MatrixComparable[] e, int zero) {
        contents = new VirtualArray(zero, this, "[contents of the table]");
        for(int i=0; i < e.length; i++)
            contents.setObject(e[i], i+zero);
    }

    public VanillaTable(MatrixComparable[] e) {
        this(e, 0);
    }

    public VanillaTable(Object[] o, int zero) {
        contents = new VirtualArray(zero, this, "[contents of the table]");
        for(int i=0; i < o.length; i++)
            contents.setObject(o[i], i+zero);
    }

    public VanillaTable(Object[] o) {
        this(o, 0);
    }

    
    public VanillaTable(int[] elements) {
       contents = new VirtualArray(this, "[contents of the table]");
       rnd = true;
       for (int i = 0; i < elements.length; i++) {
          contents.setObject(new Key(elements[i]+""), i);
       }
    }

    public VanillaTable(){
        contents = new VirtualArray(this, "[contents of the table]");
    }
    
    /**
     * Constructs a new <code>Table</code> with random two-character keys.
     * @param size the size of the table
     */
    public VanillaTable(int size) {
        contents = new VirtualArray(this, "[contents of the table]");
        
        rnd = true;
        java.util.Random r = new java.util.Random(System.currentTimeMillis());
        if(size < 1)
            size = 1;
   
        for(int i = 0; i < size; i++)
            //setObject(new Key(RandomKey.createRandomKey(r)), i);
            contents.setObject(new Key(RandomKey.createRandomKeyWithLetterAndNumber(r)), i);
    }

    /**
     * Tells if this table random table or not.
     * @return true if table is random.
     */
    public boolean isRandomTable() {
        return rnd;
    }

    /**
     * Randomizes the contents of this table.
     */
    public void randomize() {
	matrix.util.Note.out(this, "RANDOMIZE");
	java.util.Random r = new java.util.Random(System.currentTimeMillis());
        for(int i = contents.getFirst(); i <= contents.getLast(); i++)
	    //this.setObject(new Key(RandomKey.createRandomKey(r)), i);
		contents.setObject(new Key(RandomKey.createRandomKeyWithLetterAndNumber(r)), i);
    }

    // FDT interface Methods ????
    /*
    public FDT getNewStructure() {
        return new Table("");
    }
    */

    static final long serialVersionUID = -6755658672874936167L;

    public Array getArray() {
        return this;
    }

    public void swap(int i, int j){
        contents.swap(i, j);
    }
    
    public Object getObject(int address) {
        
        return contents.getObject(address);
    }

    public void setObject(Object o, int address) {
        contents.setObject(o,address);
    }

    public int getFirst() {
        return contents.getFirst();
    }

    public int getLast() {
        return contents.getLast();
    }

    public void setFirst(int i) {
        contents.setFirst(i);
    }

    public void setLast(int i) {
        contents.setLast(i);
    }
    
    public int size(){
        return getLast() - getFirst() + 1;
    }

    public boolean equals(Object o) {
        
        if(o instanceof Array) {
            return equals((Array)o);
        }
        return false;
    }

    public boolean equals(Array a) {
        return Comparer.compareArray(this,a);
    }

    public String toString(){
        return "[VanillaTable +"+contents.toString()+"]";
    }
    
}
