
/**
 * VertexImpl -objects are used as vertices of DirectedGraphImpl. VertexImpl implements the
 * styled interface in order for graph traversing algorithms to work.
 *
 * @author Jussi Nikander
 */
package matrix.structures.FDT.probe;

import java.awt.Color;

import matrix.decoration.*;
import matrix.structures.FDT.*;
import matrix.structures.FDT.substructures.Vertex;
import matrix.structures.memory.*;
import matrix.util.Note;

public class VertexImpl implements Vertex, CommonTree, Tree, Styled, LabelDecorator {
    
    /* successors of this vertex. successor[0] is the key of this vertex */
    
    private VirtualArray successors;
    private VertexStyleSheet ss = null;
    private VirtualObject parent;
    
    private VirtualArray weights;
    
    public VertexImpl() {
        successors = new VirtualArray(this, "Successors of the node");
        successors.setObject("", 0);
        parent = new VirtualObject(this, "Parent of the node");
        weights = new VirtualArray(this, "Weights of the edges");
        weights.setObject("1", 0);
    }
    
    public VertexImpl(Object o) {
        successors = new VirtualArray(this, "Successors of the node");
        successors.setObject(o, 0);
        parent = new VirtualObject(this, "Parent of the node");
        weights = new VirtualArray(this, "Weights of the edges");
        weights.setObject("1", 0);
    }
    
    public VertexImpl getNewVertex(Object o) {
        return new VertexImpl(o);
    }
    
    
    /** Inserts the param vertex to the list of successor vertices */
    public void addSuccessor(Vertex v) {
        addSuccessor(v, successors.size());
    }
    
    private VertexStyleSheet getVertexStyleSheet() {
        return (VertexStyleSheet)getStyleSheet();
    }
    
    /** Private method for actually adding successors or child trees to
     * the graph/tree.
     * Should not be called directly, use addSuccessor(Vertex) or
     * setSubTree(Tree, int) instead
     */
    private void addSuccessor(Vertex v, int pos) {
        int i = 1;
        // HS: Fixed bug. Invariant Pos was Last+1 when exiting Loop
        for(; i <= pos && successors.getObject(i) != null; i++) {
            if(v == successors.getObject(i))
                return;
            /*
             if(successors.getObject(i) == null)
             break;
             */
        }
        if(v instanceof VertexImpl) {
            successors.setObject(v, i);
        } else {
            VertexImpl tmp = getNewVertex(v.getElement());
            successors.setObject(tmp, i);
        }
        getVertexStyleSheet().addSuccessor(i);
        weights.setObject("1", i);
    }
    
    /** Removes the param vertex from the list of successor vertices. */
    public void removeSuccessor(Vertex v) {
        //Note.out(this, "removing successor " + v);
        int i;
        Vertex tmp;
        for(i = 1; i <= successors.size(); i++) {
            tmp = (Vertex)successors.getObject(i);
            if(tmp != null && tmp.equals(v))
                break;
        }
        getVertexStyleSheet().removeSuccessor(i);
        //Note.out(this, "Successor located, is successor nro. " + i);
        for(; i < successors.size(); i++) {
            successors.setObject(successors.getObject(i + 1), i);
            weights.setObject(weights.getObject(i + 1), i);
        }
        successors.setLast(i - 2);
        //Note.out(this, "Successor removed");
        
    }
    
    // interface Vertex
    
    public Vertex[] getSuccessors() {
        Vertex[] tmp = new Vertex[successors.size() - 1];
        for(int i = 1 /*successors.getFirst()*/ ; i < successors.size(); i++) {
            tmp[i - 1] = (Vertex) successors.getObject(i);
        }
        return tmp;
    }
    
    public void setSuccessors(Vertex[] v) {
        int i = 0;
        for(; i < v.length; i++) {
            successors.setObject(v[i], i + 1);
            weights.setObject("1", i + 1);
        }
        for(; i < successors.size(); i++) {
            successors.setObject(null, i + 1);
            weights.setObject(null, i + 1);
        }
    }
    
    public Object getElement() {
        return successors.getObject(0);
    }
    
    public void setElement(Object o) {
        successors.setObject(o, 0);
    }
    
    // interface Styled
    
    public StyleSheet getStyleSheet() {
        if(ss != null)
            return ss;
        else {
            ss = new VertexStyleSheet();
            return ss;
        }
    }
    
    public void setColor(Color c) {
        getStyleSheet().setDefaultBackgroundColor(c);
    }
    
    public Color getColor() {
        return getStyleSheet().getDefaultBackgroundColor();
    }
    
    /** Sets the parent Vertex of this Vertex.
     * TODO : Currently only an instance of VertexImpl can be the
     * parent; XXX change to interface?
     * Sets the color of the reference leading from parent to this
     * to be black in color, changes the color of the reference
     * leading from old parent (if present) back to blue
     */
    public void setParent(VertexImpl p) {
        if(parent.getObject() != null) {
            VertexImpl old = getParent();
            old.getStyleSheet().setReferenceColor(this, Color.blue);
        }
        if(p != null)
            p.getStyleSheet().setReferenceColor(this, Color.red);
    }
    
    public VertexImpl getParent() {
        Object o = parent.getObject();
        if(o instanceof VertexImpl)
            return (VertexImpl)o;
        return null;
    }
    
    // interface Tree
    
    public int getSubTreeCount() {
        return successors.size() - 1;
    }
    
    public Tree[] getSubTrees() {
        Tree[] subtrees = new Tree[successors.size() - 1];
        for(int i = 1; i < successors.size(); i++) {
            subtrees[i - 1] = (Tree) successors.getObject(i);
        }
        return subtrees;
    }
    
    // interface SimulationTree
    
    public void setSubTree(Tree t, int child) {
        if(t instanceof Vertex) {
            //successors.setObject(t, child+1);
            addSuccessor((Vertex)t, child);
            weights.setObject("1", child);
        }
    }
    
    public Tree getNewNode(Object key) {
        return getNewVertex(key);
    }
    
    class VertexStyleSheet extends StyleSheetAdapter {
        /** Color information */
        private VirtualObject colors;
        /** Labeling information */
        private VirtualObject label;
        private VirtualArray successorColors;
        private VirtualArray focusedSuccessors;
        private VirtualArray modifiedSuccessors;
        private VirtualBoolean isLabelEnabled;
        
        VertexStyleSheet() {
            super();
            colors = new VirtualObject(VertexImpl.this, "Node background color");
            label = new VirtualObject(VertexImpl.this, "Node label");
            successorColors = new VirtualArray(VertexImpl.this, "Successor colors");
            focusedSuccessors = new VirtualArray(VertexImpl.this, "Focused Successors");
            modifiedSuccessors = new VirtualArray(VertexImpl.this, "Modified successors");
            isLabelEnabled = new VirtualBoolean(false, VertexImpl.this, "Node's label is enabled");
        }
        
        public Color getDefaultBackgroundColor() {
            if(colors.getObject() == null)
                colors.setObject(Color.white);
            return (Color) colors.getObject();
        }
        
        public void setDefaultBackgroundColor(Color c) {
            colors.setObject(c);
        }
        
        public void setLabel(String s) {
            label.setObject(s);
        }
        
        public boolean isLabelEnabled() {
            return isLabelEnabled.eval();
        }
        
        public void setLabelEnabled(boolean flag) {
            isLabelEnabled.assign(flag);
        }
        
        public String getLabel() {
            return (String) label.getObject();
        }
        
        private int getReferenceNumber(FDT f) {
            for(int i = 1; i <= successors.getLast(); i++) {
                if(f == successors.getObject(i))
                    return i;
            }
            return Integer.MIN_VALUE;
        }
        
        private void setRefCol(FDT f, Color c, VirtualArray a) {
            if(c != null && f != null) {
                int ref = getReferenceNumber(f);
                if(ref != Integer.MIN_VALUE)
                    a.setObject(c, ref);
                else
                    Note.err(this, "Successor not found; reference color not changed");
            } else {
                Note.err(this, "Null reference, color not changed.");
            }
        }
        
        private Color getRefCol(FDT f, VirtualArray a) {
            if(f != null && a != null) {
                int ref = getReferenceNumber(f);
                if(ref != Integer.MIN_VALUE) {
                    Color col = (Color)a.getObject(ref);
                    //Note.out(this, "retrieved color " + col + " for reference " + ref);
                    return col;
                } else {
                    Note.err(this, "Cannot get color for MIN_VALUE reference " + ref);
                    return Color.pink;
                }
            } else {
                Note.err(this, "Cannot get color for null successor");
                return Color.pink;
            }
        }
        
        public void setReferenceColor(FDT f, Color c) {
            setRefCol(f, c, successorColors);
        }
        
        public void setFocusedReferenceColor(FDT f, Color c) {
            setRefCol(f, c, focusedSuccessors);
        }
        
        public void setModifiedReferenceColor(FDT f, Color c) {
            setRefCol(f, c, modifiedSuccessors);
        }
        
        public Color getReferenceColor(FDT f) {
            return getRefCol(f, successorColors);
        }
        
        public Color getFocusedReferenceColor(FDT f){
            return getRefCol(f, focusedSuccessors);
        }
        
        public Color getModifiedReferenceColor(FDT f) {
            return getRefCol(f, modifiedSuccessors);
        }
        
        void clearSuccessors() {
            successorColors = new VirtualArray(VertexImpl.this, "Successor colors");
            focusedSuccessors = new VirtualArray(VertexImpl.this, "Focused Successors");
            modifiedSuccessors = new VirtualArray(VertexImpl.this, "Modified successors");
        }
        
        void addSuccessor(int i) {
            successorColors.setObject(Color.blue, i);
            focusedSuccessors.setObject(Color.red, i);
            modifiedSuccessors.setObject(Color.green, i);
        }
        
        void removeSuccessor(int i) {
            for(; successorColors.getObject(i + 1) != null && i < successors.size(); i++) {
                successorColors.setObject(successorColors.getObject(i + 1), i);
                focusedSuccessors.setObject(focusedSuccessors.getObject(i + 1), i);
                modifiedSuccessors.setObject(modifiedSuccessors.getObject(i + 1), i);
            }
        }
        
        static final long serialVersionUID = -3475865231803741060L;
    }
    
    // LabelDecorator interface
    
    boolean isReferenceLabelEnabled = false;
    
    public void setReferenceLabel(int i, String s) {
        weights.setObject(s, i);
    }
    
    public String getReferenceLabel(int i) {
        return (String)weights.getObject(i);
    }
    
    public boolean isReferenceLabelEnabled() {
        return isReferenceLabelEnabled;
    }
    
    public void setReferenceLabelEnabled(boolean b) {
        isReferenceLabelEnabled = b;
        //Note.warning(this, "setReferenceLabelEnabled(boolean) not implemented.");
    }
    
    public boolean isLabelEnabled() {
        return getVertexStyleSheet().isLabelEnabled();
    }
    
    public void setLabelEnabled(boolean isLabeled) {
        getVertexStyleSheet().setLabelEnabled(isLabeled);
    }
    
    public void setLabel(String s) {
        getVertexStyleSheet().setLabel(s);
    }
    
    public String getLabel() {
        return getVertexStyleSheet().getLabel();
    }
    
    static final long serialVersionUID = -4442957703584744342L;
}
