/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2006 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package matrix.structures.code;

import matrix.structures.FDT.FDT;

/**
 * This interface defines an FDT for holding pseudocode and
 * the current location within it.
 *
 * The class for visualizing the data is matrix.visual.VisualPseudocode.
 *
 * The location inside the pseudo-code is essentially a region inside the
 * code. These regions are defined through annotations in the pseudo-code.
 * A sample of the annotation follows:
 *  
 *   _3,7,/6,9 
 *   
 * This would mark a point from where regions 3,7 and 9 begin and region 6 ends.
 * 
 * The format is as follows:
 * 
 * The underscore marks the beginning of the annotation, commas separate the numbers
 * and a space or the end of line marks the end of the annotation. A slash in front of a
 * number denotes the end of a region, if there is no slash it is a beginning.
 * The choice of the underscore and space as delimiters is to make them as invisible as possible
 * to make it easier to read the original code.
 *
 * For example: _1,2 This is a _3 nice _/2  example_/1,/3
 * 
 * There are three regions here: "This is a nice example", "This is a nice" and "nice example",
 * each of which can be selected by calling setLocation(int location).
 *
 * @author Otto Seppl
 * @version $Revision: 1.2 $
 * 
 * @see matrix.visual.VisualPseudocode
 */


public interface AnnotatedPseudocode extends FDT{

    /**
     * Returns the name of the algorithm visualized
     * 
     *  @return the name of the algorithm visualized.
     */
    public String getCodeName();

    /**
     * Sets the name of the algorithm visualized
     * 
     *  @param name the name of the algorithm visualized.
     */
    public void setCodeName(String name);
    
    /**
     * Sets the annotated codelines for the algorithm visualized
     * 
     *  @param annotatedCodelines an array of Strings holding the annotated codelines for the algorithm visualized.
     */    
    public void setCodeLines(String[] annotatedCodelines);

    /**
     * Returns the annotated codelines for the algorithm visualized
     * 
     *  @return An array of Strings holding the annotated codelines for the algorithm visualized.
     */        
    public String[] getCodeLines();

    /**
     * Returns the number of the currently active region inside the code.
     * @return the number of the currently active region inside the code
     */
    public int getLocation();

    /**
     * Sets the number of the currently active region inside the code.
     * @param location the number of the currently active region inside the code
     */

    public void setLocation( int location );

    /**
     * Marks the currently active region as erroneous.
     * Error flag is cleared when the location is changed the next time.
     */
    
    public void markError();
    
    /**
     * returns whether the current region is erroneous
     * @return the error flag status
     */
    public boolean getErrorState();
    
    
}
