/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: VirtualArray.java,v 1.37 2007/08/22 10:04:47 oseppala Exp $ */

package matrix.structures.memory;

/*
    Data Structure library - 4/96 PNL
*/
import matrix.animation.Animator;
import matrix.structures.FDT.FDT;
import matrix.structures.other.Comparer;
import matrix.util.Note;

/**
 * VirtualArray is a Matrix <code>MemoryStructure</code> that can be visualised as an
 * array. For Matrix animation features to be available, the instance variables of the
 * class representing the memory structure to be animated must be held in a class that
 * implements <code>MemoryStructure</code>. VirtualArray is a dynamic memory structure that
 * can store objects. It is rather similar to java.util.Vector.
 */
public class VirtualArray implements MemoryStructure {

    /** Array holding the objects in memory */
    private Object[] memory;
    protected String name;
    protected FDT fdt;

    /** Indices of the first and last object in the array */
    private VirtualInteger low, high;
    /** indicates the logical index of the memory position 0 */
    private int zero;

    private boolean duplicate = false;

    /**
     * Constructs a new <code>VirtualArray</code>.
     * @param animated tells if this VirtualArray supports animation or not
     */
    protected VirtualArray(int zero) {
        low  = new VirtualInteger(zero, null, "lowest index in this array");
        high = new VirtualInteger(zero, null, "lowest index in this array");
        memory = new Object[1];
    }

    /** Constructs a new <code>VirtualArray</code>. */
    protected VirtualArray() {
        this(0);
    }

    /**
     * Constructs a new VirtualArray.
     * @param o a table of objects that are initially stored in the array
     * @param animated tells if this VirtualArray supports animation or not
     */
    protected VirtualArray(Object[] o, int zero) {
        this(zero);
        for (int i=zero; i<o.length+zero; i++)
            setObject(o[i], i);
    }

    /**
     * Constructs a new VirtualArray.
     * @param o a table of objects that are initially stored in the array
     */
    protected VirtualArray(Object[] o) {
        this(o, 0);
    }
    
    public VirtualArray(FDT fdt, String name) {
        this(0, fdt, name);
    }   
    
    public VirtualArray(int zero, FDT fdt, String name) {
        this(zero);
                
        setFDT(fdt);
        setName(name);
        
        fixVirtuals();
    }

    public VirtualArray(Object[] o, FDT fdt, String name) {
        this(o, 0, fdt, name);
    }

    public VirtualArray(Object[] o, int zero, FDT fdt, String name) {
        this(o, 0);
        setFDT(fdt);
        setName(name);

        fixVirtuals();
    }

    /**
     * This method fixes the Virtual structure information after the simple constructors
     * have been called. This is because the fields needed in the "innermost" constructor
     * could not have been set yet.
     */
    
    public void fixVirtuals(){
        low.setFDT(fdt);
        low.setName("low index in VirtualArray named ("+name+")");
        
        high.setFDT(fdt);
        high.setName("high index in VirtualArray named ("+name+")");
    }
    
    
    // ---- END OF CONSTRUCTORS ----


    // FDT interface Methods
    //public FDT getNewStructure() { return new VirtualArray(); }

    // MemoryStructure interface
    /**
     * Returns the Object stored in the parameter index. The <code>getObject()</code>
     * method should be used instead of this, as it is protected so it won't throw a
     * IndexOutOfBoundsError.
     * @param index the index of the stored Object
     * @return the stored object
     */
    public Object getMem(int index) {
        return read(index); //memory[index];
    }

    /**
     * Sets the param index to hold the param Object. The <code>SetObject()</code>
     * method should be used instead of this, as it is protected so it won't throw
     * a IndexOutOfBoundsError.
     * @param index the index of the Object to be set
     * @param obj the new contents of the index
     */
    public void setMem(int index, Object obj) {
	//Note.out(this, "setMem "+this);
        memory[index - zero] = obj; // not write(obj, index)!
    }

    // CDT

    /**
     * Swaps two Objects.
     * @param i the index of the first Object
     * @param j the index of the second Object
     */
    public void swap(int i, int j) {
        Object o = getObject(i);
        setObject(getObject(j), i);
        setObject(o, j);
    }

    private Object[] realloc(int size, int offset, Object[] mem) {
	//	Note.out(this, "realloc "+ size);
        if (size > mem.length) {
	    //Note.out(this, "resize by allocating more memory");
            int newSize = mem.length > 1 ? mem.length : 2;
            while (newSize < size) newSize*=2;
            Object[] newMem = new Object[newSize];
            for (int i=0; i<mem.length; i++)
                newMem[i] = mem[i];
            return realloc(size, offset, newMem);
        } else { // there is already enough memory in mem
            if (offset == zero) return mem; // no need to shift
            if (offset < zero) {
		//Note.out(this, "shift needed");
		for (int i=size()-1; i-(zero-offset) >= 0; i--) {
		    //Note.out(this, "Mem.length=" + mem.length + ", index=" + i + ", i-(zero-offset)="
		    //+ (i-(zero-offset)));
		    mem[i] = mem[i-(zero-offset)];
		    
		}
		// clear new empty slots
		for (int i=0; i<(zero-offset); i++) {
		    mem[i] = null;
		}
		zero = offset;
            } else { // offset > zero
                zero = offset;
		// this should not happend
                Note.err(this, "reallocation error: offset > zero");
            }
        }
        return mem;
    }

    /** This method is rarely invoked directly; invoked by public interface method setObject(). */
    private void write(Object obj, int address) {
        //Note.out(this, "write "+address);
        //if (zero > low.eval()) zero = low.eval();
        // save write operation into animator
        int index = address - zero; //- low.eval() + (low.eval() - zero);
        Animator.getActiveAnimator().putAssignment(this, obj, address);
        // do the actual write operation
        if ((index < 0) || (index >= memory.length)) {
	    //Note.out(this, "index before = " + index + " (zero="+zero+")");
            memory = realloc(size(), low.eval(), memory);
            index = address - zero;
	    //Note.out(this, "index after = " + index + " (zero="+zero+")");
        }
        memory[index] = obj;
    }

    /** This method is rarely invoked directly; invoked by public interface method getObject(). */
    private Object read(int address) {
        int index = address - zero; //- low.eval() + (low.eval() - zero);
        if ((index < 0) || (index >= memory.length)) {
            //Note.warning(this, "index out of bounds: index =" + index);
            return null;
        }
        else
            return memory[index];
    }


    /* public interface for an VirtualArray */

    /**
     * Sets the param Object in the param address index in this VirtualArray.
     * @param obj the Object to be inserted
     * @param address the index where the param Object is to be inserted
     */
    public void setObject(Object obj, int address) {
	//Note.out(this, "setObject "+this);
        //System.out.println("low = " + low + " high = " + high + " addr = " + address);
        if (address < low.eval()) {
            low.assign(address);
            memory = realloc(size(), address, memory);
            write(obj, address);
        } else if (address > high.eval()) {
            high.assign(address);
            memory = realloc(size(), zero, memory);
            write(obj, address);
        } else { // array entry allready exists
            write(obj, address);
        }
    }

    /**
     * Sets the param Object in the param address index in this VirtualArray.
     * @param obj the Object to be inserted
     * @param address the index where the param Object is to be inserted
     */
    public void setObject(Object obj, VirtualInteger address) {
        setObject(obj, address.eval());
    }

    /**
     * Returns the Object in the param address.
     * @param address the index of the Object to be retrieved
     */
    public Object getObject(int address) {
        return read(address);
    }


    /**
     * Returns the Object in the param address.
     * @param address the index of the Object to be retrieved
     */
    public Object getObject(VirtualInteger address) {
        return getObject(address.eval());
    }

    /**
     * Returns the current size of this VirtualArray.
     * @return the size
     */
    public int size() {
        return high.eval() - low.eval() + 1;
        }

    public boolean equals(Object o) {
	
        if(o instanceof VirtualArray) {
            return equals((VirtualArray)o);
        }
        return false;
    }

    // return memory blocks 0...n-1 as a string
    public String toString(int n) {
        //String s = "VirtualArray["+low.eval()+"..."+(high.eval()-1)+"]:";
        //
        //if (n > size()) System.out.println("Error in toString: n = " + n + " >= " + " size = " + size());
        /*
        String s = "";
        for (int i=0; i<=n; i++)
            s += memory[i] != null ? memory[i] : "_"; //memory[i+NIV] + " ";
        */
        StringBuffer s = new StringBuffer();
        for (int i=low.eval(); i<=high.eval(); i++)
            s.append(getObject(i) != null ? getObject(i) : "_");
        duplicate = false;
        return s.toString();
    }

    // Loadable
    public String toString() {
        if (duplicate) return "[...]";
        duplicate = true;
        return "["+toString(size())+"]";
    }

    // Array
    public int getFirst() { return low.eval(); }
    public int getLast() { return high.eval(); }
    public void setFirst(int first) {
        low.assign(first);
    }
    public void setLast(int last) {
        high.assign(last);
    }
    
    public boolean equals(VirtualArray a) {
      return Comparer.compareArray(this,a);
    }

    public String getName() {
        return name;
    }

    public void setName(String name) {
        this.name = name;
    }

    public FDT getFDT() {
        return fdt;
    }

    public void setFDT(FDT fdt) {
        this.fdt = fdt;
    }

    static final long serialVersionUID = 4736482423992020460L;
}
