package matrix.structures.spatial.CDT.probe;

import matrix.structures.FDT.FDT;
import matrix.structures.FDT.Tree;
import matrix.structures.FDT.probe.Key;
import matrix.util.Note;
import matrix.structures.spatial.Area;
import matrix.structures.spatial.SpatialComparable;
import matrix.structures.spatial.FDT.probe.Point;
import matrix.structures.util.Iterator;

/**
 * Defines the functionality and structure of a node in a 
 * Point-Region-quadtree. Each non-leaf, i.e. internal node has four 
 * children. Additionally all nodes contain a data/key field and 
 * coordinates, which define what rectangular area the node corresponds 
 * to in 2-dimensional space. In a PR-quadtree all data is stored in
 * the leaves and when nodes are split the area is divided into 
 * four equally-sized quadrants.
 * 
 * @see PRQuadTree, QuadTreeNode
 * @author Juha Helminen
 */
public class PRQuadTreeNode extends QuadTreeNode implements Tree, FDT, Iterator, Area {

	// -- Constructors --

	/**
	 * Creates an empty node.
	 */
	public PRQuadTreeNode() {
		super();
	}

	/**
	 * Creates an empty node and inserts key into it.
	 * @param key The key to be inserted in the tree.
	 */
	public PRQuadTreeNode(Point key) {
		super(key);
	}

	/**
	 * Creates an empty node corresponding to the area given 
	 * as parameter and inserts key in it.
	 * 
	 * @param key The key/data to be inserted in the node.
	 * @param minX The x-coordinate of the lower left corner of the rectangular area.
	 * @param maxX The x-coordinate of the upper right corner of the rectangular area.
	 * @param minY The y-coordinate of the lower left corner of the rectangular area.
	 * @param maxY The y-coordinate of the upper right corner of the rectangular area.
	 */
	public PRQuadTreeNode(Point key, double minX, double maxX, double minY, double maxY) {
		super(key, minX, maxX, minY, maxY);
	}
	


	// -- Implementations of inherited abstract methods --
	
	public Tree getNewNode(Object key) {
		if (key instanceof Point) {
            PRQuadTreeNode node = new PRQuadTreeNode((Point)key);
            // Inherit the decorator.
            if (this.hasPaintingStyleDecorator()) {
                node.setPaintingStyleDecorator(this.getPaintingStyleDecorator());
            }
			return node;
		} else {
			Note.warning(this, "Cannot create a new node with key of type " + key.getClass());
			return null;
		}
	}

    public PRQuadTreeNode getNewNode(Point key, 
                                     double minX, 
                                     double maxX, 
                                     double minY, 
                                     double maxY) {
        PRQuadTreeNode node = new PRQuadTreeNode(key, minX, maxX, minY, maxY);
        // Inherit the decorator.
        if (this.hasPaintingStyleDecorator()) {
            node.setPaintingStyleDecorator(this.getPaintingStyleDecorator());
        }
        return node;
    }

    public void delete(SpatialComparable e) {
		this.delete((Object)e);
	}

	public void insert(SpatialComparable e) {
		this.insert((Object)e);
	}

	
	
	// -- QuadTree operation implementations --

	public void insert(Object o) {
		if (o == null) {
			// Nothing to insert.
		} else if (o instanceof Point) {
			// Insert in the root if the tree is empty.
			if (this.isLeaf() && 
				this.isEmpty() &&
				this.selectQuadrant((Point)o) != OUTSIDE) {
				this.setKey(o);
			} else {
				this.insert((Point)o);
			}
		} else if (o instanceof Key) {
			this.insert(((Key)o).getObject());
		} else if (o instanceof Iterator) {
			// o is a collection of data, insert data elements separately
			while (((Iterator)o).hasNext()) {
				this.insert(((Iterator)o).next());
			}
		} else {
			Note.warning(this, "Cannot insert object of type " + o.getClass());
		}
	}
	
	public void delete(Object o) {
		if (o == null) {
			// Nothing to delete.
		} else if (o instanceof Point) {
			if (!this.isEmpty() &&
				this.getKey().equals(o)) {
				// The tree only has this node.
				this.setKey(null);
				this.numberOfKeys.assign(0);
			} else {
				this.delete((Point)o, null);
			}
		} else if (o instanceof Key) {
			this.delete(((Key)o).getObject());
		} else {
			Note.warning(this, "Cannot delete object of type " + o.getClass());
		}
	}

	/**
	 * Inserts the key into one of the subtrees of this node.

	 * @param key The key to be inserted.
	 */
	private void insert(Point key) {
		if (!this.isEmpty()) {
			// Case 1: Leaf node, which already contains a key -> split the node.
			// Select a branch for the current node's key.
			int nodeKeyQuadrant = this.selectQuadrant((Point)this.getKey());
			if (nodeKeyQuadrant == INDIVISIBLE) {
				// The area is so small that it cannot be divided any more
				// -> insert the key at the end of a list of keys
				this.setKey(key, this.numberOfKeys.eval());
				this.numberOfKeys.inc();
			} else {
				// Store the key from the current node.
				Point nodeKey = (Point)this.getKey();
				// Delete the key in the current node.
				this.setKey(null);
				this.numberOfKeys.assign(0);
				// Create a new child node and store the key there.
				this.setChild(nodeKey, nodeKeyQuadrant);
				// Recall the insert for the current node and the key to be inserted.
				// -> It is chosen a branch to insert into and either 
				// 1) If the branch is empty, it's inserted there. or
				// 2) The nodeKey was just inserted there and a new split occurs.
				this.insert(key);
			}
		} else /* this is empty */ {
			if (this.isLeaf()) {
				// Case 2: Insert in this node, which is an empty leaf.
				this.setKey(key);
			} else {
				// Case 3: This node is an empty internal node
				// -> insert in a subtree.
				// Select the correct branch/child/quadrant and insert key in that branch.
				((PRQuadTreeNode)this.getChild(this.selectQuadrant(key))).insert(key);
			}
		}
	}

	/**
	 * Deletes the key from the subtrees of this node.
	 * 
	 * @param key The key to be deleted.
	 * @param collapseNode The node to which to collapse to if collapsing occurs.
	 */
	private void delete(Point key, PRQuadTreeNode collapseNode) {
		// Select the right branch/child/quadrant for the key.
		int keyQuadrant = selectQuadrant(key);
		PRQuadTreeNode childNode = (PRQuadTreeNode)this.getChild(keyQuadrant); 
		if (childNode == null) {
			// No node found containing the key.
			return;
		} else if (childNode.isLeaf()) {
			int keyIndex = -1;
			for (int i = 0; i < childNode.numberOfKeys.eval(); i++) {
				if (((Point)childNode.getKey(i)).equals(key)) {
					keyIndex = i;
					break;
				}
			}
			if (keyIndex > -1) {
				// The key was found -> delete the key and possibly the node.
				childNode.deleteKey(keyIndex);
				// Collapse nodes if possible
				if ((this.getChildCount()) == 1) {
					// Only one child left -> find it and move its key up the tree.
					childNode = null;
					int i = 0;
					while ((childNode = (PRQuadTreeNode)this.getChild(i++)).isEmpty()) {}
					// Do not collapse internal nodes or nodes with many keys.
					if (childNode.numberOfKeys.eval() == 1) {
						if (collapseNode == null) {
							// Collapse only a single level
							collapseNode = this;
						}
						collapseNode.setKey(childNode.getKey());
						collapseNode.deleteChildren();
					}
				} else {
					// The node cannot be collapsed.
					return;
				}
			} else {
				// No node found containing the key.
				return;
			}
		} else {
			// Node is internal
			if (this.getChildCount() == 1) {
				// Mark the node as a possible candidate for collapsing to,
				// if none already exist.
				if (collapseNode == null) {
					childNode.delete(key, this);
				} else {
					childNode.delete(key, collapseNode);
				}
			} else {
				// Any possible earlier collapsing points are erased, because
				// the node can't be collapsed.
				childNode.delete(key, null);
			}
		}
	}

	
	/**
	 * Deletes the key at index.
	 * @param index
	 */
	private void deleteKey(int index) {
		this.setKey(null, index);
		this.numberOfKeys.dec();
		this.keys.swap(index, this.numberOfKeys.eval());
		this.keys.setLast(this.numberOfKeys.eval() - 1);
	}

	
	/**
	 * Determines the quadrant of the area of the node, to which the key should be stored.
	 * Values 0-3 correspond to quadrants NE,SE,SW and NW respectively. Boundary conditions 
	 * are resolved by always choosing the quadrant to the right (= along the x-axis) or up 
	 * (= along the y-axis).
	 *
	 * @param key A key.
	 * @return 0-3 corresponding to quadrants NE,SE,SW and NW or -1 if the key is outside 
	 * the boundaries and -2 if the area is too small to divide;
	 * QUADRANT_NE = 0, QUADRANT_SE = 1, QUADRANT_SW = 2, QUADRANT_NW = 3, OUTSIDE = -1,
	 * INDIVISIBLE = -2.
	 */
	private int selectQuadrant(Point key) {
		if ((key.getX() >= this.minX) &&
			(key.getY() >= this.minY) &&
			(key.getX() < this.maxX) &&
			(key.getY() < this.maxY)) {
			double halfX = ((this.maxX - this.minX) / 2);
			double halfY = ((this.maxY - this.minY) / 2);
			double threshold = 0.5;
			if ((halfX > threshold) || (halfY > threshold)) {
				if (key.getX() < (this.minX + halfX)) {
					if (key.getY() < (this.minY + halfY)) {
						return QUADRANT_SW;
					} else {
						return QUADRANT_NW;
					}
				} else {
					if (key.getY() < (this.minY + ((this.maxY - this.minY) / 2))) {
						return QUADRANT_SE;
					} else {
						return QUADRANT_NE;
					}
				}
			} else {
				// The area is too small to be divided any more.
				return INDIVISIBLE;
			}
		} else {
			// The key is located outside the area corresponding to this quadtree node.
			return OUTSIDE;
		}
	}

	/**
	 * Splits this node creating four new children, in one of which key
	 * is stored as instructed by parameter quadrant.
	 * 
	 * @param key The key to be stored in the child node.
	 * @param quadrant Specifies the child in which the key is to be stored.
	 */
	private void setChild(Point key, int quadrant) {
		PRQuadTreeNode node;
		// When a new node is created it is assigned to occupy 
		// a fourth of the area the current node covers.
		Point keyNE = null, keySE = null, keySW = null, keyNW = null;
		switch(quadrant) {
			case QUADRANT_NE:
				keyNE = key;
				break;
			case QUADRANT_SE:
				keySE = key;
				break;
			case QUADRANT_SW:
				keySW = key;
				break;
			case QUADRANT_NW:
				keyNW = key;
				break;
			default: /* do nothing */ return;
		}
		
		node = this.getNewNode(keyNE,
							   (this.minX + ((this.maxX - this.minX) / 2)),
							   this.maxX,
							   (this.minY + ((this.maxY - this.minY) / 2)),
							   this.maxY);
		this.setChild(node, QUADRANT_NE);
		node = this.getNewNode(keySE,
							   (this.minX + ((this.maxX - this.minX) / 2)),
							   this.maxX,
							   this.minY,
							   (this.minY + ((this.maxY - this.minY) / 2)));
		this.setChild(node, QUADRANT_SE);
		node = this.getNewNode(keySW,
							   this.minX,
						       (this.minX + ((this.maxX - this.minX) / 2)),
							   this.minY,
							   (this.minY + ((this.maxY - this.minY) / 2)));
		this.setChild(node, QUADRANT_SW);
		node = this.getNewNode(keyNW,
							   this.minX,
							   (this.minX + ((this.maxX - this.minX) / 2)),
							   (this.minY + ((this.maxY - this.minY) / 2)),
							   this.maxY);
		this.setChild(node, QUADRANT_NW);
	}

	public PRQuadTreeNode search(Point key) {
		if (!this.isEmpty()) {
			for (int i = 0; i < this.numberOfKeys.eval(); i++) {
				if (this.getKey(i).equals(key)) {
					return this;
				}
			}
			// Search failed.
			return null;
		} else {
			PRQuadTreeNode keyChild = ((PRQuadTreeNode)this.getChild(this.selectQuadrant(key)));
			if (keyChild != null) {
				return keyChild.search(key);
			} else {
				// Search failed.
				return null;
			}
		}
	}
	
	public boolean equals(Object obj) {
		if (obj instanceof PRQuadTreeNode) {
			return super.equals(obj);
		} else {
			return false;
		}
	}
	
	private static final long serialVersionUID = 4135528722587995434L;
}
