package matrix.structures.spatial.CDT.probe;

import java.awt.Rectangle;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.Vector;

import matrix.decoration.LabelDecorator;
import matrix.structures.FDT.FDT;
import matrix.structures.FDT.Tree;
import matrix.structures.FDT.probe.Key;
import matrix.structures.FDT.probe.Table;
import matrix.structures.memory.VirtualBoolean;
import matrix.structures.memory.VirtualInteger;
import matrix.structures.memory.VirtualObject;
import matrix.structures.spatial.Area;
import matrix.structures.spatial.PaintingStyleDecorator;
import matrix.structures.spatial.SpatialComparable;
import matrix.structures.spatial.FDT.probe.SpatialElement;
import matrix.structures.util.Iterator;
import matrix.util.Note;

/**
 * Class defining the basic functionality of an R-tree, which is B-tree-like 
 * data structure used to index spatial data. 
 * 
 * The data structure divides the space into hierarchically nested, and 
 * possibly overlapping, boxes. Each node has its own (bounding) box, the 
 * smallest rectangle that covers the area of its entries' bounding boxes. 
 * Each node has a variable number of entries ranging from a minimum 
 * value to a maximum, where min <= Sup(MAX/2), e.g. (1,2) and (2,3).
 * In internal nodes the entries are child nodes and in leaves spatial data,
 * that is, all data is stored in the leaves.
 *   
 * @see RTree
 * @author Juha Helminen
 */
public class RTreeNode implements FDT, Tree, Iterator, Area, LabelDecorator {

    private VirtualObject paintingStyleDecorator = 
        new VirtualObject(null, "PaintingStyleDecorator");

    // An empty tree has no real bounding box and this is used instead.
	protected static final Rectangle DUMMY_BOUNDS =	new Rectangle(0, 0, 0, 0);

	// These are used to generate meaningful IDs for the nodes 
	// (and corresponding covering rectangles).
	// They are visualized as labels for the nodes/bounding boxes.
	private static VirtualInteger idCounter = new VirtualInteger(-1, null, "RTreeNode.idCounter");
	private VirtualInteger id;
    private transient boolean labelEnabled = true;
    private String label = null;

	// Max
	private static final int DEFAULT_MAXIMUM_ENTRIES = 3;
	// min <= Sup(MAX/2)
	private static final int DEFAULT_MINIMUM_ENTRIES = 2;
	// Define which split method to use.
	public static final int SPLIT_QUADRATIC = 0;
	public static final int SPLIT_LINEAR = 1;
	private static int splitAlgorithm = SPLIT_QUADRATIC;
	
	// Stores the smallest rectangle which covers the area of all
	// this or child nodes' entries.
	protected VirtualObject rectangle;
	// References to the children nodes of this node or
	// in leaf nodes the keys in this node.
	protected Table entries;
	// The maximum allowed number of entries=children/keys in the node
	protected int max;
	// The minimum allowed number of entries=children/keys in the node
	protected int min;
	protected VirtualInteger numberOfEntries;
	// Reference to the parent node of this node, null if root.
	protected VirtualObject parent;
	// True, if this node is a leaf node, false otherwise.
	protected VirtualBoolean isLeaf;
	

	
	// -- Constructors --

	public RTreeNode() {
		this(null, DEFAULT_MINIMUM_ENTRIES, DEFAULT_MAXIMUM_ENTRIES, true, RTreeNode.getNewId());
	}

	/**
	 * @param key A key that is inserted in the empty tree after creation.
	 */
	public RTreeNode(SpatialComparable key) {
		this(null, DEFAULT_MINIMUM_ENTRIES, DEFAULT_MAXIMUM_ENTRIES, true, RTreeNode.getNewId());
	}

	/**
	 * @param minEntries The minimum allowed number of entries in a node.
	 * @param maxEntries The maximum allowed number of entries in a node.
	 */
	public RTreeNode(int minEntries, int maxEntries) {
		this(null, minEntries, maxEntries, true, RTreeNode.getNewId());
	}

	/**
	 * @param minEntries The minimum allowed number of entries in a node.
	 * @param maxEntries The maximum allowed number of entries in a node.
	 * @param isLeaf True if the created node is to be a leaf node, false otherwiae.
	 */
	public RTreeNode(int minEntries, int maxEntries, boolean isLeaf) {
		this(null, minEntries, maxEntries, isLeaf, RTreeNode.getNewId());
	}

	/**
	 * @param minEntries The minimum allowed number of entries in a node.
	 * @param maxEntries The maximum allowed number of entries in a node.
	 * @param isLeaf True if the created node is to be a leaf node, false otherwiae.
	 * @param id The id number for the new node.
	 */
	public RTreeNode(int minEntries, int maxEntries, boolean isLeaf, int id) {
		this(null, minEntries, maxEntries, isLeaf, id);
	}
	
	/**
	 * @param key A key that is inserted in the empty tree after creation.
	 * @param minEntries The minimum allowed number of entries in a node.
	 * @param maxEntries The maximum allowed number of entries in a node.
	 * @param isLeaf True if the created node is to be a leaf node, false otherwiae.
	 * @param id The id number for the new node.
	 */
	public RTreeNode(SpatialComparable key, int minEntries, int maxEntries, boolean isLeaf, int id) {
		this.id = new VirtualInteger(id, this, "id number of the node");
		// Check that the values are legal
		if (!((minEntries >= 1) &&
			  (maxEntries >= 2) &&
			  (minEntries <= Math.ceil(maxEntries / 2.0)))) {
			Note.err(this.getClass(), "Illegal argument in class RTreeNode: must be m <= Sup(M/2).");
		}
		this.min = minEntries;
		this.max = maxEntries;
		this.entries = new Table();
		// Set the limits for the Table containing the keys
		// so that also the empty "slots" are visualized.
		this.entries.setFirst(0);
		this.entries.setLast(maxEntries - 1);

		this.isLeaf = new VirtualBoolean(this, "isLeaf");
        this.isLeaf.assign(isLeaf);
		this.parent = new VirtualObject(null, this, "parent");
		if (key != null) {
			this.setEntry(key, 0);
			this.numberOfEntries = new VirtualInteger(1, this, "numberOfEntries");
			// Get the dimensions of the smallest rectangle that covers the key
			this.rectangle = new VirtualObject(null, this, "rectangle");
			this.setBoundingBox(key.getBoundingBox());
		} else {
			this.numberOfEntries = new VirtualInteger(0, this, "numberOfEntries");
			// Initially the node has no corresponding area.
			this.rectangle = new VirtualObject(null, this, "rectangle");
			this.setBoundingBox(null);
		}
	}

	

	// -- Getters = accessors --
	
	protected Object getEntry(int index) {
		return this.getEntries().getObject(index);
	}

	protected RTreeNode getParent() {
		return (RTreeNode)this.parent.getObject();
	}
	
	protected int getMinEntries() {
		return this.min;
	}

	protected int getMaxEntries() {
		return this.max;
	}
	
	protected Table getEntries() {
		return this.entries;
	}
	
	
	
	// -- Setters = mutators --
	
	protected void setEntry(Object entry, int index) {
		this.getEntries().setObject(entry, index);
	}

	protected void setParent(RTreeNode p) {
		this.parent.setObject(p);
	}

	protected void setLeaf(boolean bool) {
		this.isLeaf.assign(bool);
	}
	
	public static void setSplitAlgorithm(int algorithm) {
		RTreeNode.splitAlgorithm = algorithm;
	}
	
	
	
	// -- Implementation of interface FDT --
	
	public Object getElement() {
		// This defines what is shown in the node in the tree visualization.
		if (this.isLeaf()) {
			return this.getEntries();
		} else {
			return null;
		}
	}

	public void setElement(Object o) {
		if (o instanceof Table) {
			this.entries = (Table)o;
		} else {
			Note.warning(this, "Cannot insert " + o);
		}
	}
	

	
	// -- Implementation of interface Tree --
	
	public int getSubTreeCount() {
		if (this.isLeaf()) {
			return 0;
		} else {
			return (this.numberOfEntries());  
		}
	}

	public Tree[] getSubTrees() {
		if (this.isLeaf()) {
	        return new Tree[0];
		} else {
	        Tree[] subTrees = new Tree[this.numberOfEntries()];
	        for (int i = 0; i <= this.getEntries().getLast(); i++) {
	        	if (this.getEntry(i) != null) {
	        		subTrees[i] = (Tree)this.getEntry(i);
	        	}
	        }
	        return subTrees;
		}
	}

	public void setSubTree(Tree t, int child) {
        if ((t instanceof RTreeNode) && (child <= this.getEntries().getLast())) {
        	this.setEntry((RTreeNode)t, child);
        } else {
        	Note.warning(this, "Operation not allowed.");
        }
	}

	public Tree getNewNode(Object key) {
		if (key instanceof SpatialComparable) {
			return new RTreeNode((SpatialComparable)key);
		} else {
			Note.warning(this, "Cannot create a new node with key of type " + key.getClass());
			return null;
		}
	}
	
    protected RTreeNode getNewNode(int minEntries, int maxEntries, boolean isLeaf, int id) {
        return new RTreeNode(minEntries, maxEntries, isLeaf, id);
    }
	
    private RTreeNode getNewNode(int minEntries, int maxEntries, boolean isLeaf) {
        return this.getNewNode(minEntries, maxEntries, isLeaf, RTreeNode.getNewId());
    }

	// -- Implementation of interface Area --

	public SpatialElement[] getSpatialElements() {
		// Only leaves contain data.
		if (this.isLeaf()) {
			Vector keys = new Vector(this.numberOfEntries());
			for (int i = 0; i <= this.getEntries().getLast(); i++) {
				if (this.getEntry(i) != null) {
					keys.add(this.getEntry(i));
				}
			}
			return (SpatialElement[])keys.toArray(new SpatialElement[keys.size()]);
		} else {
			return new SpatialElement[0];
		}
	}

	public Area[] getSubAreas() {
		// Only internal nodes have children (and therefore subareas).
		if (this.isLeaf()) {
			return new Area[0];
		} else {
			Vector children = new Vector(this.numberOfEntries());
			for (int i = 0; i <= this.getEntries().getLast(); i++) {
				if (this.getEntry(i) != null) {
					children.add(this.getEntry(i));
				}
			}
			return (Area[])children.toArray(new Area[children.size()]);
		}
	}

	public Rectangle getBoundingPolygon() {
		if (this.getBoundingBox() == null) {
			if (this.getParent() == null) {
				// An empty tree.
				return RTreeNode.DUMMY_BOUNDS;
			} else {
				// An empty node.
				return this.getParent().getBoundingPolygon();
			}
		} else {
			// Return a _copy_ of the rectangle that 
			// represents the area that this node covers.
			return this.getBoundingBox().getBounds();
		}
	}

	public void insert(SpatialComparable e) {
		this.insertData(e);
	}

	public void delete(SpatialComparable e) {
		this.deleteKey(e);
	}

	public void addSubArea(Area a) {
		Note.warning(this, "Method addSubArea() not implemented.");
	}

	public void removeSubArea(Area a) {
		Note.warning(this, "Method removeSubArea() not implemented.");
	}

    public boolean hasPaintingStyleDecorator() {
        return (this.paintingStyleDecorator.getObject() != null);
    }

    public PaintingStyleDecorator getPaintingStyleDecorator() {
        return (PaintingStyleDecorator)this.paintingStyleDecorator.getObject();
    }

    public void setPaintingStyleDecorator(PaintingStyleDecorator psd) {
        this.paintingStyleDecorator.setObject(psd);
    }

    
	// -- Implementation of interface Iterator --
	
	private java.util.Iterator keyIterator = null;

	/**
	 * Recursively traverses the tree and inserts all the keys into the list
	 * given as parameter.
	 * @param keyList The list where to insert the keys in this (sub)tree.
	 */
	private void recursiveTraverse(ArrayList keyList) {
		if (this.isLeaf()) {
			for (int i = 0; i <= this.getEntries().getLast(); i++) {
				if (this.getEntry(i) != null) {
					keyList.add(this.getEntry(i));
				}
			}
		} else {
			for (int i = 0; i <= this.getEntries().getLast(); i++) {
				if (this.getEntry(i) != null) {
					((RTreeNode)this.getEntry(i)).recursiveTraverse(keyList);
				}
			}
		}
	}
	
	public boolean hasNext() {
		if (keyIterator == null) {
			ArrayList keys = new ArrayList();
			// Recursively insert the keys in the list
			this.recursiveTraverse(keys);
			keyIterator = keys.iterator();
		}
		return keyIterator.hasNext();
	}

	public Object next() {
		return keyIterator.next();
	}

	public void reset() {
		keyIterator = null;
	}

	
	
	// ** Implementation of interface LabelDecorator **
	
	public static synchronized void resetIDCounter() {
		RTreeNode.idCounter.assign(-1);
	}

	protected static synchronized int getNewId() {
		return RTreeNode.idCounter.inc();
	}
	
	protected synchronized int getId() {
		return this.id.eval();
	}
	
	protected synchronized void setId(int id) {
		this.id.assign(id);
	}
	
    public boolean isLabelEnabled() {
		return this.labelEnabled;
    }

    public void setLabelEnabled(boolean b) {
        this.labelEnabled = b;
    }

    public void setLabel(String s) {
        this.label = s;
    }

    public String getLabel() {
    	return (label == null) ? ("R" + this.id.toString()) : label;
    }

    public boolean isReferenceLabelEnabled() {
        return false;
    }

    public void setReferenceLabelEnabled(boolean b) {
    }

    public void setReferenceLabel(int i, String s) {
    }

    public String getReferenceLabel(int i) {
        return null;
    }

    
    
    // ** Methods for executing R-tree operations **
	
	public RTreeNode insert(Object o) {
		if (o == null) {
			return this;
		} else if (o instanceof SpatialComparable) {
			return this.insertData(o);
		} else if (o instanceof Key) {
			return this.insert(((Key)o).getObject());
		} else if (o instanceof Iterator) {
			RTreeNode root = this;
			while (((Iterator)o).hasNext()) {
				root = this.insert(((Iterator)o).next());
			}
			return root;
		} else {
			Note.warning(this, "Cannot insert object of type " + o.getClass());
			return this;
		}
	}
	
	public RTreeNode delete(Object o) {
		if (o == null) {
			return this;
		} else if (o instanceof SpatialComparable) {
			return this.deleteKey((SpatialComparable)o);
		} else if (o instanceof Key) {
			return this.delete(((Key)o).getObject());
		}  else {
			Note.warning(this, "Cannot delete object of type " + o.getClass());
			return this;
		}
	}

	protected boolean isLeaf() {
		return (boolean)this.isLeaf.eval();
	}

	protected boolean isRoot() {
		return (this.getParent() == null);
	}

	protected boolean isFull() {
		return (this.numberOfEntries() == this.getMaxEntries());
	}

	protected boolean isUnderfull() {
		return (this.numberOfEntries() < this.getMinEntries());
	}

	protected boolean isEmpty() {
		return (this.numberOfEntries() == 0);
	}

	protected int numberOfEntries() {
		return this.numberOfEntries.eval();
	}
	
	protected Rectangle2D getBoundingBox() {
		if (this.rectangle.getObject() == null) {
			return null;
		} else {
			// Return the rectangle that represents the area that this node covers.
			return (Rectangle2D)this.rectangle.getObject();
		}
	}
	
	protected void setBoundingBox(Rectangle2D r) {
		this.rectangle.setObject(r);
	}

	/**
	 * @return The area of this node's bounding box.
	 */
	protected double calculateArea() {
		if (getBoundingBox() != null) {
			return (this.getBoundingBox().getHeight() * 
					this.getBoundingBox().getWidth());
		} else {
			return 0;
		}
	}

	/**
	 * Inserts an entry in the tree.
	 * @param entry the entry to be inserted, either a node or a SpatialComparable key.
	 * @return Possibly new root.
	 */
	private RTreeNode insertData(Object entry) {
		RTreeNode root = this, node = null;
		if (entry instanceof SpatialComparable) {
			// Choose a leaf where to insert the key.
			node = chooseLeaf((SpatialComparable)entry);
		} else if (entry instanceof RTreeNode) {
			// Inserting a subtree into the tree.
			// Choose a node where to insert the subtree.
			node = chooseNode((RTreeNode)entry);
		} else {
			Note.warning(this, "Cannot insert object of type " + entry.getClass());
		}
		if (!node.isFull()) {
			// Insert the key in the leaf node.
			node.insertEntry(entry);
			// Adjust the bounding boxes.
			node.adjustTree(null);
		} else {
			// The node is full -> split.
			RTreeNode[] newNodes;
			// Store the key temporarily in the node (even though it's already full),
			// then split the node into two new nodes, among which all the keys
			// are divided.
			node.insertEntry(entry);
			newNodes = node.split();
			// Replace the split node with one of the two new nodes.
			if (node.getParent() != null) {
				node.getParent().replace(node, newNodes[0]);
			}
			// Propagate the changes up the tree to the root. 
			root = newNodes[0].adjustTree(newNodes[1]);
		}
		return root;
	}
	
	protected RTreeNode[] split() {
		// Split this node into two new nodes and divide the entries between them. 
		RTreeNode[] nodes = new RTreeNode[2];
		// nodes[0] will replace this node, so take its id.
		nodes[0] = this.getNewNode(this.getMinEntries(), this.getMaxEntries(), this.isLeaf(), this.getId());
		nodes[1] = this.getNewNode(this.getMinEntries(), this.getMaxEntries(), this.isLeaf());
        // Propagate the painting style.
        if (this.hasPaintingStyleDecorator()) {
            nodes[0].setPaintingStyleDecorator(this.getPaintingStyleDecorator());
            nodes[1].setPaintingStyleDecorator(this.getPaintingStyleDecorator());
        }
		this.splitDivideEntries(nodes);
		return nodes;
	}
	
	/**
	 * This method defines how the keys are divided between the nodes
	 * when a split occurs.
	 * 
	 * @param nodes An array containing two empty nodes, among whic to divide
	 * all the entries in this node.
	 */
	protected void splitDivideEntries(RTreeNode[] nodes) {
		if (this.isEmpty()) {
			return;
		} else {
			// Pick one entry from each of the nodes.
			Object[] seeds;
			if (RTreeNode.splitAlgorithm == RTreeNode.SPLIT_QUADRATIC) {
				seeds = this.pickSeedsQuadratic();
			} else /*if (RTreeNode.SPLIT_ALGORITHM == RTreeNode.SPLIT_LINEAR)*/{
				seeds = this.pickSeedsLinear();
			}
			nodes[0].insertEntry(seeds[0]);
			if (seeds[1] == null) {
				// This node contained only one entry.
				return;
			} else {
				nodes[1].insertEntry(seeds[1]);
				Object nextEntry = null;
				double enlargement0, enlargement1;
				while (!this.isEmpty()) {
					if (nodes[0].numberOfEntries() + this.numberOfEntries() <= nodes[0].getMinEntries()) {
						// Assign the rest to node 0 and terminate.
						for (int i = 0; i <= this.getEntries().getLast(); i++) {
							if (this.getEntry(i) != null) {
								nodes[0].insertEntry(this.getEntry(i));
								this.setEntry(null, i);
							}
						}
						break;
					} else if (nodes[1].numberOfEntries() + this.numberOfEntries() <= nodes[1].getMinEntries()) {
						// Assign the rest to node 1 and terminate.
						for (int i = 0; i <= this.getEntries().getLast(); i++) {
							if (this.getEntry(i) != null) {
								nodes[1].insertEntry(this.getEntry(i));
								this.setEntry(null, i);
							}
						}
						break;
					}
					// Pick the entry from this node which to insert next.
					if (RTreeNode.splitAlgorithm == RTreeNode.SPLIT_QUADRATIC) {
						nextEntry = this.pickNextQuadratic(nodes[0], nodes[1]);
					} else /*if (RTreeNode.SPLIT_ALGORITHM == RTreeNode.SPLIT_LINEAR)*/{
						nextEntry = this.pickNextLinear();
					}
					// Insert the entry in one of the nodes.
					enlargement0 = enlargementArea(nextEntry, nodes[0]);
					enlargement1 = enlargementArea(nextEntry, nodes[1]);
					if (enlargement0 < enlargement1) {
						// Insert in node 0.
						nodes[0].insertEntry(nextEntry);
					} else if (enlargement0 > enlargement1) {
						// Insert in node 1.
						nodes[1].insertEntry(nextEntry);
					} else {
						// Resolve tie by comparing the areas.
						double area0 = nodes[0].calculateArea();
						double area1 = nodes[1].calculateArea();
						if (area0 < area1) {
							// Insert in node 0.
							nodes[0].insertEntry(nextEntry);
						} else if (area0 > area1) {
							// Insert in node 1.
							nodes[1].insertEntry(nextEntry);
						} else {
							// Resolve tie by comparing the number of entries
							// and finally resolve tie by choosing node 0.
							if (nodes[0].numberOfEntries() <= nodes[1].numberOfEntries()) {
								// Insert in node 0.
								nodes[0].insertEntry(nextEntry);
							} else {
								// Insert in node 1.
								nodes[1].insertEntry(nextEntry);
							}
						}
					}
				}
				return;
			}
		}
	}

	/**
	 * Propagates the changes done in this node to the root
	 * adjusting bounding boxes and performing splits as needed.
	 * 
	 * @param  newSibling  A node that should be inserted in the tree as a sibling
	 *                     of this node.
	 * @return The possibly new root.
	 */
	protected RTreeNode adjustTree(RTreeNode newSibling) {
		RTreeNode parent = this.getParent();
		if (parent == null) {
			// This is the root node.
			if (newSibling == null) {
				// The root is unchanged.
				return this;
			} else {
				// The root must be split ->
				// Grow the tree taller and return the new root.
				RTreeNode newRoot = this.getNewNode(this.getMinEntries(), 
                                                    this.getMaxEntries(), 
                                                    false);
                // Copy the style.
                if (this.hasPaintingStyleDecorator()) {
                    newRoot.setPaintingStyleDecorator(this.getPaintingStyleDecorator());
                }
				newRoot.insertEntry(this);
				newRoot.insertEntry(newSibling);
				return newRoot;
			}
		} else {
			// Add the newSibling.
			if (newSibling == null) {
				// Only adjust the covering rectangle in the ancestors.
				parent.setBoundingBox(RTreeNode.add(parent.getBoundingBox(), 
													this.getBoundingBox()));
				return parent.adjustTree(null);
			} else if (!parent.isFull()) {
				// The newSibling can be inserted in the parent -> no more splits.
				// Insert the sibling next to this node, in order to have split nodes
				// appear next to each other.
				parent.insertNextTo(newSibling, this);
				parent.numberOfEntries.inc();
				parent.setBoundingBox(RTreeNode.add(parent.getBoundingBox(), 
													RTreeNode.add(this.getBoundingBox(), 
														          newSibling.getBoundingBox())));
				return parent.adjustTree(null);
			} else {
				// Another split occurs in the parent.
				// Insert the entry in the node (even though it's already full),
				// split the node into two new nodes and divide the entries between them.
				
                // Insert the sibling next to this node, in order to have split nodes
                // appear next to each other.
                parent.insertNextTo(newSibling, this);
                parent.numberOfEntries.inc();
                parent.setBoundingBox(RTreeNode.add(parent.getBoundingBox(), 
                                                    RTreeNode.add(this.getBoundingBox(), 
                                                                  newSibling.getBoundingBox())));
				
                
                RTreeNode[] newNodes = parent.split();
				// Replace the split node with one of the two new nodes.
				if (parent.getParent() != null) {
					parent.getParent().replace(parent, newNodes[0]);
				}
				// Propagate the changes up the tree to the root. 
				return newNodes[0].adjustTree(newNodes[1]);
			}
		}
	}

	protected void replace(RTreeNode nodea, RTreeNode nodeb) {
		if ((nodea == null) || (nodeb == null)) {
			return;
		} else {
			for (int i = 0; i <= this.getEntries().getLast(); i++) {
				if (this.getEntry(i) == nodea) {
					if (nodeb instanceof RTreeNode) {
						nodeb.setParent(nodea.getParent());
					}
					this.setEntry(nodeb, i);
				}
			}
		}
	}

	protected Object[] pickSeedsQuadratic() {
		Object[] seeds = new Object[2];
        // For the case that the node contains only one/two entry/ies. 
        seeds[0] = this.getEntry(0);
        seeds[1] = this.getEntry(1);
		double wastedArea, maxWastedArea = 0;
		for (int i = 0; i < this.getEntries().getLast(); i++) {
			for (int j = i + 1; j <= this.getEntries().getLast(); j++) {
				wastedArea = wastedArea(this.getEntry(i), this.getEntry(j));
				// Compare wastedArea and set new values if worse,
				// ties are resolved by choosing the first pair encountered,
				// i.e. the pair who has the lowest index in the list of
				// entries.
				if (wastedArea > maxWastedArea) {
					maxWastedArea = wastedArea;
					seeds[0] = this.getEntry(i);
					seeds[1] = this.getEntry(j);
				}
			}
		}
        // Remove the seeds from the keylist and return them.
        this.deleteEntry(seeds[0]);
        this.deleteEntry(seeds[1]);
        return seeds;
	}

	protected Object pickNextQuadratic(RTreeNode node0, RTreeNode node1) {
		double enlargementDiff, maxEnlargementDiff = -1;
		Object nextEntry = null;
		// Pick the next entry to insert in one of the new nodes.
		for (int i = 0; i <= this.getEntries().getLast(); i++) {
			if (this.getEntry(i) != null) {
				enlargementDiff = Math.abs(enlargementArea(this.getEntry(i), node0)
										   - enlargementArea(this.getEntry(i), node1));
				if (enlargementDiff > maxEnlargementDiff) {
					maxEnlargementDiff = enlargementDiff;
					nextEntry = this.getEntry(i);
				}
			}
		}
		this.deleteEntry(nextEntry);
		return nextEntry;
	}
	
	protected Object[] pickSeedsLinear() {
        Object[] seeds = new Object[2];
        if (this.getBoundingBox() == null) {
            // The node contains only descendants that have no entries.
            if (this.numberOfEntries() > 1) {
                // Just select the first two because all entries are equally good
                // choices when they are empty.
                seeds[0] = this.getEntry(0);
                seeds[1] = this.getEntry(1);
                this.deleteEntry(seeds[0]);
                this.deleteEntry(seeds[1]);
                return seeds;
            } else {
                // There is only one entry.
                seeds[0] = this.getEntry(0);
                seeds[1] = null;
                this.deleteEntry(seeds[0]);
                return seeds;
            }
        } else {
    		// Choose the pair of entries who has the greatest normalized
    		// separation along any dimension.
    		int[] xEntryPair = new int[2];
    		int[] yEntryPair = new int[2];
    		double xScale = this.getBoundingBox().getWidth();
    		double yScale = this.getBoundingBox().getHeight();
    		double xSeparation;
    		double highestLow = Double.NEGATIVE_INFINITY;
    		double lowestHigh = Double.POSITIVE_INFINITY;
    		// x-dimension
    		for (int i = 0; i <= this.getEntries().getLast(); i++) {
    			if (getLowSideX(this.getEntry(i)) > highestLow) {
    				highestLow = getLowSideX(this.getEntry(i));
    				xEntryPair[0] = i;
    			}
    			if ((getHighSideX(this.getEntry(i)) < lowestHigh) &&
    				(xEntryPair[0] != i)) {
    				lowestHigh = getHighSideX(this.getEntry(i));
    				xEntryPair[1] = i;
    			}
    		}
    		xSeparation = (highestLow - lowestHigh) / xScale;
    		highestLow = Double.NEGATIVE_INFINITY;
    		lowestHigh = Double.POSITIVE_INFINITY;
    		// y-dimension
    		for (int i = 0; i <= this.getEntries().getLast(); i++) {
    			if (getLowSideY(this.getEntry(i)) > highestLow) {
    				highestLow = getLowSideY(this.getEntry(i));
    				yEntryPair[0] = i;
    			}
    			if ((getHighSideY(this.getEntry(i)) < lowestHigh) &&
    			    (yEntryPair[0] != i)) {
    				lowestHigh = getHighSideY(this.getEntry(i));
    				yEntryPair[1] = i;
    			}
    		}
    		// Compare the normalized separations along the two dimensions.
    		if (((highestLow - lowestHigh) / yScale) > xSeparation) {
    			seeds[0] = this.getEntry(yEntryPair[0]);
    			seeds[1] = this.getEntry(yEntryPair[1]);
    		} else {
    			seeds[0] = this.getEntry(xEntryPair[0]);
    			seeds[1] = this.getEntry(xEntryPair[1]);
    		}
    		// If the node has only one child, store it at index 0.
    		if ((seeds[0] == null) || (seeds[0] == seeds[1])) {
    			seeds[0] = seeds[1];
    			seeds[1] = null;
    		}
    		this.deleteEntry(seeds[0]);
    		this.deleteEntry(seeds[1]);
    		return seeds;
        }
	}
	
	protected Object pickNextLinear() {
		// Just pick the first entry and remove & return it.
		Object nextEntry = null;
		for (int i = 0; i <= this.getEntries().getLast(); i++) {
			if (this.getEntry(i) != null) {
				nextEntry = this.getEntry(i);
				break;
			}
		}
		this.deleteEntry(nextEntry);
		return nextEntry;
	}
	
	
	static private double getLowSideX(Object entry) {
		if (entry == null) {
			return Double.NEGATIVE_INFINITY;
		} else {
			if (entry instanceof SpatialComparable) {
				return ((SpatialComparable)entry).getBoundingBox().getMinX();
			} else /* if (entry instanceof RTreeNode)*/ {
				if (((RTreeNode)entry).getBoundingBox() != null) {
					return ((RTreeNode)entry).getBoundingBox().getMinX();
				} else {
					return Double.NEGATIVE_INFINITY;
				}
			}
		}
	}

	static private double getHighSideX(Object entry) {
		if (entry == null) {
			return Double.POSITIVE_INFINITY;
		} else {
			if (entry instanceof SpatialComparable) {
				return ((SpatialComparable)entry).getBoundingBox().getMaxX();
			} else /* if (entry instanceof RTreeNode)*/ {
				if (((RTreeNode)entry).getBoundingBox() != null) {
					return ((RTreeNode)entry).getBoundingBox().getMaxX();
				} else {
					return Double.POSITIVE_INFINITY;
				}
			}
		}
	}

	static private double getLowSideY(Object entry) {
		if (entry == null) {
			return Double.NEGATIVE_INFINITY;
		} else {
			if (entry instanceof SpatialComparable) {
				return ((SpatialComparable)entry).getBoundingBox().getMinY();
			} else /* if (entry instanceof RTreeNode)*/ {
				if (((RTreeNode)entry).getBoundingBox() != null) {
					return ((RTreeNode)entry).getBoundingBox().getMinY();
				} else {
					return Double.NEGATIVE_INFINITY;
				}
			}
		}
	}

	static private double getHighSideY(Object entry) {
		if (entry == null) {
			return Double.POSITIVE_INFINITY;
		} else {
			if (entry instanceof SpatialComparable) {
				return ((SpatialComparable)entry).getBoundingBox().getMaxY();
			} else /* if (entry instanceof RTreeNode)*/ {
				if (((RTreeNode)entry).getBoundingBox() != null) {
					return ((RTreeNode)entry).getBoundingBox().getMaxY();
				} else {
					return Double.POSITIVE_INFINITY;
				}
			}
		}
	}

	/**
	 * Adds an entry in this node.
	 * 
	 * @param entry The entry to be inserted.
	 */
	protected void insertEntry(Object entry) {
		// Adjust the covering rectangle.
		if (entry instanceof SpatialComparable) {
			// Adding keys
			if (this.isEmpty()) {
				this.setBoundingBox(((SpatialComparable)entry).getBoundingBox());
			} else {
				this.setBoundingBox(RTreeNode.add(this.getBoundingBox(), 
												  ((SpatialComparable)entry).getBoundingBox()));
			}
		} else /* if (entry instanceof RTreeNode)*/ {
			// Adding children
			if (this.getBoundingBox() == null) {
				this.setBoundingBox(((RTreeNode)entry).getBoundingBox());
			} else {
				this.setBoundingBox(RTreeNode.add(this.getBoundingBox(), 
												  ((RTreeNode)entry).getBoundingBox()));
			}
			((RTreeNode)entry).setParent(this);
		}
		// Insert the entry at the end of the array of entries.
		this.setEntry(entry, this.numberOfEntries());
		this.numberOfEntries.inc();
	}

	/**
	 * Insert an entry of class RTreeNode in this node next to 
	 * a node given as parameter.
	 * 
	 * @param node The node to be inserted.
	 * @param precedingNode The node after which to insert.
	 */
	protected void insertNextTo(RTreeNode node, RTreeNode precedingNode) {
		node.setParent(this);
		int i = 0;
		while (this.getEntry(i) != precedingNode) {
			if (this.getEntry(i) == null) {
				// Make room for the new entry.
				this.setEntry(this.getEntry(i + 1), i);
			}
			i++;
		}
		if ((i > 0) && (this.getEntry(i - 1) == null)) {
			this.setEntry(precedingNode, i - 1);
			this.setEntry(node, i);
		} else {
			RTreeNode tempa = node;
			RTreeNode tempb;
			while (this.getEntry(++i) != null) {
				tempb = (RTreeNode)this.getEntry(i);
				this.setEntry(tempa, i);
				tempa = tempb;
			}
			this.setEntry(tempa, i);
		}

	}

	/**
	 * Delete the entry in this node at index.
	 * @param index
	 */
	protected void deleteEntry(int index) {
		this.setEntry(null, index);
		this.numberOfEntries.dec();
		for (int i = index; i <= this.getEntries().getLast(); i++) {
			this.setEntry(this.getEntry(i + 1), i);
		}
		// Note that the covering rectangle isn't adjusted here 
		// even though it might be affected.
	}

	/**
	 * Delete the node given as parameter if found in this node.
	 * @param node
	 */
	protected void deleteEntry(Object node) {
		if (node ==  null) {
			return;
		} else {
			for (int i = 0; i <= this.getEntries().getLast(); i++) {
				if (this.getEntry(i) == node) {
					this.deleteEntry(i);
				}
			}
		}
	}

	/**
	 * Choose a leaf in which it would be best to insert the key
	 * given as parameter.
	 * 
	 * @param key The key to be inserted.
	 * @return The node to which the key should be inserted.
	 */
	protected RTreeNode chooseLeaf(SpatialComparable key) {
		if (this.isLeaf()) {
			return this;
		} else {
			// Choose the node among the children 
			// that needs least enlargement to include key.
			RTreeNode chosenNode = null;
			double enlargement, minEnlargement = Double.POSITIVE_INFINITY;
			for (int i = 0; i < this.numberOfEntries(); i++) {
				enlargement = enlargementArea(key, (RTreeNode)this.getEntry(i));
				if (enlargement < minEnlargement) {
					chosenNode = (RTreeNode)this.getEntry(i);
					minEnlargement = enlargement;
				} else if (enlargement == minEnlargement) {
					// Resolve ties by choosing the entry with the smallest area.
					if (((RTreeNode)this.getEntry(i)).calculateArea() < chosenNode.calculateArea()) {
						chosenNode = (RTreeNode)this.getEntry(i);
					} /* else
						// The area of the chosen node is smaller or they are equal
					    // -> do nothing, i.e. ties are resolved by choosing the node
					    // that was encountered first.
					}*/
				}
			}
			return chosenNode.chooseLeaf(key);
		}
	}
	
	/**
	 * @return The height of the tree whose root this node is.
	 */
	private int height() {
		if (this.isLeaf()) {
			return 1;
		} else {
			return (1 + ((RTreeNode)this.getEntry(0)).height());
		}
	}
	
	/**
	 * Choose the node to which to insert the subtree, whose root node
	 * is given as parameter.
	 * 
	 * @param subTree The subtree to insert.
	 * @return the node where the subtree should be inserted.
	 */
	protected RTreeNode chooseNode(RTreeNode subTree) {
		return this.chooseNode(subTree, subTree.height(), this.height());
	}

	private RTreeNode chooseNode(RTreeNode subTree, int subTreeHeight, int treeHeight) {
		if (treeHeight == subTreeHeight + 1) {
			return this;
		} else {
			RTreeNode chosenNode = null;
			double area, enlargement = Double.POSITIVE_INFINITY;
			for (int i = 0; i < this.numberOfEntries(); i++) {
				area = enlargementArea(subTree, (RTreeNode)this.getEntry(i));
				if (area < enlargement) {
					chosenNode = (RTreeNode)this.getEntry(i);
					enlargement = area;
				} else if (area == enlargement) {
					// Resolve ties by choosing the entry with the smallest area.
					if (((RTreeNode)this.getEntry(i)).calculateArea() < chosenNode.calculateArea()) {
						chosenNode = (RTreeNode)this.getEntry(i);
					} /* else
					// The area of the chosen node is smaller or they are equal
				    // -> do nothing, i.e. ties are resolved by choosing the node
				    // that was encountered first.
				    }*/
				}
			}
			return chosenNode.chooseNode(subTree, subTreeHeight, treeHeight - 1);
		}
	}
	
	/**
	 * Calculates the area of a rectangle that covers both the entry and the node's
	 * current coverage minus the node's current area.
	 */
	static public double enlargementArea(Object entry, RTreeNode node) {
		if (entry instanceof SpatialComparable) {
			return (RTreeNode.area(RTreeNode.add(node.getBoundingBox(), 
								 			    ((SpatialComparable)entry).getBoundingBox())) -
  	 			    node.calculateArea());
		} else /*if (entry instanceof RTreeNode*/{
			if (((RTreeNode)entry).getBoundingBox() == null) {
				return 0;
			} else {
				if (node.getBoundingBox() == null) {
					return RTreeNode.area(((RTreeNode)entry).getBoundingBox());
				} else {
					return (RTreeNode.area(RTreeNode.add(node.getBoundingBox(), ((RTreeNode)entry).getBoundingBox())) -
							node.calculateArea());
				}
			}
		}
	}

	/**
	 * @return The size of the area that would be wasted if entries e0 and e1 would
	 * be put in a node together.
	 */
	static protected double wastedArea(Object e0, Object e1) {
		if ((e0 == null) || (e1 == null)) {
			return 0;
		} else if (e0 instanceof SpatialComparable) /*&& (e1 instanceof SpatialComparable)*/ {
			SpatialComparable k0 = (SpatialComparable)e0;
			SpatialComparable k1 = (SpatialComparable)e1;
			return RTreeNode.area(RTreeNode.add(k0.getBoundingBox(), k1.getBoundingBox())) -
				   RTreeNode.area(k0.getBoundingBox()) -
			       RTreeNode.area(k1.getBoundingBox());
		} else /*if (e0 instanceof RTreeNode*/ {
			RTreeNode n0 = (RTreeNode)e0;
			RTreeNode n1 = (RTreeNode)e1;
			if ((n0.getBoundingBox() == null) ||
				(n1.getBoundingBox() == null)) {
				// No area is wasted if both or the other one has an area of zero.
				return 0;
			} else {
				return RTreeNode.area(RTreeNode.add(n0.getBoundingBox(), n1.getBoundingBox())) -
					   RTreeNode.area(n0.getBoundingBox()) -
				       RTreeNode.area(n1.getBoundingBox());
			}
		}
	}
	
	/**
	 * @return The first leaf found that contains the key or null if not found.
	 */
	private RTreeNode search(SpatialComparable key, boolean deleteEnabled) {
		if (this.isLeaf()) {
			for (int i = 0; i < this.numberOfEntries(); i++) {
				if (((SpatialComparable)this.getEntry(i)).equals(key)) {
					// The key was found
					// -> delete the key if required.
					if (deleteEnabled) {
						this.deleteEntry(i);
					}
					// -> return the node containing the key
					return this;
				}
			}
			// The key wasn't found in the leaf -> search failed.
			return null;
		} else {
			// Search those children that could contain the key.
			for (int i = 0; i < this.numberOfEntries(); i++) {
				RTreeNode result;
				// Check if the key is contained in the node's bounding box.
				if (RTreeNode.contains(((RTreeNode)this.getEntry(i)).getBoundingBox(), (key.getBoundingBox()))) {
					// The key could be in this child -> search deeper
					if ((result = ((RTreeNode)this.getEntry(i)).search(key, deleteEnabled)) != null) {
						// The key was found in the child
						return result;
					}
					// The key wasn't in this child -> check the rest.
				}
			}
			// No child could contain the key -> search failed.
			return null;
		}
	}
	
	public RTreeNode search(SpatialComparable key) {
		return this.search(key, false);
	}
	
	private void condenseTree() {
		this.condenseTree(new ArrayList());
	}
	
	/**
	 * Calculate a bounding box for this node by iterating through its entries.
	 * 
	 * @return The bounding box.
	 */
	protected Rectangle2D calculateBoundingBox() {
		Rectangle2D box = null;
		if (this.isLeaf()) {
			for (int i = 0; i <= this.getEntries().getLast(); i++) {
				if (this.getEntry(i) != null) {
					if (box == null) {
						box = ((SpatialComparable)this.getEntry(i)).getBoundingBox();
					} else {
						box = RTreeNode.add(box, ((SpatialComparable)this.getEntry(i)).getBoundingBox());
					}
				}
			}
		} else {
			for (int i = 0; i <= this.getEntries().getLast(); i++) {
				if (this.getEntry(i) != null) {
					if (((RTreeNode)this.getEntry(i)).getBoundingBox() != null) {
						if (box == null) {
							box = ((RTreeNode)this.getEntry(i)).getBoundingBox();
						} else {
							box = RTreeNode.add(box, ((RTreeNode)this.getEntry(i)).getBoundingBox());
						}
					}
				}
			}
		}
		return box;
	}

	/**
	 * Create and return a rectangle that is the union of rectangles r0 and r1 
	 * given as parameters, that is, the smallest rectangle tha contains both 
	 * r0 and r1.
	 */
	protected static Rectangle2D add(Rectangle2D r0, Rectangle2D r1) {
		if (r0 == null) {
			return r1;
		} else if (r1 == null) {
			return r0;
		} else {
			return r0.createUnion(r1);
		}
	}

	/**
	 * calculate and return the area of the rectangle given as parameter.
	 */
	protected static double area(Rectangle2D r) {
		return (r.getHeight() * r.getWidth());
	}
	
	/**
	 * Test whether rectangle r0 contains r1, that is, their intersection
	 * equals r1.
	 * @return True if r0 contains r1, false otherwise.
	 */
	protected static boolean contains(Rectangle2D r0, Rectangle2D r1) {
		return r1.createIntersection(r0).equals(r1);		
	}
	
	
	
	/**
	 * Condense the tree by removing underfull nodes and reinserting
	 * thos entries that were stored n them.
	 * 
	 * @param eliminatedNodes The entries that have yet to be reinserted.
	 */
	private void condenseTree(ArrayList eliminatedNodes) {
		if (this.isRoot()) {
			for (int i = 0; i < eliminatedNodes.size(); i++) {
				RTreeNode node = (RTreeNode)eliminatedNodes.get(i);
				if (node.isLeaf() && node.isUnderfull()) {
					// Reinsert the keys from the eliminated node.
					this.insert(node);
				} else {
					// Reinsert the subtree.
					this.insertData(node);
				}
			}
		} else {
			if (this.numberOfEntries() < this.min) {
				// The node is under-full -> remove it.
				this.getParent().deleteEntry(this);
				if (this.isLeaf()) {
					eliminatedNodes.add(this);
				} else {
					for (int i = 0; i < this.numberOfEntries(); i++) {
						eliminatedNodes.add(this.getEntry(i));
					}
				}
			} else {
				// Adjust the covering rectangle.
				this.setBoundingBox(this.calculateBoundingBox());
			}
			// Advance to the parent.
			this.getParent().condenseTree(eliminatedNodes);
		}
	}

	/**
	 * Find and delete the key given as parameter.
	 * 
	 * @param key The key to be removed.
	 * @return The possibly new root.
	 */
	private RTreeNode deleteKey(SpatialComparable key) {
		// Find and _remove_ the key
		RTreeNode node = this.search(key, true);
		if (node == null) {
			// The key wasn't found -> return the unchanged root
			return this;
		} else {
			node.condenseTree();
			// If the root has only one child left
			// -> delete it and set the child as new root
			if (!this.isLeaf() && (this.numberOfEntries() == 1)) {
				((RTreeNode)this.getEntry(0)).setParent(null);
				return (RTreeNode)this.getEntry(0);
			} else {
				return this;
			}
		}
	}

	
	
	// -- Overridden methods inherited from class Object --

	public boolean equals(Object obj) {
		if (obj == null) {
			return false;
		} else if (obj instanceof RTree) {
			return this.equals(((RTree)obj).getElement());
		} else if (obj instanceof RTreeNode) {
			RTreeNode node = (RTreeNode)obj;
			// Compare references.
			if (this == obj) {
				return true;
			} else {
				// Compare entries.
				if ((this.getMinEntries() == node.getMinEntries())  &&
					(this.getMaxEntries() == node.getMaxEntries()) &&
					(this.numberOfEntries() == node.numberOfEntries())) {
					for (int i = 0; i <= this.getEntries().getLast(); i++) {
						if (this.getEntry(i) == null) {
                            if (node.getEntry(i) != null) {
                                // Found a difference.
                                return false;
                            }
						} else {
                            if (!this.getEntry(i).equals(node.getEntry(i))) {
                                // Found a difference.
                                return false;
                            }
                        }
					}
                    // Entries match.
                    return true;
				} else {
					return false;
				}
			}
		} else {
			return false;
		}
	}

	public String toString() {
		return "\n" + this.toString("\n ");
	}

	private String toString(String indent) {
		String[] children = new String[this.getEntries().getLast() + 1];
		int lastIndex = this.getEntries().getLast();
		if (this.isLeaf()) {
			if (this.getElement() == null) {
				return "-o";
			} else {
				String str = "-[ ";
				for (int i = 0; i <= lastIndex; i++) {
					if (this.getEntry(i) == null) {
						str += "_ ";
					} else {
						str += this.getEntry(i).toString() + " ";
					}
				}
				return str + "]";
			}
		} else {
			for (int i = 0; i < lastIndex; i++) {
				if (this.getEntry(i) == null) {
					children[i] = "-o";
				} else {
					children[i] = ((RTreeNode)this.getEntry(i)).toString(indent + "|  ");
				}
			}
			if (this.getEntry(lastIndex) == null) {
				children[lastIndex] = "-o";
			} else {
				children[lastIndex] = ((RTreeNode)this.getEntry(lastIndex)).toString(indent + "   ");
			}
			String str;
			str = "-o";
			for (int i = 0; i < lastIndex; i++) {
				str += indent + "|-" + children[i];
			}
			str += indent + "\\-" + children[lastIndex] + indent;
			return str;
		}
	}
	
	private static final long serialVersionUID = -1824041007013989293L;
}