package matrix.structures.spatial.FDT.probe;

import java.awt.geom.Rectangle2D;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.geom.Area;
import java.io.Serializable;

import matrix.structures.spatial.PaintingStyleDecorator;
import matrix.structures.spatial.SpatialComparable;
import matrix.structures.util.MatrixComparable;
import matrix.structures.FDT.probe.LexInf;
import matrix.structures.FDT.FDT;
import matrix.structures.memory.VirtualObject;

/**
 * Abstract base class for all reference implementations of different spatial data elements in 
 * Matrix. Contains all functionality common to the spatial elements. 
 * 
 * @author Jussi Nikander
 *
 */

public abstract class SpatialElement implements SpatialComparable, MatrixComparable, FDT, Serializable {

	private VirtualObject /*PaintingStyleDecorator*/ paintingStyleDecorator = new VirtualObject(null, "PaintingStyleDecorator");
	protected static int idCounter = 0;
	public static final String EMPTY = "";

	public static final double EQUALITY_THRESHOLD = 0.01; // the threshold where two doubles are considered equal
	private int id;
	private VirtualObject comparison, identity, visualization;
    
    
    /**
     * The following two variable define whether a spatial element
     * is selectable or draggable from within a VisualArea
     * 
     * This "unorthodox" approach is required because spatial elements
     * of same the type can have different properties. Therefore an approach
     * based on VisualTypeConf does not work.
     */
    
    private boolean selectable = false;
    private boolean draggable  = false;
    
	//private String comparisonString, identityString, visualizationString;
	
	public SpatialElement() {
		this("");
	}
	
	public SpatialElement(String key) {
		id = idCounter++;
		if(key == null) {
			comparison = new VirtualObject(EMPTY, this, "comparison");
			visualization = new VirtualObject(EMPTY, this, "visualization");
			identity = new VirtualObject(EMPTY + id, this, "identity");
		} else {
			comparison = new VirtualObject(key, this, "comparison");
			visualization = new VirtualObject(key, this, "visualization");
			identity = new VirtualObject(key + id, this, "identity");
		}
	}

	protected int getId() {
		return id;
	}
	
	public String getComparisonString() {
		return (String)comparison.getObject();
	}

	public void setComparisonString(String s) {
		if(s == null) {
			comparison.setObject(EMPTY);
		} else {
			comparison.setObject(s);
		}
	}
	
	public String getIdentityString() {
		return (String)identity.getObject();
	}

	public void setIdentityString(String s) {
		if(s == null) {
			identity.setObject(EMPTY+ id);
		} else {
			identity.setObject(s+id);
		}
	}
	
	public String getVisualizationString() {
		return (String)visualization.getObject();
	}

	public void setVisualizationString(String s) {
		if(s == null) {
			visualization.setObject(EMPTY);
		} else {
			visualization.setObject(s);
		}
	}

	public void setKey(Object o) {
		String s = EMPTY;
		if(o != null) {
			s = o.toString();
		}
		setComparisonString(s);
		setIdentityString(s);
		setVisualizationString(s);
	}
	
	public boolean equals(double a, double b) {
		return Math.abs(a-b) < EQUALITY_THRESHOLD;
	}

	public boolean spatialEquals(Object o) {
        if(o instanceof Point) {
            return false; // spatialEquals overridden in Point
        } else if(o instanceof SpatialComparable) {
			Area a = new Area(getShape());
			Area ca = new Area(((SpatialComparable)o).getShape());
			return a.equals(ca);
		}
		return false;
	}

	public abstract boolean intersects(SpatialComparable sc);

	public abstract boolean contains(SpatialComparable sc);
	
	public abstract Shape getShape();

	public Rectangle2D getBoundingBox(){
        Rectangle bounds = getShape().getBounds();
        return new Rectangle(bounds.x, bounds.y, bounds.width, bounds.height);

    }

    public boolean equals(SpatialElement k) {
    	if (comparison == null)
            return false;            
        return eq(k);
    }
 
	
    public boolean equals(Object k) {
        if (k instanceof SpatialElement) {
            return equals((SpatialElement)k);
        } else {
        	return false;
        }
    }

	
	public boolean eq(MatrixComparable k) {
	       if (k == null) {
	            return (this.getComparisonString().equals(""));	    	   
	       }
	        if(k instanceof LexInf) {
	            return false;
	        }
	        return (this.getComparisonString().equals(k.getComparisonString()));
	}

	public boolean lt(MatrixComparable k) {
        if (k == null) {
            return false; 
        }
        if(k instanceof LexInf) {
            return true;
        }
        return (this.getComparisonString().compareTo(k.getComparisonString()) < 0);
	}

	public boolean gt(MatrixComparable k) {
        if (k == null) {
            return (this.getComparisonString().equals("")); 
        }
        if(k instanceof LexInf) {
            return false;
        }
        return (this.getComparisonString().compareTo(k.getComparisonString()) > 0);
	}

	public boolean leq(MatrixComparable k) {
		if(k instanceof LexInf) {
			return true;
		}
		return (this.eq(k) || this.lt(k));
	}

	public boolean geq(MatrixComparable k) {
        if(k instanceof LexInf) {
            return false;
        }
        return (this.eq(k) || this.gt(k));
	}
	
	public Object getElement() {
		return this;
	}

	public void setElement(Object o) {
		
	}

    public boolean hasPaintingStyleDecorator(){
        return (paintingStyleDecorator.getObject() != null);
    }

    public void setPaintingStyleDecorator(PaintingStyleDecorator psd){
        paintingStyleDecorator.setObject(psd);
    }

    public PaintingStyleDecorator getPaintingStyleDecorator(){
        return (PaintingStyleDecorator)paintingStyleDecorator.getObject();
    }

// Focusable / draggable
    
    public boolean isDraggable(){
        return draggable;
    }
    
    public void setDraggable(boolean value){
        draggable = value;
    }

    
	public String toString() {
		return getVisualizationString();
	}
	
	
}
