/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package matrix.uitools;

import javax.swing.*;

import java.awt.Toolkit;
import java.awt.event.*;
import java.lang.reflect.*;
import java.util.*;

import matrix.util.Note;
import matrix.util.configuration.*;

/**
 * MenuFactory is an <strong>singleton</strong> class where
 * prototypes for all menus defined in the Matrix configuration
 * file are stored. The MenuFactory is capable of creating
 * new instances of these menus from the menu prototypes
 * it holds.
 */

public class MenuFactory {

    private static MenuFactory theInstance;
    private Vector menus;
    private HashMap radioButtonGroups;
    private HashMap globals;
    private Initializator initializator;

    // ---- static methods ----

    /**
     * Returns the instance of MenuFactory
     */
    public static MenuFactory getInstance() {
        if(theInstance == null) {
            theInstance = new MenuFactory();
        }
        return theInstance;
    }

    private MenuFactory() {
        menus = new Vector();
        initializator = Initializator.getInstance();
        globals = new HashMap();
    }

    /**
     * Adds the prototype of a new <code>MenuBar</code>
     * to this class. The different menubars are identified
     * by the type of anchor class they have.
     *
     * For more information about anchor classes see documentation
     * on the Matrix configuration file.
     */
    public void addMenuBar(CMenuBar mb) {
        menus.add(mb);
    }

    /**
     * Adds the prototype of a new <code>PopupMenu</code>
     * to this class. The different popupmenus are identified
     * by the type of anchor class they have.
     *
     * For more information about anchor classes see documentation
     * on the Matrix configuration file.
     */
    public void addPopupMenu(CPopupMenu mb) {
        menus.add(mb);
    }

    // ---- menubar creation methods ----

    /**
     * Creates a menubar for the parameter
     * object. The parameter object must be an instance of the anchor
     * class of the required menubar. The created menubar corresponds to
     * the given instance of the parameter object.
     *
     * @param obj the object for which the menubar is created
     */
    public JMenuBar getMenuBar(Object obj) {
        radioButtonGroups = new HashMap();
        //Note.out(this, "Fetching menubar for " + obj);
        ConfItem menubar = null;
        // Since new entries to the menus Vector are added to the
        // end of the Vector, we iterate through the Vector's 
        // members from the end to the start.
        int i = menus.size();
        while(i > 0) {
            menubar = (ConfItem) menus.get(--i);
            if(menubar.getItemType() == ConfItem.MENUBAR) {
                Class anchor = ((CMenuBar)menubar).getAnchor();
                if(anchor.isInstance(obj)) {
                    break;
                }
            }
            menubar = null;
        }
        if(menubar == null) {
            Note.out(this, "Could not find menubar for object " + obj);
            return null;
        }
        return createMenuBar((CMenuBar)menubar, obj);
    }

    private JMenuBar createMenuBar(CMenuBar menuBar, Object obj) {
        JMenuBar mb = new JMenuBar();
        for(int i = 0; i < menuBar.getChildCount(); i++) {
            ConfItem child = menuBar.getChild(i);
            if(child.getItemType() == ConfItem.MENU) {
                mb.add(createMenu((CMenu)child, obj));
            } else if(child.getItemType() == ConfItem.INITIALIZATION) {
                Object[] objs = initializator.initialize((CInitialization)child, obj);
                for(int j = 0; j < objs.length; j++) {
                    if(objs[j] instanceof JMenu) {
                        mb.add((JMenu)objs[j]);
                    } else {
                        Note.out(this, "Intialization returned a non-Menu object " + objs[j].toString() + ". Ignoring.");
                    }
                }
            } else {
                Note.err(this, "Illegal child " + child.toString() + " of ConfItem CMenuBar.");
            }
        }
        return mb;
    }

    private JMenu createMenu(CMenu cMenu, Object obj) {
        String menuName = cMenu.getName();
        JMenu menu = new JMenu(menuName);
        /*This sets the menu to heavyweight popup menu. This makes the menus to be drawn
         * always on top. Without this, structures are drawn on top of the open menus when 
         * structures are highlighted.*/
        menu.getPopupMenu().setLightWeightPopupEnabled(false);
        // add mnemonic
        char mnemonic = cMenu.getMnemonic();
        if ((int)mnemonic != 0)
            menu.setMnemonic((int)mnemonic);
        for(int i = 0; i < cMenu.getChildCount(); i++) {
            ConfItem child = cMenu.getChild(i);
            switch(child.getItemType()) {
            case ConfItem.MENU:
                menu.add(createMenu((CMenu)child,obj));
                break;
            case ConfItem.MENUITEM:
                menu.add(createMenuItem((CMenuItem)child,obj));
                break;
            case ConfItem.INITIALIZATION:
                Object[] objs = initializator.initialize((CInitialization)child,obj);
                for(int j = 0; j < objs.length; j++) {
                    if(objs[j] instanceof JMenuItem) {
                        menu.add((JMenuItem)objs[j]);
                    } else {
                        Note.out(this, "Initialization returned a non-MenuItem object " + objs[j] + ". Ignoring.");
                    }	
                }
                break;
            case ConfItem.MENUSEPARATOR:
                menu.addSeparator();
                break;
            default:
                Note.err(this, "Illegal child component of menu " + child.toString());
            }
        }
        return menu;
    }

    private JMenuItem createMenuItem(CMenuItem mi, Object obj) {
        String type = mi.getType();
        String name = mi.getName();
        JMenuItem item = null;
        boolean checkbox = false;
        // checking if we need to use a JCheckBoxMenuItem
        if(type != null && type.equals("checkbox")) {
            checkbox = true;
            item = new JCheckBoxMenuItem(name);
            if(mi.getChecked()) {
                ((JCheckBoxMenuItem)item).setState(true);
            }
        } else if (type != null && type.equals("radio")) { //we need to use JRadioButtonMenuItem
            item = new JRadioButtonMenuItem(name);
            if (mi.getChecked()) {
                ((JRadioButtonMenuItem)item).setSelected(true);
            }
        } else {
            item = new JMenuItem(name);
        }
        if (mi.getAction() != null) {
            Action a = ActionFactory.getInstance().createAction(mi.getAction(), obj);
            item.setAction(a);
        }

        // add shortcutkey
        int ac = mi.getAccessKey();
        if(ac >= 0) {
            item.setAccelerator(KeyStroke.getKeyStroke(ac, 
                    Toolkit.getDefaultToolkit().getMenuShortcutKeyMask()));
        }
        // add mnemonic
        char mnemonic = mi.getMnemonic();
        if ((int)mnemonic != 0)
            item.setMnemonic((int)mnemonic);
        // see if disabled
        if(mi.getDisabled()) {
            item.setEnabled(false);
            return item;
        }

        // if there are no children, we stop here
        if(mi.getChildCount() < 1) {
            item.setEnabled(false);
            return item;
        }
        // a radiobutton needs the group to which it belongs
        if(type != null && type.equals("radio")) {
            String groupName = mi.getGroup();
            ButtonGroup buttonGroup = (ButtonGroup)radioButtonGroups.get(groupName);
            if (buttonGroup == null) {
                buttonGroup = new ButtonGroup();
                radioButtonGroups.put(groupName, buttonGroup);
            }
            buttonGroup.add(item); //adding radio button to the button group it belongs to
        }

        // A MenuItem has two possible children: a checked and a method.
        // We handle these in a for loop, though the item order is
        // set. Since MenuItem has 1 or 2 children (if the method reaches
        // here) it is easier this way.
        for(int i = 0; i < mi.getChildCount(); i++) {
            ConfItem child = mi.getChild(i);
            if(child.getItemType() == ConfItem.CHECKED) {
                if(item instanceof JCheckBoxMenuItem) {
                    ((JCheckBoxMenuItem)item).setState(getCheckedValue((CChecked)child,obj));
                }
            } else if(child.getItemType() == ConfItem.METHOD) {
                Object[] params = initializator.createParams(child,obj);
                Method m = initializator.createMethod((CMethod)child, params,obj);
                if(m == null) {
                    item.setEnabled(false);
                    return item;
                }
                if(checkbox) { //This handles JCheckBoxMenuItems
                    ItemListener il = null;
                    if(mi.isGlobal()) {
                        Vector v = (Vector)globals.get(mi);
                        if(v == null) {
                            v = new Vector();
                            globals.put(mi,v);
                        }
                        il = new GlobalItemListener(m,params,obj,(JCheckBoxMenuItem)item,v);
                        v.add(item);
                    } else {
                        il = new MatrixItemListener(m,params,obj,(JCheckBoxMenuItem)item);
                    }
                    ((JCheckBoxMenuItem)item).addItemListener(il);
                } else { //This handles MenuItems and JRadioButtonMenuItems.
                    item.addActionListener(new MatrixActionListener(m,params,obj));
                }
            }
        }
        return item;
    }

    private boolean getCheckedValue(CChecked check, Object obj) {
        ConfItem child = check.getChild(0);
        Boolean ret = null;
        if(child.getItemType() == ConfItem.METHOD) {
            Object[] params = initializator.createParams(child,obj);
            Method m = initializator.createMethod((CMethod)child,params,obj);
            ret = (Boolean)initializator.invokeMethod(m,params,obj);
            if(ret != null)
                return ret.booleanValue();
            return false;
        } else if(child.getItemType() == ConfItem.FIELD) {
            Field f = ((CField)child).getField();
            try {
                ret = (Boolean)f.get(obj);
                return ret.booleanValue();
            } catch(IllegalAccessException e) {
                Note.err(this,"Could not access field " +f);
                return false;
            } catch(ClassCastException e) {
                Note.out(this, "Non-boolean field for checked");
                return false;
            }
        }
        return false;
    }


    // ---- popupmenu creation methods ----

    /**
     * Fetches the popupmenu for the given object. The popupmenu
     * creation routine looks through the whole given popupmenu
     * hierarchy and select those MenuItems that have a corresponding
     * functionality in the parameter object. Those menuitems that
     * do not map to the parameter object are not shown in the
     * resulting popupmenu. The parameter object must be an instance of
     * the anchor class of the required menubar.
     *
     * @param obj the Object for which the popupmenu is created.
     */
    public JPopupMenu getPopupMenu(Object obj) {
        radioButtonGroups = new HashMap();
        Note.out(this, "Fetching popupmenu for " + obj);
        ConfItem popupmenu = null;
        int i = menus.size();
        while(i>0) {
            popupmenu = (ConfItem) menus.get(--i);
            if(popupmenu.getItemType() == ConfItem.POPUPMENU) {
                Class anchor = ((CPopupMenu)popupmenu).getAnchor();
                if(anchor.isInstance(obj)) {
                    break;
                }
            }
            popupmenu = null;
        }
        if(popupmenu == null) {
            Note.out(this, "Could not find popupmenu for " + obj);
        }
        return createPopupMenu((CPopupMenu)popupmenu,obj);
    }

    private JPopupMenu createPopupMenu(CPopupMenu ppm, Object obj) {
        JPopupMenu pm = new JPopupMenu();
        for(int i = 0; i < ppm.getChildCount();i++) {
            ConfItem child = ppm.getChild(i);
            if(child.getItemType() == ConfItem.MENU) {
                JMenu m = createMenu((CMenu)child,obj);
                if(m.getItemCount() > 0) {
                    pm.add(m);
                }
            } else if(child.getItemType() == ConfItem.MENUITEM) {
                JMenuItem mi = createMenuItem((CMenuItem)child,obj);
                if(mi.isEnabled()) {
                    pm.add(mi);
                }
            } else if(child.getItemType() == ConfItem.INITIALIZATION) {
                Object[] objs = initializator.initialize((CInitialization)child,obj);
                for(int j = 0; j < objs.length; j++) {
                    if(objs[j] instanceof JMenu) {
                        if(((JMenu)objs[j]).getItemCount() > 0) {
                            pm.add((JMenu)objs[j]);
                        }
                    } else if(objs[j] instanceof JMenuItem /*&& ((MenuItem)objs[j]).isEnabled()*/) {
                        pm.add((JMenuItem)objs[j]);
                    } else {
                        if (objs[j]!=null)
                            Note.out(this, "Non-menuitem " + objs[j] + " ignored");
                    }
                }
            } else if(child.getItemType() == ConfItem.MENUSEPARATOR) {
                pm.addSeparator();
            }
        }
        pruneUnused(pm);
        return pm;
    }

    private void pruneUnused(JMenu m) {
        int i = 0;
        while(i < m.getItemCount()) {
            JMenuItem mi = m.getItem(i);
            if(!mi.isEnabled()) {
                //Note.out(this, i + "Pruning unused menuitem " + mi);
                m.remove(mi);
            } else if(mi instanceof JMenu) {
                pruneUnused((JMenu)mi);
                if(((JMenu)mi).getItemCount() <= 0) {
                    m.remove(mi);
                    //Note.out(this, i + "Pruning unused menu " + mi);
                } else {
                    i++;
                }
            } else {
                //Note.out(this, i + "Not pruning " + mi);
                i++;
            }
        }
    }

    private void pruneUnused(JPopupMenu m) {
        int i = 0;
        while(i < m.getComponentCount()) {
            if (m.getComponent(i) instanceof JMenuItem) {
                JMenuItem mi = (JMenuItem)m.getComponent(i);
                if(!mi.isEnabled()) {
                    //Note.out(this, i + "Pruning unused menuitem " + mi);
                    m.remove(mi);
                } else if(mi instanceof JMenu) {
                    pruneUnused((JMenu)mi);
                    if(((JMenu)mi).getItemCount() <= 0) {
                        m.remove(mi);
                        //Note.out(this, i + "Pruning unused menu " + mi);
                    } else {
                        i++;
                    }
                } else {
                    //Note.out(this, i + "Not pruning " + mi);
                    i++;
                }
            } else {
                i++;
            }
        }
    }


    // ---- inner classes start here ----

    /**
     * XXX Not a good implementation!
     *
     * This is an implementation of "global" switches, where toggling one switch will cause
     * all of them to be switched. This is a bad implementation which leaks memory. Even if
     * the frame to which a menuitem is stored is removed, the menuitem is still stored in
     * the vector global, which is shared between several GlobalItemListeners. Therefore,
     * a link to the CheckboxMenuItem is always stored, and the MenuItem is never destroyed.
     * This might even cause the whole menu structure to be stored after the Frame to which
     * it belongs is destroyed.
     */
    class GlobalItemListener extends MatrixItemListener {
        private Vector global;

        public GlobalItemListener(Method m, Object[] par, Object obj, JCheckBoxMenuItem i, Vector g) {
            super(m,par,obj,i);
            global = g;
        }

        public void itemStateChanged(ItemEvent e) {
            super.itemStateChanged(e);
            boolean val = getMenuItem().getState();
            for(int i = 0; i < global.size();i++) {
                JCheckBoxMenuItem mi = (JCheckBoxMenuItem) global.get(i);
                mi.setState(val);
            }
        }
    }

    class MatrixItemListener extends MatrixActionListener implements ItemListener {
        private JCheckBoxMenuItem mi; 

        public MatrixItemListener(Method m, Object[] par, Object obj, JCheckBoxMenuItem i) {
            super(m,par,obj);
            mi = i;
        }

        protected JCheckBoxMenuItem getMenuItem() {
            return mi;
        }

        public void itemStateChanged(ItemEvent e) {
            actionPerformed(null);
        }
    }

    class MatrixActionListener implements ActionListener {
        private Method method;
        private Object[] params;
        private Object anchor;

        public MatrixActionListener(Method m, Object[] par, Object obj) {
            method = m;
            params = par;
            anchor = obj;
        }

        public void actionPerformed(ActionEvent e) {
            if(method == null) {
                Note.out(this, "Null method invoked");
                return;
            }
            initializator.invokeMethod(method,params,anchor);
            return;
        }

    }

}
