/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package matrix.uitools;

import java.awt.*;

import javax.swing.*;
import java.awt.event.*;
import java.awt.print.PrinterGraphics;
import java.io.PrintStream;
import java.util.*;

import matrix.animation.*;
import matrix.simulation.Picker;
import matrix.structures.FDT.FDT;
import matrix.util.*;
import matrix.util.export.MatrixGraphics;
import matrix.visual.*;

public class StructurePanel extends JPanel implements Application {
    
    public static final Insets DEFAULT_INSETS = new Insets(20,20,20,20);
    
    private ActionListener listener = null;
    
    private boolean dblbuf = false;
    
    private Animator animator;
    
    private Picker picker;
    
    protected Vector fdtList = new Vector();
    
    private VisualAnimator visualAnimator = null;
    
    private MatrixFrame frame = null;
    
    private HashMap constraints = new HashMap();
    
    private static Color bgColor = Color.lightGray;
    
    private JPopupMenu pop;
    
    private int swapSemantics = VisualType.INSERT_SEMANTICS;
    
    public void setSwapSemantics(int sem){
        swapSemantics = sem;      
    }
    
    public int getSwapSemantics(){
        return swapSemantics;      
    }
    
    public void toggleSwapSemantics(){
        swapSemantics = (swapSemantics+1)%2;      
    }
    
    public static Color getBgColor() {
        return bgColor;
    }
    
    public static void setBgColor(Color c) {
        bgColor = c;
    }
    
    // constructors
    
    
    /**
     */
    public StructurePanel(Animator animator) {
        super(new GridBagLayout());
        setBackground(getBgColor());
        
        addEmptyComponentToTheBottom();
        
        this.animator = animator;
        addPicker();
        addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent evt) {
                validateAnimator();
            }
        });
    }
    
    public void addPicker() {
        add(picker = new Picker());
    }
    
    public StructurePanel(FDT[] ss, Animator animator) {
        this(animator);
        addStructures(ss);
    }
    
    // layout
    
    
    protected void addEmptyComponentToTheBottom() {
        GridBagConstraints gbc = new GridBagConstraints();
        JPanel p = new JPanel();
        p.setBackground(StructurePanel.getBgColor());
        gbc.weightx = 0.0;
        gbc.weighty = 1.0;
        gbc.gridx = 0;
        gbc.gridy = 100;
        gbc.anchor =  GridBagConstraints.SOUTH;
        gbc.fill =  GridBagConstraints.NONE;
        super.add(p,gbc);
    }
    
    
    
    /**
     */
    public void addStructure(FDT ss) {
        fdtList.addElement(ss);
        VisualType vc = null;
        vc = RepresentationFactory.getRepresentation(ss);
        addComponent(vc);
    }
    
    /**
     * Adds structure to StructurePanel.
     * 
     * @param ss FDT which is going to be added
     * @param name name of the structure
     */
    public void addStructure(FDT ss, String name) {
        fdtList.addElement(ss);
        VisualType vc = null;
        vc = RepresentationFactory.getRepresentation(ss);
        addComponent(vc);
        vc.rename(name);
    }
    
    
    public FDT[] getStructures() {
        FDT[] tmp = new FDT[fdtList.size()];
        for(int i = 0; i < tmp.length; i++) {
            tmp[i] = (FDT)fdtList.elementAt(i);
        }
        return tmp;
    }
    
    
    public VisualType[] addStructures(FDT[] ss) {
        VisualType[] vt =  new VisualType[ss.length];
        for (int i = 0; i < ss.length; i++) {
            fdtList.addElement(ss[i]);
            vt[i] = RepresentationFactory.getRepresentation(ss[i]);
            addComponent(vt[i]);
        }
        return vt;
    }
    
    
    public VisualType[] addStructures(FDT[] ss, String[] visualizations) {
        VisualType[] vt =  new VisualType[ss.length];
        for (int i = 0; i < ss.length; i++) {
            fdtList.addElement(ss[i]);
            if (visualizations[i] != null) {
                vt[i] = RepresentationFactory.getRepresentation(ss[i], visualizations[i]);
            } else {
                vt[i] = RepresentationFactory.getRepresentation(ss[i]);
            }
            addComponent(vt[i]);
        }
        return vt;
    }
    
    /**
     * Add structures to this panel with given set of GridBagConstraints. 
     * Mixing structures added with this method and methods that do not supply
     * the GridBagConstraints may cause odd combinations. It is better to either
     * have all structures us constraints or none use them.
     * @param ss
     * @param constraints
     * @return
     */
    public VisualType[] addStructures(FDT[] ss, GridBagConstraints[] constraints) {
        VisualType[] vt =  new VisualType[ss.length];
        for (int i = 0; i < ss.length; i++) {
            fdtList.addElement(ss[i]);
            vt[i] = RepresentationFactory.getRepresentation(ss[i]);
            add(vt[i], constraints[i], gridy++);
            this.constraints.put(vt[i], constraints[i]);
        }
        return vt;
    }
    
    
    /**
     * Add structures to this panel with given set of GridBagConstraints. 
     * Mixing structures added with this method and methods that do not supply
     * the GridBagConstraints may cause odd combinations. It is better to either
     * have all structures us constraints or none use them.
     * @param ss
     * @param visualizations
     * @param constraints
     * @return
     */
    public VisualType[] addStructures(FDT[] ss, String[] visualizations, GridBagConstraints[] constraints) {
        VisualType[] vt =  new VisualType[ss.length];
        for (int i = 0; i < ss.length; i++) {
            fdtList.addElement(ss[i]);
            if (visualizations[i] != null) {
                vt[i] = RepresentationFactory.getRepresentation(ss[i], visualizations[i]);
            } else {
                vt[i] = RepresentationFactory.getRepresentation(ss[i]);
            }
            add(vt[i], constraints[i], gridy++);
            this.constraints.put(vt[i], constraints[i]);
        }
        return vt;
    }
    
    
    protected int gridy = 0;
    
    
    
    protected GridBagConstraints addComponent(JComponent c) {
        return addComponent(c, gridy++,DEFAULT_INSETS);
    }
    
    protected GridBagConstraints addComponent(JComponent c, Insets insets) {
        return addComponent(c, gridy++,insets);
    }
    
    
    
    private GridBagConstraints addComponent(JComponent c, int position, Insets insets) {
        GridBagConstraints gbc = new GridBagConstraints();
        gbc.weightx = 0.0;
        gbc.weighty = 0.0;
        gbc.insets = insets;
        gbc.gridx = 0;
        gbc.gridy = position;
        gbc.anchor =  GridBagConstraints.NORTH;
        gbc.fill =  GridBagConstraints.NONE;
        super.add(c,gbc,position);
        if(c instanceof VisualType)
            constraints.put(c, gbc);
        return gbc;
    }
    
    
    
    protected void removeComponents() {
        int count = getComponentCount();
        int firstComponent = 0;
        for (int i=0; i<count; i++) {
            if(!(getComponent(firstComponent) instanceof VisualType))
                firstComponent++;
            else
                remove(getComponent(firstComponent));
        }
        fdtList.removeAllElements();
        constraints = new HashMap();
        gridy = 0;
    }
    
    
    
    
    // getters and setters
    
    public Animator getAnimator() {
        return animator;
    }
    
    public void setAnimator(Animator a) {
        this.animator = a;
        if (getVisualAnimator() != null){
            getVisualAnimator().changeAnimator(a);
        }
    }
    
    public Picker getPicker(){
        return picker;
    }
    
    
    public void setVisualAnimator(VisualAnimator v) {
        if (visualAnimator != null) {
            throw new IllegalStateException("Visual animator can be set only once.");
        }
        this.visualAnimator = v;
        v.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent evt) {
                validate();
                fireActionEvent(evt);
            }
        });
    }
    
    
    public VisualAnimator getVisualAnimator() {
        return visualAnimator;
    }
    
    public void setDoubleBuffering(boolean b) {
        dblbuf = b;
    }
    
    
    public MatrixFrame getFrame() {
        return frame;
    }
    
    public void setFrame(MatrixFrame f) {
        frame = f;
    }
    
    // misc methods
    
    
    public void addActionListener(ActionListener a) {
        listener = AWTEventMulticaster.add(listener, a);
    }
    
    
    public void removeActionListener(ActionListener a) {
        listener = AWTEventMulticaster.remove(listener, a);
    }
    
    
    
    public void dump(PrintStream out) {
        for (int i=0; i<getComponentCount(); i++){
            if (getComponent(i) instanceof VisualType) {
                ((VisualType)getComponent(i)).dump(out);
            } else {
                out.println("Non-VisualType: "+getComponent(i));
            }
        }
    }
    
    
    
    // Double Buffering
    
    Image offscreen;
    
    public Image getOffscreen() {
        return offscreen;
    }
    
    /**
     * null out the offscreen buffer as part of invalidation
     */
    public void invalidate() {
        offscreen = null;
        super.invalidate();
    }
    
    
    
    
    
    /**
     * override update to *not* erase the background before painting
     */
    public void update(Graphics g) {
        if (dblbuf)
            paint(g);
        else
            super.update(g);
    }
    
    /**
     * paint children into an offscreen buffer, then blast entire image
     * at once.
     */
    public void paintComponent(Graphics g) {
        if (g instanceof MatrixGraphics) { // exporting
            super.paintComponent(g);
            return;
        }
        if (g instanceof matrix.util.export.ConfigurableGraphics) { // exporting
            g.setColor(Color.lightGray);
            Rectangle r = g.getClip().getBounds();
            g.fillRect(r.x,r.y,r.width,r.height);
            super.paintComponent(g);
            return;
        }
        if (g instanceof PrinterGraphics) { // just printing...
            super.paintComponent(g);
            return;
        }
        if (dblbuf) {
            //to get rid of the IllegalArgumentException caused by creating
            //an image of size 0x0
            if (getSize().width == 0 && getSize().height == 0) 
                return;
            
            if(offscreen == null) {
                offscreen = createImage(getSize().width, getSize().height);
            }
            
            Graphics og = offscreen.getGraphics();
            og.setColor(Color.lightGray);
            og.setClip(g.getClip());
            Rectangle r = g.getClip().getBounds();
            og.clearRect(r.x,r.y,r.width,r.height);
            super.paintComponent(og);
            g.drawImage(offscreen, 0, 0, this);
            og.dispose();
        } else{ // no double buffering
            Rectangle r = g.getClip().getBounds();
            g.clearRect(r.x,r.y,r.width,r.height);
            super.paintComponent(g);
        }
    }
    
    
    
    
    
    // application
    
    
    public void changeRepresentation(VisualContainer oldType, String newType) {
        int index = containerIndex(oldType);
        FDT s = oldType.getStructure();
        remove(index);
        String name = oldType.getName();
        VisualType vc =
            RepresentationFactory.getRepresentation(s, newType);
        vc.setNameOfRepresentation(newType);
        addComponent(vc, index, DEFAULT_INSETS);
        vc.setName(name);
        validate();
        repaint();
    }
    
    
    private int containerIndex(JComponent c) {
        Component[] comps = getComponents();
        for (int i = 0; i < comps.length; i++) {
            if (comps[i] == c) return i;
        }
        return -1;
    }
    
    
    
    public void showInfoMessage(VisualType vt, String info){
        Note.out(this,"started");
        InfoDialog.getInfoDialog().setInfoTxt(info);
    }
    
    
    public void removeInfoMessage(VisualType vt){
        InfoDialog.getInfoDialog().clear();
    }
    
    
    public void showMessage(String title, String message){
        JButton dismiss = new JButton("Dismiss");
        JFrame mf = getFrame();
        if (mf == null) mf = new JFrame();
        final JDialog dialog = new JDialog(mf, title, false);
        JComponent text = null;
        
        // In case of a long message, show text area. Short messages
        // can be shown in a label.
        if (message.indexOf("\n") >= 0) {
            JTextArea ta = new JTextArea(message);
            ta.setColumns(40);
            ta.setRows(20);
            JScrollPane pane = new JScrollPane(ta);
            text = pane;
        }
        else {
            text = new JLabel(" "+ message + " ");
        }
        dialog.getContentPane().add("Center", text);
        dialog.getContentPane().add("South", dismiss);
        dismiss.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) { dialog.dispose(); }
        });
        
        dialog.addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) { dialog.dispose(); }
        });
        
        dialog.pack();
        dialog.setVisible(true);
    }
    
    
    public String getInput(String title, String defaultString){
        return QueryString.ask(title, defaultString);
    }
    
    // XXX Is a final param. the right way to do this?
    public void engage(final Application app){
        app.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                validateAnimator();
                validate();
            }
        });
        this.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                app.validateAnimator();
                app.validate();
            }
        });
    }
    
    
    public void fireActionEvent(ActionEvent a){
        if (listener != null) listener.actionPerformed(a);
    }
    
    
    public Application getNewInstance(FDT[] fdt, Animator a){
        StructurePanel p = new StructurePanel(fdt, a);
        return p;
    }
    
    
    
    public JFrame getNewWindow() {
        if(getParent() instanceof MatrixFrame) {
            Note.out(this, "Parent instanceof MatrixFrame");
            return ((MatrixFrame)getParent()).getNewFrame(this);
        } else {
            Note.out(this, "Parent "+ getParent() + " not instanceof MatrixFrame");
            //return new MatrixFrame(this);
            return new MenuFrame(this);
        }
    }
    
    
    public void delete(VisualType vt){
        FDT struc = vt.getStructure();
        for(int i = 0; i < fdtList.size(); i++) {
            if(fdtList.elementAt(i) == struc) {
                fdtList.removeElementAt(i);
            }
        }
        // Here we move all components after the deleted one one
        // place up in the layout.
        int count = getComponentCount();
        int pos = 0;
        for(; pos < count; pos++) {
            if(getComponent(pos) == vt)
                break;
        }
        pos++;
        for(; pos <count; pos++) {
            Component c = getComponent(pos);
            GridBagConstraints gbc = (GridBagConstraints)constraints.get(c);
            if(gbc != null) {
                gbc.gridy = gbc.gridy - 1;
                remove(c);
                super.add(c, gbc, pos - 1);
            } else if(c instanceof VisualType) {
//              Note.err(this, "No GridBagConstraints found for " + c + ", ignoring this component");
            }
            
        }
        constraints.remove(vt);
        remove(vt);
        
        gridy--;
        invalidate();
        validate();
        repaint();
    }
    
    
    public void invalidateComponents() {
        Component[] components = getComponents();
        for(int i = 0; i < components.length; i++) {
            components[i].invalidate();
        }
    }
    
    
    public void validate(){
        //Note.out(this, "Validate");
        int count = getComponentCount();
        for (int i=0; i<count; i++) {
            getComponent(i).validate();
        }
        doLayout(); /*kstalnac: Added another doLayout call here to draw the duplicates 
        in VisualList correctly when moving animator*/
        super.validate();
        doLayout();
    }
    
    
    
    public void validateAnimator(){
        //Note.out(this, "Validating animator");
        if (visualAnimator != null) {
            visualAnimator.validate();
        }
    }
    
    public void updateReferences() {
        getAnimator().operation(new Command() {
            public void execute() {
                Validator.validate();
            }});
        if (getFrame() != null) {
            getFrame().validateComponents();
        } else {
            validate();
            validate(); /*kstalnac: added another validate call here to draw the duplicates
            in VisualList correctly after updating references*/ 
            validateAnimator();
        }
        
    }
    
    
    /**
     Disables the visual components of this container.
     */
    public void disableVisualComponents() {
        VisualType[] vt = getVisualComponents();
        for (int i=0; i<vt.length; i++) {
            vt[i].enableSubComponents(false);
        }
    }
    
    
    public VisualType[] getVisualComponents() {
        int count = 0;
        for (int i=0; i<getComponentCount(); i++) {
            if (getComponent(i) instanceof VisualType) {
                count++;
            }
        }
        VisualType[] vt = new VisualType[count];
        count = 0;
        for (int i=0; i<getComponentCount(); i++) {
            if (getComponent(i) instanceof VisualType) {
                vt[count++] = (VisualType)getComponent(i);
            }
        }
        return vt;
    }
    
    
    public void setFontSize(int size) {
        
        Font fnt = getFont();
        int style = getFont().getStyle();
        
        setFont(new Font(fnt.getName(), style, size));
        
        // notify VisualKey that fontSize changed
        matrix.visual.VisualKey.fontSizeChanged(size);
        
        validate();
    }
    
    
    public int getFontSize() {
        return getFont().getSize();
    }
    
    public void nameStructure(FDT s, String name) {
        VisualType[] vt = getVisualComponents();
        for (int i = 0; i < vt.length; i++) {
            if (vt[i].isRepresenting(s)) {
                vt[i].setName(name);
                vt[i].setInvalid();
                return;
            }
        }
    }
    
    public void setPopupMenu(JPopupMenu pop) {
        if (this.pop != null)
            this.remove(this.pop);
        this.pop = pop;
        this.add(pop);
        this.addMouseListener(new MouseListener() {
            public void mousePressed(MouseEvent e) {
                maybeShowPopup(e);
            }
            
            public void mouseReleased(MouseEvent e) {
                maybeShowPopup(e);
            }
            public void mouseClicked(MouseEvent e) {}
            public void mouseEntered(MouseEvent e) {}
            public void mouseExited(MouseEvent e) {}
        });
    }
    
    public JPopupMenu getPopupMenu() {
        return pop;
    }
    
    private void maybeShowPopup(MouseEvent e) {
        if (e.isPopupTrigger())
            pop.show(e.getComponent(), e.getX(), e.getY());
    }
    
    public void setStructures(FDT[] newStructures) {
        removeComponents();
        addStructures(newStructures);
        invalidateComponents();
        validate();
        repaint();
    }
    
}


