/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2006 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package matrix.uitools;

import java.awt.event.ActionListener;
import java.lang.reflect.Method;
import java.util.Vector;

import javax.swing.*;

import matrix.util.Note;
import matrix.util.configuration.*;

public class ToolbarFactory {
    private static ToolbarFactory theInstance;
    private Initializator initializator;
    private Vector toolbars;
    
    protected ToolbarFactory() {
        super();
        toolbars = new Vector();
        initializator = Initializator.getInstance();
    }
    
    /**
     * Returns the instance of ToolbarFactory
     */
    public static ToolbarFactory getInstance() {
        if(theInstance == null) {
            theInstance = new ToolbarFactory();
        }
        return theInstance;
    }
    
    
    /**
     * Adds the prototype of a new <code>Toolbars</code>
     * to this class. The different toolbars are identified
     * by the type of anchor class they have.
     *
     * For more information about anchor classes see documentation
     * on the Matrix configuration file.
     */
    public void addToolbars(CToolbars tb) {
        toolbars.add(tb);
        System.out.println("added toolbars "+tb);
    }
      
    public JToolBar[] getToolbars(Object obj) {
        //Note.out(this, "Fetching menubar for " + obj);
        ConfItem tbs = null;
        // Since new entries to the menus Vector are added to the
        // end of the Vector, we iterate through the Vector's 
        // members from the end to the start.
        int i = toolbars.size();
        while(i > 0) {
            tbs = (ConfItem) toolbars.get(--i);
            if(tbs.getItemType() == ConfItem.TOOLBARS) {
                Class anchor = ((CToolbars)tbs).getAnchor();
                if(anchor.isInstance(obj)) {
                    break;
                }
            }
            tbs = null;
        }
        if(tbs == null) {
            Note.out(this, "Could not find toolbars for object " + obj);
            return null;
        }
        return createToolbars((CToolbars)tbs, obj);
    }
    
    private JToolBar[] createToolbars(CToolbars tbs, Object obj) {
        JToolBar[] jtbs = new JToolBar[tbs.getChildCount()];
        for(int i = 0; i < tbs.getChildCount(); i++) {
            ConfItem child = tbs.getChild(i);
            if(child.getItemType() == ConfItem.TOOLBAR) {
                jtbs[i] = createToolbar((CToolbar)child, obj);
            } else if(child.getItemType() == ConfItem.INITIALIZATION) {
                Object[] objs = initializator.initialize((CInitialization)child, obj);
                for(int j = 0; j < objs.length; j++) {
                    if(objs[j] instanceof JToolBar) {
                        jtbs[i] = (JToolBar)objs[j];
                    } else {
                        Note.out(this, "Intialization returned a non-JToolbar object " + objs[j].toString() + ". Ignoring.");
                    }
                }
            } else {
                Note.err(this, "Illegal child " + child.toString() + " of ConfItem CToolbar.");
            }
        }
        return jtbs;
    }
    
    private JToolBar createToolbar(CToolbar cToolbar, Object obj) {
        String tbName = cToolbar.getName();
        JToolBar tb = new JToolBar(tbName);
        for(int i = 0; i < cToolbar.getChildCount(); i++) {
            ConfItem child = cToolbar.getChild(i);
            switch(child.getItemType()) {
                case ConfItem.TOOLBARITEM:
                    JComponent jc = createToolbarItem((CToolbarItem)child,obj);
                    if (jc != null)
                        tb.add(jc);
                    break;
                case ConfItem.INITIALIZATION:
                    Object[] objs = initializator.initialize((CInitialization)child,obj);                    
                    for(int j = 0; j < objs.length; j++) {
                        System.out.println("initialization results: "+objs[j]);
                        if(objs[j] instanceof JComponent && objs[j] != null) {
                            tb.add((JComponent)objs[j]);
                        } else if (objs[j] instanceof Action && objs[j] != null) {
                            tb.add(new JButton((Action)objs[j]));
                        } else {
                            Note.out(this, "Initialization returned a non-valid object for toolbar " + objs[j] + ". Ignoring.");
                        }   
                    }
                    break;
                case ConfItem.MENUSEPARATOR:
                    tb.addSeparator();
                    break;
                default:
                    Note.err(this, "Illegal child component of toolbar " + child.toString());
            }
        }
        return tb;
    }
    
    private JComponent createToolbarItem(CToolbarItem mi, Object obj) {
        JComponent item = null;
        if (mi.getAction() != null) {
            Action a = ActionFactory.getInstance().createAction(mi.getAction(), obj);
            item = new JButton(a);
            if (((JButton)item).getIcon() != null)
                ((JButton)item).setText(null);
            return item;
        }
        if (mi.getClassAttr() != null) {
            Class c = mi.getClassAttr();
            try {
                item = (JComponent) c.newInstance();
            if (obj != null && obj instanceof MenuFrame) {
                Method m = c.getMethod("setFrame", new Class[] {Class.forName("matrix.uitools.MenuFrame")});
                if (m != null)
                    m.invoke(item, new Object[] { (MenuFrame)obj });
            }           
            } catch (Exception e) {
                e.printStackTrace();
            }            
            if (item instanceof ActionListener) {
                ((MenuFrame)obj).getStructurePanel().addActionListener((ActionListener)item);
            }
            return item;
        }
        return null;
    }
}
