/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2002  Ari Korhonen and others
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package matrix.util.configuration;

import java.util.Vector;
import java.io.Serializable;

/**
 * This class is the basic class of the intermediate representation
 * of menus created by <code>MatrixConfiguration</code>. The intermediate
 * representation is stored in the class <code>matrix.uitools.MenuFactory</code>
 * and used to instantiate new instances of the various menus upon request.
 *
 * In the intermediate representation each menu is stored as a tree, where each
 * tree node contains one menu, submenu, menuitem, etc.. Knowing the specifics
 * of the intermediate representation shouldn't be necessary unless you wish to
 * expand or modify what can be stored in Matrix menus.
 *
 * The static final int fields of this class each represent one XML tag used
 * in the configuration file.
 */
public class ConfItem implements Serializable {

    public static final int CONFIGURATION = 0;
    public static final int MENUBAR = 1;
    public static final int MENU = 2;
    public static final int MENUITEM = 3;
    public static final int CHECKED = 4;
    public static final int MENUSEPARATOR = 5;
    public static final int FIELD = 6;
    //    public static final int CLASS = 7;
    public static final int CONSTRUCTOR = 8;
    public static final int METHOD = 9;
    public static final int INT = 10;
    public static final int STRING = 11;
    public static final int BOOLEAN = 12;
    public static final int CHAR = 13;
    public static final int POPUPMENU = 14;
    public static final int INITIALIZATION = 15;
    public static final int VISUALIZATIONS = 16;
    public static final int STRUCTURE = 17;
    public static final int VISUALIZATION = 18;
    public static final int PICKEDVISUALIZATIONS = 19;
    public static final int DUMMY = 20;
    public static final int ACTION = 21;
    public static final int TOOLBAR = 22;
    public static final int TOOLBARITEM = 23;
    public static final int TOOLBARS = 24;

    private Vector children;
    private int itemType;
	
    public ConfItem(int t) {
	itemType = t;
	children = new Vector();
    }
    
    /**
     * Tells what kind of menu element this ConfItem is.
     *
     * @return A numeric value corresponding to one of the valid values
     */
    public int getItemType() {
	return itemType;
    }
    
    public void setItemType(int type) {
        itemType = type;
    }
    
    /**
     * Returns the n:th child of this ConfItem.
     *
     * @param index the number of the child
     * 
     * @return the n:th child
     */
    public ConfItem getChild(int index) {
	return (ConfItem)children.get(index);
    }
    
    /**
     * Tells how many child components this ConfItem has
     *
     * @return the number of children
     */
    public int getChildCount() {
	return children.size();
    }
    
    /**
     * Adds a new child to this ConfItem. The child is added
     * to the end of the list of children.
     * 
     * @param newChild the new child to be added
     */
    public void addChild(ConfItem newChild) {
	children.add(newChild);
    }

    public String toString() {
	return "CI:" + itemType;
    }

}
