/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package matrix.util.configuration;

import java.lang.reflect.*;

import matrix.util.Note;

/**
 * Initializator is a class that holds a collection of methods used to
 * manage the methods, fields and constructors defined in the Matrix
 * Configuration file. This class holds all methods required for
 * changing the method field and constructor representations from one
 * intermediate format to another, as well as executing methods and
 * instantiating new classes (via constructors defined in
 * &lt;constructor&gt; -elements), plus creating actual run-time parameters
 * for methods and constructors.
 *
 * The first intermediate format creation is handled by the configuration
 * reader.  
 *
 * Initializator uses the <strong>Singleton</strong> pattern.
 */

public class Initializator {

    private static Initializator theInstance = null;

    public static Initializator getInstance() {
	if(theInstance == null)
	    theInstance = new Initializator();
	return theInstance;
    }

    private Initializator() {
    }


    /**
     * Creates a method from a given intermediate representation
     *
     * @param cm the intermediate representation of the method
     * @param params the parameters of the method, as returned by method createParams
     *
     * @return appropriate instance of Method or null, if no Method is found
     */
    public Method createMethod(CMethod cm, Object[] params, Object obj) {

        Class[] formal = createFormalParams(params);

        Class c  = cm.getWrappedClass();
        // if CMethod corresponds to a static method, c is now
        // non-null.
        if(c == null) {
            c = obj.getClass();
        }

        Method candidate = cm.getCandidate(c, formal);
        if (candidate != null){
            return candidate;
        }

        String mName = cm.getMethodName();


        Method meth = null;

        do{
            Method[] methods = c.getMethods();
            
            outerLoop: for(int i = 0; i < methods.length; i++) {
                
                if (!methods[i].getDeclaringClass().equals(c))
                    continue; //Not defined here -> continue
                              //(Cannot use getDeclaredMethods, which would give only the declared methods above,
                              // as it lists also private members, whih is not allowed by the security manager)                
                
                if(methods[i].getName().equals(mName)) {
                    Class[] mParams = methods[i].getParameterTypes();
                    if(mParams.length == formal.length) {
                        for(int j = 0; j < formal.length;j++) {
                            if(!mParams[j].isAssignableFrom(formal[j])) {
                                continue outerLoop;
                            }
                        }
                        if(meth == null) {
                            meth = methods[i];
                            // if return type of (old) meth is same or superclass of the return type of
                            // methods[i], we replace the old value of meth. Therefore we get the most
                            // spesific return value for meth, as per method getMethod in Class.
                        } else if(meth != null && meth.getReturnType().isAssignableFrom(methods[i].getReturnType())) {
                            meth = methods[i];
                        }
                    }
                }
            }
            if (meth != null){

                cm.addCandidate(meth, c, formal); //Added by Otto
                return meth;

            }

            c = c.getSuperclass();

        }while(c != null);

        // This method is said to return a null if no method is found. Therefore
        // this warning is not needed.
        // if(meth == null) {
        //      Note.warning(this, "Method " + mName + " not found.");
        // }
        return null;
    }


    /**
     * Creates a constructor from a given intermediate representation
     *
     * @param cc the intermediate representation of the method
     * @param params the parameters of the constructor, as returned by createParams
     *
     * @return the appropriate constructor or null if none is found
     */
    public Constructor createConstructor(CConstructor cc, Object[] params) {
	Class c = cc.getInstClass();
	Class[] formal = createFormalParams(params);
	try {
	    return c.getConstructor(formal);
	} catch(NoSuchMethodException e) {
		Constructor[] constrs = c.getConstructors();
		loop: for(int i = 0; i < constrs.length; i++) {
			Constructor co = constrs[i];
			Class[] prs = co.getParameterTypes();
			if(prs.length != formal.length) {
				continue loop;
			}
			for(int j = 0; j < prs.length; j++) {
				if(!prs[j].isAssignableFrom(formal[j])) {
					continue loop;
				}
			}
			return co;
		}
		Note.out(this, "Could not create constructor for class " + c);
	}
	return null;
    }

    /**
     * Creates an intermediate representation of parameters for a method or
     * a constructor. This method is called when a menu item that can call the
     * method or constructor in the param ConfItem is created.
     * 
     * @param ci the ConfItem holding the method or constructor
     *
     * @return an Object array holding the intermediate representation.
     */
    public Object[] createParams(ConfItem ci,Object obj) {
	Object[] params = new Object[ci.getChildCount()];
	for(int i = 0; i < params.length; i++) {
	    ConfItem child = ci.getChild(i);
	    switch(child.getItemType()) {
	    case ConfItem.METHOD:
		Object[] mpar = createParams(child,obj);
		Method me = createMethod((CMethod)child,mpar, obj);
		params[i] = new MethodWrapper(me,mpar);
		break;
	    case ConfItem.INT:
		params[i] = new Integer(((CInt)child).getInt());
		break;
	    case ConfItem.STRING:
		params[i] = ((CString)child).getString();
		break;
	    case ConfItem.FIELD:
		params[i] = ((CField)child).getField();
		break;
	    case ConfItem.BOOLEAN:
		params[i] = new Boolean(((CBoolean)child).getBoolean());
		break;
	    case ConfItem.CHAR:
		params[i] = new Character(((CChar)child).getChar());
		break;
	    case ConfItem.CONSTRUCTOR:
		Object[] cpar = createParams(child,obj);
		Constructor c = createConstructor((CConstructor)child,cpar);
		params[i] = new ConstructorWrapper(c,cpar);
		break;
	    default:
		Note.err(this, "unknown method param " + child.toString());
	    }
	}
	return params;
    }

    /**
     * Creates formal parameters required for method or constructor creation.
     * The given parameters are assumed to be in intermediate format given
     * by createParams.
     *
     * @param params the parameters with which 
     */
    public Class[] createFormalParams(Object[] param) {
	Class[] formal = new Class[param.length];
	for(int i = 0; i < formal.length; i++) {
	    if(param[i] instanceof Boolean) {
		formal[i] = Boolean.TYPE;
	    } else if(param[i] instanceof Integer) {
		formal[i] = Integer.TYPE;
	    } else if(param[i] instanceof Character) {
		formal[i] = Character.TYPE;
	    } else if(param[i] instanceof MethodWrapper) {
		formal[i] = ((MethodWrapper)param[i]).getMethod().getReturnType();
	    } else if(param[i] instanceof ConstructorWrapper) {
		formal[i] = ((ConstructorWrapper)param[i]).getConstr().getDeclaringClass();
	    } else if(param[i] instanceof Field) {
		formal[i] = ((Field)param[i]).getType();
	    } else {
		formal[i] = param[i].getClass();
	    }
	}
	return formal;
    }

    /**
     * Returns the formal Class of a parameter. Called by createMethod and
     * createConstructor.
     * 
     * @param o the parameter
     */
    private Class getParamClass(Object o) {
	if(o instanceof ConstructorWrapper) {
	    return ((ConstructorWrapper)o).getConstr().getDeclaringClass();
	} else if(o instanceof MethodWrapper) {
	    return ((MethodWrapper)o).getMethod().getDeclaringClass();
	} else {
	    return o.getClass();
	}
    }

    /**
     * Creates the actual run-time parameters for a method or a constructor. 
     * Called by initializations as well as ActionListeners listening to
     * various MenuItems.
     * 
     * @param params the parameters created by createParams method
     * @param obj the anchor object for which all non-static methods should be invoked
     *
     * @return an Object array holding the run-time parameters.
     */
    public Object[] getActualParams(Object[] params, Object obj) {
	Object[] actual = new Object[params.length];
	for(int i = 0; i < actual.length; i++) {
	    if(params[i] instanceof ConstructorWrapper) {
		ConstructorWrapper cw = (ConstructorWrapper)params[i];
		actual[i] = instantiateClass(cw.getConstr(),cw.getParams(),obj);
	    } else if(params[i] instanceof MethodWrapper) {
		MethodWrapper mw = (MethodWrapper)params[i];
		Method m = mw.getMethod();
		actual[i] = invokeMethod(m,mw.getParams(),obj);
	    } else if(params[i] instanceof Field) {
		try {
		    actual[i] = ((Field)params[i]).get(obj);
		} catch(IllegalAccessException e) {
		    Note.out(this, "Could not access field " + params[i]);
		    actual[i] = null;
		}
	    } else {
		actual[i] = params[i];
	    }
	}
	return actual;
    }

    /**
     * Invokes a method with certain parameters. It is assumed that
     * the parameters are in the intermediate representation created
     * by createParams -method. This method calls the getActualParams
     * -method before trying to invoke the parameter method.
     *
     * @param m the method to be invoked
     * @param params the method parameters.
     * @param obj the object for which the method should be invoked
     *
     * @return the return value of the method, or null if method cannot be invoked.
     */
    public Object invokeMethod(Method m, Object[] params, Object obj) {
	Object[] actual = getActualParams(params, obj);
	if(m == null) {
		return null;
	}
	try {
	    return m.invoke(obj,actual);
	} catch(Exception e) {
	    Note.out(this, "Could not call method " + m+"\n"+e);
	    e.printStackTrace();
	    return null;
	}
    }
    /**
     * Invokes the parameter constructor with the given paramters to create
     * a new instance of the class defining the Costructor object.
     * The parameters are assumed to be in the intermediate format
     * returned by createParams method.
     * 
     * @param c the constructor to be invoked
     * @param params the parameters
     * 
     * @return a new instance of the class or null if constructor could not be invoked
     */
    public Object instantiateClass(Constructor c, Object[] params,Object obj) {
	Object[] cParams = getActualParams(params,obj);
	try {
	    return c.newInstance(cParams);
	} catch(Exception e) {
	    // XXX Should have better error handling?
	    Note.out(this, "Unable to instantiate class " + c.getDeclaringClass());
	    e.printStackTrace();
	    return null;
	} 
    }

    /**
     * Runs initialization specified by an &lt;initialization&gt; -element in
     * the configuration file.
     * 
     * @param ci the intermediate representation of an initialization element
     * @param obj the anchor object for the initialization
     * 
     * @return and Object array containing the return values of each method called
     */
    public Object[] initialize(CInitialization ci, Object obj) {
	Object[] returns = new Object[ci.getChildCount()];
	for(int i = 0; i < returns.length; i++) {
	    CMethod cm = (CMethod)ci.getChild(i);
	    Object[] params = createParams(cm,obj);
	    Method m = createMethod(cm,params,obj);
	    returns[i] = invokeMethod(m,params,obj);
	}
	return returns;
    }

    // ---- Inner classes ---- 

    class ConstructorWrapper {
	private Constructor constr;
	private Object[] params;

	public ConstructorWrapper(Constructor c, Object[] o) {
	    constr = c;
	    params = o;
	}

	public Constructor getConstr() {
	    return constr;
	}

	public Object[] getParams() {
	    return params;
	}

    }

    class MethodWrapper {
	private Method meth;
	private Object[] params;

	public MethodWrapper(Method m, Object[] o) {
	    meth = m;
	    params = o;
	}

	public Method getMethod() {
	    return meth;
	}

	public Object[] getParams() {
	    return params;
	}

    }


}
