/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2002  Ari Korhonen and others
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package matrix.util.configuration;

import java.io.*;
import java.lang.reflect.Field;
import java.net.*;
import java.util.Vector;
import java.util.jar.JarFile;
import java.util.HashMap;

import java.io.ByteArrayOutputStream;
import java.io.ObjectOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ObjectInputStream;

import javax.xml.parsers.*;

import org.xml.sax.*;
import org.xml.sax.helpers.DefaultHandler;

import matrix.uitools.*;
import matrix.util.Note;
import matrix.visual.RepresentationFactory;
/**
 * This class parses a matrix configuration file and creates a configuration,
 * including all menus, for matrix. The configuration file must be an XML
 * file conforming to the matrix.dtd. The parsing is started by calling the
 * <code>parse()</code> -method.
 *
 * <code>MatrixConfiguration</code> is a singleton class. To get the instance
 * of this class, use the <code>getInstance()</code> -method.
 *
 * @author Jussi Nikander
 */

public class MatrixConfiguration extends DefaultHandler {
    
    // ---- static fields ----
    
    private static MatrixConfiguration theInstance = null;
    private static String confFile = null;
    private static String url = null; 
    private static String codebaseURL = null; 
    private static boolean validatingParser = false;
    private static ConfItem configuration = null;
    private static ConfItem dummy = new ConfItem(ConfItem.DUMMY);
    private static String jarFile = null;
    private static InputStream in = null;
    
    private static final String DEFAULT = "matrixconf.xml";
    private static final String JARFILE = "matrix.jar";
    private static final String JARCONFFILE = "matrixconf.xml";
    
    
    // ---- static methods ----
    
    /**
     * Retrieves the singleton instance of this class.
     */
    public static MatrixConfiguration getInstance() {
        if(theInstance == null) {
            theInstance = new MatrixConfiguration();
        }
        return theInstance;
    }
    
    /**
     * Sets the local path from which the configuration
     * file is read. It is best to use the full, absolute
     * path to the configuration file in order to make sure
     * that it is found. Most Java implementations are capable
     * of reading relative paths, but not necessarely all.
     *
     * @param s the path to the configuration file
     */
    public static void setConfPath(String s) {
        confFile = s;
    }
    
    public static String getConfPath() {
        return confFile;
    }
    
    /**
     * Sets the URL to the configuration file, making it
     * possible to read the configuration file from the
     * Internet. This is used by the applet version
     * of the framework to search for the configuration
     * file.
     *
     * @param s the URL to the configuration file
     */
    public static void setUrl(String s) {
        url = s;
    }
    
    public static String getUrl() {
        return url;
    }
    
    public static void setJarFile(String s) {
        jarFile = s;
    }
    
    public static String getJarFile() {
        return jarFile;
    }
    
    
    /*
     * needed by applets
     */
    public static void setCodebaseURL(String s) {
        codebaseURL = s;
    }
    
    public static String getCodebaseURL() {
        return codebaseURL;
    }
    
    
    public static InputStream getInputStream() {
        return in;
    }
    
    public static void setInputStream(InputStream stream) {
        in = stream;
    }
    
    
    /**
     * Sets the XML parser to validate the configuration
     * file using the DTD specified in the XML.
     * 
     * @param v if true, the XML will be validated, otherwise
     */
    public static void setValidatingParser(boolean v) {
        validatingParser = v;
    }
    
    public static boolean getValidatingParser() {
        return validatingParser;
    }
    
    /**
     * Returns the root item of the whole Matrix
     * Configuration.
     *
     * @return The ConfItem corresponding to the &lt;configuration&gt; -element
     */
    public static ConfItem getConfiguration() {
        return configuration;
    }
    
    // ---- fields ----
    
    private ConfStack stack;
    private Class currentAnchor;
    // the lock object tells if the program is currently parsing a referenced
    // menu element. When lock != null this is the case.
    // A menu element that is referenced by another should not be modified in any
    // way and therefore its state is locked, hence the name.
    private Object lock;
    // since xml specification requires menu, menubar and popupmenu ids to be
    // globally unique in an xml file, these three HashMaps could be combined
    // to one. For now three are used for clarity.
    private HashMap menuStorage;
    private HashMap menuBarStorage;
    private HashMap popupMenuStorage;
    private HashMap actionStorage;
    private HashMap toolbarStorage;
    private HashMap toolbarsStorage;
    
    
    // ---- constructor ----
    
    private MatrixConfiguration() {
        stack = new ConfStack();
        currentAnchor = null;
        lock = null;
        menuStorage = new HashMap();
        menuBarStorage = new HashMap();
        popupMenuStorage = new HashMap();
        actionStorage = new HashMap();
        toolbarStorage = new HashMap();
        toolbarsStorage = new HashMap();
        try {
            parse();
        } catch(Exception e) {
            Note.warning(this, "Could not parse configuration file due to exception");
            e.printStackTrace();
        }
    }
    
    // ---- XML parsing methods ----
    
    // parse has been lifted line-by-line from old MatrixConfiguration. There is 
    // no need for changes in this method.
    
    /**
     * Tries to open configuration file for reading and parse it. If Configuration file location has
     * been set using <code>setConfPath</code> that location is tried. If 
     * configuration file cannot be opened from specified location, the program tries
     * to find file "matrixconf.xml" in the current working directory. 
     * <p>
     * If configuration file is not set or could not be opened for reading, the 
     * program will try to open configuration file from an URL, if configuration file
     * URL has been set using <code>setUrl</code>. 
     * <p>
     * If neither configuration file location nor URL has been set, or configuration could not
     * successfully be parsed from either, the program will try to see if it finds the
     * Matrix jar package holding the configuration file. 
     * <p>
     * If configuration could not be read from jar, the program will give up and 
     * exit.
     * <p>
     */
    public void parse() {
        Note.out(this, "Starting parsing");
        SAXParserFactory factory = SAXParserFactory.newInstance();
        factory.setValidating(getValidatingParser());
        SAXParser parser = null;
        try {
            parser = factory.newSAXParser();
        } catch(Exception e) {
            throw new MatrixConfigurationException("Could not create SAX Parser; cannot read configuration.");
        }
        File conf = null;
        // First, if confFile is set, we try to open it.
        if(confFile != null) {
            try {
                conf = new File(confFile);
                if(!(conf.exists()) || !(conf.isFile())) {
                    Note.out(this,"Configuration file not found, trying default.");
                    conf = null;
                }
            } catch(SecurityException e) {
                Note.out(this,"Could not open configuration file, trying default.");
                conf = null;
            }
        }
        // if confFile could not be opened, we try the default.
        if(conf == null) {
            try {
                conf = new File(DEFAULT);
                if(!(conf.exists()) || !(conf.isFile())) {
                    Note.out(this,"Default configuration file not found.");
                    conf = null;
                }
            } catch(SecurityException e) {
                Note.out(this,"Could not open default configuration file.");
                conf = null;
            }
        }
        
        // Now we try to parse the configuration, if we have an InputStream for the
        // confFile or an URL where the configuration file can be found at.
        try {
            if(conf != null) {
                Note.out(this, "Parsing...");
                parser.parse(conf, this);
                Note.out(this, "Ended parsing");
                return;
            } else if(url != null) {
                Note.out(this, "Parsing URL...");
                parser.parse(url, this);
                Note.out(this, "Ended parsing");
                return;
            }
        } catch(SAXException e) {
            Note.warning(this,"XML Parsing exception while trying to parse configuration file.");
            e.printStackTrace();
        } catch(IOException e) {
            Note.warning(this,"O/I-exception while parsing configuration file.");
            e.printStackTrace();
        }
        // added 8.6.2004 -- if InputStream is set, use it - vkaravir
        if (in != null) {
            Note.out(this, "Using InputStream");
            try {
                parser.parse(in, this);
                return;
            } catch(SAXException e) {
                Note.warning(this,"XML Parsing exception while parsing configuration file.");
                e.printStackTrace();
            } catch(IOException e) {
                Note.warning(this,"I/O-exception while parsing configuration file.");
                e.printStackTrace();
            }
        }
        // If no confFile or URL was present or could not be parsed, we try to find
        // the Matrix jar and open the file from it. 
        try {
            Note.out(this, "Trying jar");
            if (codebaseURL != null) {
                String jarURL = "jar:" + codebaseURL + ((jarFile == null) ? JARFILE : jarFile) + "!/";
                Note.out(this, "Reading configuration file from " + jarURL);
                URL u = new URL(jarURL);
                JarURLConnection conn = (JarURLConnection)u.openConnection();
                JarFile jar = conn.getJarFile();
                parser.parse(jar.getInputStream(jar.getEntry(JARCONFFILE)),  this);
                return;
            } else {
                JarFile jar = new JarFile((jarFile == null) ? JARFILE : jarFile);
                Note.out(this, "Reading configuration file from " + ((jarFile == null) ? JARFILE : jarFile));
                parser.parse(jar.getInputStream(jar.getEntry(JARCONFFILE)),  this);
                return;
            }
        } catch(Exception e) {
            Note.warning(this, "Exception while trying to parse configuration from JAR file");
            e.printStackTrace();
        }
        // Give up, if configuration could not be created
        throw new MatrixConfigurationException("Could not configure Matrix");
    }
    
    public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException {
        //Note.out(this, "Starting element " + qName);
        try {
            if(lock != null) {
                stack.push(dummy);
                return;
            }
            else if(qName.equals("configuration")) {
                // configuration has no attributes
                startConfiguration();
            } else if(qName.equals("menubar")) {
                startMenuBar(attributes);
            } else if(qName.equals("menu")) {
                startMenu(attributes);
            } else if(qName.equals("menuitem")) {
                startMenuItem(attributes);
            } else if(qName.equals("checked")) {
                // checked has no attributes
                startChecked();
            } else if(qName.equals("menuseparator")) {
                //menuseparator has no attributes
                startMenuSeparator();
            } else if(qName.equals("field")) {
                startField(attributes);
            } else if(qName.equals("constructor")) {
                startConstructor(attributes);
            } else if(qName.equals("method")) {
                startMethod(attributes);
            } else if(qName.equals("int")) {
                startInt(attributes);
            } else if(qName.equals("string")) {
                startString(attributes);
            } else if(qName.equals("boolean")) {
                startBoolean(attributes);
            } else if(qName.equals("char")) {
                startChar(attributes);
            } else if(qName.equals("popupmenu")) {
                startPopupMenu(attributes);
            } else if(qName.equals("initialization")) {
                // initialization has no attributes
                startInitialization();
            } else if(qName.equals("visualizations")) {
                // visualizations has no attributes
                startVisualizations();
            } else if(qName.equals("structure")) {
                startStructure(attributes);
            } else if(qName.equals("visualization")) {
                // not to be confused with visualizations
                startVisualization(attributes);
            } else if(qName.equals("picked-visualizations")) {
                // picked-visualizations has no attributes
                startPickedVisualizations();
            } else if(qName.equals("action")) {
                startAction(attributes);
            } else if (qName.equals("toolbars")) {
                startToolbars(attributes);
            } else if (qName.equals("toolbar")) {
                startToolbar(attributes);
            } else if (qName.equals("item")) {
                startToolbarItem(attributes);
            } else {
                Note.warning(this, "Unknown element " + qName + " encountered during parsing!");
                stack.push(dummy);
            }
        } catch(MatrixConfigurationException e) {
            Note.warning(this, "Exception caught while opening element " + qName);
            //Note.out(this, e.getMessage());
            e.printStackTrace();
        }
        
    }
    
    private void startToolbars(Attributes attributes){
        Class anchor = null;
        CToolbars tbs = new CToolbars();
        for(int i = 0; i < attributes.getLength(); i++) {
            String attr = attributes.getQName(i);
            String val = attributes.getValue(i);
            if(attr.equals("anchor")) {
                try {
                    anchor = Class.forName(val);
                } catch(ClassNotFoundException e) {
                    throw new MatrixConfigurationException("unable to create anchor " + val + " for Toolbars");
                }
                currentAnchor = anchor;
                tbs.setAnchor(anchor);
            } else if(attr.equals("id")) {
                CToolbars tmp = retrieveToolbars(val);
                // if this particular menu is already defined
                if(tmp != null && tmp.getChildCount() > 0) {
                    tbs = tmp;
                    continue;
                }
                storeToolbars(val, tbs);
            } else if(attr.equals("extends")) {
                CToolbars org = retrieveToolbars(val);
                if(org == null) {
                    Note.warning(this, "Tried to extend non-existent toolbars" + val);
                    continue;
                }
                tbs = (CToolbars)deepCopy(org);
                try {
                    anchor = Class.forName(attributes.getValue("anchor"));
                } catch(ClassNotFoundException e) {
                    throw new MatrixConfigurationException("unable to create anchor " + val + " for Toolbars");
                }
                currentAnchor = anchor;
                tbs.setAnchor(anchor);
            }
        }
        if(tbs.getAnchor() == null) {
            Note.err(this, "No anchor defined for toolbars " + tbs);
        }
        stack.push(tbs);
    }
    
    private void startToolbar(Attributes attributes){
        CToolbar menu = new CToolbar();
        for(int i = 0; i < attributes.getLength(); i++) {
            String attr = attributes.getQName(i);
            String val = attributes.getValue(i);
            if(attr.equals("name")) {
                menu.setName(val);
            } else if(attr.equals("id")) {
                CToolbar tmp = retrieveToolbar(val);
                // if this particular menu is already defined
                if(tmp != null && tmp.getChildCount() > 0) {
                    menu = tmp;
                    continue;
                }
                storeToolbar(val, menu);
            } else if(attr.equals("extends")) {
                CToolbar org = retrieveToolbar(val);
                if(org == null) {
                    Note.warning(this, "Tried to extend non-existent toolbar " + val);
                    continue;
                }
                menu = (CToolbar)deepCopy(org);
            } else if(attr.equals("references")) {
                CToolbar tmp = retrieveToolbar(val);
                if(tmp == null) {
                    // This toolbar is not yet defined. We trust it will be defined later.
                    tmp = new CToolbar();
                }
                menu = tmp;
                lock = menu;
                // since this is a reference to an already defined toolbar, we
                // do not need to modify it any more.
                break;
            }
        }
        stack.push(menu);
    }
    
    private void startToolbarItem(Attributes attributes){
        CToolbarItem ti = new CToolbarItem();
        for(int i = 0; i < attributes.getLength(); i++) {
            String attr = attributes.getQName(i);
            String val = attributes.getValue(i);
            if(attr.equals("name")){
                ti.setName(val);
            } else if(attr.equals("disabled")) {
                if(val.toLowerCase().equals("true")) {
                    ti.setDisabled(true);
                } else {
                    ti.setDisabled(false);
                }
            } else if (attr.equals("action")) {
                ti.setAction(val);
            } else if (attr.equals("class")) {
                Class cls = null;
                try {
                    cls = Class.forName(val);
                } catch(ClassNotFoundException e) {
                    throw new MatrixConfigurationException("unable to create " + val + " for Toolbars");
                }
                ti.setClassAttr(cls);
            }
        }
        stack.push(ti);
    }
    
    private void startAction(Attributes attributes) {
        CAction ca = new CAction();
        for (int i=0; i < attributes.getLength(); i++) {
            String attr = attributes.getQName(i);
            String val = attributes.getValue(i);
            if(attr.equals("id")) {
                ca.setId(val);
            } else if (attr.equals("class")) {
                ca.setImplClass(val);
            }
        }
        stack.push(ca);
    }
    
    private void startConfiguration(){
        ConfItem item = new Configuration();
        stack.push(item);
        return;
    }
    
    private void startMenuBar(Attributes attributes){
        Class anchor = null;
        CMenuBar mb = new CMenuBar();
        for(int i = 0; i < attributes.getLength(); i++) {
            String attr = attributes.getQName(i);
            String val = attributes.getValue(i);
            if(attr.equals("anchor")) {
                try {
                    anchor = Class.forName(val);
                } catch(ClassNotFoundException e) {
                    throw new MatrixConfigurationException("unable to create anchor " + val + " for MenuBar");
                }
                currentAnchor = anchor;
                mb.setAnchor(anchor);
            } else if(attr.equals("id")) {
                CMenuBar tmp = retrieveMenuBar(val);
                // if this particular menu is already defined
                if(tmp != null && tmp.getChildCount() > 0) {
                    mb = tmp;
                    continue;
                }
                storeMenuBar(val, mb);
            } else if(attr.equals("extends")) {
                CMenuBar org = retrieveMenuBar(val);
                if(org == null) {
                    Note.warning(this, "Tried to extend non-existent menubar " + val);
                    continue;
                }
                mb = (CMenuBar)deepCopy(org);
            }
        }
        if(mb.getAnchor() == null) {
            Note.err(this, "No anchor defined for menubar " + mb);
        }
        stack.push(mb);
    }
    
    private void startMenu(Attributes attributes){
        CMenu menu = new CMenu();
        for(int i = 0; i < attributes.getLength(); i++) {
            String attr = attributes.getQName(i);
            String val = attributes.getValue(i);
            if(attr.equals("name")) {
                menu.setName(val);
            } else if(attr.equals("id")) {
                CMenu tmp = retrieveMenu(val);
                // if this particular menu is already defined
                if(tmp != null && tmp.getChildCount() > 0) {
                    menu = tmp;
                    continue;
                }
                storeMenu(val, menu);
            } else if(attr.equals("extends")) {
                CMenu org = retrieveMenu(val);
                if(org == null) {
                    Note.warning(this, "Tried to extend non-existent menu " + val);
                    continue;
                }
                menu = (CMenu)deepCopy(org);
            } else if(attr.equals("mnemonic")) {
                menu.setMnemonic(val.charAt(0));   
            } else if(attr.equals("references")) {
                CMenu tmp = retrieveMenu(val);
                if(tmp == null) {
                    // This menu is not yet defined. We trust it will be defined later.
                    tmp = new CMenu();
                }
                menu = tmp;
                lock = menu;
                // since this is a reference to an already defined menu, we
                // do not need to modify it any more.
                break;
            }
        }
        stack.push(menu);
    }
    
    private void startMenuItem(Attributes attributes){
        CMenuItem mi = new CMenuItem();
        for(int i = 0; i < attributes.getLength(); i++) {
            String attr = attributes.getQName(i);
            String val = attributes.getValue(i);
            if(attr.equals("name")){
                mi.setName(val);
            } else if(attr.equals("disabled")) {
                if(val.toLowerCase().equals("true")) {
                    mi.setDisabled(true);
                } else {
                    mi.setDisabled(false);
                }
            } else if(attr.equals("type")) {
                mi.setType(val);
            } else if(attr.equals("checked")) {
                if(val.toLowerCase().equals("true")) {
                    mi.setChecked(true);
                } else {
                    mi.setChecked(false);
                }
            } else if(attr.equals("accesskey")) {
                Class keys = null;
                Field key = null;
                try {
                    keys = Class.forName("java.awt.event.KeyEvent");
                    key = keys.getField(val);
                    mi.setAccessKey(key.getInt(null));
                } catch(Exception e) {
                    //e.printStackTrace();
                    Note.warning(this, "Could not create access key " + val);
                }
            } else if(attr.equals("group")) {
                mi.setGroup(val);
            } else if(attr.equals("mnemonic")) {
                mi.setMnemonic(val.charAt(0));   
            } else if (attr.equals("action")) {
                mi.setAction(val);
            }
        }
        stack.push(mi);
    }
    
    private void startChecked(){
        stack.push(new CChecked());
    }
    private void startMenuSeparator(){
        stack.push(new CMenuSeparator());
    }
    private void startField(Attributes attributes){
        CField f = new CField();
        Class anchor = currentAnchor;
        String fName = null;
        for(int i = 0; i < attributes.getLength(); i++) {
            String aName = attributes.getQName(i);
            String val = attributes.getValue(i);
            if(aName.equals("name")) {
                fName = val;
            } else if(aName.equals("class")) {
                try {
                    anchor = Class.forName(val);
                } catch(ClassNotFoundException e) {
                    //e.printStackTrace();
                    Note.warning(this, "Could not initalize static anchor " + val + " for field. Substituting dummy.");
                    stack.push(dummy);
                }
            }
        }
        try {
            f.setField(anchor.getField(fName));
        } catch(Exception e) {
            //e.printStackTrace();
            Note.warning(this, "Could not create required field " + fName + ". Substituting dummy.");
            stack.push(dummy);
            return;
        }
        stack.push(f);
    }
    
    private void startConstructor(Attributes attributes){
        String cName = attributes.getValue(0);
        Class c = null;
        try {
            c = Class.forName(cName);
        } catch(ClassNotFoundException e) {
            //e.printStackTrace();
            Note.warning(this, "Class " + cName + " not found, using dummy");
            stack.push(dummy);
            return;
        }
        ConfItem ci = new CConstructor(c);
        stack.push(ci);
    }
    
    private void startMethod(Attributes attributes){
        String mName = null;
        String cName = null;
        for(int i = 0; i < attributes.getLength(); i++) {
            String curr = attributes.getQName(i);
            if(curr.equals("name")) {
                mName = attributes.getValue(i);
            } else if(curr.equals("class")) {
                cName = attributes.getValue(i);
            }
        }
        Class anchor = null;
        if(cName != null) {
            try {
                anchor = Class.forName(cName);
            } catch(ClassNotFoundException e) {
                //e.printStackTrace();
                //throw new MatrixConfigurationException("Class not found");
                Note.warning(this, "Class " + cName + " not found, using dummy");
                stack.push(dummy);
                return;
            }
        }
        ConfItem ci;
        if(cName != null)
            ci = new CMethod(anchor, mName, true);
        else
            ci = new CMethod(anchor, mName);
        stack.push(ci);
    }
    
    private void startInt(Attributes attributes){
        String intValue = attributes.getValue(0);
        try {
            ConfItem ci = new CInt(Integer.parseInt(intValue));		
            stack.push(ci);
        } catch(NumberFormatException e) {
            Note.warning(this, "" + intValue + " not a number. Substituting 0");
            stack.push(new CInt(0));
        }
    }
    
    private void startString(Attributes attributes){
        String str = attributes.getValue(0);
        ConfItem ci = new CString(str);
        stack.push(ci);
    }
    private void startBoolean(Attributes attributes){
        String bVal = attributes.getValue(0);
        ConfItem ci = null;
        if(bVal.toLowerCase().equals("true")) {
            ci = new CBoolean(true);
        } else {
            ci = new CBoolean(false);
        }
        stack.push(ci);
    }
    private void startChar(Attributes attributes){
        String cVal = attributes.getValue(0);
        ConfItem ci = new CChar(cVal.charAt(0));
        stack.push(ci);
    }
    private void startPopupMenu(Attributes attributes){
        Class anchor = null;
        CPopupMenu pm = new CPopupMenu();
        for(int i = 0; i < attributes.getLength(); i++) {
            String attr = attributes.getQName(i);
            String val = attributes.getValue(i);
            if(attr.equals("anchor")) {
                try {
                    anchor = Class.forName(val);
                } catch(ClassNotFoundException e) {
                    throw new MatrixConfigurationException("unable to create anchor " + val + " for MenuBar");
                }
                currentAnchor = anchor;
                pm.setAnchor(anchor);
            } else if(attr.equals("id")) {
                CPopupMenu tmp = retrievePopupMenu(val);
                // if this particular menu is already defined
                if(tmp != null && tmp.getChildCount() > 0) {
                    pm = tmp;
                    continue;
                }
                storePopupMenu(val, pm);
            } else if(attr.equals("extends")) {
                CPopupMenu org = retrievePopupMenu(val);
                if(org == null) {
                    Note.warning(this, "Tried to extend non-existent popupmenu " + val);
                    continue;
                }
                pm = (CPopupMenu)deepCopy(org);
            }
        }
        if(pm.getAnchor() == null) {
            Note.err(this, "No anchor defined for popupmenu " + pm);
        }
        stack.push(pm);
    }
    private void startInitialization(){
        stack.push(new CInitialization());
    }
    
    private void startVisualizations(){
        stack.push(new CVisualizations());
    }
    private void startStructure(Attributes attributes){
        String sName = attributes.getValue(0);
        Class str = null;
        try {
            str = Class.forName(sName);
        } catch(ClassNotFoundException e) {
            //e.printStackTrace();
            //throw new MatrixConfigurationException("No class " + sName + " found");
            Note.warning(this, "Class " + sName + " not found. Using dummy.");
            stack.push(dummy);
            return;
        }
        stack.push(new CStructure(str));
    }
    private void startVisualization(Attributes attributes){
        String name = null;
        String cName = null;
        for(int i = 0; i < attributes.getLength(); i++) {
            String attr = attributes.getQName(i);
            if(attr.equals("name")) {
                name = attributes.getValue(i);
            } else if(attr.equals("class")) {
                cName = attributes.getValue(i);
            }
        }
        Class visual = null;
        try {
            visual = Class.forName(cName);
        } catch(ClassNotFoundException e) {
            //e.printStackTrace();
            //throw new MatrixConfigurationException("Could not find class " + cName);
            Note.warning(this, "Class " + cName + " not found. Using dummy");
            stack.push(dummy);
            return;
        }
        ConfItem ci = new CVisualization(name,visual);
        stack.push(ci);
    }
    
    private void startPickedVisualizations(){
        stack.push(new CPickedVisualizations());
    }
    
    public void endElement(String uri, String localName, String qName) throws SAXException {
        //Note.out(this, "Ending element " + qName);
        try {
            if(lock != null) {
                ConfItem item = stack.peek();
                if(item == lock) {
                    lock = null;
                } else {
                    stack.pop();
                    return;
                }
            }
            ConfItem item = stack.peek();
            if(item == dummy) {
                Note.warning(this, "Dummy item encountered outside locks, skipping.");
                stack.pop();
                return;
            }
            if(qName.equals("configuration")) {
                endConfiguration();
            } else if(qName.equals("menubar")) {
                endMenuBar();
            } else if(qName.equals("menu")) {
                endMenu();
            } else if(qName.equals("menuitem")) {
                endMenuItem();
            } else if(qName.equals("checked")) {
                endChecked();
            } else if(qName.equals("menuseparator")) {
                endMenuSeparator();
            } else if(qName.equals("field")) {
                endField();
            } else if(qName.equals("constructor")) {
                endConstructor();
            } else if(qName.equals("method")) {
                endMethod();
            } else if(qName.equals("int")) {
                endInt();
            } else if(qName.equals("string")) {
                endString();
            } else if(qName.equals("boolean")) {
                endBoolean();
            } else if(qName.equals("char")) {
                endChar();
            } else if(qName.equals("popupmenu")) {
                endPopupMenu();
            } else if(qName.equals("initialization")) {
                // initialization has no 
                endInitialization();
            } else if(qName.equals("visualizations")) {
                endVisualizations();
            } else if(qName.equals("structure")) {
                endStructure();
            } else if(qName.equals("visualization")) {
                // not to be confused with visualizations
                endVisualization();
            } else if(qName.equals("picked-visualizations")) {
                endPickedVisualizations();
            } else if (qName.equals("toolbars")) {
                endToolbars();
            } else if (qName.equals("toolbar")) {
                endToolbar();
            } else if (qName.equals("item")) {
                endToolbarItem();
            } else if (qName.equals("action")) {
                endAction();
            }
        } catch(MatrixConfigurationException e) {
            Note.warning(this, "Exception caught while closing element " + qName);
            Note.out(this, e.getMessage());
            //e.printStackTrace();
        }
    }
    
    private void endToolbars(){
        ConfItem ci = stack.pop();
        if(ci.getItemType() != ConfItem.TOOLBARS) {
            unbalancedStack(ci.getItemType(), ConfItem.TOOLBARS);
        }
        ConfItem parent = stack.peek();
        parent.addChild(ci);
        ToolbarFactory tf = ToolbarFactory.getInstance();
        tf.addToolbars((CToolbars)ci);
    }
    
    private void endToolbar(){
        endCommonElement(ConfItem.TOOLBAR);
    }
    
    private void endToolbarItem(){
        endCommonElement(ConfItem.TOOLBARITEM);
    }
    
    private void endAction() {
        ConfItem ci = stack.pop();
        if (ci.getItemType() == ConfItem.ACTION) {
            String actionId = ((CAction)ci).getId();
            if (actionStorage.containsKey(actionId))
                Note.out(this, "Warning: Action " + actionId +
                "already defined.");
            actionStorage.put(actionId, ci);
            ActionFactory af = ActionFactory.getInstance();
            af.addAction((CAction)ci);
        } else
            unbalancedStack(ci.getItemType(), ConfItem.ACTION);
    }
    
    private void endConfiguration(){
        ConfItem ci = stack.pop();
        if(ci.getItemType() != ConfItem.CONFIGURATION) {
            unbalancedStack(ci.getItemType(), ConfItem.CONFIGURATION);
        }
        configuration = ci;
    }
    
    private void endMenuBar(){
        ConfItem ci = stack.pop();
        if(ci.getItemType() != ConfItem.MENUBAR) {
            unbalancedStack(ci.getItemType(), ConfItem.MENUBAR);
        }
        ConfItem parent = stack.peek();
        parent.addChild(ci);
        MenuFactory mf = MenuFactory.getInstance();
        mf.addMenuBar((CMenuBar)ci);
    }
    
    private void endMenu(){
        endCommonElement(ConfItem.MENU);
    }
    
    private void endMenuItem(){
        endCommonElement(ConfItem.MENUITEM);
    }
    
    private void endChecked(){
        endCommonElement(ConfItem.CHECKED);
    }
    
    private void endMenuSeparator(){
        endCommonElement(ConfItem.MENUSEPARATOR);
    }
    
    private void endField(){
        endCommonElement(ConfItem.FIELD);
    }
    private void endConstructor(){
        endCommonElement(ConfItem.CONSTRUCTOR);
    }
    
    private void endMethod(){
        ConfItem ci = stack.pop();
        if(ci.getItemType() != ConfItem.METHOD) {
            unbalancedStack(ci.getItemType(),ConfItem.METHOD);
        }
        ConfItem parent = stack.peek();
        parent.addChild(ci);
        if(((CMethod)ci).isStatic() && 
                parent.getItemType() == ConfItem.MENUITEM) {
            ((CMenuItem)parent).setGlobal(true);
        }	
    }
    
    private void endInt(){
        endCommonElement(ConfItem.INT);
    }
    
    private void endString(){
        endCommonElement(ConfItem.STRING);
    }
    private void endBoolean(){
        endCommonElement(ConfItem.BOOLEAN);
    }
    private void endChar(){
        endCommonElement(ConfItem.CHAR);
    }
    private void endPopupMenu(){
        ConfItem ci = stack.pop();
        if(ci.getItemType() != ConfItem.POPUPMENU) {
            unbalancedStack(ci.getItemType(), ConfItem.POPUPMENU);
        }
        ConfItem parent = stack.peek();
        parent.addChild(ci);
        MenuFactory mf = MenuFactory.getInstance();
        mf.addPopupMenu((CPopupMenu)ci);
    }
    
    private void endInitialization(){
        ConfItem ci = stack.pop();
        if(ci.getItemType() != ConfItem.INITIALIZATION) {
            unbalancedStack(ci.getItemType(), ConfItem.INITIALIZATION);
        }
        ConfItem parent = stack.peek();
        parent.addChild(ci);
        // Initializations that are immediate children of the configuration -element
        // are run immediately when they've been completely parsed
        if(parent.getItemType() == ConfItem.CONFIGURATION) {
            Initializator.getInstance().initialize((CInitialization)ci,null);
        }
    }
    private void endVisualizations(){
        ConfItem ci = stack.pop();
        if(ci.getItemType() != ConfItem.VISUALIZATIONS) {
            unbalancedStack(ci.getItemType(), ConfItem.VISUALIZATIONS);
        }
        ConfItem parent = stack.peek();
        parent.addChild(ci);
        RepresentationFactory rf = RepresentationFactory.getInstance();
        rf.addVisualizations((CVisualizations)ci);
    }
    
    private void endStructure(){
        endCommonElement(ConfItem.STRUCTURE);
    }
    
    private void endVisualization(){
        endCommonElement(ConfItem.VISUALIZATION);
    }
    
    
    private void endPickedVisualizations(){
        ConfItem ci = stack.pop();
        if(ci.getItemType() != ConfItem.PICKEDVISUALIZATIONS) {
            unbalancedStack(ci.getItemType(), ConfItem.PICKEDVISUALIZATIONS);
        }
        ConfItem parent = stack.peek();
        parent.addChild(ci);
        RepresentationFactory rf = RepresentationFactory.getInstance();
        rf.addPickedVisualizations((CPickedVisualizations)ci);
    }
    
    private void endCommonElement(int type) {
        ConfItem ci = stack.pop();
        if(ci.getItemType() != type) {
            unbalancedStack(ci.getItemType(),type);
        }
        ConfItem parent = stack.peek();
        parent.addChild(ci);
    }
    
    
    // ---- debug methods ----
    
    /**
     * Prints the intermediate representation tree to System.out. Used for
     * debug purposes.
     */
    public void printTree(ConfItem ci) {
        printTree(ci, 0);
    }
    
    private void printTree(ConfItem ci, int offset) {
        String hdr = "";
        for(int i = 0; i < offset; i++) {
            hdr += "   ";
        }
        System.out.println(hdr + ci.toString());
        for(int i = 0; i < ci.getChildCount(); i++) {
            printTree(ci.getChild(i), offset+1);
        }
        if(ci.getChildCount() > 0) {
            System.out.println(hdr+"/" + ci.toString());
        }
    }
    
    // ---- menu, menubar and popupmenu storage and retrieval methods ----
    
    /* these methods are used by the MatrixConfiguration class to store 
     menus, menubars and popupmenus defined in the configuration file 
     in order to use references and inheritance.
     */
    
    public void storeMenu(Object id, CMenu menu) {
        Object o = menuStorage.put(id, menu);
        if(o != null) {
            //throw new RuntimeException("Duplicate menu id for menus " + menu + " and " + o);
            Note.warning(this, "Duplicate menu id for menus " + menu + " and " + o);
            Note.warning(this, "" + o + " has been overwritten.");
        }
    }
    
    public CMenu retrieveMenu(Object id) {
        return (CMenu)menuStorage.get(id);
    }
    
    public void storeMenuBar(Object id, CMenuBar mbar) {
        Object o = menuBarStorage.put(id, mbar);
        if(o != null) {
            //throw new RuntimeException("Duplicate menu bar id for menubars " + mbar + " and " + o);
            Note.warning(this, "Duplicate menu bar id for menubars " + mbar + " and " + o);
            Note.warning(this, ""+ o + " has been overwritten.");
        }
        
    }
    
    public CMenuBar retrieveMenuBar(Object id) {
        return (CMenuBar)menuBarStorage.get(id);
    }
    
    public void storePopupMenu(Object id, CPopupMenu pop) {
        Object o = popupMenuStorage.put(id, pop);
        if(o != null) {
            //throw new RuntimeException("Duplicate popupmenu id for popupmenus " + pop + " and " + o);
            Note.warning(this, "Duplicate popupmenu id for popupmenus " + pop + " and " + o);
            Note.warning(this, ""+ o + " has been overwritten.");
        }
    }
    
    public CPopupMenu retrievePopupMenu(Object id) {
        return (CPopupMenu)popupMenuStorage.get(id);
    }
    
    
    
    public void storeToolbars(Object id, CToolbars menu) {
        Object o = toolbarsStorage.put(id, menu);
        if(o != null) {
            //throw new RuntimeException("Duplicate menu id for menus " + menu + " and " + o);
            Note.warning(this, "Duplicate menu id for menus " + menu + " and " + o);
            Note.warning(this, "" + o + " has been overwritten.");
        }
    }
    
    public CToolbars retrieveToolbars(Object id) {
        return (CToolbars)toolbarsStorage.get(id);
    }
    
    public void storeToolbar(Object id, CToolbar mbar) {
        Object o = toolbarStorage.put(id, mbar);
        if(o != null) {
            //throw new RuntimeException("Duplicate menu bar id for menubars " + mbar + " and " + o);
            Note.warning(this, "Duplicate menu bar id for menubars " + mbar + " and " + o);
            Note.warning(this, ""+ o + " has been overwritten.");
        }
        
    }
    
    public CToolbar retrieveToolbar(Object id) {
        return (CToolbar)toolbarStorage.get(id);
    }
    
    // ---- miscallenous methods ----
    
    private void unbalancedStack(int ci, int req) {
        Note.warning(this, "Stack error: type " + ci + " item popped from stack. Should have been "
                     + "type " + req);
    }
    
    private ConfItem deepCopy(ConfItem item) {
        try {
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            ObjectOutputStream out = new ObjectOutputStream(baos);
            out.writeObject(item);
            ByteArrayInputStream bais = new ByteArrayInputStream(baos.toByteArray());
            ObjectInputStream in = new ObjectInputStream(bais);
            ConfItem copy = (ConfItem)in.readObject();
            return copy;
        } catch (Exception e) {
            //e.printStackTrace();
            Note.err(this, "Failed to copy ConfItem " + item);
        }
        return null;
    }
    
    // ---- inner classes ----
    
    class ConfStack extends Vector {
        public ConfItem peek() {
            return (ConfItem) lastElement();
        }
        public ConfItem pop() {
            return (ConfItem) remove(size()-1);
        }
        
        public void push(ConfItem c) {
            add(c);
        }
    }
}
