/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2005 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: ConfigurableGraphics.java,v 1.2 2005/05/24 10:50:58 kstalnac Exp $ */

package matrix.util.export;

import java.awt.Color;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.image.ImageObserver;
import java.text.AttributedCharacterIterator;
import java.util.Properties;
import java.io.*;

/**
 * ConfigurableGraphics class for selecting what to paint and what to not paint.
 * 
 * This class is an adapter which mediates the drawing commands into the Graphics 
 * object given in constuctor. This class adds the functionality that can be used 
 * to select what to paint and what to not paint. Before painting, the wanted 
 * functionality/visualization can be disabled by using the addDisabled method 
 * and giving it a corresponding keyword. When painting the Graphics object, 
 * Visual classes can check in suitable places if the functionality is disabled
 * in the ConfigurableGraphics object (that is, the keyword is set) by using 
 * isDisabled method. In this way Visual classes can take into account during painting 
 * configurations made before painting.
 * 
 * @author kstalnac, oseppala
 */
public class ConfigurableGraphics extends Graphics {

    /**Graphics to which all the drawing commands are mediated.*/
    private Graphics   g;
    /**Properties contains configurations for configuring the visualization/drawing.*/
    private Properties configuration; 
    
    /**
     * Loads the configuration from InputStream
     * @param is inpustream containing Properties
     * @throws IOException
     */
    public void loadConfiguration(InputStream is) throws IOException{
    	configuration = new Properties();
        configuration.load(is);
    }

    /**
     * Constructor
     * @param g Graphics object where the actual drawing commands are mediated.
     */
	public ConfigurableGraphics(Graphics g){
		this.g = g;
        configuration = new Properties();
    }
    
    /**
     * Adds information about disabled functionality/part of the visualization into Properties.
     * @param disableKey name of the disabled functionality or part of the visualization
     */
    public void addDisable(String disableKey){
    	configuration.setProperty(disableKey, "disabled");
    }
    
    /**
     * Returns information is the given functionality disabled
     * @param key name of the functionality
     * @return true, if the functionality is disabled, false if it is not disabled
     */
    public boolean isDisabled(String key){
    	return "disabled".equals(configuration.getProperty(key));
    }
    
    
    /*--  Methods that overrides abstract methods in Graphics class  --*/
    
    public Graphics create() {
        ConfigurableGraphics cg = new ConfigurableGraphics(g.create());
        cg.configuration = configuration;
		return cg;
	}

    public Graphics create(int x, int y, int width, int height){
        ConfigurableGraphics cg = new ConfigurableGraphics(g.create(x,y,width,height));
        cg.configuration = configuration;
        return cg;
    }
    
	public void translate(int arg0, int arg1) {
		g.translate(arg0, arg1);		
	}

	public Color getColor() {
		return g.getColor();
	}

	public void setColor(Color arg0) {
		g.setColor(arg0);
	}

	public void setPaintMode() {
		g.setPaintMode();
	}

	public void setXORMode(Color arg0) {
		g.setXORMode(arg0);
	}

	public Font getFont() {
		return g.getFont();
	}

	public void setFont(Font arg0) {
		g.setFont(arg0);
	}

	public FontMetrics getFontMetrics(Font arg0) {
		return g.getFontMetrics(arg0);
	}

	public Rectangle getClipBounds() {
		return g.getClipBounds();
	}

	public void clipRect(int arg0, int arg1, int arg2, int arg3) {
		g.clipRect(arg0, arg1, arg2, arg3);
	}

	public void setClip(int arg0, int arg1, int arg2, int arg3) {
		g.setClip(arg0, arg1, arg2, arg3);
	}

	public Shape getClip() {
		return g.getClip();
	}

	public void setClip(Shape arg0) {
		g.setClip(arg0);
	}

	public void copyArea(int arg0, int arg1, int arg2, int arg3, int arg4, int arg5) {
		g.copyArea(arg0, arg1, arg2, arg3, arg4, arg5);
		
	}

	public void drawLine(int arg0, int arg1, int arg2, int arg3) {
        g.drawLine(arg0, arg1, arg2, arg3);				    
	}

	public void fillRect(int arg0, int arg1, int arg2, int arg3) {
		g.fillRect(arg0, arg1, arg2, arg3);
	}

	public void clearRect(int arg0, int arg1, int arg2, int arg3) {
		g.clearRect(arg0, arg1, arg2, arg3);
	}

	public void drawRoundRect(int arg0, int arg1, int arg2, int arg3, int arg4, int arg5) {
		g.drawRoundRect(arg0, arg1, arg2, arg3, arg4, arg5);
	}

	public void fillRoundRect(int arg0, int arg1, int arg2, int arg3, int arg4, int arg5) {
		g.fillRoundRect(arg0, arg1, arg2, arg3, arg4, arg5);
	}

	public void drawOval(int arg0, int arg1, int arg2, int arg3) {
		g.drawOval(arg0, arg1, arg2, arg3);
	}  

	public void fillOval(int arg0, int arg1, int arg2, int arg3) {
		g.fillOval(arg0, arg1, arg2, arg3);
	}

	public void drawArc(int arg0, int arg1, int arg2, int arg3, int arg4, int arg5) {
		g.drawArc(arg0, arg1, arg2, arg3, arg4, arg5);
    }

	public void fillArc(int arg0, int arg1, int arg2, int arg3, int arg4, int arg5) {
		g.fillArc(arg0, arg1, arg2, arg3, arg4, arg5);
	}

	public void drawPolyline(int[] arg0, int[] arg1, int arg2) {
		g.drawPolyline(arg0, arg1, arg2);
    }

	public void drawPolygon(int[] arg0, int[] arg1, int arg2) {
		g.drawPolygon(arg0, arg1, arg2);
    }

	public void fillPolygon(int[] arg0, int[] arg1, int arg2) {
		g.fillPolygon(arg0, arg1, arg2);
    }

	public void drawString(String arg0, int arg1, int arg2) {
		g.drawString(arg0, arg1, arg2);
    }

	public void drawString(AttributedCharacterIterator arg0, int arg1, int arg2) {
		g.drawString(arg0, arg1, arg2);
    }

	public boolean drawImage(Image arg0, int arg1, int arg2, ImageObserver arg3) {
		return g.drawImage(arg0, arg1, arg2, arg3);
	}

	public boolean drawImage(Image arg0, int arg1, int arg2, int arg3, int arg4, ImageObserver arg5) {
		return g.drawImage(arg0, arg1, arg2, arg3, arg4, arg5);
	}

	public boolean drawImage(Image arg0, int arg1, int arg2, Color arg3, ImageObserver arg4) {
		return g.drawImage(arg0, arg1, arg2, arg3, arg4);
	}

	public boolean drawImage(Image arg0, int arg1, int arg2, int arg3, int arg4, Color arg5, ImageObserver arg6) {
		return g.drawImage(arg0, arg1, arg2, arg3, arg4, arg5, arg6);
	}

	public boolean drawImage(Image arg0, int arg1, int arg2, int arg3, int arg4, int arg5, int arg6, int arg7, int arg8, ImageObserver arg9) {
		return g.drawImage(arg0, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9);
	}

	public boolean drawImage(Image arg0, int arg1, int arg2, int arg3, int arg4, int arg5, int arg6, int arg7, int arg8, Color arg9, ImageObserver arg10) {
		return g.drawImage(arg0, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10);
	}

	public void dispose() {
        g.dispose();
        g = null;
	}

}
