/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: LatexWriter.java,v 1.14 2007/09/05 10:29:02 oseppala Exp $ */

package matrix.util.export;

import java.awt.Rectangle;
import java.io.*;
import java.util.Enumeration;

import matrix.util.Note;

/** 
 * Class used to export LaTeX documents from Matrix.
 */
public class LatexWriter {

    /** 
     * <code>MatrixGraphicsContents</code> which contents are to be
     * exported 
     */
    private MatrixGraphicsContents mgc;
    /** Default value for determining to export complete LaTeX document. */
    public static final boolean	DEFAULT_DOC = true; 
    /** 
     * Default value for determining to overwrite existing files,
     * instead of appending them. 
     */
    public static final boolean DEFAULT_APPEND = false;

    /**
     * Constructs a new <code>LatexWriter</code> for exporting the given 
     * <code>MatrixGraphicsContents</code>.
     *
     * @param mgc The <code>MatrixGraphicsContents</code> to be exported.  */
    public LatexWriter(MatrixGraphicsContents mgc) {
	this.mgc = mgc;
    }

    /** 
     * Writes LaTeX document of the contents this object contains to
     * the given file. The document will be a complete LaTeX document
     * and the file will be overwritten if it exists. Same as write(f,
     * DEFAULT_DOC, DEFAULT_APPEND).
     * 
     * @param f Output file 
     */
    public void write(File f) {
	write(f, DEFAULT_DOC, DEFAULT_APPEND);
    }

    /** 
     * Writes LaTeX document of the contents this object contains to
     * the given file. If parameter doc is true, a complete document
     * is exported, otherwise only the Texdraw representation. If
     * parameter append is true, everything is appended to the given
     * file, else an existing file is overwritten.
     * 
     * @param f Output file 
     * @param doc Boolean representing whether to export complete
     * LaTeX document or just the Texdraw part. If true, complete
     * document is exported.
     * @param append If true, everything is appended to the given
     * file, else an existing file is overwritten. 
     */
    public void write(File f, boolean doc, boolean append) {
	Rectangle bounds = mgc.getBounds();
	int x = bounds.width - bounds.x;
	int y = bounds.height - bounds.y;
	if (Math.max(x, y) > 540)  // for a4paper portrait
	    mgc.setScale(540.0/Math.max(x, y));
	try {
	    Writer out;
	    if (append) 
		out = new FileWriter(f.getPath(), true);
	    else
		out = new FileWriter(f);
	    if (doc) {
		out.write("\\documentclass{article}\n\\usepackage{color}\n\\begin{document}\n");
		out.write("\\input texdraw\n");
	    }
	    int fontsize = mgc.getFontsize();
	    out.write("\\centertexdraw{\n");
	    out.write("%Here you can set the scale of the figure. Just " +
		      "replace the one (1)\n% with the scale you want.\n");
	    out.write("\\newcommand{\\scale}{1}\n" +
		      "\\realmult {" + fontsize + "} {\\scale} \\mfs\n" +
		      "\\newcommand{\\matrixfontselect}{" +
		      "\\fontsize{\\mfs pt}{0pt}\\selectfont}\n");
	    out.write("\\drawdim pt\n");
	    out.write("\\textref h:L v:C\n");
	    out.write("\\relunitscale " + 
		      ((fontsize != -1) ? "{\\scale}" : " " + mgc.getScale()) 
		      + "\n");
	    out.write("\\linewd 1\n");
	    writeGroup(mgc.getRoot(), out);
	    out.write("}\n");
	    if (doc)
		out.write("\\end{document}\n");
	    out.close();
	} catch (IOException e) {
	    Note.out(e,"Error: " + e.toString()); 
	    e.printStackTrace();
	}
    }


    /**
     * Writes the given group's Texdraw representation to the given 
     * Writer.
     *
     * @param grp The group to be written.
     * @param out Output <code>Writer</code>.
     */
    private void writeGroup(MG_Group grp, Writer out) 
                  throws IOException {
	Enumeration subs = grp.getGroups();
	Enumeration contents = grp.getContents();

    while (contents.hasMoreElements()) {
	    Object obj = contents.nextElement();
	    if (obj instanceof MG_Text) { // Text need to be scaled
		out.write(((MG_Text)obj).toLatex(mgc.getScale()));
	    }
	    else if (obj instanceof MG_Element) 
		out.write(((MG_Element)obj).toLatex());
	}
	while (subs.hasMoreElements())
	    writeGroup((MG_Group)subs.nextElement(), out);
    }
}
	    
