/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: MG_Polygon.java,v 1.14 2004/05/26 08:06:03 vkaravir Exp $ */

package matrix.util.export;

import java.awt.Color;
import java.io.*;

/**
 * Represents polygon in matrix graphics. 
 */
public class MG_Polygon extends MG_Element {

    private int[] xPoints, yPoints;
    /** Fillcolor of this element. */
    private Color fill;

    public MG_Polygon(int[] x, int[] y, Color c, Color f) {
	super();
	xPoints = x;
	yPoints = y;
	color = c;
	fill = f;
    }

    public MG_Polygon(int[] x, int[] y, Color c, Color f,
			  int step) {
	this(x, y, c, f);
	int i = 0;
	setOpacity(0);
	while (i <= step) {
	    addStep(this, i);
	    i++;
	}	
    }

    public int[][] getPoints() { 
	return new int[][] { xPoints, yPoints };
    }

    public Color getFillColor() { return fill; }

    public String toString() {
	return "MG_Polygon: points:" + pointsToString(xPoints, yPoints) + " " 
	    + "color:" + color + " fill:" + fill;
    }

    // interface MG_Element

    public MG_Element getStep(int step) {
	MG_Polygon temp = (MG_Polygon)steps.get(new Integer(step));
	if (temp == null) return null;
	return new MG_Polygon(temp.getPoints()[0], temp.getPoints()[1],
			      temp.getColor(), temp.getFillColor());
    }

    public String toLatex() {
	float gLevel = ColorConverter.toGrayLevel(color);
	String tex = "\\move (" + xPoints[0] + " " + (-yPoints[0]) + ")\n";
	int i = 1;
	while (i < xPoints.length) {
	    tex += "\\lvec (" + xPoints[i] + " " + (-yPoints[i]) + ")\n";
	    i++;
	}
	tex += "\\lvec (" + xPoints[0] + " " + (-yPoints[0]) + ")\n"; 
	if (fill != null) {
	    float fillLevel = ColorConverter.toGrayLevel(fill);
	    tex += "\\lfill f:" + fillLevel + "\n";
	    if (gLevel != 0)
		tex = "\\setgray " + gLevel + "\n" + tex + "\\setgray 0\n";
	} else if (gLevel == 1)
	    tex = "\\setgray 1\n" + tex + "\\setgray 0\n";
	return tex;
    }

    public void toSVG(Writer out, SVGUtil util)
                throws IOException {
	int[] colors = ColorConverter.toRGB(color);
	out.write("<polygon");
	if (getOpacity() != 1)
	    out.write(" opacity=\"" + getOpacity() + "\"");
	if (fill != null) {
	    int[] fillc = ColorConverter.toRGB(fill);
	    out.write(" stroke=\"none\"");
	    out.write(" fill=\"rgb(" + fillc[0] + "," + fillc[1] + "," + 
		fillc[2] + ")\"");
	} else {
	    out.write(" stroke=\"rgb(" + colors[0] + "," + colors[1] + "," +
		      colors[2] + ")\"");
	    out.write(" fill=\"none\"");
	}
	out.write(" points=\"" + pointsToString(xPoints, yPoints));
	out.write("\">");
	animate(out, util);
	out.write("</polygon>\n");
    }

    public int getType() {
	return MG_Element.MG_POLYGON;
    }

    // interface MG_Element done

    /**
     * Creates the SVG elements needed to animate this element.
     *
     * @param out Output Writer
     * @param util SVGUtil used exporting this animation
     * @see SVGUtil
     */
    private void animate(Writer out, SVGUtil util) 
                 throws IOException {
	int i = 0;
	MG_Polygon curr = this;
	while (hasStep(i+1)) {
	    MG_Polygon next = (MG_Polygon)steps.get(new Integer(i+1));
	    String currPoints = pointsToString(curr.getPoints());
	    String nextPoints = pointsToString(next.getPoints());
	    if (!currPoints.equals(nextPoints))
		out.write(util.animate("points", i, currPoints, nextPoints));
	    if (!curr.color.equals(next.color)) {
		int[] c1 = ColorConverter.toRGB(curr.color);
		int[] c2 = ColorConverter.toRGB(next.color);
		String fr = "rgb(" + c1[0] + "," + c1[1] + "," + c1[2] + ")";
		String to = "rgb(" + c2[0] + "," + c2[1] + "," + c2[2] + ")";
		out.write(util.animate("animateColor", "stroke", i, fr, to));
	    }
	    if (curr.fill != null && next.fill != null) 
		if (!curr.fill.equals(next.fill)) {
		    int[] c1 = ColorConverter.toRGB(curr.fill);
		    int[] c2 = ColorConverter.toRGB(next.fill);
		    String fr = "rgb(" + c1[0] + "," + c1[1] + "," 
			+ c1[2] + ")";
		    String to = "rgb(" + c2[0] + "," + c2[1] + "," 
			+ c2[2]	+ ")";
		    out.write(util.animate("animateColor", "fill", i, fr, to));
		}
	    if (curr.getOpacity() != next.getOpacity())
		out.write(util.animate("opacity", i, "" + curr.getOpacity(), 
				       "" + next.getOpacity()));
	    i++;
	    curr = next;
	}
	float stepLength = util.getStepLength();
	float pause = util.getPause();
	float endTime = (i)*(pause + stepLength) + pause;
	if (!util.addPanel()) {
	    String opac = "0";
	    if (i == util.getMaxStep())
		opac = "0.5";
	    out.write(SVGUtil.animateAdditive("opacity", endTime, stepLength, 
					      opac, "sum"));
	}
	else if (i < util.getMaxStep() && util.addPanel())
	    out.write(util.animate("opacity", i, ""+curr.getOpacity(), "0"));
    }

    /**
     * Converts the given (x,y) value pairs to a <code>String</code>.
     * The result is in the form x[0],y[0] x[1],y[1] ...
     */
    private String pointsToString(int[] x, int[] y) {
	String pts = "";
	int i = 0;
	while (i < x.length) {
	    pts += " " + x[i] + "," + y[i];
	    i++;
	}
	return pts;
    }

    /**
     * Converts the given (x,y) values to <code>String</code>. Points must 
     * be 2-dimensional array with x-values at index 0 and y-values at
     * index 1 for this method to behave rationally. Same as 
     * pointsToString(points[0], points[1]).
     */
    private String pointsToString(int[][] points) {
	return pointsToString(points[0], points[1]);
    }
}
