/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: GraphFromFile.java,v 1.39 2007/08/20 13:30:50 oseppala Exp $ */

package matrix.util.io;

import java.awt.Color;
import java.io.*;
import java.util.*;

import matrix.structures.FDT.*;
import matrix.structures.FDT.advanced.GeometricGraph;
import matrix.structures.FDT.advanced.StyledArray;
import matrix.structures.FDT.probe.Key;
import matrix.structures.FDT.probe.VertexImpl;
import matrix.structures.FDT.substructures.Vertex;
import matrix.structures.memory.*;
import matrix.structures.simulationextensions.CallableMethods;
import matrix.structures.util.MatrixComparable;
import matrix.util.Note;


/**
 * Creates a graph from data in a file.
 *
 * Input file format: The file contains edges, one per line, in one of
 * the following formats:
 * nodename<TAB>adjname<TAB>weight
 * nodename<TAB>adjname                   (default weight)
 * nodename                               (for isolated nodes)
 *
 * nodename = name of source node
 * adjname = name of adjacent node
 * weight = a string specifying the weight of this edge
 *
 * All named nodes are created. Isolated nodenames may appear in the
 * file even if they have edges specified on other lines. Multiedges
 * are supported if the layout supports them.
 * <p>
 * This can also open a file which has been saved by Save To ASCII -method
 * and which contains a graph and any number of inner structures. 
 * <p>
 * Moreover, lines can contain additional comments, which are skipped when loading the file.
 * Comments must start with string "//". Space characters are also removed from the end of 
 * each line. Empty characters before the comment must be space characters.
 * <p>
 * example:<br>
 * <br>
 * #matrix graph adjacency-list     //this is the header of the file<br>
 * A:B C D                          //first node and its adjacents<br>
 * B:G                              //second node and its adjacents<br>
 * ...<br>
 *
 * @author Riku Saikkonen
 * @author Ari Korhonen
 * @version $Revision: 1.39 $
 */
public class GraphFromFile implements Graph, Tree, StyledArray, CallableMethods, GeometricGraph {
    public static final String GRAPH_HEADER = "#matrix graph";
    public static final String EDGELIST = "edge-list";
    public static final String ADJLIST = "adjacency-list";
    
//  private VertexImpl[] nodes;
//  Changed nodes to VirtualArray to enable adding of new nodes.
    private VirtualArray nodes;
    private Hashtable root = new Hashtable();
//  private VertexImpl[] roots;
    private VertexImpl forest;
    private Hashtable/*<String->VertexImpl>*/ nodeHash;
    boolean isGraph = false;	
    matrix.structures.memory.VirtualArray dad;
    matrix.structures.coordinates.StaticXYCoordinates ggraph = 
        new matrix.structures.adaptation.XYCoordinatesAdapter(0); // needed to store coordinates
    
    // after parsing the file this contains structures in strings
    // This shouldn't be serialized. If this is serialized, it causes problems
    // when trying to load serialized animation. That is a reason for "transient".
    transient private HashMap structuresInASCII = new HashMap();
    
    public GraphFromFile(String filename) throws IOException{
        Note.out(this, "Loading graph description...");
        nodeHash = new Hashtable();
        
        try {
            parseFile(filename);
            
            String temp = (String)structuresInASCII.get(filename);
            String header = null;
            StringTokenizer st = new StringTokenizer(temp, "\n", false);
            if (st.hasMoreTokens())
                header = st.nextToken();
            
            if ((isGraph = header.trim().equalsIgnoreCase(GRAPH_HEADER) == true) ||
                    (isGraph = header.trim().equalsIgnoreCase(GRAPH_HEADER+" "+EDGELIST) == true))
                readAndParseEdgeList(st);
            else if (isGraph = header.trim().equalsIgnoreCase(GRAPH_HEADER+" "+ADJLIST) == true)
                readAndParseAdjList(st);
        } catch (IOException e) {
            Note.err(null, "GraphFromFile reading "+filename+": "+e);
            isGraph = false;
        }
        
        if (isGraph)
            makeNodes();
        
    }
    
    
    /**
     * Constructor. This is used when opening the inner structures. The purpose
     * of this is that the original file need not to be parsed again. 
     *
     * @param sia contains all the structures in strings
     */
    public GraphFromFile(HashMap sia) {
        structuresInASCII = sia;
        nodeHash = new Hashtable();
    }
    
    
    // StructuresFromFile is not serializable. When serializing GraphFromFile,
    // this variable is not serialized.
    transient StructuresFromFile sff = null;
    /**
     * Constructor. This is used when opening the inner structures. The
     * purpose of this is that the original file need not to be parsed 
     * again.
     *
     * @param sia contains all the structures in strings
     * @param sff StructureFromFile which starts the saving process
     */    
    public GraphFromFile(HashMap sia, StructuresFromFile sff) {
        structuresInASCII = sia;
        this.sff = sff;
        nodeHash = new Hashtable();
    }
    
    
    /**
     * Returns true, if the file contains a graph.
     *
     * @param fileName file which needs to be checked
     * @return true, if the file contains a graph
     */
    public static boolean isGraph(String fileName) {
        try {
            BufferedReader reader = new BufferedReader(new FileReader(fileName));
            String line = reader.readLine();
            line = handleLine(line);
            while(!(line.toLowerCase().trim().startsWith("#matrix"))) {
                line = reader.readLine();
                line = handleLine(line);
            }
            //file contains the graph if the first header in
            //the file is one of the following
            if(line.trim().equalsIgnoreCase(GRAPH_HEADER) || 
                    line.trim().equalsIgnoreCase(GRAPH_HEADER+" "+EDGELIST) ||
                    line.trim().equalsIgnoreCase(GRAPH_HEADER+" "+ADJLIST))
                return true;
        } catch (Exception e) {
            return false;
        }
        return false;
    }
    
    /**
     * Parses a file and inserts the string representation of the each structure
     * into hashmap
     *
     * @param fileName file, which contains the structures and needs to be parsed
     */
    private void parseFile(String fileName) {
        String structureName = fileName; //the name of the top structure is the name of the file
        try {
            BufferedReader reader = new BufferedReader(new FileReader(fileName));
            String gia = ""; //graph in ascii
            String line = null;
            while((line = reader.readLine()) != null) {
                line = handleLine(line);
                while (line.equals("")) {
                    line = reader.readLine();
                    line = handleLine(line);
                }
                if(line.equals("#EOS")) {
                    // the description of last structure has ended and 
                    // the description of next structure is beginning
                    structuresInASCII.put(structureName, gia);
                    line = reader.readLine(); //this contains now name of the new structure
                    if(line != null) {
                        line = handleLine(line);
                        while (line.equals("")) {
                            line = reader.readLine();
                            line = handleLine(line);
                        }
                        structureName = line;
                        gia = "";
                    }
                    //System.out.println("creating new structureInASCII: " + line);
                }
                else if(line.startsWith("#representation") ||
                        line.startsWith("#rotated") ||
                        line.startsWith("#minimized")) {
                    //do nothing, skip these lines
                    System.out.println("Skipped: " + line);
                }
                else {
                    if(line.equals(""))
                        gia = gia.concat(" \n");
                    else
                        gia = gia.concat(line + "\n");
                    //System.out.println("adding line into gia");
                }
            }
            structuresInASCII.put(structureName, gia); //adding the last structure
            
        } catch(Exception e) {
            System.out.println("Error in parsing the file: " + fileName);
            e.printStackTrace();
        }
    }
    
    
    private static String handleLine(String str) {
        String temp = str;
        temp = removeSpacesFromBeginning(temp);
        temp = removeCommentsFromLine(temp);
        return temp;
    }
    
    /**
     * Removes space characters from the beginning of the string.
     *
     * @param str string which needs to be handled
     * @return string where the possible space characters are removed
     */
    private static String removeSpacesFromBeginning(String str) {
        String temp = str;
        while(temp.indexOf(" ") == 0) {
            if (temp.length() == 1)
                return "";
            temp = temp.substring(1);
        }
        return temp;
    }
    
    
    /**
     * Removes comments from the string. Comments starts with "//".
     *
     * @param str string which needs to be handled
     * @return string from where possible comments are removed
     */
    private static String removeCommentsFromLine(String str) {
        String temp = str;
        int indexOfComment = temp.indexOf("//");
        if (indexOfComment == -1)
            ;
        else
            temp = temp.substring(0, indexOfComment);
        temp = removeSpacesFromEnd(temp);
        return temp;
    }
    
    
    /**
     * Removes space characters from the end of a string.
     *
     * @param str string which needs to be handled
     * @return string from where possible space charachters are removed
     */
    private static String removeSpacesFromEnd (String str) {
        String temp = str;
        int indexOfLastSpace = temp.lastIndexOf(" ");
        while (indexOfLastSpace == (temp.length() - 1)) {
            if (temp.length() == 0)
                return temp;
            temp = temp.substring(0, indexOfLastSpace);
            indexOfLastSpace = temp.lastIndexOf(" ");
        }
        if (temp.lastIndexOf("'") == (temp.length() - 1))
            return temp.concat(" ");
        else
            return temp;
    }
    
    
    public boolean isGraph() { return isGraph; }
    
    private void readAndParseEdgeList(StringTokenizer st) throws IOException {
        String line;
        while(st.hasMoreTokens()) {
            line = st.nextToken();
            int tab1 = line.indexOf('\t');
            if (tab1 == -1) { //tab not found, parsing the line
                VertexImpl n = null;
                VertexImpl child = null;
                String tempName = "";
                String temp = null;
                boolean isChildFound = false;
                
                for (int index = 0; index < line.length(); index++) {
                    if (index == (line.length() - 1))
                        temp = line.substring(index);
                    else
                        temp = line.substring(index, index + 1);
                    
                    if(temp.equals(" ")) {
                        if(tempName != "") {
                            if(n == null) {
                                n = getNode(tempName);
                                tempName = "";
                            }
                            else {
                                if(!isChildFound) {
                                    n.setSubTree(child = getNode(tempName), 
                                        n.getSubTreeCount());
                                    removeFromRoots(child);
                                    isChildFound = true;
                                }
                                tempName = "";
                            }
                        }
                    }
                    else if(temp.equals("'")) {
                        if(index == line.length() - 1)
                            ;
                        else {
                            index++;
                            tempName = tempName.concat(temp + line.substring(index, index + 1));
                        }
                    }
                    else {
                        tempName = tempName.concat(temp);
                    }
                }
                if((tempName != "") && (!isChildFound)) {
                    n.setSubTree(child = getNode(tempName),
                        n.getSubTreeCount());
                    removeFromRoots(child);
                }
                
            } else {
                VertexImpl child;
                VertexImpl n = getNode(line.substring(0, tab1));
                int tab2 = line.indexOf('\t', tab1 + 1);
                if (tab2 == -1) {
                    n.setSubTree(child = getNode(line.substring(tab1 + 1)),
                        n.getSubTreeCount());
                } else {
                    n.setSubTree(child = getNode(line.substring(tab1 + 1, 
                        tab2)),
                        n.getSubTreeCount());
                    // XXX weights not supported!
                    // String weight = line.subString(tab2 + 1).trim();
                }
                removeFromRoots(child);
            }
        }
    }
    
    private void readAndParseAdjList(StringTokenizer st) throws IOException {
        String line;
        while(st.hasMoreTokens()) {
            line = st.nextToken();
            int tab1 = line.indexOf(':');
            if (tab1 == -1) {
                //getNode(line.trim());
                getNode(line);
            } else {
                VertexImpl child;
                VertexImpl n = getNode(line.substring(0, tab1));
                int index = tab1 + 1;
                String temp = null;
                String tempName = "";
                if(index == (line.length() - 1)) {
                    n.setSubTree(child = getNode(line.substring(index)),
                        n.getSubTreeCount());
                    removeFromRoots(child);
                }
                else {
                    for (index = tab1 + 1; index < line.length(); index++) {
                        if (index == (line.length() - 1))
                            temp = line.substring(index);
                        else
                            temp = line.substring(index, index + 1);
                        
                        if(temp.equals(" ")) {
                            if(tempName != "") {
                                n.setSubTree(child = getNode(tempName), 
                                    n.getSubTreeCount());
                                removeFromRoots(child);
                                tempName = "";
                            }
                        }	    
                        else if(temp.equals("'")) {
                            if(index == line.length() - 1)
                                ;
                            else {
                                index++;
                                tempName = tempName.concat(temp + line.substring(index, index + 1));
                            }
                        }
                        else {
                            tempName = tempName.concat(temp);
                        }   
                    }
                    if(tempName != "") {
                        n.setSubTree(child = getNode(tempName),
                            n.getSubTreeCount());
                        removeFromRoots(child);
                    }
                }
                
                /*	
                 //this is old way to create nodes from adjacency-list
                  int tab2;
                  do {
                  tab2 = line.indexOf(' ', tab1 + 1);
                  if (tab2 == -1) {
                  n.setSubTree(child = getNode(line.substring(tab1 + 1)),
                  n.getSubTreeCount());
                  } else {
                  n.setSubTree(child = getNode(line.substring(tab1 + 1, 
                  tab2)),
                  n.getSubTreeCount());
                  tab1 = tab2;
                  // XXX weights not supported!
                   // String weight = line.subString(tab2 + 1);
                    }
                    removeFromRoots(child);
                    } while (tab2 != -1);
                    */
                
            }
        }
    }
    
    
    
    // first root saved in case there is no source in the graph
    private VertexImpl firstRoot = null;
    private VertexImpl getNode(String key) {
        //String realKey = key.trim();
        String realKey = key;
        VertexImpl node = (VertexImpl) nodeHash.get(realKey);
        if (node == null) {
            if(structuresInASCII.get(realKey) != null) {
                String temp = (String)structuresInASCII.get(realKey);
                StringTokenizer st = new StringTokenizer(temp, "\n", false);
                String visualSettings = ""; //visualization settings
                String line = null;
                line = st.nextToken();
                while(!(line.toLowerCase().startsWith("#matrix"))) {
                    visualSettings = visualSettings.concat(line + "\n");
                    line = st.nextToken();
                }
                
                //String firstLine = st.nextToken();
                //if(firstLine.equals(TableFromFile.TABLE_HEADER)) {
                if(line.equalsIgnoreCase(TableFromFile.TABLE_HEADER) ||
                        line.equalsIgnoreCase(TableFromFile.TABLE_HEADER2)) {
                    TableFromFile tff = new TableFromFile(structuresInASCII, sff);
                    FDT fdttemp = tff.getTableFromASCII(st);
                    //saves the visualization settings for the current fdt
                    sff.setVisualSettingsForFDT(fdttemp, visualSettings);
                    node = new VertexImpl(fdttemp);
                    
                    //Loading of inner structures that are marked with "autopolymorph" tag
                } else if(line.trim().startsWith(StructuresFromFile.AUTO_POLYMORPH_HEADER)) {
                    if (line.trim().length() > StructuresFromFile.AUTO_POLYMORPH_HEADER.length() + 1) {
                        line = line.trim().substring(StructuresFromFile.AUTO_POLYMORPH_HEADER.length() + 1);
                        FDT fdttemp = sff.getPolyMorphFromASCII(line, st);
                        sff.setVisualSettingsForFDT(fdttemp, visualSettings);
                        node = new VertexImpl(fdttemp);
                    }
                    
                }
                else {
                    GraphFromFile gff = new GraphFromFile(structuresInASCII, sff);
                    FDT fdttemp = gff.getGraphFromASCII(st);
                    //saves the visualization settings for the current fdt
                    sff.setVisualSettingsForFDT(fdttemp, visualSettings);
                    node = new VertexImpl(fdttemp);
                }
            }
            else {
                /*
                 If the key is a dublicate then it has "_number" in its end, 
                 so that it is possible to differentiate it from other 
                 dublicates. Next clauses removes that postfix so that 
                 only the original keys are shown. Process name removes also
                 all the escape characters "'" from the key so that the
                 original saved key can be returned. Escape character 
                 makes it possible that the key can contain space characters
                 and other special characters (eg. "_", "#", "'") which can
                 have a special meaning when parsing the keys.
                 */  
                String originalKey = realKey;
                
                originalKey = processName(originalKey);
                
                //node = new VertexImpl(realKey);
                node = new VertexImpl(originalKey);
            }
            nodeHash.put(realKey, node);
//          addToRoots(node);
            if (firstRoot == null)
                firstRoot = node;
        }
        
        return node;
    }
    
    
    /**
     * Processes the name of the key. This removes all the escape characters "'"
     * from the key and returns back original keys. With help of escape character
     * it is possible that the key can contain space characters and other special
     * characters (eg. "_", "#", "'") which can have a special meaning when 
     * parsing the keys.
     *
     * @param name name of the key in a saved form
     * @return name of the original key where all the markup characters have been removed
     */
    private String processName(String name) {
        String tempName = "";
        String temp = null;
        int index;
        if (name != "") {
            for(index = 0; index < name.length(); index++) {
                if(index == (name.length() - 1))
                    temp = name.substring(index);
                else
                    temp = name.substring(index, index + 1);
                
                if(temp.equals("'")){
                    if(index == name.length() - 1) {
                    } else {
                        index++;
                        tempName = tempName.concat(name.substring(index, index + 1));
                    }
                }
                else if(temp.equals(" "))
                    return tempName;
                else if(temp.equals("_")) {
                    int i = name.lastIndexOf("_");
                    if((i == index) && (index > 0) && (index < name.length() - 1)){
                        String postfix = name.substring(index + 1);
                        try {
                            if(Integer.parseInt(postfix) > 0) {
                                return tempName;
                            }
                            else {
                                tempName = tempName.concat(temp);
                            }
                        } catch (Exception e) {
                            //catches possible exception in Integer.parseInt(temp)
                            //exception occurs if the String temp does not contain a parsable integer.
                            tempName = tempName.concat(temp);
                        }
                    }
                    else {
                        tempName = tempName.concat(temp);
                    }
                }
                else {
                    tempName = tempName.concat(temp);
                }
            }
        }
        return tempName;		       
    }
    
    private void removeFromRoots(VertexImpl node) { 
        root.remove(node);
    }
    
    private void addToRoots(VertexImpl node) {
        root.put(node, node);
    }
    
    private void makeNodes() {
        Enumeration e = nodeHash.elements();
        int n = nodeHash.size();
        //      nodes = new VertexImpl[n];
        nodes = new VirtualArray(null, "Nodes of the graph (read from a file)");
        for (int i = n-1; i >= 0; i--) { // reverse order in enumeration 
//          nodes[i] = (VertexImpl) e.nextElement();
            nodes.setObject((VertexImpl)e.nextElement(), i);
        }
        /*
         roots = new VertexImpl[root.size()];
         int next = 0;
         for (int i=0; i<n; i++) {
         if (root.containsKey(nodes[i]))
         roots[next++] = nodes[i];
         }
         
         if (next > 1) { // many roots -> we have forest
         forest = new VertexImpl(new Key("Forest"));
         for (int i=0; i<next; i++) 
         forest.setSubTree(roots[i],i);
         } else if (root.size() == 0) { // no root at all, maybe cycle
         roots = new VertexImpl[1];
         //roots[0] = (forest = firstRoot);
          roots[0] = firstRoot;
          forest = new VertexImpl(new Key("Forest"));
          forest.setSubTree(roots[0], 0);
          }
          else {// only one root, no separate forest-root needed
          //forest = roots[0];
           forest = new VertexImpl(new Key("Forest"));
           forest.setSubTree(roots[0], 0);
           }
           */
        forest = new VertexImpl(new Key("Forest"));
        for (int i = 0; i < nodes.size(); i++) {
            forest.setSubTree((VertexImpl)nodes.getObject(i), i);
        }
        dad = new matrix.structures.memory.VirtualArray(1, null, "fathers of nodes in the graph (spanningtree)");
    }
    
    
    /**
     * Returns FDT which corresponds the string representation which is
     * given through the StringTokenizer. Graph in the string is in
     * adjacency-list format.
     * 
     * @param st StringTokenizer which contains the keys of the structure
     * @return returns the corresponding FDT
     */
    public FDT getGraphFromASCII(StringTokenizer st) {
        try {
            //graph is in adjacency-list format
            readAndParseAdjList(st);
            makeNodes();
        } catch (Exception e) {
            e.printStackTrace();
        }
        return this;
    }
    
    
    /**
     * Returns FDT which corresponds the string representation which is
     * given through the StringTokenizer. Graph in the string is in
     * edge-list format.
     *
     * @param st StringTokenizer which contains the keys fo the structure
     * @return returns the corresponding FDT
     */
    public FDT getGraphFromASCII2(StringTokenizer st) {
        try {
            //graph is in edge-list format
            readAndParseEdgeList(st);
            makeNodes();
        } catch (Exception e) {
            e.printStackTrace();
        }
        return this;
    }
    
    
    public Vertex[] getVertices() { // for Graph
        VertexImpl[] temp = new VertexImpl[nodes.size()];
        for (int i = 0; i < nodes.size(); i++) {
            temp[i] = (VertexImpl)nodes.getObject(i);
        }
        return temp;
    }
    
    public void setVertices(Vertex[] v) {
        int i = 0;
        for(; i < v.length; i++)
            nodes.setObject(v[i], i);
        for(; i < nodes.size(); i++)
            nodes.setObject(null, i);
    }
    
    private VertexImpl getNewVertex(Object o) {
        return new VertexImpl(o);	
    }
    
    public void addVertex(Object o) {
        VertexImpl vi = null;
        if(o instanceof VertexImpl)
            vi = (VertexImpl)o;
        else if(o instanceof MatrixComparable){
            vi = getNewVertex(o);
        }
        else if(o instanceof FDT) {
            vi = getNewVertex(new Key(((FDT)o).getElement()));
        }
        else
            vi = getNewVertex(new Key(o));
        for(int i =  0; i < nodes.size(); i++) {
            if(nodes.getObject(i) == null) {
                nodes.setObject(vi, i);
                return;
            }
        }
        nodes.setObject(vi, nodes.size());
    }
    
    
    public void removeVertex(Vertex v) {
        int i;
        Note.out(this, "deleting, size " + nodes.size());
        for (i = 0; i < nodes.size(); i++) {
            if(((Vertex)nodes.getObject(i)).equals(v)) {
                break;	
            } else
                ((VertexImpl)nodes.getObject(i)).removeSuccessor(v);
        }
        for(;i < nodes.size(); i++) {
            if(i < nodes.size() - 1) {
                VertexImpl vi = (VertexImpl)nodes.getObject(i + 1);	
                vi.removeSuccessor(v);
                nodes.setObject(vi, i);
            }
        }
        nodes.setLast(i - 2);
    }
    
    
    
    // Old, removed method from Graph
    public void setVertex(Graph dg, int i) { // for Graph
        Note.err(this, "Not implemented (yet?)");
    }
    
    public void insertVertex(Object o) {
        Note.err(this, "Not implemented (yet?)");
    }
    
    public void deleteVertex(Vertex v) {
        Note.err(this, "Not implemented (yet?)");
    }
    public void insertEdge(Vertex from, Vertex to) {}
    
    public void deleteEdge(Vertex from, Vertex to) {}
    
    public int getSubTreeCount() {
        return nodes.size();/*roots.length;*/ 
    } // for Tree

    /** Return: table of referencies to subtrees */
    public Tree[] getSubTrees() {
        return (Tree[])getVertices();/*roots;*/
    }
 
    /** Return: the element stored in this node */
    public Object getElement() {
        return forest.getElement();
    }

    /** Set new element into this node */
    public void setElement(Object v) {
        forest.setElement(v);
    }
        
    public Object getObject(int address) { 
        return dad.getObject(address);
    }
    
    public void setObject(Object o, int address) { 
        dad.setObject(o, address);
    }
    
    public int getFirst() { 
        return dad.getFirst();
    }
    
    public int getLast() { 
        return dad.getLast();
    }
    
    public void setFirst(int i) { 
        dad.setFirst(i);
    }
    
    public void setLast(int i) { 
        dad.setLast(i);
    }
    
    public String getIndexName(int address) {
        //return nodes[address].getElement().toString();
        return ((VertexImpl)nodes.getObject(address)).getElement().toString();
    }
    
    // ----- Simulation Extensions -----
    
    public String[] getCallableMethodNames() {
        return new String[] { "DFS", "BFS" };
    }
    
    private int getIndex(Vertex vi) {
        for (int i=0; i<nodes.size(); i++)
            //if (nodes[i].equals(vi))
            if (((VertexImpl)nodes.getObject(i)).equals(vi))
                return i;
        return -1;
    }
    
    private void setParent(VertexImpl node, VertexImpl parent) {
        if (parent != null)
            dad.setObject(new Key(parent.getElement().toString()), getIndex(node));
        else 
            dad.setObject(new Key("?"), getIndex(node));
        node.setParent(parent);
    }
    
    private int time;
    
    /*    public void DFS() {
     for (int i=0; i<nodes.length; i++) {
         nodes[i].setColor(Color.white);
         setParent(nodes[i], null);
     }
     time = 0;
     for (int i=0; i<nodes.length; i++)
         if (nodes[i].getColor().equals(Color.white))
     DFSVisit(nodes[i]);
     }
     */
    
    
    public void DFS(Object startNode) {
        for (int i=0; i<nodes.size(); i++) {
//          nodes[i].setColor(Color.white);
            ((VertexImpl)nodes.getObject(i)).setColor(Color.white);
            //setParent(nodes[i], null);
            setParent((VertexImpl)nodes.getObject(i), null);
//          nodes[i].setLabel(null);
            ((VertexImpl)nodes.getObject(i)).setLabel(null);
            ((VertexImpl)nodes.getObject(i)).setLabelEnabled(false);
        }
        time = 0;
        
        if (startNode instanceof VertexImpl) {
            VertexImpl temp = (VertexImpl)startNode;
            for (int i=0; i<nodes.size(); i++) {
                if((VertexImpl)nodes.getObject(i) == temp) {
                    DFSVisit((VertexImpl)nodes.getObject(i));
                    break;
                }
            }
        }
        
        for (int i=0; i<nodes.size(); i++)
            if (((VertexImpl)nodes.getObject(i)).getColor().equals(Color.white))
                DFSVisit((VertexImpl)nodes.getObject(i));
    }
    
    
    private void DFSVisit(VertexImpl u) {
        u.setColor(Color.gray);
        u.setLabel(time+++"/");
        VertexImpl[] v = (VertexImpl[])u.getSuccessors();
        if (v != null)
            for (int i=0; i<v.length; i++)
                if (v[i].getColor().equals(Color.white)) {
                    setParent(v[i], u);
                    DFSVisit((VertexImpl)v[i]);
                }
        u.setColor(Color.black);
        u.setLabel(u.getLabel()+time++);
        u.setLabelEnabled(true);
    }
    
    
    /*
     public void BFS() {
         for(int i = 0; i<nodes.length;i++) {
             nodes[i].setColor(Color.white);
             setParent(nodes[i], null);
             nodes[i].setLabel(null);
         }
         time = 0;
         Vector queue = new Vector();
         nodes[0].setColor(Color.gray);
         nodes[0].setLabel(""+time);
         queue.addElement(nodes[0]);
         while(!queue.isEmpty()) {
             VertexImpl v = (VertexImpl)queue.elementAt(0);
             queue.removeElementAt(0);
             time++;
             Vertex[] adj = v.getSuccessors();
             for(int i = 0; i < adj.length;i++) {
                 VertexImpl a = (VertexImpl)adj[i];
                 if(a.getColor().equals(Color.white)) {
                     setParent(a, v);
                     a.setColor(Color.gray);
                     a.setLabel(""+time);
                     queue.addElement(a);
                 }
             }
             v.setColor(Color.black);
         }
     }
     */
    
    
    
    public void BFS(Object startNode) {
        for(int i = 0; i<nodes.size();i++) {
            ((VertexImpl)nodes.getObject(i)).setColor(Color.white);
            setParent((VertexImpl)nodes.getObject(i), null);
            ((VertexImpl)nodes.getObject(i)).setLabel(null);
            ((VertexImpl)nodes.getObject(i)).setLabelEnabled(false);
        }
        time = 0;
        Vector queue = new Vector();
        
        if (startNode instanceof VertexImpl) {
            VertexImpl temp = (VertexImpl)startNode;
            for(int i=0; i<nodes.size(); i++) {
                if((VertexImpl)nodes.getObject(i) == temp) {
                    ((VertexImpl)nodes.getObject(i)).setColor(Color.gray);
                    ((VertexImpl)nodes.getObject(i)).setLabel(""+time);
                    ((VertexImpl)nodes.getObject(i)).setLabelEnabled(true);
                    queue.addElement((VertexImpl)nodes.getObject(i));
                    break;
                }
            }
        }
        else {
            ((VertexImpl)nodes.getObject(0)).setColor(Color.gray);
            ((VertexImpl)nodes.getObject(0)).setLabel(""+time);
            ((VertexImpl)nodes.getObject(0)).setLabelEnabled(true);
            queue.addElement((VertexImpl)nodes.getObject(0));
        }
        
        while(!queue.isEmpty()) {
            VertexImpl v = (VertexImpl)queue.elementAt(0);
            queue.removeElementAt(0);
            time++;
            Vertex[] adj = v.getSuccessors();
            for(int i = 0; i < adj.length;i++) {
                VertexImpl a = (VertexImpl)adj[i];
                if(a.getColor().equals(Color.white)) {
                    setParent(a, v);
                    a.setColor(Color.gray);
                    a.setLabel(""+time);
                    a.setLabelEnabled(true);
                    queue.addElement(a);
                }
            }
            v.setColor(Color.black);
        }
    }
    
    // interface GeometricGraph, ie, XYCoorinates
    public int getX(Vertex vertex) { 
        return ggraph.getX(getIndex(vertex)); 
    }

    public int getY(Vertex vertex) {
        return ggraph.getY(getIndex(vertex)); 
    }
    
    public void setCoordinates(Vertex vertex, int x, int y) {
        ggraph.setCoordinates(getIndex(vertex), x, y);
    }
    
    static final long serialVersionUID = 6006491725339480263L;

    public void setSubTree(Tree t, int child) {
        // TODO Auto-generated method stub
        
    }


    public Tree getNewNode(Object key) {
        // TODO Auto-generated method stub
        return null;
    }
}

