/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package matrix.visual;

import java.lang.reflect.*;
import java.util.*;

import matrix.structures.CDT.CDT;
import matrix.structures.FDT.FDT;
import matrix.structures.FDT.probe.Key;
import matrix.util.Note;
import matrix.util.configuration.*;


/**
 * RepresentationFactory creates visual representations for data structures
 * visualized using Matrix. The functionality of RepresentationFactory is
 * configured from the configuration file.
 *
 * Java's own configuration class, java.util.Properties, isn't used since
 * it cannot handle duplicate keys. Also, there are several different
 * fields that need to be configured, which is impossible using Properties.
 *
 * @author Jussi Nikander
 */

/*$Id: RepresentationFactory.java,v 1.80 2006/05/30 10:02:31 oseppala Exp $*/

public class RepresentationFactory {


    private static RepresentationFactory theInstance = null;
    private static String url = null;
    //    private static String confPath = null;


    private RepresentationFactory() {
	try {
	    // force MatrixConfiguration to parse configuration file 
	    // (psilvast 11.9.2002)
	    //matrix.util.configuration.MatrixConfiguration.getInstance();
	    // removed - jtn 25.8.03
	}
	catch(Exception e) {
	    e.printStackTrace();
	}
    }

    public void addVisualizations(CVisualizations c) {
	int size = 0;
	// Now we count the number of different visualizations
	for(int i = 0; i < c.getChildCount(); i++) {
	    // ci should be a CStructure, but that's not important
	    ConfItem ci = c.getChild(i);
	    size += ci.getChildCount();
	}
	// create the array holding the visualizations. This used to be
	// created in MatrixConfiguration. 
	// format: Class to be visualized(Class), visualization(Class)
	// visualization name (String)
	visualizations = new Object[3][size];
	size = 0;
	for(int i = 0; i < c.getChildCount(); i++) {
	    CStructure ci = (CStructure)c.getChild(i);
	    Class cls = ci.getStructure();
	    for(int j = 0; j < ci.getChildCount(); j++) {
		CVisualization cv = (CVisualization)ci.getChild(j);
		Class vis = cv.getVisualization();
		String vName = cv.getName();
		//Note.out(this, "Adding: " + cls + ":" + vis + ":" +vName+".");
		visualizations[0][size] = cls;
		visualizations[1][size] = vis;
		visualizations[2][size] = vName;
		size++;
	    }
	}
    }

    public void addPickedVisualizations(CPickedVisualizations c) {
	pickedVisualizations = new HashMap();
	for(int i = 0; i < c.getChildCount(); i++) {
	    CStructure cs = (CStructure)c.getChild(i);
	    CVisualization cv = (CVisualization)cs.getChild(0);
	    Class str = cs.getStructure();
	    java.util.StringTokenizer st = new java.util.StringTokenizer(str.toString());
	    st.nextToken();
	    String key = st.nextToken();
	    //Note.out(this, "Putting " + key + ", " + cv.getVisualization());
	    pickedVisualizations.put(key, cv.getVisualization());
	}
    }

    

    /**
     * Retrieves the single instance of RepresentationFactory
     * 
     * @return the instance of RepresentationFactory
     */
    public static RepresentationFactory getInstance() {
	if(theInstance == null)
	    theInstance = new RepresentationFactory();
	return theInstance;
    }

    /**    
     * Sets the URL of the matrix configuration file.
     * Eg. "http://www.cs.hut.fi/Research/Matrix/applet/matrixconf.xml"
     * <p>
     * If the URL is set, then matrix configuration is read from the network,
     * not from the file system. This can be convenient with applets.
     */
    public static void setUrl(String s) {
        url = s;
    }

    /**    
     * Returns the URL of the matrix.cnf file.
     * Eg. "http://www.cs.hut.fi/Research/Matrix/applet/matrixconf.xml"
     */
    public static String getUrl() {
        return url;
    }


    // -=- static interface starts here. -=-

    private static Object[][] visualizations = null;
    private static HashMap pickedVisualizations = null;

    public static void setPickedVisualizations(HashMap h) {
	pickedVisualizations = h;
    }
    
    public static void setVisualizations(Object[][] o) {
	visualizations = o;
    }

    /* helper function for finding the right constructor and to create the VisualType
     * argument is the Object that the representation should be created for
     * vt is the desired visualization (an instance of VisualType)
     */
    private static VisualType getVisualType(Object argument, Class vt) {
	Constructor[] c = null;
	Object o;
	VisualType type;
	//Class pc = getInstance().pc;
	Class ar = argument.getClass();
	Class[] args;
	try {
	    c = vt.getConstructors();
	}catch(SecurityException e) {
	    Note.out(getInstance(), e.getMessage());
	    e.printStackTrace();
	}

	try {

	    for(int i = 0; i < c.length; i++) {
		args = c[i].getParameterTypes();
		if(args.length == 1 && args[0].isAssignableFrom(ar)) {
		    //Note.out(null, "New Visualtype: " + c[i] + " with " + argument);
		    o = c[i].newInstance(new Object[] { argument } );
		    if(o instanceof VisualType)
			type = (VisualType)o;
		    else
			type = null;
		    return type;
		}
	    }
	}catch(InvocationTargetException f) {
	    Note.out(getInstance(), "InvocationTargetException!");
	    f.getTargetException().printStackTrace();
	}catch(Exception e) {
	    Note.out(getInstance(), "Exception caught: " + e.getMessage());
	    e.printStackTrace();
	}
	return null;
    }

    /**
     * Creates a default VisualType representation for the parameter
     * Object. If there is no representation for the parameter Object
     * a VisualClass -representation is used for all FDTs and a VisualKey
     * -representation for all non-FDTs. 
     *
     * @param o the Object the VisualType is to represent
     * @return the VisualType representing parameter Object
     */

    public static VisualType getRepresentation(Object o) {

	// force parsing of conf file
	getInstance();

	VisualType vt;
	if(o == null) {
	    return new VisualKey(new Key(o));
	}

	Class cl = o.getClass();
	for(int i = 0; i < visualizations[0].length; i++) {
	    if(((Class)visualizations[0][i]).isAssignableFrom(cl)) {
		vt = getVisualType(o, (Class)visualizations[1][i]);
		if(vt != null)
		    return vt;
	    }
	}

	if (o instanceof java.awt.Component) {
	    return new VisualAWTComponent((java.awt.Component)o);
	}

	if(o instanceof CDT) {
	    vt = getRepresentation(((CDT)o).getElement());
	    if(vt != null) {
		vt.setStructure((FDT)o);
		return vt;
	    }
	}
	if(o instanceof FDT)
	    return new VisualClass(o);
	else
	    return new VisualKey(new Key(o));
    }

    /**
     * Creates a representation of the parameter Object according to the
     * parameter String. If no named representation is found, a default
     * representation is returned.
     * 
     * @param ss the Object to be given a representation
     * @param type the name of the representation
     * @return a VisualType representing the param Object
     */
    public static VisualType getRepresentation(Object ss, String type) {
	getInstance();
	VisualType vt = null;
	for(int i = 0; i < visualizations[0].length; i++) {
	    if(type.equals((String)visualizations[2][i])) {
		vt = getVisualType(ss, (Class)visualizations[1][i]);
		if(vt != null) {
		    return vt;
		}
	    }
	}
	if(ss instanceof CDT) {
	    vt = getRepresentation(((CDT)ss).getElement(), type);
	    if(vt != null) {
		vt.setStructure((FDT)ss);
		return vt;
	    }
	}
	return getRepresentation(ss);
    }

    /**
     * Creates a new VisualContainer representation for the parameter VisualType.
     * This method is used by matrix.simulation.Picker to create new VisualContainers
     * to hold Objects when Objects are dragged in Matrix. The return type is VisualType
     * in order to preserve compability with other getRepresentation -methods.
     *
     * @param vt the VisualType holding the param Object.
     * @return a VisualContainer that holds the same Object.
     */
    public static VisualType getRepresentation(VisualType vt) {
	getInstance();
	if(vt == null)
	    return new VisualKey(new Key(vt));
	String name = vt.getClass().getName();
	if(pickedVisualizations.containsKey(name)) {
	    VisualType visual = getVisualType(vt.getStructure(), (Class) pickedVisualizations.get(name));
	    if(visual != null) {
		Note.out(getInstance(), "Picking visualization " + visual);		
		return visual;
	    }
	}
	return getRepresentation(vt.getStructure());
    }

    /** 
     * Returns the representation name associated to the parameter VisualType. If, for some reason,
     * several representation names are associated with the same VisualType, the first encountered
     * match is returned.
     * 
     * @param type the VisualType
     * @return a String containing the representation name or null if none found
     */
    
    public static String getRepresentationName(VisualType type) {
    	if(type == null) {
    		Note.out(null, "Cannot find representation name for VisualType null");
    		return null;
    	}
    	Class c = type.getClass();
    	return getRepresentationName(c);
    }
    
    /**
     * Returns a representation name associated with the parameter Class. For the representation
     * name to be found the parameter must be Class that represents a class for which a representation
     * has been configured in the visualizations -part of the configuration file. If, for somre reason,
     * several representation names are associated with the same VisualType, the first encountered match
     * is returned.
     * 
     * @param c the Class for which a representation name must be found
     * @return a String containing the representation name or null if none found 
     */
    
    public static String getRepresentationName(Class c) {
    	if(c == null) {
    		Note.out(null, "Cannot find representation for Class null");
    		return null;
    	}
    	getInstance();
    	String cName = c.getName();
    	for(int i = 0; i < visualizations[1].length;i++) {
    		if(((Class)visualizations[1][i]).getName().equals(cName)) {
    			return (String)visualizations[2][i];
    		}
    	}
    	return null;
    }
    
    /**
     * Creates a String array holding the names of the possible representations
     * of the parameter FDT.
     * 
     * @param ss the FDT to be examined.
     * @return an array of the names of possible representations.
     */
    public static String[] getRepresentationNames(FDT ss) {
	Note.out(null, "getting reprnames for class " + ss);
	getInstance();
	if(ss == null) {
	    Note.out(null, "Returning null repr names");
	    return new String[] {};
	}
	Vector v = new Vector();
	Class c = ss.getClass();
	for(int i = 0; i < visualizations[0].length; i++) {
	    if(((Class)visualizations[0][i]).isAssignableFrom(c))
		v.addElement(visualizations[2][i]);
	}
	if(ss instanceof CDT) {
	    FDT fdt = (FDT) ss.getElement();
	    // it is possible for an CDT to return a key, if the 
	    // CDT also can be used as an FDT. 
	    // We do not want to have "key" as possible representation
	    if(!(fdt instanceof Key)) {
		String[] more = getRepresentationNames(fdt);
		for(int i = 0; i < more.length; i++) {
		    if(v.contains(more[i]))
			continue;
		    v.addElement(more[i]);
		}
	    }
	}
	Enumeration e = v.elements();
	String s[] = new String[v.size()];
	int i = 0;
	while(e.hasMoreElements())
	    s[i++] = (String) e.nextElement();
	Note.out(null, "Returning reprnames for " + ss + " : ");
	for(i = 0; i < s.length; i++) {
	    Note.out(null, "   " + s[i]);
	}
	return s;
    }

    //-=- Dummy implementations for creating add -menu -=-
    

    private static String[] structureNames = null;

    public static void setStructureNames(String[] s) {
	structureNames = s;
    }

    /** Return an array of possible representations known to the system. 
     * @see #getFDT(int)
     */
    public static String[] getStructureNames() {
	getInstance();
        return structureNames;
    }

    private static MatrixConstructor[] fdts = null;

    public static void setFDTs(MatrixConstructor[] m) {
	fdts = m;
    }
    
    /** Return an example of the i'th known representation. 
     * @see #getStructureNames()
     */
    public static FDT getFDT(int i) {
	getInstance();
	if(i > 0 && i < fdts.length && fdts[i] != null) 
	    try {
		return (FDT) fdts[i].newInstance();
	    }catch(Exception e) {
		Note.err(null, "RepresentationFactory: Cannot instantiate " + fdts[i]);
	    }
	return null;

    }

}
