/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2006 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * Erilaisia toimintoja, joita kyttjn on pystyttv tekemn aluevisualisaatioon:                            
                                                                                                             
Kyttjn on voitava list visualisaatioon uusia elementtej raahamalla niit hiiren avulla toisista        
rakenteista.                                                                                                 
                                                                                                             
Kyttjn on voitava poistaa visualisaatiosta elementtej.                                                   
                                                                                                             
aluevisualisaatioon on pystyttv lismn ennalta mrtyn tyyppisi elementtej (esimerkiksi pisteit)    
hiiren avulla. Kyttj voi esimerkiksi napsauttaa tietty kohtaa visualisaatiosta, jolloin thn kohtaan    
(koordinaateihin joissa hiiren nappia napsautettiin) listn uusi piste. Elementin lisminen               
delegoidaan alla olevalle rakenteelle insert-metodin avulla. Rakenteelle on annettava tietona listtvn     
elementin koordinaatit sek tyyppi.                                                                          
                                                                                                             
Aluevisualisaatiosta on pystyttv valitsemaan aktiiviseksi yksi ( kenties jopa useampia) elementti,         
joille pystytn sitten tekemn jokin mielivaltainen operaatio tehtvien buttoneita kyttmll.            
Valinnassa oltava kaksi toimitilaa. Toisessa valitulle rakenteen osalle tehdn vlittmsti operaatio ja    
toisessa aktiivinen osa valitaan ensin ja operaatio tehdn valitsemalla sopiva button. (vrt. esimerkiksi    
Dijkstran algoritmi-tehtvt).                                                                               
                                                                                                             
Aluevisualisaatioon on pystyttv piirtmn rakenteita raahaamalla hiiren avulla. Tyypillisesti             
piirrettvt rakenteet ovat suoria viivoja tai murtoviivoja. Yhdess tehtvss on pystyttv piirtmn     
janan keskinormaali, mutta tm voidaan luultavasti toteuttaa buttoneiden avulla.  
 */

package matrix.visual;

import java.awt.AlphaComposite;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Composite;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.RenderingHints;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.geom.AffineTransform;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import matrix.decoration.LabelDecorator;
import matrix.structures.CDT.CDT;
import matrix.structures.FDT.FDT;
import matrix.structures.simulationextensions.Selectable;
import matrix.structures.spatial.Area;
import matrix.structures.spatial.PaintingStyleDecorator;
import matrix.structures.spatial.SpatialComparable;
import matrix.structures.spatial.FDT.probe.Point;
import matrix.structures.spatial.FDT.probe.PolyLine;
import matrix.structures.spatial.FDT.probe.Polygon;
import matrix.structures.spatial.FDT.probe.SpatialElement;
import matrix.structures.spatial.FDT.probe.SpatialElementTable;

/**
 * This class is used to visualize any spatial data structures 
 * that implement the interface Area.
 *
 * @author Otto Seppl, Juha Helminen
 * @version $Revision: 1.40 $
 * 
 * @see matrix.structures.spatial.Area
 * @see matrix.structures.spatial.FDT.probe.SpatialElement
 */

public class VisualArea extends VisualContainer {

    // Defines the radius of the circular area around a point,
    // where the focus is moved to the point entity (for highlighting/selection).
    private static final int POINT_FOCUS_THRESHOLD = 8;
    // The stroke used to generate the selection area for spatial elements' outlines.
    private static final int SELECTION_WIDTH = 10;
    private static final Stroke SELECTION_STROKE = new BasicStroke(SELECTION_WIDTH); 

    
    // ** Default color and style scheme **
   
    // The color for the background of the drawing area.
    private static final Color BACKGROUND_COLOR = Color.LIGHT_GRAY;
    // The colors for the axis lines and numbers.
    private static final Color X_AXIS_COLOR = Color.DARK_GRAY;
    private static final Color Y_AXIS_COLOR = Color.DARK_GRAY;
    private static final double AXIS_FONT_SCALING = 0.7;
    private static final double POINT_LABEL_FONT_SCALING = 0.8;
    // The radius of a circle visualizing a single point.
    private static final int POINT_SIZE = 4;

    // The decorator used for spatial elements that haven't got one.
    private static final PaintingStyleDecorator DEFAULT_ELEMENT_DECORATOR;
    // The decorator used for areas that haven't got one.
    private static final PaintingStyleDecorator DEFAULT_AREA_DECORATOR;
    // The decorator used for points that haven't got one.
    private static final PaintingStyleDecorator DEFAULT_POINT_DECORATOR;
    
    static {
        DEFAULT_ELEMENT_DECORATOR = new PaintingStyleDecorator();
        DEFAULT_ELEMENT_DECORATOR.setLineColor(Color.GREEN);
        DEFAULT_ELEMENT_DECORATOR.setStroke(3, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND, 10.0f, 
                                            new float[]{4f, 4f}, 0);
        DEFAULT_ELEMENT_DECORATOR.setFillColor(Color.GREEN);
        DEFAULT_ELEMENT_DECORATOR.setTransparency(0.1f);
        
        DEFAULT_AREA_DECORATOR = new PaintingStyleDecorator();
        DEFAULT_AREA_DECORATOR.setLineColor(Color.RED);
        DEFAULT_AREA_DECORATOR.setStroke(3, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND, 10.0f, 
                                         new float[]{4f, 4f}, 0);
        DEFAULT_AREA_DECORATOR.setFillColor(Color.RED);
        DEFAULT_AREA_DECORATOR.setTransparency(0.1f);

        DEFAULT_POINT_DECORATOR = new PaintingStyleDecorator();
        DEFAULT_POINT_DECORATOR.setLineColor(Color.BLUE);
        DEFAULT_POINT_DECORATOR.setStroke(3, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND, 10.0f, 
                                          new float[]{4f, 4f}, 0);
        DEFAULT_POINT_DECORATOR.setFillColor(null);
        DEFAULT_POINT_DECORATOR.clearTransparency();
    }

    // Overshoot of the grid
    private static final int MAJOR_OVERSHOOT = 5;

    private static final int VERTICAL_ADD   = 25;
    private static final int HORIZONTAL_ADD = 25;
    private static final int VERTICAL_BASE  = 5;
    private static final int HORIZONTAL_BASE= 5;

    private matrix.structures.spatial.Area area;
    private AreaVisualization areaPanel;

    private static final int MIN_SIZE = 50;
    private Dimension layoutSize = new Dimension(MIN_SIZE, MIN_SIZE);
    private Rectangle2D minimumBoundary = new java.awt.geom.Rectangle2D.Double(0, 0, MIN_SIZE, MIN_SIZE); 

    // Axis direction
    private static final boolean MATH_AXIS = true;
    
    private int xOrigin;
    private int yOrigin;
    
    private int xMax;
    private int yMax;

    private int rightInset;    
    private int leftInset;
    private int topInset;
    private int bottomInset;

    public static final int next10(int a) {
        return (int)(Math.pow(10, 1 + Math.floor(Math.log(a) / Math.log(10) - 0.0001f)));
    }

    
    public VisualArea(matrix.structures.spatial.Area area) {
        this((FDT)area);
    }
    
    public VisualArea(FDT area) {
        super(area);
                
        if (area instanceof matrix.structures.spatial.Area == false){
            return;
        } else {
            this.area = (matrix.structures.spatial.Area)area;
        }
    }

    public void insert(Object structure) {
        if (getStructure() instanceof CDT) {
            CDT cdt = (CDT)getStructure();
            FDT newRoot = cdt.insert(structure);
            if (newRoot != getStructure()) {
                remove(areaPanel);
                areaPanel = null;
                setStructure(newRoot);
                init();
            }
        } else { 
            getApplication().showMessage("Sorry", "Cannot insert into non CDT " + this);
        }
    }

    public void delete(Object structure) {
        if (getStructure() instanceof CDT) {
            CDT cdt = (CDT)getStructure();
            FDT newRoot = cdt.delete(structure);
            if (newRoot != getStructure()) {
                remove(areaPanel);
                areaPanel = null;
                setStructure(newRoot);
                init();
            }
        } else { 
            getApplication().showMessage("Sorry", "Cannot delete from non CDT " + this);
        }
    }
    
    public void validateComponents() {
        super.validateComponents();
        init();
    }

    
    private boolean calledFromInit = false;
    
    public void init() {
        // if (this.areaPanel == null || getFont().equals(oldFont) == false){
        // Oldfont is used to check for font size changes --- should possible be
        // changed to track inset changes

        if (this.areaPanel == null) {
            this.areaPanel = new AreaVisualization((matrix.structures.spatial.Area)area);
            add(this.areaPanel);
        }

        xOrigin = (int) area.getBoundingPolygon().getX();
        yOrigin = (int) area.getBoundingPolygon().getY();
        xMax = (int) area.getBoundingPolygon().getWidth() + xOrigin;
        yMax = (int) area.getBoundingPolygon().getHeight() + yOrigin;

        leftInset = HORIZONTAL_BASE + HORIZONTAL_ADD + getInsets().left;
        rightInset = HORIZONTAL_BASE + getInsets().right;

        if (MATH_AXIS) {
            topInset = VERTICAL_BASE + getInsets().top;
            bottomInset = VERTICAL_BASE + VERTICAL_ADD + getInsets().bottom;
        } else {
            topInset = VERTICAL_BASE + VERTICAL_ADD + getInsets().top;
            bottomInset = VERTICAL_BASE + getInsets().bottom;
        }

        calledFromInit = true;
        doActualLayout();
        calledFromInit = false;

        // System.err.println("Added");
        // }
    }

    protected LayoutSize doActualLayout() {
        Rectangle2D bound = ((Rectangle2D)(area.getBoundingPolygon()).clone());
        
        // Rectangle2D bound = new java.awt.geom.Rectangle2D.Double(r.getX(),
        // r.getY(), r.getWidth()+10, r.getHeight()+10);
        bound.add(minimumBoundary);
        
        if (!calledFromInit) {
            init();
        }
        
        areaPanel.setBounds(leftInset, topInset, 
                            (int)bound.getX() + (int)bound.getWidth() + 1, 
                            (int)bound.getY() + (int)bound.getHeight() + 1);
        
        layoutSize = new Dimension(leftInset + rightInset + (int)bound.getWidth() + 1, 
                                   topInset + bottomInset + (int)bound.getHeight() + 1);
        return new LayoutSize((int)layoutSize.getWidth(), (int)layoutSize.getHeight());
    }

    
    public void paintComponent(Graphics g) {
        super.paintComponent(g);
        if (isPicked()) {
            return; // no frames
        }

        Graphics2D g2 = (Graphics2D)g;

        Object antialiasOriginal = null;
        antialiasOriginal =  g2.getRenderingHint(RenderingHints.KEY_ANTIALIASING);
        g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        
        Font origFont = g2.getFont();
        g2.setFont(origFont.deriveFont(AffineTransform.getScaleInstance(AXIS_FONT_SCALING, AXIS_FONT_SCALING)));
                
        int height = getHeight();
        int width  = getWidth();
        
        int visibleWidth  = width  - rightInset - leftInset;
        int visibleHeight = height - topInset - bottomInset;
                
        int dataWidth  = xMax - xOrigin;
        int dataHeight = yMax - yOrigin;

        dataWidth = Math.max(dataWidth, MIN_SIZE);
        dataWidth = Math.max(dataHeight, MIN_SIZE);
        
        int maxMajor = next10(Math.max(dataWidth, dataHeight)) / 10;
        int maxMinor = maxMajor / 10;
                
        int startX = xOrigin - (xOrigin % maxMajor);
        int startY = yOrigin - (yOrigin % maxMajor);
        
        int endX = xMax - (xMax % maxMajor) + maxMajor;
        int endY = yMax - (yMax % maxMajor) + maxMajor;
        
        // TESTING OVERRIDE - DISABLES SCALING
        startX = 0;
        startY = 0;
        endX   = xMax;
        endY   = yMax;

        endX = Math.max(endX, MIN_SIZE);
        endY = Math.max(endY, MIN_SIZE);
        
        double scale = Math.min((double)visibleWidth / (endX - startX), 
                                (double)visibleHeight / (endY - startY));
        
        for (int x = startX; x <= endX; x += maxMinor) {
            if (x % maxMajor == 0) {
                g2.setColor(X_AXIS_COLOR);
                
                if (maxMajor > 1.5 * AXIS_FONT_SCALING * getStringWidth(x + "") || (x % (2 * maxMajor) == 0)) {
                    if (MATH_AXIS) {
                        g2.drawString(x+"",(int)(leftInset + (x - startX)*scale) - (int)(AXIS_FONT_SCALING*getStringWidth(x+"")/2), height - bottomInset/2);
                    } else {
                        g2.drawString(x+"",(int)(leftInset + (x - startX)*scale) - (int)(AXIS_FONT_SCALING*getStringWidth(x+"")/2), topInset - VERTICAL_ADD / 2);
                    }
                }
            } else {
                g2.setColor(Color.WHITE);
                if (maxMinor < 10) {
                    continue;
                }
            }
            
            if (MATH_AXIS) {
                g2.drawLine((int)(leftInset + (x - startX) * scale), 
                            height - bottomInset + MAJOR_OVERSHOOT, 
                            (int)(leftInset + (x - startX) * scale), 
                            topInset);
            } else {
                g2.drawLine((int)(leftInset + (x - startX) * scale), 
                            height - bottomInset - 1, 
                            (int)(leftInset + (x - startX) * scale), 
                            topInset - MAJOR_OVERSHOOT - 1);     
            }
        }
        
        
        for (int y = startY; y <= endY; y += maxMinor) {
            if (y % maxMajor == 0){
                g2.setColor(Y_AXIS_COLOR);
                
                if (maxMajor > 1.5 * AXIS_FONT_SCALING * getStringHeight() || (y % (2 * maxMajor) == 0)) {
                    if (MATH_AXIS) {
                        g2.drawString(y + "", leftInset - 1 - (int)(AXIS_FONT_SCALING * getStringWidth(y + "")) - MAJOR_OVERSHOOT - 5,getStringHeight()/2 + (int)(height - bottomInset - (y - startY) * scale));
                    } else {
                        g2.drawString(y + "", leftInset - (int)(AXIS_FONT_SCALING * getStringWidth(y + "")) - MAJOR_OVERSHOOT - 5, getStringHeight() / 2 + (int)(topInset + (y - startY) * scale));
                    }
                }
            } else {
                g2.setColor(Color.WHITE);
                if (maxMinor < 10) {
                    continue;
                }
            }

            
            if (MATH_AXIS) {
                g2.drawLine(leftInset - MAJOR_OVERSHOOT -1, 
                            (int)(height - bottomInset - (y - startY) * scale) - 1, 
                            width - rightInset - 1, 
                            (int)(height - bottomInset - (y - startY) * scale) - 1 );
            } else {
                g2.drawLine(leftInset - MAJOR_OVERSHOOT -1, 
                            (int)(topInset + (y - startY) * scale) , 
                            width - rightInset - 1, 
                            (int)(topInset + (y - startY) * scale));
            }
        }
        
        g2.setFont(origFont);
        g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, antialiasOriginal);
    }

    public void paintChildren(Graphics g){
        super.paintChildren(g);
    }
    
    public Dimension getPreferredSize(){
        return layoutSize;
    }

    public VisualType getVisualType(){
        return areaPanel;
    }
    
    public final int getStringWidth(String text){
        return getMetrics().stringWidth(text);
    }
    
    public final int getStringHeight(){
        return getMetrics().getAscent();
    }
    

    class AreaVisualization extends VisualComponent {

        private matrix.structures.spatial.Area rootArea;        
        
        // The structures that hold the elements.
        private ArrayList areas    = new ArrayList();
        private ArrayList points   = new ArrayList();
        // polyline, parabola
        private ArrayList lineElements = new ArrayList();
        // polygon, circle, rectangle, ellipse
        private ArrayList filledElements = new ArrayList();
        // For each element a corresponding selection area 
        // is stored in these structures.
        private ArrayList lineElementSelectionAreas = new ArrayList();
        private ArrayList filledElementSelectionAreas = new ArrayList();

        private Image doubleBuffer;
        // These variables are used to figure out when to allocate a new double buffer
        private int oldWidth = 0, oldHeight = 0;

        private int mouseX, mouseY;
        
        public AreaVisualization(matrix.structures.spatial.Area rootArea) {
            super((FDT) rootArea);
            setHotSpotEnabled(false);
            this.rootArea = rootArea;

            addComponentListener(new ComponentListener() {
                private int oldWidth = -1;
                private int oldHeight = -1;

                public void componentResized(ComponentEvent arg0) {
                    if (doubleBuffer == null || (oldHeight != getHeight() && oldWidth != getWidth())) {
                        doubleBuffer = createImage(getWidth(), getHeight());
                    }
                    oldWidth = getWidth();
                    oldHeight = getHeight();
                }

                public void componentMoved(ComponentEvent arg0) {
                }

                public void componentShown(ComponentEvent arg0) {
                    doubleBuffer = createImage(getWidth(), getHeight());
                }

                public void componentHidden(ComponentEvent arg0) {
                }
            });

            unpack();
        }

        public void unpack(){
            java.util.ArrayList recursedAreas = new java.util.ArrayList();
            recursedAreas.add(rootArea);
            
            points.clear();
            lineElements.clear();
            filledElements.clear();
            lineElementSelectionAreas.clear();
            filledElementSelectionAreas.clear();
            areas.clear();

            Area area;
            Area[] addThese;
            while (!recursedAreas.isEmpty()) {
                area = (matrix.structures.spatial.Area)recursedAreas.remove(0);
                areas.add(area);
                
                addThese = area.getSubAreas();
                for (int i = 0; i < addThese.length; i++) {
                    if (addThese[i] != null) {
                        recursedAreas.add(addThese[i]);
                    }
                }
                
                if (area != null) {
                    SpatialElement[] structures = area.getSpatialElements();
                    java.awt.geom.Area selectionArea;
                    for (int i = 0; i < structures.length; i++) {
                        if (structures[i] instanceof Point) {
                            points.add(structures[i]);
                        } else if ((structures[i] instanceof PolyLine) && (!(structures[i] instanceof Polygon))) {
                            lineElements.add(structures[i]);
                            lineElementSelectionAreas.add(SELECTION_STROKE.createStrokedShape(structures[i].getShape()));
                        } else {
                            // Filled elements have their interior included in the selection area.
                            filledElements.add(structures[i]);
                            selectionArea =  new java.awt.geom.Area(SELECTION_STROKE.createStrokedShape(structures[i].getShape()));
                            selectionArea.add(new java.awt.geom.Area(structures[i].getShape()));
                            filledElementSelectionAreas.add(selectionArea);
                        }
                    }
                }
            }
            
        }
        
        /* (non-Javadoc)
         * @see matrix.visual.VisualType#selectEntity(boolean)
         */
        public void selectEntity(boolean isShiftDown) {
            // Go through the elements in the visualization to find out if the click occured inside one.
            // Points have precedence over "line" elements (e.g. polyline, parabola) and they have 
            // precedence over filled elements (e.g. polygon, circle).
            for (int i = 0; i < points.size(); i++) {
                if (points.get(i) instanceof Selectable) {
                    if (((Point)points.get(i)).getPoint2D().distance(mouseX, mouseY) < POINT_FOCUS_THRESHOLD) {
                        if (((Selectable)points.get(i)).getSelectionTime() == Selectable.NOT_SELECTED) {
                            ((Selectable)points.get(i)).setSelect(Selectable.clock.getNextTimeStamp());
                        } else {
                            // Clear selection if the point already was selected.
                            ((Selectable)points.get(i)).setSelect(Selectable.NOT_SELECTED);
                        }
                        return;
                    }
                }
            }
            for (int i = 0; i < lineElements.size(); i++) {
                if ((lineElements.get(i) instanceof Selectable)) {
                    if (((Shape)lineElementSelectionAreas.get(i)).contains(mouseX, mouseY)) {
                        if (((Selectable)lineElements.get(i)).getSelectionTime() == Selectable.NOT_SELECTED) {
                            ((Selectable)lineElements.get(i)).setSelect(Selectable.clock.getNextTimeStamp());
                        } else {
                            // Clear selection if the element already was selected.
                            ((Selectable)lineElements.get(i)).setSelect(Selectable.NOT_SELECTED);
                        }
                        return;
                    }
                }
            }            
            for (int i = 0; i < filledElements.size(); i++) {
                if ((filledElements.get(i) instanceof Selectable)) {
                    if (((Shape)filledElementSelectionAreas.get(i)).contains(mouseX, mouseY)) {
                        if (((Selectable)filledElements.get(i)).getSelectionTime() == Selectable.NOT_SELECTED) {
                            ((Selectable)filledElements.get(i)).setSelect(Selectable.clock.getNextTimeStamp());
                        } else {
                            // Clear selection if the element already was selected.
                            ((Selectable)filledElements.get(i)).setSelect(Selectable.NOT_SELECTED);
                        }
                        return;
                    }
                }
            }            
        }
        

        /* (non-Javadoc)
         * @see matrix.visual.VisualType#pick(int, int)
         */
        public void pick(int x, int y) {
            if (MATH_AXIS) {
                y = getHeight() - y;
            }
            
            for (int i = 0; i < points.size(); i++) {
                if (((Point)points.get(i)).isDraggable() && 
                    ((Point)points.get(i)).getPoint2D().distance(x,y) < POINT_FOCUS_THRESHOLD) {
                    picked = true;
                    read = getRepresentation((Point)points.get(i));
                    read.setApplication(getApplication());
                    drag = read;
                    pickedType = picker.pick(read, getFrameLocation(), 0, 0);
                    return;
                }
            }
            
            for (int i = 0; i < lineElementSelectionAreas.size(); i++) {
                if ((((SpatialElement)lineElements.get(i)).isDraggable() || 
                     (lineElements.get(i) instanceof Selectable)) && 
                    ((Shape)lineElementSelectionAreas.get(i)).contains(x,y)) {
                    picked = true;
                    read = getRepresentation((SpatialComparable)lineElements.get(i));
                    read.setApplication(getApplication());
                    drag = read;
                    pickedType = picker.pick(read, getFrameLocation(), 0,0);
                    return;
                }
            }
            
            for (int i = 0; i < filledElementSelectionAreas.size(); i++) {
                if ((((SpatialElement)filledElements.get(i)).isDraggable() || 
                     (filledElements.get(i) instanceof Selectable)) && 
                    ((Shape)filledElementSelectionAreas.get(i)).contains(x,y)) {
                    picked = true;
                    read = getRepresentation((SpatialComparable)filledElements.get(i));
                    read.setApplication(getApplication());
                    drag = read;
                    pickedType = picker.pick(read, getFrameLocation(), 0,0);
                    return;
                }
            }

            picked = false;
            read = null;
            drag = null;
            return;
        }
        
        public LayoutSize doActualLayout() {
            Rectangle2D r = rootArea.getBoundingPolygon();
            return new LayoutSize((int)r.getWidth() + 1, (int)r.getHeight() + 1);
        }
                
        public Dimension getPreferredSize() {
            Rectangle2D r = rootArea.getBoundingPolygon();
            return new Dimension((int)r.getWidth() + 1, (int)r.getHeight() + 1);
        }
        
        public Dimension getMinimumSize(){
            return getPreferredSize();
        }        
                
        public void addComponents(){}
        
        public void validateComponents(){
            unpack();
        }
        
        /* (non-Javadoc)
         * @see matrix.visual.VisualType#moved(int, int)
         */
        public void moved(int x, int y){
            mouseX = x;
            mouseY = y;
            if (MATH_AXIS) {
                mouseY = getHeight() - mouseY;
            }
            repaint();
        }
        
        
        public void paintComponent(Graphics gOld) {
            if ((getWidth() != oldWidth) || (getHeight() != oldHeight)) {
                oldWidth = getWidth();
                oldHeight = getHeight();
                doubleBuffer = createImage(oldWidth, oldHeight);
            }

            Graphics2D g = (Graphics2D)doubleBuffer.getGraphics();
            g.setColor(BACKGROUND_COLOR);
            g.fillRect(0, 0, getWidth(), getHeight());
            
            AffineTransform oldTransform = g.getTransform();
            AffineTransform newTransform = (AffineTransform)oldTransform.clone();
            if (MATH_AXIS) {                
                newTransform.concatenate(AffineTransform.getTranslateInstance(0, getHeight() - 1));
                newTransform.concatenate(AffineTransform.getScaleInstance(1, -1));
            }
            g.setTransform(newTransform);
            
            Composite savedComposite = g.getComposite();
            Stroke savedStroke = g.getStroke();
            Object savedAntialias = g.getRenderingHint(RenderingHints.KEY_ANTIALIASING);
            g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
            Font savedFont = g.getFont();

            // The element currently in focus.
            SpatialElement focusElement = null;

            // If the component has focus go through the points and line elements to find
            // if the mouse is hovering over one of them. Filled elements are handled when
            // they are drawn. These other elements are checked here to give them precedence
            // in selection over the filled elements.
            if (hasFocus()) {
                Point p;
                for (int i = 0; i < points.size(); i++) {
                    p = (Point)points.get(i);
                    if ((p.isDraggable() || (p instanceof Selectable)) &&
                        (p.getPoint2D().distance(mouseX, mouseY) < POINT_FOCUS_THRESHOLD) ) {
                        focusElement = p;
                        break;
                    }
                }
                if (focusElement == null) {
                    SpatialElement se;
                    for (int i = 0; i < lineElements.size(); i++) {
                        se = (SpatialElement)lineElements.get(i);
                        if ((se.isDraggable() || (se instanceof Selectable)) && 
                            ((Shape)lineElementSelectionAreas.get(i)).contains(mouseX, mouseY)) {
                            focusElement = se;
                            break;
                        }
                    }
                }
            }

            // Areas to be filled are separated according to their color and transparency to
            // buckets that are filled separately. The buckets are stored in a Hashmap,
            // which uses a string constructed from the color and transparency to recognize
            // the correct bucket.
            //
            // This approach avoids transparency problems with overlapping areas of the 
            // same color.

            // Fill hashmap with the drawing instructions.
            HashMap fillAreas = new HashMap(); 
            Area area;
            PaintingStyleDecorator psd;
            java.awt.geom.Area fillArea;
            String key;
            for (int i = 0; i < areas.size(); i++) {
                area = (Area)areas.get(i);
                psd = area.hasPaintingStyleDecorator() ? area.getPaintingStyleDecorator() : DEFAULT_AREA_DECORATOR;
                if (psd.hasFill()) {
                    key = psd.getFillColor().getRGB() + ",";
                    key += psd.hasTransparency() ? (psd.getTransparency() + "") : "";
                    
                    if (!fillAreas.containsKey(key)) {
                        fillArea = new java.awt.geom.Area();
                        fillAreas.put(key, fillArea);
                    } else {
                        fillArea = (java.awt.geom.Area)fillAreas.get(key);
                    }
                    
                    fillArea.add(new java.awt.geom.Area(area.getBoundingPolygon()));
                }
            }

            // Fill the areas stored in the hashmap.
            Iterator iterator = fillAreas.entrySet().iterator();
            Map.Entry fillAreaEntry;
            int separator;
            String color, transparency, desc;
            while (iterator.hasNext()) {
                fillAreaEntry = (Map.Entry)iterator.next();
                desc = (String)fillAreaEntry.getKey();
                separator = desc.indexOf(",");
                color = desc.substring(0, separator);
                transparency = desc.substring(separator + 1, desc.length());
                
                if (transparency.trim().length() > 0) {
                    g.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 
                                   Float.parseFloat(transparency)));
                } else {
                    g.setComposite(savedComposite);
                }
                
                if (color.trim().length() > 0) {
                    g.setColor(Color.decode(color));
                    g.fill((java.awt.Shape)fillAreaEntry.getValue());
                } 
            }
            g.setComposite(savedComposite);            

            // Draw the outlines for the areas.
            area = null;
            psd = null;
            for (int i = 0; i < areas.size(); i++) {
                area = (Area)areas.get(i);
                psd = area.hasPaintingStyleDecorator() ? area.getPaintingStyleDecorator() : DEFAULT_AREA_DECORATOR;
                if (psd.hasOutline()) {
                    g.setColor(psd.getLineColor());
                    g.setStroke(psd.getStroke());
                    g.draw(area.getBoundingPolygon());
                    
                }
            }
            
            // Draw area labels.
            area = null;
            psd = null;
            Font labelFont = savedFont.deriveFont(AffineTransform.getScaleInstance(0.6, 0.6));
            for (int i = 0; i < areas.size(); i++) {
                area = (Area) areas.get(i);

                if (((area instanceof LabelDecorator) && 
                     ((LabelDecorator)area).isLabelEnabled())) {
    
                    psd = area.hasPaintingStyleDecorator() ? area.getPaintingStyleDecorator() : DEFAULT_AREA_DECORATOR;
                    g.setColor(psd.getLineColor());
    
                    g.setFont(labelFont);
                    String idString = ((LabelDecorator)area).getLabel();
    
                    java.awt.Point topCorner = 
                        new java.awt.Point((int)area.getBoundingPolygon().getX(), 
                                           (int)area.getBoundingPolygon().getY());
                    if (MATH_AXIS) {
                        g.setTransform(oldTransform);
                        g.drawString(idString, 
                                     (int)topCorner.getX() + 3,
                                     getHeight() - (int)topCorner.getY() - 3);
                        g.setTransform(newTransform);
                    } else {
                        g.drawString(idString, 
                                     (int)topCorner.getX() + 3,
                                     (int)topCorner.getY());
                    }
                }
            }
            
            // Fill the hashmap with the drawing instruction for the filled spatial elements fill.
            fillAreas = new HashMap();
            fillArea = null;
            key = null;
            psd = null;
            SpatialElement se;
            for (int i = 0; i < filledElements.size(); i++) {
                se = (SpatialElement)filledElements.get(i);
                psd = se.hasPaintingStyleDecorator() ? se.getPaintingStyleDecorator() : DEFAULT_ELEMENT_DECORATOR;
                
                if (psd.hasFill()) {
                    if ((se instanceof Selectable) && 
                        (((Selectable)se).getSelectionTime() != Selectable.NOT_SELECTED)) {
                        if (focusElement == se) {
                            key = PaintingStyleDecorator.mixColors(psd.getSelectionFillColor(), 
                                                                   psd.getFocusFillColor(), 
                                                                   .5f).getRGB() + ",";
                        } else {
                            key = psd.getSelectionFillColor().getRGB() + ",";
                        }
                    } else if (hasFocus() && 
                               (focusElement == null) &&
                               (se.isDraggable() || (se instanceof Selectable)) &&
                               ((Shape)filledElementSelectionAreas.get(i)).contains(mouseX, mouseY)) {
                               focusElement = se;
                        key = psd.getFocusFillColor().getRGB() + ",";
                    } else {
                        key = psd.getFillColor().getRGB() + ",";
                    }
                    key += psd.hasTransparency() ? (psd.getTransparency() + "") : "";

                    if (!fillAreas.containsKey(key)) {
                        fillArea = new java.awt.geom.Area();
                        fillAreas.put(key, fillArea);
                    } else {
                        fillArea = (java.awt.geom.Area)fillAreas.get(key);
                    }
                    fillArea.add(new java.awt.geom.Area(se.getShape()));
                }
            }

            // Fill the spatial elements according to the constructed hashmap.
            iterator = fillAreas.entrySet().iterator();
            fillAreaEntry = null;
            color = null;
            transparency = null;
            desc = null;
            separator = -1;
            while (iterator.hasNext()) {
                fillAreaEntry = (Map.Entry)iterator.next();
                desc = (String)fillAreaEntry.getKey();
                separator = desc.indexOf(",");
                color = desc.substring(0, separator);
                transparency = desc.substring(separator + 1, desc.length());
                
                if (transparency.trim().length() > 0){
                    g.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 
                                   Float.parseFloat(transparency)));
                } else {
                    g.setComposite(savedComposite);
                }                

                if (color.trim().length() > 0){
                    g.setColor(Color.decode(color));
                    g.fill((java.awt.Shape)fillAreaEntry.getValue());
                }   
            }
            g.setComposite(savedComposite);            
            
            // Draw the outlines for filled spatial elements.
            se = null;
            psd = null;
            for (int i = 0; i < filledElements.size(); i++) {
                se = (SpatialElement)filledElements.get(i);
                psd = se.hasPaintingStyleDecorator() ? se.getPaintingStyleDecorator() : DEFAULT_ELEMENT_DECORATOR;
                // Do not draw the outlines if the stroke has been explicitly disabled.
                if (psd.hasOutline()) {
                    g.setStroke(psd.getStroke());
                    
                    if ((se instanceof Selectable) && 
                        (((Selectable)se).getSelectionTime() != Selectable.NOT_SELECTED)) {
                        if (focusElement == se) {
                            g.setColor(psd.getSelectionFocusLineColor());
                        } else {
                            g.setColor(psd.getSelectionLineColor());
                        }
                    } else if (focusElement == se) {
                        g.setColor(psd.getFocusLineColor());
                    } else {
                        g.setColor(psd.getLineColor());
                    }
                    g.draw(se.getShape());
                }
            }

            // Draw the line elements, e.g. polyline.
            se = null;
            for (int i = 0; i < lineElements.size(); i++) {
                se = (SpatialElement)lineElements.get(i);
                psd = se.hasPaintingStyleDecorator() ? se.getPaintingStyleDecorator() : DEFAULT_ELEMENT_DECORATOR;
                // Do not draw if the stroke has been explicitly disabled.
                if (psd.hasOutline()) {
                    g.setStroke(psd.getStroke());
                    
                    if ((se instanceof Selectable) && 
                        (((Selectable)se).getSelectionTime() != Selectable.NOT_SELECTED)) {
                        if (focusElement == se) {
                            g.setColor(psd.getSelectionFocusLineColor());
                        } else {
                            g.setColor(psd.getSelectionLineColor());
                        }
                    } else if (focusElement == se) {
                        g.setColor(psd.getFocusLineColor());
                    } else {
                        g.setColor(psd.getLineColor());
                    }
                    g.draw(se.getShape());
                }
            }
            
            // Draw the points and their labels.
            psd = null;
            Point p;
            labelFont = savedFont.deriveFont(AffineTransform.getScaleInstance(POINT_LABEL_FONT_SCALING, POINT_LABEL_FONT_SCALING));
            for (int i = 0; i < points.size(); i++) {
                p = (Point)points.get(i);                
                psd = p.hasPaintingStyleDecorator() ? p.getPaintingStyleDecorator() : DEFAULT_POINT_DECORATOR;
                
                if (psd.hasOutline()) {
                    g.setStroke(psd.getStroke());
                    g.setColor(psd.getLineColor());
                    
                    if ((p instanceof Selectable) && 
                        (((Selectable)p).getSelectionTime() != Selectable.NOT_SELECTED)) {
                        if (focusElement == p) {
                            g.setColor(psd.getSelectionFocusLineColor());
                        } else {
                            g.setColor(psd.getSelectionLineColor());
                        }
                    } else if (focusElement == p) {
                        g.setColor(psd.getFocusLineColor());
                    } else {
                        g.setColor(psd.getLineColor());
                    }

                    // Draw the labels for points.
                    g.setFont(labelFont);
                    if (MATH_AXIS) {
                        g.setTransform(oldTransform);
                        g.drawString((p.getField(0)).toString(),
                                     (int)p.getX() + 7,
                                     getHeight() - (int)p.getY() + getStringHeight() / 2 - 2);                    
                        g.setTransform(newTransform);
                    } else {
                        g.drawString((p.getField(0)).toString(),
                                     (int)p.getX() + 7,
                                     (int)p.getY() - getStringHeight() / 2 + 2);
                    }

                    g.fillOval((int)p.getX() - POINT_SIZE, 
                               (int)p.getY() - POINT_SIZE, 
                               POINT_SIZE * 2, 
                               POINT_SIZE * 2);
                }
            }
            
            // Restore original settings.
            g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, savedAntialias);
            g.setTransform(oldTransform);
            g.setStroke(savedStroke);
            g.setFont(savedFont);
            
            // Paint the double buffer on the real buffer.
            gOld.drawImage(doubleBuffer,0,0, null);
        }

        
        // @Override
        public VisualType getVisualType() {
            return this;
        }

        // @Override
        public FDT getKey() {
            // TODO Auto-generated method stub
            return null;
        }

        // @Override
        public void assign() {
            VisualArea.this.insert(read.getStructure());
        }

        public void insert(Object o) {
            VisualArea.this.insert(o);
        }
    }
    
	// A hack to make SpatialElementTable work. 
    // If possible, should be replaced with a better way to do it. - 16.6.06 jtn
	public void setType(int i) {
		if (getStructure() instanceof SpatialElementTable) {
			((SpatialElementTable)getStructure()).setType(i);
			setInvalid();
		}
	}
}
