/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: VisualArrayComponent.java,v 1.47 2006/12/07 10:26:51 oseppala Exp $ */

package matrix.visual;

import java.awt.*;
import java.io.PrintStream;

import matrix.decoration.*;
import matrix.structures.CDT.CDT;
import matrix.structures.FDT.FDT;
import matrix.structures.FDT.probe.Table;

public class VisualArrayComponent extends VisualComponent {
    protected int place;
    protected VisualType type;
    private Object component;
    private boolean rotated=false;

    /**
     * Choose whether the array component has the label below or to the left
     * of the value.
     * @param r <code>false</code> to place the label below, <code>true</code>
     * to place it to the left.
     */
    public void setRotated(boolean r) {
        rotated=r;
    }

    public void dump(PrintStream out) {
        super.dump(out);
        out.println("[[[indexed="+isIndexed()+
                "; component = " + component+ "]]]");
    }

    /** This constructor creates a new VisualArrayComponent that holds
	the specified Object. The object cannot be changed during the
	lifetime of this component, thus a new component is created if
	the content of this index has changed. The creation of new
	components is handled in VisualArray. 
     */
    VisualArrayComponent(int place, Object o) {
        super(null);
        this.place = place;
        this.component = o;
    }

    protected VisualArrayComponent(int place, Object o, boolean rotated) {
        this(place, o);
        this.rotated=rotated;
    }

    public void addComponents() {
        if (getComponentCount() == 0) {
            add(type=getRepresentation(component));
        }
    }

    /** This method does nothing. VisualArrayComponents are validated
	elsewhere. Actually we implement this methos just for the case we
	do not have to invoke super.validateComponents() which has no use
	in here. 
     */
    public void validateComponents() {    }

    public void setIndex(int i) { 
        place = i; 
    }

    public Dimension getPreferredSize() {
        Dimension d=super.getPreferredSize();
        if (rotated) {
            if (isIndexed())
                if(d.height<getMetrics().getHeight()+2)
                    d.height=getMetrics().getHeight()+2;
        } else {
            // This will automatically set the Dimension d to the size of the object within this place
            // However, if the object above is smaller than the space for index, we should adjust the size
            if (isIndexed()) {
                Insets in = getInsets();

                int w = getMetrics().stringWidth(getMaxIndex()) + in.left + in.right;
                int h = getMetrics().getHeight() + in.top + in.bottom;
                if(d.width < w) // true if index is wider than the object
                    d.width = w;
                if(d.height < h) // true if index (font size) is higher than the object
                    d.height = h;
            }
        }
        return d;
    }


    public boolean isIndexed() {
        return getComponentParent().isIndexed() && !getComponentParent().isDuplicate();
    }

    public Dimension getMinimumSize() {
        Dimension d = new Dimension(1,1);
        if (getFont() != null) {
            d.width = getMetrics().stringWidth(getName());
            d.height += getMetrics().getHeight();
        }
        else return super.getMinimumSize();
        return d;
    }

    public Insets getInsets() {
        if (!isIndexed()) 
            return new Insets(1,1,1,1);
        else if (getFont() != null) {
            if (rotated)
                return new Insets(2,getMetrics().
                        stringWidth(getName())+2,
                        1,1);
            else
                return new Insets(2,2,
                        getMetrics().getHeight()+2,3);
        }
        else
            return super.getInsets();
    }

    public VisualType getVisualType() { 
        return getComponentCount() > 0 ? (VisualType)getComponent(0) : null;
    }

    public VisualArray getComponentParent() { 
        return (VisualArray)getParent(); 
    }

    public void write() { 
        getComponentParent().write(place); 
    }

    /** Note: This method (in a way) uses the RepresentationFactory to
	determine the corresponding FDT type for the component
	object. Thus, the VisualType for the component should be set
	in order to be able to determine the key structure of this
	VisualArrayComponent.  */
    public FDT getStructure() { 
        if (type != null)
            return type.getStructure(); 
        else
            return null; 
    }           

    public FDT getKey() {
        return getStructure(); 
    }

    public void assign() { 
        write(); 
    }

    public void erase() {
        getComponentParent().erase(component);
        setInvalid();
    }

    public String getLabel() {
        return component.toString();
    }

    public String getName() {
        //return labeled ? getLabel() : getComponentParent().getIndexName(place);
        return getComponentParent().getIndexName(place);
    }

    public String getMaxIndex() {
        return getComponentParent().getMaxIndexName();
    }


    private StyleSheet defaultStyleSheet = null;
    public StyleSheet getDefaultStyleSheet() {
        if (defaultStyleSheet == null) {
            defaultStyleSheet = new StyleSheetAdapter();
            defaultStyleSheet.setDefaultPenColor(Color.white);
            defaultStyleSheet.setFocusedPenColor(Color.red.brighter());
        }
        return defaultStyleSheet;
    }


    public void paintComponent(Graphics g) {
        super.paintComponent(g);
        Rectangle r = getBounds();
        Insets in = getInsets();
        int x = 1;
        int y = 0;
        int w = r.width-3;
        int h = r.height-2;
        int borderY = y+h-in.bottom;

        setBorderColor(g);

        if(((VisualType)getParent()).getStructure() instanceof CDT)
            g.setColor(((VisualContainer)getParent()).getCDTColor());

        g.drawRect(x, y, w, h);
        if (isIndexed()) {
            String name = getName();
            int iWidth = getMetrics().stringWidth(name);
            int iHeight = getMetrics().getHeight();
            int yGap = (in.bottom-iHeight) > 0 ? (in.bottom - iHeight) / 2 : 0;
            if (rotated) {
                g.fillRect(x, y, in.left, h); 
                setPenColor(g); // index
                g.drawString(name, x, h-yGap);
            } else {
                g.fillRect(x, borderY, w+1, in.bottom+1); 
                setPenColor(g); // index
                g.drawString(name, w/2 - iWidth/2 + 2, h - yGap);
            }
        }
    }     

    public boolean hasFocus(){
        if (getComponentParent().getStructure() instanceof Table) {
            Table  t = (Table)getComponentParent().getStructure();
            if (!t.isSelection())
                return super.hasFocus();
            else {
                if (read == this)
                    return true;
                return super.hasFocus() ^ (getIndex() >= t.selectionStart() 
                        && getIndex()<=t.selectionEnd()); 
            }
        }
        return super.hasFocus();
    }


    //Called by visualType.click
    public void selectEntity(boolean isShiftDown) {
        if (getComponentParent().getStructure() instanceof Table) {
            Table  t = (Table)getComponentParent().getStructure();
            t.selectIndex(place);
            if (!t.isSelection())
                read = null; // clear also currently read object
        }
        validate();
    }

    public String toString() {
        return super.toString() + " at index " + place;
    }


    public int getIndex() {
        return place;
    }
}
