/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: VisualComponent.java,v 1.62 2006/12/11 13:46:34 oseppala Exp $ */

package matrix.visual;

import java.awt.*;
import java.io.PrintStream;

import matrix.decoration.*;
import matrix.structures.FDT.FDT;
import matrix.structures.FDT.probe.Key;

/** A VisualComponent is the abstract superclass for all (non-reference)
 ** components a VisualContainer may include.
 * In addition, two VisualComponents (items)
 * can be connected by a VisualReference (link) in a VisualContainer.
 * @see VisualContainer
 *
 * @author Ari Korhonen
 * @version $Revision: 1.62 $
 */
public abstract class VisualComponent extends VisualType {

    /* Current color of this VisualType */
    protected Color color;

    protected boolean running = false;

    /** Debug method.
     * Subclasses should invoke super.debug() in order
     * to provide all debug information. Every subclass has its own
     * responsibility to provide debug info about its instance variables.
     * The debug info should be printed out with Note.out() utility.
     */
    public void dump(PrintStream out) {
        super.dump(out);
        boolean ISCEFlag;
        ISCEFlag = ((VisualType)getParent()).isSubComponentsEnabled();
        out.println("[[parent.subcomponents enabled="+ISCEFlag+
                 "; parent = " + getParent() +
                 "; color="+color+"]]");
    }

    // --- constructors ---

    VisualComponent(FDT s) {
        super(s);
        this.labeled = false;
    }

    // --- end of constructors ---

    public void addComponents() {
        if (getParent() instanceof VisualContainer)
            this.labeled = ((VisualContainer)getParent()).isLabeled();
    }
    /** Return true if the specified parameters equals. */
    protected boolean isSame(FDT o1, Object o2) {
        if (o1 == null) return (o2 == null);
        if (o2 == null)
	    return ((o1 instanceof Key) && (((Key)o1).equals(Key.Empty)));
        if (o1 == null)
	    return ((o2 instanceof Key) && (((Key)o2).equals(Key.Empty)));
        return ((o1.equals(o2)) || (o2.equals(o1)));

    }

    /** Return a representation for the specified FDT. */
    public VisualType getRepresentation(Object fdt) {
        if (getVisualContainer() != null)
            return getVisualContainer().getRepresentation(fdt);
        else
            return RepresentationFactory.getRepresentation(fdt);
    }

    public void enableSubComponents(boolean val) {
        super.enableSubComponents(val);
        // if modified to prevent endless loops in case getVisualType() returns this
        if (getVisualType() != null && getVisualType() != this)
            getVisualType().enableSubComponents(val);
    }

    /** Toggles whether the label of this component is visible. */
    public void Labeled() {
        labeled = !labeled;
        setInvalid();
    }

    /** Is the label of this component visible?
     * @return true if the label of this component is shown beside the
     * component; false otherwise.
     */
    public boolean isLabeled() {
	return labeled;
    }

    /** Sets this component's labeled flag. */
    public void setLabeled(boolean value) {
        labeled = value;
    }


    /** Determines the visual representation of the key value of the component.
     * @return the VisualType of the key value nested in this component.
     */
    public abstract VisualType getVisualType();

    /** Determines the insets of this component,
     ** which indicate the size of the component's border.
     * <TT>
     * return new Insets(getMetrics().getHeight(),1,1,1);
     * </TT>
     */
    /*
    public Insets getInsets() {
        if (getFont() != null)
            return new Insets(getMetrics().getHeight(),1,1,1);
        else
            return super.getInsets();
    }
    */

    /** Show or hide this component and the nested representation.
     * */
    // Behavior changed when getVisualType returns this (default behavior in visualType
    // originally resulted in endless recursion.
    public void setVisible(boolean b) {
        super.setVisible(b);
        if (getVisualType() != null && getVisualType() != this)
            getVisualType().setVisible(b);
        //setInvalid(); <--- do not put this here! (will cause an error)
    }

    /** Returns the preferred size of this component.
     * By default the preferred size of this component
     * consists of the insets and the size of the nested representation.
     */
    public Dimension getPreferredSize() {
        Dimension dd = super.getPreferredSize();
        Insets in = getInsets();
        VisualType vt = getVisualType();
        Dimension d = vt != null ? vt.getPreferredSize() : getMinimumSize();
        width = (d.width += (in.left + in.right));
        height = (d.height += (in.top + in.bottom));
        if (dd == null)
            return new Dimension(width, height);
        else { // just want to prevent creating new objects all the time
            if ((dd.width == width) && (dd.height == height))
                return dd;
            else 
                return new Dimension(width, height);
        }
    }

    /** Lay out a single key within this component border.
     * First, super doLayout() is invoked.
     * Second, this method recalculates the width and height of this
     * component depending on the size of the nested structure.
     * Finally, the correct positioning and size of the nested structure
     * is calculated and the bounds of the nested strucures are set.
     * @see VisualType#doLayout()
     */
    public void doLayout() {
        super.doLayout();
        getPreferredSize(); // set width and height
        VisualType vt;
        if ((vt = getVisualType()) != null) {
        	if(vt != this) {
        		vt.doLayout();
        	}
            Dimension d = vt.getPreferredSize();
            Insets in = getInsets();
            int x = in.left + (width - in.left - in.right - d.width)/2;
            int y = in.top + (height - in.top - in.bottom - d.height)/2;
            vt.setBounds(x, y-1, d.width, d.height);
        }
    }

    public void validateComponents() {
        VisualType vt = (getComponentCount() > 0) ? getVisualType() : null;
        FDT oldType = (vt != null) ? vt.getStructure() : null;
        Object currType = getKey();

        if (vt == null) { // create new visualElement
            add(getRepresentation(currType));
	    return;
        }
        else if (!sameStructure(oldType, currType)) {
            remove(0); // remove old type and
	    // add new type
            add(getRepresentation(currType));
	    return;
        }
    }

    private boolean sameStructure(FDT o1, Object o2) {
        if (o1 == null) return (o2 == null);
        if (o2 == null)
	    return ((o1 instanceof Key) && (((Key)o1).equals(Key.Empty)));
        if (o1 == null)
	    return ((o2 instanceof Key) && (((Key)o2).equals(Key.Empty)));
	//Note.out(this, "eq " + ":" + o1 + ": == :" + o2 + ":" + o1.equals(o2));
        //otherwise objects need to be same
        return o1 == o2;
    }

    /** Return the key attached to this component.
     * @return FDT - the Fundamental Data Type this visual
     * object is representing.
     * @see VisualType#getStructure()
     */
    public abstract FDT getKey();

    /** Gets the component insets of this component.
     * Override this if you need to draw a label outside the component.
     * @see #getComponentBounds()
     */
    protected Insets getComponentInsets() {
        return new Insets(0, 0, 0, 0);
    }

    // saved for efficiency
    private Rectangle cb = new Rectangle(0,0,0,0);
    /** Returns the bounds of the inside of this component.
     * The inside is the area inside the component insets, and
     * represents the area where the component itself is drawn
     * (excluding the possible label outside it). The return value is
     * in the same coordinate system as getBounds.
     *
     * @return immutable rectangle containing component bounds
     */
    public Rectangle getComponentBounds() {
        Rectangle b = getBounds();
        Insets ci = getComponentInsets();
        cb.x = b.x + ci.left;
        cb.y = b.y + ci.top;
        cb.width = b.width - ci.left - ci.right;
        cb.height = b.height - ci.top - ci.bottom;
        return cb;
    }

    /** Returns the center point of the "component bounds". */
    public Point getCenter() {
        Rectangle r = getComponentBounds();
        return new Point(r.x + r.width/2, r.y + r.height/2);
    }

    // saved for efficiency
    private Rectangle cbfp = new Rectangle(0,0,0,0);
    /** As getComponentBounds, but for the paint method.
     * The return value is in the "local" coordinate system used by
     * paint methods (origin at upper left corner of bounds).
     * @return immutable Rectangle containing the bounds
     * @see #getComponentBounds()
     */
    public Rectangle getComponentBoundsForPaint() {
        Rectangle b = getBounds();
        Insets ci = getComponentInsets();
        cbfp.x = ci.left;
        cbfp.y = ci.top;
        cbfp.width = b.width - ci.left - ci.right;
        cbfp.height = b.height - ci.top - ci.bottom;
        return cbfp;
    }

    /**
     * Calculates the coordinates of the border of this component at a
     * specified angle (from the center point returned by getCenter).
     * Searches for the border by using the contains method and
     * getComponentBounds; subclasses may override this if necessary.
     *
     * @param angle - angle in radians; 0=to the right, positive values
     *              go counterclockwise, normalized to the range [0..2pi)
     * @return coordinates of the border point
     * @see #getCenter()
     */
    public Point getBorderAtAngle(double angle) {
        Rectangle b = getComponentBounds();
        Point c = getCenter();
        double x, y;
        double urCornerA = Math.atan2(b.height, b.width);
        double ulCornerA = Math.PI - urCornerA;
        double lrCornerA = 2*Math.PI - urCornerA;
        double llCornerA = urCornerA + Math.PI;

        // Find a starting point on the getComponentBounds rectangle
        if (angle < urCornerA || angle > lrCornerA) { // on right side
            x = b.x + b.width;
            y = c.y - (b.width/2.0) * Math.tan(angle);
        } else if (angle > ulCornerA && angle < llCornerA) { // left
            x = b.x;
            y = c.y + (b.width/2.0) * Math.tan(angle - Math.PI);
        } else if (angle <= ulCornerA) { // top
            x = c.x + (b.height/2.0) / Math.tan(angle);
            y = b.y;
        } else { // on bottom side
            x = c.x - (b.height/2.0) / Math.tan(angle - Math.PI);
            y = b.y + b.height;
        }

        if (!contains((int)x, (int)y)) {
            // Binary search
            double dist = Math.sqrt((x-c.x)*(x-c.x) + (y-c.y)*(y-c.y));
            double distInc = dist/2.0;
            do {
                if (contains((int)x, (int)y))
                    dist += distInc;
                else dist -= distInc;

                distInc /= 2.0;
                x = c.x + Math.cos(angle) * dist;
                y = c.y - Math.sin(angle) * dist;
            } while (Math.abs(distInc) > 1.0);
        }
        return new Point((int) x, (int) y);
    }

    private StyleSheet defaultStyleSheet = null;
    public StyleSheet getDefaultStyleSheet() {
	if (defaultStyleSheet == null) {
            defaultStyleSheet = new StyleSheetAdapter();
	    defaultStyleSheet.setDefaultPenColor(Color.blue);
	    defaultStyleSheet.setFocusedPenColor(Color.blue);
	}
        return defaultStyleSheet;
    }

    /*
     * For transforming to/from layout coordinates: ask our container
     * to do the transformations according to its rotation etc.
     * settings
     */

    public LayoutSize lGetPreferredSize() {
        VisualContainer cont = getVisualContainer();
        if (cont == null) { // no transformations if no container
            Dimension d = getPreferredSize();
            return new LayoutSize(d.width, d.height);
        } else return cont.dimensionToLayoutSize(getPreferredSize());
    }

    public LayoutBounds lGetBounds() {
        VisualContainer cont = getVisualContainer();
        if (cont == null) { // no transformations if no container
            Rectangle r = getBounds();
            return new LayoutBounds(r.x, r.y, r.width, r.height);
        } else return cont.rectangleToLayoutBounds(getBounds());
    }

    public void lSetBounds(LayoutBounds lb) {
        VisualContainer cont = getVisualContainer();
        if (cont == null) { // no transformations if no container
            setBounds(lb.x, lb.y, lb.width, lb.height);
        } else setBounds(cont.layoutBoundsToRectangle(lb));
    }

    public void lSetBounds(int x, int y, int width, int height) {
        lSetBounds(new LayoutBounds(x, y, width, height));
    }

}
