/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: VisualContainer.java,v 1.113 2007/09/05 16:53:52 oseppala Exp $ */

package matrix.visual;

import java.awt.*;
import javax.swing.*;
import java.awt.event.*;
import java.io.PrintStream;
import java.util.WeakHashMap;

import matrix.simulation.VisualTypeConf;
import matrix.decoration.*;
import matrix.structures.CDT.CDT;
import matrix.structures.FDT.*;
import matrix.structures.util.Iterator;
import matrix.util.*;
import matrix.util.export.ConfigurableGraphics;

/** A VisualContainer is the abstract superclass for all containers
 ** holding components and references.
 * VisualContainer implements all the default functionality shared
 * among containers.
 *
 * @author Ari Korhonen
 * @version $Revision: 1.113 $
 */
public abstract class VisualContainer extends VisualType implements ActionListener, ItemListener {
    private Color color = null;
    
    /** Flag indicating if the container title is enabled */
    private boolean titled = true; // default: container show its title bar
    private boolean rotated = false, flippedX = false, flippedY = false;
    private VisualMenuHotSpot visualMenuHotSpot = null;
    private VisualNullHotSpot visualNullHotSpot = null;
    private VisualSearchHotSpot visualSearchHotSpot = null;
    protected VisualResizeHotSpot visualResizeHotSpot = null;
    private boolean hotSpotEnabled = true;
    private boolean nullHotSpotEnabled = true;
    private boolean searchHotSpotEnabled = true;
    
    //A map for resolving the desired locations of empty nodes in a tructure
    private WeakHashMap emptyNeighborMap = new WeakHashMap();
    
    
    
    private JMenu representationMenu;
    
    public void dump(PrintStream out) {
        super.dump(out);
        out.println("items="+getItemCount()+"; #links="+getLinkCount()+"; #others="+getOtherCount());
        for (int i=0; i<getItemCount(); i++)
            out.println(" "+i+". "+getItem(i)+"; visible="+getItem(i).isShowing());
        for (int i=0; i<getLinkCount(); i++)
            out.println(" "+i+". "+getLink(i)+"; visible="+getLink(i).isShowing());
        for (int i=0; i<getOtherCount(); i++)
            out.println(" "+i+". "+getOther(i)+"; visible="+getOther(i).isShowing());
        out.println("titled="+titled+"; rotated="+rotated+"; flipX="+flippedX+"; flipY="+flippedY);
    }
    
    /** Construct the container of the specified fundamental data type. */
    public VisualContainer(FDT s) {
        super(s);
        this.enableEvents(AWTEvent.MOUSE_EVENT_MASK);
        setTitled(!isNestedContainer());
        
        if (s != null)
            setName(getStructureName());
        //setName(getRepresentationName() + " Representation of " + getStructureName());
    }
    
    /** Default constructor: VisualContainer(null) */
    public VisualContainer() {
        this(null);
    }
    
    // --- end of constructors ---
    
    public void setHotSpotEnabled(boolean value) {
        hotSpotEnabled = value;
        if (visualMenuHotSpot != null)
            visualMenuHotSpot.setVisible(value);
    }
    
    public boolean isHotSpotEnabled() {
        return hotSpotEnabled;
    }
    
    
    /* [debugger code commented out]
     public void setBreakPoint(String methodName) {
     Note.out(this, "Call Debugger for " + methodName);
     Class clazz = getStructure().getClass();
     String className = clazz.getName();
     DebuggerClient d = DebuggerClient.getInstance();
     if (d != null) {
     d.createDebugWindow(-1, className, methodName);
     } else {
     Note.err(this, "Cannot get debugger.");
     }
     }
     */
    
    /** Return the representation name for this container.
     * By default, the class name is returned.
     * @return the name of this representation.
     */
    public String getRepresentationName() {
        return getClass().toString();
    }
    
    /** By default, this method adds visual null hot spot in
     * this container. Subclasses should override this method to do nothing if
     * the default behaviour is not wanted.
     */
    public void addNullHotSpot() {
        add(visualNullHotSpot=new VisualNullHotSpot(getStructure()));
        visualNullHotSpot.setVisible(isTitled() && nullHotSpotEnabled);
    }
    
    /** By default, this method adds visual search hot spot in
     * this container. Subclasses should override this method to do nothing if
     * the default behaviour is not wanted.
     */
    public void addSearchHotSpot() {
        add(visualSearchHotSpot=new VisualSearchHotSpot(getStructure()));
        visualSearchHotSpot.setVisible(isTitled() && searchHotSpotEnabled);
    }
    
    /** 
     * Sets the null hot spot enabled or disabled depending on the
     * given value. If null hot spot is not added, does nothing.
     */
    public void setNullHotSpotEnabled(boolean value) {
        nullHotSpotEnabled = value;
        if (visualNullHotSpot != null)
            visualNullHotSpot.setVisible(value);
    }
    
    /** Returns true, if null hot spot is added and enabled, false
     * otherwise.*/
    public boolean isNullHotSpotEnabled() {
        if (visualNullHotSpot == null)
            return false;
        return visualNullHotSpot.isVisible();
    }
    
    /** 
     * Sets the search hot spot enabled or disabled depending on the
     * given value. If search hot spot is not added, does nothing.
     */
    public void setSearchHotSpotEnabled(boolean value) {
        searchHotSpotEnabled = value;
        if (visualSearchHotSpot != null)
            visualSearchHotSpot.setVisible(value);
    }
    
    /** Returns true, if search hot spot is added and enabled, false
     * otherwise.*/
    public boolean isSearchHotSpotEnabled() {
        if (visualSearchHotSpot == null)
            return false;
        return visualSearchHotSpot.isVisible();
    }
    
    public void addComponents() {
        super.addComponents();
        setTitled(!isNestedContainer());
        add(visualMenuHotSpot=new VisualMenuHotSpot(getStructure()));
        visualMenuHotSpot.setVisible(isTitled() && hotSpotEnabled);
        addNullHotSpot();
        if (getStructure() instanceof CDT){
            addSearchHotSpot();
        }
    }
    
    public void validateComponents() {
        super.validateComponents();
        if (visualMenuHotSpot != null)
            visualMenuHotSpot.validate();
        if (visualNullHotSpot != null)
            visualNullHotSpot.validate();
        if (visualSearchHotSpot != null)
            visualSearchHotSpot.validate();
    }
    
    protected void removePopUpMenu() {
        // do nothing because visual hot spot employs the same popUp objects and thus, cannot be removed
    }
    
    public JMenu createRepresentationMenu() {
        JMenu repr = new JMenu("Change layout");
        String[] reprNames =
            RepresentationFactory.getRepresentationNames(getStructure());
        for (int j=0; j<reprNames.length; j++) {
            JCheckBoxMenuItem mi = new JCheckBoxMenuItem(reprNames[j]);
            // next clause is always true at the beginning before any changes in 
            // representation has occured
            if (j == 0 && getNameOfRepresentation() == "") {
                //default representation is the first one in the menu
                mi.setState(true);
            } else {
                mi.setState((getNameOfRepresentation()).equals(reprNames[j]));
            }
            mi.addItemListener(this);
            mi.setActionCommand(reprNames[j]);
            repr.add(mi);
        }
        return repr;
    }
    
    /** This method is invoked by CheckboxMenuItems created by
     createRepresentationMenu in order to change the
     representation. */
    public void itemStateChanged(ItemEvent event) {
        if (event.getItem() instanceof JCheckBoxMenuItem) {
            changeRepresentation(((JCheckBoxMenuItem)event.getItem()).getText());
        }
    }
    
    /** This method may be unneccessary. Previously invoked by
     createRepresentationMenu at least in some platforms (but not all of
     them). */
    public void actionPerformed(ActionEvent event) {
        changeRepresentation(event.getActionCommand());
    }
    
    
    public void changeRepresentation(String repr) {
        Container c=getParent();
        
        if (c == null) {
            Note.err(this, "No parent found. Cannot change representation.");
            return;
        }
        
        if (c instanceof Application)
            ((Application)c).changeRepresentation(this, repr);
        if (c instanceof VisualType)
            ((VisualType)c).changeRepresentation(this, repr);
    }
    
    
    /** This method is invoked by VisualComponent in order to determine
     * whether all components in this container should be labeled.
     * @return by default false
     * @see VisualAdjList
     */
    public boolean isLabeled() {
        return false;
    }
    
    /** Toggle the title label. */
    public void titled() {
        setTitled(!titled);
        setInvalid();
    }
    
    public boolean isTitled() {
        return titled;
    }
    
    public void setTitled(boolean val) {
        titled = val;
        if (visualMenuHotSpot != null)
            visualMenuHotSpot.setVisible(val && hotSpotEnabled);
        if (visualNullHotSpot != null)
            visualNullHotSpot.setVisible(val);
        if (visualSearchHotSpot != null)
            visualSearchHotSpot.setVisible(val);
    }
    
    public void setRotated(boolean val) {
        rotated = val;
    }
    
    public void setFlippedX(boolean val) {
        flippedX = val;
    }
    
    public void setFlippedY(boolean val) {
        flippedY = val;
    }
    
    public void rotated() {
        rotated = !rotated;
        setInvalid();
    }
    
    public void flipX() {
        flippedX = !flippedX;
        setInvalid();
    }
    
    public void flipY() {
        flippedY = !flippedY;
        setInvalid();
    }
    
    public void rename() {
        rename(getApplication().getInput("Rename: ", getName()));
    }
    
    public void rename(String newName) {
        setName(newName);
        if (getStructure() instanceof LabelDecorator) {
            LabelDecorator ld = (LabelDecorator)getStructure();
            ld.setLabel(getName());
        }
        setInvalid();
    }
    
    /** Determines whether the specified FDT has a representing
     ** component between index zero and the specified index.
     */
    public boolean hasRepresentation(FDT fdt, int i) {
        int k = lookUpVisualItem(fdt, 0);
        if ((k == -1) || (k >= i)) return false;
        return true;
    }
    
    
    /** @return true if this visualization is nested into another one. */
    public boolean isNestedContainer() {
        //Note.out(this, "is nested? "+getParent());
        if (getParent() == null) return false;
        if (getParent() instanceof Application) return false;
        return true;
    }
    
    /** This method returns always false because
     * VisualContainer cannot be duplicate.
     XXXX This method should not always return false becase otherwise
     arrays wont recognize the infinite loop in validation procedure
     public boolean isDuplicate() {
     return false;
     }
     */
    
    /** Are the contents of this container rotated 90 degrees? */
    public boolean isRotated() {
        return rotated;
    }
    
    /** Are the contents' X axes flipped from the default orientation? */
    public boolean isFlippedX() {
        return flippedX;
    }
    
    /** Are the contents' Y axes flipped from the default orientation? */
    public boolean isFlippedY() {
        return flippedY;
    }
    
    /** Returns the horizontal gap between nodes in layout coordinates. */
    public int getGapWidth() { return 5; }
    /** Returns the vertical gap between nodes in layout coordinates. */
    public int getGapHeight() { return 10; }
    
    /** Convert a Dimension to a LayoutSize, applying rotation.
     * @param d the Dimension object to convert
     * @return a new LayoutSize object
     */
    public LayoutSize dimensionToLayoutSize(Dimension d) {
        if (isRotated())
            return new LayoutSize(d.height, d.width);
        else
            return new LayoutSize(d.width, d.height);
    }
    
    /** Convert a LayoutSize to a Dimension, applying rotation.
     * @param d the LayoutSize object to convert
     * @return a new Dimension object
     */
    public Dimension layoutSizeToDimension(LayoutSize ls) {
        if (isRotated())
            return new Dimension(ls.height, ls.width);
        else
            return new Dimension(ls.width, ls.height);
    }
    
    /** Convert a Rectangle to a LayoutBounds, applying rotation and
     ** flipping.
     * @param r the Rectangle object to convert
     * @return a new LayoutBounds object
     */
    public LayoutBounds rectangleToLayoutBounds(Rectangle r) {
        int x, y, width, height, cWidth, cHeight, tmp;
        Rectangle realBounds = getBounds();
        Insets realInsets = getInsets();
        
        // This code should calculate the inverse of layoutBoundsToRectangle
        
        cWidth = realBounds.width - realInsets.left - realInsets.right;
        cHeight = realBounds.height - realInsets.top - realInsets.bottom;
        
        // Remove insets
        x = r.x - realInsets.left;
        y = r.y - realInsets.top;
        width = r.width;
        height = r.height;
        
        if (isFlippedY()) {
            y = -y + cHeight - height;
        }
        
        if (isFlippedX()) {
            x = -x + cWidth - width;
        }
        
        if (isRotated()) {
            tmp = width; width = height; height = tmp;
            tmp = x;
            x = -y + cHeight - width;
            y = tmp;
        }
        
        return new LayoutBounds(x, y, width, height);
    }
    
    public Rectangle layoutBoundsToRectangle(LayoutBounds lb) {
        int x, y, width, height, cWidth, cHeight;
        Rectangle realBounds = getBounds();
        Insets realInsets = getInsets();
        
        cWidth = realBounds.width - realInsets.left - realInsets.right;
        cHeight = realBounds.height - realInsets.top - realInsets.bottom;
        
        // Do rotation before flipping (perhaps more intuitive this way)
        if (isRotated()) {
            width = lb.height;
            height = lb.width;
            x = lb.y;
            y = cHeight - lb.x - lb.width;
        } else {
            x = lb.x; y = lb.y; width = lb.width; height = lb.height;
        }
        
        if (isFlippedX()) {
            x = cWidth - x - width;
        }
        
        if (isFlippedY()) {
            y = cHeight - y - height;
        }
        
        // Add insets (no need to add realBounds.{x,y})
        x = x + realInsets.left;
        y = y + realInsets.top;
        
        return new Rectangle(x, y, width, height);
    }
    
    /** This is the default doLayout method to lay out the components
     ** of this container.
     * This method invokes the doLayout method for all the components.
     * Space is reserved for the default frame and the actual component
     * layout is determined by invoking the abstract doActualLayout method.
     * @see #doActualLayout()
     */
    public void doLayout() {
        super.doLayout();
        
        Insets in = getInsets();
        int upperLeftCornerLabelWidth = 0;
        int upperRightCornerLabelWidth = 0;
        Dimension size;
        
        if (hotSpotEnabled) {
            if (visualMenuHotSpot != null)
                upperLeftCornerLabelWidth =
                    visualMenuHotSpot.getPreferredSize().width;
            if (visualSearchHotSpot != null) 
                upperRightCornerLabelWidth =
                    visualSearchHotSpot.getPreferredSize().width;
            else if (visualResizeHotSpot != null) 
                upperRightCornerLabelWidth =
                    visualResizeHotSpot.getPreferredSize().width;
        }
        if (getItemCount() > 0) {
            size = layoutSizeToDimension(doActualLayout());
            
            height = in.top + size.height + in.bottom;
            width = in.left + size.width + in.right;
            if (isTitled()){
                int titleWidth = Math.max(getMetrics().stringWidth(getName()) + upperLeftCornerLabelWidth + upperRightCornerLabelWidth + 2 + ROUNDING*2, width);
                if (titleWidth > width)
                    width = titleWidth;
            }
        } else {
            size = getPreferredSize();
            height = size.height;
            width = size.width;
            if (isTitled()){
                int titleWidth = getMetrics().stringWidth(getName()) + upperLeftCornerLabelWidth + upperRightCornerLabelWidth + 2 + 2*ROUNDING;
                if (titleWidth > width)
                    width = titleWidth;
            }
        }
        if (visualMenuHotSpot != null) {
            Dimension d = visualMenuHotSpot.getPreferredSize();
            int w = width > d.width ? d.width : width;
            visualMenuHotSpot.setBounds(ROUNDING, 1, w, d.height);
        }
        if (visualSearchHotSpot != null) {
            Dimension r = visualSearchHotSpot.getPreferredSize();
            visualSearchHotSpot.setBounds(getBounds().width-r.width-ROUNDING, 1, r.width, r.height-1);
        }
        if (visualResizeHotSpot != null) {
            Dimension r = visualResizeHotSpot.getPreferredSize();
            visualResizeHotSpot.setBounds(getBounds().width-r.width-ROUNDING, 1, r.width, r.height);
        }
        if (visualNullHotSpot != null) {
            Dimension d = visualNullHotSpot.getPreferredSize();
            visualNullHotSpot.setBounds(getBounds().width-d.width-ROUNDING, getBounds().height-d.height-1, d.width, d.height);
        }
        
        //Center the components
        Dimension offset = new Dimension(width - in.left - size.width - in.right, height - size.height - in.top - in.bottom);
        if (getItemCount() > 0 && (offset.width != 0 || offset.height != 0)) {
            for (int i=0; i < getItemCount(); i++) {
                VisualComponent vt = (VisualComponent)getItem(i);
                LayoutBounds l = vt.lGetBounds();
                LayoutSize ls = dimensionToLayoutSize(offset);
                l.x += ls.width/2;
                l.y += ls.height;
                Rectangle r = layoutBoundsToRectangle(l);
                vt.setBounds(r.x, r.y, r.width, r.height);
            }
        }
        int cCount = getLinkCount();
        for (int i = 0; i < cCount; i++)
            getLink(i).doLayout();
    }
    
    /** Do the actual layout of this container.
     * A container needs to override this method in order to lay out
     * its components. The layout must be done in layout coordinates,
     * using the method lGetPreferredSize to get the preferred size of
     * each component. The lSetBounds method should be called on each
     * component to set its position and actual size. If needed,
     * lGetBounds can be used to query a component's current size and
     * position in layout coordinates.
     *
     * The layout should be done downwards (increasing y) from (0, 0);
     * appropriate rotations etc. will be done as the layout
     * coordinates are converted into actual coordinates. As a special
     * case, lists and arrays can be laid out left-to-right (from (0,
     * 0), increasing x); this determines their default orientation.
     *
     * @return non-null object that determines the actual size of the container
     */
    protected abstract LayoutSize doActualLayout();
    
    /** Return a representation for the specified FDT. */
    public VisualType getRepresentation(Object fdt) {
        return RepresentationFactory.getRepresentation(fdt);
    }
    
    /** This method returns the default insets for a framed container.
     * Insets include an empty border around the image. */
    public Insets getInsets() {
        if (!isTitled())
            return new Insets(3,1,3,1);
        else if (getFont() != null) {
            int h = getMetrics().getHeight() + getGapHeight();
            int w = 1 + getGapWidth();
            return new Insets(h, w, h, w);
        } else
            return super.getInsets();
    }
    
    /** Look for an "empty" VisualItem for given FDT starting from the specified index.
     * @return -1 if no visualItem found for s;
     * index n >= i if visualItem(n) is representing s
     */  
    public int lookUpEmptyVisualItem(FDT s, VisualComponent neighbor, int i) {
        if (i < 0)
            Note.err(this, "lookUpVisualItem()::Illegal parameter i = " + i);
        while (i < getItemCount()) {
            if (getItem(i).isRepresenting(s)
                    && ((matrix.structures.FDT.probe.Key)emptyNeighborMap.get(getItem(i))).equals(neighbor.getKey())) {                
                return i;
            }
            i++;
        }
        return -1;
    }        
    
    
    /**Registers a new "empty" VisualItem as being connected to another visual entity.
     * This information is used when figuring out the correct locations of reused empty nodes.
     */
    public void registerEmptyElement(VisualComponent empty, VisualComponent neighbor){
        emptyNeighborMap.put(empty, neighbor.getKey());
    }
    
    /** Look for VisualItem for given FDT starting from the specified index.
     * @return -1 if no visualItem found for s;
     * index n >= i if visualItem(n) is representing s
     */
    public int lookUpVisualItem(FDT s, int i) {
        if (i < 0)
            Note.err(this, "lookUpVisualItem()::Illegal parameter i = " + i);
        while (i < getItemCount()) {
            if (getItem(i).isRepresenting(s)
                    /* DONT REMOVE THIS CODE, Otto
                     * related to VisualTree setcomponents fix 
                     * 
                     * && emptyNeighborMap.containsKey(getItem(i))==false */) {
                
                return i;
            }
            i++;
        }
        return -1;
    }
    
    /** Looking for a VisualReference between specified FDTs.
     * @return -1 if no such a reference was found
     * index n >= i if visualReference exists at position n
     */
    public int lookUpVisualReference(VisualType from, VisualType to, int i) {
        if (i < 0)
            Note.err(this, "lookUpVisualRef():Illegal parameter i = " + i);
        if ((from == null) || (to == null))
            Note.err(this, "lookUpVisualRef():Illegal parameters: " +
                    "from == " +from+
                    "to === " + to);
        while (i < getLinkCount()) {
            VisualReference vr = getLink(i);
            VisualType vc_from = vr.getSource();
            VisualType vc_to = vr.getTarget();
            if (from.equals(vc_from) && to.equals(vc_to))
                return i;
            i++;
        }
        return -1;
    }
    
    /** This method returns the default minimum size for a container. */
    public Dimension getMinimumSize() {
        return getPreferredSize();
    }
    
    private StyleSheet defaultStyleSheet = null;
    public StyleSheet getDefaultStyleSheet() {
        if (defaultStyleSheet == null) {
            defaultStyleSheet = new StyleSheetAdapter();
            defaultStyleSheet.setDefaultPenColor(Color.white);
            defaultStyleSheet.setFocusedPenColor(Color.green);
        }
        return defaultStyleSheet;
    }


    /** This method first calls super.paintChildren and
     * after that paintFrame.
     * This causes that search and null hotspots and head reference of
     * VisualTrees (which are thus children of VisualContainer) are 
     * drawn under the borders of VisualContainer by default.
     * They are highlighted, when pointed with the mouse.*/
    public void paintChildren(Graphics g) {
        super.paintChildren(g);
        paintFrame(g);
    }
    
    /** This method draws the default frame representation into
     ** the specified Graphics. Invoked by Java.awt.
     */
    public void paintFrame(Graphics g) {
        if (g instanceof ConfigurableGraphics) {
            if (((ConfigurableGraphics) g).isDisabled("StructureBorder")) {
                // if exporting and structure borders are not wanted then stop
                // here
                return;
            }
        }

        // Nested containers do not have a frame, thus we do this only in case
        // this is not a nested container
        if (!isNestedContainer() || hasFocus()) {
            Rectangle r = getBounds();
            int x = 0;
            int y = 0;
            int w = r.width - 1;
            int h = r.height - 1;
            // draw frame and title of this array
            setBorderColor(g);

            if (getStructure() instanceof CDT)
                g.setColor(getCDTColor());
            g.drawRoundRect(x, y, w, h, ROUNDING*2, ROUNDING*2); // bounding box
            if (isTitled()) { // calculate, e.g., places for the hot spots
                // only if the container is titled
                String name = getName();
                int nWidth = getMetrics().stringWidth(name);
                int nHeight = getMetrics().getHeight();
                int descent = getMetrics().getDescent();

                // MenuHotSpot+SearchHotSpot and nullHotSpot
                int menuHotSpotLabelWidth = 0;
                int upperRightCornerHotSpotWidth = 0;
                int nullHotSpotLabelWidth = 0;
                if (hotSpotEnabled) {
                    menuHotSpotLabelWidth = visualMenuHotSpot
                            .getPreferredSize().width;
                    if (visualSearchHotSpot != null)
                        upperRightCornerHotSpotWidth = visualSearchHotSpot
                                .getPreferredSize().width;
                    else if (visualResizeHotSpot != null)
                        upperRightCornerHotSpotWidth = visualResizeHotSpot
                                .getPreferredSize().width;
                    if (visualNullHotSpot != null)
                        nullHotSpotLabelWidth = visualNullHotSpot
                                .getPreferredSize().width;
                }
                int roundDiam = 2 * ROUNDING;

                // top bar begins
                // left section consists of an arc on top and a rectangle below it
                g.fillArc(0, 0, roundDiam, roundDiam, 90, 90);
                g.fillRect(0, ROUNDING, ROUNDING, nHeight - ROUNDING);
                // center section is a rectangle
                g.fillRect(ROUNDING + menuHotSpotLabelWidth, y, w + 1
                        - roundDiam - menuHotSpotLabelWidth
                        - upperRightCornerHotSpotWidth, nHeight);
                // right section is again an arc and a rectangle
                g.fillArc(w - roundDiam + 1, 0, roundDiam, roundDiam, 0, 90);
                g.fillRect(w - ROUNDING + 1, ROUNDING, ROUNDING - 1, nHeight
                        - ROUNDING);
                // top bar ends

                // bottom bar begins
                // left section is a rectangle on top and an arc below it
                g.fillArc(0, y + h - roundDiam, roundDiam, roundDiam, 180, 90);
                g.fillRect(0, y + h - nHeight, ROUNDING, nHeight - ROUNDING);
                // center section
                g.fillRect(ROUNDING, y + h - nHeight, w + 1 - roundDiam
                        - nullHotSpotLabelWidth, nHeight);
                // right section
                g.fillArc(w - roundDiam + 1, y + h - roundDiam, roundDiam,
                          roundDiam, 0, -90);
                g.fillRect(w - ROUNDING + 1, y + h - nHeight, ROUNDING, nHeight
                        - ROUNDING);
                // bottom bar ends

                // calculate the correct position and space left for title name
                // after hot spots
                x += menuHotSpotLabelWidth;
                // cut width according to both possible hot spots
                w -= (menuHotSpotLabelWidth + upperRightCornerHotSpotWidth);

                if (w < 0)
                    return; // does not fit, just return

                while ((name.length() > 1) && (nWidth > w)) { // try to have
                    // at least some
                    // title name
                    name = name.substring(0, name.length() - 1);
                    nWidth = getFontMetrics(getFont()).stringWidth(name);
                }
                setPenColor(g); // title drawing color
                g.drawString(name, x + w / 2 - nWidth / 2, y + nHeight
                        - descent);
            } else { // if not titled, just paint the single line borders
                g.fillRect(x, y + h - 3, w, 3); // bottom bar
                g.fillRect(x, y, w, 3); // top bar
            }
        }
    }
    
    /** This is the default paint method for minimized visual objects.
     * The default representation illustrates the visual object in
     * shape of Polygon that is determined by invoking the
     * getMinimizedPolygon method.
     * @see VisualType#getMinimizedPolygon(int, int, int, int)
     */
    public void paint_minimized(Graphics g) {
        Rectangle r = getBounds();
        int w = r.width;
        int h = r.height;
        int x = 1+r.width/2;
        int y = 1+h/2;
        
        //Note.out(this, "Minimized + " + this + " b= " + r);
        
        setBorderColor(g);
        g.drawPolygon(getMinimizedPolygon(x-w/2, y-h/2, w, h));
        setBackgroundColor(g);
        g.fillPolygon(getMinimizedPolygon(x-w/2, y-h/2, w, h));
        String key = "@";
        x -= getMetrics().stringWidth(key)/2;
        y += getMetrics().getHeight()/2;
        setPenColor(g);
        g.drawString(key, x, y-1);
    }
    
    //--- methods for handling insertion ---
    
    /**
     * Tests if insert can be done to this visual container
     * WHAT A FINE FEATURE WAS THIS?
     public abstract boolean canInsert();
     */
    
    /** Inserts an object into the data structure this container is
     * representing. The subclasses should override this method
     * in order to provide proper behaviour.
     * @param structure the Object to be inserted
     */
    public void insert(Object structure) {
        Note.show(this, "Insert operation not implemented for " + this);
    }
    
    /** Deletes an object from the data structure this container is
     * representing. The subclasses should override this method
     * in order to provide proper behaviour.
     * @param structure the Object to be deleted
     */
    public void delete(Object structure) {
        Note.show(this, "Delete operation not implemented for " + this);
    }
    
    /** Searches an object from the data structure this container is
     * representing. The subclasses should override this method in
     * order to provide proper behaviour.
     * @param structure the Object to be searched
     */
    public void search(Object structure) {
        if (getStructure() instanceof CDT)
            ((CDT)getStructure()).search(structure);
        else
            Note.show(this, "Search operation not implemented for non-CDT");
        
    }
    
    /** This method is invoked when an assign operation is performed
     * for an CDT during the simulation process. The assign operation
     * is interpreted to be one of the following: insert, delete or
     * search depending on the current default method determined by
     * the getDefaultCallee() method.
     */
    public void assign(Object structure) {
        getAnimator().startOperation();
        if (getDefaultCallee().equals("insert") ||
                getDefaultCallee().equals("assign"))
            insert(structure);
        else if (getDefaultCallee().equals("delete")) 
            delete(structure);
        else if (getDefaultCallee().equals("search"))
            search(structure);
        else
            Note.err(this, "Illegal operation: " + getDefaultCallee());
        getAnimator().endOperation();
    }
    
    /** This method is invoked when an assign operation is performed
     * during the simulation process. By default, the assign(Object)
     * is invoked for CDTs and insert(Object) for non-CDTs.
     */
    public void assign() {
        if (getStructure() instanceof CDT) {
            if(read.getStructure() instanceof Iterator) {
                Iterator iter = (Iterator) read.getStructure();
                iter.reset();
                while(iter.hasNext()) {
                    getAnimator().startOperation();
                    assign(iter.next());
                    getAnimator().endOperation();
                }
            }
            else{
                assign(read.getStructure());
            }
        } else
            insert(read.getStructure());
    }
    
    /** This method is invoked when an remove operation is performed
     * during the simulation process. By default, the assign(Object)
     * is invoked for CDTs and delete(Object) for non-CDTs.
     */
    public void remove() {
        if (getStructure() instanceof CDT) {
            if(read.getStructure() instanceof Iterator) {
                Iterator iter = (Iterator) read.getStructure();
                iter.reset();
                while(iter.hasNext()) {
                    getAnimator().startOperation();
                    assign(iter.next());
                    getAnimator().endOperation();
                }
            }
            else{
                assign(read.getStructure());
            }
        } else
            delete(read.getStructure());
    }
    
    private boolean autoLabeled = false;
    public void setAutoLabeling(boolean enab) {
        autoLabeled = enab;
    }
    public boolean isAutoLabeled() {
        return autoLabeled;
    }
    
    public void configure(VisualTypeConf vtc) {
        if (vtc != null) {
            String value = vtc.getVisualConfValue(this.getClass().getName(), VisualTypeConf.FLIPPEDY);
            //In the following way, no need to check whether value is null or not.
            if ("true".equals(value)) {
                setFlippedY(true);
            }
            value = vtc.getVisualConfValue(this.getClass().getName(), VisualTypeConf.FLIPPEDX);
            if ("true".equals(value)) { 
                setFlippedX(true);
            }
            value = vtc.getVisualConfValue(this.getClass().getName(), VisualTypeConf.ROTATED);
            if("true".equals(value)) {
                setRotated(true);
            }
        }
        super.configure(vtc);
    }
}



