/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: VisualLayeredGraph.java,v 1.34 2007/09/05 10:21:06 oseppala Exp $ */

package matrix.visual;

import java.awt.Component;
import java.util.*;

import matrix.structures.FDT.*;
import matrix.structures.FDT.substructures.Vertex;
import matrix.util.Note;

/**
 * A version of the DAG drawing algorithm from Chapter 9 ("Layered
 * Drawings of Digraphs") in Battista et. al., _Graph Drawing_,
 * supporting arbitrary graphs and variable-size nodes.
 *
 * @author Riku Saikkonen
 * @version $Revision: 1.34 $
 */
public class VisualLayeredGraph extends VisualGraph {
    private Vector/*<Vector<VisualGraphComponent>>*/ layers;
    private boolean orderChanged; // for crossingReduction

    public VisualLayeredGraph(Vertex v) { super(v); }
    public VisualLayeredGraph(Graph g) { super(g); }

    protected VisualGraphComponent createNewComponent(Vertex v) {
        return new VisualLayeredGraphComponent(v);
    }    

    protected LayoutSize doActualLayout() {
        resetComponentData();
        removeCyclesGreedily();
        assignToLayers();
        addDummyNodes();
        crossingReduction();
        return placeNodes(0, 0);
    }

    public int getGapWidth() { return 15; }
    public int getGapHeight() { return 15; }

    /** Causes components to forget their internal data */
    private void resetComponentData() {
        Component[] comps = getItems();
        for (int i = 0; i < comps.length; i++)
            ((VisualLayeredGraphComponent)comps[i]).resetData();

        Component[] links = getLinks();
        for (int i = 0; i < links.length; i++)
            ((VisualReference)links[i]).forgetMidPoints();
    }

    /** Assigns components to (improper) layers.
     * Uses longest path layering. The input graph must be acyclic!
     * Call addDummyVertices() after this to make the layering proper.
     */
    private void assignToLayers() {
        layers = new Vector();

        java.awt.Component[] comps = getItems();
        int numAssigned = 0, currentLayer = 0;

        do {
            Vector layer = new Vector();

        componentLoop:
            for (int i = 0; i < comps.length; i++) {
                VisualLayeredGraphComponent c =
                    (VisualLayeredGraphComponent) comps[i];

                if (c.hasLayer())
                    continue;

                // reversed edges included here (no dummy nodes exist yet)
                VisualLayeredGraphComponent succs[] = c.getPseudoSuccs();

                for (int j = 0; j < succs.length; j++) {
                    VisualLayeredGraphComponent succ = succs[j];

                    if (!succ.hasLayer() || succ.getLayer() == currentLayer)
                        continue componentLoop;
                }
                
                // no unassigned successors, so assign this node to
                // the current layer
                // Note.out(this, "layer "+currentLayer+" has node "+c);
                layer.addElement(c);
                c.setLayer(currentLayer);
                numAssigned++;
            }

            if (layer.isEmpty()) {
                Note.err(this, "Cycle removal failed?");
            } else {
                // Switch to next (upper) layer
                currentLayer++;
                layers.addElement(layer);
            }
        } while (numAssigned < comps.length);
    }

    /** Greedy cycle removal.
     * Does greedy cycle removal (alg. 9.4 steps 2b and 2c from _Graph
     * Drawing_), modified so that nodes to be inserted into Sl are
     * marked instead, and edges are reversed when marking instead of
     * later.
     */
    private void removeCyclesGreedily() {
        java.awt.Component[] comps = getItems();
        boolean found;

        // Mark sinks repeatedly with markRCGSink (alg. 9.4 2a)
        do {
            found = false;

        componentLoop:
            for (int i = 0; i < comps.length; i++) {
                VisualLayeredGraphComponent c =
                    (VisualLayeredGraphComponent) comps[i];
                
                if (c.hasRCGMark())
                    continue;

                // The real successors are enough for us here, because
                // edges haven't been reversed and dummy nodes haven't
                // been created yet.
                VisualGraphComponent succs[] = c.getSuccessorComponents();

                for (int j = 0; j < succs.length; j++) {
                    VisualLayeredGraphComponent succ =
                        (VisualLayeredGraphComponent) succs[j];

                    if (!succ.hasRCGMarkSink())
                        // not a sink, try next component
                        continue componentLoop;
                }

                // c is a sink; mark it. No back edges here.
                c.markRCGSink();
                found = true;                
            }
        } while (found);

        // Mark sources repeatedly with markRCGSource (alg. 9.4 2b)
        do {
            found = false;
        componentLoop:
            for (int i = 0; i < comps.length; i++) {
                VisualLayeredGraphComponent c =
                    (VisualLayeredGraphComponent) comps[i];
                
                if (c.hasRCGMark())
                    continue;

                // The real predecessors are enough for us here,
                // because edges haven't been reversed and dummy nodes
                // haven't been created yet.
                VisualGraphComponent preds[] = c.getPredecessorComponents();

                for (int j = 0; j < preds.length; j++) {
                    VisualLayeredGraphComponent pred =
                        (VisualLayeredGraphComponent) preds[j];

                    if (!pred.hasRCGMarkSource())
                        // not a source, try next component
                        continue componentLoop;
                }

                // c is a source; mark it. No back edges here: a
                // source is a node with no unmarked predecessors, so
                // a new source can't have a previously marked source
                // as a successor.
                c.markRCGSource();
                found = true;
            }
        } while (found);
        
        // Choose unmarked nodes (alg. 9.4 2c), mark them and reverse
        // edges that point to a source-marked node (=leftward edges)
        do {
            VisualLayeredGraphComponent foundC = null;
            int foundCValue = 0; // not used when foundC==null

            for (int i = 0; i < comps.length; i++) {
                VisualLayeredGraphComponent c =
                    (VisualLayeredGraphComponent) comps[i];
                
                if (c.hasRCGMark())
                    continue;

                // If we get here, this node still needs to be processed.

                // The real predecessors and successors are enough for
                // us here, because unmarked nodes' edges haven't been
                // reversed and dummy nodes haven't been created yet.

                // Calculate inDeg, with only unprocessed nodes
                VisualGraphComponent preds[] = c.getPredecessorComponents();
                int inDeg = 0;
                for (int j = 0; j < preds.length; j++) {
                    VisualLayeredGraphComponent pred =
                        (VisualLayeredGraphComponent) preds[j];

                    if (pred.hasRCGMark())
                        continue;

                    inDeg++;
                }

                // Calculate outDeg, with only unprocessed nodes
                VisualGraphComponent succs[] = c.getSuccessorComponents();
                int outDeg = 0;
                for (int j = 0; j < succs.length; j++) {
                    VisualLayeredGraphComponent succ =
                        (VisualLayeredGraphComponent) succs[j];

                    if (succ.hasRCGMark())
                        continue;

                    outDeg++;
                }

                int value = outDeg - inDeg;

                if (foundC == null || foundCValue < value) {
                    foundC = c;
                    foundCValue = value;
                }
            }

            if (foundC != null) {
                // mark foundC and reverse its left edges
                // this is a new source after edge reversal
                foundC.markRCGSource();

                VisualGraphComponent succs[] = foundC.getSuccessorComponents();
                for (int j = 0; j < succs.length; j++) {
                    VisualLayeredGraphComponent succ =
                        (VisualLayeredGraphComponent) succs[j];

                    if (succ.hasRCGMarkSource()) {
                        // This is a left edge; reverse it
                        // Note.out(this, "reversing edge "+foundC+"->"+succ);
                        foundC.reverseEdge(succ);
                    }
                }
            } else break; // all nodes processed
        } while (true);

        // Remove reflective arcs from pseudo successors (they are
        // also cycles, and would disturb the remaining phases of the
        // layout)
        for (int i = 0; i < comps.length; i++) {
            VisualLayeredGraphComponent c =
                (VisualLayeredGraphComponent) comps[i];
            c.removeReflectivePseudos();
        }
    }

    /** Add dummy vertices to make the layering proper. */
    private void addDummyNodes() {
        for (int curLayer = layers.size() - 1; curLayer >= 0; curLayer--) {
            Enumeration e = ((Vector)layers.elementAt(curLayer)).elements();

            while (e.hasMoreElements()) {
                VisualLayeredGraphComponent c =
                    (VisualLayeredGraphComponent) e.nextElement();
                VisualLayeredGraphComponent succs[] = c.getPseudoSuccs();

                for (int i = 0; i < succs.length; i++) {
                    if (succs[i].getLayer() < curLayer - 1) {
                        // Add dummy node on next layer between c and succs[i]
                        VisualLayeredGraphComponent dummy = c.addDummyNode(i);

                        // curLayer should always be >= 1 here,
                        // because succ is on a following layer
                        dummy.setLayer(curLayer - 1);
                        ((Vector)layers.elementAt(curLayer - 1)).
                            addElement(dummy);
                    }
                }
            }
        }
    }

    /** Sorts layer contents to reduce crossings.
     * Uses the median method, sweeping in both directions.
     */
    private void crossingReduction() {
        Enumeration le = layers.elements();

        if (!le.hasMoreElements())
            return; // no layers

        // Keep bottom layer's ordering as is
        assignPosFromLayer((Vector)le.nextElement());

        // Sweep until nothing changes, to a maximum of 5 sweeps
        // (Odd numbers of sweeps seem to work better.)
        // XXX do adjacent-exchange afer median method?
        for (int i = 0; i < 5; i++) {
            orderChanged = false;

            // Sweep bottom-to-top
            while (le.hasMoreElements())
                medianCrossingReduction((Vector)le.nextElement(),
                                        false);
            
            // Sweep top-to-bottom
            for (int curLayer = layers.size() - 2; curLayer >= 0; curLayer--)
                medianCrossingReduction((Vector)layers.elementAt(curLayer),
                                        true);

            if (!orderChanged) // exit loop if nothing changed
                break;
        }
    }

    /** Sorts a layer to reduce crossings using the median method.
     * The method is described in section 9.2.3 of _Graph Drawing_.
     * @param layer the layer to sort
     * @param pred if true, reduce predecessor crossings instead of
     *   successor crossings
     */
    private void medianCrossingReduction(Vector layer, boolean pred) {
        // Calculate medians for each element
        Enumeration e = layer.elements();
        while (e.hasMoreElements()) {
            VisualLayeredGraphComponent c =
                (VisualLayeredGraphComponent)e.nextElement();            
            VisualLayeredGraphComponent adjs[] =
                pred ? c.getPseudoPreds() : c.getPseudoSuccs();
            double median = 0.0;
            
            if (adjs.length > 0) {
                // Quickselect modified slightly from page 146 of
                // Sedgewick: _Algorithms_
                int k = (adjs.length - 1) / 2;
                int l = 0, r = adjs.length - 1;
                while (r > l) {
                    double v = adjs[r].getPosInLayer();
                    int i = l - 1;
                    int j = r;

                    for (;;) {
                        do { i++; }
                        while (i < j && adjs[i].getPosInLayer() < v);
                        do { j--; }
                        while (i < j && adjs[j].getPosInLayer() > v);
                        if (i >= j)
                            break;
                        // swap adjs[i] and adjs[j]
                        VisualLayeredGraphComponent t = adjs[i];
                        adjs[i] = adjs[j]; adjs[j] = t;
                    }

                    if (i >= k)
                        r = i - 1;
                    else l = i + 1;
                }
                median = adjs[k].getPosInLayer();
                
                // Handle tie-breaking: the median is always an integer,
                // so to sort even degree vertices after odd degree ones,
                // we can add 0.5 to their median.
                if (adjs.length % 2 == 0)
                    median += 0.5;
            } else median = 0.0;

            c.setPosInLayer(median);
        }

        sortLayerByPos(layer);
        assignPosFromLayer(layer);
    }

    /** Sorts a layer vector by posInLayer in the components.
     * Ties are broken arbitrarily. Sets orderChanged if something changes.
     */
    private void sortLayerByPos(Vector layer) {
        // Insertion sort (fast enough because there usually aren't
        // many nodes, they are often almost sorted already, and the
        // sort can be done in place)
        int n = layer.size();
        for (int i = 1; i < n; i++) {
            VisualLayeredGraphComponent c =
                (VisualLayeredGraphComponent) layer.elementAt(i);

            for (int j = i - 1; j >= 0; j--) {
                VisualLayeredGraphComponent prev =
                    (VisualLayeredGraphComponent) layer.elementAt(j);

                if (prev.getPosInLayer() < c.getPosInLayer())
                    break; // c is in correct place

                // swap prev and c
                layer.setElementAt(prev, j + 1); // c was here
                layer.setElementAt(c, j); // prev was here
                orderChanged = true;
            }
        }
    }
    
    /** Assigns posInLayer values from the ordering of a layer vector. */
    private void assignPosFromLayer(Vector layer) {
        Enumeration e = layer.elements();
        double pos = 1.0;

        while (e.hasMoreElements()) {
            VisualLayeredGraphComponent c =
                (VisualLayeredGraphComponent) e.nextElement();
            c.setPosInLayer(pos);
            pos += 1.0;
        }
    }

    /** Assigns coordinates for nodes, and sets their bounds.
     * Places layers from top to bottom, nodes from left to right.
     *
     * XXX try to hide dummy vertices?
     *
     * @param ulx x coordinate of upper left corner of graph
     * @param uly y coordinate of upper left corner of graph
     * @return size of complete graph
     */
    private LayoutSize placeNodes(int ulx, int uly) {
        int width = 0, height = 0, x = ulx, y = uly;
        int layerWidth[] = new int[layers.size()];
        int layerHeight[] = new int[layers.size()];

        for (int curLayer = layers.size() - 1; curLayer >= 0; curLayer--) {
            Enumeration e = ((Vector) layers.elementAt(curLayer)).elements();

            x = ulx;
            layerHeight[curLayer] = 0;

            while (e.hasMoreElements()) { // for each node
                VisualNode c = (VisualNode) e.nextElement();
                LayoutSize cSize = c.lGetPreferredSize();
                c.lSetBounds(x, y, cSize.width, cSize.height);

                x += cSize.width + getGapWidth();

                if (layerHeight[curLayer] < cSize.height)
                    layerHeight[curLayer] = cSize.height;
            }
            
            if (x > ulx) // if there were elements
                x -= getGapWidth(); // remove extra gap

            layerWidth[curLayer] = x - ulx;

            if (width < x - ulx)
                width = x - ulx;

            y += layerHeight[curLayer] + getGapHeight();            
        }

        if (y > uly)
            y -= getGapHeight(); // remove extra gap

        height = y - uly;

        // Center layers horizontally and their components vertically
        for (int curLayer = layers.size() - 1; curLayer >= 0; curLayer--) {
            int xTrans = (width - layerWidth[curLayer]) / 2;
            Enumeration e = ((Vector)layers.elementAt(curLayer)).elements();
            while (e.hasMoreElements()) {
                VisualNode c = (VisualNode) e.nextElement();
                LayoutBounds cb = c.lGetBounds();
                int yTrans = (layerHeight[curLayer] - cb.height) / 2;
                c.lSetBounds(cb.x + xTrans, cb.y + yTrans,
                             cb.width, cb.height);
            }
        }

        return new LayoutSize(width, height);
    }
}
