/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: VisualLayeredGraphComponent.java,v 1.18 2006/05/24 13:58:53 vkaravir Exp $ */

package matrix.visual;

import java.util.*;

import matrix.structures.FDT.substructures.Vertex;

/**
 * Layout-specific storage for the VisualLayeredGraph layout.
 *
 * @author Riku Saikkonen
 * @version $Revision: 1.18 $
 * @see VisualLayeredGraph
 */
public class VisualLayeredGraphComponent extends VisualGraphComponent {
    private int layer;
    // marks for VLG.removeCyclesGreedily
    private boolean rcgMarkSource, rcgMarkSink;
    private double posInLayer;
    private Vector/*<VisualLayeredGraphComponent>*/ pseudoSuccs, pseudoPreds;
    private Vector/*<VisualReference>*/ pseudoSuccRefs;

    public VisualLayeredGraphComponent(Vertex v) {
        super(v);
        resetData();
    }
    
    public void resetData() {
        layer = -999;
        rcgMarkSource = false;
        rcgMarkSink = false;
        posInLayer = 0.0;

        VisualReference[] refs = getSuccessors();
        pseudoSuccs = new Vector(refs.length);
        pseudoSuccRefs = new Vector(refs.length);
        for (int i = 0; i < refs.length; i++) {
            pseudoSuccs.addElement(refs[i].getTarget());
            pseudoSuccRefs.addElement(refs[i]);
        }

        VisualComponent[] preds = getPredecessorComponents();
        pseudoPreds = new Vector(preds.length);
        for (int i = 0; i < preds.length; i++)
            pseudoPreds.addElement(preds[i]);
    }

    public boolean hasLayer() {
        return (layer != -999);
    }

    public int getLayer() {
        return layer;
    }

    public void setLayer(int layer) {
        this.layer = layer;
    }

    public boolean hasRCGMark() {
        return rcgMarkSource || rcgMarkSink;
    }

    public boolean hasRCGMarkSource() {
        return rcgMarkSource;
    }

    public boolean hasRCGMarkSink() {
        return rcgMarkSink;
    }

    public void markRCGSource() {
        rcgMarkSource = true;
    }

    public void markRCGSink() {
        rcgMarkSink = true;
    }

    public double getPosInLayer() {
        return posInLayer;
    }

    public void setPosInLayer(double p) {
        posInLayer = p;
    }

    /** Returns the current pseudo successors of this node. 
     * The caller may modify the returned array at will.
     */
    public VisualLayeredGraphComponent[] getPseudoSuccs() {
        VisualLayeredGraphComponent[] t =
            new VisualLayeredGraphComponent[pseudoSuccs.size()];
        pseudoSuccs.copyInto(t);
        return t;
    }

    /** Returns the current pseudo predecessors of this node. 
     * The caller may modify the returned array at will.
     */
    public VisualLayeredGraphComponent[] getPseudoPreds() {
        VisualLayeredGraphComponent[] t =
            new VisualLayeredGraphComponent[pseudoPreds.size()];
        pseudoPreds.copyInto(t);
        return t;
    }

    /** Removes reflective arcs from pseudo predecessors and successors.
     */
    public void removeReflectivePseudos() {
        Enumeration e = pseudoSuccs.elements();
        while (e.hasMoreElements()) {
            Object o = e.nextElement();
            if (o == this)
                pseudoSuccs.removeElement(o);
        }

        e = pseudoPreds.elements();
        while (e.hasMoreElements()) {
            Object o = e.nextElement();
            if (o == this)
                pseudoPreds.removeElement(o);
        }
    }

    /** Adds a dummy node between this node and its ith successor.
     * @param i specifies the successor via an index into the pseudo
     *   successors
     * @return the new dummy node
     */
    public VisualLayeredGraphComponent addDummyNode(int i) {
        VisualLayeredGraphComponent succ =
            (VisualLayeredGraphComponent)pseudoSuccs.elementAt(i);
        VisualReference ref = (VisualReference)pseudoSuccRefs.elementAt(i);
        VisualLayeredGraphComponent dummy =
            new VisualLayeredGraphDummyComponent(getVisualContainer(), ref);

        // Replace my successor with dummy
        pseudoSuccs.setElementAt(dummy, i);
        // pseudoSuccRefs doesn't change

        // Add predecessor and successor to dummy
        dummy.pseudoPreds.addElement(this);
        dummy.pseudoSuccs.addElement(succ);
        dummy.pseudoSuccRefs.addElement(ref);

        // Replace old successor's predecessor with dummy
        succ.pseudoPreds.setElementAt(dummy, succ.pseudoPreds.indexOf(this));

        return dummy;
    }

    /** Reverses the edge from this node to the argument node.
     * The argument must currently be a pseudo successor of this node.
     */
    public void reverseEdge(VisualLayeredGraphComponent to) {
        int i = pseudoSuccs.indexOf(to);
        VisualReference ref = (VisualReference)pseudoSuccRefs.elementAt(i);

        ref.setMidPointOrdering(true); // reverse midpoints after this

        // my successor => predecessor
        pseudoSuccs.removeElementAt(i);
        pseudoSuccRefs.removeElementAt(i);
        pseudoPreds.addElement(to);

        // to's predecessor => successor
        to.pseudoPreds.removeElement(this);
        to.pseudoSuccs.addElement(this);
        to.pseudoSuccRefs.addElement(ref);
    }
}
