/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: VisualList.java,v 1.49 2006/05/30 10:02:31 oseppala Exp $ */

package matrix.visual;

import java.awt.Graphics;

import matrix.animation.StructureChangeOperation;
import matrix.structures.CDT.CDT;
import matrix.structures.FDT.*;
import matrix.structures.FDT.probe.Key;
import matrix.util.Note;

public class VisualList extends VisualContainer {
    
    public VisualList(LinkedList l) {
        super(l);
    }
    
    public String getRepresentationName() { return "VisualList"; }

    public boolean canInsert() {
	return true;
    }

    public void insert(Object structure) {
	if (getStructure() instanceof CDT) {
	    CDT cdt = (CDT)getStructure();
        if (structure == cdt) {
            getApplication().showMessage("Error","Structure cannot be added to itself.");
            return;
        }
	    FDT tmp = cdt.insert(structure);
	    setList((LinkedList)tmp);
	} else if (getStructure() == null) { // first node
	    if (structure instanceof LinkedList)
		setList((LinkedList)structure);
	    else if (structure instanceof FDT) {
		LinkedList tmp = new matrix.structures.FDT.probe.LinkedListImpl(structure);
		setList(tmp);
	    } else {
		LinkedList tmp = new matrix.structures.FDT.probe.LinkedListImpl(new Key(structure));
		setList(tmp);
	    }
	} else if (getStructure().getElement() == null) { // getStructure() != null
	    if (structure instanceof LinkedList) { // replace empty list with new one
		setList((LinkedList)structure);
	    } else if (structure instanceof FDT) { // replace the empty key only
		getStructure().setElement(structure);
	    } else {
		getStructure().setElement(new Key(structure));
	    }
	} else { // getStructure() != null & getStructure().getElement() != null
	    if (structure instanceof LinkedList) { // append list at the end
		LinkedList tmp = (LinkedList)getStructure();
		while (tmp.getNext() != null)
		    tmp = tmp.getNext();
		tmp.setNext((LinkedList)structure);
	    } else if (structure instanceof FDT) { // insert new key at the front
		LinkedList tmp = ((LinkedList)getStructure()).getNewNode(structure);
		tmp.setNext((LinkedList)getStructure());
		setList(tmp);
	    } else {
		LinkedList tmp = ((LinkedList)getStructure()).getNewNode(structure);
		tmp.setNext((LinkedList)getStructure());
		setList(tmp);
	    }
	}
    }

    public void delete(Object structure) {
	if (getStructure() instanceof CDT) {
	    CDT cdt = (CDT)getStructure();
	    FDT tmp = cdt.delete(structure);
	    setList((LinkedList)tmp);
	}
    }
    
    protected void removeKeyHelp(Object o) {
	if(getStructure() instanceof CDT)
	    delete(o);
	else
	    super.removeKeyHelp(o);	    
    }
    
    protected void setList(LinkedList newList) {
	LinkedList oldList = (LinkedList)getStructure();
	getAnimator().put(new StructureChangeOperation(oldList, newList));
	setStructure(newList);
    }

    // -=- validation -=-

    private VisualListComponent createNewVisualListComponent(LinkedList curr, 
							     int index) {
	VisualListComponent tmp = new VisualListComponent(curr);
	tmp.setIndex(index);
	add(tmp, index);
	return tmp;
    }

    private VisualListComponent moveVisualListComponent(int from, 
							int to) {
	VisualListComponent tmp = (VisualListComponent)getItem(from);
	moveItem(from, to);
	tmp.setIndex(to);
	return tmp;
    }						
    
    private void validateReference(int index, 
				   VisualListComponent from, 
				   VisualListComponent to) {
	from.setNext(to);
	to.setPrevious(from);
	int n = getLinkCount();
	VisualReference tmp;
	if(n > index) {
	    tmp = getLink(index);
	    tmp.setSource(from);
	    tmp.setTarget(to);
	}
	else {
	    tmp = new VisualReference(from, to);
	    add(tmp, index);
	}
    } 


    /**
     * Validates components of this VisualList.
     * This iterates through the underlying LinkedList structure and checks
     * that every node in LinkedList has a corresponding VisualListComponent.
     * If it finds a node which doesn't have a visualization it creates new 
     * VisualListComponent for that node. It also updates the VisualListHeadReference
     * to point to the first VisualListComponent of the VisualList or creates a new 
     * one if it doesn't already exist. (VisualListHeadReference doesn't exist 
     * for CDT structures.) It also validates the references, handles the
     * duplicates and removes unused components and links.
     */
    public void validateComponents() {
        int index = 0;
        VisualListComponent curr = null, prev = null;
        LinkedList lst = (LinkedList)getStructure();
        if (lst != null) {
            while (lst != null /*&& lst.getElement() != null*/) {

                int k = lookUpVisualItem(lst, index);

                if (k == index) { // found and valid
                    curr = (VisualListComponent) getItem(index); // already validated
                } else if (k < 0) { // not found
                    //Note.out(this, "create new VLC for " + lst + " at " + index);
                    curr = createNewVisualListComponent(lst, index);
                    /*Updates (or creates a new) VisualListHeadReference to point to first 
                     * VisualListComponent of the list*/
                    updateVisualListHeadReference(); // should this be done only once at the end?
                    
                } else { // k > index, i.e. found but not valid
                    curr = moveVisualListComponent(k, index);
                    /*Updates (or creates a new) VisualListHeadReference to point to first 
                     * VisualListComponent of the list*/
                    updateVisualListHeadReference(); // should this be done only once at the end?
                }
                if (prev != null) { //validate reference
                    /* if structure is not CDT it has VisualListHeadReference in index 0 
                     * and therefore index of the reference differs from the CDT structure which
                     * doesn't have VisualListHeadReference
                     * */
                    if (!(getStructure() instanceof CDT))
                        validateReference(index, prev, curr);
                    else
                        validateReference(index-1, prev, curr);
                }
                if (curr.isDuplicate()) {
                    //Note.warning(this, "duplicate: " + curr);
                    index++; //this has to be updated before break because otherwise duplicates are not drawn
                    break; // end of validation if duplicate encountered
                }
                
                prev = curr;
                index++;
                lst = lst.getNext();
            }
            // remove unused components. Currently not working as it should?
            //Note.out(this, "index: " + index + " items: " + getItemCount() + " links: " + getLinkCount());
            while (index < getItemCount()) { // too many VLC's (can be deleted)
                VisualListComponent comp = (VisualListComponent)getItem(index);
                if (comp != null) {
                    VisualListComponent p = comp.getPrevious();
                    VisualListComponent n = comp.getNext();
                    if (p != null) {
                        p.setNext(n);
                    }
                    if (n != null) {
                        n.setPrevious(p);
                    }
                    comp.setNext(null);
                    comp.setPrevious(null);
                }
                removeItem(index);
            }
	    /*Updates (or creates a new) VisualListHeadReference to point to first 
	     * VisualListComponent of the list*/
	    //updateVisualListHeadReference();
                    
            if (!(getStructure() instanceof CDT)) {
                //if not CDT, this contains one extra reference (that is, HeadReference in index 0)
                while(getLinkCount()>0 && getLinkCount() > getItemCount())
                    removeLink(getLinkCount() - 1); // too many references
            } else {
                while(getLinkCount()>0 && getLinkCount() >= getItemCount())
                    removeLink(getLinkCount() - 1); // too many references                
            }
        } else { // empty list
            // delete all components if there is any
            removeAll();
        }
    }
    
    
    /**
     * Returns the VisualListComponent of the root of the list.
     * @return the first actual VisualListComponent of this list
     */
    protected VisualListComponent getRootNode() {
        if (getItemCount() > 0)
            return (VisualListComponent) getItem(0);
        else
            return null; // no root node found
    }
    
    /**
     * Updates, and if needed creates, VisualListHeadReference and
     * sets it target to the root node of the list (that is, the first
     * actual VisualListComponent of this list.)
     */
    private void updateVisualListHeadReference() {
        /* Head reference is not made for CDTs because changing the headreference
         * for a CDT would cause problems. (E.g., Underlying structure (CDT) differs from
         * the visualized one.)*/
        if (!(getStructure() instanceof CDT)) { 
            if (getLinkCount() == 0) { // reference not found
                // create new VisualListHeadReference
                add(new VisualListHeadReference(getRootNode()),0); 
            } else if (getLink(0) instanceof VisualListHeadReference) { // ref found
                // updates VisualListHeadReference target
                ((VisualListHeadReference)getLink(0)).setTarget(getRootNode());
            } else {
                Note.err(this, "setRoot():Cannot find VisualHeadReference: "+ getLink(0));
            }
        }        
    }

    
    // -=- layout -=-

    public int getGapWidth() { return 10; }
    public int getGapHeight() { return getGapWidth(); }

    public LayoutSize doActualLayout() {
        //Note.out(this, "doActualLayout "+getStructure());
        if (getStructure() == null)
            return new LayoutSize(40, 20); // size of empty list
        LayoutSize ls = getListSize((VisualListComponent)getItem(0),
                           new LayoutBounds(0, 0, 0, 0));
	//Note.out(this, "LayoutSize = " + ls.toString());
	return ls;
    }

    // This is the actual method to recursively count the size of the list and
    //lay out its representation. 
    private LayoutSize getListSize(VisualListComponent vtc,
                                   LayoutBounds bounds) {
        // x + empty space in left
        int x = bounds.x + getGapWidth();
        int y = bounds.y;
        bounds.width = 0;
        bounds.height = 0;
        while (vtc != null) {
	    //Note.out(this, "vtc="+vtc);
            LayoutSize vtcSize = vtc.lGetPreferredSize();
            vtc.lSetBounds(x, y, vtcSize.width, vtcSize.height);
            x += vtcSize.width + getGapWidth();
            bounds = bounds.union(vtc.lGetBounds());
            vtc = vtc.getNext();
        }
        // return bounds + empty space in right
        return new LayoutSize(bounds.width+getGapWidth(), bounds.height);
    }

    public void paintComponent(Graphics g) {
        super.paintComponent(g);
        if (isMinimized()) { paint_minimized(g); return; }
        if (isPicked()) return; // no frames
        paintFrame(g);
    }
}
