/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2005 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: VisualListHead.java,v 1.4 2007/09/03 12:25:25 oseppala Exp $ */

package matrix.visual;

import java.awt.Graphics;

import matrix.structures.FDT.LinkedList;
import matrix.util.Command;
import matrix.util.Validator;

/**
 * This class works as a source node for a VisualListHeadReference.
 * It is not visualized. In VisualListHeadReference, it's location is
 * set to the left border of the VisualList so that
 * VisualListHeadReference starts from there. This handles also the
 * moving of the VisualListHeadReference instances.
 * 
 * @author kstalnac
 */
public class VisualListHead extends VisualListComponent {
    VisualListComponent visualRoot;
    
    /**
     * Constructor.
     * Constructs a new VisualListHead and sets it not visible, and not enabled.
     * @param t the VisualListHeadReference that is the first visualized reference in the list
     */
    public VisualListHead(VisualComponent t) {
    	super(null);
	visualRoot = (VisualListComponent)t;
        setVisible(false);
        setEnabled(false);
    }

    /**
     * Returns the VisualContainer. Needed to be overridden because this component's parent is
     * not the correct container, but the visual root's parent. Actually, this component doesn't
     * have a container as this is not visualized. 
     * @return VisualContainer (VisualList) of the first element of the list
     */
    public VisualContainer getVisualContainer() {
        return (VisualContainer)visualRoot.getParent(); 
    }

    public void setVisualRoot(VisualListComponent vlc) {
	visualRoot = vlc;
    }

    /**
     * Handles the moving of the VisualListHeadReference, i.e., the reference pointing to the first item in the list.
     * @param r visualReference which is moved (in fact VisualListHeadReference)
     * @param oldTarget VisualType where the reference earlier pointed to 
     * @param newTarget VisualType where the reference should now point to
     */
    public void moveReference(VisualReference r, VisualListComponent oldTarget, VisualListComponent newTarget) {
	// delay execution until validateAll invoked
	Validator.add(new Command(this, newTarget) {
		public void execute() {
		    VisualListHead vlh = (VisualListHead)parameter;
		    VisualList visualList = (VisualList)vlh.getVisualContainer();
		    VisualListComponent visualTarget = (VisualListComponent)parameter2;
		    LinkedList target = (LinkedList)visualTarget.getStructure();
		    if (target == null) 
			target = new matrix.structures.FDT.probe.LinkedListImpl(null);
		    visualList.setList(target);
		}
	    });
	Validator.add(new Command(r) {
		public void execute() {
		    ((VisualType)parameter).setVisited(false);
		}
	    });
	r.setVisited(true);
    }

    /**
     * Paint method.
     * This method does nothing because VisualListHead should not be painted at all.
     */
    public void paintComponent(Graphics g) {}
    
}
