/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package matrix.visual;

import java.awt.*;
import javax.swing.JPopupMenu;

import matrix.structures.FDT.FDT;
import matrix.util.Note;
import matrix.util.export.ConfigurableGraphics;
import matrix.util.export.MatrixGraphics;

/** VisualMenuHotSpot is an abstraction for popUpMenu visualization
 * @see VisualContainer
 *
 * @author Ari Korhonen
 * @version $Revision: 1.14 $
 */
public class VisualMenuHotSpot extends VisualHotSpot {

    // --- constructors ---

    /** Note: The parent of this visual component should always be another visual component. */
    public VisualMenuHotSpot(FDT fdt) {
        super(fdt);
        setLabel("");
        setToolTipText("Click to see the menu");
    }
    
    protected void removePopupMenu() {
	// do nothing because we share the visual hot spot created by visual container
    }

    protected JPopupMenu getPopUp() {
	return ((VisualType)getParent()).getPopUp();
    }

    protected void setPopUp(JPopupMenu pm) {
	Note.err(this, "Cannot change popUp for " + this);
    }

    public void selectEntity(boolean isShiftDown) {
        if (getPopUp() != null && popUpEnabled()) {
            getPopUp().show(this, -10, -10);
        }
    }

    /* The following three methods make hot spots imitate the configuration of
       their parent objects when the parent object has overridden these
       methods. */

    /**
     * Is drag enabled for this object?
     * @return Is drag enabled for the parent object?
     */
    protected boolean dragEnabled() {
	try {
	    return ((VisualType)getParent()).dragEnabled();
	} catch(Exception e) {
	    return false;
	}
    }

    /**
     * Is pop-up menu enabled for this object?
     * @return Is pop-up menu enabled for the parent object?
     */
    protected boolean popUpEnabled() {
	try {
	    return ((VisualType)getParent()).popUpEnabled();
	} catch(Exception e) {
	    return false;
	}
    }

    /**
     * Is activation enabled for this object?
     * @return Is activation enabled for the parent object?
     */
    protected boolean activationEnabled() {
	try {
	    return ((VisualType)getParent()).popUpEnabled();
	} catch(Exception e) {
	    return false;
	}
    }

    public void setInvalid() {
	
    }
    
    public Dimension getPreferredSize() {
        int h = 20;
        Font f = getFont();
        if (f != null)
            h = getFontMetrics(f).getHeight();
        return new Dimension((int)Math.ceil(18*(f.getSize()/16.0)), h);
    }
    
    public void paintComponent(Graphics g) {
    	if (g instanceof MatrixGraphics)
    		return;
        if (g instanceof ConfigurableGraphics) {
            if (((ConfigurableGraphics)g).isDisabled("StructureBorder")) {
                //if exporting and structure borders are not wanted then stop here
                return;
            }
        }
        Rectangle r = getBounds();
        setBorderColor(g);
        g.fillRect(0, 0, r.width, r.height - 1);
        if (g instanceof ConfigurableGraphics) {
        // If g is ConfigurableGraphics, we have to check whether the hotspots are 
        // wanted to be painted or not.
        //
            if (((ConfigurableGraphics)g).isDisabled("VisualHotSpot")) {
                super.paintComponent(g);
                return;            
            }
        }
        ((Graphics2D)g).scale(getFont().getSize()/16.0, getFont().getSize()/16.0);
        if (hasFocus()) {
            g.setColor(opaqueWhite);
            g.fillRoundRect(1, 2, 15, 14, 1, 1);
            g.setColor(opaqueWhite.darker().darker());
            g.drawRoundRect(1, 2, 14, 13, 1, 1);
            g.setColor(g.getColor().darker().darker());
            g.drawRoundRect(1, 2, 15, 14, 1, 1);
            
        }
        g.setColor(Color.white);
        g.fillPolygon(new int[] {3, 13, 8}, new int[] {6, 6, 12}, 3);
       //super.paintComponent(g);
        ((Graphics2D)g).scale(1, 1);
    }
}
