/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2006 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package matrix.visual;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Composite;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Insets;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.font.FontRenderContext;
import java.awt.font.LineBreakMeasurer;
import java.awt.font.TextAttribute;
import java.awt.font.TextLayout;
import java.awt.geom.Area;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.geom.RoundRectangle2D;
import java.text.AttributedCharacterIterator;
import java.text.AttributedString;
import java.util.StringTokenizer;

import javax.swing.JComponent;

import matrix.structures.FDT.FDT;
import matrix.structures.code.AnnotatedPseudocode;

/**
 * This class is used to visualize both pseudocode and
 * the current location within it.
 *
 * The class holding the data to be visualized is matrix.structures.code.AnnotatedPseudocode.
 *
 * @author Otto Seppl
 * @version $Revision: 1.4 $
 * 
 * @see matrix.structures.code.AnnotatedPseudocode
 * @see matrix.structures.code.probe.PseudoCode
 */

public class VisualPseudocode extends VisualContainer{

    private AnnotatedPseudocode code;
    private CodePanel codePanel;
    private Dimension layoutSize = new Dimension(100,100);
    
    private int previousLocation = -1;
    
    public VisualPseudocode(FDT t) {
        super(t);
        
        if (t instanceof AnnotatedPseudocode == false)
            return;

        code = (AnnotatedPseudocode)t;
        
        Parser p = new Parser();
        
        p.parse(code.getCodeLines());
        
        codePanel = new CodePanel(
            p.getDeannotatedCode(),
            p.getStartLines(),
            p.getStartPositions(),
            p.getEndLines(),
            p.getEndPositions());
        
        add(codePanel);
    }

    public void assign() {
        //Nothing here - no support for simulation yet
    }

    public void validateComponents() {
        if (previousLocation != code.getLocation()){
            previousLocation = code.getLocation();

            codePanel.setSelectedBlock(previousLocation);
            codePanel.setErrorState(code.getErrorState());
        }
        
    }

    public void paintComponent(Graphics g) {
        super.paintComponent(g);
        if (isPicked())
            return; // no frames
          
        codePanel.paintComponent(g);
    }

    public Dimension getPreferredSize(){
        if (getGraphics() != null /*&& sizeDetermined == false*/){
            codePanel.setAdditionalInsets(getInsets());
            codePanel.getDimensions(getGraphics());
        }
        
        return layoutSize;
    }

    protected LayoutSize doActualLayout() {        
        if (getGraphics() != null /*&& sizeDetermined == false*/){
            codePanel.setAdditionalInsets(getInsets());
            codePanel.getDimensions(getGraphics());
        }

        return new LayoutSize((int)layoutSize.getWidth(), (int)layoutSize.getHeight());
    }

    public String getName(){
        if (code.getCodeName() == null)
            return super.getName();
        else
            return "Pseudo-code : "+code.getCodeName();        
    }
    
    
    class CodePanel extends JComponent{

        private int selection = 0;
        private String[] visualizedCode;
        private int[] startLines; 
        private int[] startLocations;
        private int[] endLines;
        private int[] endLocations;
        
        private Insets additionalInsets;
        
        private boolean errorState = false;

        
        public CodePanel(
                String[] visualizedCode,
                int[] startLines, 
                int[] startLocations,
                int[] endLines,
                int[] endLocations){
        
            
            this.visualizedCode = visualizedCode;
            this.startLines     = startLines;
            this.startLocations = startLocations;
            
            this.endLines       = endLines;
            this.endLocations   = endLocations;
            
            this.selection      = 0;
            
            additionalInsets = new Insets(0,0,0,0);
            
        }
        
        public void setAdditionalInsets(Insets insets){
            additionalInsets = insets;
        }
        
        public void setSelectedBlock(int selection){
            this.selection = selection;
        }
        
        public void setErrorState(boolean errorState){
            this.errorState = errorState;
        }

        int desiredWidth = 1024;

        public void getDimensions(Graphics g){
            paintComponent(g, false);
            
        }
        
        public void paintComponent(Graphics g){
            paintComponent(g, true);
        }
        
        //TODO: rivitys sotkee valinnat, sisennyskn ei silloin toimine.
        //tm kuitenkin tapahtuu vasta kun koodin leveys ylitt sdetyn maksimileveyden
        //joka on nykyiselln 1024 pistett.
        
        public void paintComponent(Graphics gOld, boolean performPaints){
            Rectangle overallBounds = new Rectangle(0,0,0,0);
            
            Graphics2D g = (Graphics2D)gOld;
            int selectionRadius = 4;            
            Object antialiasOriginal = null;
            
            if (performPaints){
                
                antialiasOriginal =  g.getRenderingHint(RenderingHints.KEY_ANTIALIASING);
                g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);

                Color original = g.getColor();
                
                g.setColor(Color.WHITE);
                
                g.fillRect(1,1,layoutSize.width-2, layoutSize.height-2);
                g.setColor(original);
            }
            
            //Starting point
            Point2D topLeft = 
                new Point2D.Double(
                    additionalInsets.left + selectionRadius,
                    additionalInsets.top  + selectionRadius);
            
            FontRenderContext frc = g.getFontRenderContext();

            Area a = new Area();
            
            int vertical = 0;            
            
            for (int i=0; i<visualizedCode.length; i++){
                
                //Forcing more indentation
                int tabbing = 10 * (visualizedCode[i].length() - visualizedCode[i].trim().length());

                AttributedString as = new AttributedString(visualizedCode[i]+" ");
                as.addAttribute(TextAttribute.FAMILY, "sansserif");
                //If font size is greater than 24 some computations throw exceptions
                as.addAttribute(TextAttribute.SIZE, new Float(Math.min(24, g.getFont().getSize())));                
                //TODO: when font is changed the pseudo code area's new size is not updated correctly on the first redraw
                
                int lineLength = (visualizedCode[i]+" ").length();

                AttributedCharacterIterator aci = as.getIterator();
                LineBreakMeasurer lineMeasurer = new LineBreakMeasurer(aci, frc);            

                Rectangle2D bounds = null;
                int startVertical = 0;
                
                while (lineMeasurer.getPosition() < aci.getEndIndex()){
                    TextLayout layout = lineMeasurer.nextLayout(desiredWidth - tabbing);

                    overallBounds.add(new Rectangle2D.Double(
                        (double)topLeft.getX() + tabbing,
                        (double)topLeft.getY() + vertical,
                        layout.getBounds().getWidth(),
                        layout.getBounds().getHeight()));
                    
                    vertical += layout.getAscent();
     
                    if (startVertical == 0)
                        startVertical = vertical;
                    
                    if (performPaints)
                        layout.draw(g, (float)topLeft.getX() + tabbing, (float)topLeft.getY() + vertical);
                    
                    Rectangle2D bounds2 = null;
                    
                    if (startLines[selection] == i){
                        if (endLines[selection] == i){
                            bounds2 = layout.getLogicalHighlightShape(startLocations[selection], endLocations[selection]).getBounds2D();
                        }
                        else{
                            bounds2 = layout.getLogicalHighlightShape(startLocations[selection], lineLength).getBounds();
                        }
                    }else{
                        if (startLines[selection] < i){
                            if (endLines[selection] > i){
                        
                                //Kokonaisia rivej
                                Rectangle2D bnds = layout.getBounds();
                        
                                bounds2 = new Rectangle2D.Double(
                                    bnds.getMinX(), bnds.getMinY()+vertical-startVertical,
                                    bnds.getWidth(), bnds.getHeight());
                            }
                            else if(endLines[selection] == i){
                                //osittainen rivi
                                Rectangle2D bnds = layout.getLogicalHighlightShape(0, endLocations[selection]).getBounds();
                        
                                bounds2 = new Rectangle2D.Double(
                                    bnds.getMinX(), bnds.getMinY()+vertical-startVertical,
                                    bnds.getWidth(), bnds.getHeight());
                            }
                        }
                    }
                    
                    if (bounds == null)
                        bounds = bounds2;
                    else
                        bounds.add(bounds2);
                    
                    vertical += layout.getDescent() + layout.getLeading();
                }

                if (bounds != null){
                                    
                    RoundRectangle2D nbounds = new RoundRectangle2D.Double();
                    nbounds.setRoundRect(
                        bounds.getX()+topLeft.getX() - selectionRadius + tabbing,
                        bounds.getY()+topLeft.getY() - selectionRadius + startVertical,
                        bounds.getWidth()  + 2*selectionRadius,
                        bounds.getHeight() + 2*selectionRadius,
                        2*selectionRadius,2*selectionRadius);
                  
                    a.add(new Area(nbounds));
                }
            }

            if (performPaints){
                Composite normalComp = g.getComposite();

                if (errorState){
                    g.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 0.6f));

                    g.setColor(Color.RED);                    
                }
                else{
                    g.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 0.4f));

                    g.setColor(Color.ORANGE);
                }
                g.fill(a);
            
                g.setColor(Color.BLACK);            
                g.draw(a);
                
                g.setComposite(normalComp);
                g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, antialiasOriginal);
            }
                
            //Add space for the selection shape. This approach avoids constantly changing the size
            //of the component when the selection changes.
            
            overallBounds.setRect(overallBounds.getMinX(), overallBounds.getMinY(),
                overallBounds.getMaxX() + 2* selectionRadius + additionalInsets.right,
                overallBounds.getMaxY() + 2* selectionRadius + additionalInsets.bottom);
            
            overallBounds.add(a.getBounds());
            
            layoutSize.setSize(overallBounds.getWidth(), overallBounds.getHeight());
        }
        
        
    }

    
    /**
     * This parser class processes a piece of annotated code, and separates the annotation markings
     * from the code, creating five arrays. One holds the now de-annotated code lines and the others
     * hold the selection areas - starting and ending lines and locations within these lines.
     */
    
    class Parser{
        private int[] startLines, startPos, endLines, endPos;
        private String[] deannotatedCode;
        
        public int[] getStartLines(){
            return startLines;
        }
        
        public int[] getEndLines(){
            return endLines;
        }
        
        public int[] getStartPositions(){
            return startPos;
        }
        
        public int[] getEndPositions(){
            return endPos;
        }
        
        public String[] getDeannotatedCode(){
            return deannotatedCode;
        }
        
        public void parse(String[] code) {
                        
            int highID = -1;
            //Search for the highest id
            for (int i=0; i<code.length; i++){
                StringTokenizer st = new StringTokenizer(code[i],"_", true);
                boolean foundFirst = false;
                
                while (st.hasMoreElements()){
                    
                    String potentialStarts = st.nextToken();
                    
                    if (foundFirst == false && !potentialStarts.equals("_"))
                        continue;
                    
                    foundFirst = true;
                    if (potentialStarts.equals("_"))
                        continue;
                    
                    StringTokenizer st2 = new StringTokenizer(potentialStarts, "/, ", true);
                    while (st2.hasMoreElements()){
                        String ids = st2.nextToken();
                        if (ids.trim().equals(""))
                            break;
                        
                        if (ids.trim().equals(",") || ids.trim().equals("/"))
                            continue;
                        
                        int numericalID = Integer.parseInt(ids.trim());
                        if (highID < numericalID)
                            highID = numericalID;                    
                    }                
                }
            }
            
            if (highID == -1)
                return;
            
            //Reserve memory
            startLines = new int[highID+1];
            startPos   = new int[highID+1];
            endLines   = new int[highID+1];
            endPos     = new int[highID+1];
            
            deannotatedCode = new String[code.length];
            
            boolean readingEnd = false;
            
            for (int i=0; i<code.length; i++){
                StringTokenizer st = new StringTokenizer(code[i],"_", true);

                String line = "";
                int localPosition = 0;
                boolean first = false;
                
                while (st.hasMoreElements()){
                    
                    String potentialStarts = st.nextToken();
                    if (!first && !potentialStarts.equals("_")){
                        localPosition += potentialStarts.length();
                        line += potentialStarts;
                        first = true;
                        continue;
                    }
                    first = true;
                    if (potentialStarts.equals("_"))
                        continue;
                    
                    StringTokenizer st2 = new StringTokenizer(potentialStarts, "/, ", true);
                    while (st2.hasMoreElements()){
                        String ids = st2.nextToken();
                        if (ids.trim().equals("")){
                            while (st2.hasMoreElements()){
                                String part = st2.nextToken();
                                line += part;
                                localPosition += part.length();
                            }
                            break;
                        }
                        
                        if (ids.trim().equals(","))
                            continue;
                        
                        if (ids.trim().equals("/")){
                            readingEnd = true;
                            continue;
                        }
                        
                        int numericalID = Integer.parseInt(ids.trim());
                        
                        if (readingEnd){
                            
                            endLines[numericalID] = i;
                            endPos[numericalID] = localPosition;
                            
                        }else{
                            startLines[numericalID] = i;
                            startPos[numericalID] = localPosition;
                        }
                        
                        readingEnd = false;
                        
                        
                    }                
                }
                
                deannotatedCode[i] = line;
            }
            
            /* DEBUG 
            for (int i=0; i<startLines.length; i++){
             
                if (startLines[i] == endLines[i])
                    Note.out(deannotatedCode[startLines[i]].substring(startPos[i], endPos[i]));
                else{
                    Note.out(this, deannotatedCode[startLines[i]].substring(startPos[i]));
                    
                    for (int j=startLines[i]+1; j<endLines[i]-1; j++)
                        Note.out(this, deannotatedCode[j]);                    
                    
                    Note.out(this, deannotatedCode[endLines[i]].substring(0,endPos[i]));
                }
            }
            */
        }       
    }

    
}
