/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package matrix.visual;

import java.awt.*;

import matrix.structures.FDT.*;
import matrix.util.Note;
import matrix.util.export.ConfigurableGraphics;
import matrix.util.export.MatrixGraphics;

/** VisualReziseHotSpot is an abstraction for special active resize components within containers.
 * @see VisualContainer
 *
 * @author Ari Korhonen
 * @version $Revision: 1.12 $
 */
public class VisualResizeHotSpot extends VisualHotSpot {

    // --- constructors ---

    public VisualResizeHotSpot(FDT fdt) {
        super(fdt);
        setLabel("");
        setToolTipText("Drag to resize the structure");
    }

    public Array getArray() {
        if (getStructure() instanceof Array) {
            return (Array)getStructure();
        } else Note.show(this, "Cannot resize this kind of structure (not an Array)");
        return null;
    }

    private boolean even(int x) {
	return (x/2 == (x+1)/2); // division is truncated, thus this should work for all x >= 0
    }

    int counter = 0;
    public void selectEntity(boolean isShiftDown) {
	/* See task #347: select is not working correctly in all environments*/
	if (even(counter++) && !picked) { // we don't want to show this every time this hotspot is clicked by mistake
	    Note.show(this, "This operation is not defined. Drag this to resize the array.");
	}
	
    }

    public void assign() {
	Note.show(this, "This operation is not defined. Drag this hotspot to resize the array.");
    }

    int slotSize;
    int pick_x;
    public void pick(int x, int y) {
        picked = true;
        Array a = getArray();
        if (a != null) {
            slotSize = getBounds().width;
            pick_x = x;
        }
    }

    public void drag(int x, int y) {
        Array a = getArray();
        if ((a != null) && (slotSize > 0)) {
            if (x > pick_x + slotSize/2) {
                a.setLast(a.getLast()+2);
                setInvalid();
            }
            else if (x < pick_x) {
                if (a.getLast() - a.getFirst() > 0)
                    a.setLast(a.getLast()-1);
                setInvalid();
            }
        }
    }

    public void released() {
        picked = false;
        setInvalid();
    }
    
    public Dimension getPreferredSize() {
        int h = 20;
        Font f = getFont();
        if (f != null)
            h = getFontMetrics(f).getHeight();
        return new Dimension((int)Math.ceil(25*(f.getSize()/16.0)), h);
     }
    
    public void paintComponent(Graphics g) {
        if (g instanceof ConfigurableGraphics) {
            if (((ConfigurableGraphics) g).isDisabled("StructureBorder")) {
                // if exporting and structure borders are not wanted then stop
                // here
                return;
            }
        }
    	if (g instanceof MatrixGraphics)
    		return;
        Rectangle r = getBounds();
        setBorderColor(g);
        g.fillRect(0, 0, r.width, r.height - 1);
        if (g instanceof ConfigurableGraphics) {
            // If g is ConfigurableGraphics, we have to check whether the
            // hotspots are
            // wanted to be painted or not.
            //
            if (((ConfigurableGraphics) g).isDisabled("VisualHotSpot")) {
                super.paintComponent(g);
                return;
            }
        }
        ((Graphics2D) g).scale(getFont().getSize() / 16.0,
                               getFont().getSize() / 16.0);
        if (hasFocus()) {
            g.setColor(opaqueWhite);
            g.fillRoundRect(1, 1, 19, 14, 1, 1);
            g.setColor(opaqueWhite.darker().darker());
            g.drawRoundRect(1, 1, 18, 13, 1, 1);
            g.setColor(g.getColor().darker().darker());
            g.drawRoundRect(1, 1, 19, 14, 1, 1);

        }
        g.setColor(Color.white);
        g.fillPolygon(new int[] { 3, 8, 8 }, new int[] { 8, 3, 13 }, 3);
        g.fillPolygon(new int[] { 14, 19, 14 }, new int[] { 3, 8, 13 }, 3);
        g.drawLine(8, 8, 14, 8);
        ((Graphics2D) g).scale(1, 1);
    }
}
