package matrix.visual;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Insets;
import java.awt.Rectangle;
import java.awt.RenderingHints;

import matrix.decoration.StyleSheet;
import matrix.decoration.StyleSheetAdapter;
import matrix.structures.FDT.FDT;
import matrix.structures.FDT.Struct;
import matrix.structures.FDT.probe.Key;

public class VisualStruct extends VisualContainer{

    private static final long serialVersionUID = -3652972750113539105L;

    private VisualField[] children;
    private int commonLabelWidth = 0;
    private LayoutSize currentSize = new LayoutSize(50,50); //Dummy values before actual calculations
    public static final int VERTICAL_GAP = 1;
    private String title;
    
    public static final int TITLE_INSET = 1;
    
    public static final int max(int a, int b){
        return a >= b ? a : b;
    }

    public VisualStruct(Struct s) {
        super(s);
                
        title = s.getField(0).toString();
    }
    
    private StyleSheet defaultStyleSheet;
    
    public StyleSheet getDefaultStyleSheet() {
        if (defaultStyleSheet == null) {
            defaultStyleSheet = new StyleSheetAdapter();
            
            defaultStyleSheet.setDefaultPenColor(Color.black);
            defaultStyleSheet.setFocusedPenColor(Color.black);
            defaultStyleSheet.setSelectedPenColor(Color.BLUE);
            
            defaultStyleSheet.setFocusedBackgroundColor(Color.white);
            defaultStyleSheet.setDefaultBackgroundColor(Color.white);
            defaultStyleSheet.setSelectedBackgroundColor(Color.white);
            
            defaultStyleSheet.setDefaultBorderColor(Color.white);
            defaultStyleSheet.setFocusedBorderColor(Color.BLUE);
            defaultStyleSheet.setSelectedBorderColor(Color.GREEN);

        }
        return defaultStyleSheet;
    }
    
    public boolean isTitled(){
        return false;
    }
    
    public Insets getInsets() {

            return new Insets(2, 2, 2, 2);
    }

    /**
     * Insert method for Visual Struct. Since it does not make any sense
     * to insert anything to a struct (you should insert something to a part of the 
     * struct instead, this method will defer the insert to the parent object, if it
     * is a VisualType.
     */
    public void insert(Object structure) {
    	if(getParent() instanceof VisualType) {
    		((VisualType)getParent()).write();
    	}
    }
    
    protected LayoutSize doActualLayout() {
             
        // Estimate maximum widths for labels & values
        int maxLabelWidth = 0;
        int maxValueWidth = 0;
        int totalHeight   = 0;
        
        int titleHeight = getMetrics().getHeight() + 2*TITLE_INSET;
        int titleWidth  = getMetrics().stringWidth(title);

        Insets in = getInsets();
        
        if (isMinimized()){
            currentSize = new LayoutSize(titleWidth + 4 + in.left+in.right, titleHeight + in.top+in.bottom);
            return currentSize;
        }
        
        if (title.trim().equals("")){
            titleHeight = 0;
            titleWidth = 0;
        }
        
        
        totalHeight += titleHeight;
        
        for (int i=0; i<children.length; i++){
            maxLabelWidth = max(maxLabelWidth, children[i].getLabelWidth());
            maxValueWidth = max(maxValueWidth, children[i].getValueWidth());
            totalHeight  += children[i].getFieldHeight() + VERTICAL_GAP;
        }

        // needed by visualField to make all label widths equal
        commonLabelWidth = maxLabelWidth;
        
        // set the bounds for each component

        int commonFullWidth = maxLabelWidth + maxValueWidth + VisualField.GAP + children[0].getInsets().left + children[0].getInsets().left;
        
        if (commonFullWidth < titleWidth)
            commonFullWidth = titleWidth + children[0].getInsets().left + children[0].getInsets().left;
        
        int y = in.top + titleHeight;
        for (int i=0; i<children.length; i++){
            int height = children[i].getFieldHeight();
            children[i].lSetBounds(in.left, y,  commonFullWidth, height);
            y += height + VERTICAL_GAP;
        }
        currentSize = new LayoutSize(commonFullWidth + in.left+in.right, totalHeight+in.top+in.bottom);
        return currentSize;
    }

    public Dimension getPreferredSize(){
        return new Dimension(currentSize.width, currentSize.height);
    }
    
    // Intentionally empty method. The default paintFrame
    // is executed at wrong times and paints over the component
    public void paintFrame(Graphics g2){
    }
    
    public void paintComponent(Graphics g){
                
        Object antialiasState = null;
        if (g instanceof Graphics2D){
            antialiasState = ((Graphics2D)g).getRenderingHint(RenderingHints.KEY_ANTIALIASING);
            ((Graphics2D)g).setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        }
        
        Insets in = getInsets();
        int titleWidth  = getMetrics().stringWidth(title);
        int titleHeight = getMetrics().getAscent() + TITLE_INSET;

        if (title.trim().equals("")){
            titleHeight = 0;
            titleWidth = 0;
        }
        
        int titleOffset = (getWidth() - titleWidth) / 2;        
        
        setBackgroundColor(g);
        
        g.fillRoundRect(2,2,getWidth()-3,getHeight()-3,5,5);
        
        setPenColor(g);        
        g.drawString(title, titleOffset, in.top + titleHeight);

        setBorderColor(g);
        g.drawRoundRect(3,3,getWidth()-6,getHeight()-6,5,5);
                
        if (g instanceof Graphics2D){
            ((Graphics2D)g).setRenderingHint(RenderingHints.KEY_ANTIALIASING, antialiasState);
        }
    }
        
    
    public void validateComponents(){
        super.validateComponents();
        
        
        Struct structure = (Struct)getStructure(); 
        if (structure == null){
            // null structure
        }else{
            if (children == null){
                children = new VisualField[structure.getFieldNames().length - 1];
                for (int i=0; i<children.length; i++){  //Starts from 1 because 0 is id
                    VisualField child = new VisualField(
                        structure.getFieldName(i+1), 
                        structure.getField(i+1));
                    
                    children[i]= child;
                    add(child);
                        
                }
            }
            
            repaint();
            
            for (int i=0; i<children.length; i++){
                
                children[i].updateValue(structure.getField(i+1));
                
                children[i].validate();
                children[i].setVisible( !isMinimized() );
                children[i].repaint();
            }
        }
            
    }
    
    
    public class VisualField extends VisualComponent{
        
        private static final long serialVersionUID = -3648045982946034652L;

        private Object value;
        private String title;
        private VisualType valueVisualization;
        
        public static final int GAP = 4;
        
        public VisualField(String title, Object value){
            super(/*value instanceof FDT ? (FDT)value : */new Key(value)); // For some reason, not wrapping the value in a key makes visualization erradic
            
            this.title = title;
            this.value = value;
                   
        }

      /*  public boolean isDoubleBuffered(){
            return true;
        }*/

        public void updateValue(Object value){
            if (!this.value.equals(value)){
                remove(valueVisualization);
                this.value = value;
                addComponents();
            }
        }
        
        public void addComponents() {
            if (getComponentCount() == 0) {
                valueVisualization = getRepresentation(value);             
                add(valueVisualization);
            }
        }
        
        public Insets getInsets(){
            return new Insets(1,1,1,1);
        }
        
        public void assign() { 
            write(); 
        }
       
        public void doLayout(){
            Dimension insides = valueVisualization.getPreferredSize();
            Insets insets    = getInsets();
            
            valueVisualization.setBounds(commonLabelWidth + insets.left + GAP/2, insets.top, insides.width, insides.height);
        }
        
        public FDT getKey(){
            if (value instanceof FDT)
                return (FDT)value;
            else
                return new Key(value);
        }

        public VisualType getVisualType() {            
            return valueVisualization;
        }
        
        public int getLabelWidth(){
            return getMetrics().stringWidth(title);
        }
        
        public int getValueWidth(){
            return valueVisualization.getPreferredSize().width;
        }
        
        public int getFieldHeight(){
            Insets insets    = getInsets();

            return insets.top + insets.bottom + max(
                getMetrics().getMaxAscent() + getMetrics().getMaxDescent(),
                valueVisualization.getPreferredSize().height);
        }
        
        public Dimension getPreferredSize(){
            Insets insets    = getInsets();
            return new Dimension(getLabelWidth() + getValueWidth() + GAP + insets.left + insets.right, getFieldHeight()+insets.top+insets.right);
        }
        
        public Dimension getMinimumSize(){
            return getPreferredSize();
        }
        
        public void paintComponent(Graphics g) {           
            Insets insets    = getInsets();

            Rectangle bounds = getBounds();
            
            //Use the stylesheet through VisualStruct
            VisualStruct.this.setBackgroundColor(g);
            
            //Debug draw : box around the field
            //g.drawRect(0,0,bounds.width-1, bounds.height-1);
            
            int x = insets.left;
            int h = bounds.height-insets.top-insets.bottom;
            
            VisualStruct.this.setPenColor(g);
                        
            g.drawString(title, x, h - (h-getMetrics().getAscent())/2);
            
            
        }  
        
        public void validateComponents(){
            
        }

    }
    
}
