/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: VisualTreeComponent.java,v 1.54 2007/09/03 12:25:23 oseppala Exp $ */

package matrix.visual;

import java.awt.*;
import java.io.PrintStream;
import java.util.*;

import matrix.structures.CDT.CDT;
import matrix.structures.FDT.*;
import matrix.util.*;

/** This abstraction is used to represent the nodes of an abstract tree.
 * A subclass of the VisualTree could have its own concrete subclass of
 * the VisualTreeComponent in order to have a special look and feel of nodes.
 * @see VisualTree
 *
 * @author Ari Korhonen
 * @version $Revision: 1.54 $
 */
public class VisualTreeComponent extends VisualNode {
    //protected static Tree empty; // empty tree
    private Tree tree; // the node which this component represents
    private VisualTreeComponent father; // father node representation
    //private int child; // this is parent's #child; 0 = first child
    //private Object component;
    private Vector children; // children of this node

    // debug
    public void dump(PrintStream out) {
        super.dump(out);
        out.println("children("+children.size()+")="+children);
        out.println("father="+father);
    }

    // constructor
    public VisualTreeComponent(Tree t, VisualTreeComponent father) {
        super(t);   
        this.tree = t;
        setFather(father);
        children = new Vector();
    }

    // --- end of constructors ---
    /*
    public void addComponents() {
        if (getComponentCount() == 0)
            add(getRepresentation(tree != null ? tree.getElement() : null));
    }
    */
    public VisualTree getTreeContainer() {
        return (VisualTree)getVisualContainer();
    }

    /** Set the specified component to be this components father.
     * The old father is informed that it's not anymore the father
     * of this component and the new father is informed that it has
     * received a new child. 
     */
    protected void setFather(VisualTreeComponent newFather) {
        if (father != null) father.removeChild(this);
        if (newFather != null) newFather.addChild(this);
        father = newFather;
    }
    
    /** Return the representation of this component's father. */
    protected VisualTreeComponent getFather() { return father; }

    /** Determines whether this node has the specified predecessor.
     */
    protected boolean hasPredecessor(Tree predecessor) {
        if (getStructure() == null) return (predecessor == null);
        if (getStructure().equals(predecessor)) return true;
	if (predecessor == null) return false;
        VisualTreeComponent fatherComponent = getFather();
        if (fatherComponent == null) return false; // root encountered
        return fatherComponent.hasPredecessor(predecessor);
    }
    
    /** Add the specified component to be one of this components child. */
    private void addChild(VisualTreeComponent child) {
        children.addElement(child);
    }
    
    /** Remove the specified component from this components children. */
    private void removeChild(VisualTreeComponent child) {
        children.removeElement(child);
    }
    
    /** Returns an array of the children. */
    protected VisualTreeComponent[] getChildren() {
        VisualTreeComponent[] vtc = new VisualTreeComponent[children.size()];
        children.copyInto(vtc);
        return vtc;
    }

    /** Return the position of the specified component within children. 
     * @return the index of the child; or -1 if the child was not found 
     */
    public int getReferenceNumber(VisualType to) {
        Enumeration e = children.elements();
        int i = 0;
        while (e.hasMoreElements())
            if (to == e.nextElement())
                return i;
            else
                i++;
        return -1;
    }

    // simulation

    /** This method should be invoked whenever a new key is attached into this node. 
	The key is changed for this structure.
     */
    public void write() {
	Note.out(this, "WRITE "+tree);
        if (read != null) {
	    /*
            if ((read.getStructure() != null) && 
		(read.getStructure().equals(this.getStructure()))) { 
                getApplication().showMessage("Sorry", "Component cannot be it's own key");
                return;
            }
	    Note.out(this, "read = " + read.getStructure());
	    Note.out(this, "this = " + this.getStructure());
	    */
            if (!empty())
                tree.setElement(read.getStructure()); 
	    else {
                if (father == null) { // root node
                    if (tree != null) tree.setElement(read.getStructure()); 
                    else getApplication().showMessage("Error", "No root found ");
                } else { // father != null
                    FDT f = father.getStructure();
                    if (f instanceof Tree) { 
                        // only dynamic trees can expand
                        Tree t = (Tree)f;//father.getStructure();
                        t.setSubTree(t.getNewNode(read.getStructure()), 
                                     father.getReferenceNumber(this));
                    } else if (f == null) {// new root
                        tree.setElement(read.getStructure()); 
                    } else if (tree != null) { // static tree
			//getApplication().showMessage("Error", "Unknown error in write()");
			tree.setElement(read.getStructure()); 
		    } else getApplication().showMessage("Error", "Missing underlying structure (tree == null)");
                }
            }
        } else getApplication().showMessage("Sorry", "Nothing to write");
        setInvalid();
	/*invalidate();*/
    }

    public void selectEntity(boolean isShiftDown) {
	if (getStructure() instanceof matrix.structures.simulationextensions.Selectable) {
	    getAnimator().startOperation();
        //Default behavior, only an animated version of it...
	    super.selectEntity(isShiftDown);
        getAnimator().endOperation();
	}	
	setInvalid();
    }


    /** This method should be invoked whenever a new structure is inserted into this node. 
	The structure is added into this structure if possible.
     */
    public void assign() {
	//write();
	Note.out(this, "ASSIGN");
        if (read instanceof VisualTreeComponent) {
            if (father == null) {
		// here root changes. 
                // If possible, the changing the root is put in the animator.
		if(getVisualContainer() instanceof VisualTree) {
		    try {
			((VisualTree)getVisualContainer()).setRoot((Tree)read.getStructure());
		    } catch(ClassCastException e) {
			getVisualContainer().setStructure(read.getStructure());		
		    }
		}
		else
		    getVisualContainer().setStructure(read.getStructure());		
            } else if (father.getStructure() instanceof Tree) {
                Tree f = (Tree)father.getStructure();
                // only dynamic trees can expand
                    f.setSubTree((Tree)read.getStructure(),
                                 father.getReferenceNumber(this));
            } else {
                Note.err(this, "Illegal structure " + 
                         father.getStructure() + " of " + father);             
                getApplication().showMessage("Error", "Illegal structure " + 
                          father.getStructure() + " of " + father);                
            }           
        } else {
            if (!empty() && (tree instanceof Tree)) {
		if(tree instanceof CommonTree) {
			Tree t = ((Tree)tree).getNewNode(read.getStructure());
//			((SimulationTree)tree).setSubTree(((SimulationTree)tree).getNewNode(read.getStructure()), tree.getSubTreeCount());
			((Tree)tree).setSubTree(t, tree.getSubTreeCount());
			if(getTreeContainer().getStructure() instanceof Graph)
				((Graph)getTreeContainer().getStructure()).addVertex(t);
		}
		else
		    tree.setElement(read.getStructure()); 
	    }
            else write();
        }
    }

    public void erase() {
        if (father == null /*|| father.getStructure() instanceof Head*/) {
	    Tree t = (Tree) getStructure();
		int n = t.getSubTreeCount();
		t.setElement(null);
		for(int i = 0; i < n; i++) 
		    t.setSubTree(null, i);
		setInvalid();

        } else if (father.getStructure() instanceof Tree) {
            Tree f = (Tree)father.getStructure();
            f.setSubTree(null, father.getReferenceNumber(this));
            setInvalid();
        }
    }

    protected void removeKeyHelp(Object o) {
      if(!(getStructure() instanceof CDT)) {
	Component c = this;
	while(c.getParent() instanceof VisualType)
	  c = c.getParent();
	if(((VisualType)c).getStructure() instanceof CDT) {
	  super.removeKeyHelp(o);
	  return;
	}
	FDT tree = getStructure();
	if(tree instanceof Tree) 
	  tree.setElement(null);
	else
	  Note.show(this, "Cannot delete key, not a Dynamic Tree");
	return;
      }
      else
	super.removeKeyHelp(o);
    }	    


    public void remove() {
        tree.setElement(null);
    }

    /** Move the specified reference to point to another component.
     * @param r the reference
     * @param oldTarget the old target
     * @param newTarget the new target
     */
    public void moveReference(VisualReference r, 
                              VisualType oldTarget, 
                              VisualType newTarget) {
	//Note.out(this, "moveReference");
        Container target = newTarget;
        
        while ((target instanceof VisualType) && 
	       (!(target instanceof VisualTreeComponent)))
            target = target.getParent();
        
        if (target instanceof VisualTreeComponent) {
            VisualTreeComponent nTarget = (VisualTreeComponent)target;
            VisualTreeComponent oTarget = (VisualTreeComponent)oldTarget;
            Tree f = (Tree)oTarget.father.getStructure();
                // delay execution until validateAll invoked
                // set subtree as follows:
                // the old target asks from its father which
                // (number n) child it is and the command replaces
                // the n'th child with the new one after
                // the references are validated
                int index = oTarget.getFather().getReferenceNumber(oTarget);
                Validator.add(new Command(f,
                                          nTarget.getStructure(),
                                          index) {
                        public void execute() {
                            Tree father = (Tree)parameter;
                            Tree tree = (Tree)parameter2;
                            int i = parameter3;
                            father.setSubTree(tree, i);
                        }
                    });
                Validator.add(new Command(r) {
                        public void execute() {
                            ((VisualType)parameter).setVisited(false);
                        }
                    });
                r.setVisited(true);
        } else {
            getApplication().showMessage("Sorry", "Cannot move to " + 
					 ((target != null) ? target.getName() : "null"));
        }
        
    }

    public boolean isHiding() {
        if (!empty()) return false;
	if (getTreeContainer() == null) {
	    Note.warning(this, "No container present");
	    return true;
	}
        return !getTreeContainer().isEmptySubTreesEnabled();
    }
}
