// MatrixPro - algorithm simulation system
// Copyright (C) 2000-2007 See the AUTHORS file and the author tag
// below for more information on the authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package matrixpro.animation;

import java.awt.EventQueue;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.lang.reflect.InvocationTargetException;

import javax.swing.Icon;
import javax.swing.JButton;

import matrix.animation.VisualAnimator;

import matrixpro.util.FileHelper;

/**
 * VisualAnimator for the MatrixPro. Contains small buttons with pictures for
 * the animation control commands (begin, backward, play, forward, end).
 * 
 * @version $Id: SmallVisualAnimator.java,v 1.12 2007/09/03 13:46:13 vkaravir
 *          Exp $
 */
public class SmallVisualAnimator extends VisualAnimator implements
        ActionListener {
    /** The buttons. */
    private JButton undo, redo, begin, end, play;

    /** Time to wait in one step when playing the animation. */
    private int waitTime = 500;

    private Icon playIcon, stopIcon;

    private boolean playing = false;

    /**
     * Constructs a new SmallVisualAnimator visualizing the given Animator.
     */
    public SmallVisualAnimator(CountingAnimator a) {
        super(a);
    }

    /** Add the components. */
    protected void layoutComponents() {
        setLayout(new GridBagLayout());

        GridBagConstraints gbc = new GridBagConstraints();
        gbc.fill = GridBagConstraints.WEST;
        gbc.gridx = 0;
        gbc.gridy = 0;
        gbc.weightx = 0.0;

        begin = FileHelper.getJButton("/pics/Rewind24.gif", "Begin");
        begin.setToolTipText("Begin");
        begin.setMargin(new Insets(3, 6, 3, 6));
        begin.setBackground(getBackground());
        undo = FileHelper.getJButton("/pics/StepBack24.gif", "Backward");
        undo.setToolTipText("Backward");
        undo.setMargin(new Insets(3, 6, 3, 6));
        undo.setBackground(getBackground());
        redo = FileHelper.getJButton("/pics/StepForward24.gif", "Forward");
        redo.setToolTipText("Forward");
        redo.setMargin(new Insets(3, 6, 3, 6));
        redo.setBackground(getBackground());
        end = FileHelper.getJButton("/pics/FastForward24.gif", "End");
        end.setToolTipText("End");
        end.setMargin(new Insets(3, 6, 3, 6));
        end.setBackground(getBackground());
        play = FileHelper.getJButton("/pics/Play24.gif", "Play");
        play.setToolTipText("Play");
        play.setMargin(new Insets(3, 6, 3, 6));
        play.setBackground(getBackground());
        playIcon = play.getIcon();
        stopIcon = FileHelper.getJButton("/pics/Stop24.gif", "Stop").getIcon();

        add(begin, gbc);
        gbc.gridx = 1;
        add(undo, gbc);
        gbc.gridx = 2;
        add(play, gbc);
        gbc.gridx = 3;
        add(redo, gbc);
        gbc.gridx = 4;
        add(end, gbc);

        gbc.gridy = 1;
        gbc.gridx = 0;
        gbc.gridwidth = GridBagConstraints.REMAINDER;
        undo.setActionCommand("undo");
        undo.addActionListener(this);
        redo.setActionCommand("redo");
        redo.addActionListener(this);
        begin.setActionCommand("begin");
        begin.addActionListener(this);
        end.setActionCommand("end");
        end.addActionListener(this);
        play.addActionListener(new ActionListener() {
            public void actionPerformed(final ActionEvent e) {
                if (playing) {
                    playing = false;
                    play.setIcon(playIcon);
                    play.setToolTipText("Play");
                    if (animator.hasPreviousOperation())
                        animator.undo();
                    return;
                }
                playing = true;
                play.setIcon(stopIcon);
                play.setToolTipText("Stop");
                new Thread() {
                    public void run() {
                        while (animator.hasNextOperation() && playing) {
                            try {
                                Thread.sleep(waitTime);
                                try {
                                    EventQueue.invokeAndWait(new Runnable() {
                                        public void run() {
                                            animator.redo();
                                            fireActionEvent((ActionEvent) e);
                                        }
                                    });
                                } catch (InvocationTargetException ite) {
                                    ite.getTargetException().printStackTrace();
                                }
                            } catch (InterruptedException ie) {
                            }
                        }
                        if (!animator.hasNextOperation()) {
                            playing = false;
                            play.setIcon(playIcon);
                            play.setToolTipText("Play");
                        }
                    }
                }.start();
            }
        });
        validate();
    }

    /** Validates this SmallVisualAnimator. */
    public void validate() {
        super.validate();
        if (play == null || redo == null || end == null || undo == null
                || begin == null)
            return;
        boolean hasNext = animator.hasNextOperation();
        play.setEnabled(hasNext);
        redo.setEnabled(hasNext);
        end.setEnabled(hasNext);
        boolean hasPrev = animator.hasPreviousOperation();
        undo.setEnabled(hasPrev);
        begin.setEnabled(hasPrev);
    }

    /** Overridden because no reset button is in use. Does nothing. */
    public void setResetButtonVisible(boolean value) {
    }

    /**
     * Sets the play button either visible on invisible depending on the given
     * value.
     */
    public void setPlayButtonVisible(boolean value) {
        play.setVisible(value);
    }

    /**
     * Sets the time to wait in one step when playing the animation.
     */
    public void setWaitTime(int time) {
        waitTime = time;
    }

    public void actionPerformed(ActionEvent event) {
        String command = event.getActionCommand();
        if (command.equals("undo")) {
            if ((event.getModifiers() & ActionEvent.SHIFT_MASK) != 0)
                GoTo(animator.getState() - 1, animator.getMaximum());
            else
                undo();
            fireActionEvent(event);
        } else if (command.equals("redo")) {
            if ((event.getModifiers() & ActionEvent.SHIFT_MASK) != 0)
                GoTo(animator.getState() + 1, animator.getMaximum());
            else
                redo();
            fireActionEvent(event);
        } else if (command.equals("end")) {
            end();
            fireActionEvent(event);
        } else if (command.equals("begin")) {
            rewind();
            fireActionEvent(event);
        }
    }
}
