// MatrixPro - algorithm simulation system
// Copyright (C) 2000-2007 See the AUTHORS file and the author tag
// below for more information on the authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/* $Id: MatrixProAVLTree.java,v 1.14 2007/09/10 08:32:32 vkaravir Exp $ */

package matrixpro.structures;

import matrix.decoration.StyleSheet;
import matrix.decoration.StyleSheetAdapter;
import matrix.decoration.Styled;
import matrix.structures.CDT.CDT;
import matrix.structures.CDT.probe.BinSearchTree;
import matrix.structures.FDT.FDT;
import matrix.structures.FDT.Tree;
import matrix.structures.FDT.probe.BinTree;
import matrix.structures.FDT.probe.Key;
import matrix.structures.FDT.probe.Table;
import matrix.structures.memory.VirtualInteger;
import matrix.structures.other.Rotations;
import matrix.structures.simulationextensions.CallableMethods;
import matrix.structures.util.MatrixComparable;
import matrix.util.Note;

/**
 * Labeled AVL-tree with node's height in node's label. Automatic balancing can
 * be turned off to make it easier to simulate the rotations.
 * 
 * @author Ville Karavirta
 */
public class MatrixProAVLTree extends BinSearchTree implements CallableMethods {

    protected class MatrixProAVLNode extends BinTree implements Styled,
            CallableMethods {

        protected VirtualInteger height = new VirtualInteger(0, this,
                "height of the tree");

        protected StyleSheet myStyle = null;

        public MatrixProAVLNode() {
            super();
        }

        public MatrixProAVLNode(MatrixComparable e) {
            super(e);
        }

        public StyleSheet getStyleSheet() {
            if (myStyle != null)
                return myStyle;
            else {
                myStyle = new StyleSheetAdapter() {
                    public boolean isInfoEnabled() {
                        return false;
                    }

                    public boolean isLabelEnabled() {
                        return !empty();
                    }

                    public String getLabel() {
                        return "" + getHeight();
                    }
                };
                return myStyle;
            }
        }

        public int getHeight() {
            return height.eval();
        }

        public void setHeight(int value) {
            height.assign(value);
        }

        public boolean empty() {
            return getElement() == null;
        }

        public String[] getCallableMethodNames() {
            return new String[] { "rotateLeft", "rotateRight" };
        }

        public void rotateRight() {
            Rotations.rotateWithLeft_(this);
            evalHeight(this);
        }

        public void rotateLeft() {
            Rotations.rotateWithRight_(this);
            evalHeight(this);
        }

    }

    public MatrixProAVLTree() {
        super();
    }

    public MatrixProAVLTree(MatrixComparable e) {
        super(e);
    }

    public FDT getNewStructure() {
        return new MatrixProAVLNode();
    }

    public Tree getNewNode(Object key) {
        return new MatrixProAVLNode((MatrixComparable) key);
    }

    // Methods for interface CDT

    /**
     * Inserts a new object to this CDT.
     * 
     * @param o
     *            the Object to be inserted
     * @return the CDT wrapper of the data structure after insertion
     */
    public CDT insert(Object o) {
        Note.out(this, "INSERT" + o);

        if (o == null) {
            return this;
        } else if (o instanceof MatrixComparable) {
            if (empty()) {
                setElement((MatrixComparable) o);
            } else
                setElement(AVLInsert((MatrixProAVLNode) getElement(),
                        (MatrixComparable) o));
        } else if (o instanceof Table) {
            Table t = (Table) o;
            for (int i = 0; i < t.size(); i++)
                insert(t.getObject(i));
        } else {
            return insert(new Key(o));
        }
        return this;
    }

    /**
     * Deletes the the parameter object from this CDT.
     * 
     * @param o
     *            the Object to be deleted
     * @return the CDT wrapper of the data structure after deletion
     */
    public CDT delete(Object o) {
        Note.show(this, "Cannot delete, Delete is not implemented");
        return this;
    }

    /**
     * Inserts the given element to the <code>MatrixProAVLNode</code>.
     */
    protected MatrixProAVLNode AVLInsert(MatrixProAVLNode t, MatrixComparable e) {
        if (t == null || t.getElement() == null) {
            t = new MatrixProAVLNode();
            t.setHeight(0);
            t.setElement(e);
        } else if (((MatrixComparable) t.getElement()).gt(e)) {
            t.setLeft(AVLInsert((MatrixProAVLNode) t.getLeft(), e));
            if (!balance()) { // if not autobalancing, don't do it
                evalHeight((MatrixProAVLNode) t.getRight());
                evalHeight((MatrixProAVLNode) t.getLeft());
                evalHeight(t);
                return t;
            }
            if (getHeight((MatrixProAVLNode) t.getLeft())
                    - getHeight((MatrixProAVLNode) t.getRight()) == 2) {
                if (((MatrixComparable) t.getLeft().getElement()).gt(e))
                    t = (MatrixProAVLNode) Rotations.withLeftChild(t);
                else
                    t = (MatrixProAVLNode) Rotations.doubleWithLeftChild(t);
            }

        } else {
            t.setRight(AVLInsert((MatrixProAVLNode) t.getRight(), e));
            if (!balance()) { // if not autobalancing, don't do it
                evalHeight((MatrixProAVLNode) t.getRight());
                evalHeight((MatrixProAVLNode) t.getLeft());
                evalHeight(t);
                return t;
            }
            if (getHeight((MatrixProAVLNode) t.getRight())
                    - getHeight((MatrixProAVLNode) t.getLeft()) == 2) {
                if (!((MatrixComparable) t.getRight().getElement()).gt(e))
                    t = (MatrixProAVLNode) Rotations.withRightChild(t);
                else
                    t = (MatrixProAVLNode) Rotations.doubleWithRightChild(t);
            }
        }

        evalHeight((MatrixProAVLNode) t.getRight());
        evalHeight((MatrixProAVLNode) t.getLeft());

        evalHeight(t);
        return t;
    }

    protected static int evalHeight(MatrixProAVLNode t) {
        if (t == null)
            return -1;
        if (t.getLeft() == null && t.getRight() == null) {
            t.setHeight(0);
            return 0;
        }
        // int h = Math.max( evalHeight((MatrixProAVLNode)t.getLeft()),
        // evalHeight((MatrixProAVLNode)t.getRight()) ) + 1;
        int h = Math.max(getHeight((MatrixProAVLNode) t.getLeft()),
                getHeight((MatrixProAVLNode) t.getRight())) + 1;
        t.setHeight(h);
        return h;

    }

    public static int getHeight(MatrixProAVLNode t) {
        return (t == null) ? -1 : t.getHeight();
    }

    /**
     * Returns an array of strings describing the names of methods that could be
     * invoked during the simulation. The methods should not take any argument.
     */
    public String[] getCallableMethodNames() {
        return new String[] { "toggleBalance" };
    }

    private boolean balance = true;

    /**
     * Returns the current value of the automatic balancing.
     */
    public boolean balance() {
        return balance;
    }

    /**
     * Method used to toggle on or off the automatic balancing of this AVL tree.
     * To get the current value of the balancing, use the balance() method.
     * 
     * @see #balance()
     */
    public void toggleBalance() {
        balance = !balance;
    }
}
