// MatrixPro - algorithm simulation system
// Copyright (C) 2000-2007 See the AUTHORS file and the author tag
// below for more information on the authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package matrixpro.toolbar;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Arrays;

import matrix.util.Note;

import matrixpro.ui.MainFrame;
import matrixpro.util.PropertiesHandler;
import matrixpro.util.Property;

/**
 * A class that can be used to get an instance of <code>ToolbarPanel</code>
 * that has the components set in the preferences.
 * 
 * @version $Id: ToolbarFactory.java,v 1.9 2007/09/10 08:32:32 vkaravir Exp $
 */
public class ToolbarFactory {

    /**
     * Returns a <code>ToolbarComponent</code> of the given class. The
     * returned component should be added to the toolbar given as a parameter,
     * otherwise it might not function properly.
     * 
     * @param componentClass
     *            Class name of the new component.
     * @param tb
     *            ToolbarPanel that the component will be added to.
     * @param name
     *            Name of the component.
     * @return A new <code>ToolbarComponent</code> of the given class.
     */
    private static ToolbarComponent getComponent(String componentClass,
            ToolbarPanel tb, String name) throws NoSuchMethodException,
            IllegalAccessException, ClassNotFoundException,
            InvocationTargetException {
        Class co = Class.forName(componentClass);
        Class toolbar = Class.forName("matrixpro.toolbar.Toolbar");
        Method m = co.getDeclaredMethod("getNewInstance",
                new Class[] { toolbar });
        return (ToolbarComponent) m.invoke(null, new Object[] { tb });
    }

    /**
     * Returns a <code>ToolbarPanel</code> that contains the Tools given as
     * parameter. The returned ToolbarPanel should be added to the
     * <code>MatrixFrame</code> given as parameter. Otherwise it might not
     * function properly.
     * 
     * @param tools
     *            The Tools to be added to the toolbar.
     * @param mf
     *            The MatrixFrame the ToolbarPanel will be added to.
     * @return <code>ToolbarPanel</code> that contains the given Tools.
     */
    public static ToolbarPanel getToolbar(Tool[] tools, MainFrame mf) {
        ToolbarPanel tb = new ToolbarPanel(mf);
        int errors = 0;
        for (int i = 0; i < tools.length; i++) {
            try {
                ToolbarComponent comp = getComponent(tools[i].getClassName(),
                        tb, tools[i].getName());
                tb.addComponent(comp, tools[i].getName(), tools[i]
                        .getPosition()
                        - errors);
            } catch (ClassNotFoundException e) {
                Note.err(new ToolbarFactory(), "Class "
                        + tools[i].getClassName() + " not found!");
                errors++;
            } catch (NoSuchMethodException e) {
                Note.err(new ToolbarFactory(), "Object of class "
                        + tools[i].getClassName() + " coudn't be instantiated");
                errors++;
            } catch (IllegalAccessException e) {
                Note.err(new ToolbarFactory(), "Object of class "
                        + tools[i].getClassName() + " coudn't be instantiated");
                errors++;
            } catch (InvocationTargetException e) {
                Note.err(new ToolbarFactory(), "Object of class "
                        + tools[i].getClassName() + " coudn't be instantiated");
                errors++;
            } catch (IllegalArgumentException e) {
                errors++;
                i--;
            }
        }
        return tb;
    }

    /**
     * Returns a <code>ToolbarPanel</code> that contains the components set in
     * the preferences. The returned ToolbarPanel must be added to the given
     * <code>MatrixFrame</code>.
     */
    public static ToolbarPanel getToolbarFromProperties(MainFrame mf) {
        Property[] components = PropertiesHandler
                .getPropertiesStartingWith("toolbar-component-");
        String[] toolClasses = new String[components.length];
        String[] toolNames = new String[components.length];
        int[] toolPositions = new int[components.length];
        boolean[] show = new boolean[components.length];
        int count = 0;
        for (int i = 0; i < components.length; i++) {
            String name = components[i].getName();
            String value = components[i].getValue();
            toolNames[i] = value.substring(value.indexOf(";") + 1, value
                    .lastIndexOf(";"));
            int pos = (name.indexOf(";") == -1) ? name.length() : name
                    .indexOf(";");
            toolClasses[i] = name.substring(name.indexOf("toolbar-component-")
                    + "toolbar-component-".length(), pos);
            toolPositions[i] = Integer.parseInt(value.substring(value
                    .lastIndexOf(";") + 1));
            if (value.substring(0, value.indexOf(";")).equals("show")) {
                show[i] = true;
                count++;
            }
        }
        Tool[] tools = new Tool[count];
        count = 0;
        for (int i = 0; i < components.length; i++) {
            if (show[i]) {
                tools[count] = new Tool(toolClasses[i], toolNames[i],
                        toolPositions[i]);
                count++;
            }
        }
        Arrays.sort(tools, Tool.getComparator());
        ToolbarPanel tb = getToolbar(tools, mf);
        return tb;
    }
}
