// MatrixPro - algorithm simulation system
// Copyright (C) 2000-2007 See the AUTHORS file and the author tag
// below for more information on the authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package matrixpro.ui;

import java.awt.Component;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Enumeration;

import javax.swing.BorderFactory;
import javax.swing.DefaultListCellRenderer;
import javax.swing.DefaultListModel;
import javax.swing.JButton;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.ListSelectionModel;
import javax.swing.ToolTipManager;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import matrix.util.Note;

import matrixpro.toolbar.Tool;
import matrixpro.toolbar.Toolbar;
import matrixpro.toolbar.ToolbarComponent;
import matrixpro.util.PropertiesHandler;
import matrixpro.util.Property;

/**
 * A panel used to select the visible components in the toolbar as well as the
 * order of the components.
 * 
 * @version $Id: ToolbarTab.java,v 1.10 2007/09/10 08:35:54 vkaravir Exp $
 */
public class ToolbarTab extends JPanel implements ActionListener,
        ListSelectionListener {
    private static final long serialVersionUID = 3946203725562063302L;

    /**
     * JLists containing the names of the hidden and visible toolbar components.
     */
    private JList hidden, visible;

    /** Listsmodels of the hidden and visible lists. */
    private DefaultListModel h, v;

    /** Array of the tools configured in the properties file. */
    private Tool[] tools;

    /** Buttons. */
    private JButton show, hide, up, down, add;

    /**
     * Boolean representing whether or not the properties have been modified.
     */
    private boolean valChanged = false;

    /** Dialog that this tab is added to. */
    private MainFrame frame;

    /** Constructs a new <code>ToolbarTab</code>. */
    public ToolbarTab(MainFrame f) {
        super();
        frame = f;
        loadData();
        setLayout(new GridBagLayout());
        addComponents();
        addListeners();
    }

    /** Initializes the components. */
    private void initializeComponents(Object[] hide, Object[] show) {
        h = new DefaultListModel();
        for (int i = 0; i < hide.length; i++)
            h.addElement(hide[i]);
        v = new DefaultListModel();
        for (int i = 0; i < show.length; i++)
            v.addElement(show[i]);
        hidden = new JList(h);
        ToolTipManager.sharedInstance().registerComponent(hidden);
        hidden.setVisibleRowCount(4);
        hidden.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        hidden.setCellRenderer(new MPListCellRenderer(frame.getToolbar()));
        visible = new JList(v);
        ToolTipManager.sharedInstance().registerComponent(visible);
        visible.setVisibleRowCount(5);
        visible.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        visible.setCellRenderer(new MPListCellRenderer(frame.getToolbar()));
        this.show = new JButton("Show");
        this.show.setEnabled(false);
        this.show.setToolTipText("Set the selected component visible.");
        this.hide = new JButton("Hide");
        this.hide.setEnabled(false);
        this.hide.setToolTipText("Set the selected component hidden.");
        up = new JButton("Move up");
        up.setEnabled(false);
        up.setToolTipText("Move the selected component up in the toolbar.");
        down = new JButton("Move down");
        down.setEnabled(false);
        down.setToolTipText("Move the selected component down in the toolbar.");
        add = new JButton("Add");
        add.setToolTipText("Add new component to the toolbar.");
    }

    /** Load the properties. */
    private void loadData() {
        Property[] components = PropertiesHandler
                .getPropertiesStartingWith("toolbar-component-");
        ArrayList show = new ArrayList();
        ArrayList hide = new ArrayList();
        for (int i = 0; i < components.length; i++) {
            String name = components[i].getName();
            String value = components[i].getValue();
            String toolClass = name.substring(name
                    .indexOf("toolbar-component-")
                    + "toolbar-component-".length());
            String toolName = value.substring(value.indexOf(";") + 1, value
                    .lastIndexOf(";"));
            int toolPosition = Integer.parseInt(value.substring(value
                    .lastIndexOf(";") + 1));
            if (value.substring(0, value.indexOf(";")).equals("show")) {
                show.add(new Tool(toolClass, toolName, toolPosition));
            } else {
                hide.add(new Tool(toolClass, toolName, toolPosition));
            }
        }
        Object[] obj = show.toArray();
        Arrays.sort(obj, Tool.getComparator());
        tools = new Tool[components.length];
        for (int i = 0; i < obj.length; i++)
            tools[i] = (Tool) obj[i];
        for (int i = 0; i < hide.size(); i++)
            tools[i + obj.length] = (Tool) hide.get(i);
        initializeComponents(hide.toArray(), obj);
    }

    /** Add the components to this panel. */
    private void addComponents() {
        JScrollPane hi = new JScrollPane(hidden);
        hi.setBorder(BorderFactory.createTitledBorder(BorderFactory
                .createEmptyBorder(), "Hidden components"));
        GridBagConstraints gbc = new GridBagConstraints();
        gbc.gridx = gbc.gridy = 0;
        gbc.gridwidth = 6;
        gbc.fill = GridBagConstraints.BOTH;
        gbc.anchor = GridBagConstraints.CENTER;
        add(hi, gbc);
        gbc.fill = GridBagConstraints.NONE;
        gbc.gridy = 1;
        gbc.gridwidth = 3;
        add(show, gbc);
        gbc.gridx = 3;
        add(hide, gbc);
        JScrollPane vi = new JScrollPane(visible);
        vi.setBorder(BorderFactory.createTitledBorder(BorderFactory
                .createEmptyBorder(), "Visible components"));
        gbc.gridx = 0;
        gbc.gridy = 2;
        gbc.gridwidth = 6;
        add(vi, gbc);
        gbc.gridy = 3;
        gbc.gridwidth = 2;
        add(up, gbc);
        gbc.gridx = 2;
        add(down, gbc);
        gbc.gridx = 4;
        add(add, gbc);
    }

    /** Adds listeners to the components. */
    private void addListeners() {
        show.setActionCommand("show");
        show.addActionListener(this);
        hide.setActionCommand("hide");
        hide.addActionListener(this);
        up.setActionCommand("up");
        up.addActionListener(this);
        down.setActionCommand("down");
        down.addActionListener(this);
        hidden.addListSelectionListener(this);
        visible.addListSelectionListener(this);
        add.setActionCommand("add");
        add.addActionListener(this);
    }

    public void actionPerformed(ActionEvent e) {
        valChanged = true;
        String command = e.getActionCommand();
        if (command.equals("show")) {
            int index = hidden.getSelectedIndex();
            Object o = h.remove(index);
            v.addElement(o);
        } else if (command.equals("hide")) {
            int index = visible.getSelectedIndex();
            Object o = v.remove(index);
            h.addElement(o);
        } else if (command.equals("up")) {
            int index = visible.getSelectedIndex();
            Object o = v.remove(index);
            v.add(index - 1, o);
            visible.setSelectedIndex(index - 1);
        } else if (command.equals("down")) {
            int index = visible.getSelectedIndex();
            Object o = v.remove(index);
            v.add(index + 1, o);
            visible.setSelectedIndex(index + 1);
        } else if (command.equals("add")) {
            addComponentClass();
        }
    }

    /**
     * Called by AddComponentDialog to add a new item to the componentlist.
     */
    private void addComponentClass() {
        AddComponentDialog addClass = new AddComponentDialog(frame, true);
        if (addClass.showDialog()) {
            String className = addClass.getClassName();
            try {
                Class.forName(className);
            } catch (ClassNotFoundException ex) {
                Note.show(this, "Class " + className + " not found!");
                return;
            }
            Tool t = new Tool(className, addClass.getComponentName(), v
                    .getSize());
            v.addElement(t);

        }
    }

    /**
     * Helper attribute used in valueChanged method. Keeps track when a JLists
     * selection is being cleared and the events should be ignored.
     */
    private boolean clearing = false;

    public void valueChanged(ListSelectionEvent e) {
        if (e.getValueIsAdjusting() == false && !clearing) {
            Object source = e.getSource();
            if (source == hidden) {
                clearing = true;
                visible.clearSelection();
                hide.setEnabled(false);
                clearing = false;
                boolean enable = hidden.getSelectedIndex() != -1;
                show.setEnabled(enable);
                up.setEnabled(false);
                down.setEnabled(false);
            } else if (source == visible) {
                boolean enable = (visible.getSelectedIndex() != -1);
                hide.setEnabled(enable);
                up.setEnabled(enable && visible.getSelectedIndex() != 0);
                down.setEnabled(enable
                        && visible.getSelectedIndex() != v.size() - 1);
                clearing = true;
                hidden.clearSelection();
                clearing = false;
                show.setEnabled(false);
            }
        }
    }

    /**
     * Sets the properties in <code>PropertiesHandler</code> to the currently
     * selected values.
     */
    public void setProperties() {
        Enumeration elem = v.elements();
        int pos = 0;
        while (elem.hasMoreElements()) {
            Tool t = (Tool) elem.nextElement();
            PropertiesHandler.setProperty("toolbar-component-"
                    + t.getClassName(), "show" + ";" + t.getName() + ";" + pos);
            pos++;
        }
        elem = h.elements();
        while (elem.hasMoreElements()) {
            Tool t = (Tool) elem.nextElement();
            PropertiesHandler.setProperty("toolbar-component-"
                    + t.getClassName(), "hide" + ";" + t.getName() + ";-1");
        }
    }

    /**
     * Returns true if the selections have changed since this instance was
     * created.
     */
    public boolean isChanged() {
        return valChanged;
    }

    /**
     * Custom ListCellRenderer that does what the default renderer plus adds the
     * description of the ToolbarComponent as the tooltip text.
     */
    private class MPListCellRenderer extends DefaultListCellRenderer {

        private static final long serialVersionUID = 7389465081130764015L;

        public MPListCellRenderer(Toolbar tb) {
            super();
        }

        public Component getListCellRendererComponent(JList list, Object value,
                int index, boolean isSelected, boolean cellHasFocus) {
            super.getListCellRendererComponent(list, value, index, isSelected,
                    cellHasFocus);
            try {
                Tool tool = (Tool) value;
                Class co = Class.forName(tool.getClassName());
                Class toolbar = Class.forName("matrixpro.toolbar.Toolbar");
                Toolbar tb = frame.getToolbar();
                Method m = co.getDeclaredMethod("getNewInstance",
                        new Class[] { toolbar });
                ToolbarComponent tbc = (ToolbarComponent) m.invoke(null,
                        new Object[] { tb });
                String text = tool.getClassName();
                setText(text.substring(text.lastIndexOf(".") + 1));
                setToolTipText(tbc.getDescription());
            } catch (Exception e) {
                e.printStackTrace();
            }
            return this;
        }
    }
}
