// MatrixPro - algorithm simulation system
// Copyright (C) 2000-2007 See the AUTHORS file and the author tag
// below for more information on the authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package matrixpro.util;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.JarURLConnection;
import java.net.URL;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Properties;
import java.util.jar.JarFile;

import matrix.util.Note;

/**
 * Class that handles the loading and saving of the properties used in the
 * MatrixPro.
 * 
 * @version $Id: PropertiesHandler.java,v 1.15 2007/09/10 08:35:54 vkaravir Exp $
 */
// XXX Could be done smarter by extending the java's properties.
public class PropertiesHandler {

    private static String confFile;

    private static String url, codebaseURL;

    private static Properties properties = null;

    /** Name of the default properties file. */
    public static final String DEFAULT_FILE = "matrixpro.properties";

    /** Name of the default jar file. */
    public static final String DEFAULT_JAR = "matrixpro-full.jar";

    private static boolean changed = false;

    /** Sets the codebase that should be used when loading the properties. */
    public static void setCodebaseURL(String s) {
        codebaseURL = s;
    }

    /** Sets the path of the file properties should be loaded from. */
    public static void setConfPath(String s) {
        confFile = s;
    }

    /** Sets the url that the properties should be loaded from. */
    public static void setUrl(String s) {
        url = s;
    }

    /**
     * Returns the given property.
     * 
     * @param property
     *            The name of the property.
     * @return The property as an <code>Property</code> object.
     * @see Property
     */
    public static Property getProperty(String property) {
        if (properties == null)
            properties = loadProperties();
        return new Property(property, properties.getProperty(property));
    }

    /**
     * Returns the given property. If the property is not found a property with
     * the given default value is returned.
     * 
     * @param property
     *            The name of the property.
     * @param defaultVal
     *            Default value of the property.
     * @return The property or a property with defaultVal if it is not found.
     * @see Property
     */
    public static Property getProperty(String property, String defaultVal) {
        if (properties == null)
            properties = loadProperties();
        String val = properties.getProperty(property, defaultVal);
        return new Property(property, val.equals("") ? defaultVal : val);
    }

    /**
     * Returns the value of the given property. If the property is not found the
     * default value is returned.
     * 
     * @param property
     *            The name of the property.
     * @param defaultVal
     *            Default value of the property.
     * @return The value of the property or defaultVal if the property is not
     *         found.
     */
    public static String getPropertyValue(String property, String defaultVal) {
        return getProperty(property, defaultVal).getValue();
    }

    /**
     * Returns the value of the given property.
     * 
     * @param property
     *            The name of the property.
     * @return The value of the property.
     */
    public static String getPropertyValue(String property) {
        return getProperty(property).getValue();
    }

    /**
     * Sets the property to the given value.
     * 
     * @param property
     *            The name of the property.
     * @param value
     *            The new value of the property.
     */
    public static void setProperty(String property, String value) {
        changed = true;
        properties.setProperty(property, value);
    }

    /**
     * Sets the given property.
     * 
     * @param p
     *            The property which is set.
     * @see Property
     */
    public static void setProperty(Property p) {
        setProperty(p.getName(), p.getValue());
    }

    /**
     * Stores the properties to the given OutputStream.
     * 
     * @param out
     *            The <code>OutputStream</code> to write the properties to.
     */
    public static void store(OutputStream out) {
        store(out, "");
    }

    /**
     * Stores the properties to the default file.
     */
    public static void store() {
        try {
            store(new FileOutputStream(DEFAULT_FILE));
        } catch (Exception e) {
            Note.err(new PropertiesHandler(), "Properties not saved - "
                    + e.getMessage());
            // e.printStackTrace();
            Note.show(new PropertiesHandler(), "Preferences coudn't be saved!");
        }
    }

    /**
     * Stores the properties to the given OutputStream and adding the given
     * header.
     * 
     * @param out
     *            The <code>OutputStream</code> to write the properties to.
     * @param header
     *            A description of the property list.
     */
    public static void store(OutputStream out, String header) {
        try {
            if (properties != null && changed)
                properties.store(out, header);
        } catch (SecurityException e) {
            // applet
        } catch (Exception e) {
            Note.err(new PropertiesHandler(), "Properties not saved - "
                    + e.getMessage());
            // e.printStackTrace();
            Note.show(new PropertiesHandler(), "Preferences coudn't be saved!");
        }
    }

    /**
     * Returns the properties that start with the given name.
     * 
     * @param start
     *            The <code>String</code> that the properties start with.
     * @return A <code>Property</code> array containing the properties
     *         starting with the given text.
     */
    public static Property[] getPropertiesStartingWith(String start) {
        if (properties == null)
            properties = loadProperties();
        Enumeration en = properties.propertyNames();
        ArrayList a = new ArrayList();
        ;
        while (en.hasMoreElements()) {
            String s = (String) en.nextElement();
            if (s.startsWith(start))
                a.add(new Property(s, properties.getProperty(s)));
        }
        Property[] p = new Property[a.size()];
        Object[] o = a.toArray();
        // System.arrayCopy(o, 0, p, 0, o.length);
        for (int i = 0; i < o.length; i++)
            p[i] = (Property) o[i];
        return p;
    }

    /** Loads the properties. */
    private static Properties loadProperties() {
        Properties p = new Properties();
        if (confFile != null) { // confFile is set
            try {
                FileInputStream in = new FileInputStream(confFile);
                p.load(in);
                in.close();
                return p;
            } catch (Exception ex) {
                // ex.printStackTrace();
                // XXX
            }
        } else if (url != null) { // read from url
            try {
                URL inUrl = new URL(url);
                InputStream in = inUrl.openStream();
                p.load(in);
                in.close();
                return p;
            } catch (Exception ex) {
                // ex.printStackTrace();
                // XXX
            }
        }
        try {
            FileInputStream in = new FileInputStream(DEFAULT_FILE);
            p.load(in);
            in.close();
            return p;
        } catch (Exception ex) {
            // ex.printStackTrace();
            // XXX
        }
        try {
            InputStream in = PropertiesHandler.class.getResourceAsStream("/"
                    + DEFAULT_FILE);
            p.load(in);
            in.close();
            return p;
        } catch (Exception e) {
            // e.printStackTrace();
            // XXX
        }
        try {
            if (codebaseURL != null) { // jar
                String jarURL = "jar:" + codebaseURL + DEFAULT_JAR + "!/";
                URL u = new URL(jarURL);
                JarURLConnection conn = (JarURLConnection) u.openConnection();
                JarFile jar = conn.getJarFile();
                InputStream in = jar.getInputStream(jar.getEntry(DEFAULT_FILE));
                p.load(in);
                in.close();
                return p;
            } else {
                JarFile jar = new JarFile(DEFAULT_JAR);
                InputStream in = jar.getInputStream(jar.getEntry(DEFAULT_FILE));
                p.load(in);
                in.close();
                return p;
            }
        } catch (Exception ex) {
            // ex.printStackTrace();
            // XXX
        }
        try {

            JarFile jar = new JarFile(System.getProperties().getProperty(
                    "java.class.path"));
            InputStream in = jar.getInputStream(jar.getEntry(DEFAULT_FILE));
            p.load(in);
            in.close();
            return p;
        } catch (Exception ex) {

        }
        try {
            return loadDefault(p);
        } catch (Exception e) {
        }
        return p;
    }

    private static Properties loadDefault(Properties p) throws IOException {
        p
                .setProperty(
                        "toolbar-component-matrixpro.toolbar.components.SetEndComponent",
                        "show;setend;3");
        p
                .setProperty(
                        "toolbar-component-matrixpro.toolbar.components.DisjoinStepsComponent",
                        "show;disjoinsteps;6");
        p
                .setProperty(
                        "toolbar-component-matrixpro.toolbar.components.RenameComponent",
                        "show;rename;11");
        p
                .setProperty(
                        "toolbar-component-matrixpro.toolbar.components.AnimatorDumpComponent",
                        "hide;animatordump;-1");
        p
                .setProperty(
                        "toolbar-component-matrixpro.toolbar.components.RepresentationComponent",
                        "show;representation;9");
        p.setProperty(
                "toolbar-component-matrixpro.toolbar.components.DumpComponent",
                "hide;dump;-1");
        p
                .setProperty(
                        "toolbar-component-matrixpro.toolbar.components.ContextualPanel",
                        "show;contextual;16");
        p
                .setProperty(
                        "toolbar-component-matrixpro.toolbar.components.RemoveBreakComponent",
                        "show;removebreak;5");
        p
                .setProperty(
                        "toolbar-component-matrixpro.toolbar.components.MicroStepForwardComponent",
                        "hide;microforward;-1");
        p
                .setProperty(
                        "toolbar-component-matrixpro.toolbar.components.DeleteComponent",
                        "show;delete;12");
        p
                .setProperty(
                        "toolbar-component-matrixpro.toolbar.components.MicroStepBackwardComponent",
                        "hide;microback;-1");
        p
                .setProperty(
                        "toolbar-component-matrixpro.toolbar.components.SpeedAdjustComponent",
                        "show;speedadjust;1");
        p.setProperty(
                "toolbar-component-matrixpro.toolbar.components.FileComponent",
                "show;file;8");
        p
                .setProperty(
                        "toolbar-component-matrixpro.toolbar.components.ChangeEdgeLengthComponent",
                        "show;edgelength;17");
        p
                .setProperty(
                        "toolbar-component-matrixpro.toolbar.components.LabelNodesComponent",
                        "show;labelnodes;10");
        p
                .setProperty(
                        "toolbar-component-matrixpro.toolbar.components.NewWindowComponent",
                        "show;newwindow;14");
        p.setProperty("matrixpro.file.url",
                "http\\://www.cs.hut.fi/Research/MatrixPro/");
        p
                .setProperty(
                        "toolbar-component-matrixpro.toolbar.components.StepLabelComponent",
                        "show;step;0");
        p
                .setProperty(
                        "toolbar-component-matrixpro.toolbar.components.SetBeginComponent",
                        "show;setbegin;2");
        p
                .setProperty(
                        "toolbar-component-matrixpro.toolbar.components.DebugComponent",
                        "hide;debug;-1");
        p
                .setProperty(
                        "toolbar-component-matrixpro.toolbar.components.InsertBreakComponent",
                        "show;insertbreak;4");
        p.setProperty("shortcutkey.9", "Table of Keys");
        p.setProperty("shortcutkey.8", "Stack");
        p
                .setProperty(
                        "toolbar-component-matrixpro.toolbar.components.JoinStepsComponent",
                        "show;joinsteps;7");
        p.setProperty("shortcutkey.7", "AVL Tree");
        p.setProperty("shortcutkey.6", "Binary Heap");
        p.setProperty("shortcutkey.5", "Red-Black Tree");
        p.setProperty("shortcutkey.4", "Binary Search Tree");
        p.setProperty("shortcutkey.3", "Undirected Graph");
        p.setProperty(
                "toolbar-component-matrixpro.toolbar.components.SaveComponent",
                "hide;save;-1");
        p.setProperty("shortcutkey.2", "Directed Graph");
        p.setProperty("shortcutkey.1", "Linked List");
        p.setProperty("shortcutkey.0", "Table of Random Keys");
        p
                .setProperty(
                        "toolbar-component-matrixpro.toolbar.components.NewRepresentationComponent",
                        "show;newrepresentation;13");
        p
                .setProperty(
                        "toolbar-component-matrixpro.toolbar.components.InsertEdgeComponent",
                        "show;insertedge;15");
        return p;
    }
}
