// MatrixPro - algorithm simulation system
// Copyright (C) 2000-2007 See the AUTHORS file and the author tag
// below for more information on the authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package matrixpro.util.export;

import java.awt.GridLayout;
import java.awt.Rectangle;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.zip.GZIPOutputStream;

import javax.swing.BorderFactory;
import javax.swing.JCheckBox;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.filechooser.FileFilter;

import matrix.animation.Animator;
import matrix.util.ApplicationAdapter;
import matrix.util.Note;
import matrix.util.export.MatrixGraphics;
import matrix.util.export.SVGWriter;

import matrixpro.ui.MainFrame;
import matrixpro.util.FileHelper;

/**
 * SVG animation export format for MatrixPro. This format is capable of
 * exporting the animation sequence of MatrixPro as an SVG animations with a
 * animation control panel.
 * 
 * @version $Id: SVGExportFormat.java,v 1.6 2007/09/10 08:35:56 vkaravir Exp $
 */
public class SVGExportFormat extends MPExportFormat {
    /** The <code>JPanel</code> used when svg is selected. */
    private JPanel svg;

    /** <code>JCheckBox</code> for the svg compression selection. */
    private JCheckBox compress;

    /** <code>JCheckBox</code> for the svg animator selection. */
    private JCheckBox animator;

    /** <code>JTextField</code> for the svg scale selection. */
    private JTextField scale;

    /** <code>JTextField</code> for the svg pause length selection. */
    private JTextField pause;

    /** <code>JTextField</code> for the svg step length selection. */
    private JTextField step;

    public SVGExportFormat() {
        super();
    }

    public FileFilter getFileFilter() {
        return new SVGFileFilter();
    }

    public String getFileExtension() {
        if (compress.isSelected())
            return "svgz";
        else
            return "svg";
    }

    public boolean export(MainFrame mf, File saveFile) {
        JPanel p = mf.getStructurePanel();
        MatrixGraphics matrix = new MatrixGraphics(new Rectangle(0, 0, 10000,
                10000), p.getGraphics());
        matrix.addDisable("VisualHotSpot");
        boolean compress = this.compress.isSelected();
        boolean control = animator.isSelected();
        double scale = 1.3;
        float stepLength = SVGWriter.DEFAULTSTEPLENGTH;
        float pause = SVGWriter.DEFAULTPAUSE;
        try {
            scale = Double.parseDouble(this.scale.getText());
        } catch (NumberFormatException e) {
            Note.show(new FileHelper(),
                    "Invalid scale. Using default scale 1.3.");
        }
        try {
            stepLength = Float.parseFloat(step.getText());
        } catch (NumberFormatException e) {
            Note.show(new FileHelper(), "Invalid step length. Using default "
                    + stepLength + ".");
        }
        try {
            pause = Float.parseFloat(this.pause.getText());
        } catch (NumberFormatException e) {
            Note.show(new FileHelper(), "Invalid pause length. Using default "
                    + pause + ".");
        }
        Animator animator = ApplicationAdapter.getApplication().getAnimator();
        animator.rewind();
        p.validate();
        mf.validate();
        p.paint(matrix);
        int steps = 0;
        matrix.nextStep();
        while (animator.hasNextOperation()) {
            animator.redo();
            p.validate();
            mf.validate();
            p.paint(matrix);
            matrix.nextStep();
            steps++;
        }
        SVGWriter toSVG = new SVGWriter(matrix.getContents());
        toSVG.setMaxStep(steps);
        toSVG.setBackground(p.getBackground());
        matrix.setScale(scale);
        if (!compress) // svg
            toSVG.write(stepLength, pause, saveFile, control);
        else { // gzipped svg
            try {
                GZIPOutputStream zos = new GZIPOutputStream(
                        new FileOutputStream(saveFile));
                toSVG.write(stepLength, pause, zos, control);
                zos.flush();
                zos.close();
            } catch (IOException e) {
                Note.err(e, "Error: " + e.toString());
            }
        }
        return true;
    }

    public JComponent getAccessory() {
        if (svg == null)
            initSvgPanel();
        return svg;
    }

    /**
     * Method that adds components to the panel used when svg is selected.
     */
    private void initSvgPanel() {
        svg = new JPanel();
        svg.setLayout(new GridLayout(6, 1));
        svg.setBorder(BorderFactory.createTitledBorder(BorderFactory
                .createBevelBorder(2), "SVG"));
        compress = new JCheckBox("Compress (gzip)");
        compress.setToolTipText("The svg file can be compressed with gzip.");
        svg.add(compress);
        animator = new JCheckBox("Animator panel", true);
        animator.setToolTipText("Add animation control buttons (play, "
                + "stop, etc)to the animation.");
        animator.addItemListener(new ItemListener() {
            public void itemStateChanged(ItemEvent e) {
                animatorValueChanged();
            }
        });
        svg.add(animator);
        svg.setVisible(true);
        JPanel s = new JPanel();
        s.setLayout(new GridLayout(1, 2));
        JLabel scaleLabel = new JLabel("Scale");
        scaleLabel.setToolTipText("Size of the animation. To create a "
                + "larger animation, set this value larger.");
        s.add(scaleLabel);
        scale = new JTextField(4);
        scale.setText("1.3");
        scale.setToolTipText("Size of the animation. To create a "
                + "larger animation, set this value larger.");
        s.add(scale);
        svg.add(s);
        JPanel paus = new JPanel();
        paus.setLayout(new GridLayout(1, 2));
        JLabel pauseLabel = new JLabel("Pause length");
        pauseLabel.setToolTipText("Length of the pause between steps in "
                + "the animation");
        paus.add(pauseLabel);
        pause = new JTextField(4);
        pause.setText(SVGWriter.DEFAULTPAUSE + "");
        pause.setToolTipText("Length of the pause between steps in "
                + "the animation");
        paus.add(pause);
        pause.setEnabled(false);
        svg.add(paus);
        JPanel ste = new JPanel();
        ste.setLayout(new GridLayout(1, 2));
        JLabel stepLabel = new JLabel("Step length");
        stepLabel.setToolTipText("Time it takes to animate one step in "
                + "the animation.");
        ste.add(stepLabel);
        step = new JTextField(4);
        step.setText(SVGWriter.DEFAULTSTEPLENGTH + "");
        step.setToolTipText("Time it takes to animate one step in "
                + "the animation.");
        ste.add(step);
        step.setEnabled(false);
        svg.add(ste);
    }

    /**
     * Method called when the value of the animator control panel -checkbox is
     * selected/deselected. Sets the pause and step length fields enabled if
     * animator panel is not selected and unenabled if it is.
     */
    private void animatorValueChanged() {
        boolean state = !animator.isSelected();
        step.setEnabled(state);
        pause.setEnabled(state);
    }

}
